/*                                             -*- c-file-style: "Ellemtel" -*-

`directest.c' - Skeleton agent cli that runs under director

This file is part of version 3 of the Sample agent program for director.
Read the `License' file for terms of use and distribution.
Copyright 2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

/*
 * * Example application designed to run under the `director' cli wrapper.
 * This program serves both as an example for programmers wishing to make
 * an `agent' program for director, and also as a diagnostic tool for
 * testing the director itself.
 */

#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include "cli/cli.h" /* including with "" means it will be in the dependencies */

#ifndef UNUSED
#define UNUSED
#endif

static BOOLEAN done = FALSE;	/* flag to set when program should exit */
static Command comlist[];	/* contains an entry for each command */

/* Command Set
 * -----------
 * Each of these must take a single const char* argument (even if they
 * don't use it for anything) and return PASS or FAIL depending on
 * whether the command was successful or not.  If needed, libcli
 * provides further functions for breaking up the single string
 * into argc/argv-style arguments and removing quotes from string
 * constants.  If you find that you are implementing something in your
 * code that would be generally useful to other programs, let me know
 * and we will include it in libcli. -si
 */

PASSFAIL
com_exit( const char* args UNUSED ) { done = TRUE; return PASS; }
/*
 * The exit command just sets a flag to tell the main while loop to
 * exit cleanly on the next iteration
 */

PASSFAIL
com_help( const char* args ) { return cli_help( args ); }
/*
 * The help command returns specific help on a command (if an argument
 * was given) or the full list of commands, using the library call cli_help.
 * Usually it is easier to just hardcode the help screen in right here,
 * but then you have to update it every time you add something.
 * If all that com_help() does is return cli_help, you can simply include
 * cli_help in the Command table instead.
 */

PASSFAIL
com_list( const char* arg ) { return cli_system("ls -lF", arg); }
/*
 * Runs the system command "ls", with a couple of default flags
 */

PASSFAIL
com_count( const char* args )
/*
 * This command is just here to help test libcli's cli_argv() call.
 */
{
   static char** argv = NULL;
   int i, argc;
   
   if (argv) { if (argv[0]) free(argv[0]); free(argv); }

   argv = cli_argv(&argc, args);
   printf("Got %d arguments.\n", argc);
   for (i=0; i<argc; i++)
      printf("Argument #%d = `%s'\n", i+1, argv[i]);

   return PASS;
}

PASSFAIL
com_rename( const char* args )
/*
 * Example of a command that must have 2 arguments.  Anything
 * else causes it to return FAIL;
 */
{
   static char** argv = NULL;
   int argc;

   if (argv) { if (argv[0]) free(argv[0]); free(argv); }

   argv = cli_argv(&argc, args);
   if (argc!=2)
   {
      printf("please specify 2 arguments: 'oldname' 'newname'\n");
      return FAIL;
   }

   return cli_system("mv", args);
}

PASSFAIL
com_comment( const char* arg )
/*
 * Example of a command that takes the whole rest of the command line
 * as a single argument, or prints its current status if no argument
 * is found at all.  Note that in this case you don't want to call
 * cli_argv() at all since it is more convient to keep the string as
 * one piece.  Use cli_arg1() instead, so that quotes will be handled
 * correctly.
 */
{
   static char* comment = NULL;

   if (!arg || !*arg)
   {
      printf("Comment is currently `%s'.  Give an argument to change.\n",
	     comment?comment:"(not set)");
   }
   else
   {
      if (comment) free(comment);
      comment = cli_arg1(arg); /* cli_arg1() mallocs a new copy for us */
      printf("status: Comment now: `%s'\n", comment);
   }
   return PASS;
}

PASSFAIL com_sync(const char* arg UNUSED)
{
   printf("If I had a par file, I'd be re-reading it now...\n");
   return PASS;
}

PASSFAIL com_fork(const char* arg)
{
   switch (fork())
   {
      case -1:
      {
	 perror("fork");
	 return FAIL;
      }
      case 0:
      {
	 int i;

	 if (arg && *arg)
	 {
	    fprintf(stderr, "child is creating own group...\n");
	    setsid();
	 }
	 for (i=5; i; i--)
	 {
	    fprintf(stderr, "Child will exit in %d...\n", i);
	    sleep(1);
	    if (cli_signal_event(SIGINT))
	       fprintf(stderr, "child got SIGINT\n");

	    if (cli_signal_event(SIGTSTP))
	       fprintf(stderr, "child got SIGTSTP\n");

	    if (cli_signal_event(SIGALRM))
	       fprintf(stderr, "child got SIGALRM\n");

	    if (cli_signal_event(SIGQUIT))
	       fprintf(stderr, "child got SIGQUIT\n");
	 }
	 _exit(EXIT_SUCCESS);
      }
      default:
      {
	 printf("forked test-child process\n");
	 return PASS;
      }
   }
}


PASSFAIL com_sleep(const char* arg)
{
   int seconds = 0;

   if (arg && *arg) seconds = atoi(arg);
   if (seconds == 0) seconds = 1;
   while (seconds)
   {
      printf("progress: Sleeping for %d seconds\n", seconds--);
      if (cli_signal_event(SIGINT))
      {
	 printf("error: Sleeping interrupted by SIGINT!\n");
	 return FAIL;
      }
      if (cli_signal_event(SIGTSTP))
      {
	 printf("progress:\n");
	 printf("error: Sleeping interrupted by SIGTSTP!\n");
	 return FAIL;
      }
      if (cli_signal_event(SIGQUIT))
      {
	 printf("warning: sleep command ignoring SIGQUIT!\n");
      }
      if (sleep(1)!=0)
      {
	 printf("progress:\n");
	 printf("error: Sleeping interrupted!\n");
	 return FAIL;
      }
   }
   printf("progress:\n");
   return PASS;
}

PASSFAIL com_hack(const char* arg)
{
#if 0
   printf("\rok> ");
   fflush(stdout);
   sleep(atoi(arg));
#endif
   puts("\rok> @alarm");
   fflush(stdout);
   sleep(atoi(arg));
   return PASS;
}

static Command comlist[] = {
   /*
    * List of all commands this cli understands in a specific mode.
    * The cli_help() functions specifically keys off of any commands
    * descriptions that start with "Synonym" and skips them when listing
    * the full set of commands.
    *
    * NOTE: some work is needed on the formatting of the help strings
    *       so that they clearly indicate what arguments each command
    *       requires and what arguments are optional.
    */
   { "",     com_sync, "Synonym for sync" },
   { "cd",   cli_cd,   "Change working directories" }, /* all agents should have this! */
   { "list", com_list, "List files in DIR" },
   { "ls",   com_list, "Synonym for list" },
   { "ll",   com_list, "Synonym for list" },
   { "rename", com_rename, "Rename a file" },
   { "count", com_count, "Counts args in argv" },
   { "comment", com_comment, "Sets the comment header" },
   { "sync", com_sync, "Re-read par file" },
   { "fork", com_fork, "Fork a test-child process" },
   { "sleep", com_sleep, "Sleep for a certain number of seconds" },
   { "hack", com_hack, "Hack to for testing a director feature" },
   { "exit", com_exit, "Exit the program" }, /* all agents must have this! */
   { "help", com_help, "Displays help for directest" }, /* all agents must have this! */
   { "?",    com_help, "Synonym for help" },
   { 0, 0, 0 } /* This terminates the command list */
};

static void
do_low_priority_stuff(void)
{
   static int count = 0;

   count++;
   CLI_DEBUG(printf("debug: Doing low priority stuff (%d)\n", count));
   printf("info1: I've done low priority stuff %d time%s so far\n", 
	  count, count==1?"":"s");
}

extern void wait_on_child(void);

static void
cleanup(void)
{
   printf("\rdirectest cleaning up on TERM signal.\n");
   exit(EXIT_SUCCESS);
}

int
main(int argc, const char* argv[])
{
   int i;

   if (cli_init("directest", comlist, 2)==FAIL) return EXIT_FAILURE;

   /*
    * Echo back the program's name and any args passed on the command line
    */
   printf("status: ");
   for (i=0; i<argc; i++)
      printf("`%s' ", argv[i]);
   printf("starting up . . .\n");

   /*
    * Indicate which signals should be detected and kept track of
    */
   cli_signal(SIGTERM, cleanup);
   cli_signal(SIGINT, CLI_SIG_TRAP);  /* usually ^C if on a tty */
   cli_signal(SIGTSTP, CLI_SIG_TRAP); /* usually ^Z if on a tty */
   cli_signal(SIGQUIT, CLI_SIG_TRAP); /* usually ^\ if on a tty */
   cli_signal(SIGCHLD, CLI_SIG_TRAP); /* %%% WARNING: don't do this in your code! */

   sleep(2);

   printf("status: ready.\n");

   while (!done)
   {
      static char* line = NULL;

      /*
       * Stuff marked with LPS is only needed if you have low-priority
       * stuff that has to execute every N seconds of idle time.  Otherwise
       * you can ignore all the stuff with SIGALRM!
       */

      cli_signal(SIGALRM, CLI_SIG_TRAP);	/* LPS */
      cli_signal_event(SIGALRM);		/* LPS */
      alarm(10);				/* LPS */

      CLI_DEBUG(printf("debug: Calling cli_getline.\n"));

      if (line!=NULL) free(line);	/* free memory used by last command */
      line = cli_getline();		/* get the next command */

      /*
       * Next check for any unusual situations.  Do this immediately before
       * errno gets clobbered by another call!  If you don't need LPS,
       * and if you only want to exit with the "exit" command, you don't
       * need this whole if-clause at all!
       */
      if (!line)
      {
	 if (errno==0)
	    done = 1;			/* Only if you want exit-on-EOF */
	 else if (cli_signal_event(SIGALRM))	/* LPS */
	    do_low_priority_stuff();		/* LPS */
      }

      alarm(0);					/* LPS */
      cli_signal(SIGALRM, CLI_SIG_DFL);	/* LPS */

      /*
       * Note about detecting signals:  There are many ways to handle signals.  When
       * cli_signal() is used to set up signal trapping,  a system call that was in
       * progress when a signal happens will return an error and set errno to EINTR.
       * Then you can use cli_signal_event(signo) to test AND CLEAR the flag that
       * indicated a particular signal happened.  Always run cli_signal_event before
       * doing whatever it is that gets done when a particular signal is received to
       * avoid the potential problem of a new signal arriving between the time the old
       * one was handled and the flag was cleared (by cli_signal_event.).  cli_getline
       * will return immediately if a signal is received while the program is waiting
       * for user input, but remember that signals can arrive during almost any other
       * part of the program's execution also.  It may be wise to use cli_signal_block()
       * to delay having to deal with a signal (and to avoid errno=EINTR's) during other
       * blocks of code.  The moment you call cli_signal_unblock(), any signals that
       * happened during the blocked time are immediately handled and will show up when
       * you test cli_signal_event.  It all depends on what kind of behavior you need
       * from the code.  In any case you probably almost always want to check for these
       * three signals immediately after cli_getline():
       */
      /*
       * %%% Warning: you probably don't want to mess with SIGCHLD!
       */
      if (cli_signal_event(SIGCHLD)) /* DIRECTEST ONLY!!! */
	 wait_on_child(); /* REMOVE THIS FROM REAL AGENTS! */
      if (cli_signal_event(SIGQUIT)) /* And remove these too probably */
	 printf("\astatus: caught SIGQUIT\n"); /* \a rings the bell */
      if (cli_signal_event(SIGTSTP))
	 printf("warning: caught SIGTSTP\n");
      if (cli_signal_event(SIGINT))
      {
	 fprintf(stderr,"er"); /* just testing parsing here... */
	 fflush(stderr); sleep(1);
	 fprintf(stderr,"ror: caught SIGINT\n");
      }

      /*
       * If there was a real command typed, execute the right function here
       */
      if (line)
      {
	 cli_execute(line);	/* lookup and run the command */
      }
   }
   /* The while loop exits on EOF or when com_exit runs */
   puts("exiting...");
   return EXIT_SUCCESS;
}
/* end of directest.c */
