/*                                             -*- c-file-style: "Ellemtel" -*-

`Curs.cc' - See Curs.hh

This file is part of version 2.12 of the DIRECTOR cli wrapper.
Read the `License' file for terms of use and distribution.
Copyright 2000, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/termios.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
extern "C" {
#include <cfht/cfht.h>
}

#include "cli/cli.h"

#include "Curs.hh"

#define CONTROL(c) (1-'A'+c)
#define ESCAPE (27)

class History
/*
 * Keeps track of command lines as they are entered so the
 * user can scroll back and edit previous commands.  The size
 * of the history (number of commands it can hold before it
 * starts throwing away the oldest ones) is set at compile-time
 * with the define below.  The lines themselves, however, are
 * copied with strdup and can be arbitrarily long.
 */
#define HISTORYSIZE 100
#define HISTORYFILE ".director/history"
{
   private:
      char histname[1024];
      char* hist[HISTORYSIZE];
      int hist_pos; // the current command the user is editing
      int hist_end; // last command (the one to jump to with ENTER)

      void del(int pos) {
	 if (hist[pos]!=0) { free(hist[pos]); hist[pos]=0; }
      }

   public:
      History() {
	 sprintf(histname, "%.800s/"HISTORYFILE, cli_gethome());

	 FILE* histfile = fopen(histname, "rb");
	 char histline[1024];

	 hist_pos = 0;
	 hist_end = 0;
	 for (int i=0; i<HISTORYSIZE; i++) hist[i]=0;

	 if (histfile==0) return;

	 /*
	  * Re-read old history file if there was one found.
	  */
	 while (fgets(histline, 1023, histfile))
	 {
	    int len = strlen(histline);
	    if (len>1)
	    {
	       histline[len-1]='\0';
	       add(histline);
	    }
	 }
	 fclose(histfile);
      }

      ~History() {
	 int i;
	 FILE* histfile = fopen(histname, "wb");
	 
	 /*
	  * If can't open history file, silently ignore it.
	  */
	 if (histfile==0) return;
	 
	 /*
	  * Otherwise, write entire history to HISTORYFILE
	  */
	 for (i=(hist_end+1)%HISTORYSIZE; i!=hist_end; i=(i+1)%HISTORYSIZE)
	    if (hist[i])
	    { fputs(hist[i], histfile); fputc('\n', histfile); }
	 fclose(histfile);
	 for (i=0; i<HISTORYSIZE; i++) del(i);
      }

      void save(const char* s) {
	 char* saved_s;
	 char* end_s;
	 /*
	  * This is called whenever the user moves off of a line
	  * without hitting enter.  Just update the current history
	  * entry so that whatever edits they had made as still
	  * there if they scroll back.
	  */

	 while (*s==' ') s++; /* trim white space from front */

	 saved_s = strdup(s); /* make a copy of the string */
	 end_s = saved_s + strlen(saved_s)-1;
	 while (end_s >= saved_s && *end_s==' ')
	    *end_s--='\0'; /* trim white space from end */
	 
	 del(hist_pos);
	 hist[hist_pos] = saved_s;
      }
      
      const char* add(const char* s) {
	 /*
	  * This is called when the user hits enter.  The return
	  * value is a pointer to the copied version of the command
	  * and the history position is moved to the end of the
	  * list.
	  */
	 int hist_prev = (hist_end + HISTORYSIZE - 1) % HISTORYSIZE;
	 int hist_next = (hist_end + 1) % HISTORYSIZE;
	 char* saved_s;
	 char* end_s;

	 hist_pos = hist_end;

	 if (!s) return "";

	 while (*s==' ') s++; /* trim white space from front */
	 if (!*s) return "";

	 saved_s = strdup(s); /* make a copy of the string */
	 end_s = saved_s + strlen(saved_s)-1;
	 while (end_s >= saved_s && *end_s==' ')
	    *end_s--='\0'; /* trim white space from end */
	 
	 if (hist[hist_prev]!=0 &&
	     strcmp(hist[hist_prev], saved_s)==0)
	 {
	    /*
	     * don't save duplicate commands in the history
	     */
	    free(saved_s);
	    return hist[hist_prev];
	 }

	 del(hist_pos);
	 
	 hist[hist_pos] = saved_s;
	 hist_pos = hist_end = hist_next;
	 return saved_s;
      }
      
      const char* prev() {
	 /*
	  * Go to previous history item.  Returns a pointer
	  * to the saved string, or 0 if already at the top.
	  */
	 int hist_prev;
	 
	 hist_prev = (hist_pos + HISTORYSIZE - 1) % HISTORYSIZE;
	 if (hist_prev == hist_end) return 0;
	 if (hist[hist_prev]==0) return 0;
	 else return hist[hist_pos=hist_prev];
      }
      
      const char* next() {
	 /*
	  * Go to the next history item.  Returns a pointer
	  * the the save string, or 0 if already at the bottom.
	  */
	 int hist_next;
	 
	 if (hist_pos == hist_end) return 0;
	 hist_next = (hist_pos + 1) % HISTORYSIZE;
	 if (hist[hist_next]==0) return 0;
	 else return hist[hist_pos=hist_next];
      }
} history;

/*
 * To avoid problems when resizing the window, ignore
 * re-size events that are too small, as defined here:
 */
#define MIN_COLS 10
#define MIN_LINES 3

/*
 * First, some simple macro definitions to check for
 * proper return codes from various curses and form
 * library calls.  These assume that cfhtlog and and
 * feedback window are both ready.
 */

static int _cr_errors=20;
/*
 * use this counter to prevent flooding the logs with errors
 * related only to curses failures.  Print only the first 20.
 */

#define want_E_OK(fn, desc) {\
  int r;\
  if ((r=(fn))!=E_OK) {\
    if (_cr_errors && _cr_errors--)\
      cfht_logv(CFHT_MAIN, CFHT_WARN, "%s failed (return=%d)", desc, r);\
    wprintw(feedback, "warning: %s failed (return=%d)\n", desc, r);\
  }\
}

#define want_not_NULL(fn, desc) {\
  if ((fn)==NULL) {\
    if (_cr_errors && _cr_errors--)\
      cfht_logv(CFHT_MAIN, CFHT_WARN, "%s failed", desc);\
    wprintw(feedback, "%s failed", desc);\
  }\
}

#define want_not_ERR(fn, desc) {\
  if ((int)(fn)==ERR) {\
    if (_cr_errors && _cr_errors--)\
      cfht_logv(CFHT_MAIN, CFHT_WARN, "%s failed", desc);\
    wprintw(feedback, "error: %s failed", desc);\
  }\
}

/*
 * These could be defined like the above except they would print
 * errors instead of warnings, to distinguish between bad problems
 * and _really_ bad problems.  Right now, they are the same though.
 */
#define need_E_OK     want_E_OK
#define need_not_NULL want_not_NULL
#define need_not_ERR  want_not_ERR


/*
 * These colors are out of order to try and help hpterms.  On
 * xterms, it doesn't matter, but on hpterms ncurses seems to
 * have trouble setting the colors.
 */
typedef enum { COLOR_NORM=6, 
	       COLOR_INFO=2,
	       COLOR_ERROR=3,
	       COLOR_STATUS=4,
	       COLOR_STATUSBAR=5,
	       COLOR_VERBOSE=1,
	       COLOR_WARN=7,
	       COLOR_HIGHLIGHT=8 } color_t;

Curs::Curs(int infolines, int keep_bgcolor)
{
   const char* term;
   /* 
    * This gets run at the point the variable is declared.
    */

   /*
    * Set everything to zero
    */
   feedback = statbar = prompt = info = 0;
   form = 0;
   field[0] = field[1] = 0;
   statbar_showing_ = 0;
   prompt_showing_ = 1;
   infolines_ = infolines;
   keep_bgcolor_ = keep_bgcolor;
   pendsize_ = 0;
   titleclock_ = 0;
   verbose_ = 0;
   debug_ = 0;
   clone_ = 0;
   jumpscroll_ = -1;
   input_buffer_empty_ = 1;
   instant_ = strdup("");

   if (!(term=getenv("COLORTERM")) || !*term)
      if (!(term=getenv("TERM")) || !*term)
	 putenv((term="TERM=vt100"));
   if (strncmp(term,"hp",2)==0)
      always_refresh_ = 1;
   else
      always_refresh_ = 0;

   if (strncmp(term,"xterm",5)==0 ||
       strncmp(term,"rxvt",5)==0)
      titlebar_ok_ = 1;
   else
      titlebar_ok_ = 0;

   /*
    * Create the window and take over the terminal
    */
   resize(); /* init(); */
}


void
Curs::cleanup(void)
{
   if (form!=0)
   {
      want_E_OK(unpost_form(form), "unpost_form");
      want_E_OK(free_form(form), "free_form");
      form = 0;
   }
   if (field[0]!=0)
   {
      want_E_OK(free_field(field[0]), "free_field");
      field[0] = 0;
   }
   endwin();
}

static int initialized = FALSE;

void
Curs::init(int clear_feedback)
{
   const char* saveline = 0;

   /*
    * NOTE NOTE NOTE: The order of the following steps is very
    * important to the curses library.  If things are not initialized
    * in the correct order, you may get an error returned from the
    * appropriate curses call, but not always!  So test this thoroughly
    * if anything needs to be adjusted between here and the main input
    * loop.  See the ncurses man pages for more on all of these calls.
    */

   if (initialized==FALSE)
   {
      initialized = TRUE;

#define RXVT_TERMINFO
#ifdef RXVT_TERMINFO
      if (getenv("TERM")
	  && (!strcmp(getenv("TERM"), "rxvt")))
      {
#include "terminfo-rxvt.h"

	 SCREEN* tmp = newterm(0, stdout, stdin);
	 
	 if (!tmp)
	 {
	    char tienv[128];
	    char* tiname = tienv + 9;
	    int tifd;
	    
	    sprintf(tienv, "TERMINFO=/tmp/terminfo.%d", (int)geteuid());
	    putenv(strdup(tienv));
	    mkdir(tiname, 0755);
	    strcat(tiname, "/r");
	    mkdir(tiname, 0755);
	    strcat(tiname, "/rxvt");
	    tifd = open(tiname, O_RDWR|O_CREAT|O_EXCL, 0644);
	    if (tifd)
	    {
	       write(tifd,
		     terminfo_rxvt,
		     sizeof(terminfo_rxvt));
	    }
	 }
	 else
	 {
	    endwin();
	    delscreen(tmp);
	 }
      }
#endif

      initscr(); cbreak(); noecho(); nonl(); start_color();
      /*
       * Create the main feedback window
       */
      if (infolines_<0) infolines_=0;
      if (infolines_>9) infolines_=9;
      if (infolines_>LINES-3) infolines_=LINES-3;
      feedback_size_ = LINES - statbar_showing_ - prompt_showing_ - infolines_;

      need_not_NULL(feedback = 
		    newwin(feedback_size_, 0, 0, 0),
		    "newwin");
   }
   else
   {
      instant("");
      form_driver(form, REQ_BEG_FIELD);
      saveline = history.add(field_buffer(field[0], 0));
      cleanup();

      /*
       * Delete the old statbar and prompt windows (they will be
       * re-created below)
       */
      if (statbar) { delwin(statbar); statbar=0; }
      if (info)    { delwin(info); info=0; }
      if (prompt)  { delwin(prompt); prompt = 0; }

      /*
       * Re-calculate the number of lines allocated to the feedback area
       */
      if (infolines_<0) infolines_=0;
      if (infolines_>9) infolines_=9;
      if (infolines_>LINES-3) infolines_=LINES-3;
      feedback_size_ = LINES - statbar_showing_ - prompt_showing_ - infolines_;
      /*
       * Keep the feedback window and just re-size it appropriately
       * so that the contents are not erased.
       */
      wresize(feedback, feedback_size_, COLS);
   }

   /*
    * Create the windows for status bar and command prompt.
    */
   if (statbar_showing_)
      need_not_NULL(statbar= newwin(1, 0, feedback_size_, 0), "newwin");
   if (prompt_showing_)
      need_not_NULL(prompt = newwin(1, 0, feedback_size_+statbar_showing_, 0), "newwin");

   if (infolines_)
      need_not_NULL(info =
		    newwin(infolines_, 0, 
			   feedback_size_+statbar_showing_+prompt_showing_,
			   0), "newwin");

   if (prompt)
   {
      nodelay(prompt, TRUE);
      intrflush(prompt, TRUE);
      keypad(prompt, TRUE);
   }
   leaveok(feedback, TRUE);
   if (statbar) leaveok(statbar, TRUE);
   if (info)    leaveok(info, TRUE);
   scrollok(feedback, TRUE);
   wmove(feedback, feedback_size_-1, 0);

   if (has_colors()==TRUE)
   {
      int bg = -1;

      /*
       * Unless the -b option was given on the command line (meaning
       * do not clobber user's background color or pixmap) just set
       * a default black background for everything.
       */
      if (!keep_bgcolor_) bg = COLOR_BLACK;
      else use_default_colors();

      want_not_ERR(init_pair(COLOR_NORM, COLOR_WHITE, bg), "init_pair");
      want_not_ERR(init_pair(COLOR_WARN, COLOR_YELLOW, bg), "init_pair");
      want_not_ERR(init_pair(COLOR_ERROR, COLOR_RED, COLOR_WHITE), "init_pair");
      want_not_ERR(init_pair(COLOR_STATUS, COLOR_GREEN, bg), "init_pair");
      want_not_ERR(init_pair(COLOR_VERBOSE, COLOR_YELLOW, bg), "init_pair");
      want_not_ERR(init_pair(COLOR_STATUSBAR,COLOR_WHITE,COLOR_BLUE),"init_pair");
      want_not_ERR(init_pair(COLOR_INFO, COLOR_BLACK, COLOR_WHITE), "init_pair");
      want_not_ERR(init_pair(COLOR_HIGHLIGHT,COLOR_BLACK, COLOR_WHITE),"init_pair");
   }

   /*
    * Set the default attributes of each screen area.
    */

   want_not_ERR(wattrset(feedback, A_NORMAL), "wattrset");
   if (prompt) want_not_ERR(wattrset(prompt, A_NORMAL), "wattrset");
   if (statbar)
   {
      if (has_colors()==TRUE)
      {
	 want_not_ERR(wattrset(statbar, COLOR_PAIR(COLOR_STATUSBAR)|A_BOLD), 
		      "wattrset");
      }
      else
      {
	 want_not_ERR(wattrset(statbar, COLOR_PAIR(COLOR_STATUSBAR)|A_STANDOUT), 
		      "wattrset");
      }
   }
   if (info)
   {
      want_not_ERR(wattrset(info, COLOR_PAIR(COLOR_INFO)|A_STANDOUT),
		   "wattrset");
   }

   /*
    * Create the form's fields
    */

   field[1] = NULL;
   need_not_NULL((field[0]=
		  new_field(/*height=*/1, /*width=*/COLS-3,
			    /*toprow=*/0, /*leftcol=*/2,
			    /*offscreen=*/0, /*nbuffers=*/0)),
		 "new_field");
   want_E_OK(field_opts_off(field[0], O_BLANK), "field_opts_off(O_BLANK)");
   want_E_OK(field_opts_off(field[0], O_AUTOSKIP), "field_opts_off(O_AUTOSKIP)");
   want_E_OK(field_opts_off(field[0], O_STATIC), "field_opts_off(O_STATIC)");
   want_E_OK(set_max_field(field[0], 0), "set_max_field");

   /*
    * Create the form
    */

   need_not_NULL((form = new_form(field)), "new_form");

   if (has_colors())
   {
      want_E_OK(set_field_fore(field[0], A_BOLD), "set_field_fore");
      want_E_OK(set_field_back(field[0], A_NORMAL), "set_field_back");
   }

   /*
    * Associate it with the window prompt
    */

   need_E_OK(set_form_win(form, prompt), "set_form_win");

   /*
    * Post it to the screen
    */

   need_E_OK(post_form(form), "post_form");

   /*
    * If this was a resize, and the current input line was
    * saved, restore it now.
    */
   if (saveline != 0)
   {
      form_driver(form, REQ_FIRST_FIELD);
      set_field_buffer(field[0], 0, saveline);
      form_driver(form, REQ_END_LINE);
   }

   wbkgdset(feedback, COLOR_PAIR(COLOR_NORM)|A_NORMAL);

   if (clear_feedback) wclear(feedback);
   if (statbar)        wclear(statbar);
   if (info)           wclear(info);

   mvwaddch(prompt, 0, 0, '>'|A_BOLD);

   repaint();

   if (statbar) statusbar("");

   if (info)
      for (int i=0; i<infolines_; i++) infoline("",i+1);

   curs_set(1); /* Make sure cursor is visible */
   curs_set(2); /* Try to select large cursor */
   update();
}

void
Curs::infosize(int infolines)
{
   if (infolines>infolines_) wscrl(feedback, infolines-infolines_);
   infolines_=infolines;
   init(0);
}

void
Curs::repaint(void)
{
   clearok(curscr, TRUE);
   if (statbar) wrefresh(statbar);
   if (info)    wrefresh(info);
   wrefresh(feedback);
   wrefresh(curscr);
   update();
}

void
Curs::update(char prompt_char)
{
   if (prompt_char!='\0')
      mvwaddch(prompt, 0, 0, prompt_char|A_BOLD);
   mvwaddch(prompt, 0, 1, (data_behind(form)==TRUE)?('<'|A_STANDOUT):(' '|A_BOLD));
   mvwaddch(prompt, 0, COLS-1, (data_ahead(form)==TRUE)?('>'|A_STANDOUT):(' '));

   want_E_OK(pos_form_cursor(form),"pos_form_cursor");

   if (prompt_char!='>' && prompt_char!='S' && input_buffer_empty_)
   {
      int y, x;

      getyx(prompt, y, x);

      if (x==2)
      {
	 mvwaddch(prompt, 0, 0, ' '|A_BOLD);
	 wmove(prompt, 0, 0);
      }
   }

   wnoutrefresh(prompt);
   doupdate();
}

static PASSFAIL
autosize_vt100(short unsigned int* rows, short unsigned int* cols)
{
   char buf[200];
   int stdin_flags;
   struct termios tio, tioorig;
   int tmp = 1, size = 0;
   int wait = 10 * 4; /* Wait up to 10 seconds */
   int lrows, lcols;
   int tty = 0;

   if (tcgetattr(STDIN_FILENO, &tioorig)>=0)
   {
      tty = 1;

      /*
       * Turn off echo and put the terminal in raw mode
       */
      
      tio = tioorig;
      tio.c_iflag &= ~(ICRNL | IUCLC);
      tio.c_lflag &= ~(ICANON | ECHO);
      tio.c_cflag |= CS8|CREAD;
      tio.c_cc[VMIN] = 1;
      tio.c_cc[VTIME] = 0;
      if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &tio)!=0) return FAIL;
   }

   if ((stdin_flags = fcntl(STDIN_FILENO, F_GETFL, 0))<0
       || fcntl(STDIN_FILENO, F_SETFL, stdin_flags|O_NONBLOCK)<0)
   {
      if (tty) tcsetattr(STDIN_FILENO, TCSAFLUSH, &tioorig);
      return FAIL;
   }

   /*
    * Flush
    */
   while (read(STDIN_FILENO, buf, sizeof(buf))>0);

   /*
    * Cause VT100's to measure their rows and columns (requires
    * setting the full screen as the scrolling region and moving
    * to the lower left corner.)
    */
   write(STDOUT_FILENO, "\0337\033[r\033[999;999H\033[6n", 19);

   /*
    * Wait up to 4 seconds for the response
    */
   while (size < 10 && tmp != 0 && wait > 0)
   {
      struct timeval tv = { 0, 250000 };
      if ((tmp = read(STDIN_FILENO, buf+size, 10-size))==-1)
      {
	 SELECT(0, 0, 0, 0, &tv);
	 wait--;
      }
      else
      {
	 size += tmp;
	 if (size && buf[size-1]=='R') break;
      }
   }

   fcntl(STDIN_FILENO, F_SETFL, stdin_flags);
   if (tty) tcsetattr(STDIN_FILENO, TCSAFLUSH, &tioorig);

   if (size && buf[size-1]=='R' && sscanf(buf, "\033[%d;%dR", &lrows, &lcols)==2)
   {
      *rows = lrows;
      *cols = lcols;
      return PASS;
   }

   /*
    * Default to 80x24 and return failure.
    */
   *rows = 24;
   *cols = 80;
   return FAIL;
}


void
Curs::resize(void)
{
   struct winsize size;
   
   /*
    * Check the new window size and make sure everything will fit.
    */
   if (TEMP_FAILURE_RETRY(ioctl(STDOUT_FILENO, TIOCGWINSZ, &size))<0 &&
       /*
	* Should really check first to see that we are on a VT100-type
	* terminal.  This will cause some garbage to appear in HPTerms,
	* but it should get cleaned up when curses redraws everything.
	* (Unfortunately, on HPTerms where ncurses doesn't have control
	* of the pseudo terminal, the size will be wrong though.)
	*/
       autosize_vt100(&size.ws_row, &size.ws_col)==FAIL)
   {
      cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
		"Failed to resize terminal properly; assuming 80x24");
      size.ws_col = 80;
      size.ws_row = 24;
   }
   
   if (size.ws_col>=MIN_COLS && size.ws_row>=MIN_LINES)
   {
      static char lines_env[80];
      static char columns_env[80];
      
      sprintf(lines_env, "LINES=%d", size.ws_row);
      putenv(lines_env);
      sprintf(columns_env, "COLUMNS=%d", size.ws_col);
      putenv(columns_env);

      if (initialized==TRUE)
	 resizeterm(size.ws_row, size.ws_col);

      init();
   }
   else
   {
      /*
       * Ignore resizes that would make the window too small.
       */
      if (initialized==FALSE) init();
   }
   if (jumpscroll_==0) jumpscroll_=-1; /* momentarily enable jumpscroll */
}

int
Curs::confirm(const char* s)
{
   int ch;
   chtype* savedline = (chtype*)malloc((COLS+1)*sizeof(chtype));

   instant("");
   mvwinchnstr(prompt, 0, 0, savedline, COLS);
   want_not_ERR(wattrset(prompt, COLOR_PAIR(COLOR_WARN)|A_BOLD), "wattrset");
   mvwaddnstr(prompt, 0, 0, s, COLS);
   beep();
   wrefresh(prompt);
   ch = wgetch(feedback);
   want_not_ERR(wattrset(prompt, A_NORMAL), "wattrset");
   mvwaddchnstr(prompt, 0, 0, savedline, COLS);
   clearok(prompt,1);
   wrefresh(prompt);
   free(savedline);
   if (ch=='y' || ch=='Y') return 1;
   else return 0;
}

int
Curs::password(const char* prompt_string, const char* passwd)
{
   int ch;
   chtype* savedline = (chtype*)malloc((COLS+1)*sizeof(chtype));
   char passwd_entered[32];
   int passwd_length = 0;
   int rtn = 0;

   instant("");
   mvwinchnstr(prompt, 0, 0, savedline, COLS);
   want_not_ERR(wattrset(prompt, COLOR_PAIR(COLOR_WARN)|A_BOLD), "wattrset");
   mvwaddnstr(prompt, 0, 0, prompt_string, COLS);
   beep();
   wrefresh(prompt);
   while (passwd_length < 31 && (ch = wgetch(feedback))>=' ')
      passwd_entered[passwd_length++] = char(ch);
   passwd_entered[passwd_length] = '\0'; /* Null terminate the string */

   want_not_ERR(wattrset(prompt, A_NORMAL), "wattrset");
   mvwaddchnstr(prompt, 0, 0, savedline, COLS);
   clearok(prompt,1);
   wrefresh(prompt);
   free(savedline);
   if (!strcmp(passwd, crypt(passwd_entered, passwd))) rtn = 1;
   memset(passwd_entered, 0, sizeof(passwd_entered));
   return rtn;
}

void
Curs::pending(const char* s)
{
   pendsize_++;

   wsetscrreg(feedback, 0, feedback_size_-1);
   wscrl(feedback, 1);
   wmove(feedback, feedback_size_-1, 0);
   wattrset(feedback, A_BOLD);
   wprintw(feedback, "* %s", s);
   wattrset(feedback, COLOR_PAIR(COLOR_WARN));
   wprintw(feedback, "   (press ^G or cancel to undo)");
   wattrset(feedback, A_NORMAL);
   wnoutrefresh(feedback);
   if (jumpscroll_==0) update();
}

void
Curs::print(const char *s, const char* prefix)
{
   const char* p = s;
   int indent = 1;
   int maxx, maxy;

   getmaxyx(feedback, maxy, maxx);
   
   if (*s=='>' && pendsize_>0)
   {
      pendsize_--;
      wsetscrreg(feedback, 0, feedback_size_-1-pendsize_);
   }
   else
   {
      wsetscrreg(feedback, 0, feedback_size_-1-pendsize_);
      wscrl(feedback, 1);
   }
   wmove(feedback, feedback_size_-1-pendsize_, 0);

   wclrtoeol(feedback);
   if (prefix && !has_colors())
   {
      wprintw(feedback, prefix);
      indent = strlen(prefix);
   }

   if (*s=='>') wattrset(feedback, A_BOLD);
   for (p=s; *p; p++)
   {
      /* Note: === This routine is quite inefficient */
      int x, y, charsleft;
      const char* nextsp;
      char c = (*p)&0x7F;

      if ((c<' ' && c!='\t' && c!=CONTROL('H')) || c==127) c='*';

      if (!(nextsp = strpbrk(p+1," \t"))) charsleft = strlen(p+1);
      else charsleft = nextsp-p;

      getyx(feedback, y, x);
      
      if (x==maxx || (c<=' ' && charsleft > (maxx-x-1)))
      {
	 int i;

	 waddch(feedback, '\n');
	 for (i=0; i<indent+1; i++)
	    waddch(feedback, ' ');
	 if (c!=' ') waddch(feedback, c);
      }
      else
	 waddch(feedback, c);
   }
   wattrset(feedback, A_NORMAL);

   wnoutrefresh(feedback);
   if (jumpscroll_==0) update();
}

void
Curs::message(const char* s, int logtype)
{
   int c = A_NORMAL;	// color to pass to ncurses' wattrset()
   const char* p = "";	// prefix to print if no colors available

   switch (logtype)
   {
      case CFHT_DEBUG:   if (!debug_) return; /* suppress */
      case CFHT_LOGONLY: if (!verbose_) return; /* suppress */
                                    c=COLOR_PAIR(COLOR_VERBOSE); break;
      case CFHT_ERROR:p="error: ";  c=COLOR_PAIR(COLOR_ERROR)|A_REVERSE; break;
      case CFHT_WARN: p="warning: ";c=COLOR_PAIR(COLOR_WARN)|A_BOLD; break;
      case CFHT_STATUS:             c=COLOR_PAIR(COLOR_STATUS)|A_BOLD; break;
      case CFHT_HIGHLIGHT:          c=COLOR_PAIR(COLOR_INFO)|A_STANDOUT|A_BOLD; break;
   }

   /*
    * Change colors, print the message, and change back to "normal" color
    */
   wattrset(feedback, c);
   print(s, p);
   wattrset(feedback, A_NORMAL);
}

/* === look for *s=='>' in a more elegant way? */

void
Curs::titleclock(int onoff)
{
   if (titleclock_ != onoff)
   {
      titleclock_ = onoff;
      titlebar();
   }
}


const char*
Curs::titlename(const char* u)
{
   static char* u_ = strdup("???");

   if (u)
   {
      if (u_) free(u_);
      u_ = strdup(u);
   }
   return u_;
}

void
Curs::titlebar(const char* s)
{
#define TITLESIZE 1024
   static char* s_ = 0;
   static char timestr[100] = "";
   static char current_[TITLESIZE] = "";
   static char last_[TITLESIZE] = "";
   static int last_state_ = 0;
   int new_state_ = debug_|(verbose_<<1)|(titleclock_<<2);
   time_t date;

   if (!titlebar_ok_) return;
   if (!s && !titleclock_ && last_state_==new_state_) return;

   last_state_ = new_state_;

   if (titleclock_)
   {
      /*
       * Update the clock in the status line
       */
      time(&date);
      strftime(timestr, 99, "%B %d %H:%M:%S  -  ", localtime(&date));
   }
   else timestr[0]='\0';

   if (s)
   {
      if (s_) free(s_);
      s_ = strdup(s);
   }

   if (!s_) s_ = strdup("[no title]");

   sprintf(current_, "\033]2;%.100s%.100s  -  %.100s%.100s%.100s%.100s\a",
	   timestr, titlename(), s_,
	   clone_?" CLONE":"",
	   debug_?" DEBUG":"",
	   verbose_?" VERBOSE":"");
   if (strcmp(current_, last_)==0) return;
   /*
    * Only output the escape sequence if the title actually changed
    */
   strcpy(last_, current_);
   fflush(stderr);
   fputs(current_, stdout);
   fflush(stdout);
}


int
Curs::statusbar(const char *s)
{
   int rtn = 0;

   if (!statbar)
   {
      rtn = 1;
      wscrl(feedback,1);
      statbar_showing_ = 1;
      init(0);
   }

   if (always_refresh_) clearok(statbar, TRUE);
   mvwaddnstr(statbar, 0, 0, s, COLS);
   whline(statbar, (int)' ', COLS);
   wnoutrefresh(statbar);
   update();

   return rtn;
}

void
Curs::infoline(const char* s, int n)
{
   int y, x;

   if (!info || n>infolines_ || n<1) return;
   if (always_refresh_) clearok(info, TRUE);
   mvwaddnstr(info, n-1, 0, s, COLS);
   getyx(info, y, x);
   if (x>0 || !*s)
      whline(info, (int)' ', COLS);
   wnoutrefresh(info);
   update();
}

void
Curs::instant(const char* s)
{
   static chtype* savedline = 0;
   static int savedsize = 0;

   if (*s==' ') s++;
   if (strcmp(s, instant_)==0) return;

   int len = strlen(s);

   if (len==0)
   {
      free(instant_); instant_ = strdup("");

      if (savedline)
      {
	 mvwaddchnstr(prompt, 0, 2, savedline, savedsize>COLS-2?(COLS-2):savedsize);
	 clearok(prompt,1);
	 wrefresh(prompt);
	 free(savedline);
	 savedline = 0;
      }
   }
   else if (input_buffer_empty_)
   {
      if (savedline)
      {
	 if (*savedline) instant("");
	 free(savedline);
      }
      free(instant_); instant_ = strdup(s);
      savedline = (chtype*)malloc((COLS+1)*sizeof(chtype));
      mvwinchnstr(prompt, 0, 2, savedline, savedsize=COLS-2);

      want_not_ERR(wattrset(prompt, COLOR_PAIR(COLOR_WARN)|A_BOLD), "wattrset");
      mvwaddch(prompt, 0, 2, ' ');
      waddch(prompt, '(');
      waddnstr(prompt, s, COLS-5);
      waddch(prompt, ')');
      want_not_ERR(wattrset(prompt, A_NORMAL), "wattrset");
      wrefresh(prompt);
   }
}

const char*
Curs::check_line()
{
   static int insert_mode = 1; /* 1 if edit mode is "insert" */
   static int escape = 0; /* 1 if reading an escape sequence */
   static char escape_seq[1024];
   int ch, result = E_OK;
   const char* rtn = 0;
	 
   do
   {
      ch = wgetch(prompt);
      
      if (ch!=ERR) 
      {
	 input_buffer_empty_ = 0;
	 instant("");
      }

      if (escape)
      {
	 escape_seq[escape-1] = ch;
	 if (!(('0'<=ch && ch<=';') ||
	       ch=='(' || ch=='[' || ch==ESCAPE))
	    /* filter out escape sequences not handled by curses */
	 {
	    escape_seq[escape]='\0';
	    escape=0;
	    if (!strcmp(escape_seq, "[5~"))
	       ch = KEY_PPAGE;
	    else if (!strcmp(escape_seq, "[6~"))
	       ch = KEY_NPAGE;
	    else if (!strcmp(escape_seq, "[3~"))
              ch = KEY_DC;
	    else
	    {
	       beep();
	       wprintw(feedback, "Unrecognized escape sequence: `%s'\n", escape_seq);
	       wrefresh(feedback);
	       ch = ERR;
	    }
	 }
	 else
	 {
	    if (escape<1022) escape++;
	 }
      }

      if (!escape) switch (ch)
      {
	 case ERR: break; /* no chars were ready */
	 
	 case KEY_ENTER:
	 case CONTROL('M'):
	 case CONTROL('J'):
	 {
	    form_driver(form, REQ_BEG_FIELD);
	    rtn = history.add(field_buffer(field[0], 0));
	    form_driver(form, REQ_FIRST_FIELD);
	    set_field_buffer(field[0], 0, "");
	    break;
	 }
	 case ESCAPE: escape=1; break;
	 case KEY_CANCEL: return "cancel";
	 case KEY_HELP: return "help";
	 case KEY_EXIT:
	 case KEY_CLOSE:
	 case KEY_PRINT:
	 case KEY_BREAK: /* exit the program ? */
	 case KEY_SUSPEND:
	 /*
	  * If stty is correct, these should never be seen
	  */
	 case CONTROL('Z'):
	 if (confirm("[^Z] Are you sure you want to send a stop? (Y/N)"))
	    return "stop";
         break;
	 case CONTROL('C'): 
	 if (confirm("[^C] Are you sure you want to send an abort? (Y/N)"))
	    return "abort";
         break;
	 case CONTROL('\\'):
	 if (confirm("[^\\] Are you sure you want to send a quit? (Y/N)"))
	    return "quit";
         break;
	 case CONTROL('G'): return "cancel";
	 /*
	  * === the following keys are still left unbound
	  */
	 case CONTROL('I'):
	 case CONTROL('O'):
	 case CONTROL('Q'):
	 case CONTROL('R'):
	 case CONTROL('S'):
	 case CONTROL('T'):
	 case CONTROL('V'):
	 case CONTROL('W'):
	 case CONTROL('X'):
	 case CONTROL('Y'):
	 case CONTROL(']'):
	 case CONTROL('^'):
	 case CONTROL('_'):
	 waddstr(feedback, "unbound key!\n");
	 wrefresh(feedback);
	 break;
	 case KEY_IC:
	 case KEY_EIC:
	 {
	    insert_mode ^= 1;
	    result=form_driver(form, insert_mode?REQ_INS_MODE:REQ_OVL_MODE);
	    break;
	 }
	 
	 case KEY_CLEAR: /* refresh the screen */
	 case KEY_REFRESH:
	 case CONTROL('L'): repaint(); break;
	 case CONTROL('U'): result=form_driver(form, REQ_CLR_FIELD); break;
	 case KEY_EOL: /* kill to end of line === save killed region for a yank? */
	 case CONTROL('K'): result=form_driver(form, REQ_CLR_EOL); break;
	 case KEY_HOME:
	 case KEY_A1:
	 case CONTROL('A'): result=form_driver(form, REQ_FIRST_FIELD); break;
	 case KEY_END:
	 case KEY_C1:
	 case CONTROL('E'): result=form_driver(form, REQ_END_LINE); break;
	 case KEY_LEFT:
	 case CONTROL('B'): result=form_driver(form, REQ_LEFT_CHAR); break;
	 case KEY_RIGHT:
	 case CONTROL('F'): result=form_driver(form, REQ_RIGHT_CHAR); break;
	 case KEY_UP:
	 case CONTROL('P'):
	 {
	    form_driver(form, REQ_BEG_FIELD);
	    history.save(field_buffer(field[0], 0));
	    {
	       const char* h = history.prev();
	       
	       if (!h) beep();
	       else
	       {
		  set_field_buffer(field[0], 0, h);
		  result=form_driver(form, REQ_END_LINE);
	       }
	    }
	    break;
	 }
	 case KEY_DOWN:
	 case CONTROL('N'):
	 {
	    form_driver(form, REQ_BEG_FIELD);
	    history.save(field_buffer(field[0], 0));
	    {
	       const char* h = history.next();
	       
	       if (!h) beep();
	       else
	       {
		  set_field_buffer(field[0], 0, h);
		  result=form_driver(form, REQ_END_LINE);
	       }
	    }
	    break;
	 }
	 case KEY_BACKSPACE:
	 case KEY_DC:
	 case 127:
	 case CONTROL('H'): result=form_driver(form, REQ_DEL_PREV); break;
	 case CONTROL('D'): result=form_driver(form, REQ_DEL_CHAR); break;
	 /* REQ_PREV_WORD */
	 /* REQ_NEXT_WORD */
	 default:
	 if (' '<=ch && ch<='~') result=form_driver(form, ch);
	 else { beep(); }
      }
      
      if (result != E_OK) beep();
	
      if (ch!=ERR) update();

   } while (rtn==0 && ch!=ERR);

   if (rtn) input_buffer_empty_ = 1;
   else input_buffer_empty_ = ((mvwinch(prompt, 0, 2)&A_CHARTEXT)<=' '); /* %%% ineff. */
   return rtn;
}
/* end of Curs.cc */
