/*                                             -*- c-file-style: "Ellemtel" -*-

`Pipes.hh' - reads line-by-line from pipe or fifo sources

This file is part of version 2.12 of the DIRECTOR cli wrapper.
Read the `License' file for terms of use and distribution.
Copyright 2000, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */
#ifndef _INCLUDED_Pipes
#define _INCLUDED_Pipes 1

#include <fcntl.h>
extern "C" {
#include <cfht/cfht.h>
}
typedef int PASSFAILBUSY;
#define BUSY (PASS*2 + FAIL*3) /* === argh.  fix this */

#define CMD_MAXLEN 1023

class LineBuffer
/*
 * This class reads any pending input from a file descriptor
 * each time "check_line()" is called.  The fd passed to set_fd
 * will be set to non-blocking i/o, so that each call to
 * check_line() will immediately return either
 *
 *  0 if no complete line has been read yet, or
 *  a character pointer to a complete string, with the \n stripped.
 */
{
   private:
      int fd;
      int* command_counter;
      PASSFAILBUSY* command_result;
      BOOLEAN found_prompt;
      char buf[2][CMD_MAXLEN+2]; /* +2 for newline hack */
      char* bufp;
      int bufnum;
      char lastnewline;

   public:
      LineBuffer();

      void set_command_counter(int* counter, int* result) {
	 command_counter = counter;
	 command_result = result;
      }

      void set_fd(int newfd, int non_blocking=1) {
	 fd = newfd;
	 if (fd!=-1 && non_blocking)
	    if (fcntl(fd, F_SETFL, fcntl(fd, F_GETFL, 0)|O_NONBLOCK)==-1)
	    {
	       cfht_logv(CFHT_MAIN, CFHT_ERROR,
			 "can't set non-blocking i/o for fd %d (%s)\n",
			 fd, strerror(errno));
	       fprintf(stderr, "can't set non-blocking i/o for fd %d (%s)\n",
		       fd, strerror(errno));
	    }
      }
      
      int get_fd() { return fd; }

      const char* check_line();
};


class CommandPipe : public LineBuffer
{
   private:
      int command_fd;
      int autoremove;
      const char* pipe_name;
      
   public:
      CommandPipe(const char* name, int wronly=0);
      ~CommandPipe();
      void send(const char* command);
      void shutdown(); /* Stop accepting input */
};

class LogPipe : public LineBuffer
{
   private:
      int log_fd;
      int autoremove;
      const char* pipe_name;

   public:
      LogPipe(const char* name);
      ~LogPipe();
};


class ProcessPipe : public LineBuffer
{
   private:
      enum { PIPE_NOT_USED, PIPE_TO_DEVICE, PIPE_TO_PROCESS, PIPE_TO_RCMD } type;
      int child_input[2];
      int child_output[2];
      int child_stderr[2];
      pid_t child;
      int signalmap[NSIG]; /* === argh.  different host types assign signals differently! */
      char** argv_;
      void shutdown();

   public:
      int commands_pending;
      PASSFAILBUSY command_result;
      LineBuffer err;

      ProcessPipe(const char** argv);
      ~ProcessPipe();

      PASSFAIL restart();

      int signal(int signo);

      BOOLEAN running() {
	 if (get_fd()==-1 && err.get_fd()==-1)	return FALSE;
	 else					return TRUE;
      }

      PASSFAIL send(const char* command);

      PASSFAILBUSY pend() {
	 if (command_result==BUSY && !running())
	 {
	    commands_pending--;
	    command_result = FAIL;
	 }
	 return command_result;
      }
         /*
	  * Returns PASS: if the last command completed successfully (prompt was "ok>")
	  *         FAIL: if the last command failed (prompt was "failed>")
	  *         BUSY: if the last command is still executing. (no prompt yet)
	  */
      
      const char* postmortem();
	 /*
	  * tries to conduct an autopsy on an exited process and
	  * returns a string describing what happened, or 0 if the
	  * process is apparently still running.
	  */

      const char* check_line();
         /*
	  * Krusty: prescans LineBuffer's check_line() for "signalmap:" directives!
	  */
};

#endif /* !_INCLUDED_Pipes */
