/*                                             -*- c-file-style: "Ellemtel" -*-

`Roll.hh' - holds of the latest N lines echoed in shared memory

This file is part of version 2.12 of the DIRECTOR cli wrapper.
Read the `License' file for terms of use and distribution.
Copyright 2000, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#ifndef _INCLUDED_ROLL
#define _INCLUDED_ROLL 1

#define INFO_LINES    9	/* maximum number of lines in info panel */
#define ROLL_LINES 3000	/* maximum number of lines in roll buffer at once */
#define ROLL_MEM 120000	/* maximum bytes kept in roll buffer at once */
#define LINE_WIDTH 1000	/* maximum number of characters per line */

#define CFHT_HIGHLIGHT -10 /* === hack, add field to RollLine struct instead? */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>

extern "C" {
#include <cfht/cfht.h>
}

/* NOTE: RollPtr, RollLine, RollShm are just structures instead of classes
 * because they need to exist in shared memory.  Functions that would
 * be members of these classes are part of the Roll class instead.
 */

struct RollPtr
/*
 * Information about each line in the roll buffer is kept in one of these
 * structures.  It stores the type, length, and the offset in buf[] where
 * the null terminated string begins.
 */
{
   int offset;			// offset into buffer where string starts
   int length_handle;		// length of the "handle" part of the message (e.g. "directest: ")
   int length;			// length of the message or -1 if not yet known
   int logtype;			// cfht_log type of message or -1 if not yet known
   time_t timestamp;		// time that the message was entered into the log
   unsigned char reserved[8];	// reserved for future use, fill with 0's
};

struct RollLine
/*
 * For static screen lines, this structure is used instead.  Every time the
 * line is modified, the serial number is incremented so others watching
 * the shared memory know when to re-display it.
 */
{
   char str[LINE_WIDTH+1];	// nul terminated string with the message
   int length;			// length of the message or -1 if not yet known
   int serial;			// serial number (incremented if str rewritten)
};

struct RollShm
/*
 * An array of ROLL_LINES RollLine's is stored in the shared memory segment.
 * New entries are added to line[] in increasing order.  `tail' always points
 * to the most recent line.  `(tail+1)%ROLL_LINES' points to the head.  Unused
 * slots contain the empty string "".
 */
{
#define DIRECTOR_MAGIC 0xEDDE
   unsigned short magic;	// director magic number (should stay the same forever)
   int version;			// director version number (i.e. 2, 3, ...)
   int roll_key;		// uid of the session user (usually = shm key)
   int time_to_exit;		// flag set by parent when it wants clones to exit
   int tail;			// tail entry in roll[], or -1 if time to exit
   int head;			// head entry in roll[]
   int nextofs;			// offset into buf where next message will go
   int infosize;		// current infosize for master-director (for new clones)
   char buf[ROLL_MEM];		// space for the actual text
   RollPtr roll[ROLL_LINES];	// circular buffer of feedback lines
   RollLine info[INFO_LINES];	// text for info panel (if displayed)
   RollLine statusbar;		// current text for the status bar
   RollLine titlebar;		// current text for the title bar
   RollLine instant;		// current text for "instant" message area
   unsigned char reserved[1024];// reserved for future use, fill with 0's
};

typedef enum { CLONE_OR_MASTER, CLONE_ONLY, MASTER_ONLY } clone_mode_t;

class Roll
{
   private:
      RollShm* shm;		// pointer to attached shared memory segment
      int shmid;		// shared memory id returned by shmget()
      int clone_;		// flag indicating whether in clone mode or not
      int disp;			// display is all caught up when this == shm->tail
      int statusbar_serial_disp, titlebar_serial_disp, instant_serial_disp,
	 info_serial_disp[INFO_LINES];

      void init_RollLine(RollLine& r);
      void set_RollLine(RollLine& r, const char* s, int eatspace=0);
      void set_RollPtr(const char* s, int logtype=-1, int eatspace=0);

   public:
      Roll(int roll_key, clone_mode_t clone_mode);
	 /*
	  * Constructor for Roll class attaches to the shared memory block
	  * If must_clone is zero, Roll will automatically determine if it
	  * is a clone monitoring another director's RollShm, or if it should
	  * actually allocate it.  If must_clone is non-zero, having to
	  * allocate the Shm is treated as an error.
	  */
      
      ~Roll();
	 /*
	  * Destructor for roll class detaches from shared memory
	  * If there are clones, the master process also waits for
	  * clones to exit before removing the shared memory id.
	  */

      int is_clone();
	 /*
	  * Returns 0 if master process
	  *         1 if clone and master process is still running
	  *        -1 if clone and master process is trying to exit
	  */

      int nattch();
	 /*
	  * Get number of processes attached to shm segment.  Returns -1
	  * on error from shmctl()
	  */
	 
      void rewind(int lines);
	 /*
	  * Moves the disp index back so that get_line() will
	  * return `lines' previous lines from the current position
	  */
      
      RollPtr* get_line();
	 /*
	  * Returns a pointer to the next roll buffer line, or 0 if
	  * there are no new lines since the last call.
	  */

      const char* get_str(RollPtr* rp);
         /*
	  * Returns a pointer to a string in shared memory that corresponds
	  * to the message structure returned by get_line, or NULL if rp is NULL
	  */

      RollLine* get_statusbar();
	 /*
	  * Returns a pointer to the statusbar line if changed, else 0.
	  */

      RollLine* get_titlebar();
	 /*
	  * Returns a pointer to the titlebar line if changed, else 0.
	  */

      RollLine* get_instant();
	 /*
	  * Returns a pointer to the titlebar line if changed, else 0.
	  */

      RollLine* get_info(int n);
	 /*
	  * Returns a pointer to the n-th info line if changed, else 0.
	  * n should be in the range 1..INFO_LINES
	  */

      void add_line(const char* s, int logtype=-1, int nolog=0, int no_echo=0);
	 /*
	  * Adds a new line to the buffer and logs it to cfht_log
	  * if necessary.
	  *
	  * As long as nolog is 0, the message is also logged to cfht_log.
	  *
	  * If logtype is a valid cfht_log type (eg. CFHT_ERROR), the message
          * will be shown as that type if no other "logtype:" key is found
	  * at the beginning of the string.
	  */

      void instant_off();
	 /*
	  * This gets called when a command pipe receives the '>' line
	  * and clears the "instant:..." message (if any) currently being
	  * displayed in the prompt area.
	  */
      
      void set_infosize(int lines);
      int get_infosize();
         /*
	  * A master can use this to store a hint for the starting infopanel
	  * size of new clones.  It is only provided as a hint, and the user
	  * can always set their own infosize within their clone using the
	  * infosize builtin command.
	  */
      
};
#endif /* !_INCLUDED_ROLL */
