/*                                             -*- c-file-style: "Ellemtel" -*-

`director.cc' - main event loop for director program

This file is part of version 2.12 of the DIRECTOR cli wrapper.
Read the `License' file for terms of use and distribution.
Copyright 2000, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */


#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <termios.h>
#include <pwd.h>

extern "C" {
#include <cfht/cfht.h>
}

#include "cli/cli.h"
#include "cli/clidefs.h"

void do_sleep(int seconds); /* temporarily disable the interval timer and do
			     * a regular sleep 
			     */

const char* ps1      = DEFAULT_PROMPT; /* === don't override this one! */
const char* ps2      = DEFAULT_PROMPT;
const char* psok     = DEFAULT_OK_PROMPT;
const char* psfailed = DEFAULT_FAILED_PROMPT;
const char* pscont   = DEFAULT_CONT_PROMPT;
const char* cli_newline = "\n";

int ps1len, ps2len, psoklen, psfailedlen, pscontlen; /* === hack */

static int autorestart = 0;
#define RESTART_THRESH 60 /* don't autorestart if agent dies faster than about every 30 seconds! */

/*
 * 0                 - don't restart
 * 1..RESTART_THRESH - don't restart, but increment autorestart about twice a second
 * >RESTART_THRESH   - allow restart if necessary, reset counter to 1
 */

static char hostname[81] = "unknown";

static int sessionhost_created = 0;

static int get_out = 0;
/*
 * When the program exits, it looks at this to determine
 * how "quickly" it should cleanup and exit.
 *
 * A value of 0 indicates the program is probably NOT exiting as a
 * direct result of a "detach" or "exit" command, so pause for 3 seconds
 * to let the user see what happened.
 *
 * A value of 1 indicates a "detach" or "exit" command, so don't pause.
 *
 * A value >1 indicates something unexpected happened during cleanup and
 * we should call _exit() to get out as soon as possible.
 */
                                                                             
/*
 * Order is important here . . . See Index for some descriptions.
 */

#include "Roll.hh"

Roll* roll=0; /* === Should be static but I want access to it from Pipes.cc */

#include "Curs.hh"

static Curs* curs=0;

#include "Pipes.hh"

static ProcessPipe* process=0;
static CommandPipe* commandPipe=0;
static LogPipe* logPipe=0;

static void
return_to_clicmd(pid_t pid, PASSFAIL result, int no_fail = 0)
{
   if (!pid || pid==-1) return;
   if (no_fail && result!=PASS) result = PASS;
   if (kill(pid, result==PASS?SIGUSR1:SIGUSR2)==-1)
   {
      char junkbuf[1024];

      sprintf(junkbuf,
	      "%s: Can't signal pid %d to tell it command %s.  (error was: `%s')",
	      errno==ESRCH?"logonly":"error", (int)pid,
	      result==PASS?"passed":"failed", strerror(errno));
      if (roll) roll->add_line(junkbuf);
      else cfht_log(CFHT_MAIN, CFHT_ERROR, junkbuf);
   }
}


/* --------------------------------------------------- */

#define MAX_PEND 100

typedef struct scripttag
{
   char filename[128];
   FILE* fp;
   int lineno;
   struct scripttag* parent;
} ScriptFile;


#define MAX_SOURCE_NESTING 10

static class Scripter {
   private:
      ScriptFile* scriptFile;
      int nest_level; /* just a sanity check */
      char junkbuf[1024];

   public:
      Scripter() { scriptFile = 0; nest_level = 0; }
      ~Scripter() { cancel(); }

      PASSFAIL add(const char* filename) {
	 if (nest_level > MAX_SOURCE_NESTING)
	 {
	    if (roll) roll->add_line("error: too many nested `source' commands!");
	    return FAIL;
	 }

	 ScriptFile* s = new ScriptFile;
	 
	 if ((s->fp = fopen(filename,"r"))==NULL)
	 {
	    sprintf(junkbuf,
		    "error: source[%d]: cannot run `%.500s' -- %.100s",
		    nest_level, filename, strerror(errno));
	    if (roll) roll->add_line(junkbuf);
	    delete s;
	    return FAIL;
	 }
	 s->lineno = 0;
	 strncpy(s->filename, filename, 127); s->filename[127]='\0';
	 s->parent = scriptFile;
	 scriptFile = s;
	 nest_level++;
	 return PASS;
      }
      
      const char* check_line() {
	 static char scriptLine[1024];
	 
	 if (!scriptFile) return 0;

	 if (fgets(scriptLine, sizeof(scriptLine), scriptFile->fp))
	 {
	    char* last;
	    char* first = scriptLine;
	    /*
	     * Successfully read a line from the file.  Return it.
	     */
	    if (++scriptFile->lineno == 1)
	    {
	       sprintf(junkbuf,
		       "status: source[%d]: Now executing `%s'",
		       nest_level,
		       scriptFile->filename);
	       if (roll) roll->add_line(junkbuf);
	    }
	    while (*first && *first<=' ') first++;
	    last = first + strlen(first) - 1;
	    while (last >= first && *last<=' ') *last-- = '\0';
	    if (*first && *first!='#') return first; else return check_line();
	 }
	 else
	 {
	    /*
	     * close the file and go back to taking commands from
	     * where-ever we were taking them from before.
	     */
	    ScriptFile* finished = scriptFile;
	    
	    fclose(finished->fp);
	    sprintf(junkbuf, "status: source[%d]: Leaving `%s' (%d line%s read)",
		    nest_level,
		    finished->filename,
		    finished->lineno,
		    finished->lineno==1?"":"s");
	    if (roll) roll->add_line(junkbuf);
	    scriptFile = finished->parent;
	    delete finished;
	    nest_level--;
	    return check_line();
	 }
      }

      void cancel() {
	 /*
	  * Close all scripts without reading any further
	  */

	 while (scriptFile)
	 {
	    ScriptFile* aborted = scriptFile;
	    
	    fclose(aborted->fp);
	    sprintf(junkbuf, "error: source[%d]: script aborted at line %d",
		    nest_level, aborted->lineno);
	    if (roll) roll->add_line(junkbuf);
	    scriptFile = aborted->parent;
	    delete aborted;
	    nest_level--;
	 }
      }
      BOOLEAN waiting() { return scriptFile==0?FALSE:TRUE; }

} scripter;


static class PendBuffer {
   private:
      char* line[MAX_PEND];
      pid_t pids[MAX_PEND];
      int _no_echo[MAX_PEND];
      int _no_fail[MAX_PEND];
      int pos_add;
      int pos_get;

   public:
      PendBuffer() {
	 pos_add = pos_get = 0;
	 for (int i=0; i<MAX_PEND; i++) { line[i]=0; pids[i]=0; _no_echo[i]=0; _no_fail[i]=0; }
      }

      void add_line(const char* s, pid_t pid, int no_echo, int no_fail) {
	 int next = (pos_add + 1)%MAX_PEND;

	 if (line[next]) free(line[next]);
	 line[next] = strdup(s);
	 pids[next] = pid;
	 _no_echo[next] = no_echo;
	 _no_fail[next] = no_fail;
	 pos_add = next;
      }
      
      const char* check_line() {
	 if (pos_get != pos_add)
	 {
	    pos_get = (pos_get + 1)%MAX_PEND;
	    return line[pos_get];
	 }
	 else 
	 {
	    return 0;
	 }
      }

      void cancel() {
	 /*
	  * Used after delivering a signal to cancel queued up commands
	  */
	 const char* curline;
	 char junkbuf[1024];

#if 0
	 /*
	  * === hack to make sure other about-to-be queued commands also get canceled.
	  */
	 do_sleep(1);
#endif

	 /*
	  * Dump all pending input lines
	  */
	 return_to_clicmd(check_pid(), FAIL);

	 while ((curline=check_line())!=0)
	 {
	    sprintf(junkbuf, "> %.500s    *** canceled ***", curline);
	    /* make sure clicmd's don't hang */
	    return_to_clicmd(check_pid(), FAIL);
	    if (roll) roll->add_line(junkbuf);
	 }
	 if (process && process->commands_pending > 0)
	    process->commands_pending = 0;
      }

      pid_t check_pid() { return pids[pos_get]; }
      void reset_pid()  { pids[pos_get] = 0; }
      int no_echo()     { return _no_echo[pos_get]; }
      int no_fail()     { return _no_fail[pos_get]; }

} pendBuffer;


class FDSet
{
/*
 * Manages file descriptor sets for select.
 * It is acceptable to pass a -1 to as the fd to the
 * functions below and they will ignore that particular
 * file descriptor
 */
   private:
      fd_set fdset;
      int fdnum;

   public:
      FDSet() { fdnum=0; FD_ZERO(&fdset); }
      FDSet(const FDSet& f) { *this = f; }
      FDSet& operator=(const FDSet& f) { fdset = f.fdset; fdnum = f.fdnum; return *this; }

      void add_fd(int fd) { 
	 if (fd!=-1)
	 {
	    if (fdnum<=fd) fdnum = fd+1; FD_SET(fd, &fdset);
	 }
      }

      int get_fdnum() { return fdnum; }
      fd_set* get_fdsetp() { return &fdset; }
      
      int is_set(int fd) { if (fd==-1) return 0; else return FD_ISSET(fd, &fdset); }
};

static char* real_username = "Unknown";

static class Who
/*
 * Handles figuring out the correct roll key (uid) $HOME and $USER
 */
{
   private:
      clone_mode_t clone_mode_;
      uid_t uid_;
      int key_;
      char* name_;
      char* home_;
      char* passwd_;

      void save_pwd(struct passwd* pwd) {
	 uid_ = pwd->pw_uid;
	 if (name_) free(name_);
	 name_ = strdup(pwd->pw_name);
	 if (home_) free(home_);
	 home_ = strdup(pwd->pw_dir);
	 if (passwd_) free(passwd_);
	 passwd_ = strdup(pwd->pw_passwd);
      }

   public:
      Who() {
	 /*
	  * Constructor just sets up assuming we are going to
	  * run using getuid()'s session and shared memory.
	  */
	 struct passwd* pwd;

	 clone_mode_ = CLONE_OR_MASTER;
	 name_ = home_ = passwd_ = 0;
	 uid_ = getuid();
	 key_ = -2;
	 pwd = getpwuid(uid_);
	 if (!pwd) 
	 {
	    cfht_logv(CFHT_MAIN, CFHT_ERROR, "Can't find userid %d in password file", uid_);
	    perror("getpwuid"); 
	    exit(EXIT_FAILURE); 
	 }
	 save_pwd(pwd);
      }
	 
      void setup(const char* arg) {
	 /*
	  * This is called to override the default uid when director
	  * is invoked with `-k<user>'
	  */
	 struct passwd* pwd = 0;
	 
	 clone_mode_ = CLONE_ONLY;
	 errno = 0;
	 uid_ = strtoul(arg, NULL, 10);
	 if (uid_!=0 && errno==0)
	 {
	    pwd = getpwuid(uid_);
	 }
	 if (!pwd)
	 {
	    pwd = getpwnam(arg);
	 }
	 if (!pwd)
	 {
	    cfht_logv(CFHT_MAIN, CFHT_ERROR, "Can't find user %s in password file", arg);
	    perror(arg); 
	    exit(EXIT_FAILURE);
	 }
	 save_pwd(pwd);
      }
      
      void force_master() { clone_mode_ = MASTER_ONLY; }
      
      const char* name() { return name_; } /* user name from passwd */
      const char* home() { return home_; } /* home directory from passwd */
      const char* passwd() { return passwd_; } /* encrypted password */
      void setkey(int key) { key_ = key; }
      int getkey() { return (key_!=-2)?key_:uid_; } /* user id number */
      clone_mode_t clone_mode() { return clone_mode_; } /* was -k/-m forced? */
} who;

/* --------------------------------------------------- */

static int debugflg = 0;

void /* %%% Was static ! */
flush_to_screen()
   /*
    * Make the info in on the curs (the screen)
    * catch up with whats in Roll (the shared memory)
    */
{
   RollPtr* rp;
   
   if (!roll || !curs) return;

   while ((rp=roll->get_line())!=0)
   {
      /*
       * the master director process strips out \a's and rings the
       * local terminal bell immediately.  Clones only ring the bell
       * when an error is printed (or re-printed after a resize).
       */
      if (roll->is_clone() && rp->logtype==CFHT_ERROR)
	 fputc('\a', stderr);	
      curs->message(roll->get_str(rp), rp->logtype);
   }
   
   RollLine* rl;
   
   if ((rl=roll->get_instant())!=0)
      curs->instant(rl->str);
   
   if ((rl=roll->get_statusbar())!=0)
      if (curs->statusbar(rl->str))
	 roll->rewind(0); /* redraw the info bars */
   
   if (!debugflg && (rl=roll->get_titlebar())!=0)
      curs->titlebar(rl->str);
   
   for (int l=1; l<=9; l++)
   {
      if ((rl=roll->get_info(l))!=0)
	 curs->infoline(rl->str, l);
   }
   
   /*
    * If jumpscroll was temporarily turned on, turn it off again now.
    */
   curs->jumpscroll_check();
   
   curs->update(
      scripter.waiting()?'S':
      ((process&&process->commands_pending<0)?'>':'*'));
}

void /* %%% Was static ! */
flush_inputs()
   /*
    * Add lines from process stdout/stderr and CFHTLOGU
    * to the shared memory roll buffer
    */
{
   int max_repeat = 25; /* Don't "lock up" here if overflowing with input */
   const char* curline = "";

   if (!roll) return;

   while (max_repeat-- && curline && process)
   {
      if ((curline=process->check_line())!=0)
	 roll->add_line(curline, -1, 0, pendBuffer.no_echo());
      
      else if ((curline=process->err.check_line())!=0)
	 roll->add_line(curline, CFHT_WARN, 0, pendBuffer.no_echo());
   }
   
   max_repeat = 25;

   while (max_repeat-- && logPipe && (curline=logPipe->check_line())!=0)
   {
      /*
       * When session wants to end and make the feedback window go
       * away, it echos ^QUIT to the CFHTLOGU pipe.  For easy integration
       * into the sessions, we'll just have director do the same for now.
       */
      if (strncmp(curline, "^QUIT", 5)==0)
      {
	 flush_inputs();
	 roll->add_line("logonly: session cleanup in progress");
	 flush_to_screen();
	 get_out = 1;
	 exit(EXIT_SUCCESS);
      }
      else
	 roll->add_line(curline, /*logtype=*/-1, /*nolog=*/1);
   }
}

#include "builtins.cc"

void
set_interval_timer(int onoff)
{
   struct itimerval timeout;

   if (onoff)
   {
      cfht_signal(SIGALRM, CFHT_SIG_TRAP);
      cfht_signal_block(SIGALRM);
   }

   timeout.it_value.tv_sec=0;
   timeout.it_value.tv_usec=(onoff!=0)*500000;
   timeout.it_interval.tv_sec=0;
   timeout.it_interval.tv_usec=(onoff!=0)*500000;

   setitimer(ITIMER_REAL, &timeout, NULL);

   if (!onoff)
   {
      cfht_signal_unblock(SIGALRM);
      cfht_signal_event(SIGALRM);
      cfht_signal(SIGALRM, CFHT_SIG_DFL);
   }      
}

void
do_sleep(int sec)
{
   set_interval_timer(0);
   sleep(sec);
   set_interval_timer(1);
}

static const char*
do_tilde_substitution(const char* orig)
{
   static char buf[1024];
   char user[10];
   char* userp = user;
   char* bufp = buf;
   const char* origp = orig;
   size_t len, ulen = 0, slen = 0;
   struct passwd* pwd;

   while (1)
   {
      const char* subst;

      /*
       * Copy one character at a time until a ~ after a space is found
       */
      while (*origp &&
	     (*origp!='~' || origp==orig || *(origp-1)>' ') &&
	     slen < sizeof(buf)-1)
      { *bufp++ = *origp++; slen++; }

      if (*origp++ != '~') break; /* end of string or buf full */

      /*
       * Get what could be a possible username
       */
      for (userp = user; 
	   *origp > ' ' && *origp != '/' && ulen<9;
	   *userp++ = *origp++, ulen++);
      *userp = '\0'; /* terminate username string */

      if (ulen==0)
	subst = who.home();	// empty means use default username
      else if (ulen<9 && (pwd=getpwnam(user)))
	 subst = pwd->pw_dir;	// found in passwd table, substitute
      else
      { *bufp++ = '~'; slen++; subst = user; } // otherwise leave alone

      slen += (len = strlen(subst));
      if (slen > sizeof(buf)-1) break; /* overflow */
      strcpy(bufp, subst);
      bufp += len;
   }
   *bufp = '\0'; /* terminate the string */
   return buf; /* return pointer to static area */
}

static int unlocked = FALSE; /* Has the account password been entered? */

static void
final_cleanup(void)
   /*
    * This takes care of making sure the destructors for
    * roll, process, and curs always take place.  The
    * final_cleanup() function should be registered with
    * atexit() so it runs even if the program exits abnormally.
    *
    * There is no harmful side-effect from calling this function
    * more than once at exit.
    */
{
   cfht_signal(SIGHUP, CFHT_SIG_TRAP);
   cfht_signal(SIGTERM, CFHT_SIG_TRAP);

   if (roll && roll->is_clone()>0 && commandPipe)
   {
      char junkbuf[1024];

      sprintf(junkbuf,
	      "say_ %s%s has detached from the session.",
	      unlocked?"warning:":"* ", real_username);
      commandPipe->send(junkbuf); /* === krusty */
   }
   flush_inputs();

   if (roll)
   {
      roll->add_line("logonly: cleaning up");
      if (curs) curs->update(' ');
      flush_to_screen();
   }

   if (sessionhost_created)
   {
      if (roll) roll->add_line("logonly: removing sessionhost file");
      chdir(who.home());
      unlink(".director/sessionhost");
      sessionhost_created = 0;
   }

   if (process) { delete process; process=0; }

   if (commandPipe) 
   {
      const char* curline;

      if (roll)
      {
	 roll->add_line("logonly: cancelling commands");
	 flush_to_screen();
      }

      scripter.cancel();
      pendBuffer.cancel();

      if (roll)
      {
	 roll->add_line("logonly: closing command pipe");
	 if (curs) curs->update(' ');
	 flush_to_screen();
      }

      commandPipe->shutdown();

      while ((curline = commandPipe->check_line()))
      {
	 const char* endbrace;
	 int killpid;

	 while (*curline!='\0' && *curline<=' ') curline++; /* eat white space */

	 /* check for `{PID}command' notation */

	 if (*curline=='{' &&
	     (endbrace=strchr(curline,'}'))!=0)
	 {
	    killpid = atoi(curline+1);
	    curline = endbrace + 1;
	    if (killpid != 0)
	    {
	       if (roll) roll->add_line("warning: director: trashing pending commandPipe input");
	       return_to_clicmd(killpid, FAIL);
	    }
	 }
      }

      delete commandPipe; commandPipe=0;
   }

   if (logPipe)
   {
      if (roll)
      {
	 roll->add_line("logonly: closing log pipe");
	 if (curs) curs->update(' ');
	 flush_to_screen();
      }
      delete logPipe; logPipe=0;
   }

   if (roll)
   {
      roll->add_line("logonly: closing roll"); 
      if (curs) curs->update(' ');
      flush_to_screen();
      delete roll; roll=0;
   }

   if (curs)
   {
      cfht_logv(CFHT_MAIN, CFHT_LOGONLY, "restoring terminal");
      /*
       * Restore the terminal to a usable state
       */
      curs->update(' ');
      printf("\033]2;%s\a", hostname);
      delete curs; curs=0;
   }
   printf("\r\n");

   if (cfht_signal_event(SIGHUP) ||
       cfht_signal_event(SIGTERM)) get_out = 2;

   if (!get_out)
   {
      cfht_logv(CFHT_MAIN, CFHT_LOGONLY, "pausing for 5 seconds");
      do_sleep(5); /* Give the user some time to read the screen */
      get_out = 2;
   }
   else if (get_out>1)
   {
      cfht_logv(CFHT_MAIN, CFHT_DONE, "_exiting");
      _exit(EXIT_SUCCESS);
   }

   cfht_logv(CFHT_MAIN, CFHT_DONE, "exiting");
}

static void
signal_cleanup(int signo)
{
   static char buf[81];

   cfht_signal(signo, CFHT_SIG_DFL);

   if (signo==SIGHUP || signo==SIGTERM) get_out = 2;

   sprintf(buf, "director on host %s caught fatal signal #%d", hostname, signo);
   cfht_logv(CFHT_MAIN, CFHT_LOGONLY, buf);
   fprintf(stderr, "\r%s\n", buf);

   final_cleanup();
   kill(getpid(), signo);
}

BOOLEAN
send_ok(void)
{
   if (unlocked==TRUE || !roll->is_clone()) return TRUE;

   if (curs->password("Enter account password to unlock your clone:", who.passwd()) && commandPipe)
   {
      char junkbuf[1024];

      sprintf(junkbuf,
	      "say_ warning: %s is sending commands from a clone window.",
	      real_username);
      commandPipe->send(junkbuf); /* === krusty */
      unlocked=TRUE;
   }

   return unlocked;
}

static void
autosense_rxvt()
{
   char buf[200];
   int stdin_flags;
   struct termios tio, tioorig;
   int tmp = 1, size = 0;
   int wait = 5 * 4; /* Wait up to 5 seconds */
   int tty = 0;

   if (tcgetattr(STDIN_FILENO, &tioorig)>=0)
   {
      tty = 1;

      /*
       * Turn off echo and put the terminal in raw mode
       */
      
      tio = tioorig;
      tio.c_iflag &= ~(ICRNL | IUCLC);
      tio.c_lflag &= ~(ICANON | ECHO);
      tio.c_cflag |= CS8|CREAD;
      tio.c_cc[VMIN] = 1;
      tio.c_cc[VTIME] = 0;
      if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &tio)!=0) return;
   }

   
   if ((stdin_flags = fcntl(STDIN_FILENO, F_GETFL, 0))<0
       || fcntl(STDIN_FILENO, F_SETFL, stdin_flags|O_NONBLOCK)<0)
   {
      if (tty) tcsetattr(STDIN_FILENO, TCSAFLUSH, &tioorig);
      return;
   }

   /*
    * Flush
    */
   while (read(STDIN_FILENO, buf, sizeof(buf))>0);

   /*
    * A control-E will cause RXVT to send back it's "I'm an enhanced
    * VT102" response.
    */
   write(STDOUT_FILENO, "\005", 1);

   /*
    * Wait up to 4 seconds for the response
    */
   while (size < 7 && tmp != 0 && wait > 0)
   {
      struct timeval tv = { 0, 250000 };
      if ((tmp = read(STDIN_FILENO, buf+size, 7-size))==-1)
      {
	 SELECT(0, 0, 0, 0, &tv);
	 wait--;
      }
      else size += tmp;
   }

   /*
    * Is it an RXVT?
    */
   if (size == 7 && !strncmp(buf, "\033[?1;2c", 7))
      putenv("TERM=rxvt");

   fcntl(STDIN_FILENO, F_SETFL, stdin_flags);
   if (tty) tcsetattr(STDIN_FILENO, TCSAFLUSH, &tioorig);
}

const char* logpipename = 0;

/*
 * --------------- MAIN -----------------
 */

BOOLEAN waiting = FALSE; /* TRUE if waiting on a command sent to process */
/* %%% Hack: This was a local variable inside main, but com_agent() needs it */

int
main(int argc, const char* argv[])
{
   int c, done = 0;
   char helpscreen='\0';
   char termtype[128]="";
   int 
      infolines = -1,
      clockflg = 1,
      jumpflg = 0,
      errflg = 0,
      verboseflg = 0,
      privateflg = 0,
      keep_bgcolor = 0;
   char junkbuf[1024];

   extern char *optarg;
   extern int optind, optopt;

   setuid(getuid()); /* make sure director can't be setuid root by accident */
   
   {
      struct passwd* pwd;
      pwd = getpwuid(getuid());
      if (pwd!=0) real_username = strdup(pwd->pw_gecos);
   }

   atexit(final_cleanup); /* make sure destructors happen */

   cfht_signal(SIGTERM, (PFV)signal_cleanup);
   cfht_signal(SIGHUP,  (PFV)signal_cleanup);
   cfht_signal(SIGPIPE, (PFV)signal_cleanup);
   cfht_signal(SIGSEGV, (PFV)signal_cleanup); /* should really dump a core! */
   cfht_signal(SIGBUS,  (PFV)signal_cleanup); /* ditto */

   logpipename = getenv("CFHTLOGU"); /* save this before it's clobbered */

   if (gethostname(hostname,80)==-1)
	 strcpy(hostname, "unknown host");
   /*
    * Initialize cfht_log facility.  Duplicate messages go there for
    * easier debugging and tracing.
    */

   putenv("CFHTLOGU=/dev/null"); /* we _are_ the user's feedback window */
   if (!getenv("CFHTLOGS")) putenv("CFHTLOGS=/tmp/pipes/syslog.np");
   if (!getenv("CFHTCONFPATH")) putenv("CFHTCONFPATH=/usr/local/cfht/dev/conf");
   putenv("LINES=0"); putenv("COLUMNS=0"); /* force ncurses not to rely on these */

   cfht_logv( CFHT_MAIN, CFHT_LOG_ID, "director-"VERSION_STRING);

   /*
    * This must come after director's own CFHT_LOG_ID call, because this
    * call will look for the presence of DIRECTOR and cause future cfht_log
    * calls to print to stderr.  For director itself, this doesn't work.
    */
   putenv(DIRECTOR_ENV"="VERSION_STRING); /* pass this on to the cli */

   while ((c = getopt(argc, (char**)argv, ":t:bCcdvji:K:k:mh:Pp:")) != -1)
   {
      switch (c)
      {
	 case 't': sprintf(termtype, "COLORTERM=%.100s", optarg);
                   putenv(termtype);
	           break;
	 case 'b': keep_bgcolor=1; break;
	 case 'C': clockflg=0; break;
	 case 'c': cli_newline="\r"; break;
	 case 'd': debugflg=1; break;
	 case 'v': verboseflg=1; break;
	 case 'j': jumpflg=1; break;
	 case 'i': infolines = atoi(optarg); break;
	 case 'K': who.setkey(atoi(optarg)); break;
	 case 'k': who.setup(optarg); break;
	 case 'm': who.force_master(); break;
	 case 'h': if (strlen(optarg)<2) helpscreen = *optarg; errflg++; break;
	 case 'P': privateflg=1; who.setkey(-1); break;
	 case 'p': ps2 = optarg; break;
         case ':': if (optopt!='h') fprintf(stderr, "-%c needs argument\n", optopt);
	 case '?': errflg++; break;
      }
   }

   /*
    * === Calculate prompt string length now
    */
   ps1len = strlen(ps1);
   ps2len = strlen(ps2);
   psoklen = strlen(psok);
   psfailedlen = strlen(psfailed);
   pscontlen = strlen(pscont);
   
   if (!getenv("PATH"))
   {
      putenv("PATH=.:/usr/local/cfht/bin:/usr/local/bin:/usr/bin:/bin");
      /* === warning: using default path */
   }
   
   /*
    * Always startup running in the user's home directory
    * so director and agents agree on what the cwd is.
    * %%% This messes up things like "director ./directest"
    */
   chdir(who.home());
   sprintf(junkbuf, "HOME=%.500s", who.home());
   putenv(strdup(junkbuf));
   sprintf(junkbuf, "USER=%.500s", who.name());
   putenv(strdup(junkbuf));

   if (!privateflg)
   {
      int len;

      strcpy(junkbuf, "SESSIONHOST=");
      len = readlink(".director/sessionhost", junkbuf+12, sizeof(junkbuf)-13);
      if (len!=-1)
      {
	 junkbuf[len+12]='\0';
	 putenv(strdup(junkbuf));
      }
   }

   if (!getenv("FFTEMPLATE"))
   {
      sprintf(junkbuf,"FFTEMPLATE=%.500s/.,template", who.home());
      putenv(strdup(junkbuf));
   }
   if (!getenv("SESSIONHOST"))
   {
      sprintf(junkbuf, "SESSIONHOST=%s", hostname);
      putenv(strdup(junkbuf));
   }
   if (!getenv("DISPLAY"))
   {
      sprintf(junkbuf, "DISPLAY=%s:0.0", getenv("SESSIONHOST"));
      putenv(strdup(junkbuf));
   }

   if (helpscreen=='e')
      /*
       * Display Enviroment Variable Help Screen
       */
   {
      fprintf(stderr,"\n"
"Environment and director:\n"
"\n"
DIRECTOR_ENV" (this is version `%s')\n"
"  DO NOT SET this externally!  Director sets it to an appropriate value\n"
"  so agents know what version of director they are running under.\n"
"SESSIONHOST (you're running on `%s')\n"
"  If unset, director will set this to the current hostname.  If it is set\n"
"  but it doesn't agree with the current hostname, director will issue a warning\n"
"  and keep the value you have set. File ~/.director/sessionhost overrides this.\n"
"DISPLAY (currently `%s')\n"
"  If unset, director sets this to $SESSIONHOST:0.0.\n"
"USER and HOME (currently `%s' and `%s', uid/roll key is %d)\n"
"  After a userid for a director session has been determined (either that\n"
"  passed with `-k' or getuid() if no `-k') director overrides these two\n"
"  variables to agree with information from the passwd table. The original\n"
"  settings of these variables have no effect on director or its agents.\n"
"PATH (currently `%s')\n"
"  User's current path is sent to agent as-is.\n"
"LINES and COLUMNS\n"
"  Ignored (set to 0 to force ncurses to determine window size on its own.)\n"
"CFHTCONFPATH (currenty `%s')\n"
"  Defaults to /usr/local/cfht/dev/conf.\n"
"CFHTLOGS (currently `%s')\n"
"  Log messages from agents running within director are also written here\n"
"  using cfhtlog() calls.  The message type is determined from the first\n"
"  few charaters in a line printed by the agent.  For example, a string\n"
"  starting with `error:' is written as CFHT_ERROR, etc.  See also `-hs'.\n"
"  Defaults to /tmp/pipes/syslog.np.\n"
"CFHTLOGU (currently `%s')\n"
"  If this file (fifo) does not exist, or if no other process is currently\n"
"  reading from it, it is opened.  Messages written to this node by other\n"
"  programs are echoed in the director feedback area.\n"
"FFTEMPLATE (currently `%s')\n"
"  Passed on to the agent.  If not found, defaults to ~/.,template\n"
"When the agent runs on another host, these are the ONLY variables it\n"
"should expect to have defined.\n"
	      ,getenv(DIRECTOR_ENV)
	      ,getenv("SESSIONHOST")
	      ,getenv("DISPLAY")
	      ,getenv("USER")
	      ,getenv("HOME")
	      ,who.getkey()
	      ,getenv("PATH")
	      ,getenv("CFHTCONFPATH")
	      ,getenv("CFHTLOGS")
	      ,logpipename?logpipename:"not set"
	      ,getenv("FFTEMPLATE"));
      exit(EXIT_FAILURE);
   }

   if (helpscreen=='f')
      /*
       * Display .director/ files help screen
       */
   {
      fprintf(stderr,"\n"
"Files used by director:\n\n"
"%s (see also: director -he)\n"
"  Opened for reading session feedback messages from other sources.\n"
"\n"
"~/.director/sessionhost\n"
"  If this file exists as a symlink, director will first attempt to\n"
"  `remsh' to the machine name given by the contents of the link.\n"
"\n"
"~/.director/startup\n"
"  This file is automatically sourced at startup (non-clones only).\n"
"  Providing an agent on the command line (after --) suppresses this.\n"
"\n"
"~/.director/history\n"
"  Command history is saved here between sessions.  The `~' is always\n"
"  replaced by the home directory of the original user starting director.\n"
"\n"
"~/.director/command.np\n"
"  Non-clones create (and remove at exit) this fifo and read from it.\n"
"  Clones (and other external programs) must only write commands to it.\n"
"  If `-k' is present, clones use the `~' for that account.  `~' expansion\n"
"  at the command prompt also follows the `-k' user after in that case.\n"
,logpipename?logpipename:"CFHTLOGU is currently not set");
      exit(EXIT_FAILURE);
   }

   if (helpscreen=='s')
      /*
       * Display string parsing rules
       */
   {
      fprintf(stderr,"\n"
"String parsing rules: ('^'->beginning of a line)\n\n"
"OUTPUT FROM AGENT:\n"
"^status:     displays green,  logs as CFHT_STATUS\n"
"^logonly:    displays brown,  logs as CFHT_LOGONLY\n"
"^debug:      displays brown,  logs as CFHT_DEBUG\n"
"^warning:    displays yellow, logs as CFHT_WARN\n"
"^error:      displays red,    logs as CFHT_ERROR\n"
"Any line on stderr is at least a CFHT_WARN, so C agents can use perror().\n"
"All other lines are not logged, but display in various areas of the screen:\n"
"^statusbar:  creates status bar (if needed); displays text in reverse video\n"
"^titlebar:   displays text in window title (if TERM=xterm or rxvt only)\n"
"^infoN:      (where N is 1..9) displays on info panel if infosize >= N\n"
"^instant:    displays text in user prompt area until command completes\n"
"^*           displays with highlighted background (including the *)\n"
"^ok>         command completed ok *handled internally by cli_getline()\n"
"^failed>     last command failed  *handled internally by cli_getline()\n"
"\\a(bell)     rings the terminal bell immediately (and eats the character)\n"
"...          all other lines scroll in white in the feedback area\n"
"             long lines are word-wrapped to fit the current window size\n"
"\n"
"INPUT AT PROMPT, COMMAND PROMPT, OR FROM SCRIPT:\n"
" ~ or ~user  basic shell-style tilde-substitution is done by director\n"
"builtin      (where `builtin' is a director builtin command) runs immediately\n"
"other        (where `other' is any other command) put in queue for agent\n"
"@cmd         runs cmd as above but does not echo the command itself to window\n"
"-cmd         runs cmd as above but does not abort a script if this command fails\n"
"{PID}cmd     runs cmd as above; when it returns director signals PID with:\n"
"               SIGUSR1 if command returned PASS (prompt was ^ok>)\n"
"               SIGUSR2 if command returned FAIL (prompt was ^failed>)\n"
"               *handled internally by cli_cmd() or the shell utility `clicmd'\n");
      exit(EXIT_FAILURE);
   }

   if (errflg)
   {
      fprintf(stderr,"\n"
"usage: %s [director options] [ -- [rhost:]agent [agent options] ]\n"
"or     %s [director options] -- [{9600,N81,see_manual...}]serialdevice\n"
"or     %s           (no startup and no args, just tails log pipe)\n"
"or     %s -k <user> (attach to another user's session without su'ing)\n"
"\n"
"agent options are passed without change to the actual cli program\n"
"\n"
"director options are:\n"
"    -i <n>    displays <n> info lines at bottom of screen (default is 0)\n"
"    -j        enables jumpscroll mode for slow terminals\n"
"    -v        turns on verbose mode which shows logonly and debug\n"
"    -d        turns on debug messages (also requires -v to see on screen)\n"
"    -m        used by sessions to ensure they get a master, not a clone\n"
"    -c        sends a CR character after each command instead of a newline\n"
"    -C        turn off the titlebar clock for slow connections\n"
"    -t <type> override terminal auto-sensing and TERM var (e.g., -t vt100)\n"
"    -P        starts a private director (can't be cloned, no CFHTLOGU...)\n"
"    -p <str>  sets the prompt string to expect after each command completes\n"
"    -h        displays this help information\n"
"    -hf       displays summary of director-related files and resources\n"
"    -hs       displays string parsing rules used on output from realprogram\n"
"    -he       displays summary of how the environment affects director\n"
"              (hint: -he can be combined with -k<user> for testing)\n"
"\n"
"All options above can also be controlled by director builtin commands.\n"
"Type `builtin' at the director command prompt for more help on these.\n"
"Type `license' to read information on (lack of) warranty and copying.\n"
"\n"
	      ,argv[0], argv[0], argv[0], argv[0]);
      exit(EXIT_FAILURE);
   }

   /*
    * Don't count on an ncurses that checks for COLORTERM.  Instead,
    * Just set TERM equal to COLORTERM before initializing curses.
    */
   if (getenv("COLORTERM") && *getenv("COLORTERM"))
   {
      static char tmp[128];
      sprintf(tmp, "TERM=%.100s", getenv("COLORTERM"));
      putenv(tmp);
   }
   else
   {
      /*
       * If all we have to go on is the TERM variable, try and autodetect
       * if this is an rxvt window.  Rxvt should reply with ESC [ ? 1 ; 2 c
       * when given a ^E.
       */
      autosense_rxvt();
   }

   if (strcmp(getenv("SESSIONHOST"), hostname)!=0)
   {
      /*
       * Remote shell to another machine if sessionhost is different
       * from the current host.
       */

      fprintf(stderr,
	      "Please wait; Switching to the session host . . .\n"
	      "(Note: logging in to `%s' manually first may work better.)\n",
	      getenv("SESSIONHOST"));

      sprintf(junkbuf, "CFHTLOGU=%.900s", logpipename?logpipename:"/dev/null");
      putenv(strdup(junkbuf));

      cli_runon(getenv("SESSIONHOST"), /*rawtty=*/1, argc, argv);
      exit(EXIT_FAILURE);
   }
	      
   /*
    * Create a roll buffer which will hold the last N lines echoed
    * to the terminal.  uid is used as a roll key so other director
    * clones will be able to attach to a shared memory segment with
    * that key and view the same log messages.
    */

   roll = new Roll(who.getkey(), who.clone_mode()); /* === shmdt on exec? */

#if 0
   if (!roll->is_clone() && strcmp(getenv("SESSIONHOST"), hostname)!=0)
   {
      sprintf(junkbuf, "warning: SESSIONHOST (%.100s) does not match `%.100s'",
	      getenv("SESSIONHOST"),
	      hostname);
      roll->add_line(junkbuf);
   }
#endif

   /*
    * constructor for this class creates a node in the file
    * system and begins accepting commands from other programs
    * through the named pipe.  If pipename is NULL, it is not
    * an error.  It just means we aren't using the pipe feature.
    */

   if (!privateflg)
   {
      char pipename[512];

      sprintf(pipename, "%s/.director", who.home());
      mkdir(pipename, 00755);
      strcat(pipename, "/command.np");

      commandPipe = new CommandPipe(pipename, /*wronly=*/roll->is_clone());
      /* === set close on exec flag? */

      if (!roll->is_clone() && commandPipe->get_fd()==-1)
      {
	 cfht_logv(CFHT_MAIN, CFHT_ERROR, "Can't open command pipe `%s'",
		   pipename);
	 fprintf(stderr,
"Can't open command pipe.  Perhaps a copy is running on another machine?\n"
"If not, manually remove the file `%s' and try again.\n",
		 pipename);
	 exit(EXIT_FAILURE);
      }
   }

   if (!privateflg && !roll->is_clone())
   {
      struct stat stbuf;

      if (lstat(".director/sessionhost", &stbuf)==-1 && errno==ENOENT)
      {
	 roll->add_line("logonly: creating sessionhost file");
	 if (symlink(getenv("SESSIONHOST"), ".director/sessionhost")==0)
	    sessionhost_created = 1;
	 /*
	  * Set a flag so we remember to clean it up too
	  */
	 else
	 {
	    roll->add_line("warning: trouble creating sessionhost link");
	 }
      }
   }
	 
   /*
    * Ignore the setting of CFHTLOGU if this is a clone
    */
   if (!privateflg && !roll->is_clone() && 
       logpipename && *logpipename && strcmp(logpipename,"/dev/null")!=0)
   {
      logPipe = new LogPipe(logpipename);

      if (logPipe->get_fd()==-1)
      {
	 cfht_logv(CFHT_MAIN, CFHT_ERROR, "Can't read log pipe `%s'",
		   logpipename);
	 fprintf(stderr,
"Can't use log pipe.  Is someone else already reading it?\n"
"If not, manually remove the file `%s' and try again.\n"
"Or, to disable CFHTLOGU, unset or set to empty $CFHTLOGU.\n",
		 logpipename);
	 exit(EXIT_FAILURE);
      }
   }

   /*
    * the following forks and actually starts up the _real_
    * underlying process that commands will be piped to
    */

   if (optind<argc && !roll->is_clone())
      process = new ProcessPipe(&argv[optind]);

   /*
    * The variable `infolines' will contain a -1 unless the user
    * started director with -i#.  If it still contains -1, check if
    * there is a hint saved in shared memory for the new size.
    */
   if (infolines==-1) infolines = roll->get_infosize();
   roll->set_infosize(infolines);

   /*
    * constructor for this class starts up curses in the terminal window
    */
   curs = new Curs(infolines, keep_bgcolor);

   com_clock(clockflg?"on":"off");
   com_verbose(verboseflg?"on":"off");
   com_debug(debugflg?"on":"off");
   curs->clone(roll->is_clone());

   curs->titlename(who.name());
   if (optind<argc && !roll->is_clone())
   {
      sprintf(junkbuf, "titlebar: %.500s", argv[optind]);
      roll->add_line(junkbuf);
   }

   if (jumpflg) curs->jumpscroll();

   /*
    * Install a signal handlers to catch window size change events,
    * and control C/Z/| at the keyboard
    */
#if defined(SIGWINCH) && (SIGWINCH!=0)
   cfht_signal(SIGWINCH, CFHT_SIG_TRAP);
#endif
   cfht_signal(SIGINT, CFHT_SIG_TRAP);
   cfht_signal(SIGTSTP, CFHT_SIG_TRAP);
   cfht_signal(SIGCHLD, CFHT_SIG_TRAP);
   cfht_signal(SIGQUIT, CFHT_SIG_TRAP);

   tzset();

   /*
    * Source a startup script if needed
    */
   if (optind>=argc && !roll->is_clone() && !privateflg)
   {
      struct stat stbuf;
      
      if (stat(".director/startup", &stbuf)==0)
      {
	 scripter.add(".director/startup");
      }
   }
   
   /*
    * Start an interval timer that delivers SIGALRM at a steady
    * rate of 0.5Hz.  We use this to ensure that `seconds' on the
    * clock are really `seconds'.  On side-effect is that director
    * cannot use alarm() or sleep() without de-installing this
    * interval timer since they use the same signal.
    */

   cfht_logv( CFHT_MAIN, CFHT_START, "initialized; entering main event loop");

   set_interval_timer(1);

   /*
    * ------------------- INPUT LOOP ----------------------
    */

   if (roll) if (!roll->is_clone())
   {
      roll->add_line("------------------------------------------------------------------------");
      roll->add_line("                  Director CFHT version "VERSION_STRING" ready.");
      /*
       * OR,
       * If you modify the code, edit this line so the program clearly indicates
       * this on startup.  (Or, call it something other than director, but do not
       * remove the original Copyright from derivative works.
       roll->add_line("Director version "VERSION_STRING" with modifications by ...");
       */
      roll->add_line("Copyright 1998 Canada-France-Hawaii Telescope, <daprog@cfht.hawaii.edu>");
      roll->add_line("Distributed under the GNU Public License 2.  Type `license' for details.");
      roll->add_line("------------------------------------------------------------------------");
   }
   else if (commandPipe)
   {
      /*
       * Announce it when someone joins the session
       */
      sprintf(junkbuf,
	      "say_ * %s has started a clone window.",
	      real_username);
      commandPipe->send(junkbuf); /* === krusty */
   }

   while (!done)
   {
      const char* curline;
      FDSet readfds, exceptfds;

      if (process)
      {
	 readfds.add_fd(process->get_fd());
	 readfds.add_fd(process->err.get_fd());
	 /*
	  * This requires that process->pend() return either PASS or FAIL
	  * if the agent has died or exited.
	  */
	 if (waiting==FALSE && readfds.get_fdnum()==0)
	 {
	    flush_inputs();
	    scripter.cancel();
	    pendBuffer.cancel();

	    if (autorestart && autorestart > RESTART_THRESH)
	    {
	       roll->add_line("error: director: agent exited; autorestart in progress");
	       flush_to_screen();
	       do_sleep(5);
	       com_agent("restart");
	       autorestart -= RESTART_THRESH;
	    }
	    else
	    {
	       if (autorestart)
	       {
		  roll->add_line("error: agent dying too quickly; turning off autorestart");
		  autorestart = 0;
		  if (who.clone_mode()==MASTER_ONLY)
		  {
		     roll->add_line("error: restart it manually with `agent restart'");
		     flush_to_screen();
		  }
	       }
	       if (who.clone_mode()!=MASTER_ONLY)
	       {
		  roll->add_line("logonly: agent program has exited");
		  flush_to_screen();
		  exit(EXIT_SUCCESS);
	       }
	    }
	 }
      }
      if (commandPipe) readfds.add_fd(commandPipe->get_fd());
      if (logPipe) readfds.add_fd(logPipe->get_fd());

      if (roll->is_clone()==-1)
      {
	 fprintf(stderr, "clone exiting because parent exited\n");
	 exit(EXIT_SUCCESS);
      }

      readfds.add_fd(STDIN_FILENO);
      exceptfds.add_fd(STDIN_FILENO);
      exceptfds.add_fd(STDOUT_FILENO);

      /*
       * Now wait for input or interruption from a signal.
       *
       * Skip the select though if we are reading from a script
       * and the agent is reading for the next command (->director
       * not waiting for command to complete, scripter is waiting
       * to send commands.)
       */
      if (!scripter.waiting() || waiting)
      {
	 int rtn;

	 cfht_signal_unblock(SIGALRM);
	 if ((rtn=SELECT(readfds.get_fdnum(), readfds.get_fdsetp(), 0, exceptfds.get_fdsetp(), 0))<0
	     && errno!=EINTR)
	    perror("select");
	 cfht_signal_block(SIGALRM);

	 if (debugflg)
	 {
	    sprintf(junkbuf, "rtn:%d stdin-[%c] stdout-[%c] stderr-[%c] command-[%c] log-[%c]",
		    rtn, 
		    exceptfds.is_set(STDIN_FILENO)?'X':(readfds.is_set(STDIN_FILENO)?'*':' '),
		    (!process||(process->get_fd()==-1))?'-':
		    (exceptfds.is_set(process->get_fd())?'X':(readfds.is_set(process->get_fd())?'*':' ')),
		    (!process||(process->err.get_fd()==-1))?'-':(readfds.is_set(process->err.get_fd())?'*':' '),
		    !commandPipe?'-':(readfds.is_set(commandPipe->get_fd())?'*':' '),
		    !logPipe?'-':(readfds.is_set(logPipe->get_fd())?'*':' '));
	    curs->titlebar(junkbuf);
	 }
	 else if (cfht_signal_event(SIGALRM))
	 {
	    if (autorestart && autorestart<=2*RESTART_THRESH) autorestart++;
	    curs->titlebar(); /* this updates the clock */
	 }

	 if (feof(stdin))
	 {
	    roll->add_line("warning: eof on stdin");
	    exit(EXIT_FAILURE);
	 }

	 if (rtn>0 &&
	     (exceptfds.is_set(STDIN_FILENO) ||
	      exceptfds.is_set(STDOUT_FILENO)))
	 {
	    roll->add_line("warning: i/o exception; cleaning up");
	    exit(EXIT_FAILURE);
	 }
      }
      else if (debugflg)
	 curs->titlebar("scripter running");

      /*
       * Instead of checking which bits are set in readfds,
       * or checking the return value of select, we just
       * check all possible events that this loop should handle
       * each time, and do all reads as non-blocking i/o.
       */

      /*
       * If the child process has anything to tell us,
       * scroll it up in the feedback window.
       */

      curline="";

      flush_inputs();

      /*
       * If control-Z was pressed at the terminal, pass it on ...
       */
      if (cfht_signal_event(SIGTSTP))
	 if (curs->confirm("Are you sure you want to send a stop? (Y/N)"))
	    com_sigsend("TSTP");
      /*
       * If control-C was pressed at the terminal, pass it on ...
       */
      if (cfht_signal_event(SIGINT))
	 if (curs->confirm("Are you sure you want to send an abort? (Y/N)"))
	    com_sigsend("INT");
      
      /*
       * If control-| was pressed at the terminal, pass it on ...
       */
      if (cfht_signal_event(SIGQUIT))
	 if (curs->confirm("Are you sure you want to send a quit? (Y/N)"))
	    com_sigsend("QUIT");
      
      /*
       * If a window change size event has occurred...
       */
#if defined(SIGWINCH) && (SIGWINCH!=0)
      if (cfht_signal_event(SIGWINCH))
      {
	 if (!roll->is_clone())
	    roll->add_line("logonly: window size change requested...");
	 curs->resize();
	 roll->rewind(LINES);
      }
#endif
      
      /*
       * If the child exited...
       */
      if (cfht_signal_event(SIGCHLD))
      {
	 const char* s = process?process->postmortem():0;

	 if (s) roll->add_line(s);
      }

      /*
       * check_line returns a pointer to a string if
       * a complete command line has been entered, otherwise
       * it returns 0.
       *
       * This should really be a while loop, but then `waiting'
       * will be FALSE all the time when reading from a script
       * and I want it to not read any further in the file than
       * the current line that is executing.
       */
      if ((curline=curs->check_line())!=0 ||
	  (curline=commandPipe?commandPipe->check_line():0)!=0 ||
	  (!waiting &&
	   (curline=scripter.check_line())!=0))
      {
	 const char* endbrace = 0;
	 const char* send_curline;
	 int killpid = 0;
	 int no_echo = 0;
	 int no_fail = 0;
	 
	 sprintf(junkbuf,"debug: parsing new command: `%.800s'", curline);
	 roll->add_line(junkbuf);

	 while (*curline!='\0' && *curline<=' ') curline++; /* eat white space */

	 /* check for `{PID}command' notation */

	 if (*curline=='{' &&
	     (endbrace=strchr(curline,'}'))!=0)
	 {
	    killpid = atoi(curline+1);
	    curline = endbrace + 1;
	 }
	 
	 curline = do_tilde_substitution(curline);

	 send_curline = curline;

	 if (*curline=='@') { no_echo=1; curline++; }
	 if (*curline=='-') { no_fail=1; curline++; }
	 if (*curline=='@') { no_echo=1; curline++; }

	 /* ---------------------------------------------------------- */

	 /*
	  * Specially handled commands (aka ``hacks'')
	  */

	 /*
	  * `detach' can be used if the agent is completely hosed and
	  * you want to force director to exit.  It tries to kill the
	  * agent before exiting if it is not a clone.
	  */

	 if (strcmp(curline, "detach")==0 ||
	     strcmp(curline, "masterdetach")==0)
	 {
	    if (who.clone_mode()==MASTER_ONLY &&
		strcmp(curline, "detach")==0)
	    {
	       roll->add_line("warning: Cannot detach master.  Exit the window session to log off.");
	       return_to_clicmd(killpid, FAIL);
	       continue;
	    }
	    else
	    {
	       get_out = 1;
	       return_to_clicmd(killpid, PASS);
	       exit(EXIT_SUCCESS);
	    }
	 }

	 /*
	  * `exit' is handled like `detach' if you are a clone.  Even if
	  * not a clone, we still want to signal a {PID} waiting that the
	  * command was completed, because otherwise exit never returns.
	  */

	 if ((strcmp(curline, "exit")==0 ||
	      strncmp(curline, "exit ", 5)==0))
	 {
	    if (who.clone_mode()==MASTER_ONLY)
	    {
	       roll->add_line("warning: Cannot exit master.  Exit the window session to log off.");
	       return_to_clicmd(killpid, FAIL);
	       continue;
	    }
	    get_out = 1;
	    return_to_clicmd(killpid, PASS);
	    if (!process && roll->is_clone())
	       exit(EXIT_SUCCESS);
	 }

	 /*
	  * `cd' is handled by director AND the agent so they both
	  * agree on what the current working directory is.
	  * === Warning: director will get lost (and not complain about it)
	  * if you cd to a directory that exists only on the agent machine.
	  * === need to move, so it gets executed at the right time in queue.
	  */

	 if ((strcmp(curline, "cd")==0 ||
	      strncmp(curline, "cd ", 3)==0))
	 {
	    const char* newdir = curline + 2;
	    
	    while (*newdir && *newdir<=' ') newdir++;
	    if (*newdir) chdir(newdir);
	 }

	 /* ---------------------------------------------------------- */

	 /*
	  * Class 1A Commands:	Instant commands that affect the appearance
	  *                  	of director itself.  Handled locally, even
	  *                  	for clones.
	  */

	 if (cli_search(curline, local))
	 {
	    PASSFAIL result;

	    cli_init("director", local, 0);
	    roll->add_line("debug: executing director-local command");
	    result = cli_execute(curline);
	    return_to_clicmd(killpid, result, no_fail);
	    if (result==FAIL && !no_fail)
	    { scripter.cancel(); pendBuffer.cancel(); }
	 }
	 else if (send_ok())
	 {
	    /*
	     * Class 1B Commands:	Instant commands that affect the agent.
	     *                    	Must be sent to remote director if this
	     *                    	is a clone.
	     */

	    if (!roll->is_clone() && 
		cli_search(curline, remote))
	    {
	       PASSFAIL result;

	       if (!no_echo &&
		   strncmp(curline, "say", 3)!=0)
	       {
		  sprintf(junkbuf, "> %s", curline);
		  roll->add_line(junkbuf);
	       }
	       cli_init("director", remote, 0);
	       roll->add_line("debug: executing director-remote command"); /* %%% what about clones */
	       result = cli_execute(curline);
	       return_to_clicmd(killpid, result, no_fail);
	       if (result==FAIL && !no_fail)
	       { scripter.cancel(); pendBuffer.cancel(); }
	    }
	    else
	    {
	       /*
		* Class 2 Commands:	queued in order using the pendBuffer
		*/

	       if (process)
	       {
		  if (process->commands_pending>=0 &&
		      !scripter.waiting())
		     curs->pending(curline);
		  
		  pendBuffer.add_line(curline, killpid, no_echo, no_fail);
	       }
	       else if (roll->is_clone())
	       {
		  /*
		   * Clones must get the command to the parent's pendBuffer instead
		   */
		  if (commandPipe) commandPipe->send(send_curline);
	       }
	       else
	       {
		  sprintf(junkbuf, "* %s", curline);
		  roll->add_line(junkbuf);
		  roll->add_line("status: No process; type `builtin' for help on director");
		  return_to_clicmd(killpid, FAIL, no_fail);
		  if (!no_fail)
		  { scripter.cancel(); pendBuffer.cancel(); }
	       }
	    }
	 }
	 else curs->print("clone: must be enabled with valid password first.");
      }

      {
	 /*
	  * This chunk of code makes sure only one command is
	  * sent at a time.
	  */

	 if (waiting==TRUE)
	 {
	    PASSFAILBUSY result = process?process->pend():BUSY;
	    
	    if (result!=BUSY)
	    {
	       roll->instant_off();

	       if (result==PASS)
		  roll->add_line("debug: got successful return from process");
	       else
		  roll->add_line("debug: got FAIL from process");

	       if (pendBuffer.check_pid()!=0)
	       {
		  return_to_clicmd(pendBuffer.check_pid(), result, pendBuffer.no_fail());
		  pendBuffer.reset_pid();
	       }

	       if (result==FAIL && !pendBuffer.no_fail())
	       { scripter.cancel(); pendBuffer.cancel(); }

	       waiting = FALSE;
	    }
	 }

	 if (waiting==FALSE && process && process->running())
	 {
	    if ((curline = pendBuffer.check_line()))
	    {
	       curline = do_tilde_substitution(curline);
	       roll->add_line("debug: sending line to process");
	       if (process->send(curline)==PASS) waiting = TRUE;
	    }
	 }
      }

      flush_to_screen();

   } /* while (!done) */

   /*
    * The destructor for curs runs here and restores the
    * terminal settings when the program exits.
    */

   return EXIT_SUCCESS;
}
