/*                                             -*- c-file-style: "Ellemtel" -*-

`CliQueue.hh' - queue/stack data structure to handle nested commands

This file is part of version 3 of the DIRECTOR cli wrapper.
Read the `License' file for terms of use and distribution.
Copyright 2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#ifndef _INCLUDED_CliQueue
#define _INCLUDED_CliQueue 1

#include <sys/types.h>
#include <string.h>
#include <unistd.h>

#include "cli/cli.h"

class ProcessPipe;

typedef enum { SOURCE_NONE, SOURCE_NEED_TO_LOAD, SOURCE_PASS, SOURCE_FAIL } CliSource;
typedef enum { OUTPUT_DIRECTOR, OUTPUT_CAP, OUTPUT_DUP } CliOut;

class CliQueueNode
{
   protected:
      ProcessPipe* process_;
      CliQueueNode * next_, * prev_, * next_stack_, * prev_stack_;
      char* line_orig_;	/* The command line as entered "[handle.][@-]command [args...]" */
      char* line_;	/* Same as above but with optional handle. and `@' `-' removed */
      char* handle_;	/* Handle, without the ., or empty string if no specific handle */
      char* command_;	/* Same as line_, but no arguments */
      char* exit_message_;
      pid_t pid_;	/* PID that wants a signal when command complete */
      int fd_capture_;	/* Capture output of command? -1 -> director window */
      int no_wait_;	/* 0->wait for *all* agents to finish whatever they're doing */
      			/* 1->Execute command as soon as possible */
      int no_echo_;
      int no_fail_;
      int shell_escape_;
      int script_dir_; /* %%% Hack flag to tell source/shell_escape that this binary is
			* in $HOME/SCRIPTDIR.  Goes away with version 4.0's scripter agent.
			*/
      int level_;
      CliOut output_;
      CliSource source_;
      PASSFAIL result_;

   public:
      CliQueueNode(const char* newline, int level=0, pid_t pid=0);
      /*
       * Constructor requires at least the command line.
       * Optionally, you can also supply a desired DIRECTOR_LEVEL
       * and a process id that would like notification (either
       * SIGUSR1 or SIGUSR2 depending on whether the command passed
       * or failed/was aborted.
       */

      ~CliQueueNode(); 
      /*
       * The destructor frees memory associated with this node
       * and sends a signal to another pid if one was waiting on
       * the completion of this particular command.  The result
       * is assumed to be FAIL, unless the .pass() method is called
       * sometime before the destructor.
       *
       * Command flagged with "no_fail" are an exception.  They
       * are usually preceded by a `-' character on the command line,
       * indicating director should ignore failures, like the "make"
       * command does if given the same syntax.
       */
      
      void set_exit_message(const char* msg);
      /*
       * Call this to set a message that will print when the
       * command completes (used for "source") === ugly
       */

      void assign(ProcessPipe* p) { process_ = p; }

      void pass()		{ result_ = PASS; }
      /*
       * Set a flag indicating this command has been completed successfully
       * just before `delete'ing it.  If the command failed, just call
       * delete without calling .pass() first.
       */

      int level()		{ return level_; }
      /*
       * The DIRECTOR_LEVEL at which this command was entered into the
       * queue.
       */

      const char* line()	{ return line_; }
      /*
       * The command line, as entered, but without the special `@' and `-'
       * characters at the beginning of the line, and without the optional
       * agent "handle." that can appear at the beginning of the command.
       */
      
      const char* command() 	{ return command_; }
      /*
       * Same as line() but without optional extra command line arguments.
       */

      const char* handle() 	{ return handle_; }
      /*
       * Returns the "handle" or an empty string if no specific handle was
       * given with this command.
       */

      const char* line_orig()	{ return line_orig_; }
      /*
       * Returns "[handle.]command"
       */

      int no_wait() 		{ return no_echo_; }
      int no_fail()		{ return no_fail_; }
      int no_echo() 		{ return no_echo_; }
      int fd_capture() 		{ return fd_capture_; }
      int script_dir() 		{ return script_dir_; }
      int shell_escape() 	{ return shell_escape_; }
      CliOut output() 		{ return output_; }

      void no_echo_clear()      { no_echo_ = 0; }
      void fd_capture_close()   { if (fd_capture_!=-1)
	 	                    close(fd_capture_);
                                  fd_capture_ = -1;
				  output_ = OUTPUT_DIRECTOR; }
      friend class CliQueue;
};


class CliQueue
/*
 * This class implements "stacked queues".  check_line() always
 * returns the first-in command off the last-in stack level.
 * New commands are put at the bottom of the queue, in the
 * level given by $DIRECTOR_LEVEL in the evironment.  This way
 * commands in the queue that cause other commands to be fed
 * back together will be executed in the correct order.  See
 * director-flow.ps in the ../DOCS subdirectory for a diagram.
 */
{
   private:
      CliQueueNode* head;
      int level_;
      void builtin_source(CliQueueNode* p);
      /*
       * The "source" command is hardcoded into CliQueue.  The
       * caller will never see the "source" command itself; only
       * the commands that it expanded into.
       */

   public:
      CliQueue() { head = 0; level_ = -1; }
      ~CliQueue() { cancel(); }

      void add_line(CliQueueNode* newNode);
      /*
       * === Update the text below:
       * Add a command line given in the string `s' to queue `level'
       * (the top-level queue, 0, if omitted) and set flags indicating
       * whether any particular `pid' wants to be notified when the
       * command completes, whether the command itself should echo to
       * the screen when it comes up for execution, and whether failure
       * of this command should be ignored (i.e., should not abort a script)
       */

      CliQueueNode* check_line();
      /*
       * Returns NULL if all commands have been processed, otherwise
       * returns the first-in command off the last-in stack level.
       * "source" commands are not returned.  Instead CliQueue internally
       * expands them and returns each of their individual lines.
       * The caller should pass the node back to CliQueue when it
       * has completed executing and either remove() it or cancel() it.
       */
      
      int level() { return level_; }
      /*
       * Returns the deepest existing level still waiting to execute,
       * or -1 if no more commands (i.e., when check_line would return NULL)
       */
      
      void remove(CliQueueNode* p); /* %%% This was private */
      /*
       * This gets called automatically when a node is to be deleted.
       * It updates the linked list, prints a "completed" message (if any--
       * currently only used for "source" commands to tell user when
       * it is leaving a batch file) and frees memory associated with
       * the node.
       *
       * Some sanity checks are done on the integrity of the command
       * queue at the same time.
       */

      int cancel(CliQueueNode* p = 0);
      /*
       * Call this whenever failure of a command should cause
       * p's queue level to be aborted, or when the user hits
       * ^C.  The value returned in the number of commands that
       * were trashed from the queue.  With nested levels of queues,
       * this will typically causes the level above to get a FAIL
       * back from the sublevel that was trashed, which will in turn
       * cause .cancel to be called again, but if you really want
       * to be sure *all* commands are canceled, omit the node and
       * the entire command tree will be cleared.
       */
      
      char state();
      
      void dump(int logtype = CLI_LOG_VERBOSE, CliQueueNode* p = 0, int indent = 0);
      /*
       * For debugging, print some kind of representation of
       * self to the roll.
       */
};

#endif /* !_INCLUDED_CliQueue */
