/*                                             -*- c-file-style: "Ellemtel" -*-

`Pipes.hh' - reads line-by-line from pipe or fifo sources

This file is part of version 3 of the DIRECTOR cli wrapper.
Read the `License' file for terms of use and distribution.
Copyright 2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#ifndef _INCLUDED_Pipes
#define _INCLUDED_Pipes 1

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>

#include "CliQueue.hh"
#include "director.hh"
#include "cli/cli.h"

typedef int PASSFAILBUSY;
#define BUSY (PASS*2 + FAIL*3) /* === argh.  fix this */

class CliQueueNode;

class FDSet
{
/*
 * Manages file descriptor sets for select.
 * It is acceptable to pass a -1 to as the fd to the
 * functions below and they will ignore that particular
 * file descriptor
 */
   private:
      fd_set fdset;
      int fdnum;

   public:
      FDSet() { fdnum=0; FD_ZERO(&fdset); }
      FDSet(const FDSet& f) { *this = f; }
      FDSet& operator=(const FDSet& f) { fdset = f.fdset; fdnum = f.fdnum; return *this; }

      void add_fd(int fd) { 
	 if (fd!=-1)
	 {
	    if (fdnum<=fd) fdnum = fd+1; FD_SET(fd, &fdset);
	 }
      }

      int get_fdnum() { return fdnum; }
      fd_set* get_fdsetp() { return &fdset; }
      
      int is_set(int fd) { if (fd==-1) return 0; else return FD_ISSET(fd, &fdset); }
};

/* --------------------------------------------------- */

#define CMD_MAXLEN 1023

class LineBuffer
/*
 * This class reads any pending input from a file descriptor
 * each time "check_line()" is called.  The fd passed to set_fd
 * will be set to non-blocking i/o, so that each call to
 * check_line() will immediately return either
 *
 *  0 if no complete line has been read yet, or
 *  a character pointer to a complete string, with the \n stripped.
 */
{
   private:
      int fd;
      BOOLEAN found_prompt;
      int mtype;
      char* id;
      char* prompts;
      char buf[2][CMD_MAXLEN+2]; /* +2 for newline hack */
      char* bufp;
      int bufnum;
      char lastnewline;
      PASSFAILBUSY* resultp;
      unsigned long* timeoutp;
      int timeout_restart_;
      int timeout_initial_;

  protected:
      /*
       * Diagnostic stuff for logging errors and debug messages.
       */
      void log_error(const char* msg);
      void log_debug(const char* msg);

   public:
      LineBuffer();
      ~LineBuffer();

      void set_id(const char* p) { if (id) free(id); id = strdup(p?p:""); }
      const char* get_id(void) { return id?id:"[No ID]"; }
      void set_mtype(int default_mtype) { mtype = default_mtype; }
      int  get_mtype(void) { return mtype; }
      void set_prompts(const char* p) { if (prompts) free(prompts); prompts = strdup(p?p:""); }
      void set_resultp(PASSFAILBUSY* p) { resultp = p; }
      void set_fd(int newfd, int non_blocking=1) {
	 fd = newfd;
	 if (fd!=-1 && non_blocking)
	    if (fcntl(fd, F_SETFL, fcntl(fd, F_GETFL, 0)|O_NONBLOCK)==-1)
	    {
	       /* %%% Print to roll? */
	       fprintf(stderr, "can't set non-blocking i/o for fd %d (`%s')\n",
		       fd, get_id());
	       perror("fcntl");
	    }
      }
      int get_fd() { return fd; }

      /*
       * Timeout-related stuff.  This is for agents that do not have a
       * consistent recognizable prompt for director to detect.  Instead,
       * two timeouts are used.
       *
       * The initial timeout determines how long director will wait for
       * data to begin flowing before it assumes the the command completed
       * with no response at all.  Set this to something fairly high if
       * all commands are expected to return at least one character in response.
       *
       * The second timeout sets the longest acceptable pause in the data,
       * once it begins flowing.  This can be fairly low if the response
       * is always sent in one contiguous chunk.
       */
      int timeout_isset() { 
	 return (timeout_initial_ != 0);
      }

      void timeout_reinit() {
	 if (timeout_isset())
	    *timeoutp = director_timer + (timeout_initial_ * director_timer_hz) / 1000;
      }

      void timeout_init(unsigned long* tp, int initial, int restart) {
	 if (tp==0)
	 {
	    fprintf(stderr, "error: Pipes.hh timeout_init() NULL tp for `%s'!\a\n", get_id());
	    return;
	 }

	 timeoutp = tp;
	 timeout_restart_ = restart;
	 timeout_initial_ = initial;
	 timeout_reinit();
      }

      void timeout_restart() {
	 if (timeout_isset())
	    *timeoutp = director_timer + (timeout_restart_ * director_timer_hz) / 1000;
      }

      int timeout_expired() {
	 if (timeout_isset() && director_timer > *timeoutp) return 1; else return 0;
      }

      const char* check_line(int counteract_pagers = 0);
};


class CommandPipe : public LineBuffer
{
   private:
      int command_fd;
      int autoremove;
      
   public:
      CommandPipe(const char* name, int wronly=0);
      ~CommandPipe();
      void send_line(const char* command);

};

class LogPipe : public LineBuffer
{
   private:
      int log_fd;
      int autoremove;

   public:
      LogPipe(const char* name);
      ~LogPipe();
};

typedef enum { PIPE_NOT_USED, PIPE_TO_DEVICE, PIPE_TO_PROCESS }
ProcessType;

typedef struct
{
      const char* line;
      const char* handle;
      int pid;
      int mtype;
} process_line_t;

class ProcessPipe : public LineBuffer
{
   private:
      CliQueueNode* command_;
      ProcessType type;

      time_t autorestart_;
      /*
       * autorestart_ is normally set to 0, which means this process
       * should not be restart automatically if it dies or exits.
       *
       * When the autorestart feature is turned on for an agent,
       * the current time is stored in this variable.
       *
       * When an agent needs to be restarted, and if the timestamp meets
       * the criterion described below, the timestamp is:
       *  1. Brought up to the current time if it is in the past.
       *  2. The max. restart period is added to the time stamp.
       *
       * An agent that steadily dies at some rate just below the
       * max. restart period will always have a timestamp somewhere
       * between "now" and max. restart in the future.  An agent
       * that is dying off too rapidly will get a timestamp moving
       * further and further into the future.  Once it gets to
       * more than 5X the max. restart period, autorestart_ gets
       * set to "1", a special value meaning restart has been
       * temporarily turned off until the next manually issued
       * "agent restart" enables it again.
       */

      int default_agent_;
      int child_input[2];
      int child_output[2];
      int child_stderr[2];
      pid_t child;
      int signalmap[NSIG]; /* different hosttypes assign signals differently! */
      char** argv_;
      unsigned long command_timeout;
      PASSFAILBUSY command_result; /* %%% Make sure no UMRs here! */
      void cleanup();
      void log(const char* type, const char* msg);

   protected:
      ProcessPipe* next_;

   public:
      LineBuffer err;

      ProcessPipe(const char** argv, int director_level,
		  const char* handle, const char* prompts,
		  int timeout_init, int timeout_sust,
		  int default_mtype,
		  int default_agent, int autorestart);

      ~ProcessPipe();

      PASSFAIL restart(int accept_input = 1, int director_level = 1000);

      PASSFAIL signal(int signo);

      BOOLEAN running() {
	 if (get_fd()==-1 && err.get_fd()==-1 &&
	     type==PIPE_TO_PROCESS && child==-1)	return FALSE;
	 else						return TRUE;
      }

      pid_t childpid() { return child; }
      int default_agent() { return default_agent_; }

      PASSFAIL send_command(CliQueueNode* command);
      void assign_command(CliQueueNode* command) { command_ = command; }
      void remove_command() { command_ = 0; }

      PASSFAILBUSY pend();
      /*
       * Returns PASS: if the last command completed successfully (prompt was "ok>")
       *         FAIL: if the last command failed (prompt was "failed>")
       *         BUSY: if the last command is still executing. (no prompt yet)
       */
      
      const char* postmortem();
      /*
       * tries to conduct an autopsy on an exited process and
       * returns a string describing what happened, or 0 if the
       * process is apparently still running.
       */

      process_line_t* check_line();
      /*
       * Krusty: prescans LineBuffer's check_line() for "signalmap:" directives!
       */

      const char* handle() { return get_id(); }
      /*
       * === document "handle.command" ?
       */

      void dump(int logtype = CLI_LOG_VERBOSE);

      friend class ProcessList;
};


class ProcessList
{
   private:
      ProcessPipe* head_;

   public:
      ProcessList() { head_ = 0; }

      ~ProcessList() { killall(); }

      void add_read_fds( FDSet& fds );

      int killall(const char* handle = 0);
      
      PASSFAIL assign(CliQueueNode* command);

      PASSFAIL signal(int signo);
      
      char state();

      void cleanup(ProcessPipe* force_exit = 0);

      ProcessPipe* autorestart();

      ProcessPipe* exec(const char* argv[],
			int director_level,
			const char* handle = "",
			const char* prompts = "",
			int timeout_init = 0,
			int timeout_sust = 0,
			int default_mtype = -1,
			int default_agent = 0, /* %%% Better way to do default agent? */
			int autorestart = 0);
      /*
       * handle == 0     => build default handle from program name
       * handle == ""    => output only agent; close stdin
       * handle == "str" => force handle to given string
       */

      ProcessPipe* restart(const char* handle = 0);
      
      process_line_t* check_line();

      const char* postmortem();

      int count();
      
      void dump(int logtype = CLI_LOG_VERBOSE);
};

      
#endif /* !_INCLUDED_Pipes */
