/*                                             -*- c-file-style: "Ellemtel" -*-

`director.cc' - main event loop for director program

This file is part of version 3 of the DIRECTOR cli wrapper.
Read the `License' file for terms of use and distribution.
Copyright 2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <termios.h>
#include <pwd.h>
#ifdef SHADOW_PWD
#include <shadow.h>
#define SHADOW_LOOKUP_FAILED "sp"
#endif

#include "cli/cli.h"
#include "cli/clidefs.h"

#ifdef NO_CFHTLOG
#define cfht_logv(a,b,c,d)
#define cfht_log(a,b,c)
#endif

#include "director.hh"

void do_sleep(int seconds); /* temporarily disable the interval timer and do
			     * a regular sleep 
			     */
int default_mtype = -1;
int default_timeout_init = 0;
int default_timeout_sust = 0;
const char* default_prompts = ":"DEFAULT_OK_PROMPT":"DEFAULT_PROMPT":"DEFAULT_CONT_PROMPT":";
const char* psfailed = DEFAULT_FAILED_PROMPT;
const char* pscont   = DEFAULT_CONT_PROMPT;
const char* cli_newline = "\r"; /* %%% Some things require \n */

int psfailedlen, pscontlen; /* === hack */
unsigned int director_timer_hz = DIRECTOR_TIMER_HZ;
unsigned long director_timer;

static char hostname[81] = "unknown";

static int sessionhost_created = 0;

static int get_out = 0;
/*
 * When the program exits, it looks at this to determine
 * how "quickly" it should cleanup and exit.
 *
 * A value of 0 indicates the program is probably NOT exiting as a
 * direct result of a "detach" or "exit" command, so pause for 3 seconds
 * to let the user see what happened.
 *
 * A value of 1 indicates a "detach" or "exit" command, so don't pause.
 *
 * A value >1 indicates something unexpected happened during cleanup and
 * we should call _exit() to get out as soon as possible.
 */

/*
 * Order is important here . . . See Index for some descriptions.
 */

#include "Roll.hh"

Roll* roll=0; /* === Should be static but I want access to it from Pipes.cc */

#include "Curs.hh"

static Curs* curs=0;

#include "Pipes.hh"

static ProcessList processList;
static CommandPipe* commandPipe=0;
static LogPipe* logPipe=0;

#include "CliQueue.hh"

static CliQueue cliQueue;

static char* real_username = "Unknown";

static class Who
/*
 * Handles figuring out the correct roll key (uid) $HOME and $USER
 */
{
   private:
      clone_mode_t clone_mode_;
      uid_t uid_;
      int key_;
      char* name_;
      char* home_;
      char* passwd_;

      void save_pwd(struct passwd* pwd) {
#ifdef SHADOW_PWD
	 struct spwd* sp;

	 seteuid(0);
	 if ((sp = getspnam(pwd->pw_name)))
	 {
	    pwd->pw_passwd = sp->sp_pwdp;
	 }
	 else
	 {
	    if (!pwd->pw_passwd || strlen(pwd->pw_passwd) <= 2 && geteuid()!=0)
	       pwd->pw_passwd = SHADOW_LOOKUP_FAILED;
	 }
	 seteuid(getuid());
	 endspent();
#endif
	 uid_ = pwd->pw_uid;
	 if (name_) free(name_);
	 name_ = strdup(pwd->pw_name);
	 if (home_) free(home_);
	 home_ = strdup(pwd->pw_dir);
	 if (passwd_) free(passwd_);
	 passwd_ = strdup(pwd->pw_passwd);
      }

   public:
      Who() {
	 /*
	  * Constructor just sets up assuming we are going to
	  * run using getuid()'s session and shared memory.
	  */
	 struct passwd* pwd;

	 clone_mode_ = CLONE_OR_MASTER;
	 name_ = home_ = passwd_ = 0;
	 uid_ = getuid();
	 key_ = -2;
	 pwd = getpwuid(uid_);
	 if (!pwd) 
	 {
	    cfht_logv(CFHT_MAIN, CFHT_ERROR, "Can't find userid %d in password file", uid_);
	    perror("getpwuid"); 
	    exit(EXIT_FAILURE); 
	 }
	 save_pwd(pwd);
      }
	 
      void setup(const char* arg) {
	 /*
	  * This is called to override the default uid when director
	  * is invoked with `-k<user>'
	  */
	 struct passwd* pwd = 0;
	 
	 clone_mode_ = CLONE_ONLY;
	 errno = 0;
	 uid_ = strtoul(arg, NULL, 10);
	 if (uid_!=0 && errno==0)
	 {
	    pwd = getpwuid(uid_);
	 }
	 if (!pwd)
	 {
	    pwd = getpwnam(arg);
	 }
	 if (!pwd)
	 {
	    cfht_logv(CFHT_MAIN, CFHT_ERROR, "Can't find user %s in password file", arg);
	    perror(arg); 
	    exit(EXIT_FAILURE);
	 }
	 save_pwd(pwd);
      }
      
      void force_master() { clone_mode_ = MASTER_ONLY; }
      
      const char* name() { return name_; } /* user name from passwd */
      const char* home() { return home_; } /* home directory from passwd */
      const char* passwd() { return passwd_; } /* encrypted password */
      void setkey(int key) { key_ = key; }
      int getkey() { return (key_!=-2)?key_:uid_; } /* user id number */
      clone_mode_t clone_mode() { return clone_mode_; } /* was -k forced? */
} who;

/* --------------------------------------------------- */

static int debugflg = 0;

void /* %%% Was static! */
flush_to_screen()
   /*
    * Make the info in on the curs (the screen)
    * catch up with whats in Roll (the shared memory)
    */
{
   RollPtr* rp;
   
   if (!roll || !curs) return;

   while ((rp=roll->get_line())!=0)
   {
      /*
       * the master director process strips out \a's and rings the
       * local terminal bell immediately.  Clones only ring the bell
       * when an error is printed (or re-printed after a resize).
       */
      if (roll->is_clone() && rp->logtype==CLI_LOG_ERROR)
	 fputc('\a', stderr);
      curs->message(roll->get_line_number(),
		    roll->get_str(rp),
		    rp->logtype,
		    cliQueue.level()<=0,
		    rp->timestamp,
		    rp->length_handle);
      /*
       * === inefficient calling cliQueue.level() so much?
       * %%% doesn't work right (leaves a `*' on screen sometimes)
       * === Doesn't work for clones.
       */
   }
   
   RollLine* rl;
   
   if ((rl=roll->get_instant())!=0)
      curs->instant(rl->str);
   
   if ((rl=roll->get_progress())!=0)
      curs->progress(rl->str);

   if ((rl=roll->get_statusbar())!=0)
      if (curs->statusbar(rl->str))
	 roll->rewind(0); /* redraw the info bars */
   
   if ((rl=roll->get_titlebar())!=0)
      curs->titlebar(rl->str);
   
   for (int l=1; l<=9; l++)
   {
      if ((rl=roll->get_info(l))!=0)
	 curs->infoline(rl->str, l);
   }
   
   /*
    * If jumpscroll was temporarily turned on, turn it off again now.
    */
   curs->jumpscroll_check();
   
   curs->update(cliQueue.state());
}

void /* %%% Was static! */
flush_inputs()
   /*
    * Add lines from process stdout/stderr and CFHTLOGU
    * to the shared memory roll buffer
    */
{
   int max_repeat = 25; /* Don't "lock up" here if overflowing with input */
   const char* curline = "";
   if (!roll) return;

   while (max_repeat-- && curline)
   {
      const process_line_t* pl = processList.check_line();
      if (pl)
      {
	 curline = pl->line;
	 roll->add_line(curline, pl->mtype, 0, pl->handle, pl->pid);
      }
      else curline = 0;
   }
   
   max_repeat = 25;

   while (max_repeat-- && logPipe && (curline=logPipe->check_line())!=0)
   {
      /*
       * When session wants to end and make the feedback window go
       * away, it echos ^QUIT to the CFHTLOGU pipe.  For easy integration
       * into the sessions, we'll just have director do the same for now.
       */
      if (strncmp(curline, "^QUIT", 5)==0)
      {
	 flush_inputs();
	 roll->add_line("logonly: session cleanup in progress");
	 flush_to_screen();
	 get_out = 1;
	 exit(EXIT_SUCCESS);
      }
      else
	 roll->add_line(curline, /*logtype=*/-1, /*nolog=*/1);
   }
}

#include "builtins.cc"

static void
set_interval_timer(int onoff)
{
   struct itimerval timeout;

   if (onoff)
   {
      cli_signal(SIGALRM, CLI_SIG_TRAP);
      cli_signal_block(SIGALRM);
   }

   timeout.it_value.tv_sec=0;
   timeout.it_value.tv_usec=(onoff!=0)*(1000000/director_timer_hz);
   timeout.it_interval.tv_sec=0;
   timeout.it_interval.tv_usec=(onoff!=0)*(1000000/director_timer_hz);

   setitimer(ITIMER_REAL, &timeout, NULL);

   if (!onoff)
   {
      cli_signal_unblock(SIGALRM);
      cli_signal_event(SIGALRM);
      cli_signal(SIGALRM, CLI_SIG_DFL);
   }
}

void
do_sleep(int sec)
{
   set_interval_timer(0);
   sleep(sec);
   set_interval_timer(1);
}

static int unlocked = FALSE; /* Has the account password been entered? */

static void
final_cleanup(void)
   /*
    * This takes care of making sure the destructors for
    * roll, process, and curs always take place.  The
    * final_cleanup() function should be registered with
    * atexit() so it runs even if the program exits abnormally.
    *
    * There is no harmful side-effect from calling this function
    * more than once at exit.
    */
{
   char junkbuf[1024];

   cli_signal(SIGHUP, CLI_SIG_TRAP);
   cli_signal(SIGTERM, CLI_SIG_TRAP);

   if (roll && roll->is_clone()>0 && commandPipe)
   {
      sprintf(junkbuf,
	      "@say_ %s%.100s has detached from the session.",
	      unlocked?"warning:":"* ", real_username);
      commandPipe->send_line(junkbuf); /* %%% krusty */
   }
   flush_inputs();

   if (roll)
   {
      roll->add_line("logonly: cleaning up");
      if (curs) curs->update(' ');
      flush_to_screen();
   }

   if (sessionhost_created)
   {
      if (roll) roll->add_line("logonly: removing sessionhost file");
      sprintf(junkbuf, "%.500s/"SESSIONHOSTFILE, who.home());
      unlink(junkbuf);
      strcat(junkbuf, ".tmp");
      unlink(junkbuf);
      sessionhost_created = 0;
   }

   processList.killall();

   if (commandPipe) 
   {
      if (roll)
      {
	 roll->add_line("logonly: closing command pipe");
	 if (curs) curs->update(' ');
	 flush_to_screen();
      }
      delete commandPipe; commandPipe=0;
   }

   if (logPipe)
   {
      if (roll)
      {
	 roll->add_line("logonly: closing log pipe");
	 if (curs) curs->update(' ');
	 flush_to_screen();
      }
      delete logPipe; logPipe=0;
   }

   if (roll)
   {
      roll->add_line("logonly: closing roll"); 
      if (curs) curs->update(' ');
      flush_to_screen();
      delete roll; roll=0;
   }

   if (cli_signal_event(SIGHUP) ||
       cli_signal_event(SIGTERM)) get_out = 2;

   if (!get_out)
   {
      cfht_log(CFHT_MAIN, CFHT_LOGONLY, "pausing for 5 seconds");
      do_sleep(5); /* Give the user some time to read the screen */
   }

   if (curs)
   {
      cfht_log(CFHT_MAIN, CFHT_LOGONLY, "restoring terminal");
      /*
       * Restore the terminal to a usable state
       */
      curs->update(' ');
      fprintf(stderr,"\033]2;%s\a", hostname);
      delete curs; curs=0;
   }
   fprintf(stderr,"\r\n");

   if (!get_out)
   {
      get_out = 2;
   }
   else if (get_out>1)
   {
      cfht_log(CFHT_MAIN, CFHT_DONE, "_exiting");
      _exit(EXIT_SUCCESS);
   }

   cfht_log(CFHT_MAIN, CFHT_DONE, "exiting");
}

static void
diagnostic_dump(int logtype = CLI_LOG_VERBOSE)
{
   if (roll)
   {
      roll->add_line("###################### Director Internal Diagnostics", logtype);
      processList.dump(logtype);
      cliQueue.dump(logtype);
      roll->add_line("###################### End of Internal Diagnostics", logtype);
      flush_to_screen();
   }
}

static void
signal_cleanup(int signo)
{
   static char buf[256];

   if (signo==SIGHUP || signo==SIGTERM) get_out = 2;
   
   sprintf(buf, "logonly: director on host %.100s caught fatal signal #%d", hostname, signo);
   if (roll)
      roll->add_line(buf);
   else
      cfht_log(CFHT_MAIN, CFHT_LOGONLY, buf);
   fprintf(stderr, "%s\r\n", buf);

   cli_signal(signo, CLI_SIG_DFL);
   diagnostic_dump(CLI_LOG_WARNING);
   final_cleanup();
   kill(getpid(), signo);
}

BOOLEAN
send_ok(void)
{
   if (unlocked==TRUE || !roll->is_clone()) return TRUE;

   if (strlen(who.passwd()) <= 2) return FALSE;

   if (curs->password("Enter account password (not needed for \"say\" command):", who.passwd()) && commandPipe)
   {
      char junkbuf[1024];

      sprintf(junkbuf,
	      "@say_ warning: %.100s is sending commands from a clone window.",
	      real_username);
      commandPipe->send_line(junkbuf); /* %%% krusty */
      unlocked=TRUE;
   }

   return unlocked;
}

static void
autosense_rxvt()
{
   char buf[200];
   int stdin_flags;
   struct termios tio, tioorig;
   int tmp = 1, size = 0;
   int wait = 5 * 4; /* Wait up to 5 seconds */

   if (tcgetattr(STDIN_FILENO, &tioorig)<0) return;

   /*
    * Turn off echo and put the terminal in raw mode
    */

   tio = tioorig;
   tio.c_iflag &= ~(ICRNL | IUCLC);
   tio.c_lflag &= ~(ICANON | ECHO);
   tio.c_cflag |= CS8|CREAD;
   tio.c_cc[VMIN] = 1;
   tio.c_cc[VTIME] = 0;
   if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &tio)!=0) return;

   if ((stdin_flags = fcntl(STDIN_FILENO, F_GETFL, 0))<0
       || fcntl(STDIN_FILENO, F_SETFL, stdin_flags|O_NONBLOCK)<0)
   {
      tcsetattr(STDIN_FILENO, TCSAFLUSH, &tioorig);
      return;
   }

   /*
    * Flush
    */
   while (read(STDIN_FILENO, buf, sizeof(buf))>0);

   /*
    * A control-E will cause RXVT to send back it's "I'm an enhanced
    * VT102" response.
    */
   write(STDOUT_FILENO, "\005", 1);

   /*
    * Wait for the response
    */
   while (size < 7 && tmp != 0 && wait > 0)
   {
      struct timeval tv = { 0, 250000 };
      if ((tmp = read(STDIN_FILENO, buf+size, 7-size))==-1)
      {
	 SELECT(0, 0, 0, 0, &tv);
	 wait--;
      }
      else size += tmp;
   }

   /*
    * Is it an RXVT?
    */
   if (size == 7 && !strncmp(buf, "\033[?1;2c", 7))
      putenv("TERM=rxvt");

   fcntl(STDIN_FILENO, F_SETFL, stdin_flags);
   if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &tioorig)!=0) return;
}

const char* logpipename = 0;

/*
 * --------------- MAIN -----------------
 */
int
main(int argc, const char* argv[])
{
   int c, done = 0;
   char helpscreen='\0';
   char termtype[128]="";
   int
      infolines = -1,
      jumpflg = 1, /* %%% Change to 0 when PgUp PgDn are fixed */
      clockflg = 1,
      errflg = 0,
      verboseflg = 0,
      privateflg = 0,
      keep_bgcolor = 0;
   char junkbuf[1024];

   extern char *optarg;
   extern int optind, optopt;

   setuid(getuid()); /* make sure director can't be setuid root by accident */

   /*
    * Store the user's real name for use in talk messages
    */
   {
      struct passwd* pwd;
      pwd = getpwuid(getuid());
      if (pwd!=0) real_username = strdup(pwd->pw_gecos);
   }

   atexit(final_cleanup); /* make sure destructors happen */

   cli_signal(SIGTERM, (PFV)signal_cleanup);
   cli_signal(SIGHUP,  (PFV)signal_cleanup);
   cli_signal(SIGSEGV, (PFV)signal_cleanup); /* should really dump a core! */
   cli_signal(SIGBUS,  (PFV)signal_cleanup); /* ditto */
   cli_signal(SIGABRT, (PFV)signal_cleanup); /* ditto */
   cli_signal(SIGPIPE, CLI_SIG_IGN); /* Deal with EPIPE instead */

   logpipename = getenv("CFHTLOGU"); /* save this before it's clobbered */

   if (gethostname(hostname,80)==-1)
	 strcpy(hostname, "unknown host");
   /*
    * Initialize cfht_log facility.  Duplicate messages go there for
    * easier debugging and tracing.
    */

   putenv("CFHTLOGU=/dev/null"); /* we _are_ the user's feedback window */
   if (!getenv("CFHTLOGS") || !strcmp(getenv("CFHTLOGS"), "/dev/null"))
      putenv("CFHTLOGS=/tmp/pipes/syslog.np");
   if (!getenv("CFHTCONFPATH"))
      putenv("CFHTCONFPATH=/cfht/conf");
   putenv("LINES=0"); putenv("COLUMNS=0"); /* force ncurses not to rely on these */

   /*
    * If CFHTLOGS doesn't exist, it means roll isn't running on this machine.
    * disable CFHTLOGS for director itself, so that the messages don't trash
    * the director-curses screen.  (Because if cfht_log() can't write to CFHTLOGS,
    * it prints those messages on stdout instead.)
    */
   {
      struct stat stbuf;
      
      if (stat(getenv("CFHTLOGS"), &stbuf)==-1 && errno==ENOENT)
	 putenv("CFHTLOGS=/dev/null");
   }

   cfht_log( CFHT_MAIN, CFHT_LOG_ID, "director-"VERSION_STRING);

   while ((c = getopt(argc, (char**)argv, ":t:z:bCcndvji:K:k:mh:pP:T:M:")) != -1)
   {
      switch (c)
      {
	 case 't': sprintf(termtype, "COLORTERM=%.100s", optarg);
                   putenv(termtype);
	           break;
	 case 'b': keep_bgcolor=1; break;
	 case 'C': clockflg=0; break;
	 case 'c': cli_newline="\r"; break;
	 case 'n': cli_newline="\n"; break;
	 case 'd': debugflg=1; break;
	 case 'v': verboseflg=1; break;
	 case 'j': jumpflg=1; break;
	 case 'i': infolines = atoi(optarg); break;
	 case 'K': who.setkey(atoi(optarg)); break;
	 case 'k': who.setup(optarg); break;
	 case 'm': who.force_master(); break;
	 case 'h': if (strlen(optarg)<2) helpscreen = *optarg; errflg++; break;
	 case 'p': privateflg=1; who.setkey(-1); break;
	 case 'z': com_timerhz(optarg); break;
	 case 'P':
	 {	
	    if (strlen(optarg)<2 || *optarg != optarg[strlen(optarg)-1])
	    {
	       printf(
		  "error: agent start -P option takes a list of prompt "
		  "strings separated by a delimiter character.  This "
		  "delimiter character should appear at the beginning "
		  "AND end of the argument to -P and between each possible "
		  "prompt string.  Choose any delimiter character that does "
		  "not appear in any of your prompts.  For example: "
		  "'agent start -P /ok>/done>/'\r\n");
	       exit(EXIT_FAILURE);
	    }
	    default_prompts = optarg;
	    break;
	 }
	 case 'T':
	 {
	    char* colon = strchr(optarg, ':');
	    
	    if (!colon)
	    {
	       printf("error: -T option requires 'initial:sustained' values (milliseconds)\r\n");
	       exit(EXIT_FAILURE);
	    }
	    default_timeout_init = atoi(optarg);
	    default_timeout_sust = atoi(colon+1); /* %%% */
	    break;
	 }
	 case 'M':
	 {
	    if (!strcmp(optarg, "debug")) default_mtype = CLI_LOG_DEBUG;
	    else if (!strcmp(optarg, "logonly")) default_mtype = CLI_LOG_VERBOSE;
	    else if (!strcmp(optarg, "status")) default_mtype = CLI_LOG_STATUS;
	    else if (!strcmp(optarg, "warning")) default_mtype = CLI_LOG_WARNING;
	    else if (!strcmp(optarg, "error")) default_mtype = CLI_LOG_ERROR;
	    else
	    {
	       printf( /* === Clean up */
		  "error: -M option requires an argument "
		  "with the default message type for feedback from this "
		  "agent.  Valid message types are `debug' `logonly' "
		  "`status' `warning' and `error'.\r\n");
	       exit(EXIT_FAILURE);
	    }
	 }
         case ':': if (optopt!='h') fprintf(stderr, "-%c needs argument\r\n", optopt);
	 case '?': errflg++; break;
     }
   }

   /*
    * === Calculate prompt string length now
    */
   psfailedlen = strlen(psfailed);
   pscontlen = strlen(pscont);
   
   sprintf(junkbuf, "HOME=%.500s", who.home());
   putenv(strdup(junkbuf));
   sprintf(junkbuf, "USER=%.500s", who.name());
   putenv(strdup(junkbuf));

   if (!getenv("PATH"))
   {
      putenv("PATH=.:/cfht/bin:/usr/local/bin:/usr/bin:/bin");
      /* === warning: using default path */
   }

   /* %%% Hack until version 4.0 */
   sprintf(junkbuf, "PATH=%.100s/"SCRIPTDIR":%.500s", cli_gethome(), getenv("PATH"));
   putenv(strdup(junkbuf));

   if (!privateflg)
   {
      int len;
      char sbuf[600];

      sprintf(sbuf, "%.500s/"SESSIONHOSTFILE, who.home());
      strcpy(junkbuf, "SESSIONHOST=");
      len = readlink(sbuf, junkbuf+12, sizeof(junkbuf)-13);
      if (len!=-1)
      {
	 junkbuf[len+12]='\0';
	 putenv(strdup(junkbuf));
      }
   }

   if (!getenv("FFTEMPLATE"))
   {
      sprintf(junkbuf,"FFTEMPLATE=%.500s/.,template", who.home());
      putenv(strdup(junkbuf));
   }
   if (!getenv("SESSIONHOST"))
   {
      sprintf(junkbuf, "SESSIONHOST=%.500s", hostname);
      putenv(strdup(junkbuf));
   }
   if (!getenv("DISPLAY"))
   {
      sprintf(junkbuf, "DISPLAY=%.500s:0.0", getenv("SESSIONHOST"));
      putenv(strdup(junkbuf));
   }

   if (helpscreen=='e')
      /*
       * Display Enviroment Variable Help Screen
       */
   {
      fprintf(stderr,"\n"
"Environment and director:\n"
"\n"
DIRECTOR_ENV" (this is version `%s')\n"
"  DO NOT SET this externally!  Director sets it to an appropriate value\n"
"  so agents know what version of director they are running under.\n"
"SESSIONHOST (you're running on `%s')\n"
"  If unset, director will set this to the current hostname.  If it is set\n"
"  but it doesn't agree with the current hostname, director will issue a warning\n"
"  and keep the value you have set. File ~/"SESSIONHOSTFILE" overrides this.\n"
"DISPLAY (currently `%s')\n"
"  If unset, director sets this to $SESSIONHOST:0.0.\n"
"USER and HOME (currently `%s' and `%s', uid/roll key is %d)\n"
"  After a userid for a director session has been determined (either that\n"
"  passed with `-k' or getuid() if no `-k') director overrides these two\n"
"  variables to agree with information from the passwd table. The original\n"
"  settings of these variables have no effect on director or its agents.\n"
"PATH (currently `%s')\n"
"  User's current path is sent to agent as-is.\n"
"LINES and COLUMNS\n"
"  Ignored (set to 0 to force ncurses to determine window size on its own.)\n"
"CFHTCONFPATH (currenty `%s')\n"
"  Defaults to /cfht/conf.\n"
"CFHTLOGS (currently `%s')\n"
"  Log messages from agents running within director are also written here\n"
"  using cfhtlog() calls.  The message type is determined from the first\n"
"  few charaters in a line printed by the agent.  For example, a string\n"
"  starting with `error:' is written as CFHT_ERROR, etc.  See also `-hs'.\n"
"  Defaults to /tmp/pipes/syslog.np.\n"
"CFHTLOGU (currently `%s')\n"
"  If this file (fifo) does not exist, or if no other process is currently\n"
"  reading from it, it is opened.  Messages written to this node by other\n"
"  programs are echoed in the director feedback area.\n"
"FFTEMPLATE (currently `%s')\n"
"  Passed on to the agent.  If not found, defaults to ~/.,template\n"
"When the agent runs on another host, these are the ONLY variables it\n"
"should expect to have defined.\n"
	      ,VERSION_STRING
	      ,getenv("SESSIONHOST")
	      ,getenv("DISPLAY")
	      ,getenv("USER")
	      ,getenv("HOME")
	      ,who.getkey()
	      ,getenv("PATH")
	      ,getenv("CFHTCONFPATH")
	      ,getenv("CFHTLOGS")
	      ,logpipename?logpipename:"not set"
	      ,getenv("FFTEMPLATE"));
      exit(EXIT_FAILURE);
   }

   if (helpscreen=='f')
      /*
       * Display .director/ files help screen
       */
   {
      fprintf(stderr,"\n"
"Files used by director:\n\n"
"%s (see also: director -he)\n"
"  Opened for reading session feedback messages from other sources.\n"
"\n"
"~/"SCRIPTDIR"/\n"
"  When Director receives a command, this subdirectory is searched first.\n"
"  A non-executable file with a matching name is sourced like a batch file.\n"
"  An executable (or link to executable, or link to clicmd) is executed\n"
"  with the same arguments given on the command line.\n"
"\n"
"~/"SESSIONHOSTFILE"\n"
"  If this file exists as a symlink, director will first attempt to\n"
"  `remsh' to the machine name given by the contents of the link.\n"
"\n"
"~/"STARTUPFILE"\n"
"  This file is automatically sourced at startup (non-clones only).\n"
"  Providing an agent on the command line (after --) suppresses this.\n"
"\n"
"~/"HISTORYFILE"\n"
"  Command history is saved here between sessions.  The `~' is always\n"
"  replaced by the home directory of the original user starting director.\n"
"\n"
"~/"COMMANDPIPE"\n"
"  Non-clones create (and remove at exit) this fifo and read from it.\n"
"  Clones (and other external programs) must only write commands to it.\n"
"  If `-k' is present, clones use the `~' for that account.  `~' expansion\n"
"  at the command prompt also follows the `-k' user after in that case.\n"
,logpipename?logpipename:"CFHTLOGU is currently not set");
      exit(EXIT_FAILURE);
   }

   if (helpscreen=='s')
      /*
       * Display string parsing rules
       */
   {
      fprintf(stderr,"\n"
"String parsing rules: ('^'->beginning of a line)\n\n"
"OUTPUT FROM AGENT:\n"
"^status:     displays green,  logs as CFHT_STATUS\n"
"^logonly:    displays brown,  logs as CFHT_LOGONLY\n"
"^debug:      displays brown,  logs as CFHT_DEBUG\n"
"^warning:    displays yellow, logs as CFHT_WARN\n"
"^error:      displays red,    logs as CFHT_ERROR\n"
"Any line on stderr is at least a CFHT_WARN, so C agents can use perror().\n"
"All other lines are not logged, but display in various areas of the screen:\n"
"^statusbar:  creates status bar (if needed); displays text in reverse video\n"
"^titlebar:   displays text in window title (if TERM=xterm or rxvt only)\n"
"^infoN:      (where N is 1..9) displays on info panel if infosize >= N\n"
"^instant:    displays text in user prompt area until command completes\n"
"^*           displays with highlighted background (including the *)\n"
"^ok>         command completed ok *handled internally by cli_getline()\n"
"^failed>     last command failed  *handled internally by cli_getline()\n"
"\\a(bell)     rings the terminal bell immediately (and eats the character)\n"
"...          all other lines scroll in white in the feedback area\n"
"             long lines are word-wrapped to fit the current window size\n"
"\n"
"INPUT AT PROMPT, COMMAND PROMPT, OR FROM SCRIPT:\n"
" ~ or ~user  basic shell-style tilde-substitution is done by director\n"
"builtin      (where `builtin' is a director builtin command) runs immediately\n"
"other        (where `other' is any other command) put in queue for agent\n"
"@cmd         runs cmd as above but does not echo the command itself to window\n"
"-cmd         runs cmd as above but does not abort a script if this command fails\n"
"{PID}cmd     runs cmd as above; when it returns director signals PID with:\n"
"               SIGUSR1 if command returned PASS (prompt was ^ok>)\n"
"               SIGUSR2 if command returned FAIL (prompt was ^failed>)\n"
"               *handled internally by cli_cmd() or the shell utility `clicmd'\n");
      exit(EXIT_FAILURE);
   }

   if (errflg)
   {
      fprintf(stderr,"\n"
"usage: %s [director options] [ -- [rhost:]agent [agent options] ]\n"
"or     %s [director options] -- [{B9600,see_manual...}]serialdevice\n"
"or     %s           (no startup and no args, just tails log pipe)\n"
"or     %s -k <user> (attach to another user's session without su'ing)\n"
"\n"
"agent options are passed without change to the actual cli program\n"
"\n"
"director options are:\n"
"    -i <n>    displays <n> info lines at bottom of screen (default is 0)\n"
"    -b        use default terminal background color or pixmap\n"
"    -v        turns on verbose mode which shows logonly and debug\n"
"    -d        turns on debug messages (also requires -v to see on screen)\n"
"    -m        used by sessions to ensure they get a master, not a clone\n"
"    -c        sends a CR character after each command (this is the default)\n"
"    -n        sends a NL character after each command instead of CR\n"
"    -C        turn off the titlebar clock for slow connections\n"
"    -t <type> override terminal auto-sensing and TERM var (e.g., -t vt100)\n"
"    -p        starts a private director (can't be cloned, no CFHTLOGU...)\n"
"    -P:p1:p2: Sets list of possible prompts; first+last chars set delimiter\n"
"    -T i:s    Sets (i)nitial:(s)ustained timeouts when no prompt available\n"
"    -h        displays this help information\n"
"    -hf       displays summary of director-related files and resources\n"
"    -hs       displays string parsing rules used on output from realprogram\n"
"    -he       displays summary of how the environment affects director\n"
"              (hint: -he can be combined with -k<user> for testing)\n"
"\n"
"All options above can also be controlled by director builtin commands.\n"
"Type `builtin' at the director command prompt for more help on these.\n"
"Type `license' to read information on (lack of) warranty and copying.\n"
"\n"
	      ,argv[0], argv[0], argv[0], argv[0]);
      exit(EXIT_FAILURE);
   }

   /*
    * Don't count on an ncurses that checks for COLORTERM.  Instead,
    * Just set TERM equal to COLORTERM before initializing curses.
    */
   if (getenv("COLORTERM") && *getenv("COLORTERM"))
   {
      static char tmp[128];
      sprintf(tmp, "TERM=%.100s", getenv("COLORTERM"));
      putenv(tmp);
   }
   else
   {
      /*
       * If all we have to go on is the TERM variable, try and autodetect
       * if this is an rxvt window.  Rxvt should reply with ESC [ ? 1 ; 2 c
       * when given a ^E.
       *
       * Do this before cli_runon, since we no longer have a tty
       * device after cli_runon().
       */
      autosense_rxvt();
   }

   if (strcmp(getenv("SESSIONHOST"), hostname)!=0)
   {
      /*
       * Remote shell to another machine if sessionhost is different
       * from the current host.
       */

      fprintf(stderr,
	      "Please wait; Switching to the session host %s. . .\r\n",
	      getenv("SESSIONHOST"));

      sprintf(junkbuf, "CFHTLOGU=%.900s", logpipename?logpipename:"/dev/null");
      putenv(strdup(junkbuf));

      cli_runon(getenv("SESSIONHOST"), /*rawtty=*/1, argc, argv);
      exit(EXIT_FAILURE);
   }
	      
   /*
    * This must come after director's own CFHT_LOG_ID call, because this
    * call will look for the presence of DIRECTOR and cause future cfht_log
    * calls to print to stderr.  For director itself, this doesn't work.
    *
    * In the case where director must first switch to another sessionhost,
    * it must also come after runon.
    */
   putenv(DIRECTOR_ENV"="VERSION_STRING); /* pass this on to the cli */

   /*
    * Create a roll buffer which will hold the last N lines echoed
    * to the terminal.  uid is used as a roll key so other director
    * clones will be able to attach to a shared memory segment with
    * that key and view the same log messages.
    */

   roll = new Roll(who.getkey(), who.clone_mode()); /* === shmdt on exec? */

   if (!privateflg)
   {
      char tmpname[512];

      sprintf(tmpname, "%s/.director", who.home());
      mkdir(tmpname, 00755);
   }

   if (!privateflg && !roll->is_clone())
   {
      struct stat stbuf;

      sprintf(junkbuf, "%.500s/"SESSIONHOSTFILE, who.home());
      if (lstat(junkbuf, &stbuf)==-1 && errno==ENOENT)
      {
	 roll->add_line("logonly: creating sessionhost file");
	 if (symlink(getenv("SESSIONHOST"), junkbuf)==0)
	 {
	    sessionhost_created = 1;
	    /*
	     * Set a flag so we remember to clean it up too
	     */
	    strcat(junkbuf, ".tmp");
	    symlink(getenv("SESSIONHOST"), junkbuf);
	 }
	 else
	 {
	    roll->add_line("warning: trouble creating sessionhost link");
	 }
      }
      else
      {
	 strcat(junkbuf, ".tmp");
	 if (lstat(junkbuf, &stbuf)==0)
	 {
	    fprintf(stderr, "error: Removing stale sessionhost link.  Please re-start director now.\n");
	    unlink(junkbuf);
	    junkbuf[strlen(junkbuf)-4] = '\0';
	    unlink(junkbuf);
	    exit(EXIT_FAILURE);
	 }
      }
   }
	 
   /*
    * The variable `infolines' will contain a -1 unless the user
    * started director with -i#.  If it still contains -1, check if
    * there is a hint saved in shared memory for the new size.
    */
   if (infolines==-1) infolines = roll->get_infosize();
   roll->set_infosize(infolines);

   /*
    * constructor for this class starts up curses in the terminal window
    */
   curs = new Curs(infolines, keep_bgcolor);
   
   com_clock(clockflg?"on":"off");
   com_verbose(verboseflg?"on":"off");
   com_debug(debugflg?"on":"off");
   curs->clone(roll->is_clone());

   curs->titlename(who.name());
   if (optind<argc && !roll->is_clone())
   {
      sprintf(junkbuf, "titlebar: %.500s", argv[optind]);
      roll->add_line(junkbuf);
   }

   if (jumpflg) curs->jumpscroll();

   /*
    * Install a signal handlers to catch window size change events,
    * and control C/Z/| at the keyboard
    */
#if defined(SIGWINCH) && (SIGWINCH!=0)
   cli_signal(SIGWINCH, CLI_SIG_TRAP);
#endif
   cli_signal(SIGINT, CLI_SIG_TRAP);
   cli_signal(SIGTSTP, CLI_SIG_TRAP);
   cli_signal(SIGCHLD, CLI_SIG_TRAP);
   cli_signal(SIGQUIT, CLI_SIG_TRAP);

   tzset();

   /*
    * Ignore the setting of CFHTLOGU if this is a clone
    */
   if (!privateflg && !roll->is_clone() && 
       logpipename && *logpipename && strcmp(logpipename,"/dev/null")!=0)
   {
      logPipe = new LogPipe(logpipename);

      if (logPipe->get_fd()==-1)
      {
	 cfht_logv(CFHT_MAIN, CFHT_ERROR, "Can't read log pipe `%s'",
		   logpipename);
	 fprintf(stderr,
"Can't use log pipe.  Is someone else already reading it?\r\n"
"If not, manually remove the file `%s' and try again.\r\n"
"Or, to disable CFHTLOGU, unset or set to empty $CFHTLOGU.\r\n",
		 logpipename);
	 exit(EXIT_FAILURE);
      }
   }

#ifndef CYGWIN32
   /*
    * constructor for this class creates a node in the file
    * system and begins accepting commands from other programs
    * through the named pipe.  If pipename is NULL, it is not
    * an error.  It just means we aren't using the pipe feature.
    */

   if (!privateflg)
   {
      char pipename[512];

      sprintf(pipename, "%s/.director", who.home());
      mkdir(pipename, 00755);
      strcat(pipename, "/command.np"); /* %%% Clean up and use defines instead? */

      commandPipe = new CommandPipe(pipename, /*wronly=*/roll->is_clone());

      /* === set close on exec flag? */

      if (!roll->is_clone() && commandPipe->get_fd()==-1)
      {
	 cfht_logv(CFHT_MAIN, CFHT_ERROR, "Can't open command pipe `%s'",
		   pipename);
	 fprintf(stderr,
"Can't open command pipe.  Perhaps a copy is running on another machine?\r\n"
"If not, manually remove the file `%s' and try again.\r\n",
		 pipename);
	 exit(EXIT_FAILURE);
      }
   }
#endif

   /*
    * This converts an invocation such as "director -- directest"
    * into the corresponding "agent start ..." command.
    */
   if (optind < argc && !roll->is_clone())
   {
      sprintf(junkbuf, "agent start -D %.500s", argv[optind++]);
      while (optind < argc) /* %%% Check limits! */
	 sprintf(junkbuf + strlen(junkbuf), " %.100s", argv[optind++]);
      /* %%% strlen hack */
      cliQueue.add_line(new CliQueueNode(junkbuf));
   }
   /*
    * Source a startup script if needed, but not if an agent was given
    * on invocation of director (hence the "else" below.)
    */
   else if (!roll->is_clone() && !privateflg)
   {
      struct stat stbuf;
      
      sprintf(junkbuf, "%.500s/"STARTUPFILE, who.home());
      if (stat(junkbuf, &stbuf)==0)
      {
	 sprintf(junkbuf, "source %.500s/"STARTUPFILE, who.home());
	 cliQueue.add_line(new CliQueueNode(junkbuf));
      }
   }
   
   /*
    * Start an interval timer that delivers SIGALRM at a steady
    * rate of 4Hz.  We use this to ensure that `seconds' on the
    * clock are really `seconds'.  One side-effect is that director
    * cannot use alarm() or sleep() without removing this
    * interval timer since they use the same signal.
    */

   cfht_log( CFHT_MAIN, CFHT_START, "Director Version "VERSION_STRING" entering main event loop");

   set_interval_timer(1);

   /*
    * ------------------- INPUT LOOP ----------------------
    */
   if (roll) if (!roll->is_clone())
   {
      roll->add_line("------------------------------------------------------------------------");
      roll->add_line("               Director CFHT version "VERSION_STRING" ready."
#ifdef NO_CFHTLOG
		     " (standalone)"
#else
		     " (cfht_log support)"
#endif
		     );
      /*
       * OR,
       * If you modify the code, edit this line so the program clearly indicates
       * this on startup.  (Or, call it something other than director, but do not
       * remove the original Copyright from derivative works.
       roll->add_line("Director version "VERSION_STRING" with modifications by ...");
       */
      roll->add_line("Copyright 2001 Canada-France-Hawaii Telescope, <daprog@cfht.hawaii.edu>");
      roll->add_line("Distributed under the GNU Public License 2.  Type `license' for details.");
      roll->add_line("------------------------------------------------------------------------");
   }
   else if (commandPipe)
   {
      /*
       * Announce it when someone joins the session
       */
      sprintf(junkbuf,
	      "@say_ * %.100s has started a clone window.",
	      real_username);
      commandPipe->send_line(junkbuf); /* %%% krusty */
   }

   while (!done)
   {
      const char* curline;
      FDSet readfds, exceptfds;

      /*
       * Build up a list of all file descriptors that we are interested in.
       */
      if (!roll->is_clone())
      {
	 processList.add_read_fds(readfds);

	 if (readfds.get_fdnum()==0 && 0) /* %%% Don't exit! */
	 {
	    flush_inputs();

	    roll->add_line("logonly: no processes left");
	    flush_to_screen();
	    exit(EXIT_SUCCESS);
	 }
      }
      if (commandPipe) readfds.add_fd(commandPipe->get_fd());
      if (logPipe) readfds.add_fd(logPipe->get_fd());

      if (roll->is_clone()==-1)
      {
	 fprintf(stderr, "clone exiting because parent exited\r\n");
	 exit(EXIT_SUCCESS);
      }

      readfds.add_fd(STDIN_FILENO);
      exceptfds.add_fd(STDIN_FILENO);
      exceptfds.add_fd(STDOUT_FILENO);

      /*
       * Now wait for input or interruption from a signal.
       */
      {
	 int rtn = 0;

	 /*
	  * flow-<0>
	  */
	 cli_signal_unblock(SIGALRM);
	 if ((rtn=SELECT(readfds.get_fdnum(), readfds.get_fdsetp(), 0, exceptfds.get_fdsetp(), 0))<0
	     && errno!=EINTR)
	 {
	    roll->add_line("Internal director error (select failed)", CLI_LOG_ERROR);
	    roll->add_line(strerror(errno), CLI_LOG_ERROR);
	    diagnostic_dump();
	    do_sleep(4);
	 }
	 cli_signal_block(SIGALRM);

	 if (cli_signal_event(SIGALRM))
	 {
	    /*
	     * --- This stuff runs at director_timer_hz ---
	     */

	    /* ... */

	    if ((++director_timer % director_timer_hz)==0)
	    {
	       static int minute_timer = 0;

	       /*
		* --- This stuff runs at about 1 Hz. ---
		*/
	       curs->idle();
	       /*
		* Update the clock in the titlebar and erase the
		* progress popup if enough time has elapsed.
		*/

	       if ((++minute_timer % 60)==0)
	       {
		  static time_t time_last = 0;
		  time_t time_now = time(0);
		  /*
		   * --- This stuff runs about once per minute. ---
		   */

		  if (time_now > time_last + 3600)
		  {
		     /*
		      * --- This stuff runs about once per hour. ---
		      */
		     time_last = time_now;
		     if (!roll->is_clone())
		     {
			cfht_log(CFHT_MAIN, CFHT_START,
				 "Director Version "VERSION_STRING" time stamp");
		     }
		  }
	       }
	    }
	 }

	 /*
	  * This wouldn't normally happen, but if anything unusual happens
	  * on stdin/stdout (EOF or exceptions) then just exit.
	  */
	 if (feof(stdin))
	 {
	    roll->add_line("error: eof on stdin; exiting");
	    exit(EXIT_FAILURE);
	 }

	 if (rtn>0 &&
	     (exceptfds.is_set(STDIN_FILENO) ||
	      exceptfds.is_set(STDOUT_FILENO)))
	 {
	    roll->add_line("error: i/o exception; exiting");
	    exit(EXIT_FAILURE);
	 }
      }

      /*
       * Instead of checking which bits are set in readfds,
       * or checking the return value of select, we just
       * check all possible events that this loop should handle
       * each time, and do all reads as non-blocking i/o.
       */

      /*
       * If the child process has anything to tell us,
       * scroll it up in the feedback window.
       */

      curline="";

      /*
       * flow-<1>
       */
      flush_inputs();

      /*
       * Check if any agents that are set to auto-restart
       * have died and need to be restarted.
       */
      {
	 ProcessPipe* restarted = processList.autorestart();

	 if (restarted)
	 {
	    CliQueueNode* newNode = new CliQueueNode("autorestart", 99999);
	    cliQueue.add_line(newNode);
	    newNode->assign(restarted);
	 }
      }

      /*
       * Check if any "zombie" agents need to be cleaned up
       */
      processList.cleanup();

      /*
       * If a child exited...
       */
      if (cli_signal_event(SIGCHLD))
      {
	 const char* s;

	 while ((s = processList.postmortem()))
	 {
	    roll->add_line(s);
	    flush_to_screen();
	 }
      }

      /*
       * If control-Z was pressed at the terminal, pass it on ...
       */
      if (cli_signal_event(SIGTSTP))
	 if (curs->confirm("  Really send the stop signal? (Y/N)"))
	    com_sigsend("TSTP");
      /*
       * If control-C was pressed at the terminal, pass it on ...
       */
      if (cli_signal_event(SIGINT))
	 if (curs->confirm("  Really send the abort signal? (Y/N)"))
	    com_sigsend("INT");
      /*
       * If control-| was pressed at the terminal, pass it on ...
       */
      if (cli_signal_event(SIGQUIT))
      {
	 if (debugflg)
	    diagnostic_dump(); /* In director -d mode, this gives a dump. */
	 else if (curs->confirm("  Really send the quit (break) signal? (Y/N)"))
	    com_sigsend("QUIT");
      }
      
      /*
       * If a window change size event has occurred...
       */
#if defined(SIGWINCH) && (SIGWINCH!=0)
      if (cli_signal_event(SIGWINCH))
      {
	 curs->resize();
	 roll->rewind(LINES*2);
      }
#endif
      
      /*
       * flow-<2>
       *
       * check_line returns a pointer to a string if
       * a complete command line has been entered, otherwise
       * it returns 0.
       */
      while ((curline=curs->check_line())!=0 ||
	     (curline=commandPipe?commandPipe->check_line():0)!=0)
      {
	 const char* endbrace = 0;
	 int new_level = 0;
	 int killpid = 0;

	 while (*curline!='\0' && *curline<=' ') curline++; /* eat white space */

	 /* ---------------------------------------------------------- */

	 /*
	  * Specially handled commands (aka ``hacks'').  These commands
	  * cannot be sent through the named pipe, for security reasons.
	  */

	 /*
	  * `detach' can be used if the agent is completely hosed and
	  * you want to force director to exit.  It tries to kill the
	  * agent before exiting if it is not a clone.
	  */

	 if (strcmp(curline, "detach")==0 ||
	     strcmp(curline, "logout")==0 ||	/* logout and bye */
	     strcmp(curline, "bye")==0)		/* are aliases */
	 {
	    get_out = 1;
	    exit(EXIT_SUCCESS);
	 }

	 /*
	  * `exit' is handled like `detach' if you are a clone.  Even if
	  * not a clone, we still want to signal a {PID} waiting that the
	  * command was completed, because otherwise exit never returns.
	  */

	 if ((strcmp(curline, "exit")==0 ||
	      strncmp(curline, "exit ", 5)==0))
	 {
	    get_out = 1;
	    if (roll->is_clone() || processList.count()==0)
	       exit(EXIT_SUCCESS);
	 }

	 /* ---------------------------------------------------------- */

	 /* check for `{PID.LEVEL}command' notation */

	 if (*curline=='{' &&
	     (endbrace=strchr(curline,'}'))!=0)
	 {
	    const char* dot = strchr(curline+1, '.');

	    if (dot && dot < endbrace)
	       new_level = atoi(dot+1);
	    killpid = atoi(curline+1);
	    curline = endbrace + 1;
	 }
	 
	 /* ---------------------------------------------------------- */
	 
	 /* 
	  * Allocate a node for the command queue for this line
	  * Tilde-substitution is done on curline by the constructor
	  * for CliQueueNode.
	  */

	 CliQueueNode* newNode = new CliQueueNode(curline, new_level, killpid);

	 /*
	  * Class 1A Commands:	Instant commands that affect the appearance
	  *                  	of director itself.  Handled locally, even
	  *                  	for clones.
	  */

	 if (!*(newNode->handle()) && cli_search(newNode->line(), local))
	 {
	    cli_init("director", local, 0);
	    if (cli_execute(newNode->line())==PASS) newNode->pass();
	 }
	 /*
	  * Often people forget to type "say" at the beginning of the line
	  * in their clones, when they mean to just send a message, or they
	  * hit an extra enter at the end of each line.  This catches both
	  * of those, but it only applies to clones, and only for as long
	  * as they are "locked."  Once you give the password, mistakes get
	  * sent as commands.
	  *
	  * === TODO: Come up with better way of identifying likely "say"
	  * mistakes.  Or add a "say" mode turned off with unsay or something...
	  */
	 else if (unlocked!=TRUE && roll->is_clone() &&
		  (!*(newNode->line()) || strlen(newNode->line()) > 30))
	 {
	    if (*(newNode->line()))
	       com_say(newNode->line());
	 }
	 else if (!send_ok())
	 {
#ifdef SHADOW_PWD
	    if (!strcmp(who.passwd(), SHADOW_LOOKUP_FAILED))
	       curs->progress("Director is not setuid root; can't read shadow password database");
	    else
#endif
	       curs->progress("Incorrect password; Command not sent!");
	 }
	 else
	 {
	    /*
	     * Class 1B Commands:	Instant commands that affect the agent.
	     *                    	Must be sent to remote director if this
	     *                    	is a clone.
	     */
	    
	    if (!roll->is_clone() && !*(newNode->handle()) &&
		cli_search(newNode->line(), remote))
	    {
	       cli_init("director", remote, 0);
	       switch (cli_execute(newNode->line()))
	       {
		  case PASS: newNode->pass(); break;
		  case FAIL:
		  {
#if 0
		     /* %%% This breaks "ready" which has to return fail
		      * if things are not ready yet, but shouldn't cause
		      * the other pending commands to be deleted.
		      */
		     if (!newNode->no_fail())
		     {
			if (roll)
			   roll->add_line("logonly: Canceling command queue "
					  "(remote builtin failed)");
			cliQueue.cancel(); /* %%% is this desirable? */
		     }
#endif
		     break;
		  }
		  case BUSY:
		  {
		     extern ProcessPipe* newProcess; /* %%% */

		     if (newProcess)
		     {
			cliQueue.add_line(newNode);
			newNode->assign(newProcess);
			newProcess = 0;
			newNode = 0;
		     }
		     else if (roll)
		     {
			roll->add_line("warning: cli_execute returned BUSY but "
				       "no newProcess");
		     }
		     break;
		  }
	       }
	    }
	    else
	    {
	       /*
		* Class 2 Commands:	queued in order using the cliQueue
		*/

	       if (roll->is_clone())
	       {
		  /*
		   * Clones must get the command to the parent's cliQueue instead
		   * Sending "curline" instead of newNode->line() because
		   * the parent will create a newNode itself, and we only
		   * want tilde-substitution to happen once.
		   */
		  if (commandPipe) commandPipe->send_line(curline);
	       }
	       else
	       {
		  if (cliQueue.level()>=0 && newNode->level()==0)
		  {
		     /*
		      * Display the command with a `*' in the first column to show
		      * the user that we got their command, and it will be executed
		      * as soon as the process is ready.
		      * === This is busted.
		      */
		     curs->pending(curline);
		  }

		  cliQueue.add_line(newNode);
		  newNode = 0;
		  /* The node will get deleted when it comes up for execution and
		   * cliQueue returns it to us.
		   */
	       }
	    }
	 }
	 if (newNode) delete newNode;
      }

      CliQueueNode* ready;

      if ((ready = cliQueue.check_line()))
      {
	 const char* line = ready->line();

	 /*
	  * `cd' is handled by director AND the agent so they both
	  * agree on what the current working directory is.
	  * === Warning: director will get lost (and not complain about it)
	  * if you cd to a directory that exists only on the agent machine.
	  */
	 
	 if ((strcmp(line, "cd")==0 ||
	      strncmp(line, "cd ", 3)==0))
	 {
	    const char* newdir = line + 2;
	       
	    while (*newdir && *newdir<=' ') newdir++;
	    if (*newdir) cli_cd_internal(newdir);
	 }
	    
	 /* ---------------------------------------------------------- */
	 
	 /*
	  * Re-check for builtin commands in case they appear in a
	  * script or startup file.
	  */
	 if (!*(ready->handle()) && cli_search(line, local))
	 {
	    cli_init("director", local, 0);
	    if (cli_execute(line)==PASS) ready->pass();
	    cliQueue.remove(ready); /* If a "local" command fails, don't
				     * clear the command queue.
				     */
	 }
	 else if (!*(ready->handle()) && cli_search(line, remote))
	 {
	    cli_init("director", remote, 0);

	    switch (cli_execute(line))
	    {
	       case PASS: ready->pass(); cliQueue.remove(ready); break;
	       case FAIL:
	       {
		  if (!ready->no_fail())
		  {
#if 0
		     /* %%% See comments above. */
		     if (roll)
			roll->add_line("logonly: Canceling command queue "
				       "(remote builtin failed)");
		     cliQueue.cancel(ready); /* %%% is this desirable? */
#endif
		  }
		  else
		     cliQueue.remove(ready);
		  break;
	       }
	       case BUSY:
	       {
		  extern ProcessPipe* newProcess; /* %%% */

		  if (newProcess)
		  {
		     /* cliQueue.add_line(ready); %%% It's already in the queue, just don't remove it. */
		     ready->assign(newProcess);
		     newProcess = 0;
		     ready = 0;
		  }
		  else if (roll)
		  {
		     roll->add_line("warning: cli_execute returned BUSY but "
				    "no newProcess");
		     cliQueue.remove(ready);
		  }
		  break;
	       }
	    }
	 }
	 else
/* %%% Test sourcing empty file */
/* %%% rework to find correct agent */
	 {
	    /*
	     * Assign the line that is ready to execute to a
	     * process (this will either find the correct agent
	     * and pipe it the command, or fork a new process.
	     */
	    if (processList.assign(ready)==FAIL)
	    {
	       if (roll) roll->add_line("logonly: Canceling command queue (assign failed)");
	       cliQueue.cancel(ready);
	    }
	 }
      }

      flush_to_screen();

   } /* while (!done) */

   /*
    * The destructor for curs runs here and restores the
    * terminal settings when the program exits.
    */

   exit(EXIT_SUCCESS);
}
