/*                                             -*- c-file-style: "Ellemtel" -*-

`cli_cmd.c' - cli_cmd, cli_cap, cli_dup, and cli_cmd_nowait

This file is part of version 3 of the DIRECTOR cli client library.
Read the `License' file for terms of use and distribution.
Copyright 2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#ifndef VXWORKS
#include <sys/param.h>
#endif
#include <sys/stat.h>
#ifndef VXWORKS
#include <sys/time.h>
#endif
#include <errno.h>

#include "cli.h"
#include "clidefs.h"

/*
 * Provide wrapper function for Java.  See Makefile for definition of USE_JAVA.
 */
#ifdef USE_JAVA
#include <jni.h>

JNIEXPORT jint JNICALL
Java_neosvr_CliCmd_execute(JNIEnv *env, jobject obj, jstring instr) {
  const jbyte *str;
  jint retval = 0;
  str = (*env)->GetStringUTFChars(env, instr, NULL);
  retval = cli_cmd(str);
  (*env)->ReleaseStringUTFChars(env, instr, str);
  return retval;
}
#endif

/*
 * None of this is available under VxWorks.
 */
#ifndef VXWORKS
static PASSFAIL
cli_cmd_internal(const char* line, const char* capture_mode, int no_fork)
{
   struct sigaction oldsigusr1, oldsigusr2, oldsigchld;
   char junkbuf[1024];
   char capfile[1024] = "";
   int command_fd;
   int capture_fd = -1;
   pid_t child = 0;
   const char* director_version = getenv(DIRECTOR_ENV);
   const char* dirlevel_env = getenv(DIRECTOR_LEVEL_ENV);
   int director_level = 0;

   if (dirlevel_env) director_level = atoi(dirlevel_env);

   if (!director_level && director_version && *director_version)
   {
      /*
       * New version of director set DIRECTOR_LEVEL, and you *can*
       * in fact call a cli_cmd from within director (version >= 3.x only).
       */
      fprintf(stderr,
	      "This command/program can only be executed from a\n"
	      "regular unix shell window.  Do not run it inside\n"
	      "the director window.\n");
      return FAIL;
   }

   if (!no_fork)
      /*
        && ((getuid()==0 && geteuid()!=0))
	*
	* See comment below first ... 
	*
	* cli_cmd() now _always_ forks in order to support Java.
	* If it doesn't fork, it is possible for a multithreaded aAva
	* application to be waiting on two cli_cmd's at once with
	* the same PID.
	*/
   /*
		    || (capture_mode && *capture_mode)
    * I originally thought that forking a new PID for the sole
    * purpose of catching the completion signal from director was
    * only needed for setuid root programs (see explanation below.)
    * As it turns out, it is a good idea to do this at ANY time
    * to work around NFS problems.  Consider a C program that
    * calls cli_cap() several times with the same PID, 1234.
    * The named pipe 1234.cap is constantly being removed and
    * re-created in this case.  For some reason, if the file has
    * just been removed and re-created, it sometimes shows up
    * as not being there when director stats it.  The symptoms
    * are then that cli_cap's run as cli_cmd's by mistake.  To
    * try and solve this, *all* cli_cap's and cli_dup's will fork()
    * a new process for now.
    */
   {
      /*
       * Unless the *real* userid of the process that's using cli_cmd()
       * is the same as the user whose director we wish to send a command,
       * director won't be able to signal the process that the command
       * is complete.  This code attempts to change fork() a child for
       * the specific purpose of catching this signal.  It's used for
       * programs like RPM, which run setuid root.
       */
      uid_t u = geteuid();
      int rtn = FAIL;

      sigaction(SIGCHLD, NULL, &oldsigchld);
      if (oldsigchld.sa_handler==(void(*)(int))CLI_SIG_IGN)
	 cli_signal(SIGCHLD, CLI_SIG_DFL); /* Make sure the cli_cmd child
					 * stays around as a zombie just
					 * in case the parent hasn't even
					 * started the waitpid by the time
					 * the child exits.
					 */

      switch (child = fork())
      {
	 case -1:
	 {
	    perror("error: fork for cli_cmd failed");
	    break;
	 }
	 case 0:
	 {
	    setsid();
	    seteuid(0);
	    setuid(u);

	    /* Prevent infinite recursion, in case something here is broken
	     * (shouldn't happen) */
	    if (getuid()==0 && geteuid()!=0) { perror("setuid"); _exit(EXIT_FAILURE); }

	    /*
	     * Call cli_cmd again, in the child, with proper user ID.
	     */
	    _exit(cli_cmd_internal(line, 
				   capture_mode,
				   /*no_fork=*/1)==PASS?EXIT_SUCCESS:EXIT_FAILURE);
	 }
	 default:
	 {
	    int status = 0;
	    
	    if (TEMP_FAILURE_RETRY(waitpid(child, &status, WUNTRACED))!=child)
	    {
	       perror("error: waitpid failed for cli_cmd child");
	       kill(child, SIGTERM);
	    }
	    else
	    {
	       /*
		* Try to get the exit status of the child, which will indicate
		* whether this is a PASS or a FAIL
		*/
	       if (WIFEXITED(status) && WEXITSTATUS(status)==EXIT_SUCCESS)
		  rtn = PASS;
	       else
		  kill(child, SIGTERM);
	    }
	 }
      }
      /*
       * Return the old behavior of SIGCHLD if we had to modify it.
       */
      if (oldsigchld.sa_handler==(void(*)(int))CLI_SIG_IGN)
	 sigaction(SIGCHLD, &oldsigchld, NULL);

      return rtn;
   }
   else
   {
      if (geteuid() != getuid())
      {
	 fprintf(stderr,
		 "Effective and real userid's are not the same.  Not sending command.\n");
	 return FAIL;
      }

      sprintf(junkbuf, "%s/.director/command.np", cli_gethome());
      
      if ((command_fd = 
	   open(junkbuf, O_WRONLY|O_NONBLOCK, 0))==-1)
      {
	 const char* s = line;

	 /*
	  * In case of the special builtin command "running", don't print
	  * bunch of errors if director is not running yet, because this
	  * is what the builtin is supposed to test.
	  */
	 while (strchr("@!-*", *s)) s++;
	 if (!strncmp(s, "running", 7))
	    if (errno==ENOENT || errno==ENODEV
#ifdef ENXIO
		|| errno==ENXIO
#endif
	       ) return FAIL; /* Return FAIL silently */

	 perror(junkbuf);
	 if (errno==ENOENT)
	    fprintf(stderr, "You do not appear to have a running director.\n");
	 else if (errno==ENODEV
#ifdef ENXIO
		  || errno==ENXIO
#endif
	    )
	    fprintf(stderr, "Director may be running on a different host.\n");
	 
	 return FAIL;
      }

      /*
       * Make sure old named pipes are cleaned up for this PID
       */
      sprintf(capfile, "%.800s/.director/%d.cap", cli_gethome(), (int)getpid());
      unlink(capfile);
      sprintf(capfile, "%.800s/.director/%d.dup", cli_gethome(), (int)getpid());
      unlink(capfile);
      *capfile = '\0';

      /*
       * Create new named pipe for capture, if desired.
       */
      if (capture_mode && *capture_mode)
      {
	 sprintf(capfile, "%.800s/.director/%d.%.50s",
		 cli_gethome(), (int)getpid(), capture_mode);
	 
	 if (mknod(capfile, S_IFIFO|S_IWUSR|S_IRUSR, 0)!=0)
	    perror("mknod(capture pipe)");
	 else
	 {
	    capture_fd = open(capfile, O_RDWR|O_NONBLOCK);
	 }
      }

      sigaction(SIGUSR1, NULL, &oldsigusr1);
      sigaction(SIGUSR2, NULL, &oldsigusr2);
      
      cli_signal(SIGUSR1, CLI_SIG_TRAP);
      cli_signal(SIGUSR2, CLI_SIG_TRAP);
      
      if (director_level)
	 sprintf(junkbuf, "{%d.%d}%s\n", (unsigned int)getpid(), director_level, line);
      else
	 sprintf(junkbuf, "{%d}%s\n", (unsigned int)getpid(), line);

      if (cli_write(command_fd, junkbuf, strlen(junkbuf), 3000)==FAIL)
      {
	 perror("error: write to command pipe failed");
      }
      close(command_fd);
      
      while (!cli_signal_event(SIGUSR1) && 
	     !cli_signal_peek(SIGUSR2))
      {
	 if (capture_fd == -1)
	 {
	    sleep(1);
	 }
	 else
	 {
	    /*
	     * cli_cap() waits for the feedback from the command on
	     * the named pipe that it temporarily creates in ~/.director/
	     */
	    char buf[80];
	    int nbytes;
	    struct timeval tv = { 1, 0 };
	    fd_set fds;

	    FD_ZERO(&fds);
	    FD_SET(capture_fd, &fds);
	    if (SELECT(capture_fd + 1, &fds, 0, 0, &tv)==1)
	    {
	       switch ((nbytes = read(capture_fd, buf, sizeof(buf))))
	       {
		  case -1: if (errno==EINTR || errno==EAGAIN) { break; } perror(capfile);
		  /* And fall through... */
		  case 0: close(capture_fd); capture_fd = -1; break;
		  default: if (cli_write(STDOUT_FILENO, buf, nbytes, 1000)==FAIL)
		     perror("error: write to stdout failed");
	       }
	    }
	 }
      }
      
      sigaction(SIGUSR1, &oldsigusr1, NULL);
      sigaction(SIGUSR2, &oldsigusr2, NULL);

      /*
       * Flush any other input waiting on the capture pipe before removing it.
       */
      if (capture_fd != -1)
      {
	 char buf[80];
	 int nbytes;

	 while ((nbytes = read(capture_fd, buf, sizeof(buf)))>0)
	    if (cli_write(STDOUT_FILENO, buf, nbytes, 1000)==FAIL)
	       perror("error: write to stdout failed");
	 if (nbytes==-1 && errno!=EAGAIN)
	    perror("read from capture fd failed");
	 close(capture_fd);
	 capture_fd = -1;
      }

      /*
       * Remove the temporary capture node.
       */
      if (*capfile) unlink(capfile);

      if (cli_signal_event(SIGUSR2)) return FAIL;
      return PASS;
   }
}

PASSFAIL
cli_cmd(const char* line)
{
   return cli_cmd_internal(line, "", 0);
}

PASSFAIL
cli_cap(const char* line)
{
   return cli_cmd_internal(line, "cap", 0);
}

PASSFAIL
cli_dup(const char* line)
{
   return cli_cmd_internal(line, "dup", 0);
}


PASSFAIL
cli_cmd_nowait(const char* line)
{
   char junkbuf[1024];
   int command_fd;
   
   if (getenv(DIRECTOR_ENV)!=0 && *getenv(DIRECTOR_ENV)!='\0')
   {
      /*
       * I guess it is possible to hack cli_cmd_nowait so it
       * uses director_level like cli_cmd_internal, but it
       * doesn't seem very useful to have nested cli_cmd_nowait's
       * that can be called from within director.  If you're getting
       * the error below, though, maybe there is a use for it.
       */
      fprintf(stderr,
	      "This command/program can only be executed from a\n"
	      "regular unix shell window.  Do not run it inside\n"
	      "the director window.\n");
      return FAIL;
   }

   sprintf(junkbuf, "%s/.director/command.np", cli_gethome());

   if ((command_fd = 
	open(junkbuf, O_WRONLY|O_NONBLOCK, 0))==-1)
   {
      perror(junkbuf);
      if (errno==ENOENT)
	 fprintf(stderr, "You do not appear to have a running director.\n");
      else if (errno==ENODEV
#ifdef ENXIO
	       || errno==ENXIO
#endif
	 )
	 fprintf(stderr, "Director appears to be running on a different host.\n");
      
      return FAIL;
   }

   sprintf(junkbuf, "%s\n", line);
   if (cli_write(command_fd, junkbuf, strlen(junkbuf), 1000)==FAIL)
   {
      perror("write");
      close(command_fd);
      return FAIL;
   }
   else
   {
      close(command_fd);
      return PASS;
   }
}
#endif

/* End of cli_cmd.c */
