/*                                             -*- c-file-style: "Ellemtel" -*-

`cli_comlist.c' - cli_init, cli_help, cli_search, cli_execute

This file is part of version 3 of the DIRECTOR cli client library.
Read the `License' file for terms of use and distribution.
Copyright 2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#ifndef VXWORKS
#include <pwd.h>
#endif
#include <sys/types.h>
#include <sys/wait.h>
#ifndef VXWORKS
#include <sys/param.h>
#endif
#include <sys/stat.h>
#ifndef VXWORKS
#include <sys/time.h>
#endif
#include <errno.h>

#ifdef VXWORKS
#include <ioLib.h> /* for ioctl() */
#endif

#include "cli.h"
#include "clidefs.h"

static Command* comlist_ = 0;

#ifdef USE_READLINE
/* Generator function for command completion.  STATE lets us know whether
   to start from scratch; without any state (i.e. STATE == 0), then we
   start at the top of the list. */
static char*
command_generator (char* text, int state)
{
   static int list_index, len;
   char *name;
   
   /* If this is a new word to complete, initialize now.  This includes
      saving the length of TEXT for efficiency, and initializing the index
      variable to 0. */
   if (!state)
   {
      list_index = 0;
      len = strlen (text);
   }
   
   /* Return the next name which partially matches from the command list. */
   while (comlist_ &&
	  (name = comlist_[list_index].name)!=0)
   {
      list_index++;
      
      if (strncmp (name, text, len) == 0)
	 return (strdup(name));
   }
   
   /* If no names matched, then return NULL. */
   return ((char *)NULL);
}


/* Attempt to complete on the contents of TEXT.  START and END show the
   region of TEXT that contains the word to complete.  We can use the
   entire line in case we want to do some simple parsing.  Return the
   array of matches, or NULL if there aren't any. */
static char **
try_completion (char* text, int start, int end UNUSED)
{
   char **matches;
   
   matches = (char **)NULL;
   
   /* If this word is at the start of the line, then it is a command
      to complete.  Otherwise it is the name of a file in the current
      directory. */
   if (start == 0)
      matches = completion_matches(text, command_generator);
   
   return (matches);
}
#endif

#ifndef SIGWINCH
#define SIGWINCH (0)
#endif

PASSFAIL
cli_init(const char* name, Command* comlist, int director)
{
   static int initialized = 0;

   comlist_ = comlist;

   if (!initialized)
   {
      initialized = 1;
      /*
       * This must be done at the start of the program.  Setting this
       * ensures diagnostic messages will never be buffered as long as
       * you have newlines ('\n') at the end of all your printf()'s.
       */
      if (!isatty(STDOUT_FILENO) && setvbuf(stdout, NULL, _IOLBF, BUFSIZ)!=0)
      {
	 perror("setvbuf failed");
      }


      /*
       * For most things, director wants non-blocking i/o, but here
       * we want BLOCKING i/o.  This makes sure it is turned on, and after
       * that we only read from stdin using low-level i/o.
       */
      if (director>0)
      {
#ifdef VXWORKS
	 int flags = 0;
         if (ioctl(STDIN_FILENO, FIONBIO, (int)(&flags))==-1)
            perror("ioctl");
#else
	 if (fcntl(STDIN_FILENO, F_SETFL,
		   fcntl(STDIN_FILENO, F_GETFL, 0)&~O_NONBLOCK)==-1)
	    perror("fcntl");
#endif
      }

#ifndef VXWORKS
      /*
       * To prevent the program from being accidentally run without the
       * director, you may want to include the following.
       */
      if (director>0 && (getenv(DIRECTOR_ENV)==0 || atoi(getenv(DIRECTOR_ENV))<director))
      {
	 printf("error: this program should be run with director version %d.\n"
		"If you are unable to operate with director, it is possible\n"
		"to operate in a fallback mode by running `%s' directly, but\n"
		"with "DIRECTOR_ENV"=%d set in your shell environment.\n",
		director, name, director);
	 return FAIL;
      }
#endif

      /*
       * Argh... we might be running as an rcmd() through director, and if
       * we are rcmded to a different machine type, important signals like
       * SIGINT, SIGQUIT, and SIGTSTP might not have the same assignments.
       */
      if (director>0)
      {
	 char* cfhtdebug = getenv("CFHTDEBUG");

	 /*
	  * Get the current debugging setting from Pegasus, if applicable.
	  * For each agent, this can also be toggled later with
	  * "agentname.debug on" or "agentname.debug off" in director.
	  */
	 if (cfhtdebug && *cfhtdebug)
	    cli_debug = cli_bool(cfhtdebug);

	 printf("\rsignalmap: %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
		SIGHUP, SIGINT, SIGQUIT, SIGILL, SIGTRAP, SIGABRT, SIGFPE, SIGKILL, SIGSEGV,
		SIGPIPE, SIGALRM, SIGTERM, SIGUSR1, SIGUSR2, SIGSTOP, SIGTSTP, SIGCONT,
		SIGCHLD, SIGWINCH);

#ifndef NO_CFHTLOG
#if 0
	 /*
	  * This generates a dependency between libcfht and libcli.  This
	  * isn't necessarily a bad thing, but because we now have a shared
	  * library version for java stuff, I didn't want to build a shared
	  * libcfht to go with it.  This call really shouldn't be needed
	  * anyway, unless an agent makes other calls to libcfht, in which
	  * case it should take care of CFHT_LOG_ID itself. %%%
	  *
	  * Actually, the cfht_log(LOG_ID) is no longer needed, since I've
	  * modified libcfht to send messages to stderr when $DIRECTOR is
	  * defined.  This way, cfht_log doesn't need to know who is logging,
	  * and director will put the appropriate time stamps and names
	  * on the messages before sending them back to the system log.
	  *
	  * So... this line is not needed anymore.
	  */
	 cfht_log(CFHT_MAIN, CFHT_LOG_ID, (char*)name);
#endif
#endif
      }
   }

#ifdef USE_READLINE
   /* Allow conditional parsing of the ~/.inputrc file. */
   rl_readline_name = name;

   /* Tell the completer that we want a crack first. */
   rl_attempted_completion_function = (CPPFunction *)try_completion;
#endif
   
   return PASS;
}
     
PASSFAIL
cli_parse_signalmap(int* m, const char* s)
   /*
    * Parse the line that is sent by cli_init() above.
    */
{
   if (sscanf(s, "signalmap: %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
	      &m[SIGHUP], &m[SIGINT], &m[SIGQUIT], &m[SIGILL], 
	      &m[SIGTRAP], &m[SIGABRT], &m[SIGFPE], &m[SIGKILL], &m[SIGSEGV],
	      &m[SIGPIPE], &m[SIGALRM], &m[SIGTERM], &m[SIGUSR1], 
	      &m[SIGUSR2], &m[SIGSTOP], &m[SIGTSTP], &m[SIGCONT],
	      &m[SIGCHLD], &m[SIGWINCH])==19)
      return PASS;
   return FAIL;
}

char*
cli_stripwhite(char* string)
{
   char *s, *t;
   
   for (s = string; isspace(*s); s++) ;
   
   if (*s == 0) return s;
   
   t = s + strlen (s) - 1;
   while (t > s && isspace(*t)) t--;
   *++t = '\0';
   
   return s;
}

/* Look up NAME as the name of a command, and return a pointer to that
   command.  Return a NULL pointer if NAME isn't a command name. */
static Command*
find_command(const char* name, int namelen)
{
   int i;
   
   if (comlist_==NULL) return ((Command*)NULL);

   for (i = 0; comlist_[i].name; i++)
   {
      int len;
      char* space = strchr(comlist_[i].name, ' ');
      
      if (space) len = space - comlist_[i].name;
      else len = strlen(comlist_[i].name);
      
      if (namelen==len && strncmp(name, comlist_[i].name, len) == 0)
         return &(comlist_[i]);
   }
   
   return ((Command*)NULL);
}

Command*
cli_search(const char* line, Command* comlist)
{
   Command* com;
   const char* name = line;
   int namelen = 0;

   if (line==NULL) return NULL;

   comlist_ = comlist; /* save comlist_ for the other routines */
   
   /* Isolate the command word. */
   while (*name && isspace(*name)) name++;
   while (name[namelen] && !isspace(name[namelen])) namelen++;

   com = find_command(name, namelen);
   return com;
}

PASSFAIL
cli_execute(const char* linearg)
{
   Command *com;
   int i;

   if (linearg && linearg[0]==CHAR_SYSTEM_ESC && linearg[1]!=CHAR_SYSTEM_ESC)
      return cli_system("", linearg+1);

   com = cli_search(linearg, comlist_);

   if (com==NULL)
   {
      if (linearg && *linearg)
	 printf("error: Invalid command: `%s'; ", linearg);
      printf("Type `help' for information\n");
      return cli_setprompt(FAIL);
   }
   
   if (!(com->func))
   {
      if (linearg && *linearg)
	 printf("error: Command `%s' requires an external script.\n", linearg);
      printf("error: Script is missing, or could not be accessed.\n");
      return cli_setprompt(FAIL);
   }
   /* Get argument to command, if any. */
   i = 0;
   while (linearg[i] && isspace(linearg[i])) i++;
   while (linearg[i] && !isspace(linearg[i])) i++;
   while (isspace (linearg[i])) i++;
     
   /* Call the function. */
   return cli_setprompt((*(com->func)) (linearg + i));
}


PASSFAIL
cli_help(const char* arg)
{
   int i, printed=0;
   int longest = 14;

   if (comlist_==NULL)
   {
      printf("\nNo commands available (cli_init has not been called).\n");
      return FAIL;
   }

   if (arg && !*arg)
      printf(
"*---[ Command Set ]-------------------------------------------------------*\n");

   for (i = 0; comlist_[i].name; i++)
      if (strlen(comlist_[i].name) > longest && comlist_[i].help)
	 longest = strlen(comlist_[i].name);

   for (i = 0; comlist_[i].name; i++)
   {
      if (!arg || !*arg || (strncmp(arg, comlist_[i].name, strlen(arg)) == 0))
      {
	 if (!comlist_[i].help)
	 {
	    if (!arg || !*arg) printf("%s\n", comlist_[i].name);
	 }
	 else if ((arg && *arg) || (strncmp(comlist_[i].help, "Synonym", 7)!=0 &&
		      *comlist_[i].help!='('))
	 {
	    printed++;
	    printf("%s%*s: %s\n",
		   comlist_[i].name, 
		   (int)(longest - strlen(comlist_[i].name)), "",
		   comlist_[i].help);
	 }
      }
   }

   if (arg && !*arg)
   {
      printf("builtin%*s: Displays help for director builtin options\n",
	     (int)(longest - 7), "");
      printf("!...%*s: Passes ... to a unix shell (e.g. `!ls -l *.fits')\n",
	     (int)(longest - 4), "");
      printf(
"*-------------------------------------------------------------------------*\n");
   }
     
   if (!printed)
   {
      printf("No such command, `%s'.  Possibilties are:\n", arg);
     
      for (i = 0; comlist_[i].name; i++)
      {
	 /* Print in six columns. */
	 if (printed == 6)
	 {
	    printed = 0;
	    printf ("\n");
	 }
	
	 if ((!comlist_[i].help ||
	      ((strncmp(comlist_[i].help, "Synonym", 7)!=0 &&
		*comlist_[i].help!='('))) &&
	     comlist_[i].name[0] != '*')
	 {
	    int len;
	    char* space;

	    space = strchr(comlist_[i].name, ' ');
	    if (!space) len = strlen(comlist_[i].name);
	    else len = space - comlist_[i].name;
	    if (len > 20) len = 20;
	    printf ("%.*s\t", len, comlist_[i].name);
	    printed++;
	 }
      }
      
      if (printed) printf("\n");
   }

   return PASS;
}

/* End of cli_comlist.c */
