/*                                             -*- c-file-style: "Ellemtel" -*-

`cli_signal.c' - Same as libcfht's cfht_signal.

This file is part of version 3 of the DIRECTOR cli client library.
Read the `License' file for terms of use and distribution.
Copyright 2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

/*
 * $Id: cli_signal.c,v 1.1 2001/03/08 17:59:39 isani Locked $
 *
 * DESCRIPTION
 *
 * Call cli_signal() just as you would call signal(), with two arguments
 * (the signal, and a function pointer to your handler.)  See the man page
 * for details and examples.
 */

#include <stdio.h>
#include <signal.h>

#include "cli.h"

RCSID("$Id: cli_signal.c,v 1.1 2001/03/08 17:59:39 isani Locked $");

/*
 * This table contains an entry for each signal.  When a signal handler
 * is installed, cli_signal clears the flag for that signal.  If the
 * builtin signal handler is used, then it gets set to 1 and this can
 * be tested with cli_signal_peek(), or tested and cleared with
 * cli_signal_event().
 */
#ifndef NSIG
#define NSIG _NSIGS
#endif

static volatile BOOLEAN builtin_signal_state[NSIG];

/*
 * Use this function if you for some reason need to figure out
 * if a signal is currently blocked or not.
 */
BOOLEAN
cli_signal_isblocked(int signo)
{
   sigset_t mask;

   sigprocmask(SIG_SETMASK, 0, &mask);
   if (sigismember(&mask, signo)) return TRUE;
   else return FALSE;
}

/*
 * Use this function to start blocking a particular signal.  As long
 * as the signal is blocked, caught signals will be queued up.  (But
 * you can only determine whether a signal happened, and not how many
 * times it might have happened.)  If you actually want to DISCARD
 * signals, use cli_signal(signo,SIG_IGN).
 */
void
cli_signal_block(int signo)
{
    sigset_t mask;
    
    sigemptyset(&mask);
    sigaddset(&mask, signo);
    sigprocmask(SIG_BLOCK, &mask, 0);
}

/*
 * Use this function to allow the handler for signo to run again.  If
 * and events are queued up, the handler will run immediately.
 */
void
cli_signal_unblock(int signo)
{
    sigset_t mask;
    
    sigemptyset(&mask);
    sigaddset(&mask, signo);
    sigprocmask(SIG_UNBLOCK, &mask, 0);
}


/*
 *
 * This function sets up the signal catching for any signal.
 * `signo' is the signal you want to trap.  If you want to block
 * the signal initially, you might want to block it first, then
 * call cli_signal() to install the handler, and then unblock
 * when you're ready to handle the signal.
 *
 * If desired, a default builtin handler can be used
 * which just records whether an event (or multiple events ... there
 * is no reliable way to distinguish) has occurred.  If you go this
 * route, you should use the function cli_signal_event() to check
 * if the signal happened.
 */
void
cli_signal(int signo, PFV handler)
{
   struct sigaction vec_trap;
   
   if (signo >= NSIG)
   {
      printf("\rwarning: (cli_signal:%d) signo %d out of range (0-%d)\n",
	     __LINE__, signo, NSIG);
      return;
   }
   builtin_signal_state[signo] = FALSE;
   vec_trap.sa_handler = (void(*)(int))handler;
   sigemptyset(&vec_trap.sa_mask);
#ifdef SA_INTERRUPT
   /*
    * If this particular macro is defined, then it means we are compiling
    * on a BSD system where the default (even with sigaction()) might be to
    * restart system calls after handling a signal.  Setting this flag in
    * the sa_flags field tells BSD systems to return EINTR right away
    * instead, whenever any system call is interrupted.
    */
   vec_trap.sa_flags = SA_INTERRUPT;
#else
   vec_trap.sa_flags = 0;
#endif /* SA_INTERRUPT */
   
   /*
    * install the handler
    */
   if (sigaction(signo, &vec_trap, NULL) != 0) {
      printf("warning: (cli_signal:%d) sigaction failed to install handler\n",
	     __LINE__);
   }
}

/*
 *
 * cli_signal_handler -- saves a flag so cli_signal_event() can tell if
 *                        the signal signo has been trapped yet.
 *
 */

void
cli_signal_handler(int signo)
{
    builtin_signal_state[signo] = TRUE;
}

/*
 *
 *  NAME:  cli_signal_event
 *
 *  This function tests and clears the global flag "builtin_signal_state"
 *  
 *  The idea here is that if a SIGXYZ was set up and subsequently trapped in
 *  the handler, the global flag would have been set. When a process that is
 *  pausing (see pause(2)) on this event is awaken, it should first check 
 *  that the wakeup call was really for SIGCXYZ.
 *
 *  INVOKED: while not (cli_signal_event (SIGXYZ))
 *               sleep
 *           end while
 *           proceed with state checking
 *
 *  FUNCTION RETURNS:
 *      TRUE - global flag set, indicating at least 1 SIGXYZ event
 *      FALSE - SIGXYZ not caught since last time you called cli_signal_event
 *
 *  GLOBALS: builtin_signal_state[]
 */

BOOLEAN
cli_signal_event(int signo)
{
   if (signo >= NSIG)
   {
      printf("warning: (cli_signal_event:%d) signo %d out of range (0-%d)\n",
	     __LINE__, signo, NSIG);
      return TRUE;
   }
   if (builtin_signal_state[signo] == TRUE)
   {
      builtin_signal_state[signo] = FALSE;
      return TRUE;
   }
   else
   {
      return FALSE;
   }
}

/*
 * NAME: cli_signal_peek - see if there's been an event, but don't
 *                          reset the indicator flag
 *  INVOKED: if(cli_signal_peek(signo)) printf("we have had a signo\n");
 *
 *  FUNCTION RETURNS:
 *      TRUE if signo has happened since last of cli_signal or cli_signal_event
 *      FALSE if signal has not happened since install or reset
 *      and garbage if no handler installed yet
 *      and TRUE if signo out of range
 */

BOOLEAN
cli_signal_peek(int signo)
{
   if (signo >= NSIG)
   {
      printf("waring: (cli_signal_peek:%d) signo %d out of range (0-%d)\n",
	     __LINE__, signo, NSIG);
      return TRUE;
   }
   return (builtin_signal_state[signo]);
}

/* end of cli_signal.c */
