/*                                             -*- c-file-style: "Ellemtel" -*-

`cli_syscalls.c' - cli_system, cli_getcwd, cli_cd, cli_gethome, cli_write

This file is part of version 3 of the DIRECTOR cli client library.
Read the `License' file for terms of use and distribution.
Copyright 2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#ifndef VXWORKS
#include <pwd.h>
#endif
#include <sys/types.h>
#include <sys/wait.h>
#ifndef VXWORKS
#include <sys/param.h>
#endif
#include <sys/stat.h>
#ifndef VXWORKS
#include <sys/time.h>
#endif
#include <errno.h>

#include "cli.h"

#ifdef VXWORKS
#include <selectLib.h>
#endif

PASSFAIL
cli_system(const char* command, const char* args)
{
#ifdef VXWORKS
   printf("cli_system not supported under VxWorks\n");
   return cli_setprompt(FAIL);
#else
   char* junkbuf;
   const char* argv[4];
   pid_t child;
   int status;

   junkbuf = (char*)malloc((command?strlen(command):0) +
			   (args?strlen(args):0) + 2);

   if (!junkbuf)
   {
      perror("malloc");
      return cli_setprompt(FAIL);
   }
      
   sprintf(junkbuf, "%s%s%s", 
	   command?command:"", 
	   (command&&*command&&args&&*args)?" ":"",
	   args?args:"");

   argv[0] = "/bin/sh";
   argv[1] = "-c";
   argv[2] = junkbuf;
   argv[3] = NULL;

   if ((child=fork()) == (pid_t)-1)
   {
      perror("(cli_system)fork");
      free(junkbuf);
      return cli_setprompt(FAIL);
   }

   if (child == (pid_t)0)
   {
      setsid();
      alarm(0);
      execvp(argv[0], (char**)argv);
      free(junkbuf);
      _exit(EXIT_FAILURE);
   }
   
   switch (TEMP_FAILURE_RETRY(waitpid(child, &status, WUNTRACED)))
   {
      case -1: perror("(cli_system)waitpid"); kill(child, SIGKILL); 
      free(junkbuf);
      return cli_setprompt(FAIL);
      case 0: fprintf(stderr,"cli_system: no child process\n");
      free(junkbuf);
      return cli_setprompt(FAIL);
      default:
      {
	 if (!WIFEXITED(status))
	 {
	    if (WIFSIGNALED(status))
	    {
	       fprintf(stderr, "error: program exited on signal #%d\n",
		       WTERMSIG(status));
	       free(junkbuf);
	       return cli_setprompt(FAIL);
	    }
	    else if (WIFSTOPPED(status))
	    {
	       fprintf(stderr, "error: program stopped\n");
	       free(junkbuf);
	       return cli_setprompt(FAIL);
	    }
	    else
	    {
	       fprintf(stderr, "error: unknown problem with child process\n");
	       free(junkbuf);
	       return cli_setprompt(FAIL);
	    }
	 }
	 else
	 {
	    if (WEXITSTATUS(status)!=EXIT_SUCCESS)
	    {
	       child = -1;
	       if (WEXITSTATUS(status)==127)
	       {
		  fprintf(stderr, "error: failed to execute `%s'\n",
			  junkbuf);
	       }
	       else
	       {
		  fprintf(stderr, "error: `%s' not successful (exitcode %d)\n", 
			  junkbuf, WEXITSTATUS(status));
	       }
	       free(junkbuf);
	       return cli_setprompt(FAIL);
	    }
	    /* else, normally exiting child... */
	    free(junkbuf);
	    return cli_setprompt(PASS);
	 }
      }
   }
#endif
}

static int newcwd = 1;

const char*
cli_getcwd(void)
{
#if ( MAXPATHLEN > 1024 )
#define CLI_PATHLEN MAXPATHLEN
#else
#define CLI_PATHLEN 1024
#endif
   static char cwd[CLI_PATHLEN];

   if (newcwd)
   {
#ifdef SUNOS
      extern char* getwd(const char*);
      /*
       * getcwd() calls popen() on Sun's, which breaks if cfht_exec
       * has its handler installed, since cfht_exec waits on the popen'd
       * child before pclose() does.  Symptoms are: program locks up
       * in wait() called from pclose() (if you use getcwd).
       */
      getwd(cwd);
#else
      if (!getcwd(cwd, sizeof(cwd))) { perror("getcwd"); return "[getcwd failed]"; }
#endif
      newcwd = 0;
   }

   /*
    * Some hacks to transition from actual paths to the preferred
    * paths (which should work on remote machines.)  This hack is
    * required either here or runon.
    */
   if (!strncmp(cwd, "/tmp_mnt", 8))         return cwd+8;
   if (!strncmp(cwd, "/usr/local/cfht", 15)) return cwd+10;
   if (!strncmp(cwd, "/local/", 7))          return cwd+6;
   /*
    * === Hack to help transition from /usr/local/cfht/ (which
    * is not on _all_ machines, to /cfht/, which is automounted
    * to all machines at cfht.
    */
   else return cwd;
}

PASSFAIL
cli_cd(const char* args)
{
   newcwd = 1;
   if (!args || !*args)
      printf("status: Type `cd newpath' to change directories\n");
   else 
   {
      char* newdir = cli_arg1(args);

      if (chdir(newdir)==-1) 
      {
	 perror(newdir);
	 free(newdir);
	 return FAIL; 
      }
      free(newdir);
   }

   printf("status: Current working directory is `%s'\n", cli_getcwd());
   return PASS;
}

PASSFAIL
cli_cd_internal(const char* args)
{
   char* newdir;

   if (!args || !*args) return FAIL;
   newcwd = 1;
   newdir = cli_arg1(args);
   if (chdir(newdir)==-1) 
   {
      /* perror(newdir); */
      free(newdir);
      return FAIL; 
   }
   free(newdir);
   return PASS;
}

const char*
cli_gethome(void)
{
#ifdef VXWORKS
   return "/tmp"; /* === Not using home directories on VxWorks. */
#else
   static int last_euid = -1;
   static char* home = 0;
   int new_euid = geteuid();
   struct passwd* pwd;

   if (new_euid!=last_euid)
   {
      /*
       * Get the user's home directory.  getenv("HOME") would
       * be faster, but if someone has su'd to another account
       * to run clicmd as them, the contents of HOME could be wrong.
       * Don't trust it and go to the passwd table instead.
       */
      if (home) free(home);
      last_euid = new_euid;
      pwd = getpwuid(new_euid);
      if (!pwd) home = cli_strdup("/");
      else home = cli_strdup(pwd->pw_dir);
   }
   return home;
#endif
}


PASSFAIL
cli_write(int fd, const void* data, size_t len, int timeout_msec)
{
   int n;
   fd_set fds;
   struct timeval tv;

   while (len>0)
   {
      tv.tv_sec = timeout_msec / 1000;
      tv.tv_usec = (timeout_msec%1000) * 1000;
      FD_ZERO(&fds);
      FD_SET(fd, &fds);
      switch (TEMP_FAILURE_RETRY(SELECT(fd+1, 0, &fds, 0, &tv)))
      {
	 case -1: return FAIL;
	 case 0: errno=EAGAIN; return FAIL;
	 default:
	 {
	    if ((n = TEMP_FAILURE_RETRY(write(fd, (char*)data, len))) < 0)
	    {
	       if (errno == EAGAIN) continue;
#ifdef EWOULDBLOCK
	       if (errno == EWOULDBLOCK) continue;
#endif
	       return FAIL; /* All other errors are real failures. */
	    }
	    len -= n;
	    data += n;
	 }
      }
   }
   return PASS;
}

/* End of cli_syscalls.c */
