/*                                             -*- c-file-style: "Ellemtel" -*-

`runon.c' - Replacement for the remsh program

This file is part of version 3 of the Replacement for the remsh program.
Read the `License' file for terms of use and distribution.
Copyright 2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/socket.h>
#ifndef VXWORKS
#include <sys/termios.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <pwd.h>
#endif
#include <netdb.h>
#include <time.h>
#include <errno.h>

#include "cli/cli.h"

/*
 * Assume the connection is dead and exit after being unable to write
 * to the rcmd file descriptor for this many milliseconds.
 */
#define RCMD_TIMEOUT (10*1000)

static int runon_client(int argc, const char* argv[], const char* envp[]);
static int runon_server(void);

static void talk(int* fd_in1, int* fd_out1, int* fd_err1,
		 int* fd_in2, int* fd_out2, int* fd_err2,
		 int* client_exitstatus);
static void set_non_blocking(int fd);
static int strfreq(const char* str, char c);
static void write_or_die(int fd, const void* data, int len);
static void write_quoted(int fd, const char* string);
static void write_signal(int signo);
static void default_sigmap(void);

static const char* progpath = 0;
static const char* progname = "runon";
static const char* username = 0;
static const char* hostname = 0;
static const char* serverpath = 0;
static const char* output = 0;
static int nflag = 0, dflag = 0, xflag = 0;

/*
 * Stuff for running the terminal in raw/no echo mode.
 * (Not quite the same as a tty, but... works for some programs.)
 */
#ifndef VXWORKS
static struct termios tio, tioorig;
static int need_restore = 0;

static void
tty_restore(void)
{
   int fd;

   if (!need_restore) return;

   fd = open("/dev/tty", O_RDWR|O_NOCTTY);
   if (fd==-1) { perror("open(\"/dev/tty\"...)"); fd=STDIN_FILENO; }
   if (tcsetattr(fd, TCSAFLUSH, &tioorig)<0)
      perror("tcsetattr");
}
#endif

#ifdef VXWORKS
static void
tty_raw(void)
{
   printf("warning: raw option not supported on VxWorks\n");
}
#else
static void
tty_raw(void)
{
   if (tcgetattr(STDIN_FILENO, &tioorig)<0) return;

   need_restore = 1;

   /*
    * Turn off echo and put the terminal in raw mode
    */
   tio = tioorig;
   tio.c_iflag &= ~(ICRNL | IUCLC);
   tio.c_lflag &= ~(ICANON | ECHO);
   tio.c_cflag |= CS8|CREAD;
   tio.c_cc[VMIN] = 1;
   tio.c_cc[VTIME] = 0;
   if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &tio)!=0)
      perror("tcsetattr");
   atexit(tty_restore);
}
#endif

#define MAX_FDNAME 256
/*
 * This is not critical.  If file descriptor numbers go higher than
 * MAX_FDNAME, runon won't assign a debugging name to the descriptor,
 * but everything will still work.
 */

static const char* fdname_[MAX_FDNAME];

static void
fdname_set(int fd, const char* name)
{
   if (fd < 0 || fd >= MAX_FDNAME)
   {
      fprintf(stderr, "logonly: (%s) fd %d is out of range; `%s' not assigned.\n",
	      progname, fd, name);
   }
   else
   {
      fdname_[fd] = name;
   }
}

static const char*
fdname(int fd)
{
   static char buf[16];

   if (fd < 0 || fd >= MAX_FDNAME || fdname_[fd]==0)
   {
      sprintf(buf, "%d", fd);
      return buf;
   }
   return fdname_[fd];
}

int
main(int argc, const char* argv[], const char* envp[])
{
   int c, rshflag = 0, errflg = 0, argpos = 0;

   extern char *optarg;
   extern int optind;
   char* colon = 0;

   memset(fdname_, 0, sizeof(fdname_));

   progpath = argv[0];
   if (!(colon = strchr(progpath, ':')) && 
       (progname = strrchr(progpath, '/'))) progname++;
   else progname = progpath;

   if (strcmp(progname, "rsh") &&
       strcmp(progname, "runon") &&
       strncmp(progname, "runon-", 6) &&
       strcmp(progname, "remsh"))
   {
      hostname = progname;
      progpath = "runon";
   }
   else
   {
      /*
       * First check for server invocation
       */
      if (argc==2 && !strcmp(argv[1], "-s"))
      {
	 progname = "runon -s";
	 exit(runon_server());
      }

      rshflag = 1;
   }

   if (hostname==0 && argc >= 2 && *argv[1]!='-')
   {
      hostname = argv[1];
      argpos++;
   }

   /*
    * The "+" at the beginning of the string is needed
    * by Linux.  It's exact meaning seems to be undocumented,
    * but without it getopt() continues to munch on options
    * that belong with the remote command rather than stopping
    * at the first non '-' argument.
    */
   
   if (colon)
   {
      optind = 1;
   }
   else
   {
      while ((c = getopt(argc - argpos,
			 (char**)argv + argpos,
			 "+8KLdeirwl:np:o:x"))!=-1)
      {
	 switch (c)
	 {
	    case '8':
	    case 'K':
	    case 'L':
	    case 'e':
	    case 'i':
	    case 'w': break;
	    case 'r': tty_raw(); break;
	    case 'd': dflag = 1; break;
	    case 'n': nflag = 1; break;
	    case 'x': xflag = 1; break;
	    case 'l': username = optarg; break;
	    case 'p': serverpath = optarg; break;
	    case 'o': output = optarg; break;
	    default: errflg++;
	 }
      }
      optind += argpos;
   }

   if (!errflg && (hostname || (hostname = argv[optind++])))
   {
      char* p;
      char* q;

      /*
       * This allows for username@hostname:command sytax, not supported
       * in regular remsh, but used by director.
       */

      if ((p = colon))
      {
	 q = (char*)malloc(p - hostname + 1);
	 if (p - hostname) memcpy(q, hostname, p - hostname);
	 q[p - hostname] = '\0';
	 hostname = q;
	 (const char*)(argv[--optind]) = p + 1;
      }

      if ((p = strchr(hostname, '@')))
      {
	 q = (char*)malloc(p - hostname + 1);
	 if (p - hostname) memcpy(q, hostname, p - hostname);
	 q[p - hostname] = '\0';
	 username = q;
	 hostname = p + 1;
      }

#ifndef VXWORKS
      if (!colon && !argv[optind])
      {
	 /*
	  * No arguments means spawn rlogin.  Environment and current
	  * working directory are NOT transferred!
	  */
	 setuid(getuid());
	 if (rshflag) (const char*)(argv[0]) = "rlogin";
	 execv("/usr/bin/rlogin", (char**)argv);
	 perror("error: (runon) exec of /usr/bin/rlogin failed");
	 exit(EXIT_FAILURE);
      }
#endif

      return runon_client(argc - optind, argv + optind, envp);
   }

   fprintf(stderr,
"usage: runon <hostname> [-l login] [-n] <command> [<arguments> ...]\n"
"runon uses the shell/tcp protocol like 'remsh', but behaves more like 'on':\n"
"- Instead of <hostname>, <user>@<hostname>:<command> is also accepted.\n"
"- It attempts to make the working directory the same as on the local host.\n"
"- It transfers the environment (except a few like HOSTNAME, HOSTTYPE, etc.)\n"
"- It attempts to exit with the same circumstances as the remote process.\n"
"- Almost all signals received by runon are sent to the remote process.\n"
"  Even SIGTSTP is sent, when runon is not running on a tty device.\n"
"  Signals map correctly, even when the remote host is a different type.\n"
"- runon needs to be able to invoke itself on the remote host.  It must be\n"
"  found at exactly the same place in your PATH on the local and remote host.\n"
"- While the remsh/rsh command attempts to build a string from your arguments\n"
"  and pass them to '/bin/sh -c', runon behaves more like the 'on' command and\n"
"  runs exactly the same arguments on the remote has as were entered on the\n"
"  local host, with the 'runon remhost' removed:\n"
"    [remsh|runon] remhost ls 'a b' '$HOME' '>>' c\n"
"    remsh on remhost:     ls a b $HOME >> c\n"
"    runon/on on remhost:  ls 'a b' '$HOME' '>>' c\n"
"  To get the remsh behavior with runon, use the following syntax:\n"
"    runon remhost /bin/sh -c 'ls a b $HOME >> c'\n"
"See man page for complete information.\n");
   exit(2);
}

/*
 * --- RUNON CLIENT ---
 */

static int* fd_sig = 0;
static int sigmap[NSIG+1];

static int
runon_client(int argc, const char* argv[], const char* envp[])
{
   struct servent* servent;
   struct passwd* pwd;
   const char** p;
   const char* tmp;
   int pass;
   int exitstatus = EXIT_SUCCESS;
   int fd_rcmd[3] = { -1, -1, -1 };
   char runon_cmd[3000] = "";
   int stdin_fileno = STDIN_FILENO,
      stdout_fileno = STDOUT_FILENO,
      stderr_fileno = STDERR_FILENO;

/* cfht_logv(CFHT_MAIN, CFHT_LOG_ID, (char*)progname); */

   progname = "runon[initializing]";

   default_sigmap();

   if (xflag) cli_system("xhost", hostname);

   if (!(pwd=getpwuid(getuid())))
   {
      perror("error: (runon) getpwuid failed");
      return EXIT_FAILURE;
   }
   
   if (!(servent=getservbyname("shell", "tcp")))
   {
      fprintf(stderr, "error: (runon) shell/tcp: unknown service\n");
      return EXIT_FAILURE;
   }
   
   /*
    * Find the correct path to runon.  It must be in the same
    * place on the remote host.  If progpath doesn't start with '/',
    * then search the user's PATH variable to see which one to run.
    */
   if (!serverpath)
   {
      char* p = getenv("PATH");
      char* p_next;
      char* p_orig;
      
      if (*progpath!='/' && p)
      {
	 p = p_orig = strdup(p);
	 while (p && *p)
	 {
	    struct stat stbuf;
	 
	    if ((p_next = strchr(p, ':')))
	       *p_next++ = '\0';

	    sprintf(runon_cmd, "%.900s/%.100s", p, progpath);
	    if (stat(runon_cmd, &stbuf)==0 &&
		!S_ISDIR(stbuf.st_mode) &&
		(stbuf.st_mode&S_IXUSR))
	    {
	       /*
		* Found the runon executable in the user's path, now in `p'.
		*/

	       if (!strncmp(p, "/usr/local/cfht", 15)) p+=10;
	       if (!strncmp(p, "/local/", 7)) p+=6;
	       /*
		* === Hack to help transition from /usr/local/cfht/ (which
		* is not on _all_ machines, to /cfht/, which is automounted
		* to all machines at cfht.
		*/
	       sprintf(runon_cmd, "test -d '%.900s' && cd '%.900s' ; exec '%.900s/%.100s' -s",
		       cli_getcwd(), cli_getcwd(), p, progpath);
	       break;
	    }
	    if (!(p = p_next))
	       *runon_cmd = '\0';
	 }
	 free(p_orig);
      }
      if (*runon_cmd == '\0')
      {
	 if (!strncmp(progpath, "/usr/local/cfht/", 16)) progpath+=10;
	 if (!strncmp(p, "/local/", 7)) p+=6;
	 /*
	  * === Hack to help transition from /usr/local/cfht/ (which
	  * is not on _all_ machines, to /cfht/, which is automounted
	  * to all machines at cfht.
	  */
	 sprintf(runon_cmd, "exec %.1000s -s", progpath);
      }
   }
   
   progname = "runon[running rcmd]";

   fd_rcmd[0] = rcmd((char**)&hostname, servent->s_port,
		     pwd->pw_name, username?username:pwd->pw_name,
		     serverpath?serverpath:runon_cmd,
		     &fd_rcmd[2]);

   /*
    * Setuid root priveleges no longer needed
    */
   setuid(getuid());

   if (fd_rcmd[0]<0)
   {
      return EXIT_FAILURE;
   }
   fd_rcmd[1] = dup(fd_rcmd[0]);

   if (fd_rcmd[2]<0)
   {
      fprintf(stderr, "warning: (runon) can't establish stderr\n");
   }

   fd_sig = &fd_rcmd[2]; /* Start transmitting signals */

   fdname_set(stdin_fileno, "stdin");
   fdname_set(stdout_fileno, "stdout");
   fdname_set(stderr_fileno, "stderr");
   fdname_set(fd_rcmd[0], "fd_rcmd[0]");
   fdname_set(fd_rcmd[1], "fd_rcmd[1]");
   fdname_set(fd_rcmd[2], "fd_rcmd[2]");

   progname = "runon[setting signal handling]";

   /* %%% */
   if (dflag || (!nflag && !isatty(stdin_fileno)))
   {
      cli_signal(SIGTSTP, (PFV)write_signal);
      cli_signal(SIGCONT, (PFV)write_signal);
   }
   cli_signal(SIGHUP,   (PFV)write_signal);
   cli_signal(SIGINT,   (PFV)write_signal);
   cli_signal(SIGQUIT,  (PFV)write_signal);
   cli_signal(SIGILL,   (PFV)write_signal);
   cli_signal(SIGTRAP,  (PFV)write_signal);
   cli_signal(SIGABRT,  (PFV)write_signal);
   cli_signal(SIGFPE,   (PFV)write_signal);
   cli_signal(SIGSEGV,  (PFV)write_signal);
   cli_signal(SIGALRM,  (PFV)write_signal);
   cli_signal(SIGTERM,  (PFV)write_signal);
   cli_signal(SIGUSR1,  (PFV)write_signal);
   cli_signal(SIGUSR2,  (PFV)write_signal);
   cli_signal(SIGWINCH, (PFV)write_signal);
   cli_signal(SIGPIPE,  CLI_SIG_IGN);
   cli_signal(SIGCHLD,  (PFV)write_signal);
   /* SIGCHLD will have to go if runon itself ever needs to have any children. */

   set_non_blocking(fd_rcmd[0]);

   progname = "runon[cd]";

   /*
    * Change server's working directory to same as local, if possible.
    */
   write_or_die(fd_rcmd[0], "cd ", 3);
   write_or_die(fd_rcmd[0], cli_getcwd(), strlen(cli_getcwd()));
   write_or_die(fd_rcmd[0], "\n", 1);
   
   /*
    * Send all of local environment to server.
    * First pass sends "VARNAME=value", one per line.
    * Second pass sends "export VARNAME VARNAME2...".
    * Second pass only needed if /bin/sh is used instead of "runon -s"
    */
   /*
    * First check for $HOSTDISPLAY and stick it in $DISPLAY if it exists.
    */
   if ((tmp = getenv("HOSTDISPLAY")))
   {
      static char* new_env;
      
      new_env = (char*)malloc(strlen(tmp) + 9);
      sprintf(new_env, "DISPLAY=%s", tmp);
      putenv(new_env);
   }

   for (pass = 0; pass < 2; pass++)
   {
      /*
       * On the second pass, build an "export ..." line in case the
       * server is /bin/sh.  This is not needed by "runon -s".
       */
      if (pass && strstr(serverpath?serverpath:runon_cmd, "runon")) break;

      if (pass)
      {
	 progname = "runon[env pass 2]";
	 write_or_die(fd_rcmd[0], "export ", 7);
      }
      else
      {
	 progname = "runon[env pass 1]";
      }

      for (p = envp; *p; p++)
      {
	 char* equal = strchr(*p, '=');
	 int name_len;
	 
	 name_len = equal?(equal - *p):-1;
	 if (name_len > 0)
	 {
	    if ((name_len==4 && !strncmp(*p, "HOST", 4)) ||
		(name_len==8 && !strncmp(*p, "HOSTTYPE", 8)) ||
		(name_len==8 && !strncmp(*p, "HOSTNAME", 8)) ||
		(name_len==8 && !strncmp(*p, "MACHTYPE", 8)) ||
		(name_len==6 && !strncmp(*p, "VENDOR", 6)) ||
		(name_len==6 && !strncmp(*p, "OSTYPE", 6)) ||
		(name_len==3 && !strncmp(*p, "PWD", 3)))
	    {
	       /*
		* Skip these variables, since they may not be valid
		* on a remote host.  === Way to configure!?
		*/
	    }
	    else if (pass)
	    {
	       write_or_die(fd_rcmd[0], " ", 1);
	       write_or_die(fd_rcmd[0], *p, name_len);
	    }
	    else
	    {
	       /*
		* Like HOST, HOSTTYPE, HOSTNAME, etc., CFHTLOGU must be handled
		* specially.  If it is set locally, it should specifically be
		* set to '' remotely to indicate that there is a CFHTLOGU pipe,
		* but it exists on a different host, and shouldn't be used.
		*/
	       if (name_len==8 && !strncmp(*p, "CFHTLOGU", 8))
	       {
		  write_or_die(fd_rcmd[0], "CFHTLOGU=''\n", 12);
	       }
	       else
	       {
		  write_or_die(fd_rcmd[0], *p, name_len+1);
		  write_quoted(fd_rcmd[0], *p + name_len+1);
		  write_or_die(fd_rcmd[0], "\n", 1);
	       }
	    }
	 }
      }
      if (pass) write_or_die(fd_rcmd[0], "\n", 1);
   }


   /*
    * Send all of the command line arguments and turn the server into
    * the desired remote process.
    */
   progname = "runon[exec]";

   write_or_die(fd_rcmd[0], "exec", 4);
   for (p = argv; *p; p++)
   {
      write_or_die(fd_rcmd[0], " ", 1);
      write_quoted(fd_rcmd[0], *p);
   }
   write_or_die(fd_rcmd[0], "\n", 1);

   if (dflag)
   {
      switch (fork())
      {
	 case (pid_t)-1: perror("error: (runon) fork failed"); exit(EXIT_FAILURE);
	 case 0:
	 {
	    int devnull;
	    
	    if (setsid()==-1)
	       perror("warning: (runon) setsid() failed");
	    if ((devnull = open("/dev/null", O_RDWR, 0666))==-1)
	       perror("warning: (runon) open(\"/dev/null\") failed");
	    else
	    {
	       dup2(devnull, stdin_fileno);
	       close(devnull);
	    }
	    break;
	 }
	 default:
	 {
	    close(fd_rcmd[0]);
	    close(fd_rcmd[1]);
	    close(fd_rcmd[2]);
#ifndef VXWORKS
	    tty_restore();
#endif
	    _exit(EXIT_SUCCESS);
	 }
      }
   }
   
   /*
    * Redirect output if -o option given.  -d implies
    * -o /dev/null if no other -o option was given.
    */
   if (dflag && !output) output = "/dev/null";
   if (output)
   {
      int outfd;

      if ((outfd = open(output, O_RDWR|O_APPEND|O_CREAT|O_NONBLOCK, 0666))==-1)
      {
	 perror("warning: (runon) can't open local output");
      }
      else
      {
	 dup2(outfd, stdout_fileno);
	 dup2(outfd, stderr_fileno);
	 close(outfd);
      }
   }

   if (nflag)
   {
      close(stdin_fileno);
      stdin_fileno = -1;
   }

   progname = "runon[talk]";

   /*
    * Now go into an endless loop just feeding runon's stdin to the remote
    * process and forwarding the remote's stdout/stderr back to the local.
    */
   talk(&stdin_fileno, &stdout_fileno, &stderr_fileno,
	&fd_rcmd[0],   &fd_rcmd[1],    &fd_rcmd[2],
	&exitstatus);

   progname = "runon[cleanup]";

   /*
    * Since the process is running as an rcmd(), make sure
    * it knows it is not stopped.  (This can happen if a
    * SIGTSTP was sent to the agent, and the handling of
    * this signal is *not* set to actually stop the process.)
    * Sending a SIGCONT doesn't seem to do any harm in any case.
    */
   if (fd_rcmd[2]!=-1) write_signal(SIGCONT);
   if (fd_rcmd[0]!=-1) close(fd_rcmd[0]);
   if (fd_rcmd[1]!=-1) close(fd_rcmd[1]);
   if (fd_rcmd[2]!=-1) close(fd_rcmd[2]);
   return exitstatus;
}

/*
 * --- RUNON SERVER ---
 */

static int
runon_server(void)
{
   /*
    * This is the runon "server".  User's should never have
    * to run the runon server, but runon will invoke itself
    * on the remote host as "runon -s" to handle setting up
    * the environment, etc.  The runon server is like a simple
    * non-interactive shell that currently only accepts 3
    * commands: "cd DIRECTORY" "export VAR=value" and "exec ..."
    *
    * If runon is not available at the other end, a bourne shell
    * can be used but signal mapping and exit codes will no work.
    */
   int len, readsize = 1024; /* Will grow as needed */
   char* readbuf;
   char tmp[100];

   /*
    * Setuid root is only for the client so it can run rcmd()
    */
   setuid(getuid());

   /*
    * This must be done at the start of the program.  Setting this
    * ensures diagnostic messages will never be buffered as long as
    * you have newlines ('\n') at the end of all your printf()'s.
    */
   if (!isatty(STDOUT_FILENO) && setvbuf(stdout, NULL, _IOLBF, BUFSIZ)!=0)
      perror("warning: (runon -s) setvbuf failed");

   /*
    * Different operating systems map important signals to different number
    * values.  This helps both ends figure out what the right number values
    * are (since sending a byte with the signal number is the only way to
    * send a signal over an rcmd()).
    */
   sprintf(tmp,
	   "signalmap: %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
	   SIGHUP, SIGINT, SIGQUIT, SIGILL, SIGTRAP, SIGABRT, SIGFPE, SIGKILL, SIGSEGV,
	   SIGPIPE, SIGALRM, SIGTERM, SIGUSR1, SIGUSR2, SIGSTOP, SIGTSTP, SIGCONT,
	   SIGCHLD, SIGWINCH);
   write_or_die(STDERR_FILENO, tmp, strlen(tmp));

   readbuf = (char*)malloc(readsize);

   for (;;)
   {
      char* equal;
      len = 0;

      do
      {
	 if (len >= readsize)
	 {
	    /*
	     * Line doesn't fit entirely.  Double the amount of space.
	     */
	    readsize *= 2;
	    readbuf = (char*)realloc(readbuf, readsize);
	 }
	 switch (TEMP_FAILURE_RETRY(read(STDIN_FILENO, &readbuf[len], 1)))
	 {
	    case -1:
	    {
	       perror("error: read failed");
	       return EXIT_FAILURE;
	    }
	    case 0:
	    {
	       fprintf(stderr, "error: (runon -s) no `exec ...' command found\n");
	       return EXIT_FAILURE;
	    }
	 }
      } while (readbuf[len++] != '\n');
      
      /*
       * Got a complete line.  Parse it.
       */
      readbuf[--len] = '\0';
      if (!strncmp(readbuf, "cd ", 3))
      {
	 /*
	  * Working directory command...
	  */
	 if (chdir(readbuf+3)!=0)
	 {
	    fprintf(stderr,
		    "warning: (runon -s) chdir(%s) ",
		    readbuf+3);
	    perror("failed");
	    /* But continue anyway */
	 }
      }
      else if (!strncmp(readbuf, "exec ", 5))
      {
	 /*
	  * Execute the remote command
	  */
	 char** exec_argv;
	 int exec_argv_size = 100; /* grows as needed */
	 int exec_argv_pos = 0;
	 char* start;
	 char* stop;
	 char* copy;
	 int i, status;
	 pid_t child_pid;
	 int child_stdout_fd[2], child_stderr_fd[2], child_stdin_fd[2];
	 
	 exec_argv = (char**)malloc(exec_argv_size*sizeof(char*));
	 
	 start = readbuf + 5;
	 while (*start)
	 {
	    if (*start!='\'')
	    {
	       fprintf(stderr,
		       "error: (runon -s) exec arguments must be in (') quotes\n");
	       for (i=0; i<exec_argv_pos; i++) free(exec_argv[i]);
	       free(exec_argv);
	       free(readbuf);
	       return EXIT_FAILURE;
	    }
	    start ++;
	    stop = start;
	    while (*stop)
	    {
	       if (*stop=='\'')
	       {
		  if (!strncmp(stop, "'\\''", 4))
		  {
		     stop += 4;
		     continue;
		  }
		  break;
	       }
	       else stop++;
	    }
	    
	    if (*stop!='\'')
	    {
	       fprintf(stderr,
		       "error: (runon -s) exec missing quote (')\n");
	       for (i=0; i<exec_argv_pos; i++) free(exec_argv[i]);
	       free(exec_argv);
	       free(readbuf);
	       return EXIT_FAILURE;
	    }
	    /*
	     * 'This is an argument'
	     *  ^_start ..... stop_^	       
	     */
	    exec_argv[exec_argv_pos++] = copy =
	       (char*)malloc(stop - start + 1);
	    
	    /*
	     * Make sure there is still enough room in the argv array.
	     */
	    if (exec_argv_pos>=exec_argv_size)
	    {
	       exec_argv_size *= 2;
	       exec_argv = (char**)realloc(exec_argv, 
					   exec_argv_size*sizeof(char*));
	    }
	    while (start < stop)
	    {
	       if (!strncmp(start, "'\\''", 4))
	       {
		  *copy++ = '\'';
		  start += 4;
	       }
	       else *copy++ = *start++;
	    }
	    /*
	     * Terminate the argument string
	     */
	    *copy = '\0';
	    /*
	     * Skip to the next argument.
	     */
	    if (*(++start)==' ') start++;
	 }
	 /*
	  * Terminate the argv list
	  */
	 exec_argv[exec_argv_pos] = 0;
	 free(readbuf);
	 /*
	  * Turn into the request process.  This uses
	  * execvp, so that the PATH is searched.
	  */
	 pipe(child_stdout_fd);
	 pipe(child_stderr_fd);
	 pipe(child_stdin_fd);
	 
	 switch ((child_pid = fork()))
	 {
	    case (pid_t)-1:
	    {
	       perror("error: (runon -s) fork failed");
	       return EXIT_FAILURE;
	    }
	    case 0:
	    {
	       dup2(child_stdout_fd[1], STDOUT_FILENO);
	       dup2(child_stderr_fd[1], STDERR_FILENO);
	       dup2(child_stdin_fd[0], STDIN_FILENO);
	       close(child_stdout_fd[0]);
	       close(child_stderr_fd[0]);
	       close(child_stdin_fd[1]);
	       
	       execvp(exec_argv[0], exec_argv);
	       perror(exec_argv[0]);
	       for (i=0; i<exec_argv_pos; i++) free(exec_argv[i]);
	       free(exec_argv);
	       _exit(EXIT_FAILURE);
	    }
	    default:
	    {
	       int stdin_fileno = STDIN_FILENO,
		  stdout_fileno = STDOUT_FILENO,
		  stderr_fileno = STDERR_FILENO;
	       
	       /*
		* This is krusty, but we only want the child to see
		* these signals (which it will, because rshd sends
		* the signal to the entire process group.)
		*/
	       cli_signal(SIGTSTP,  CLI_SIG_IGN);
	       cli_signal(SIGCONT,  CLI_SIG_IGN);
	       cli_signal(SIGHUP,   CLI_SIG_IGN);
	       cli_signal(SIGINT,   CLI_SIG_IGN);
	       cli_signal(SIGQUIT,  CLI_SIG_IGN);
	       cli_signal(SIGILL,   CLI_SIG_IGN);
	       cli_signal(SIGTRAP,  CLI_SIG_IGN);
	       cli_signal(SIGABRT,  CLI_SIG_IGN);
	       cli_signal(SIGFPE,   CLI_SIG_IGN);
	       cli_signal(SIGSEGV,  CLI_SIG_IGN);
	       cli_signal(SIGALRM,  CLI_SIG_IGN);
	       cli_signal(SIGTERM,  CLI_SIG_IGN);
	       cli_signal(SIGUSR1,  CLI_SIG_IGN);
	       cli_signal(SIGUSR2,  CLI_SIG_IGN);
	       cli_signal(SIGWINCH, CLI_SIG_IGN);
	       cli_signal(SIGPIPE,  CLI_SIG_IGN);
/*	       cli_signal(SIGCHLD,  CLI_SIG_IGN); Can't ignore this one */
	       
	       for (i=0; i<exec_argv_pos; i++) free(exec_argv[i]);
	       free(exec_argv);
	       
	       close(child_stdout_fd[1]);
	       close(child_stderr_fd[1]);
	       close(child_stdin_fd[0]);
	       
	       talk(&stdin_fileno,     &stdout_fileno,     &stderr_fileno,
		    &child_stdin_fd[1],&child_stdout_fd[0],&child_stderr_fd[0],
		    0 /* NULL here means server mode */);
	       
	       if (child_stdout_fd[0]!=-1) close(child_stdout_fd[0]);
	       if (child_stderr_fd[0]!=-1) close(child_stderr_fd[0]);
	       if (child_stdin_fd[0]!=-1) close(child_stdin_fd[1]);
	       
	       if (TEMP_FAILURE_RETRY(waitpid(child_pid, &status, 0))<=0)
	       {
		  perror("error: (runon -s) waitpid failed");
		  return EXIT_FAILURE;
	       }
	       else
	       {
		  char tmp[80];
		  
		  sprintf(tmp, "\rlogonly: exitstatus %05d\n",
			  status&0xFFFF);
		  write_or_die(STDERR_FILENO, tmp, strlen(tmp));
	       }
	       return status;
	    }
	 }
      }
      else if (!strncmp(readbuf, "export ", 7))
      {
	 /*
	  * Ignore "export" commands (since _everything_ is
	  * exported by default) Export commands are sent only
	  * to allow compatability with bourne shell language.
	  */
      }
      else if ((equal = strchr(readbuf, '=')))
      {
	 /*
	  * Add an environment variable
	  */
	 char* new_env = (char*)malloc(len+1);
	 char* dest;
	 char* src;
	 int name_len = equal - readbuf;
	 
	 src = readbuf;
	 dest = new_env;
	 
	 memcpy(dest, src, name_len + 1);
	 src += name_len + 1;
	 dest += name_len + 1;
	 if (*src=='\'' && readbuf[len-1]=='\'')
	 {
	    src++; readbuf[--len] = '\0';
	 }
	 while (*src)
	 {
	    if (strncmp(src, "'\\''", 4)==0)
	    {
	       *dest++ = '\''; src += 4;
	    }
	    else *dest++ = *src++;
	 }
	 *dest = '\0';
	 putenv(new_env);
      }
      else
      {
	 fprintf(stderr,
		 "logonly: (runon -s) ignoring command line `%s'\n",
		 readbuf);
      }
   }
}

/*
 * --- UTILITY FUNCTIONS ---
 */

/*

from    fd_in1  --> fd_in2   pipe to
TTY or  fd_out1 <-- fd_out2  PROCESS
REMSHD  fd_err1 <-- fd_err2  or rcmd()

        at-a-time block                select first?
fd_in1   1-byte    Y(leave)                  Y
fd_out1    *       Y(write 1 on EAGAIN)      Y
fd_err1    *       Y(write 1 on EAGAIN)      Y
fd_in2   1-byte    N(leave)                  Y
fd_out2    *       N(required)               Y
fd_err2    *       N(required)               Y

*/

static void
talk(int* fd_in1, int* fd_out1, int* fd_err1,
     int* fd_in2, int* fd_out2, int* fd_err2,
     int* client_exitstatus) /* NULL means server mode */
{
   fd_set read_fds, rfds, write_fds, wfds;
   int* fd_read[3];
   int* fd_write[3];
   int fd_max, i;

   fd_read[0] = fd_in1;
   fd_read[1] = fd_out2;
   fd_read[2] = fd_err2;
   fd_write[0] = fd_in2;
   fd_write[1] = fd_out1;
   fd_write[2] = fd_err1;

   /*
    * Non-blocking i/o is *required* on these two.  fd_read[0]
    * is read 1-byte at a time, so it doesn't matter.
    */
   set_non_blocking(*fd_read[1]);
   set_non_blocking(*fd_read[2]);

   FD_ZERO(&write_fds);
   FD_ZERO(&read_fds);

   if (*fd_read[0]!=-1) FD_SET(*fd_read[0], &read_fds);
   if (*fd_read[1]!=-1) FD_SET(*fd_read[1], &read_fds);
   if (*fd_read[2]!=-1) FD_SET(*fd_read[2], &read_fds);

   fd_max = *fd_read[0];
   if (*fd_read[1] > fd_max) fd_max = *fd_read[1];
   if (*fd_read[2] > fd_max) fd_max = *fd_read[2];
   if (*fd_write[0] > fd_max) fd_max = *fd_write[0];
   if (*fd_write[1] > fd_max) fd_max = *fd_write[1];
   if (*fd_write[2] > fd_max) fd_max = *fd_write[2];

   while (*fd_read[1]!=-1 || *fd_read[2]!=-1)
   {
      char buf[3][BUFSIZ];
      int bufread[3];
      int bufwritten[3];

      /*
       * Start with the active set of file descriptors
       */
      rfds = read_fds;
      wfds = write_fds;

      /*
       * But don't attempt to read any more from pairs on which writes
       * are still pending.
       */
      for (i=0; i<3; i++)
      {
	 if (*fd_write[i]!=-1 && *fd_read[i]!=-1 &&
	     FD_ISSET(*fd_write[i], &wfds)) FD_CLR(*fd_read[i], &rfds);
      }

      /*
       * Wait for something to happen.
       */
      if (SELECT(fd_max + 1, &rfds, &wfds, 0, 0)<0)
      {
	 if (errno==EINTR) continue; /* Make sure the masks get reset correctly */
	 fprintf(stderr, "error: (%s) ", progname);
	 perror("select failed");
	 return;
      }

      for (i=0; i<3; i++)
      {
	 /*
	  * Check if anything needs to be written out.
	  */
	 if (*fd_write[i]!=-1 && FD_ISSET(*fd_write[i], &wfds))
	 {
	    int nbytes;

	    nbytes = write(*fd_write[i], buf[i] + bufwritten[i], bufread[i] - bufwritten[i]);
	    switch (nbytes)
	    {
	       case -1:
	       {
		  if (errno==EINTR) break;
		  if (errno==EWOULDBLOCK || errno==EAGAIN)
		  {
		     if (bufread[i] - bufwritten[i] > 1)
		     {
			if (TEMP_FAILURE_RETRY(write(*fd_write[i], buf[i] + bufwritten[i], 1))==1)
			{
			   bufwritten[i]++;
			   if (bufwritten[i] >= bufread[i])
			   {
			      FD_CLR(*fd_write[i], &write_fds);
			   }
			   continue;
			}
		     }
		     perror("error: (runon) select passed but write of 1 byte failed");
		  }
		  
		  if (errno!=EPIPE || i!=0)
		  {
		     perror("write failed");
		     fprintf(stderr, "error: (%s) closing %s; ",
			     progname,
			     i==0?"stdin":(i==1?"stdout":"stderr"));
		  }
		  FD_CLR(*fd_write[i], &read_fds);
		  FD_CLR(*fd_write[i], &write_fds);
		  close(*fd_write[i]);
		  *fd_write[i] = -1;

		  if (i==0 && *fd_read[i]!=-1)
		  {
		     FD_CLR(*fd_read[i], &read_fds);
		     FD_CLR(*fd_read[i], &write_fds);
		     close(*fd_read[i]);
		     *fd_read[i] = -1;
		  }
		     
		  break;
	       }
	       default:
	       {
		  /*
		   * A single write will not necessarily send the entire buffer,
		   * so only switch out of writing mode when the whole buffer
		   * has been sent.
		   */
		  bufwritten[i] += nbytes;
		  if (bufwritten[i] >= bufread[i])
		  {
		     FD_CLR(*fd_write[i], &write_fds);
		  }
	       }
	    }
	 }
	 /*
	  * Check if any new data is waiting to be read.
	  */
	 else if (*fd_read[i]!=-1 && FD_ISSET(*fd_read[i], &rfds))
	 {
	    switch ((bufread[i] = read(*fd_read[i], buf[i], i==0?1:sizeof(buf[i]))))
	    {
	       case -1:
	       {
		  if (errno==EINTR)
		     /* EWOULDBLOCK and EAGAIN shouldn't really happen */
		     break;

		  fprintf(stderr, "error: (%s) %s ", progname,
			  i==0?"stdin":(i==1?"stdout":"stderr"));
		  perror("read failed");
		  /* fall through and close the file descriptor */
	       }
	       case 0:
	       {
		  FD_CLR(*fd_read[i], &read_fds);
		  FD_CLR(*fd_read[i], &write_fds);
		  close(*fd_read[i]);
		  *fd_read[i] = -1;

		  if (i==0)
		  {
		     if (client_exitstatus == 0)
		     {
			if (*fd_write[0] != -1)
			{
			   FD_CLR(*fd_write[0], &read_fds);
			   FD_CLR(*fd_write[0], &write_fds);
			   close(*fd_write[0]);
			   *fd_write[0] = -1;
			}
		     }
                     else
		     {
			if (shutdown(*fd_write[0], 1/*SHUT_WR*/) != 0)
			   perror("shutdown failed");
		     }
		  }
		  break;
	       }
	       default:
	       {
		  /*
		   * This whole next if-statement just exists for the purpose
		   * of getting exit statuses and signal mappings.
		   */
		  if (client_exitstatus && i==2)
		  {
		     static int got_first_line = 0;
		     static char sigmapbuf[100];
		     static int sigmappos = 0;
		     
		     if (!got_first_line)
		     {
			int p = 0;
			while (p < bufread[i] && buf[i][p] < ' ') p++;
			while (p < bufread[i] && buf[i][p] >= ' ' && 
			       sigmappos < sizeof(sigmapbuf))
			   sigmapbuf[sigmappos++] = buf[i][p++];
			if (sigmappos == sizeof(sigmapbuf))
			{
			   got_first_line = 1;
			}
			else if (sigmappos && p < bufread[i] && buf[i][p] < ' ')
			{
			   got_first_line = 1;
			   sigmapbuf[sigmappos] = '\0';
			   if (cli_parse_signalmap(sigmap, sigmapbuf)==FAIL)
			      default_sigmap();
			   else
			   {
			      p++;
			      if (bufread[i] - p)
			      {
				 char tbuf[BUFSIZ];
				 memcpy(tbuf, buf[i] + p, bufread[i] - p);
				 memcpy(buf[i], tbuf, bufread[i] - p);
			      }
			      bufread[i] -= p;
			   }
			}
		     }

		     if (got_first_line)
		     {
			char* tail = buf[i] + bufread[i] - 1;
			char* head = buf[i];
			char* tmp;
			int status;

			if (*tail=='\n') tail--;
			while (tail > head &&
			       (tmp = memchr(head, '\r', tail - head)))
			   head = tmp + 1;
			if (tail - head + 1 == 25 &&
			    sscanf(head, "logonly: exitstatus %d", &status)==1)
			{
			   /*
			    * Immediately close stdin on the local side.  There's
			    * no place to send input now that the remote command
			    * has exited.
			    */
			   if (*fd_read[0]!=-1)
			   {
			      FD_CLR(*fd_read[0], &read_fds);
			      FD_CLR(*fd_read[0], &write_fds);
			      close(*fd_read[0]);
			      *fd_read[0] = -1;
			   }

			   if (*fd_write[0]!=-1)
			   {
			      FD_CLR(*fd_write[0], &read_fds);
			      FD_CLR(*fd_write[0], &write_fds);
			      close(*fd_write[0]);
			      *fd_write[0] = -1;
			   }

			   bufread[i] -= 26;
			   if (bufread[i]) bufread[i]--;

			   if (!WIFEXITED(status))
			   {
			      if (WIFSIGNALED(status))
			      {
				 int signo = WTERMSIG(status);
				 int nocores = 0;
				 
				 cli_signal(signo, CLI_SIG_DFL);
				 if (signo == SIGQUIT || signo == SIGILL || signo == SIGTRAP ||
				     signo == SIGABRT ||
#ifdef SIGIOT
				     signo == SIGIOT ||
#endif
#ifdef SIGEMT
				     signo == SIGEMT ||
#endif
#ifdef SIGSYS
				     signo == SIGSYS ||
#endif
				     signo == SIGFPE || signo == SIGBUS || signo == SIGSEGV)
				 {
				    /*
				     * Before runon kills itself with a signal that normally
				     * generates a core, try to disable cores, because we don't
				     * want to overwrite the core of the remote process, which
				     * may be trying to write a core in the same directory over
				     * the network.  The core from runon is useless anyway.
				     */
#ifdef RLIMIT_CORE
				    struct rlimit rlim;
				    if (getrlimit(RLIMIT_CORE, &rlim)==0)
				    {
				       rlim.rlim_cur = 0;
				       if (setrlimit(RLIMIT_CORE, &rlim)==0)
					  nocores = 1;
				    }
#endif
				    if (nocores==0) /* setrlimit() didn't work or wasn't available */
				    {
				       fprintf(stderr, "logonly: (runon) unable to disable cores\n");
				       errno = 0;
				       /*
					* This is a lame hack to try and prevent runon from
					* leaving another core that might overwrite the one
					* from the remote process, if setrlimit didn't work.
					*/
				       chdir("/");
				    }
				 }
				 kill(getpid(), signo);
				 perror("warning: (runon) kill didn't work\n");
				 *client_exitstatus = EXIT_FAILURE;
			      }
			      else if (WIFSTOPPED(status))
			      {
				 fprintf(stderr, "error: (runon) child stopped\n");
				 *client_exitstatus = EXIT_FAILURE;
			      }
			      else
			      {
				 fprintf(stderr, "error: (runon) unknown problem with child process\n");
				 *client_exitstatus = EXIT_FAILURE;
			      }
			   }
			   else *client_exitstatus = WEXITSTATUS(status);
			}
		     }
		  }

		  if (bufread[i] && *fd_write[i]!=-1)
		  {
		     bufwritten[i] = 0;
		     FD_SET(*fd_write[i], &write_fds);
		  }
	       }
	    }
	 }
      }
   }
}

static void
set_non_blocking(int fd)
{
   if (fd==-1) return;
   if (isatty(fd))
   {
      fprintf(stderr, "error: tried to set NONBLOCK on a tty\n");
      return;
   }

   if (fcntl(fd, F_SETFL, fcntl(fd, F_GETFL, 0)|O_NONBLOCK)==-1)
   {
      fprintf(stderr, "warning: (%s) ", progname);
      perror("set_non_blocking failed");
   }
}

static int
strfreq(const char* str, char c)
{
   int rtn = 0;
   while (*str) if (*str++ == c) rtn++;
   return rtn;
}

static void
write_or_die(int fd, const void* data, int len)
{
   if (cli_write(fd, data, len, RCMD_TIMEOUT)==FAIL)
   {
      fprintf(stderr, "error: (%s) cliwrite(%s,\"%.*s\",%d) ",
	      progname, fdname(fd), len, (char*)data, len);
      perror("failed");
      exit(EXIT_FAILURE);
   }
}

static void
default_sigmap(void)
{
   int i;

   for (i=0; i<sizeof(sigmap)/sizeof(int); i++) sigmap[i] = i;
   sigmap[SIGTSTP] = 0;
}

static void
write_signal(int signo)
{
   char tmp[1];

   if (signo<=0 || signo > NSIG) return;
   tmp[0] = (char)(sigmap[signo]);
#if 0
   if (fd_sig && *fd_sig != -1 && tmp[0] > 0)
      write_or_die(*fd_sig, tmp, 1);
#else
   if (fd_sig && *fd_sig != -1 && tmp[0] > 0)
   {
#if 0
      if (signo==SIGSEGV) /* %%% */
      {
	 fprintf(stderr, "error: Caught SIGSEGV.  Trying to dump core in /tmp/\n");
	 chdir("/tmp");
	 cli_signal(signo, CLI_SIG_DFL);
	 kill(getpid(), signo);
      }
#endif
      if (cli_write(*fd_sig, tmp, 1, RCMD_TIMEOUT)==FAIL)
      {
	 fprintf(stderr, "warning: (%s) cli_write(%s, SIG#%d) ",
		 progname, fdname(*fd_sig), (int)(tmp[0]));
	 perror("failed");
/*	 exit(EXIT_FAILURE); %%% */
      }
   }
#endif
   else
   {
      cli_signal(signo, CLI_SIG_DFL);
      kill(getpid(), signo);
   }
}

static void
write_quoted(int fd, const char* string)
{
   static char* buf = 0;
   static int size;
   const char* p;
   int len;

   if (buf == 0)
   {
      size = 1024;
      buf = (char*)malloc(size);
   }

   len = strlen(string)          /* The length of the original string */
      + 2 + 1                    /* Plus 2 quotes (') around it and \0 byte */
      + strfreq(string, '\'')*3; /* Plus a '\'' to escape each real ' */

   if (len > size)
   {
      size *= 2;
      if (len > size) size = len;
      buf = (char*)realloc(buf, size);
   }
   
   sprintf(buf, "'");
   len = 1;

   for (p = string; *p; p++)
   {
      if (*p=='\'')
      {
	 sprintf(buf + len, "'\\''");
	 len += 4;
      }
      else if (*p=='\n') /* Replace newlines in variables with spaces */
      {
	 buf[len++] = ' ';
      }
      else
      {
	 buf[len++] = *p; /* Warning: not null-term'd now */
      }
   }
   sprintf(buf + len, "'"); /* ok, null-term'd again */
   len++;
   write_or_die(fd, buf, len);
}
/* End of runon.c */
