/*                                             -*- c-file-style: "Ellemtel" -*-

`det_data.c' - Data handling functions   (writefits, testpat)

This file is part of version 3 of the CFHT Detector Host Command Interface.
Read the `License' file for terms of use and distribution.
Copyright 1998-2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include "cli/cli.h"
#include "fh/fh.h" /* FITS Handling routines */
#include "detcom.h"
#include "misc.h"

/*
 * Comment out this line to not do bias corrections.
 * (They are never done on BIAS exposure types.)
 */
/* #define DO_BIAS_CORRECTIONS */
/* The bias levels themselves are set around line 1198 */
/* The flipping of chips for -B8 files is around line 1165 */

#ifdef DETCOM_BSWAP
/*
 * Byte-swapping for PC or DEC machines.
 */
#include <netinet/in.h>
#define NTOHS ntohs
#else
#define NTOHS
#endif

#ifndef MAP_FAILED
#define MAP_FAILED ((caddr_t)-1)
#endif
	 
/* The list of WCS FITS cards that are chip-specific */
typedef struct
{
      double crpix1;
      double crpix2;
      double cd1_1;
      double cd1_2;
      double cd2_1;
      double cd2_2;
} wcs_per_chip;

data_state_t
data_state(data_state_t new_state)
{
   static data_state_t data_state_ = DATA_EMPTY; /* Start with empty image buffers */

   if (new_state == DATA_STATE_QUERY) return data_state_;

   switch (new_state)
   {
      case DATA_EMPTY:
      {
	 /*
	  * It's ok to EMPTY the data buffers from almost any state except:
	  *
	  * READING -> EMPTY means data was never saved.
	  * SAVING -> EMPTY should cause an error (should abort first!)
	  * SAVED -> EMPTY means time to increment the file counter.
	  */
	 if (data_state_ == DATA_READING)
	 {
	    printf("warning: Emptying unsaved image buffer.\n");
	 }
	 else if (data_state_ == DATA_SAVING)
	 {
	    printf("error: data_state internal error. Emptying image buffer while still saving.\n");
	 }
	 else if (data_state_ == DATA_SAVED)
	 {
	    /*
	     * DATA_SAVED -> DATA_EMPTY transition:
	     *
	     * DATA_EMPTY is set to prepare the image buffers for new data.
	     * (either at the start of an exposure or readout, or both.)
	     * If the previous state was DATA_SAVED, this makes sure we
	     * don't re-use the filename.
	     */
	 }
	 data_state_ = new_state;
	 break;
      }
      case DATA_READING:
      {
	 /*
	  * EMPTY -> READING is the only case that makes sense.
	  */
	 if (data_state_ != DATA_EMPTY)
	 {
	    /*
	     * It's only legal to go to READING after going to EMPTY first.
	     */
	    printf("error: data_state internal error.  READING before EMPTY.\n");
	 }
	 data_state_ = new_state;
	 break;
      }
      case DATA_SAVING:
      {
	 /*
	  * Ok to start SAVING from almost any state also.
	  * 
	  * READING -> SAVING is the normal case.
	  * EMPTY -> SAVING warns that we're doing something silly.
	  * SAVED -> SAVING warns that the last save was already successful.
	  * SAVING -> SAVING doesn't make sense.
	  * ABORTED or FAILED -> SAVING means we're retrying a previous failed attempt.
	  */
	 if (data_state_ == DATA_EMPTY)
	 {
	    printf("warning: Saving empty readout frame (no new data in buffer.)\n");
	 }
	 else if (data_state_ == DATA_SAVED)
	 {
	    printf("warning: This data was already saved successfully.  Saving it again anyway.\n");
	 }
	 else if (data_state_ == DATA_SAVING)
	 {
	    printf("error: data_state internal error.  Already SAVING.\n");
	 }
	 data_state_ = new_state;
	 break;
      }

      /*
       * Whenever a frame has been successfully saved, don't re-use
       * that frame number.
       */
      case DATA_SAVED:
      case DATA_FAILED:
      case DATA_ABORTED:
      {
	 /*
	  * SAVING -> * is the only thing that makes sense with these.
	  */
	 if (data_state_ != DATA_SAVING)
	 {
	    /*
	     * Doesn't make sense to go to any of these three states
	     * unless the current state is DATA_SAVING.
	     */
	    printf("error: data_state internal error.  No save operation in progress.\n");
	 }
	 data_state_ = new_state;
	 break;
      }
      default: printf("error: data_state internal error invalid state %d\n",
		      new_state);
   }
   return data_state_;
}

PASSFAIL
com_testpat( const char* args )
{
   shm_t* shm = im_buf_get(FALSE);
   PASSFAIL rtn = PASS;
   int n;
   unsigned int rows_per_amp, cols_per_amp, amps_per_controller;

   if (shm == 0) return FAIL;
   rows_per_amp = shm->rows_per_amp;
   cols_per_amp = shm->cols_per_amp;
   amps_per_controller = shm->n_amps_per_controller;
   
   com_readout("wait");

   for (n = 0; n < n_controllers; n++)
   {
      unsigned int warning_count, warning_suppress;
      int c, x, y;

      warning_count = 0;

      for (c = 0; c < amps_per_controller; c++)
      {
	 unsigned short* p;
	 unsigned short expected;

	 expected = c - 32768;
	 p = shm->buf_base + shm->buf_ofs[n][c];
	 warning_suppress = warning_count + 5; /* Suppress warnings after first 3 */

	 for (y = 0; y < rows_per_amp; y++)
	 {
	    if (y%100==0)
	    {
	       printf("progress: Checking buffers against test pattern (%d%%)\r",
		      (n * (amps_per_controller * rows_per_amp) + c * rows_per_amp + y) /
		      (n_controllers * amps_per_controller * rows_per_amp / 100));
	       fflush(stdout);
	    }

	    for (x = 0; x < cols_per_amp - 32; x++)
	    {
	       unsigned short val;

	       val = NTOHS(*p);
	       if (val != expected++)
		  if (warning_count++ < warning_suppress)
		     printf("warning: %s, pix(x=%d,y=%d,amp=%d) "
			    "expected 0x%04X, got 0x%04X\n",
			    leach[n].name_contr, x, y, c,
			    (expected-1)&0xFFFF,
			    val);
	       p ++;
	    }
	    for (x = cols_per_amp - 32; x < cols_per_amp; x++)
	    {
	       unsigned short val;

	       val = NTOHS(*p);
	       if (val != 0x800A)
		  if (warning_count++ < warning_suppress)
		     printf("warning: %s, pix(x=%d,y=%d,amp=%d) "
			    "expected 0x800A, got 0x%04X\n",
			    leach[n].name_contr, x, y, c, val);
	       p ++;
	    }
	 }
	 if (warning_count > warning_suppress)
	    printf("warning: Further test pattern mismatches suppressed...\n");
      }
      if (warning_count)
      {
	 printf("warning: %u of %u pixels did not match the test pattern from %s\n",
		warning_count, shm->pixels_expected[n], leach[n].name_contr);
	 rtn = FAIL;
      }
   }

   if (rtn==PASS)
   {
      printf("status: Test pattern matches.\n");
   }
   else
   {
      printf("error: Test pattern check failed.\n");
   }

   return rtn;
}

#define TESTCCD_MINPIXELS 16
#define TESTCCD_MAXPIXELS 1024

PASSFAIL
com_testccd( const char* args )
   /*
    * Very crude check to make sure we have properly powered up CCDs.
    * Right now this only checks to make sure that there is some differentiation
    * in the pixels of each amp.  If any single amp comes out with all
    * zeros for example, this will return FAIL, otherwise it returns PASS.
    */
{
   shm_t* shm = im_buf_get(FALSE);
   PASSFAIL rtn = PASS;
   int n, c;

   if (shm == 0) return FAIL;
   com_readout("wait");

   for (n = 0; n < n_controllers; n++)
   {
      if (shm->pixels_read[n] < TESTCCD_MINPIXELS)
      {
	 printf("warning: Not enough pixels from %s\n", leach[n].name_contr);
	 rtn = FAIL;
      }
      else
      {
	 int pixels_per_amp = shm->pixels_read[n];

	 if (pixels_per_amp > TESTCCD_MAXPIXELS)
	    pixels_per_amp = TESTCCD_MAXPIXELS;

	 for (c = 0; c < shm->n_amps_per_controller; c++)
	 {
	    signed short* p;
	    unsigned short compare, min, max;
	    int pixels, differ, zeropix;
	    double average;

	    p = shm->buf_base + shm->buf_ofs[n][c];
	    compare = NTOHS(*p) + 32768;
	    pixels = pixels_per_amp;
	    max = differ = zeropix = 0;
	    min = 0;
	    average = 0.;

	    while (--pixels)
	    {
	       register unsigned short pixval;

	       p++;
	       pixval = NTOHS(*p) + 32768;

	       if (pixval == 0) zeropix++;
	       if (pixval != compare) { differ = 1; }
	       if (pixval < min) min = pixval;
	       if (pixval > max) max = pixval;
	       average += pixval;
	    }
	    average /= (double)pixels_per_amp;

	    printf("logonly: %s amp %d pix average=%d min=%d max=%d\n",
		   leach[n].name_contr, c, (int)average, min, max);

	    if (average >= 65000)
	    {
	       printf("warning: %s amp %d appears to be completely saturated (avg=%d).\n",
		      leach[n].name_contr, c, (int)average);
	       /* But still return PASS */
	    }
	    else if (differ == 0)
	    {
	       printf("warning: %s amp %d seems to be all %d's\n",
		      leach[n].name_contr, c, compare);
	       rtn = FAIL;
	    }
	    else
	    {
	       /*
		* === if (zeropix) ... Allowing a few zero pixels.
		*/
	       int zpct = zeropix * 100 / pixels_per_amp;

	       if (zpct > 50) /* Only fail if > 50% zero pixels */
	       {
		  printf("warning: %s amp %d has %d%% zero pixels\n",
			 leach[n].name_contr, c, zpct);
		  rtn = FAIL;
	       }
	       if (average < 50)
	       {
		  printf("warning: %s amp %d has averages %d.\n",
			 leach[n].name_contr, c, (int)average);
		  rtn = FAIL;
	       }
	    }
	 }
      }
   }
   if (rtn == FAIL)
      printf("error: Data in buffer doesn't look like a CCD.\n");

   return rtn;
}

static pid_t pid_data = (pid_t)-1;
static pid_t pid_bin8 = (pid_t)-1;
static BOOLEAN bin8_used = FALSE;

PASSFAIL
com_abortdata( const char* args )
   /*
    * Return code indicates condition of previous data writer, if any.
    * So, if there was no data writer, or if it exited successfully before
    * it could be killed, abortdata returns PASS.
    *
    * A FAIL from abortdata means that it killed the data writer, and
    * the last data file is probably no good.  (So, this is a little
    * confusing, because if it succeeds in what it is supposed to do,
    * it actually returns FAIL! === No longer needed since not called
    * from waitdata anymore?)
    */
{
   child_exit_t exit_data;
   child_exit_t exit_bin8;

   if (pid_data==(pid_t)-1 && (!bin8_used || pid_bin8==(pid_t)-1))
   {
      printf("\rlogonly: abortdata: No FITS writer running.\n");
      return PASS;
   }

   exit_data = pid_abort(pid_data);
   exit_bin8 = pid_abort(pid_bin8);

   /*
    * Case 1: Both aborted
    */
   if ((exit_data == CHILD_EXITSIG || exit_data == CHILD_EXITFAIL) &&
       (!bin8_used || (exit_bin8 == CHILD_EXITSIG || exit_bin8 == CHILD_EXITFAIL)))
   {
      pid_data = -1;
      pid_bin8 = -1;
      data_state(DATA_ABORTED);
      printf("Filename `%s' can be re-used for next exposure.\n", filebase());
      return FAIL;
   }

   /*
    * Case 2: Both already completed (too late to abort)
    */
   if ((exit_data == CHILD_GONE || exit_data == CHILD_EXITPASS) &&
       (!bin8_used || (exit_bin8 == CHILD_GONE || exit_bin8 == CHILD_EXITPASS)))
   {
      pid_data = -1;
      pid_bin8 = -1;
      data_state(DATA_SAVED);
      return PASS;
   }

   /*
    * Case 3: Mix.
    */
   pid_data = -1;
   pid_bin8 = -1;
   data_state(DATA_FAILED);
   return PASS;
}


PASSFAIL
com_waitdata( const char* args ) /* === What should args do? */
   /*
    * The return from waitdata indicates the status of the previous
    * data writer, or PASS if there wasn't one.  If anything went
    * wrong with the previous data writer, FAIL is returned.
    * This command will wait indefinitely for the data-writer
    * PID to exit.  The user can hit ^C/abort to give up and send
    * the TERM signal to the data writer, however.
    */
{
   int poll = 0;

   if (args && !strcmp(args, "-poll")) poll=1;
   if (pid_bin8!=(pid_t)-1)
   {
      CLI_DEBUG(printf("\rdebug: com_waitdata() waiting for preview\n"));
      switch (pid_wait(pid_bin8, poll))
      {
	 case CHILD_EXITPASS: 
	 case CHILD_GONE:     break;
	 case CHILD_RUNNING:  return FAIL;
	 case CHILD_EXITSIG:  
	 case CHILD_EXITFAIL: 
	 case CHILD_ERROR:
	 {
	    printf("warning: Binned image not saved successfully.\n");
	    break;
	 }
      }
      pid_bin8=-1;
   }

   if (args && !strcmp(args, "-preview")) return PASS;

   if (pid_data==(pid_t)-1) return PASS;

   CLI_DEBUG(printf("\rdebug: com_waitdata() waiting for save to complete\n"));
   switch (pid_wait(pid_data, poll))
   {
      case CHILD_EXITPASS: 
      case CHILD_GONE:
      {
	 pid_data = -1;
	 data_state(DATA_SAVED);
	 com_readout("wait"); /* Make sure the readout thread has been
			       * waited on by now too, otherwise DetCom
			       * may never go back to "sweeping" mode.
			       */
	 /*
	  * NOTE: It would be better if readout wait implemented its own
	  * polling wait so that detcom.c could call the directly.  Right
	  * now, the statusbar says "Reading" sometimes past the true end
	  * of the readout.
	  */
	 return PASS;
      }
      case CHILD_RUNNING: return FAIL;
      case CHILD_EXITSIG:  
      case CHILD_EXITFAIL: 
      case CHILD_ERROR:
      {
	 printf("warning: Previous `save' command did not complete successfully.\n");
	 pid_data = -1;
	 data_state(DATA_FAILED);
	 com_readout("wait"); /* See above */
	 return FAIL;
      }
   }
   return FAIL; /* Not reached */
}

#define CHIPID_PHU (-1)
#define CHIPID_B8 (99)

static PASSFAIL
write_header_unit( int fd, int n_chips, int chipid, int xtension )
{
   static char imageswv[80] = "";
   static time_t date;
   const raster_t* raster = raster_exposed();
   const wcs_per_chip* wcs = 0;
   PASSFAIL rtn;
   char tmpstr[256];
   HeaderUnit hu = fh_create();

   fh_reserve(hu, detcom_reserve); /* Leave space for more FITS cards
				    * (in addition to the ones DetCom puts in.)
				    */

   if (xtension && (chipid < 0 || chipid > 11))
   {
      printf("Invalid chipid (%d) for image extension\n", chipid);
      return FAIL;
   }

   if (!xtension)
   {
      fh_set_bool(hu,	   0,	"SIMPLE",	1,		"Standard FITS");
      time(&date);
   }
   else
   {
      fh_set_str(hu,	   0,	"XTENSION",	"IMAGE",	"Image extension");
      /* Don't take a new timestamp for extensions. */
   }

   /*
    * These keywords should appear once, in the primary, or only header unit,
    * but I'm sticking them in the extensions too, just to make converting
    * between split/MEF easier.
    */
   fh_set_str(hu,	 101,	"FILENAME",	filebase(),	"Base filename at acquisition");
   fh_set_str(hu,        101.01,"PATHNAME",     cli_getcwd(),   "Original directory name at acquisition");
   strftime(tmpstr, sizeof(tmpstr)-1, "%Y-%m-%dT%T", gmtime(&date));
   fh_set_str(hu,	 104,	"DATE",		tmpstr,		"UTC Date of file creation");
   strftime(tmpstr, sizeof(tmpstr)-1, "%a %b %d %H:%M:%S %Z %Y", localtime(&date));
   fh_set_str(hu,	 104.1,	"HSTTIME",	tmpstr,		"Local time in Hawaii");
   fh_set_str(hu,	 105,	"ORIGIN",	"CFHT",		"Canada-France-Hawaii Telescope");

   if (!*imageswv)
      sprintf(imageswv, "CFHT DetCom v%.2f (%.30s)", VERSION, SOURCEDATE);
   fh_set_str(hu,	 110,	"IMAGESWV",	imageswv, 	"Image creation software version");

   fh_set_com(hu,	90.0,	"COMMENT",	"");
   fh_set_com(hu,	90.1,	"COMMENT",	"Observation summary");
   fh_set_com(hu,	90.2,	"COMMENT",	"-------------------");
   /*
    * These keywords can appear in any header unit.
    */
   fh_set_com(hu,	1000.0,	"COMMENT",	"");
   fh_set_com(hu,	1000.1,	"COMMENT",	"Detector array and CCD Information");
   fh_set_com(hu,	1000.2,	"COMMENT",	"----------------------------------");
#ifdef CAMERA_CFH12K
   fh_set_str(hu,	1001,	"DETECTOR",	"CFH12K",	"CFHT 12K x 8K Mosaic CCD camera");
   fh_set_str(hu,	1002,	"DETSIZE",	"[1:12450,1:8230]", "Total data pixels in full mosaic");
#else
#ifdef CAMERA_MDM8K
   fh_set_str(hu,       1001,   "DETECTOR",     "MDM8K",        "MDM 8K x 8K Mosaic CCD camera");
   fh_set_str(hu,	1002,	"DETSIZE",	"[1:8250,1:8250]", "Total data pixels in full mosaic");
#else

   /*
    * Set these here, even if handlers provide them, so that their
    * place near the top of the FITS header is reserved.  Since DetCom
    * has already filled in all of its keywords by the time the handler
    * keywords are merged in, they appear after internal DetCom keywords
    * even if their sorting number is lower.  (Unless they are reserved
    * by also having DetCom set them, like this.)
    */
   fh_set_str(hu, 111.0, "DETECTOR", "Unknown", "Science Detector");
   fh_set_str(hu, 112.0, "INSTRUME", "Unknown", "Instrument Name");
   fh_set_str(hu, 113.0, "INSTMODE", "Unknown", "Instrument Mode");

#endif
#endif
   sprintf(tmpstr, "%d %d", raster->nsbin, raster->npbin);
   fh_set_str(hu,	1005,	"CCDSUM",	tmpstr,		"Binning factors");
   fh_set_int(hu,	1005.1,"CCDBIN1",	raster->nsbin,	"Binning factor along first axis");
   fh_set_int(hu,	1005.2,"CCDBIN2",	raster->npbin,	"Binning factor along second axis");
   fh_set_str(hu,	1006,	"RASTER",	raster->command,"");
   /* %%% HACK: PIXSIZE and PIXSCAL are 12K-specific only! */
#if 0
   fh_set_flt(hu,	1010.1,	"PIXSIZE1",	15.0, 3,	"Pixel size for axis 1 (microns)");
   fh_set_flt(hu,	1010.2,	"PIXSIZE2",	15.0, 3,	"Pixel size for axis 2 (microns)");
   fh_set_flt(hu,	1011.1,	"PIXSCAL1",	0.206, 4,	"Pixel scale for axis 1 (arcsec/pixel)");
   fh_set_flt(hu,	1011.2,	"PIXSCAL2",	0.206, 4,	"Pixel scale for axis 2 (arcsec/pixel)");
#endif
   fh_set_str(hu,	1400,	"CONSWV",	dsp_version(),	"Controller DSPID and SERNO versions");
   fh_set_str(hu,	1500,	"DETSTAT",	temp_status(),	temp_status_comment());
   fh_set_flt(hu,	1500.1,	"DETTEM",	temp_float(),3,	temp_comment());

   if (chipid == CHIPID_PHU)
   {
      /*
       * These keywords only apply to a primary header unit with extensions.
       */
      fh_set_int(hu,	   1,	"BITPIX",	16,		"Bits per pixel (not applicable)");
      fh_set_int(hu,	   2,	"NAXIS",	0,		"No image data with primary header");
      fh_set_bool(hu,	   3,	"EXTEND",	1,		"File contains extensions");
      fh_set_int(hu,	   3.1,	"NEXTEND",      n_chips,	 "Number of extensions");
      
      sprintf(tmpstr,
	      "file %.20s, raster %.30s, etype %.20s, etime %.20s, filter %.20s",
	      filebase(),
	      raster_command(),
	      etype_command(),
	      etime_command(),
	      filter_name());
      fh_set_com(hu,	90.3,	"COMMENT",	tmpstr);
      fh_set_com(hu,	90.4,	"COMMENT",	"");

      fh_set_com(hu,	 100.0,	"COMMENT",	"");
      fh_set_com(hu,	 100.1,	"COMMENT",	"General Information");
      fh_set_com(hu,	 100.2,	"COMMENT",	"-------------------");
      fh_set_com(hu,	 100.4,	"COMMENT",	"");

     fh_set_int(hu,	1003,	"NCCDS",	n_chips,	"Number of devices");
     fh_set_int(hu,	1004,	"NAMPS",	n_chips,	"Number of amplifiers");
   }
   else
   {
      /*
       * These keywords appear in any other header unit with real data.
       */
      fh_set_int(hu,	   1,	"BITPIX",	16,		"16bit 2s complement 0..65535 (BZERO=32768)");
      fh_set_int(hu,	   2,	"NAXIS",	2,		"Number of axes");
      if (chipid == CHIPID_B8)
      {
	 /*
	  * Keywords for the binned-by-8 composite image
	  */
	 fh_set_int(hu,	   2.1,	"NAXIS1",
		    (raster->nsr + raster->nsovr)*
		    (n_chips / n_controllers)/(8/raster->nsbin),
		    "Number of pixel columns");
	 fh_set_int(hu,	   2.2,	"NAXIS2",
		    (raster->npr + raster->npovr)*
		    n_controllers/(8/raster->npbin),
		    "Number of pixels rows");
      }
      else
      {
	 fh_set_int(hu,	   2.1,	"NAXIS1",	raster->nsr + raster->nsovr,	"Number of pixel columns");
	 fh_set_int(hu,	   2.2,	"NAXIS2",	raster->npr + raster->npovr,	"Number of pixels rows");
      }
      fh_set_int(hu,	   5,	"PCOUNT",	0,			"No 'random' parameters");
      fh_set_int(hu,	   6,	"GCOUNT",	1,			"Only one group");

      if (chipid == CHIPID_B8)
      {
	 sprintf(tmpstr,
		 "file %.20s-B8, raster %.30s, etype %.20s, etime %.20s, filter %.20s",
		 filebase(),
		 raster_command(),
		 etype_command(),
		 etime_command(),
		 filter_name());
      }
      else
      {
	 sprintf(tmpstr,
		 "file %.20s%02d, raster %.30s, etype %.20s, etime %.20s, filter %.20s",
		 filebase(), chipid,
		 raster_command(),
		 etype_command(),
		 etime_command(),
		 filter_name());
      }
      fh_set_com(hu,	90.3,	"COMMENT",	tmpstr);
      fh_set_com(hu,	90.4,	"COMMENT",	"");

      /*
       * This should really only appear in true extensions, but
       * include it anyway to make generating MEF from separate
       * files easier.
       */
      fh_set_bool(hu,	 199,	"INHERIT",	0,		"No need to inherit global keywords");

      fh_set_com(hu,	 100.0,	"COMMENT",	"");
#ifdef CAMERA_CFH12K
      sprintf(tmpstr, "                          %c06%c07%c08%c09%c10%c11%c",
	      chipid==6?'[':' ',
	      chipid==6?']':(chipid==7?'[':' '),
	      chipid==7?']':(chipid==8?'[':' '),
	      chipid==8?']':(chipid==9?'[':' '),
	      chipid==9?']':(chipid==10?'[':' '),
	      chipid==10?']':(chipid==11?'[':' '),
	      chipid==11?']':' ');
      fh_set_com(hu,	 100.1,	"COMMENT",	tmpstr);
#else
#ifdef CAMERA_MDM8K
      /* %%% MDM8K */
      sprintf(tmpstr, "                          %c04%c05%c06%c07%c",
	      chipid==4?'[':' ',
	      chipid==4?']':(chipid==5?'[':' '),
	      chipid==5?']':(chipid==6?'[':' '),
	      chipid==6?']':(chipid==7?'[':' '),
	      chipid==7?']':' ');
      fh_set_com(hu,	 100.1,	"COMMENT",	tmpstr);
#endif
#endif

      if (chipid == CHIPID_B8)
	 sprintf(tmpstr, "Binned-by-8 composite");
      else
	 sprintf(tmpstr, "Image data for chip %02d", chipid);

      fh_set_com(hu,	 100.2,	"COMMENT",	tmpstr);
#ifdef CAMERA_CFH12K
      sprintf(tmpstr, "----------------------    %c00%c01%c02%c03%c04%c05%c",
	      chipid==0?'[':' ',
	      chipid==0?']':(chipid==1?'[':' '),
	      chipid==1?']':(chipid==2?'[':' '),
	      chipid==2?']':(chipid==3?'[':' '),
	      chipid==3?']':(chipid==4?'[':' '),
	      chipid==4?']':(chipid==5?'[':' '),
	      chipid==5?']':' ');
      fh_set_com(hu,	 100.3,	"COMMENT",	tmpstr);
#else
#ifdef CAMERA_MDM8K
      sprintf(tmpstr, "----------------------    %c00%c01%c02%c03%c",
	      chipid==0?'[':' ',
	      chipid==0?']':(chipid==1?'[':' '),
	      chipid==1?']':(chipid==2?'[':' '),
	      chipid==2?']':(chipid==3?'[':' '),
	      chipid==3?']':' ');
#endif
#endif
      fh_set_com(hu,	 100.4,	"COMMENT",	"");

      if (chipid == CHIPID_B8)
	 sprintf(tmpstr, "bin8");
      else
	 sprintf(tmpstr, "chip%02d", chipid);

      fh_set_str(hu,	 101.1,	"EXTNAME",	tmpstr,			"Extension name");
/*
  NOAO recommends this, but it doesn't make much sense.
      fh_set_int(hu,	 101.2,	"EXTVER",	chipid,			"Same as CHIPID or IMAGEID");
      */
      fh_set_int(hu,	 151,	"IMAGEID",	chipid,		"CCD chip number");
      fh_set_int(hu,	 151.1,	"CHIPID",	chipid,		"Use IMAGEID instead");
      fh_set_flt(hu,	1101,	"BZERO",	32768.0, 6,		"Zero factor");
      fh_set_flt(hu,	1102,	"BSCALE",	1.0, 2,			"Scale factor");
   }

   wcs = 0;

#ifdef CAMERA_CFH12K
{
const wcs_per_chip wcs_12k[12] = {
   { 6340.3,	4072.6,	5.607522e-05,	0.0000,	0.0000,	-5.666794e-05 }, /*00*/
   { 4210.6,	4042.3,	5.665741e-05,	0.0000,	0.0000,	-5.679697e-05 }, /*01*/
   { -58.5,	4042.0,	-5.705934e-05,	0.0000,	0.0000,	-5.692411e-05 }, /*02*/
   { 2020.7,	4040.2,	-5.708236e-05,	0.0000,	0.0000,	-5.699236e-05 }, /*03*/
   { -2083.0,	4053.3,	5.678322e-05,	0.0000,	0.0000,	-5.688525e-05 }, /*04*/
   { -4227.3,	4078.4,	5.604337e-05,	0.0000, 0.0000, -5.664017e-05 }, /*05*/
   { 6338.3,	4228.7,	5.606953e-05,	0.0000,	0.0000,	5.660860e-05  }, /*06*/
   { 4202.7,	4204.8,	5.664367e-05,	0.0000,	0.0000,	5.678215e-05 }, /*07*/
   { 2111.9,	4191.7,	5.696350e-05,	0.0000,	0.0000,	5.688183e-05 }, /*08*/
   { 24.1,	4192.9,	5.702455e-05,	0.0000,	0.0000,	5.689634e-05 }, /*09*/
   { -2085.4,	4204.5,	5.678862e-05,	0.0000,	0.0000,	5.677299e-05 }, /*10*/
   { -4224.4,	4218.4,	5.602513e-05,	0.0000,	0.0000,	5.660209e-05 }, /*11*/
};

const wcs_per_chip wcs_b8 =
{ 780.0, 516.0, 5.697796E-5, 2.362042E-7, -3.4851269E-7, 5.68674775E-5 };

   switch (chipid)
   {
      case CHIPID_B8:
      {
	 /*
	  * %%% These keywords are not really correct.
	  */
	 fh_set_flt(hu,	 155,	"GAIN",		1.4, 3, "Amplifier gain (electrons/ADU)");
	 fh_set_str(hu,	1501,	"DATASEC",	"[1:12450,1:8230]", "Total data pixels in full mosaic");
	 fh_set_str(hu,	1502,	"DETSEC",	"[1:12450,1:8230]", "Total data pixels in full mosaic");
	 wcs = &wcs_b8;
	 break;
      }
      case 0:
      {
	 fh_set_flt(hu,	 155,	"GAIN",		1.5, 3, "Amplifier gain (electrons/ADU)");
	 fh_set_str(hu,	1502,	"DETSEC",	"[1:2045,1:4096]", "");
	 wcs = &wcs_12k[0];
	 break;
      }
      case 1:
      {
	 fh_set_flt(hu,	 155,	"GAIN",		1.5, 3, "Amplifier gain (electrons/ADU)");
	 fh_set_str(hu,	1502,	"DETSEC",	"[2082:4126,1:4096]", "");
	 wcs = &wcs_12k[1];
	 break;
      }
      case 2:
      {
	 fh_set_flt(hu,	 155,	"GAIN",		1.6, 3, "Amplifier gain (electrons/ADU)");
	 fh_set_str(hu,	1502,	"DETSEC",	"[6199:4156,1:4096]", "");
	 wcs = &wcs_12k[2];
	 break;
      }
      case 3:
      {
	 fh_set_flt(hu,	 155,	"GAIN",		2.1, 3, "Amplifier gain (electrons/ADU)");
	 fh_set_str(hu,	1502,	"DETSEC",	"[8279:6236,1:4096]", "");
	 wcs = &wcs_12k[3];
	 break;
      }
      case 4:
      {
	 fh_set_flt(hu,	 155,	"GAIN",		1.3, 3, "Amplifier gain (electrons/ADU)");
	 fh_set_str(hu,	1502,	"DETSEC",	"[8316:10359,1:4096]", "");
	 wcs = &wcs_12k[4];
	 break;
      }
      case 5:
      {
	 fh_set_flt(hu,	 155,	"GAIN",		1.7, 3, "Amplifier gain (electrons/ADU)");
	 fh_set_str(hu,	1502,	"DETSEC",	"[10400:12443,1:4096]", "");
	 wcs = &wcs_12k[5];
	 break;
      }
      case 6:
      {
	 fh_set_flt(hu,	 155,	"GAIN",		1.8, 3, "Amplifier gain (electrons/ADU)");
	 fh_set_str(hu,	1502,	"DETSEC",	"[13:2056,8216:4121]", "");
	 wcs = &wcs_12k[6];
	 break;
      }
      case 7:
      {
	 fh_set_flt(hu,	 155,	"GAIN",		1.7, 3, "Amplifier gain (electrons/ADU)");
	 fh_set_str(hu,	1502,	"DETSEC",	"[2092:4136,8209:4114]", "");
	 wcs = &wcs_12k[7];
	 break;
      }
      case 8:
      {
	 fh_set_flt(hu,	 155,	"GAIN",		1.4, 3, "Amplifier gain (electrons/ADU)");
	 fh_set_str(hu, 1502, "DETSEC",		"[4161:6204,8208:4113]", "");
	 wcs = &wcs_12k[8];
	 break;
      }
      case 9:
      {
	 fh_set_flt(hu,	 155,	"GAIN",		1.5, 3, "Amplifier gain (electrons/ADU)");
	 fh_set_str(hu,	1502,	"DETSEC",	"[6242:8285,8215:4120]", "");
	 wcs = &wcs_12k[9];
	 break;
      }
      case 10:
      {
	 fh_set_flt(hu,	 155,	"GAIN",		1.8, 3, "Amplifier gain (electrons/ADU)");
	 fh_set_str(hu,	1502,	"DETSEC",	"[8315:10358,8208:4113]", "");
	 wcs = &wcs_12k[10];
	 break;
      }
      case 11:
      {
	 fh_set_flt(hu,	 155,	"GAIN",		1.5, 3, "Amplifier gain (electrons/ADU)");
	 fh_set_str(hu,	1502,	"DETSEC",	"[10395:12438,8201:4106]", "");
	 wcs = &wcs_12k[11];
	 break;
      }
   }
}
#endif /* CAMERA_CFH12K */
   if (wcs)
   {
      fh_set_flt(hu, 1201.1,	"CRPIX1", wcs->crpix1 / raster->nsbin,	5, "WCS Coordinate reference pixel");
      fh_set_flt(hu, 1201.2,	"CRPIX2", wcs->crpix2 / raster->npbin,	5, "WCS Coordinate reference pixel");
      fh_set_flt(hu, 1205.101,	"CD1_1",  wcs->cd1_1 * raster->nsbin,	5, "WCS Coordinate scale matrix");
      fh_set_flt(hu, 1205.102,	"CD1_2",  wcs->cd1_2,			5, "WCS Coordinate scale matrix");
      fh_set_flt(hu, 1205.201,	"CD2_1",  wcs->cd2_1,			5, "WCS Coordinate scale matrix");
      fh_set_flt(hu, 1205.202,	"CD2_2",  wcs->cd2_2 * raster->npbin,	5, "WCS Coordinate scale matrix");
   }
/*
 * If these are set to 0, the bias region starts on the exact row and column
 * that the DSP code thinks it should.  If these are set to 1, a pixel column
 * is stripped off, typically resulting in a 31 instead of 32 pixel wide
 * overscan region with our DSP code.
 */
#define BIASSEC_SSKIP 1
#define BIASSEC_PSKIP 1

/*
 * Include three more pixels in DATASEC rows ??? Why?
 * Why start at 6,5 ??? %%%
 */
#ifdef CAMERA_CFH12K
#define DATASEC_NPR_HACK 3
#else
#define DATASEC_NPR_HACK 0
#endif

   if (chipid != CHIPID_B8 && chipid != CHIPID_PHU)
   {
      shm_t* shm = im_buf_get(FALSE);

      /*
       * DETSIZE (DSP code may override this one)
       * -------
       */
      sprintf(tmpstr, "[%d:%d,%d:%d]",
	      1, shm->cols_per_amp - 32 /* %%% cols_overscan */,
	      1, shm->rows_per_amp - 32 /* %%% rows_overscan */);
      if (n_chips == 1)
      {
	 /* fh_set_str(hu, 1500.5, "DETSIZE", tmpstr, ""); %%% Add this later? */
	 fh_set_str(hu, 1500.6, "CCDSIZE", tmpstr, "");
      }
      /* fh_set_str(hu, 1500.7, "AMPSIZE", tmpstr, ""); %%% Add this later? */
      /* %%% DETSEC needed too? */

      /*
       * CCDSEC
       * ------
       */
      sprintf(tmpstr, "[%d:%d,%d:%d]",
	      1 + raster->nsskip,
	      1 + raster->nsskip + (raster->nsr + BIASSEC_SSKIP - 6 / raster->nsbin)*raster->nsbin,
	      1 + raster->npskip,
	      1 + raster->npskip + (raster->npr + BIASSEC_PSKIP - 5 / raster->npbin)*raster->npbin);
      fh_set_str(hu, 1500.8, "CCDSEC", tmpstr, "");

      /*
       * DATASEC
       * -------
       */
      sprintf(tmpstr, "[%d:%d,%d:%d]",
	      6 / raster->nsbin, raster->nsr+BIASSEC_SSKIP,
	      5 / raster->npbin, raster->npr+BIASSEC_PSKIP + DATASEC_NPR_HACK / raster->npbin);
      fh_set_str(hu,	1501,	"DATASEC",	tmpstr, "");

      /*
       * BIASSEC
       * -------
       */
      if (raster->nsovr <= BIASSEC_SSKIP &&
	  raster->npovr <= BIASSEC_PSKIP)
	 strcpy(tmpstr, "");
      else if (raster->nsovr > BIASSEC_SSKIP)
	 sprintf(tmpstr, "[%d:%d,%d:%d]",
		 raster->nsr+1+BIASSEC_SSKIP, raster->nsr+raster->nsovr,
		 1+BIASSEC_PSKIP, raster->npr+raster->npovr);
      else
	 sprintf(tmpstr, "[%d:%d,%d:%d]",
		 1+BIASSEC_SSKIP, raster->nsr+raster->nsovr,
		 raster->npr+1+BIASSEC_PSKIP, raster->npr+raster->npovr);
      fh_set_str(hu,	1503,	"BIASSEC",	tmpstr,		"Overscan area of frame");
   }

   fh_set_str(hu,	210,	"OBSTYPE",	etype_fits(), "Observation/ Exposure type");
   fh_set_str(hu,	210.1,	"EXPTYPE",	etype_fits(), "OBSTYPE is the new name for this");
   if (filesequence()!=-1)
   {
      fh_set_int(hu,	211,	"EXPNUM",	filesequence(),	 "CFHT odometer number");
      fh_set_int(hu,	211,	"OBSID",	filesequence(),	 "CFHT odometer number");
   }

   if (have_etime())
   {
      fh_set_flt(hu,	220,	"EXPTIME",	etime_meas(), 5,	"Integration time (seconds)");
      /* === CFHT used INTTIME before */
      fh_set_flt(hu,	221,	"DARKTIME",	etime_meas(), 5,	"Dark current time (seconds)");
      /* === This isn't really measured properly, but doesn't matter for 12k. */
   }
   else
   {
      fh_set_flt(hu,	220,	"EXPTIME",	0., 5,	"No integration time (bias)");
      fh_set_flt(hu,	221,	"DARKTIME",	0., 5,	"Dark current time (seconds)");
   }
   if (have_etime() && have_shutter())
   {
      fh_set_flt(hu,	2026.1,"SHUTOPEN",blade_open, 5,	"Shutter blade opening time (msec)");
      fh_set_flt(hu,	2026.2,"SHUTCLOS",blade_close,5,	"Shutter blade closing time (msec)");
   }
   else
   {
      fh_set_flt(hu,	2026.1,"SHUTOPEN",      0., 5,	"Shutter not activated");
      fh_set_flt(hu,	2026.2,"SHUTCLOS",      0.,5,	"Shutter not activated");
   }

   fh_set_str(hu,	2027.1,"FILTER",  filter_name(),     "Name of filter in position");
   fh_set_int(hu,	2027.2,"FILTERID",filter_id(),       "Filter index (see also FILTER)");

   fh_set_str(hu,	 201,	"TIMESYS",	"UTC",		"Time System for DATExxxx");
   fh_set_str(hu,	 201.1, "TSYSOBS",	"UTC",		"Time System for TIMExxx");

   {
      struct tm* tms;

      tms = gmtime(&detcom_obs_eip.tv_sec);
      sprintf(tmpstr, "%04d-%02d-%02d",
	      tms->tm_year+1900, tms->tm_mon, tms->tm_mday);
      fh_set_str(hu,	 202,	"DATE-OBS",	tmpstr,		"Date at start of observation (unix)");
      sprintf(tmpstr, "%d:%02d:%02d.%03ld",
	      tms->tm_hour, tms->tm_min, tms->tm_sec,
	      detcom_obs_eip.tv_usec / 1000);
      fh_set_str(hu,	 202.1,	"TIME-OBS",	tmpstr,		"Time at start of observation (unix)");
      tms = gmtime(&detcom_obs_don.tv_sec);
      sprintf(tmpstr, "%04d-%02d-%02d", tms->tm_year+1900, tms->tm_mon, tms->tm_mday);
      fh_set_str(hu,	 206,	"DATEEND",	tmpstr,	"Date at end of observation (unix)");
      sprintf(tmpstr, "%d:%02d:%02d.%03ld", tms->tm_hour, tms->tm_min, tms->tm_sec,
	      detcom_obs_don.tv_usec / 1000);
      fh_set_str(hu,	 206.1,	"TIMEEND",	tmpstr,		"Time at end of observation (unix)");
   }

#ifdef CAMERA_CFH12K
   fh_set_str(hu,	1100,	"INSTRUME",	"CFH12K Mosaic","Instrument used for observation");
   fh_set_str(hu,	 150,	"CCD",		"MIT/LL 2K x 4K CCD", 0);
#else
   fh_set_str(hu,	 150,	"CCD",		"Unknown", 0);
#endif
   fh_set_int(hu,	 152,	"MAXLIN",	65535, 0);
   fh_set_int(hu,	 152.1,	"SATURATE",	65535, 0);
   fh_set_str(hu,	  95,	"CMMTOBS",	fitscard("comment", 0), 0);
   fh_set_str(hu,	  95.1,	"CMMTSEQ",	fitscard("sequence", 0), 0);
   fh_set_str(hu,	  96,	"OBJECT",	fitscard("object", 0),   0);
   fh_set_str(hu,	  97,	"OBSERVER",	fitscard("observer", 0), 0);
   fh_set_str(hu,	  97.1,	"PI_NAME",	fitscard("_piname", 0), 0);
   fh_set_str(hu,	  97.2,	"RUNID",	fitscard("_runid", 0), 0);

   /*
    * Merge in arbitrary keywords specified with the nheader command.
    * dheader is used by DSP code and nheader is used by QSO.
    */
   if (dheader) fh_merge(hu, dheader);
   if (nheader) fh_merge(hu, nheader);
   if (nheader_nonsticky) fh_merge(hu, nheader_nonsticky);

   /*
    * Output all headers to the file.
    */
   rtn = fh_write(hu, fd);

   fh_destroy(hu);

   return rtn;
}

static off_t save_offset[MAX_CONTROLLERS][MAX_AMPS_PER_CONTROLLER];
static int save_fd[MAX_CONTROLLERS][MAX_AMPS_PER_CONTROLLER];
static int save_fd_main = -1;
static int save_binned = 0;
static int save_binned_fd;
static BOOLEAN save_multi_ext = TRUE;
static size_t save_size = 0;

static void
save_abort(void)
{
   int n, i;
   char filename[1024];
   char junkbuf[1024]; /* Build messages in this buffer and print them to stderr
			* in one call since this can get called from "background"
			* processes.  As long as everything else writing to stderr
			* does it the same way, it shouldn't mess up the terminal
			* too much :-/
			*/
   shm_t* shm = im_buf_get(FALSE);

   if (shm==0) return;

   sprintf(junkbuf, "\rerror: FITS writing aborted; Removing files `%.900s*.fits~'\n", filebase());
   fputs(junkbuf, stderr);

   if (save_binned==0)
   {
      for (n=0; n<n_controllers; n++)
      {
	 for (i=0; i<shm->n_amps_per_controller; i++)
	 {
	    int chipno = n*shm->n_amps_per_controller+i;

	    if (save_fd[n][i]!=-1)
	    {
	       close(save_fd[n][i]);
	       save_fd[n][i] = -1;
	    }
	    if (save_multi_ext == FALSE)
	    {
	       sprintf(filename, "%.200s/%.200s%02d.fits~",
		       filebase(), filebase(), chipno);
	       unlink(filename);
	    }
	 }
      }
      if (save_multi_ext == TRUE)
      {
	 if (save_fd_main!=-1)
	 {
	    close(save_fd_main);
	    save_fd_main = -1;
	 }
	 sprintf(filename, "%.200s.fits~", filebase());
	 unlink(filename);
	 sprintf(junkbuf, "\rinfo3: ** ABORTED ** writing file `%.900s.fits'\n", filebase());
	 fputs(junkbuf, stderr);
      }
      else
      {
	 sprintf(junkbuf, "\rinfo3: ** ABORTED ** writing files `%.900s*.fits'\n", filebase());
	 fputs(junkbuf, stderr);
      }
   }
   else
   {
      sprintf(filename, "%.200s-B8.fits~", filebase());
      close(save_binned_fd);
      unlink(filename);
      sprintf(junkbuf,
	      "\rinfo4: ** ABORTED ** writing binned composite fits `%.900s'\n",
	      filename);
      fputs(junkbuf, stderr);
   }

   _exit(EXIT_FAILURE);
}

#if defined(CAMERA_CFH12K) || defined(CAMERA_MDM8K)
static void
sigbus_handler(void)
{
   fprintf(stderr, "\rerror: FITS writer exiting on bus error.\n");
   cli_signal(SIGBUS, CLI_SIG_DFL);
   raise(SIGBUS);
}
#endif

static PASSFAIL
save_endgo(BOOLEAN preview)
{
   int maxfiles;

   switch (fork()) /* %%% Save PID and report failure!? */
   {
      case (pid_t)-1: perror("error: (save_endgo) fork failed"); return FAIL;

      case 0: break;
      default: return PASS;
   }
   setsid();
   maxfiles = sysconf(_SC_OPEN_MAX);

   while (maxfiles-- > 0) /* %%% Search for places where close-on-exec is needed! */
      close(maxfiles); /* %%% Duplication from cfht_goDaemon... add setsid there? */
   /* %%% No controller interfaces found is printed instead of detecting another detcom is running */

   open("/dev/null", O_RDWR); /* %%% Should return STDIN_FILENO (0) */
   dup2(STDIN_FILENO, STDOUT_FILENO);
   dup2(STDIN_FILENO, STDERR_FILENO);

   if (preview)
      execlp("endgo", "endgo", "--preview", filebase(), 0);
   else
      execlp("endgo", "endgo", filebase(), 0);

   _exit(EXIT_FAILURE);
   return FAIL;
}

#if defined(CAMERA_CFH12K) || defined(CAMERA_MDM8K)

/* %%% Hack.  Generalize to work for other configs. */
#ifdef CAMERA_MDM8K
#define CHIPS_PER_CONTROLLER 4
#else
#define CHIPS_PER_CONTROLLER 6
#endif

#define MAX_CONTROLLERS         2
#define ROWS_PER_CHIP           4128
#define COLS_PER_CHIP           2080
#define PIXELS_PER_CONTROLLER   (ROWS_PER_CHIP*COLS_PER_CHIP*CHIPS_PER_CONTROLLER)
#define PIXELS_PER_BIN8         (PIXELS_PER_CONTROLLER*MAX_CONTROLLERS/(8*8))

static inline signed short* imp_full(signed short* p, int r, int c)
{ return p + r * COLS_PER_CHIP + c; }

static inline signed short* imp_bin2(signed short* p, int r, int c)
{ return p + r * COLS_PER_CHIP/2 + c; }

static inline signed short* imp_bin4(signed short* p, int r, int c)
{ return p + r * COLS_PER_CHIP/4 + c; }

static inline signed short px(signed short* p)
{
   return NTOHS(*p);
}

   /* %%% #define other parameters here... */
   typedef enum { IMC_UL_R, IMC_UR_R, IMC_LL_R, IMC_LR_R,
		  IMC_UL_C, IMC_UR_C, IMC_LL_C, IMC_LR_C
   } im_corner_t;
   /*
    * IMC_XY_Z:
    * X = Upper or Lower end of chip first
    * Y = Right or Left side of chip first
    * Z = Row or Column reading (12k uses rows)
    *
    * Here are all eight possibilities (*_C modes
    * are not yet supported.)
    *
    * +----------------------------------+
    * | * ---> IMC_UL_R  IMC_UR_R <--- * |
    * | |                              | |
    * | |                              | |
    * | v IMC_UL_C            IMC_UR_C v |
    * |                                  |
    * |                                  |
    * |           Each Detector          |
    * |                                  |
    * |                                  |
    * | ^ IMC_LL_C            IMC_LR_R ^ |
    * | |                              | |
    * | |                              | |
    * | * ---> IMC_LL_R  IMC_LR_R <--- * |
    * +----------------------------------+
    */
   /* === Move this stuff to a configuration file */
#endif


PASSFAIL
com_savepreview( const char* args )
{
#if defined(CAMERA_CFH12K) || defined(CAMERA_MDM8K)
   /*
    * Corner and direction on each chip where the readout begins
    */
   const im_corner_t im_corner[MAX_CONTROLLERS][MAX_AMPS_PER_CONTROLLER] =
#ifdef CAMERA_MDM8K
   { { IMC_LL_R, IMC_LL_R, IMC_LL_R, IMC_LL_R },
     { IMC_UL_R, IMC_UL_R, IMC_UL_R, IMC_UL_R } };
#else
#ifdef CAMERA_CFH12K
   { { IMC_LL_R, IMC_LL_R, IMC_LR_R, IMC_LR_R, IMC_LL_R, IMC_LL_R },
     { IMC_UL_R, IMC_UL_R, IMC_UL_R, IMC_UL_R, IMC_UL_R, IMC_UL_R } };
#else
   /* === Was this for UH8K? */
   { { IMC_LL_R, IMC_LL_R, IMC_LR_R, IMC_LL_R },
      { IMC_UL_R, IMC_UL_R, IMC_UL_R, IMC_UL_R } };
#endif
#endif
   
   /*
    * Position on the final binned image of the mosaic of the
    * first pixel in each chip. { Row, Column }.
    */
#define B8_COLS ( COLS_PER_CHIP / 8 )
#define B8_ROWS ( ROWS_PER_CHIP / 8 )
#define B8_TOP  ( B8_ROWS * 2 - 1 )
#define B8_BOT  ( 0 )
   
   const int im_rows = B8_ROWS * MAX_CONTROLLERS;
   const int im_cols = B8_COLS * CHIPS_PER_CONTROLLER;
   const int im_pos[MAX_CONTROLLERS][CHIPS_PER_CONTROLLER][2] =

#ifdef CAMERA_MDM8K
   {{{ B8_BOT, 0 },    { B8_BOT, B8_COLS },   { B8_BOT, B8_COLS*2 },
     { B8_BOT, B8_COLS*3}},
    {{ B8_TOP, B8_COLS*3},  { B8_TOP, B8_COLS*2 },   { B8_TOP, B8_COLS },
     { B8_TOP, 0}}};
#else
#ifdef CAMERA_CFH12K
   {{{ B8_BOT, 0 },         { B8_BOT, B8_COLS },   { B8_BOT, B8_COLS*3-1 },
     { B8_BOT, B8_COLS*4-1 }, { B8_BOT, B8_COLS*4 }, { B8_BOT, B8_COLS*5 }},
    {{ B8_TOP, 0},          { B8_TOP, B8_COLS },   { B8_TOP, B8_COLS*2 },
     { B8_TOP, B8_COLS*3},  { B8_TOP, B8_COLS*4 }, { B8_TOP, B8_COLS*5 }}};
#else
   /* === UH8K? */
   {{{ B8_BOT, 0 },         { B8_BOT, B8_COLS },   { B8_BOT, B8_COLS*3-1 },
      { B8_BOT, B8_COLS*3 }},
      {{ B8_TOP, 0},          { B8_TOP, B8_COLS },   { B8_TOP, B8_COLS*2 },
	 { B8_TOP, B8_COLS*3}}};
#endif
#endif
   
#ifdef DO_BIAS_CORRECTIONS
   /*
    * Bias level to subtract from each chip.
    */
   const int imval_offset[MAX_CONTROLLERS][CHIPS_PER_CONTROLLER] =
#ifdef CAMERA_MDM8K
   { { 4125, 3625, 3300, 3485 },
     { 3900, 3700, 3655, 2905 } };
#else
#ifdef CAMERA_CFH12K
   { { 395, 456, 549, 600, 630, 705 },
     { 743, 763, 802, 888, 985, 1038 } };
#else
   /* === UH8K? */
   { { 395, 456, 549, 600 },
      { 743, 763, 802, 888 } };
#endif
#endif

   /*
    * 10X the scaling factor to multiply each chip.
    */
   const int imval_factor10[MAX_CONTROLLERS][CHIPS_PER_CONTROLLER] =
#ifdef CAMERA_MDM8K
   { { 15, 15, 16, 21 }, /* %%% These values are not the right ones! */
     { 18, 17, 14, 15 } };
#else
#ifdef CAMERA_CFH12K
   { { 15, 15, 16, 21, 13, 17 },
     { 18, 17, 14, 15, 18, 15 } };
#else
   { { 15, 15, 16, 21 },
      { 18, 17, 14, 15 } };
#endif
#endif
#endif /* DO_BIAS_CORRECTION */

   const raster_t* raster = raster_exposed();
   int save_bin;	/* Additional binning needed to get 8x8 final binning */
   int rowstep_full;	/* Number of pixels in a row of the full image */
   int rowstep_bin;	/* Number of pixels in the full image that make a binned row */

   com_waitdata(args);

   /*
    * === Check chip mask
    */
   if (xstrncasecmp(raster->command, "FULL", 4))
   {
      if (etype_char()!='x') /* This is normal for focus mode. */
	 printf("Cannot generate preview image with subraster selected.\n");
      return FAIL;
   }
   /*
    * Calculate the additional binning factor needed to get a final
    * image that is binned-8x8 of the entire mosaic.
    */

   if (raster->nsbin != raster->npbin)
   {
      printf("Cannot generate preview image because unequal X and Y binning factors.\n");
      return FAIL;
   }

   switch (raster->nsbin)
   {
      case 1: save_bin = 8; break;
      case 2: save_bin = 4; break;
      case 4: save_bin = 2; break;
      case 8: save_bin = 1; break;
      default:
      printf("Cannot generate 8x8-binned preview image with current binning factor.\n");
      return FAIL;
   }

   /*
    * Calculate the length of a row of pixels in the full image.
    */
   rowstep_full = (COLS_PER_CHIP / raster->nsbin);

   /*
    * Calculate the number of pixels in the full image that must be
    * available before a single row of the binned image can be written.
    */
   rowstep_bin = rowstep_full * save_bin;

   bin8_used = TRUE;
   switch ((pid_bin8 = fork()))
   {
      case (pid_t)-1: perror("error: savepreview fork failed"); return FAIL;

      case 0: /* Child takes care of writing to disk */
      {
	 shm_t* shm = im_buf_get(FALSE);
	 unsigned int pixels_written[MAX_CONTROLLERS];
	 unsigned char zerobuf[2880];
	 char filename0[1024];
	 char filename1[1024];
	 char junkbuf[1024];
	 int n_chips;
	 int done;
	 int nice_incr;
	 int bin8_size;
	 int bin8_ofs;
	 unsigned short* bin8_addr;

	 setsid();

	 cli_signal(SIGBUS, sigbus_handler); /* For mmap */

	 nice_incr = -nice(0) + 8;
	 if (nice_incr > 0) nice(nice_incr);

	 n_chips = shm->n_controllers * shm->n_amps_per_controller;

	 memset(zerobuf, 0, sizeof(zerobuf));
	 memset(pixels_written, 0, sizeof(pixels_written));

	 /*
	  * Generate a binned 8x8 composite FITS file
	  */
	 sprintf(filename0, "%.200s-B8.fits~", filebase());
	 save_binned_fd = open(filename0, O_RDWR|O_CREAT|O_TRUNC, 0644);
	 if (write_header_unit(save_binned_fd, n_chips, CHIPID_B8, 0)==FAIL)
	 {
	    perror("\rerror: save write_header_unit failed");
	    _exit(EXIT_FAILURE);
	 }
	 
	 bin8_size = PIXELS_PER_BIN8 * sizeof(short);
	 
	 /*
	  * Save the current offset to be used as the start of
	  * the image data for mmap later.
	  */
	 if ((bin8_ofs = lseek(save_binned_fd, 0, SEEK_CUR))==-1)
	 {
	    perror("\rlseek for binned 8x8 FITS file failed");
	    _exit(EXIT_FAILURE);
	 }

	 /*
	  * Seek to the end of the data, plus padding to make the
	  * file a multiple of 2880 bytes.
	  */
	 if (lseek(save_binned_fd,
		   ((bin8_size + 2880 - 1) / 2880) * 2880 - 1,
		   SEEK_CUR)==-1)
	 {
	    perror("\rlseek for binned 8x8 FITS file failed");
	    _exit(EXIT_FAILURE);
	 }

	 if (cli_write(save_binned_fd,
		       "\0",
		       1, TOMSEC_FILE)==FAIL)
	 {
	    perror("\rsavepreview cli_write failed to create space for data");
	    _exit(EXIT_FAILURE); /* === close file ? */
	 }

	 /*
	  * Leave the file pointer at the position where the binned
	  * data will go.  The cli_write() will copy it later.
	  */
	 if (lseek(save_binned_fd, bin8_ofs, SEEK_SET)==-1)
	 {
	    perror("\rlseek to bin8_ofs failed");
	    _exit(EXIT_FAILURE);
	 }
	 
	 bin8_addr = (unsigned short*)malloc(bin8_size);

	 /*
	  * Simplest to wait until this point before allowing abort cleanup-ups.
	  */
	 cli_signal_unblock(SIGTERM);

	 do
	 {
	    int delay_time = 250000;
	    int n;

	    done = 1;
	    for (n = 0; n<n_controllers; n++)
	    {
	       unsigned int available;

	       available =
		  shm->pixels_avail[n]/CHIPS_PER_CONTROLLER - pixels_written[n];

	       if (available>=rowstep_bin) /* %%% while? */
	       {
		  int i, j;

		  delay_time = 0;
		  for (i = 0; i<CHIPS_PER_CONTROLLER; i++)
		  {
		     register signed long acc, offset, divisor, factor;
		     register signed short* src;
		     unsigned short* dest;

		     src = shm->buf_base + shm->buf_ofs[n][i]
			+ pixels_written[n];
		     dest = bin8_addr +
			im_pos[n][i][1] + im_pos[n][i][0] * im_cols;

#ifdef DO_BIAS_CORRECTIONS
		     if (strchr("ofcx", etype_char()))
		     {
			offset = (32768 - imval_offset[n][i]) * 16;
			factor = imval_factor10[n][i];
			divisor = 160;
		     }
		     else
#endif
		     {
			offset = 32768 * 16;
			factor = 1;
			divisor = 16;
		     }

		     switch (im_corner[n][i])
		     {
			case IMC_UL_R:
			case IMC_UR_R:
			{
			   dest -= (pixels_written[n]/rowstep_bin)*im_cols;
			   break;
			}
			case IMC_LL_R:
			case IMC_LR_R:
			{
			   dest += (pixels_written[n]/rowstep_bin)*im_cols;
			   break;
			}
			case IMC_UL_C:
			case IMC_LL_C:
			{
			   dest += (pixels_written[n]/rowstep_bin)*im_rows;
			   break;
			}
			case IMC_UR_C:
			case IMC_LR_C:
			{
			   dest -= (pixels_written[n]/rowstep_bin)*im_rows;
			   break;
			}
		     }

		     for (j = 0; j < COLS_PER_CHIP / 8; j++)
		     {
			/*
			 * Take the sum of 16 of the 64 pixels to be averaged.
			 */
			switch (save_bin)
			{
			   case 8:
			   {
			      acc = offset +
				 px(imp_full(src,0,0))+ px(imp_full(src,0,2))+
				 px(imp_full(src,1,4))+ px(imp_full(src,1,6))+
				 px(imp_full(src,2,1))+ px(imp_full(src,2,7))+
				 px(imp_full(src,3,3))+ px(imp_full(src,3,5))+
				 px(imp_full(src,4,0))+ px(imp_full(src,4,2))+
				 px(imp_full(src,5,4))+ px(imp_full(src,5,6))+
				 px(imp_full(src,6,1))+ px(imp_full(src,6,7))+
				 px(imp_full(src,7,3))+ px(imp_full(src,7,5));
			      src = imp_full(src,0,8);
			      break;
			   }
			   case 4:
			   {
			      acc = offset +
				 px(imp_bin2(src,0,0))+ px(imp_bin2(src,0,1))+
				 px(imp_bin2(src,0,2))+ px(imp_bin2(src,0,3))+
				 px(imp_bin2(src,1,0))+ px(imp_bin2(src,1,1))+
				 px(imp_bin2(src,1,2))+ px(imp_bin2(src,1,3))+
				 px(imp_bin2(src,2,0))+ px(imp_bin2(src,2,1))+
				 px(imp_bin2(src,2,2))+ px(imp_bin2(src,2,3))+
				 px(imp_bin2(src,3,0))+ px(imp_bin2(src,3,1))+
				 px(imp_bin2(src,3,2))+ px(imp_bin2(src,3,3));
			      src = imp_bin2(src,0,4);
			      break;
			   }
			   case 2:
			   {
			      acc =
				 px(imp_bin4(src,0,0))+ px(imp_bin4(src,0,1))+
				 px(imp_bin4(src,1,0))+ px(imp_bin4(src,1,1));
			      acc *= 4;
			      acc += offset;
			      src = imp_bin4(src,0,2);
			      break;
			   }
			   case 1: /* === Efficiency */
			   {
			      acc = offset + px(src++) * 16;
			      break;
			   }
			   default: acc = 0;
			}
			/*
			 * Gain adjustment (different for each chip).
			 * This step also divides by 16 to turn the sum into an average.
			 */
			acc *= factor;
			acc /= divisor;
			/*
			 * Limit the average value to the range [0..65536]
			 */
			if (acc < 0) acc = 0;
			if (acc > 65535) acc = 65535;
			/*
			 * Convert back to signed 16-bit value and byte-swap
			 * if needed for writing the value to a FITS file.
			 */
			acc -= 32768;
			acc = NTOHS(acc);
			switch (im_corner[n][i])
			{
			   case IMC_UL_R:
			   case IMC_LL_R:
			   {
			      *dest++ = acc;
			      break;
			   }
			   case IMC_UR_R:
			   case IMC_LR_R:
			   {
			      *dest-- = acc;
			      break;
			   }
			   case IMC_UL_C:
			   case IMC_LL_C:
			   case IMC_UR_C:
			   case IMC_LR_C:
			   {
			      printf("error: Unsupported mosaic configuration.\n");
			      return FAIL;
			   }
			}
		     }

		     /* %%% fits_padding += rowstep_bin * sizeof(short); */
		  }
		  pixels_written[n] += rowstep_bin;
		  available -= rowstep_bin;
	       }

	       if (pixels_written[n]<shm->pixels_per_amp[n]) done = FALSE;
	    }
	    /* === Place in a second progress bar? */
	    if (shm->pixels_per_amp[0] && (n_controllers==1 || shm->pixels_per_amp[1]))
	    {
	       static time_t ts = 0;
	       time_t tn = time(0);

	       if (ts != tn)
	       {
		  ts = tn;
		  sprintf(junkbuf, "\rinfo4: Generating preview: %d%%\r",
			  n_controllers==1?
			  pixels_written[0]/((shm->pixels_per_amp[0]+99)/100):
			  (pixels_written[0]+pixels_written[1])/
			  ((shm->pixels_per_amp[0]+shm->pixels_per_amp[1]+99)/100));
		  fputs(junkbuf, stderr);
	       }
	    }
	    if (delay_time) xusleep(delay_time);
	 } while (!done);

	 fputs("\rinfo4: Copying preview to disk...\n", stderr);
	 if (cli_write(save_binned_fd, bin8_addr, bin8_size, TOMSEC_FILE)==FAIL)
	 {
	    perror("\rerror: savepreview cli_write failed");
	 }
	 free(bin8_addr);
	 bin8_addr = 0;

	 CLI_DEBUG(fprintf(stderr, "\rdebug: savepreview closing file\n"));
	 if (close(save_binned_fd)==-1)
	 {
	    perror("\rsavepreview close failed; data may be incomplete");
	    _exit(EXIT_FAILURE);
	 }

	 sprintf(filename0, "%.200s-B8.fits~", filebase());
	 sprintf(filename1, "%.200s-B8.fits", filebase());
	 if (rename(filename0, filename1)<0)
	 {
	    perror("\rrename of B8 at end of save failed");
	 }
	 sprintf(filename0, "%.500s/previous-B8.fits", cli_gethome());
	 sprintf(filename1, "%.500s/current-B8.fits", cli_gethome());
	 unlink(filename0);
	 rename(filename1, filename0);
	 sprintf(filename0, "%.500s/%.200s-B8.fits", cli_getcwd(), filebase());
	 if (link(filename0, filename1)<0 &&
	     symlink(filename0, filename1)<0)
	    perror("\rinfo4: current-B8.fits link failed");
	 else
	 {
	    sprintf(junkbuf, "\rinfo4:\nstatus: Done writing `%.900s-B8.fits', linked to `~/current-B8.fits'\n", filebase());
	    fputs(junkbuf, stderr);
	 }
	 if (save_endgo(/*preview=*/TRUE) != PASS)
	    fprintf(stderr, "\rerror: save_engo failed\n");

	 _exit(EXIT_SUCCESS);
      }
   }
#else
   if (args) ;
#endif
   return PASS;
}


PASSFAIL
com_save( const char* args )
{
   const char* cwd = cli_getcwd();
   shm_t* shm;

   /*
    * Do not allow overlapping "save" operations.
    */
   com_waitdata(args);

   /*
    * Ensure that filebase() returns an unused filename.
    * filename_next() grabs an unused filename (if necessary;
    *                 this probably already happened earlier.)
    * filename_used() flags it as currently used, since it
    *                 will be created in the next step in com_save.
    */
   filename_next();
   com_filename("used");

   shm = im_buf_get(FALSE);
   if (shm == 0) return FAIL;

   /*
    * Make sure this filename hasn't been used already.
    * NOTE: This ignores the binned-by-8 files.  They just get overwritten.
    * NOTE: Use same filename again if there's no new data.
    */
   if (data_state(DATA_STATE_QUERY) != DATA_SAVED)
   {
      char filename_test[1024];
      char filename_last[1024];
      struct stat st_buf;

      if (save_multi_ext==TRUE)
	 sprintf(filename_test, "%.200s.fits", filebase()); /* Single FITS */
      else
	 sprintf(filename_test, "%.200s", filebase()); /* Subdirectory */
      if (lstat(filename_test, &st_buf)==0)
      {
	 /*
	  * The current filename already exists!
	  */
	 sprintf(filename_last, "%.200s.LAST", filebase());
	 if (lstat(filename_last, &st_buf)==0)
	 {
	    if (unlink(filename_last) != 0) perror("logonly: unlink failed");
	    printf("warning: %s exists; removing %s and keeping last 2 exposures.\n",
		   filename_test, filename_last);
	 }
	 else
	 {
	    printf("warning: %s exists; renamed it to %s.  "
		   "Please change filenames or use a filename pattern with ?'s in it "
		   "if you don't want to lose data.\n", filename_test, filename_last);
	 }
	 
	 if (rename(filename_test, filename_last) != 0)
	 {
	    perror("logonly: rename failed");
	 }
      }
   }

   bin8_used = FALSE; /* com_savepreview() will set this back to TRUE if applicable */
   if (shm->n_controllers==2 &&
       shm->n_amps_per_controller==6)
   {
      /* %%% Hack: support preview for other mosaics too. */
      com_savepreview(args);
   }

   switch ((pid_data = fork()))
   {
      case (pid_t)-1: perror("error: (save) fork failed"); return FAIL;

      case 0: /* Child takes care of writing to disk */
      {
	 unsigned int pixels_written[MAX_CONTROLLERS];
	 int fits_padding[MAX_CONTROLLERS][MAX_AMPS_PER_CONTROLLER];
	 unsigned char zerobuf[2880];
	 char filename0[1024];
	 char filename1[1024];
	 char junkbuf[1024];
	 int done;
	 int n_chips;
	 int i, n;
	 int nice_incr;

	 setsid();

	 nice_incr = -nice(0) + 16;
	 if (nice_incr > 0) nice(nice_incr);

	 memset(zerobuf, 0, sizeof(zerobuf));
	 memset(pixels_written, 0, sizeof(pixels_written));

	 cli_signal_block(SIGTERM);
	 cli_signal(SIGTERM, save_abort);

	 n_chips = shm->n_controllers * shm->n_amps_per_controller;

	 /*
	  * Save information for PHU, even if writing separate files.
	  */
	 if (save_multi_ext==TRUE)
	 {
	    /*
	     * Remove previous/ and current/ if we are using MEF
	     */
	    sprintf(filename0, "%.500s/previous", cli_gethome());
	    sprintf(filename1, "%.500s/current", cli_gethome());
	    unlink(filename0);
	    unlink(filename1);
	    sprintf(filename0, "%.200s.fits~", filebase());
	 }
	 else
	 {
	    /*
	     * Remove previous.fits and current.fits if we are using separate files
	     */
	    sprintf(filename0, "%.500s/previous.fits", cli_gethome());
	    sprintf(filename1, "%.500s/current.fits", cli_gethome());
	    unlink(filename0);
	    unlink(filename1);

	    if (mkdir(filebase(), 0755)!=0 && errno!=EEXIST)
	    {
	       perror("\rerror: (save) can't create subdirectory for FITS files");
	       _exit(EXIT_FAILURE);
	    }
	    sprintf(filename0, "%.200s/%.200s.phu", filebase(), filebase());
	 }

	 if ((save_fd_main = open(filename0, O_RDWR|O_CREAT|O_TRUNC, 0644))==-1)
	 {
	    perror("\rerror: (save) can't open file");
	    _exit(EXIT_FAILURE);
	 }

	 if (n_chips != 1)
	 {
	    /*
	     * Write primary header unit.
	     */
	    if (write_header_unit(save_fd_main, n_chips, CHIPID_PHU, 0)==FAIL)
	    {
	       perror("\rerror: (save) can't write PHU");
	       close(save_fd_main);
	       _exit(EXIT_FAILURE);
	    }
	 }

	 if (save_multi_ext==TRUE)
	 {
	    /*
	     * Write a single multi-extension FITS file
	     * or a single single basic FITS if 1 amplifier.
	     */

	    int chipidx;
	    /*
	     * This particular order makes the MEF files display faster
	     * with current versions of Iraf and the NOAO mosaic tool.
	     * It's a hack, though, and can be confusing to people, so
	     * we store them in order and wait for a way to configure
	     * Iraf to know which chips it should start rendering first.
            int chipno[12] = { 2, 3, 9, 8, 7, 1, 4, 10, 6, 0, 5, 11 };
	     */
	       
	    int chipno[12] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 }; /* %%% Hack */

	    for (chipidx = 0; chipidx < n_chips; chipidx++)
	    {
	       int n, i, datasize;

	       n = chipno[chipidx] / shm->n_amps_per_controller;
	       i = chipno[chipidx] % shm->n_amps_per_controller;
	       datasize = (( shm->pixels_per_amp[n]*sizeof(short) + 2880 - 1 ) / 2880) * 2880; /* %%% Get 2880 from libfh */

	       /*
		* Write extension header unit.
		* or the normal header unit if 1 amplifier.
		*/
	       if (write_header_unit(save_fd_main, n_chips, chipno[chipidx], (n_chips!=1))==FAIL)
	       {
		     perror("\rerror: (save) write_header_unit failed");
		     _exit(EXIT_FAILURE);
	       }
	       fits_padding[n][i] = 0; /* write_header_unit leaves us on a 2880 boundary */
	       save_offset[n][i] = lseek(save_fd_main, 0, SEEK_CUR);
	       save_fd[n][i] = -1;
	       /*
		* Skip to start of next extension.
		*/
	       if ((save_size = lseek(save_fd_main, datasize-1, SEEK_CUR))==-1)
	       {
		  perror("\rerror: (save) lseek for main file failed");
		  _exit(EXIT_FAILURE);
	       }		  
	       if (cli_write(save_fd_main, "\0", 1, TOMSEC_FILE)==FAIL)
	       {
		  perror("\rerror: (save) can't write last byte of extension");
		  _exit(EXIT_FAILURE);
	       }
	       save_size++;
	    }
	 }
	 else
	 {
	    /*
	     * Write separate FITS files
	     */

	    close(save_fd_main); /* Nothing else but the PHU goes in there. */

	    for (n=0; n<n_controllers; n++)
	    {
	       for (i=0; i<shm->n_amps_per_controller; i++)
	       {
		  int chipno = n*shm->n_amps_per_controller+i;

		  sprintf(filename0, "%.200s/%.200s%02d.fits~",
			  filebase(), filebase(), chipno);
		  if ((save_fd[n][i] = open(filename0, O_RDWR|O_CREAT|O_TRUNC, 0644))==-1)
		  {
		     perror("\rerror: (save) can't open file");
		     /*
		      * Lazy hack: Other files get closed on exit.
		      */
		     _exit(EXIT_FAILURE);
		  }
		  if (write_header_unit(save_fd[n][i], n_chips, chipno, 0)==FAIL)
		  {
		     perror("\rerror: (save) write_header_unit failed");
		     _exit(EXIT_FAILURE);
		  }
		  fits_padding[n][i] = 0; /* write_header_unit leaves us on a 2880 boundary */
	       }
	    }
	 }

	 /*
	  * Simplest to wait until this point before allowing abort cleanup-ups.
	  */
	 cli_signal_unblock(SIGTERM);

	 do
	 {
	    int delay_time;

	    delay_time = 250000;
	    done = 1;
	    for (n = 0; n<n_controllers; n++)
	    {
	       unsigned int available;

	       available =
		  shm->pixels_avail[n]/shm->n_amps_per_controller - pixels_written[n];

	       if (available>DATA_WRITE_MAX) available = DATA_WRITE_MAX;
	       if (available>=DATA_WRITE_MIN ||
		   (available>0 && shm->pixels_avail[n]==shm->pixels_expected[n]))
	       {
		  delay_time = 0;
		  for (i = 0; i<shm->n_amps_per_controller; i++)
		  {
		     int fd;

		     if (save_multi_ext==TRUE)
		     {
			if (lseek(save_fd_main, save_offset[n][i], SEEK_SET)==-1)
			{
			   perror("\rerror: (save) lseek failed");
			   close(save_fd_main);
			   _exit(EXIT_FAILURE);
			}
			save_offset[n][i] += available * sizeof(short);
			fd = save_fd_main;
		     }
		     else
		     {
			fd = save_fd[n][i];
		     }
		     if (cli_write(fd, shm->buf_base + shm->buf_ofs[n][i] + pixels_written[n],
				   available*sizeof(short), TOMSEC_FILE)==FAIL)
		     {
			perror("\rerror: (save) cli_write failed");
			_exit(EXIT_FAILURE);
		     }
		     fits_padding[n][i] += available * sizeof(short);
		  }
		  pixels_written[n] += available;
	       }

	       if (pixels_written[n]<shm->pixels_per_amp[n]) done = 0;
	    }
	    if (shm->pixels_per_amp[0] &&
		(n_controllers==1 || shm->pixels_per_amp[1]))
	    {
	       static time_t ts = 0;
	       time_t tn = time(0);

	       if (ts != tn)
	       {
		  ts = tn;
		  sprintf(junkbuf, "\rinfo3: Saving %s: %d%%\r",
			  save_multi_ext?(n_chips==1?"Basic FITS":"MEF file"):"separate FITS files",
			  n_controllers==1?
			  pixels_written[0]/((shm->pixels_per_amp[0]+99)/100):
			  (pixels_written[0]+pixels_written[1])/
			  ((shm->pixels_per_amp[0]+shm->pixels_per_amp[1]+99)/100));
		  fputs(junkbuf, stderr);
	       }
	    }
	    if (delay_time) xusleep(delay_time);
	 } while (!done);

	 /*
	  * Zero pad (to a 2880-byte boundary) and close the files
	  */
	 if (save_multi_ext == FALSE)
	 {
	    for (n=0; n<n_controllers; n++)
	    {
	       for (i=0; i<shm->n_amps_per_controller; i++)
	       {
		  int fill = fits_padding[n][i] % 2880;
		  int chipno = n*shm->n_amps_per_controller+i;
		  
		  if (fill)
		  {
		     if (cli_write(save_fd[n][i],
				   zerobuf, 2880 - fill, TOMSEC_FILE)==FAIL)
		     {
			perror("\rwarning: (save) failed to pad FITS file; data may be bad");
		     }
		  }
		  close(save_fd[n][i]);

		  sprintf(filename0, "%.200s/%.200s%02d.fits~",
			  filebase(), filebase(), chipno);
		  sprintf(filename1, "%.200s/%.200s%02d.fits",
			  filebase(), filebase(), chipno);
		  if (rename(filename0, filename1)<0)
		  {
		     perror("\rerror: (save) rename failed");
		     _exit(EXIT_FAILURE);
		  }
	       }
	    }
	    sprintf(junkbuf, "\rinfo3:\nstatus: Done writing files `%.900s*.fits'\n", filebase());
	    fputs(junkbuf, stderr);
	 }
	 else /* save_multi_ext == TRUE */
	 {
	    if (close(save_fd_main)==-1)
	    {
	       perror("\rwarning: (save) close failed");
	    }
	    sprintf(filename0, "%.200s.fits~", filebase());
	    sprintf(filename1, "%.200s.fits", filebase());
	    if (rename(filename0, filename1)<0)
	       perror("\rrename at end of save failed");
	    else
	    {
	       sprintf(junkbuf, "\rinfo3:\nstatus: Done writing file `%.900s.fits'\n", filebase());
	       fputs(junkbuf, stderr);
	       /* %%% Make file write-protected here?  Separate files too. chmod() */
	    }
	 }

	 /*
	  * Maintain "current" and "previous" links to the latest FITS files.
	  */
	 if (save_multi_ext==FALSE)
	 {
	    sprintf(filename0, "%.500s/previous", cli_gethome());
	    sprintf(filename1, "%.500s/current", cli_gethome());
	    unlink(filename0);
	    rename(filename1, filename0);
	    sprintf(filename0, "%.500s/%.200s", cwd, filebase());
	    if (symlink(filename0, filename1)<0)
	       perror("\rlogonly: current fits subdirectory link failed");
	 }
	 else
	 {
	    sprintf(filename0, "%.500s/previous.fits", cli_gethome());
	    sprintf(filename1, "%.500s/current.fits", cli_gethome());
	    unlink(filename0);
	    rename(filename1, filename0);
	    sprintf(filename0, "%.500s/%.200s.fits", cwd, filebase());
	    if (link(filename0, filename1)<0 &&
		symlink(filename0, filename1)<0)
	       perror("\rlogonly: current.fits link failed");
	 }

	 /*
	  * FITS file post-processing:
	  * -------------------------
	  *
	  * For now, archive system hard links are still part of detcom.
	  * Once we are confident that the detcom.endgo script is working
	  * properly, it could be run as part of that script.  Until then,
	  * there is a "detcom.endgo-pre" which does any pre-processing
	  * to the FITS file which might need to be done before the archive
	  * system sees the file, then the archive links are made, and then
	  * the "detcom.endgo" happens.
	  */
	 /* %%% "detcom.endgo-pre" should go here! */

	 /*
	  * Make links for the archive system.
	  * %%% This is a hack, because we can't make a hard link directly
	  * from an automounted home directory to the shadow area unless
	  * the shadow area is within the same automount, in which case
	  * the user would be able to tamper with the shadow.
	  */
	 if (!file_from_odometer())
	 {
	    fprintf(stderr, "\rlogonly: Filename not from odometer; frame not archived.\n");
	 }
	 else
	 {
	    char* cwd_base = strdup(cwd); /* Wants to end up containing /data/hostname */

	    if (!strncmp(cwd_base, "/data/", 6) && strchr(cwd_base + 6, '/') != 0)
	       *(strchr(cwd_base + 6, '/')) = '\0';
	    sprintf(junkbuf, "\rlogonly: Making archive links for `%s' in `%s/shadow'\n",
		    filebase(), cwd_base);
	    fputs(junkbuf, stderr);

	    if (save_multi_ext==FALSE)
	    {
	       sprintf(filename0, "%s/shadow/.%.200s", cwd_base, filebase());
	       umask(0);
	       if (mkdir(filename0, 0777)!=0 && errno!=EEXIST)
	       {
		  perror("\rerror: (save) failed to create archive subdirectory");
	       }
	       if (chmod(filename0, 0777)!=0)
		  perror("\rerror: (save) failed to chmod 0777 archive subdirectory");
	       umask(022);
	       for (n=0; n<n_controllers; n++)
	       {
		  for (i=0; i<shm->n_amps_per_controller; i++)
		  {
		     int chipno = n*shm->n_amps_per_controller+i;
		     
		     sprintf(filename0, "%.200s/%.200s%02d.fits",
			     filebase(), filebase(), chipno);
		     sprintf(filename1, "%s/shadow/.%.200s/%.200s%02d.fits",
			     cwd_base, filebase(), filebase(), chipno);
		     if (link(filename0, filename1)<0)
		     {
			perror("\rerror: archive link failed");
			/* But exit success so filename not reused! */
		     }
		  }
	       }
	       sprintf(filename0, "%.200s/%.200s.phu",
		       filebase(), filebase());
	       sprintf(filename1, "%s/shadow/.%.200s/%.200s.phu",
		       cwd_base, filebase(), filebase());
	       if (link(filename0, filename1)<0)
	       {
		  perror("\rerror: archive link failed");
		  /* But exit success so filename not reused! */
	       }
		  
	       sprintf(filename0, "%s/shadow/.%.200s", cwd_base, filebase());
	       sprintf(filename1, "%s/shadow/%.200s", cwd_base, filebase());
	       if (rename(filename0, filename1)!=0)
	       {
		  perror("\rerror: archive rename failed");
	       }
	    }
	    else
	    {
	       sprintf(filename0, "%.200s.fits", filebase());
	       sprintf(filename1, "%s/shadow/.%.200s.fits", cwd_base, filebase());
	       if (link(filename0, filename1)<0)
	       {
		  perror("\rerror: archive link failed");
		  /* But exit success so filename not reused! */
	       }
	    }

            /*
             * Let endgo run in the shadow directory, so post-processing
             * and archiving is not sensitive to very early deletes by
             * the user.
             */
            sprintf(filename1, "%s/shadow", cwd_base);
            if (chdir(filename1) < 0)
            {
               perror("\rerror: archive chdir failed");
            }

	    fprintf(stderr, "\rlogonly: Archiving finished.\n");
	 }
	 if (save_endgo(/*preview=*/FALSE) != PASS)
	    fprintf(stderr, "\rerror: save_endgo failed\n");

	 _exit(EXIT_SUCCESS);
      }
   }
   data_state(DATA_SAVING);
   com_nheader("unstick");
   return PASS;
}

PASSFAIL
com_mef(const char* args)
{
   static char* arg1 = 0;
   shm_t* shm = im_buf_get(FALSE);
   BOOLEAN new_mef;

   if (shm == 0) return FAIL;

   if (!args || !*args)
   {
      return cli_help("mef");
   }

   if (arg1) free(arg1);
   arg1 = cli_arg1(args);
   new_mef = cli_bool(arg1);

   /* %%% What if shm == NULL? */
   if (shm && shm->n_controllers * shm->n_amps_per_controller == 1)
      new_mef = TRUE; /* This really means 1 basic FITS here. */

   if (new_mef)
   {
      printf("status: Saving a single FITS file for each exposure.\n");
   }
   else
   {
      printf("status: Saving separate FITS files for each amplifier.\n");
   }
   if (save_multi_ext != new_mef)
   {
      save_multi_ext = new_mef;
      statusbar_update();
      parfile_update("mef mode changed");
   }
   return PASS;
}

const char*
mef_command(void)
{
   if (save_multi_ext) return "on";
   else return "off";
}


/* End of det_data.c */
