/*                                             -*- c-file-style: "Ellemtel" -*-

`detcom.c' - Global data, command tables, main entry point.

This file is part of version 3 of the CFHT Detector Host Command Interface.
Read the `License' file for terms of use and distribution.
Copyright 1998-2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>     /* setpriority() */
#include <sys/vfs.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <signal.h>
#include <fcntl.h>
#include <time.h>

#ifdef CAMERA_CFH12K
#include <cfht/cfht.h>
#endif
#include "cli/cli.h"

#include "detcom.h"

#ifdef HPUX
extern int setpriority(int which, int who, int priority);
#endif

/*
 * Global settings
 */

int n_controllers = 0;
BOOLEAN done = FALSE;	/* flag to set when program should exit */
BOOLEAN auto_save = TRUE;
BOOLEAN auto_clean = TRUE;
BOOLEAN auto_voltage = FALSE;
LeachDev leach[MAX_CONTROLLERS];

static void
final_cleanup(void)
{
   int n;
   
   fprintf(stderr, "logonly: DetCom is exiting.\n");

   if (camera_state(CAMERA_STATE_QUERY)==CAMERA_SWEEPING)
      camera_state(CAMERA_CLEANED); /* === Stop idle mode */

   /*
    * Save current state in a file to be reread at next startup.
    */
   rcfile_update();

   /*
    * Make sure there is no data writer child process hanging around.
    */
   com_abortdata("");

   /*
    * Close the camera input devices.
    */
   for (n=0; n<n_controllers; n++)
   {
      if (leach[n].dev!=0)
      {
	 if (lio_close(leach[n].dev)!=LIO_OK) perror("lio_close");
	 leach[n].dev = 0;
      }
   }
   fprintf(stderr, "statusbar:\n");
   fprintf(stderr, "logonly: DetCom cleanup complete.\n");
}

static void
signal_cleanup(void)
{
   fprintf(stderr, "logonly: DetCom exiting on signal\n");
   exit(EXIT_SUCCESS);
}

/* Command Set
 * -----------
 * Each of these must take a single const char* argument (even if they
 * don't use it for anything) and return PASS or FAIL depending on
 * whether the command was successful or not.  If needed, libcli
 * provides further functions for breaking up the single string
 * into argc/argv-style arguments and removing quotes from string
 * constants.  If you find that you are implementing something in your
 * code that would be generally useful to other programs, let me know
 * and we will include it in libcli. -si
 */


int
main(int argc, const char* argv[])
{
   int on_tty = 0;
   int max_controllers = MAX_CONTROLLERS;
   BOOLEAN download_from_rc = TRUE; /* If false, download from argv instead */
   char* debug;

   if ((debug = getenv("CFHT_DEBUG")))
      cli_debug = cli_bool(debug);

   seteuid(getuid());
   umask(022);

   /*
    * Always start in "safe" mode (nothing automatic, and no engineering
    * commands allowed either.)  This .detcomrc will contain the last
    * mode state and restore engineering or observing mode.
    */
   if (com_mode("safe")==FAIL)
     return EXIT_FAILURE;

#ifdef CAMERA_CFH12K
   cfht_logv(CFHT_MAIN, CFHT_LOG_ID, "detcom");
#endif

   if (isatty(STDOUT_FILENO)) on_tty = 1;

#ifdef CAMERA_CFH12K
#define DETCOM_NAME "DetCom-CFH12K"
#else
#ifdef CAMERA_MDM8K
#define DETCOM_NAME "DetCom-MDM8K"
#else
#define DETCOM_NAME "DetCom"
#endif
#endif

   printf("Detector Command Interface %s %.2f %s (compiled %s)\n"
	  "Copyright 1998-2002 Canada-France-Hawaii Telescope, <daprog@cfht.hawaii.edu>\n",
	  DETCOM_NAME, VERSION, SOURCEDATE, BUILDDATE);

   if (seteuid(0)<0)
   {
      printf("warning: Readouts may fail because DetCom is not setuid root.\n");
      /* === Should really exit here unless engineering mode! */
   }
   else
   {
      printf("logonly: DetCom setpriority: %d\n", DETCOM_PRIORITY);
      if (setpriority(PRIO_PROCESS, 0, DETCOM_PRIORITY)==-1)
	 perror("setpriority failed");
   }

   seteuid(getuid());

   if (geteuid()==0)
   {
      fprintf(stderr, "error: DetCom should not be run as root.\n");
   }

   /*
    * %%% Begin camera_open() code back-ported from DetCom 4.x
    */

   /*
    * argc and argv[] are the arguments passed at the invocation of
    * "detcom" itself, from the command line.  Three cases are currently
    * handled:
    *
    * 1) The preferred syntax where argc-1 is equal to the number of
    *    controllers, and each argument contains a '=' sign.  The string
    *    before the argument is the controller port device and the
    *    string after the '=' is a list of DSP code files to be booted.
    *    argv[1..n_controllers+1] = "[/dev/]<portname>=<dspcode>[,<dspcode>]"
    *
    *    NOTE: This explicitly overrides any "boot" commands in DetCom's
    *    startup file and uses the filenames on the command line only.
    *    Manually entered boot commands can still be used to switch to
    *    a different DSP code.  Only those in the .detcomrc are ignored.
    *
    * 2) The second option is to give no arguments on the command line at
    *    at all.  In this case, camera_open() searches for as many devices
    *    as it can find, starting with "/dev/roc0" (which could be a
    *    symlink to the real device.)  No DSP code is assigned or booted
    *    until a "boot" command is issued (or found in the .detcomrc from
    *    the last time DetCom was run.)
    *
    * 3) The third option is similar to (1), but remains for backward
    *    compatibility with old DetCom versions, and applies to SDSU only.
    *    No portname or '=' is used, and only the DSP code filename(s)
    *    are given, one per argument.
    */

   /*
    * Determine the correct value for max_controllers.
    */
   if (argc <= 1)
   {
      /*
       * This is the second case... nothing was given on the command line.
       */
      max_controllers = MAX_CONTROLLERS;
   }
#ifdef OLD_DETCOM
   /* === Remove this OLD_DETCOM stuff after fixing anything that
    * depended on this syntax.
    */
   else if (strchr(argv[1], '=')==0)
   {
      /*
       * If command line arguments without '=' are given, assume the following:
       *
       * detcom [ltbM.lod lubM.lod [ltbS.lod lubS.lod]] ...
       */
      switch (argc)
      {
	 case 1: break;
	 case 2:
	 case 3: max_controllers = 1; download_from_rc = FALSE; break;
	 case 5: max_controllers = 2; download_from_rc = FALSE; break;
	 default: printf("warning: odd number of command line arguments.\n");
      }
   }
#endif
   else
   {
      /*
       * This is the "/dev/roc=bootcode[,bootcode2]" syntax.
       */
      max_controllers = argc - 1;
      download_from_rc = FALSE;
   }

   if (max_controllers > MAX_CONTROLLERS)
   {
      printf("DetCom is compiled to handle no more than %d controller%s.\n",
	     MAX_CONTROLLERS, MAX_CONTROLLERS==1?"":"s");
      printf("error: Recompile DetCom with MAX_CONTROLLERS=%d and try again.\n",
	     max_controllers);
      exit(EXIT_FAILURE);
   }

   /*
    * Open a connection to the driver(s).
    */

   printf("progress: Searching for controllers\n");

   for (n_controllers=0; n_controllers<max_controllers; n_controllers++)
   {
      char dev_file[40] = "";
      char* lod_file;
      int dev_name_len;

      /*
       * Try to open the device.
       */
      lod_file = 0;
      if (argc > n_controllers + 1 && (lod_file = strchr(argv[n_controllers+1], '='))!=0)
      {
	 dev_name_len = lod_file++ - argv[n_controllers+1];
	 if (*argv[n_controllers+1] == '/')
	 {
	    if (dev_name_len >= sizeof(dev_file)) /* Don't overflow */
	       dev_name_len = sizeof(dev_file) - 1; /* warn? */
	    sprintf(dev_file, "%.*s", dev_name_len, argv[n_controllers+1]);
	 }
	 else /* Add a "/dev/" if it didn't start with a "/". */
	 {
	    if (dev_name_len + 5 >= sizeof(dev_file)) /* Don't overflow */
	       dev_name_len = sizeof(dev_file) - 6; /* warn? */
	    sprintf(dev_file, "/dev/%.*s", dev_name_len, argv[n_controllers+1]);
	 }
      }
      else
	 *dev_file = '\0';

      memset(&leach[n_controllers], 0, sizeof(leach[n_controllers]));
      leach[n_controllers].name_contr = n_controllers?"2nd":"1st";
      leach[n_controllers].name_board[LUB] = n_controllers?"2nd utility board":"1st utility board";
      leach[n_controllers].name_board[LTB] = n_controllers?"2nd timing board":"1st timing board";
      leach[n_controllers].name_cmdarg[LUB]= n_controllers?"lubS":"lubM";
      leach[n_controllers].name_cmdarg[LTB]= n_controllers?"ltbS":"ltbM";

      printf("logonly: Trying to open %s\n", *dev_file?dev_file:"(default device)");
      if ((leach[n_controllers].dev = lio_open(*dev_file?dev_file:0))==0)
	 break;

      /*
       * Save filenames of DSP code files, if present.  These will be loaded
       * later when camera_boot("") is called.
       */

      /* === Remove this OLD_DETCOM stuff after fixing anything that
       * depended on this syntax.
       */
      if (argc > 1 && strchr(argv[1], '=')==0)
      {
#ifdef OLD_DETCOM
	 switch (n_controllers)
	 {
	    case 0:
	       if (argc<2) break;
	       leach[0].dspcode_filename[LTB] = strdup(argv[1]);
	       if (argc<3) break;
	       leach[0].dspcode_filename[LUB] = strdup(argv[2]);
	       break;
	    case 1:
	       if (argc<4) break;
	       leach[1].dspcode_filename[LTB] = strdup(argv[3]);
	       if (argc<5) break;
	       leach[1].dspcode_filename[LUB] = strdup(argv[4]);
	       break;
	 }
#else
	 fprintf(stderr, "error: This version of DetCom requires the new startup syntax, `detcom /dev/lotuspci0=ltbmaster.lod[,lubmaster.lod] [/dev/lotuspci1=ltbslave.lod[,lubslave.lod]]'\n");
	 exit(EXIT_FAILURE);
#endif
      }
      else if (lod_file)
      {
	 char* p;

	 leach[n_controllers].dspcode_filename[LTB] = strdup(lod_file);
	 if ((p = strchr(leach[n_controllers].dspcode_filename[LTB], ',')) != 0)
	 {
	    *p++ = '\0';
	    leach[n_controllers].dspcode_filename[LUB] = strdup(p);
	 }
      }
   }

   if (n_controllers == 0)
      printf("error: No controller interfaces found. (Maybe another detcom has not completely exited, or you need \"modprobe lotuspci\"?)\n");
   else if (n_controllers < max_controllers)
      printf("logonly: %d of %d controller interfaces opened.\n",
	     n_controllers, max_controllers);

   atexit(final_cleanup); /* This closes the devices on exit */

   cli_signal(SIGTERM, signal_cleanup);
   cli_signal(SIGINT, CLI_SIG_TRAP);
   cli_signal(SIGTSTP, CLI_SIG_TRAP);
   cli_signal(SIGQUIT, CLI_SIG_IGN);
   cli_signal(SIGCHLD, CLI_SIG_TRAP);

   printf("progress: Testing camera communications and fiber link\n");

   com_tdl(""); /* NOTE: com_tdl will take care of auto-detecting
		 * which generation controller is on the other end
		 * of the fiber, in the case of SDSU controllers
		 * and the PCI card.
		 */

   rcfile_reset(); /* See rcfile_getline() below. */

   while (!done)
   {
      static char* rcline = "";
      static char* line = 0;
      struct sigaction oldsigalrm;

      if (!on_tty)
      {
	 statusbar_update();
	 sigaction(SIGALRM, 0, &oldsigalrm);
	 cli_signal(SIGALRM, CLI_SIG_TRAP);	/* LPS */
	 cli_signal_unblock(SIGALRM);
	 alarm(5);				/* LPS */
      }

      if (line!=0) free(line);	/* free memory used by last command */
      line = 0;

      if (n_controllers)
	 rcline = line = rcfile_getline(download_from_rc);	/* If still reading the rcfile */

      if (!line)
	 line = cli_getline();		/* get the next command */

      if (!on_tty)
      {
	 alarm(0);					/* LPS */
	 cli_signal_block(SIGALRM); /* %%% Why? Needed by EDT driver? */
	 sigaction(SIGALRM, &oldsigalrm, 0);
      }

      /*
       * Clear old signals before executing a new command
       */
      cli_signal_event(SIGINT);
      /* === TODO: see if a FITS writer is running.  If it is, tell the
       * user they can kill it by bringing it into the "foreground"
       * with `waitdata' and then hit ^C again.
       */
      cli_signal_event(SIGTSTP);

      if (cli_signal_event(SIGCHLD))
      {
	 printf("logonly: A child thread exited.\n");
	 /*
	  * If the exiting child was the "expose" thread, it may be
	  * necessary to start the readout right away.  Check.
	  */
	 com_expose("poll");
      }

      switch (camera_state(CAMERA_STATE_QUERY))
      {
	 case CAMERA_READING:
	 case CAMERA_EXPOSING:
	 case CAMERA_CLEANING:
	 {
	    if (com_waitdata("-poll"));
	    break; /* Do not disturb controller while it is reading out */
	 }
	 default:
	 {
	    /*
	     * Flush out any unsolicited messages from the controller.
	     */
	    if (leach_poll(/*verbose=*/1)!=LEACH_POLL_NODATA)
	    {
	       /* com_download(""); Bad idea. Controller may be powering up. */
	    }
	 }
      }

      /*
       * If there was a real command typed, execute the right function here
       */
      if (line)
      {
	 /*
	  * And unblock them at all times except when we are sure it
	  * is ok to be interrupted.
	  */
	 cli_signal_block(SIGINT);
	 cli_signal_block(SIGTSTP);

	 if (camera_state(CAMERA_STATE_QUERY)==CAMERA_SWEEPING)
	    camera_state(CAMERA_CLEANED); /* === Stop idle mode before ANY command for now */

	 cli_execute(line);	/* lookup and run the command */

	 cli_signal_unblock(SIGINT);
	 cli_signal_unblock(SIGTSTP);
	 cli_signal(SIGQUIT, CLI_SIG_IGN);
	 cli_signal_unblock(SIGQUIT);
      }
      else
      {
	 static time_t sleepy_time = 0;
	 BOOLEAN do_not_disturb = FALSE;

	 switch (camera_state(CAMERA_STATE_QUERY))
	 {
	    case CAMERA_READING:
	    case CAMERA_EXPOSING:
	    case CAMERA_CLEANING:
	    {
	       /* Only CAMERA_READING currently happens. */
	       do_not_disturb = TRUE; /* Do not disturb controller! */
	       break;
	    }
	    case CAMERA_READY:
	    case CAMERA_CLEANED:
	    if (detcom_mode == DETCOM_OBSERVING)
	    {
	       if (auto_clean==TRUE &&
		   com_power("test")==PASS)
	       {
		  /*
		   * If no command has been issued for several seconds,
		   * enter SWEEPING mode.  The state machine actually
		   * takes care of turning on idling mode. === This
		   * is not a very consistent implementation: almost everything
		   * else is activated from outside of the state machine,
		   * and the state machine is told to transition to a new
		   * state just as a "sanity check".
		   */
		  /* camera_state(CAMERA_SWEEPING); %%% */
		  /*
		   * Calculate the time when the shutter should go to
		   * sleep-mode. (Hardcoded to 30 minutes for now.)
		   */
		  sleepy_time = time(0) + 30 * 60; /* %%% Make configurable. */
	       }
	       break;
	    }
	    case CAMERA_SWEEPING:
	    if (detcom_mode == DETCOM_OBSERVING)
	    {
	       if (auto_clean==TRUE &&
		   sleepy_time && time(0) >= sleepy_time)
	       {
		  printf("logonly: Calling shutter sleep.\n");
		  com_shutter("sleep"); /* This only does something with 12K */
		  sleepy_time = 0;
	       }
	       break;
	    }
	    default: ; /* Do nothing in other states. */
	 }

	 if (com_waitdata("-poll"));

	 if (do_not_disturb == FALSE)
	 {
	    /*
	     * Low priority things to check every few seconds, but only
	     * if no commands are coming in!
	     */
	    com_filter("test"); /* Check if status of the filter wheel changed */
	    com_preexp("reset"); /* Force full pre-exp checks on next go. */
	    if (detcom_mode == DETCOM_OBSERVING)
	    {
	       com_temp("-q"); /* Polling rate controlled from within com_temp() */
	       if (auto_voltage==TRUE) com_power("-q");
	    }
	 }
      }
      /*
       * Disk space check-hack.  Clean this up.  %%%
       */
      {
	 static int warned_df = 0;
	 struct statfs sfb;
	 long percent_filled;

	 if (statfs(".", &sfb) < 0)
	    perror("statfs(\".\")");
	 if (sfb.f_blocks > 100)
	    percent_filled = (sfb.f_blocks - sfb.f_bavail) / (sfb.f_blocks/100);
	 else
	    percent_filled = 100;

	 printf("info1: %s (%ld%% filled)\n",
		cli_getcwd(), percent_filled);
	 if (percent_filled > 95)
	 {
	    if (!warned_df++)
	    {
	       printf("warning: %s is %ld%% filled!\n",
		      cli_getcwd(), percent_filled);
	    }
	 }
	 else warned_df = 0;
      }
   }

   exit(EXIT_SUCCESS);
}
