
#define _GNU_SOURCE /* For IPC_INFO on Linux, which improves shm handling */
#include <sys/mman.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include "detcom.h"

static shm_t* _shm = 0;
static size_t _shm_size = 0;

PASSFAIL
mlock_shm(shm_t* shm)
{
   PASSFAIL rtn = PASS;

   if (!_shm || !_shm_size)
   {
      printf("error: mlock_shm: shm not configured.\n");
      return FAIL;
   }
   
   seteuid(0); /* Get superuser priveleges for mlock */

#ifndef HPUX
   if (mlock((void*)_shm, _shm_size)<0)
   {
      perror("mlock(shm) failed");
      rtn = FAIL;
   }
   if ((mlockall(MCL_CURRENT) < 0) |
       (mlockall(MCL_FUTURE) < 0))
   {
      char hostname[80];

      gethostname(hostname, 80);
      hostname[79]='\0';
      perror("mlockall failed");
      fprintf(stderr,
	      "warning: couldn't lock down memory, "
	      "please free memory on CCD host (%s) and retry\n",
	      hostname);
      if (errno==EAGAIN) { printf("warning:  (Insufficient memory resources)\n"); }
      rtn = FAIL;
   }
#endif
   seteuid(getuid());
   return rtn;
}


shm_t*
im_buf_cfg(unsigned int n_controllers,
	   unsigned int n_amps_per_controller,
	   unsigned int rows_per_amp,
	   unsigned int cols_per_amp)
{
   int controller, amp;

   if (n_controllers > MAX_CONTROLLERS)
   {
      printf("error: im_buf_cfg internal error: n_controllers (%d) > %d\n",
	     n_controllers, MAX_CONTROLLERS);
      return 0;
   }
   if (n_amps_per_controller > MAX_AMPS_PER_CONTROLLER)
   {
      printf("error: im_buf_cfg internal error: n_amps_per_controller (%d) > %d\n",
	     n_amps_per_controller, MAX_AMPS_PER_CONTROLLER);
      return 0;
   }

   if (_shm)
   {
      /*
       * Check if the configuration is still the same.
       */
      if (_shm->n_controllers == n_controllers &&
	  _shm->n_amps_per_controller == n_amps_per_controller &&
	  _shm->rows_per_amp == rows_per_amp &&
	  _shm->cols_per_amp == cols_per_amp)
      {
	 return _shm;
      }
      im_buf_get(/*remove=*/TRUE);
   }
   
#ifndef DONT_LOCK_SHM
   seteuid(0);
   if (mlockall(MCL_CURRENT)<0) perror("mlockall(MCL_CURRENT)"); /* %%% */
   seteuid(getuid());
#endif
   
   /*
    * Given the number of amplifiers, controllers, rows, and columns,
    * calculate the total size of the shared memory region.
    */
   _shm_size = sizeof(shm_t) + sizeof(short) * 
      (rows_per_amp * cols_per_amp * n_amps_per_controller * n_controllers);

   if (!im_buf_get(FALSE))
   {
      _shm_size = 0;
      printf("error: Shared memory allocation failed.\n");
      return 0;
   }

   CLI_DEBUG(printf("debug: Allocated %d bytes\n", _shm_size)); /* %%% */

#ifndef DONT_LOCK_SHM
   /*
    * Clear the memory to be sure it is swapped in.
    */
   CLI_DEBUG(printf("progress: Clearing shared memory buffer.\n"));
   memset(_shm, 0, _shm_size);
   CLI_DEBUG(printf("progress: Done clearing.\n"));

   /* %%% machines with low memory and the SDSU SBus driver can't mlock */
   /*
    * Lock the shared memory so it doesn't get swapped out.
    */
   if (mlock_shm(_shm)==FAIL)
      printf("warning: Readouts may fail because shared memory isn't locked down.\n");
#endif

   /*
    * A new shm structure has been allocated.  Initialize it
    * with the relevant parameters.
    */
   _shm->n_controllers = n_controllers;
   _shm->n_amps_per_controller = n_amps_per_controller;
   _shm->rows_per_amp = rows_per_amp;
   _shm->cols_per_amp = cols_per_amp;
   for (controller = 0; controller < n_controllers; controller++)
   {
      for (amp = 0; amp < n_amps_per_controller; amp++)
      {
	 _shm->buf_ofs[controller][amp] = rows_per_amp * cols_per_amp
	    * (controller * n_amps_per_controller + amp);
      }
   }
   return _shm;
}

shm_t*
im_buf_get(BOOLEAN remove)
{
   int shmid; /* Shared memory ID (SysV) or mmapped file descriptor */

   if (_shm != 0 && remove==FALSE)
      return _shm;

   if (_shm == 0 && remove==TRUE)
      return 0;

   seteuid(0); /* Get superuser priveleges for shared memory */

   if (_shm_size == 0)
   {
      printf("error: Shared memory size not configured.\n");
      seteuid(getuid());
      return 0;
   }
   
#ifndef USE_MMAP
   CLI_DEBUG(printf("debug: Using %d bytes of System V shared memory\n",
		    _shm_size));

/*
 * Does this work for other Unixes?  I took it from ipcs.c.
 * It seems Linux-specific.
 */
#ifdef LINUX
#ifndef SHM_STAT
#define SHM_STAT        13
#define SHM_INFO        14
   struct shm_info {
	 int   used_ids;
	 ulong shm_tot; /* total allocated shm */
	 ulong shm_rss; /* total resident shm */
	 ulong shm_swp; /* total swapped shm */
	 ulong swap_attempts;
	 ulong swap_successes;
   };
#endif
   /*
    * Search for a previous DetCom's shared memory segment.  We know the
    * key is SHM_KEY_DETCOM, but unfortunately shmget() will not return
    * the id that goes with this key unless we get the size right.  So,
    * search through all id's instead.  (Arrgh!)
    */
   
   {
      int id, maxid; /* Maximum ID (SysV) for searching for old key */
      struct shm_info shm_info;
      struct shminfo shminfo;
      struct shmid_ds shmbuf;
      
      maxid = shmctl(0, SHM_INFO, (struct shmid_ds *)&shm_info);
      if (maxid < 0 ||
	  (shmctl(0, IPC_INFO, (struct shmid_ds *) &shminfo)) < 0)
      {
	 printf("error: Kernel not configured for shared memory\n");
	 exit(EXIT_FAILURE);
      }
      if (maxid <= 1) maxid = shm_info.used_ids;
      
      if (shminfo.shmmax < _shm_size)
      {
	 FILE* shmmax = fopen("/proc/sys/kernel/shmmax", "w");

	 if (shmmax)
	 {
	    fprintf(shmmax, "%d\n", _shm_size + 1000000);
	    if (fclose(shmmax) != 0) shmmax = 0;
	 }

	 if (shmmax == 0)
	 {
	    fprintf(stderr,
		    "warning: Your system seems to be configured to allow "
		    "only %lu bytes of shared memory (we need %lu bytes).\n",
		    (unsigned long)shminfo.shmmax, (unsigned long)_shm_size);
	    fprintf(stderr,
		    "On Linux 2.2.x, fix this by placing the following lines in /etc/rc.d/rc.local (or similar)\n"
		    "  echo %lu > /proc/sys/kernel/shmmax\n"
		    "  echo %lu > /proc/sys/kernel/shmall\n"
		    "(Choose larger values if you anticipate reading larger CCDs.)\n",
		    (unsigned long)_shm_size + 1000000, (unsigned long)_shm_size + 1000000);
	 }
      }

      for (id = 0; id <= maxid; id++)
      {
	 if ((shmid = shmctl(id, SHM_STAT, &shmbuf)) >= 0 &&
#if defined (__GNU_LIBRARY__) && __GNU_LIBRARY__ > 1
	     (shmbuf.shm_perm.__key == SHM_KEY_DETCOM)
#else
	     (shmbuf.shm_perm.key == SHM_KEY_DETCOM)
#endif
	    )
	 {
	    if (shmbuf.shm_nattch!=0)
	    {
	       printf("error: Someone else is already using the shared memory area.  "
		      "Perhaps another DetCom is still running?\n");
	       exit(EXIT_FAILURE);
	    }
	    if (shmbuf.shm_segsz != _shm_size)
	    {
	       printf("warning: Shared memory size has changed.  Re-allocating it.\n");
	       if (shmctl(shmid, IPC_RMID, 0)==-1)
		  perror("shmctl(IPC_RMID) failed");
	    }
	 }
      }
   }
#endif

   /*
    * Check for the shared memory block
    */
   if ((shmid = shmget(SHM_KEY_DETCOM, _shm_size, 00600|IPC_CREAT)) == -1)
   {
      perror("shmget");
      if (errno==EINVAL)
      {
	 fprintf(stderr,
		 "If the shared memory segment size has changed any, "
		 "you may need to remove it manually before continuing. "
		 "Use `ipcs -am' to see if it is there and `ipcrm shm <id>' to "
		 "remove it.  Otherwise, the problem could be that the limit "
		 "on shared memory is set too low...\n\n");
#ifdef SOLARIS
	 fprintf(stderr,
		 "On solaris 2.x, Make sure you have a line like:\n"
		 "  set shmsys:shminfo_shmmax=%d\n"
		 "in the file /etc/system and reboot the machine.\n", _shm_size);
#endif
      }
      exit(EXIT_FAILURE);
   }
   
   if (remove)
   {
      /* %%% shmdt? */
      if (shmctl(shmid, IPC_RMID, 0)==-1)	
	 perror("shmctl");
      seteuid(getuid());
      _shm = 0;
      return 0;
   }
#ifndef SHM_SHARE_MMU
#define SHM_SHARE_MMU 0
#endif
   if ((_shm = (shm_t*)shmat(shmid, 0, SHM_SHARE_MMU))==(void*)-1)
   {
      perror("shmat");
      exit(EXIT_FAILURE);
   }
#else /* USE_MMAP */
   /*
    * Alternative to sysV shmget: use mmap() instead
    */
   
   if (remove)
   {
      CLI_DEBUG(printf("debug: Removing mmapped shared memory.\n"));
      if (munmap((void*)_shm, _shm_size) < 0)
	 perror("munmap(_shm) failed");
      
      if (unlink(SHM_FILENAME)==-1)
	 perror("unlink(\""SHM_FILENAME"\") failed");
      _shm = 0;
      seteuid(getuid());
      return 0;
   }
      
   CLI_DEBUG(printf("debug: Using mmap on file '%s'\n", SHM_FILENAME));
      
   if ((shmid = open(SHM_FILENAME, O_RDWR|O_CREAT, 0644))==-1)
   {
      perror("open(\""SHM_FILENAME"\")");
      exit(EXIT_FAILURE);
   }
      
   if (lseek(shmid, _shm_size, SEEK_SET)==-1)
   {
      perror("lseek");
      exit(EXIT_FAILURE);
   }
   if (write(shmid, "[EOF]", 5)==-1)
   {
      perror("write");
      exit(EXIT_FAILURE);
   }
   if (lseek(shmid, 0, SEEK_SET)==-1)
   {
      perror("lseek");
      exit(EXIT_FAILURE);
   }
   
   _shm = (shm_t*) mmap(/*addr=*/ 0, _shm_size,
			PROT_READ|PROT_WRITE, MAP_SHARED, 
			shmid, /*offset=*/ 0);
   if (_shm == MAP_FAILED)
   {
      _shm = 0;
      perror("mmap");
      exit(EXIT_FAILURE);
   }
#endif

   seteuid(getuid());
   return _shm;
}
