/*                                             -*- c-file-style: "Ellemtel" -*-

`leach_dsp.c' - Fault tolerant DSP parser and downloader.

This file is part of version 3 of the CFHT Detector Host Command Interface.
Read the `License' file for terms of use and distribution.
Copyright 1998-2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "leachcont.h"
#include "detcom.h" /* === Why? */

static char dsp_version_[256] = "unknown";

/*
 * --- ADDRESS USED CACHE ---
 *
 * The address cache stores a bit for each address so
 * the download loop can tell if a .lod file is trying
 * to re-write a memory location on the same pass (in
 * this case, the re-write is done only on the final pass.)
 */
#define address_cache_test(address)  (address_cache[(address)>>5] & (1<<((address)&0x1F)))
#define address_cache_set(address)   address_cache[(address)>>5] |= (1<<((address)&0x1F))
#define address_cache_clear(address) address_cache[(address)>>5] &= ~(1<<((address)&0x1F))
#define address_cache_zero()         memset(address_cache, 0, sizeof(address_cache))

/*
 * --- DSP SYMBOL TABLE ---
 *
 * These routines handle DSP symbol table parsing and
 * lookup.  Entries are sorted and searched with
 * standard qsort and bsearch libc routines.
 *
 * sym_lookup() is the only call accessible outside of this module.
 * The rest are all static, and are used by com_boot() to build
 * the list during parsing of a .lod file.
 */
static int
sym_compare(const void* a, const void* b)
{
   return xstrcasecmp(((LeachSym*)a)->symname, ((LeachSym*)b)->symname);
}

static void
sym_sort(LeachDev* leach, BoardID board)
{
   if (board > MAX_BOARDS)
   { printf("error: Invalid BoardID %d\n", board); return; }

   if (leach->sym_count[board]<=0)
   { printf("error: No symbols to sort\n"); return; }

   qsort(leach->symbols[board], leach->sym_count[board], sizeof(LeachSym), sym_compare);
}

const char*
sym_lookup(LeachDev* leach, BoardID board, const char* name, dsp_sym_t symtype)
{
   LeachSym tmp;
   LeachSym* match;

   if (board > MAX_BOARDS || board < 0)
   { printf("error: Invalid BoardID %d\n", board); return "invalid board ID"; }

   if (leach->sym_count[board]<=0)
   {
      if (leach->name_board[board])
	 printf("logonly: No symbols loaded for %s.\n", leach->name_board[board]);
      return name;
   }

   tmp.symname = (char*)name;
   match = (LeachSym*)bsearch(&tmp, leach->symbols[board],
			      leach->sym_count[board], sizeof(LeachSym),
			      sym_compare);
   if (match)
   {
      dsp_sym_t symtype_match = DSP_SYM_UNKNOWN;

      if (!strncmp(match->symval, "N:", 2)) symtype_match = DSP_SYM_NUMBER;
      else if (!strncmp(match->symval, "X:", 2) ||
	       !strncmp(match->symval, "Y:", 2) ||
	       !strncmp(match->symval, "P:", 2)) symtype_match = DSP_SYM_ADDRESS;

      switch (symtype)
      {
	 case DSP_SYM_NUMBER:
	 {
	    if (symtype_match != DSP_SYM_NUMBER)
	    {
	       printf("warning: Want %s DSP symbol `%s' to be numeric but got %s\n",
		      leach->name_board[board], name,
		      symtype_match==DSP_SYM_ADDRESS?"address":"unknown type");
	       return name;
	    }
	    break;
	 }
	 case DSP_SYM_ADDRESS:
	 {
	    if (symtype_match != DSP_SYM_ADDRESS)
	    {
	       printf("warning: Want %s DSP symbol `%s' to an address but got %s\n",
		      leach->name_board[board], name,
		      symtype_match==DSP_SYM_NUMBER?"numeric value":"unknown type");
	       return name;
	    }
	    break;
	 }
	 case DSP_SYM_UNKNOWN:
	 {
	    printf("logonly: Someone requested DSP_SYM_UNKNOWN... probably meant DSP_SYM_ANY\n");
	    break;
	 }
	 case DSP_SYM_ANY:
	 {
	    if (symtype_match == DSP_SYM_UNKNOWN)
	    {
	       printf("logonly: Symbol `%s' in %s doesn't seem to be an address or number\n",
		      name, leach->name_board[board]);
	    }
	    break;
	 }
      }
      return match->symval;
   }
   else return name;
}

static void
sym_destroy(LeachDev* leach, BoardID board)
{
   int i;

   if (board > MAX_BOARDS)
   { printf("error: Invalid BoardID %d\n", board); return; }

   if (leach->sym_count[board]>0)
   {
      for (i=0; i<leach->sym_count[board]; i++)
      {
	 if (leach->symbols[board][i].symname)
	 {
	    free(leach->symbols[board][i].symname);
	    free(leach->symbols[board][i].symval);
	 }
      }
      free(leach->symbols[board]);
      leach->sym_count[board] = 0;
      leach->symbols[board] = 0;
   }
}

static void
sym_create(LeachDev* leach, BoardID board, int list_size)
{
   if (board > MAX_BOARDS)
   { printf("error: Invalid BoardID %d\n", board); return; }

   sym_destroy(leach, board); /* Destroy last symbol table, if needed */
   leach->symbols[board] = (LeachSym*)malloc(list_size * sizeof(LeachSym));
   memset(leach->symbols[board], 0, list_size*sizeof(LeachSym));
   leach->sym_count[board] = list_size;
}

static void
sym_set(LeachDev* leach, BoardID board, int pos, const char* name, const char* val)
{
   if (board > MAX_BOARDS)
   { printf("error: Invalid BoardID %d\n", board); return; }

   if (pos >= leach->sym_count[board] || pos < 0)
   {
      printf("error: DSP symbol %d is out of bounds (table size %d)\n",
	     pos, leach->sym_count[board]);
      return;
   }

   if (leach->symbols[board][pos].symname)
   {
      free(leach->symbols[board][pos].symname);
      free(leach->symbols[board][pos].symval);
   }

   leach->symbols[board][pos].symname = strdup(name);
   leach->symbols[board][pos].symval = strdup(val);
}

PASSFAIL
com_symbol( const char* args )
{
   static char** cargv = 0;	/* Args given on CLI to boot ... */
   static int cargc = 0;	/* Argument count */
   lb_set_t lbset;
   BoardID board;
   int n;
   PASSFAIL rtn = PASS;

   if (cargv) { if (cargv[0]) free(cargv[0]); free(cargv); }
   cargv = cli_argv(&cargc, args);

   if (cargc > 2 || cargc < 1)
   {
      printf("usage: symbol <board(s)> [<symbol name>]");
      return FAIL;
   }

   lbset = lb_parse(cargv[0]);
   for (n=0; n<n_controllers; n++)
   {
      for (board=0; board<MAX_BOARDS; board++)
      {
	 if (leach[n].name_board[board] && LBISSET(&lbset, n, board))
	 {
	    if (cargc < 2)
	    {
	       int i;

	       for (i=0; i<leach[n].sym_count[board]; i++)
	       {
		  if (i==0 || xstrcasecmp(leach[n].symbols[board][i-1].symname,
					  leach[n].symbols[board][i].symname))
		  {
		     printf("%s\tsymbol %.19s\t= %s\n",
			    leach[n].name_board[board],
			    leach[n].symbols[board][i].symname,
			    leach[n].symbols[board][i].symval);
		  }
	       }
	    }
	    else
	    {
	       const char* sym = sym_lookup(&leach[n], board, cargv[1], DSP_SYM_ANY);

	       if (!strcmp(sym, cargv[1]))
	       {
		  sym = "NOT FOUND";
		  rtn = FAIL;
	       }
	       printf("%s\tsymbol %.19s\t= %s\n",
		      leach[n].name_board[board], /* === */
		      cargv[1], sym);
	    }
	 }
      }
   }
   return rtn;
}

/*
 * Display the current filename that's booted in the info panel.
 */
static void
show_dsp_info(int controller, int board)
{
   /*
    * Only show the first controller's LTB and LUB status.
    */
   if (controller==0 && (board==LTB || board==LUB))
   {
      unsigned long serno_v = sym_to_number(&leach[controller], board, "SERNO_V");

      printf("info%d: %s=%s(#%ld) %s\n", 8+(board==LUB?1:0),
	     leach[controller].name_board[board],
	     leach[controller].dspcode_filename[board], serno_v,
	     leach[controller].dspcode_ok[board]?"OK":"NOT booted");
   }
}

/* --- MAIN DSP BOOT COMMAND --- */

static BOOLEAN force_pon = FALSE; /* Don't test to see if power is already on? */

PASSFAIL
com_boot( const char* args )
{
   static char** cargv = 0;	/* Args given on CLI to boot ... */
   static int cargc = 0;	/* Argument count */
   FILE* lodfile;		/* File pointer to DSP .lod file */
   int pass, line = -1;		/* Pass over file (0 or 1), current line number */
   int lines_total = -1;	/* Total expected, for progress meter */
   int wrmem_worked = 0;	/* If still 0, nothing has been written successfully */
   char buf[1000];		/* Buffer holds one line of .lod file at a time */
   char* p;			/* Pointer into buf[], used by parser */
   int end_found;		/* _END has been read and time to close file */
   char in_symbol_table;	/* _SYMBOL has been found, reading symbol list */
   int in_data_block;		/* _DATA has been found, address shows location */
   BOOLEAN in_comment_block;	/* _COMMENT has been found */
   BOOLEAN dheader_reset=TRUE;	/* Need to clear previous dheader keywords? */
   unsigned long address;	/* Initially set by _DATA, then incremented */
   char memtype = 'X';		/* Set by _DATA to indicate which mem segment */
   int n;			/* Controller (0 or 1) currently being booted */
   int sym_count = 0;		/* Number of symbols found in .lod file */
   BoardID board;		/* Sets utility or timing board for boot */
   lb_set_t lbset;		/* Chooses controllers and boards for boot */
   BOOLEAN force = FALSE;	/* Force reboot even if serial number matches */
   const char* filebase;	/* Filename of .lod, not including path */
   const char* filename;	/* DSP lod filename */
   char** dspcode_filename;	/* Saved DSP lod filename */
   time_t* dspcode_timestamp;	/* Saved DSP lod file timestamp (modification time) */
   BOOLEAN new_filename=TRUE;	/* Don't allow observing mode to *change* filename */

   unsigned long address_cache[(0x800000>>5)];

   CLI_DEBUG(printf("debug: com_boot(\"%s\") starting\n", args?args:""));

   if (cargv) { if (cargv[0]) free(cargv[0]); free(cargv); }
   cargv = cli_argv(&cargc, args);

   /*
    * "boot status" command just displays what's currently booted.
    */
   if (cargc == 1 && !strcmp(cargv[0], "status"))
   {
      for (n=0; n<n_controllers; n++) for (board=0; board<MAX_BOARDS; board++)
      {
	 show_dsp_info(n, board);
	 if (leach[n].name_board[board] != 0)
	 {
	    unsigned long serno_v = sym_to_number(&leach[n], board, "SERNO_V");
	    unsigned long serno = sym_rdmem(leach[n].name_cmdarg[board], "SERNO");

	    if (serno_v != SYM_INVALID &&
		serno != serno_v)
	    {
	       leach[n].dspcode_ok[board] = 0;
	       /*
		* The next "boot" command will reboot in this case.
		*/
	    }

#ifdef LIO_16BIT_READ
	    printf("%s: %s(#%ld) %s, %s, %d symbols.\n",
		   leach[n].name_cmdarg[board],
		   leach[n].dspcode_filename[board],
		   serno_v,
		   leach[n].dspcode_ok[board]?"OK":"NOT booted",
		   leach[n].has_rm2[board]?"has RM2":"no RM2",
		   leach[n].sym_count[board]);
#else
	    printf("%s: %s(#%ld) %s, %d symbols.\n",
		   leach[n].name_cmdarg[board],
		   leach[n].dspcode_filename[board],
		   serno_v,
		   leach[n].dspcode_ok[board]?"OK":"NOT booted",
		   leach[n].sym_count[board]);
#endif
	 }
      }
      return PASS;
   }

   com_readout("wait");

   if (cargc == 3 && !strcmp(cargv[0], "-force"))
   {
      force = TRUE;
      lbset = lb_parse(cargv[1]);
      filename = cargv[2];
   }
   else if (cargc == 2)
   {
      lbset = lb_parse(cargv[0]);
      filename = cargv[1];
   }
   else if (cargc == 0 || (cargc == 1 && !strcmp(cargv[0], "-force")))
   {
      PASSFAIL rtn = PASS;
      BOOLEAN first_swv = TRUE;

      force = (cargc == 1)?TRUE:FALSE;

      /*
       * Default is to reboot to all boards on all controllers
       */
      strcpy(dsp_version_, "");
      for (n=0; n<n_controllers; n++) for (board=0; board<MAX_BOARDS; board++)
      {
	 if (board==LTB || board==LUB)
	 {
	    const char* boardname = leach[n].name_cmdarg[board];
	    const char* dspcode = leach[n].dspcode_filename[board];

	    if (!dspcode || !*dspcode)
	    {
	       printf("error: No DSP code selected for %s\n",
		      leach[n].name_board[board]);
	       rtn = FAIL;
	    }
	    else
	    {
	       char junkbuf[1024];
	       unsigned long dspid_v, serno_v;

	       sprintf(junkbuf, "%s%.100s %.900s", force?"-force ":"", boardname, dspcode);
	       /*
		* %%% Recursively calling com_boot here.  cargc and cargv are
		* no longer valid!
		*/
	       if (com_boot(junkbuf)==FAIL) rtn = FAIL;
	       /*
		* Make a comma separated list of SERNO's to be stored in the CONSWV
		* FITS header.
		*/
	       serno_v = sym_to_number(&leach[n], board, "SERNO_V");
	       dspid_v = sym_to_number(&leach[n], board, "DSPID_V");
	       sprintf(dsp_version_ + strlen(dsp_version_), "%s%c%c%c=%ld", first_swv?"":",",
		       TOCHAR(dspid_v>>16), TOCHAR((dspid_v&0xFF00)>>8), TOCHAR(dspid_v&0x00FF),
		       serno_v);
	       first_swv = FALSE;
	       if (n==0 && board==LTB)
	       {
		  /*
		   * detcom configures itself according to several parameters
		   * taken from the first (or only) controller's timing board
		   * DSP code.  This code tells how many other controllers
		   * should be present, how many amplifiers each controller
		   * reads (must be the same for each controller) and the size
		   * of each chip and its overscan regions.
		   *
		   * If the array is an infrared detector, IRFPA equals 1,
		   * and if the code should only be used with a specific
		   * dewar, DEWARID is also set to a code for that dewar.
		   */
		  unsigned long nctl = sym_to_number(&leach[n], board, "NCTL");
		  unsigned long namp = sym_to_number(&leach[n], board, "NAMP");
		  unsigned long nxpx = sym_to_number(&leach[n], board, "NXPX");
		  unsigned long nxov = sym_to_number(&leach[n], board, "NXOV");
		  unsigned long nypx = sym_to_number(&leach[n], board, "NYPX");
		  unsigned long nyov = sym_to_number(&leach[n], board, "NYOV");
		  unsigned long irfpa = sym_to_number(&leach[n], board, "IRFPA");
		  unsigned long dewarid = sym_to_number(&leach[n], board, "DEWARID");
		  unsigned long dewarid_detected = 0;

		  switch (irfpa)
		  {
		     case SYM_INVALID: /* Not defined? Assume CCD */
		     case 0:
		     {
			printf("logonly: DSP code for CCDs selected.\n");
			break;
		     }
		     case 1:
		     {
			printf("logonly: DSP code for IRFPA selected.\n");
			printf("error: This is not the IR version of DetCom\n");
			exit(EXIT_FAILURE);
			/* %%% Set some global mode flag */
			break;
		     }
		     default:
		     {
			printf("error: Don't know how to handle DSP code for IRFPA=%ld\n", irfpa);
			return FAIL;
			/* %%% Is this enough to ensure nothing will be done? */
		     }
		  }

		  /*
		   * Detect current dewarid
		   * %%% Move this into a separate routine which returns
		   * (12?) if ACE detected for CFH12K.
		   */
		  {
		     unsigned long dig_in;

		     if (rdmem(&(leach[0]), LUB, LEACH_MEMBASE('Y')|0xFFF0, &dig_in,
			       /*verify=*/0, /*verbose=*/0)==FAIL)
		     {
			printf("warning: rdmem of DIG_IN failed\n");
			dewarid_detected = -1; /* error */
		     }
		     else
		     {
			dewarid_detected = dig_in & 0x1F;
		     }
		  }

		  if (dewarid == -1)
		  {
#ifdef NEED_DEWARID
		     printf("error: The timing board DSP lod file does not define a DEWARID.\n");
		     printf("error: The DEWARID detected is %ld.\n", dewarid_detected);
		     printf("error: PLEASE ADD DEWARID TO `%s'\n",
			    leach[n].dspcode_filename[LTB]);
		     return FAIL;
#endif
		     printf("logonly: No DEWARID in DSP code (detected is %ld.)\n",
			    dewarid_detected);
		     printf("logonly: Compiled without NEED_DEWARID; continuing.\n");
		  }
		  else if (dewarid_detected != dewarid)
		  {
		     printf("warning: The DSP file `%s' is only intended for DEWAIRD %ld.\n",
			    leach[n].dspcode_filename[LTB], dewarid);
		     printf("warning: The dewar ID currently detected is %ld.\n",
			    dewarid_detected);
		     /* %%% Check for slave too? */
		     if (dewarid_detected != 31)
		     {
			printf("error: Incorrect DSP code for dewar ID!\n");
			return FAIL;
		     }
		  }

		  if (nctl != SYM_INVALID)
		  {
		     while (nctl < n_controllers)
		     {
			printf("warning: DSP code says we have only %ld controller%s.  "
			       "Ignoring extra interface card #%d.\n",
			       nctl, nctl==1?"":"s", --n_controllers);
			lio_close(leach[n_controllers].dev);
			leach[n_controllers].dev = 0;
		     }
		     if (nctl > MAX_CONTROLLERS)
		     {
			printf("error: Recompile with MAX_CONTROLLERS=%ld (now %d)\n",
			       nctl, MAX_CONTROLLERS);
		     }
		     else if (nctl > n_controllers)
		     {
			printf("warning: DSP code expected %ld controllers; only found %d.\n",
			       nctl, n_controllers);
		     }
		  }
		  nctl = n_controllers;

		  /*
		   * If there is only one controller, get rid of the "1st" and "2nd"
		   * strings in the names, which would just be confusing.
		   */
		  if (nctl == 1)
		  {
		     leach[0].name_contr = "controller";
		     leach[0].name_board[LUB] = "utility board";
		     leach[0].name_board[LTB] = "timing board";
		     leach[0].name_cmdarg[LUB] = "lub";
		     leach[0].name_cmdarg[LTB] = "ltb";
		  }
		  show_dsp_info(0, LUB);
		  show_dsp_info(0, LTB); 

		  if (namp == SYM_INVALID)
		  {
		     namp = MAX_AMPS_PER_CONTROLLER;
		     printf("warning: NAMP is not defined in DSP code.  "
			    "Assuming %ld amplifiers.\n", namp);
		  }
		  if (namp > MAX_AMPS_PER_CONTROLLER)
		  {
		     printf("error: Recompile with MAX_AMPS_PER_CONTROLLER=%ld (now %d)\n",
			    namp, MAX_AMPS_PER_CONTROLLER);
		     namp = MAX_AMPS_PER_CONTROLLER;
		  }
		  if (nxpx != SYM_INVALID && nypx != SYM_INVALID)
		  {
		     if (nxov == SYM_INVALID)
		     {
			nxov = DEFAULT_NXOV;
			printf("warning: NXOV undefined; assuming %ld pixels.\n", nxov);
		     }
		     if (nyov == SYM_INVALID)
		     {
			nyov = DEFAULT_NYOV;
			printf("warning: NXOV undefined; assuming %ld pixels.\n", nyov);
		     }
		  }
		  else
		  {
		     printf("warning: Assuming default amplifier region size.\n");
		     nxov = DEFAULT_NXOV;
		     nyov = DEFAULT_NYOV;
		     nxpx = DEFAULT_NXPX;
		     nypx = DEFAULT_NYPX;
		  }
		  printf("logonly: im_buf_cfg for %ld contr, %ld amps/contr, %ld X %ld size.\n",
			 nctl, namp, nxpx + nxov, nypx + nyov);
		  if (im_buf_cfg(nctl, namp, nypx + nyov, nxpx + nxov)==0)
		  {
		     printf("warning: Can't allocate image buffer.\n");
		  }

		  /*
		   * Re-configure the raster, in case chip size is different.
		   */
		  /* com_raster(raster_command()); %%% handle raster_id() here! */
		  /*
		   * Make sure the current exposure type is still supported.
		   */
		  com_etype(etype_command());
	       }
	    }
	 }
      }
      if (rtn==PASS)
      {
	 /*
	  * Test if the high voltage is on (we don't want to just
	  * turn it on every time this command runs, because this
	  * runs before every exposure, as part of the sequence
	  * that makes sure everything is ready for a readout.
	  */
	 if (detcom_mode != DETCOM_OBSERVING)
	 {
	    if (com_power("-q test")==FAIL)
	    {
	       printf("warning: Engineering/safe mode.  Use `power on' to enable high voltages.\n");
	    }
	    else force_pon = FALSE; /* Not used unless we switch out of
				     * engineering mode.
				     */
	 }
	 else if (force_pon || com_power("-q test")==FAIL)
	 {
	    if (detcom_mode == DETCOM_OBSERVING) /* Extra check */
	    {
	       /*
		* If "power test" (PTS) fails, that's fine.  It just
		* means the power was probably off (maybe we just did
		* a boot.)  Turn the power back on.
		*/
	       rtn = com_power("on"); /* This is the ONLY place in DetCom
				       * where the power should be turned
				       * on automatically.
				       */
	       if (rtn == PASS) force_pon = FALSE;
	    }
	 }
      }
      return rtn;
   }
   else
   {
      printf("usage: boot [-force] <board(s)> <filename.lod>\n");
      return FAIL;
   }

   if (!(lodfile=fopen(filename, "r")))
   {
      perror(filename);
      printf("error: can't open .lod file\n");
      return FAIL;
   }

   filebase = strrchr(filename, '/');
   if (!filebase) filebase = filename;
   else filebase++;

   /*
    * Main DSP .lod file parsing loop:
    */
   n = -1;
   while (++n < n_controllers) for (board=0; board<MAX_BOARDS; board++)
   {
      struct stat stbuf;

      if (!LBISSET(&lbset, n, board)) continue;

      dspcode_filename = &(leach[n].dspcode_filename[board]);
      dspcode_timestamp = &(leach[n].dspcode_timestamp[board]);

      if (!*dspcode_filename || !**dspcode_filename ||
	  strcmp(*dspcode_filename, filename)!=0)
	 new_filename = TRUE;
      else
	 new_filename = FALSE;

      if (detcom_mode == DETCOM_OBSERVING)
      {
	 /*
	  * Don't allow the observer to select a new DSP code file.
	  * If the file changes on disk, however, the observer is
	  * allowed to issue a "boot" for the same file to upgrade
	  * without switching to engineering mode.
	  */
	 if (new_filename && *dspcode_filename && **dspcode_filename)
	 {
	    printf("error: Cannot change %s DSP code from `%s' to `%s' in this mode.\n",
		   leach[n].name_board[board], *dspcode_filename, filename);
	    return FAIL;
	 }
      }

      /*
       * If fstat() on the file succeeds, and if the file has not been
       * modified in the last 60 seconds, and if the timestamp still
       * matches, then current symbol table is still valid and skip to
       * pass #2.  (But only if all three conditions are met.)
       */
      if (fstat(fileno(lodfile), &stbuf) == 0 &&
	  stbuf.st_mtime < time(0) - 60 &&
	  stbuf.st_mtime == *dspcode_timestamp)
      {
	 CLI_DEBUG(printf("debug: File not modified for %s; using old symtable.\n",
			  leach[n].name_board[board]));
	 pass = 2; /* Begin with pass 2.  Keep old symbols. */
      }
      else
      {
	 pass = 0;
	 leach[n].dspcode_ok[board] = 0;
	 wrmem_worked = 0;
	 *dspcode_timestamp = stbuf.st_mtime;
      }

      /*
       * OLD:
       * 0 - count symbols, download
       * 1 - store symbols, verify downloaded data
       *
       * NEW:
       * 0 - count symbols
       * 1 - store and sort symbols
       * Check if SERNO and SERNO_V exist, verify against code.
       * 2 - download
       * 3 - verify
       */
      for (; pass<4; pass++)
      {
	 address_cache_zero();

	 if (pass==0)
	 {
	    printf("progress: Checking DSP code for %s\n",
		   leach[n].name_board[board]);
	    sym_count = 0;
	 }

	 if (pass==1 && sym_count)
	 {
	    /* Allocate a table for the symbols on the second pass */
	    sym_create(&leach[n], board, sym_count);
	    sym_count = 0;
	 }

	 if (pass==2)
	 {
	    static time_t time_last = 0;
	    
	    if (new_filename)
	    {
	       if (*dspcode_filename) free(*dspcode_filename);
	       *dspcode_filename = strdup(filename);
	    }
	    show_dsp_info(n, board);

	    CLI_DEBUG(printf("debug: DSP code selected for %s: `%s'\n",
			     leach[n].name_board[board], filename));

	    /*
	     * Avoid a lot of confusing errors by simply exiting now
	     * if the board does not respond to TDL.
	     */
	    {
	       char cmdarg[80];

	       sprintf(cmdarg, "%s 0x55A55A debug: logonly:", leach[n].name_cmdarg[board]);
	       if (com_tdl(cmdarg)==FAIL)
	       {
		  printf("error: Can't communicate with %s; NOT booting %s\n",
			 leach[n].name_board[board], leach[n].dspcode_filename[board]);
		  fclose(lodfile);
		  return FAIL;
	       }
	       if (board==LTB)
	       {
		  sprintf(cmdarg, "-quiet %s STP", leach[n].name_cmdarg[board]);
		  com_cmd(cmdarg);
	       }
	    }

	    /*
	     * If there are two controllers (CFH12K), attempt to read
	     * a bit from DIG_IN that tells if we have the master or slave.
	     * We could *almost* use sym_rdmem() here, but if you boot
	     * the timing board before the utility board, then DIG_IN
	     * is not defined yet.  So, for now, it is hardcoded here.
	     *
	     * Don't repeat the check if it was just completed within 20
	     * seconds ago.  This will speed up the pre-exposure boot
	     * check, which gets to this point in the code for each board.
	     */
	    if (n_controllers == 2 && (time(0) > time_last + 20))
	    {
	       unsigned long dig_in0_hi=0, dig_in1_hi=0, dig_in0_lo=0, dig_in1_lo=0;
	       int found_ace = 0;
	       static BOOLEAN warned_ace = 0;

#if !defined(CAMERA_CFH12K) && !defined(CAMERA_MDM8K)
	       found_ace = 1; /* %%% */
#endif
	       time_last = time(0);
	       printf("progress: Verifying ACE controller ID bits\n");
	       while (!found_ace && (time(0) <= time_last + 1))
	       {
		  unsigned long dig_in0, dig_in1;

		  if (rdmem(&(leach[0]), LUB, LEACH_MEMBASE('Y')|0, &dig_in0,
			    /*verify=*/0, /*verbose=*/0)==FAIL ||
		      rdmem(&(leach[1]), LUB, LEACH_MEMBASE('Y')|0, &dig_in1,
			    /*verify=*/0, /*verbose=*/0)==FAIL)
		  {
		     printf("warning: rdmem of DIG_IN failed\n");
		     found_ace = -1; /* error */
		  }
		  dig_in0_hi |= dig_in0;
		  dig_in1_hi |= dig_in1;
		  dig_in0_lo |= ~dig_in0;
		  dig_in1_lo |= ~dig_in1;
		  if ((dig_in0_lo & 2) && (dig_in0_hi & 2) &&
		      (dig_in1_lo & 2) && (dig_in1_hi & 2))
		  {
		     found_ace=1; /* Detected heartbeat; don't loop anymore */
		  }
	       }
	       if (found_ace!=1)
	       {
#if defined(CAMERA_CFH12K) || defined(CAMERA_MDM8K)
		  printf("warning: Unable to verify master/slave ID bits from ACE\n");
#endif
	       }
	       else if (!(dig_in0_lo & 2) || !(dig_in0_hi & 2) ||
			!(dig_in1_lo & 2) || !(dig_in1_hi & 2))
	       {
		  printf("logonly: ACE bits: dig_in0_hi=0x%06lx dig_in1_hi=0x%06lx\n",
			 dig_in0_hi, dig_in1_hi);
		  printf("logonly: ACE bits: dig_in0_lo=0x%06lx dig_in1_lo=0x%06lx\n",
			 dig_in0_lo, dig_in1_lo);
		  printf("logonly: ACE bits: dig_in0_df=0x%06lx dig_in1_df=0x%06lx\n",
			 dig_in0_lo^dig_in0_hi, dig_in1_lo^dig_in1_hi);
#ifdef LIO_16BIT_READ
		  if (leach[0].has_rm2[LUB] && leach[1].has_rm2[LUB])
#endif
		  {
		     if ((dig_in0_lo && dig_in0_hi))
		     {
			printf("warning: ACE not connected to slave?  No heartbeat on slave.\n");
		     }
		     else if ((dig_in1_lo && dig_in1_hi))
		     {
			printf("warning: ACE not connected to master?  No heartbeat on master.\n");
		     }
		     else
		     {
			printf("warning: ACE not connected?  No heartbeat detected.\n");
		     }
		  }
#ifdef LIO_16BIT_READ
		  else
		  {
		     if (!warned_ace)
		     {
			printf("warning: Can't verify ACE connection until all DSP code is booted.\n");
			warned_ace = TRUE;
		     }
		  }
#endif
	       }
	       else if ((dig_in0_lo & 1)+(dig_in0_hi & 1) != 1 ||
			(dig_in1_lo & 1)+(dig_in1_hi & 1) != 1)
	       {
		  printf("warning: ACE master/slave ID bits are not stable.\n");
	       }
	       else
	       {
		  static BOOLEAN swapped = FALSE;

		  if (warned_ace)
		  {
		     /* Re-assure the user that they can ignore the warning previously issued. */
		     printf("status: ACE is connected to controllers.\n");
		     warned_ace = FALSE;
		  }

		  switch ((dig_in0_hi & 1) | ((dig_in1_hi & 1) << 1))
		  {
		     case 0:
		     case 3:
		     {
			printf("warning: Master DIG_IN&1 = %ld, Slave DIG_IN&1 = %ld\n",
				    dig_in0_hi & 1, dig_in1_hi & 1);
			leach[0].name_contr = "1st controller";
			leach[1].name_contr = "2nd controller";
			leach[0].name_board[LUB] = "1st utility board";
			leach[1].name_board[LUB] = "2nd utility board";
			leach[0].name_board[LTB] = "1st timing board";
			leach[1].name_board[LTB] = "2nd timing board";
			break;
		     }
		     case 1:
		     {
#if 1
			Devp tmp;

			if (swapped) swapped = FALSE; else swapped = TRUE;
			printf("warning: Swapping master and slave! (%s)\n",
			       (swapped)?"messages will be correct with respect to the controller's end of the fiber":"back to normal configuration");
			tmp = leach[0].dev;
			leach[0].dev = leach[1].dev;
			leach[1].dev = tmp;
#else
			printf("warning: Master and slave fibers may be swapped?  Not correcting.\n");
#endif
			/* Fall through an re-assign names */
		     }
		     case 2:
		     {
			printf("logonly: master and slave ACE ID bits are OK.\n");
			leach[0].name_contr = swapped?"master[2nd pci]":"master";
			leach[1].name_contr = swapped?"slave[1st pci] ":"slave ";
			leach[0].name_board[LUB] = swapped?"master[2nd pci] utility board":"master utility board";
			leach[1].name_board[LUB] = swapped?"slave[1st pci] utility board":"slave utility board";
			leach[0].name_board[LTB] = swapped?"master[2nd pci] timing board":"master timing board";
			leach[1].name_board[LTB] = swapped?"slave[1st pci] timing board":"slave timing board";
			break;
		     }
		  }
	       }
	    }

	    /*
	     * Unless the '-force' option was given, check to see
	     * if the DSP code is already booted.  If DSPID
	     * and SERNO values match, then break out of the loop.
	     */
#ifdef LIO_16BIT_READ
	    leach[n].has_rm2[board] = 1;
#endif
	    if (!force && !leach[n].dspcode_force[board])
	    {
	       unsigned long dspid_v = sym_to_number(&leach[n], board, "DSPID_V");
	       unsigned long serno_v = sym_to_number(&leach[n], board, "SERNO_V");

	       if (dspid_v != SYM_INVALID && serno_v != SYM_INVALID)
	       {
		  unsigned long dspid_check;
		  unsigned long serno_check;
		  unsigned long dspid = sym_to_addr(&leach[n], board, "DSPID");
		  unsigned long serno = sym_to_addr(&leach[n], board, "SERNO");
		  
		  /*
		   * === Krusty. What about allowing string args to the lowlevel
		   * rdmem command, or allowing a '-quiet' option to the com_rdmem?
		   */
		  if (dspid != SYM_INVALID && serno != SYM_INVALID)
		  {
		     if (rdmem(&leach[n], board, dspid, &dspid_check,
			   /*verify=*/0, /*verbose=*/0)==PASS && dspid_check == dspid_v &&
			 rdmem(&leach[n], board, serno, &serno_check,
			   /*verify=*/0, /*verbose=*/0)==PASS && serno_check == serno_v)
		     {
			CLI_DEBUG(printf("debug: %s DSPID and SERNO match; not re-booting.\n",
			       leach[n].name_board[board]));
			leach[n].dspcode_ok[board] = 1;
			mem_retries_reset();
			show_dsp_info(n, board);
			break;
		     }
		     else
		     {
			if (dspid_v == dspid_check)
			{
			   printf("warning: Upgrading %s DSP code from "
				  "version %ld to %ld to match %s.\n",
				  leach[n].name_board[board],
				  serno_check, serno_v, filename);
			   show_dsp_info(n, board);
			}
			else
			{
			   CLI_DEBUG(printf("debug: Expected DSPID 0x%06lX got 0x%06lX; "
					    "Expected SERNO 0x%06lX got 0x%06lX; re-booting.\n",
					    dspid_v, dspid_check, serno_v, serno_check));
			}
		     }
		  }
		  else
		  {
		     CLI_DEBUG(printf("debug: DSP code defines DSPID_V and SERNO_V, "
				      "but not found in P: memory.\n"));
		  }
	       }
	       else
	       {
		  CLI_DEBUG(printf("debug: DSP code does not support DSPID/SERNO feature.\n"));
	       }
	    }
#ifdef LIO_16BIT_READ
	    leach[n].has_rm2[board] = 0;
#endif
	    mem_retries_reset(); /* Keep stats on number of fiber communication retries */
	 }

	 end_found = in_data_block = address = line = 0;
	 in_symbol_table = '\0';
	 in_comment_block = FALSE;

	 if (pass==2)
	 {
	    force_pon = TRUE; /* Re-issue the PON command next pre-exposure. */
	    /*
	     * %%% Somehow resetting the LTB also resets the LUB (it no longer
	     * understands the commands PON/POF) but it doesn't clear the
	     * DSPID or SERNO memory locations, so a re-boot doesn't happen,
	     * unless it is forced like this:
	     */
	    if (board == LTB)
	       leach[n].dspcode_force[LUB] = 1;
	 }

	 /*
	  * If file still hasn't been read, make a quick pass to
	  * calculate the total number of lines for the progress meter.
	  */
	 if (lines_total == -1)
	 {
	    if (fseek(lodfile, 0L, SEEK_SET)==-1)
	    {
	       perror("fseek(lodfile)");
	       fclose(lodfile);
	       return FAIL;
	    }

	    lines_total = 0;
	    /*
	     * See how many lines there are in the file, before the
	     * boot code begins.
	     */
	    while (fgets(buf, sizeof(buf), lodfile))
	    {
	       if (sscanf(buf, "_DATA %c %lX", &memtype, &address)==2 && address >= 0x1000)
		  break;
	       else lines_total++;
	    }

	    if (lines_total==0)
	    {
	       printf("error: Can't read any application DSP code from `%s'\n", filename);
	       fclose(lodfile);
	       return FAIL;
	    }
	 }

	 /*
	  * Be sure to start reading from the beginning of the file on each pass
	  */
	 if (fseek(lodfile, 0L, SEEK_SET)==-1)
	 {
	    perror("fseek(lodfile)");
	    fclose(lodfile);
	    return FAIL;
	 }

	 while (!end_found && (p = fgets(buf, sizeof(buf), lodfile)) && ++line)
	 {
	    if (pass==2 || pass==3)
	    {
	       printf("progress: %s code for %s (%d%%)\r",
		      (pass==3)?"Verifying":"Loading",
		      leach[n].name_board[board],
		      wrmem_worked==0?0: /* Stick at 0% until the first wrmem works */
		      ((line>lines_total?lines_total:line)*100/lines_total));
	       fflush(stdout);
	    }
	    if (*p=='_')
	    {
	       in_symbol_table = '\0';
	       in_data_block = 0;
	       in_comment_block = FALSE;

	       if (!strncmp(p, "_END", 4)) end_found = 1;
	       else if (sscanf(p, "_DATA %c %lX", &memtype, &address)==2)
	       {
		  in_data_block = 1;
		  switch (memtype)
		  {
		     case 'P': if (address >= LEACH_P_ROM) in_data_block = 0; break;
		     case 'X': if (address >= LEACH_X_ROM) in_data_block = 0; break;
		     case 'Y': if (address >= LEACH_Y_ROM) in_data_block = 0; break;
		     default:
		     if (pass==0)
			printf("warning: Ignoring unknown memory space `%c'\n", memtype);
		     in_data_block = 0;
		  }
		  if (in_data_block)
		     address |= LEACH_MEMBASE(memtype);
		  else
		     if (pass==2)
			CLI_DEBUG(printf("debug: Skipping download of _DATA at %c:%04lX\n",
					 memtype, address));
	       }
	       else if (!strncmp(p, "_COMMENT", 8))
	       {
		  in_comment_block = TRUE;
	       }
	       else if (!strncmp(p, "_SYMBOL ", 8))
	       {
		  in_symbol_table = p[8];
	       }
	    }
	    /*
	     * _COMMENT blocks can contain detcom commands for configuring
	     * FITS keywords like DETECTOR and GAIN.  Just like the symbols,
	     * this gets set on pass==1, so that an agent restart will
	     * still have the correct dheader's set.
	     */
	    else if (in_comment_block)
	    {
	       if (pass==1 && !strncmp(p, "detcom:", 7))
	       {
		  if (!strncmp(p+7, "dheader ", 8))
		  {
		     if (dheader_reset)
		     {
			com_dheader("reset");
			dheader_reset = FALSE;
		     }
		     com_dheader(p+7+8);
		  }
		  else
		  {
		     printf("warning: Unknown detcom command in DSP code: `%s'\n", p+7);
		  }
	       }
	    }
	    /*
	     * Parsing code for _DATA blocks.  Just keep reading 24-bit
	     * hexadecimal values and incrementing the address pointer.
	     */
	    else if (in_data_block)
	    {
	       char* next;
	       unsigned long data;
	       
	       do
	       {
		  data = strtoul((next=p), &p, 16);
		  if (next != p)
		  {
		     if (address & ~0x7FFFFF)
		     {
			printf("error: %s line %d: address 0x%06lX is out of range\n",
			       filebase, line, address);
			return FAIL;
		     }
		     if (pass==2 && address_cache_test(address))
		     {
			CLI_DEBUG(printf("debug: %s line %d: Leaving re-write of address 0x%06lX for second pass\n",
			       filebase, line, address));
		     }
		     else if (pass>=2)
		     {
			if (wrmem(&leach[n], board, address, data,
				  (pass==3) && !address_cache_test(address))==FAIL)
			{
			   /*
			    * Retries and verification are built into wrmem(), so
			    * if it returns fail, kick the bucket and return FAIL.
			    */
			   printf("logonly: %s line %d: Can't write 0x%06lX at 0x%06lX of %s\n",
				  filebase, line, data, address, leach[n].name_board[board]);
			   if (wrmem_worked)
			      printf("error: Failure in middle of download to %s\n",
				     leach[n].name_board[board]);
			   else
			      printf("error: Download to %s failed completely\n",
				     leach[n].name_board[board]);
			   fclose(lodfile);
			   return FAIL;
			}
			else wrmem_worked = 1;
		     }

		     address_cache_set(address); /* WARNING: macro, don't put the ++ in here! */
		     address++;
		  }
	       } while (next != p);
	    }
	    /*
	     * Parsing code for _SYMBOL tables.  The in_symbol_table variable
	     * stores what type of symbol block this is (N,P,X,Y).
	     */
	    else if (in_symbol_table && *p >= ' ')
	    {
	       char symname[20];
	       unsigned long symval_int;
	       char symtype;

	       if (sscanf(p, "%s %c %lX", symname, &symtype, &symval_int)!=3)
	       {
		  printf("error: %s line %d: Can't parse symbol table\n",
			 filebase, line);
		  fclose(lodfile);
		  return FAIL;
	       }
	       if (pass==0) sym_count++;
	       else if (pass==1)
	       {
		  char symval[20];
		  int i;

		  sprintf(symval, "%c:%06lX", in_symbol_table, symval_int);
		  for (i=0; i<sym_count; i++)
		  {
		     if (!xstrcasecmp(leach[n].symbols[board][i].symname, symname))
		     {
			if (xstrcasecmp(leach[n].symbols[board][i].symval, symval))
			{
			   printf("warning: Ignoring redfinition of symbol `%s' to `%s'; "
				  "Using first value, `%s'.\n",
				  symname, symval, leach[n].symbols[board][i].symval);
			   strcpy(symval, leach[n].symbols[board][i].symval);
			}
		     }
		  }
		  sym_set(&leach[n], board, sym_count++, symname, symval);
	       }
	    }
	 }
	 if (pass==1 && sym_count)
	 {
	    CLI_DEBUG(printf("debug: Sorting %d symbols from %s\n", sym_count, filebase));
	    sym_sort(&leach[n], board);
	 }
#ifdef LIO_16BIT_READ
	 if (pass==2) leach[n].has_rm2[board] = 1;
#endif
	 if (pass==3)
#ifdef LINUX
	 {
	    /*
	     * The SBus driver took some time to switch to DMA readout mode.
	     * To be sure DetCom and the driver were ready, the DSP code
	     * had to delay a bit before sending pixels.  This delay is not
	     * needed with the PCI driver, so here's a little hack that looks
	     * for the SBus hack and unhacks it :-(
	     */
	    unsigned long delrdt = sym_to_addr(&leach[n], board, "DELRDT");

#define DELRDT_VAL 1

	    if (delrdt != SYM_INVALID)
	    {
	       if (wrmem(&leach[n], board, delrdt, DELRDT_VAL, 1)==FAIL)
		  printf("warning: Unable to set DELRDT\n");
	       else
		  printf("logonly: DELRDT set to %d for PCI driver\n", DELRDT_VAL);
	    }
	    CLI_DEBUG(printf("debug: No DELRDT symbol in %s\n", filebase));

	    leach[n].dspcode_ok[board] = 1;
	    leach[n].dspcode_force[board] = 0;
	    com_preexp("reset");
	    show_dsp_info(n, board);
	    /*
	     * Any previous preexp operations that may have been done have
	     * probably been overwritten by the reboot, so make sure the
	     * exposure sequence will re-do the preexp or fail rather than
	     * taking an exposure/readout with the wrong setup.
	     */
	 }
#endif
      }

      if (mem_retries())
      {
	 printf("warning: %d retr%s needed while downloading to %s\n",
		mem_retries(), mem_retries()>1?"ies":"y", leach[n].name_board[board]);
      }
#ifdef LIO_16BIT_READ
      if (!leach[n].has_rm2[board])
      {
	 printf("warning: RM2 command was not available for %s\n",
		leach[n].name_board[board]);
      }
#endif
   }

   if (ferror(lodfile))
   {
      perror(filename);
      printf("warning: File error at line %d; download may be incomplete\n", line);
      fclose(lodfile);
      return FAIL;
   }
   else
   {
      if (line > 0)
	 CLI_DEBUG(printf("debug: Processed %d lines of dsp code\n", line));
   }
   fclose(lodfile);

   printf("progress: DSP boot complete . . .\n");
   return PASS;
}

const char* dsp_version( void ) { return dsp_version_; }

/* End of leach_dsp.c */
