/*                                             -*- c-file-style: "Ellemtel" -*-

`leach_mem.c' - Memory read/write commands (rdmem, wrmem)

This file is part of version 3 of the CFHT Detector Host Command Interface.
Read the `License' file for terms of use and distribution.
Copyright 1998-2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>

#include "cli/cli.h"
#include "leachcont.h"
#include "detcom.h" /* === Why? */

static unsigned long
stow3(const char* s)
{
   char a = s[0];
   char b = s[1];
   char c = s[2];

   return STOW3(a, b, c);
}

static int mem_retries_ = 0;

void mem_retries_reset(void) { mem_retries_ = 0; }
int  mem_retries(void) { return mem_retries_; }

unsigned long
sym_to_number(LeachDev* dev, int board, const char* symname)
{
   const char* sym = sym_lookup(dev, board, symname, DSP_SYM_NUMBER);
   unsigned long number;
   char* endptr;

   if (!strcmp(sym, symname))
   {
      /*
       * symname may also be a direct numeric value instead
       * of an N: symbol.
       */
      number = strtoul(sym, &endptr, 0);
   }
   else
   {
      if (!*sym || sym[1]!=':')
      {
	 printf("logonly: %s is not currently a valid symbol for %s\n",
		symname, NAMESTR(dev->name_board[board]));
	 return SYM_INVALID;
      }
      number = strtoul(&sym[2], &endptr, 16);
   }
   if (!endptr || *endptr!='\0')
   {
      printf("logonly: %s is not a valid numeric value\n", sym);
      return SYM_INVALID;
   }
   return number;
}

unsigned long
sym_to_addr(LeachDev* dev, int board, const char* symname)
{
   const char* sym = sym_lookup(dev, board, symname, DSP_SYM_ADDRESS);
   unsigned long addr_offset;
   char memtype;
   char* endptr;

   if (!*sym || sym[1]!=':')
   {
      printf("logonly: %s is not currently a valid symbol for %s\n",
	     symname, NAMESTR(dev->name_board[board]));
      return SYM_INVALID;
   }
   memtype = sym[0];
   addr_offset = strtoul(&sym[2], &endptr, 16);
   if (!endptr || *endptr!='\0')
   {
      printf("logonly: %s is not a valid address\n", sym);
      return SYM_INVALID;
   }
   return addr_offset|LEACH_MEMBASE(memtype);
}

PASSFAIL
rdmem( LeachDev* leach, BoardID board,
       unsigned long addr, unsigned long* data,
       int verify, int verbose )
/*
 * Read leach memory at address `addr' and store the result in `data'.
 * If `verify' is non-zero, the operation is repeated until the same
 * result is obtained twice in a row.
 */
{
   unsigned long rdm[3];
#ifdef LIO_16BIT_READ
   /*
    * The "RM2" hack is only needed when there is no direct way to
    * read all 24 bits back from the leach system (as with the
    * fiber/parallel box in use at CFHT.)
    */
   unsigned long rm2[3];
#endif
   LREAD resp[4];
   unsigned long last_value = NO_VALUE;
   unsigned long new_value;
   int tries = MEM_RETRY + 1;

   if (addr==-1)
   {
      printf("logonly: Invalid address for rdmem\n");
      return FAIL;
   }

   LMSG3_SET(rdm, board, stow3("RDM"), addr);
#ifdef LIO_16BIT_READ
   LMSG3_SET(rm2, board, stow3("RM2"), addr);
#endif

   while (tries--)
   {
      if (tries!=MEM_RETRY)
      {
	 mem_retries_++;
	 if (verbose)
	 {
#ifdef LIO_16BIT_READ
	    printf("logonly: Retrying %s command on %s\n",
		   leach->has_rm2[board]?"RM2":"RDM",
		   NAMESTR(leach->name_board[board]));
#else
	    printf("logonly: Retrying RDM command on %s\n",
		   NAMESTR(leach->name_board[board]));
#endif
	 }
	 if (NORESP_DELAY_USEC) xusleep(NORESP_DELAY_USEC);
/*	 if (tries<2) xusleep(250000); %%% Maybe a little voodoo-delay will help */
      }
      
#ifdef LIO_16BIT_READ
      /*
       * Only use "RM2" command if it is known to be available.  Otherwise,
       * just use RDM and return the 16-bit value with all the bits flipped.
       * (*data will be greater than 0xFFFFFF in this case.)
       */
      if (leach->has_rm2[board])
      {
	 if (lio_write24(leach->dev, rm2, sizeof(rdm)/sizeof(long))==LIO_ERR)
	 {
	    perror("error: lio_write24(RM2) failed");
	    return FAIL;
	 }
      }
      else
#endif
      {
	 if (lio_write24(leach->dev, rdm, sizeof(rdm)/sizeof(long))==LIO_ERR)
	 {
	    perror("error: lio_write24(RDM) failed");
	    return FAIL;
	 }
      }
      
      new_value = NO_VALUE;
      
      /*
       * The rest of this routine automatically handles either an RDM-style
       * response or the split RM2 response with the high 8-bits in the first
       * 16-bit word and the lower 16-bits in the second word.  It figures
       * out which one it is getting by looking at the number of words in
       * the response.  (So, if we wanted to, this could handle DSP code that
       * *replaces* the original RDM command with an RM2.)
       */
      switch (LIO_LREAD(leach->dev, resp, 4, 2, TOMSEC_LEACH_FAST)) /* %%% Could be 2 or 3 */
      {
	 case 2:
	 {
	    if ((resp[0]&0xFFFF)!=0x0002)
	    {
	       if (verbose)
		  printf("logonly: garbled response from %s (2 words read)\n",
			 NAMESTR(leach->name_board[board]));
	    }
#ifdef LIO_16BIT_READ
	    else if (leach->has_rm2[board])
	    {
	       printf("%s: Response from %s to RM2 was 0x**%04X%s\n",
		      verbose?"warning":"logonly",
		      NAMESTR(leach->name_board[board]), resp[1],
		      tries < 3?"; Trying RDM command":"");
	       if (tries < 3)
	       {
		  leach->has_rm2[board] = 0;
		  mem_retries_ -= (MEM_RETRY - tries);
		  /* This does not count as a communications retry */
	       }
	    }
#endif
	    else
	    {
	       new_value = resp[1];
#ifdef LIO_16BIT_READ
	       new_value |= FLAG16;
#endif
	    }
	 }
	 break;
	 case 3:
	 {
	    if ((resp[0]&0xFFFF)!=0x0003)
	    {
	       if (verbose)
		  printf("logonly: garbled response from %s (3 words read)\n",
			 NAMESTR(leach->name_board[board]));
	    }
	    else
	    {
	       new_value = (unsigned long)resp[1]<<16|resp[2];
	    }
	 }
	 break;
	 case LIO_ERR:
	 {
	    if (detcom_mode==DETCOM_FAKE)
	    {
	       new_value = 0;
	    }
#ifdef LIO_16BIT_READ
	    else if (leach->has_rm2[board])
	    {
	       if (verbose)
		  printf("logonly: No response from %s to RM2%s\n",
			 NAMESTR(leach->name_board[board]),
			 tries < 3?"; Trying RDM command":"");
	       if (tries < 3)
	       {
		  leach->has_rm2[board] = 0;
		  mem_retries_ -= (MEM_RETRY - tries);
		  /* This does not count as a communications retry */
	       }
	    }
#endif
	    else
	    {
	       if (verbose)
		  printf("logonly: No response from %s to RDM command\n",
			 NAMESTR(leach->name_board[board]));
	    }
	 }
	 break;
	 default:
	 {
	    if (verbose)
	    {
#ifdef LIO_16BIT_READ
	       printf("logonly: invalid response length from %s to %s command\n",
		      NAMESTR(leach->name_board[board]),
		      leach->has_rm2[board]?"RM2":"RDM");
#else
	       printf("logonly: invalid response length from %s to RDM command\n",
		      NAMESTR(leach->name_board[board]));
#endif
	    }
	 }
      }
      
      if (new_value!=NO_VALUE)
      {
	 if (!verify || last_value == new_value)
	 {
	    *data = new_value;
	    return PASS;
	 }
	 else if (last_value != NO_VALUE)
	 {
	    if (last_value != new_value)
	       printf("warning: %s data at %c:%06lX changed "
		      "from 0x%06lX to 0x%06lX while reading; "
		      "(Use `rdmem -volatile ...' for unstable memory locations.)\n",
		      NAMESTR(leach->name_board[board]),
		      MEMBASE(addr), MEMMASK(addr),
		      last_value, new_value);
	 }
	 last_value = new_value;
      }
   }
   if (verbose)
      printf("logonly: giving up on trying to read address %c:%06lX from %s\n",
	     MEMBASE(addr), MEMMASK(addr),
	     NAMESTR(leach->name_board[board]));
   return FAIL;
}

PASSFAIL
wrmem( LeachDev* leach, BoardID board,
       unsigned long addr, unsigned long data,
       int verify_first )
{
   int tries = MEM_RETRY + 1;
   LREAD resp[4];
   unsigned long data_check;
   const char* warnings = "logonly";
   unsigned long wrm[4];

   LMSG4_SET(wrm, board, stow3("WRM"), addr, data);

   while (tries--)
   {
      if (tries!=MEM_RETRY)
      {
	 mem_retries_++;
	 printf("logonly: retrying WRM command\n");
	 if (NORESP_DELAY_USEC) xusleep(NORESP_DELAY_USEC);
	 verify_first = 0;
      }
      if (!tries) warnings = "warning";
      if (!verify_first)
      {
	 int nwords;

	 if (lio_write24(leach->dev, wrm, sizeof(wrm)/sizeof(long))==LIO_ERR)
	 {
	    perror("error: lio_write24 failed");
	    return FAIL;
	 }
	 nwords = LIO_LREAD(leach->dev, resp, 4, 2, TOMSEC_LEACH_FAST);
	 if (nwords==LIO_ERR)
	 {
	    if (detcom_mode!=DETCOM_FAKE)
	    {
	       printf("%s: no response to WRM command from %s\n",
		      warnings, NAMESTR(leach->name_board[board]));
	       continue;
	    }
	 }
	 else if (nwords!=2)
	 {
	    printf("%s: read %d instead of 2 words in response to WRM command\n",
		   warnings, nwords);
	    continue;
	 }
	 else if ((resp[0]&0xFFFF)!=0x0002)
	 {
	    printf("%s: response from WRM command not 0x0002\n", warnings);
	    continue;
	 }
	 else if (LREAD_MASK(resp[1])!=LREAD_MASK(STOW3('D','O','N')))
	 {
	    printf("%s: didn't receive DON from WRM command\n", warnings);
	    continue;
	 }
      }

      if (detcom_mode==DETCOM_FAKE) return PASS;

      if (rdmem(leach, board, addr, &data_check, /*verify=*/0, /*verbose=*/1)==FAIL)
      {
	 printf("%s: Failed to read back data from leach memory\n", warnings);
      }
      else if (data_check == data || /* Successful 24-bit read */
	       ((data_check & FLAG16) && /* Successful 16-bit read */
		(data_check & 0xFFFF) == (data & 0xFFFF)))
      {
	 return PASS;
      }
      else
      {
	 printf("%s: tried to set 0x%06lX but read back 0x%06lX\n",
		warnings, data, data_check); /* === FLAG16 */
      }
   }
   printf("error: WRM command to %s failed\n", NAMESTR(leach->name_board[board]));
   return FAIL;
}


PASSFAIL
com_wrmem( const char* args )
{
#define MASK24 0xFFFFFF
   static char** cargv = 0;
   static int cargc = 0;
   unsigned long addr;
   unsigned long data, data_old, data_new, mask = MASK24;
   PASSFAIL rdmem_result;
   lb_set_t lbset;
   BoardID board;
   int n;
   char* endptr;

   /*
    * Split the command line arguments.
    */
   if (cargv) { if (cargv[0]) free(cargv[0]); free(cargv); }
   cargv = cli_argv(&cargc, args);

   /*
    * First argument selects SDSU board(s) that should be written to.
    */
   if (cargc > 0) lbset = lb_parse(cargv[0]);
   else
   {
      printf("error: example: wrm ltbM Y:0012AB 0xABCDEF [<mask>]\n");
      return FAIL;
   }

   /*
    * Loop through all controllers and all boards, writing the new
    * value to the selected board(s).
    */
   for (n=0; n<n_controllers; n++)
   {
      for (board=0; board<MAX_BOARDS; board++)
      {
	 if (LBISSET(&lbset, n, board))
	 {
	    /*
	     * Look up symbol names *per board* since each board
	     * could have a different address for the same symbol.
	     */
	    if ((cargc == 4 || cargc == 5) && strlen(cargv[1])==1 &&
		strchr("PXYpxy",cargv[1][0]))
	    {
	       /*
		* Support old DetI-style memory addressing with white
		* space instead of a colon in the address.  This should
		* not be used.
		*/
	       unsigned long addr_offset;
	       char memtype;

	       printf("warning: `%s %s' is obsolete memory address syntax.\n",
		      cargv[1], cargv[2]);
	       memtype = cargv[1][0];
	       addr_offset = strtoul(cargv[2], &endptr, 16);
	       if (!endptr || *endptr!='\0')
	       {
		  printf("error: `%s' is not a valid address\n", cargv[2]);
		  return FAIL;
	       }
	       addr = addr_offset|LEACH_MEMBASE(memtype);
	       data = strtoul(cargv[3], &endptr, 0);
	       if (!endptr || *endptr!='\0')
	       {
		  printf("error: `%s' is not a valid data value\n", cargv[3]);
		  return FAIL;
	       }
	       if (cargc == 5)
	       {
		  mask = strtoul(cargv[4], &endptr, 16);
		  if (!endptr || *endptr!='\0')
		  {
		     printf("error: `%s' is not a valid mask\n", cargv[4]);
		     return FAIL;
		  }
	       }
	    }
	    else if ((cargc == 3 || cargc == 4) && strlen(cargv[1])>2)
	    {
	       /*
		* Check the address.
		*/
	       addr = sym_to_addr(&leach[n], board, cargv[1]);
	       if (addr == SYM_INVALID)
	       {
		  printf("error: `%s' is not a valid address\n", cargv[1]);
		  return FAIL;
	       }

	       /*
		* Check the data value.
		*
		* The built-in words "ON" and "OFF" are used in combination
		* with masks to turn bits on and off.  There is no setbit
		* or clearbit command, but this does the same thing.
		*/
	       if (!xstrcasecmp(cargv[2], "OFF"))	data = 0x000000;
	       else if (!xstrcasecmp(cargv[2], "ON"))	data = 0xFFFFFF;
	       else data = sym_to_number(&leach[n], board, cargv[2]);

	       if (data == SYM_INVALID)
	       {
		  printf("error: `%s' is not a valid data value\n", cargv[2]); /* === Do more of this */
		  return FAIL;
	       }
	       if (cargc == 4)
	       {
		  char* mask_arg = cargv[3];
		  int bitpos = 0;

		  if (strncmp(mask_arg, "1<<", 3)==0)
		  {
		     bitpos = 1;
		     mask_arg+=3;
		  }
		  /* Note: === Old default for direct, numeric masks (ones
		   * not found in a symbol, but just typed directly on the
		   * command line) was to assume HEX.  Now you must put
		   * the 0x in front of all hex values.
		   */
		  mask = sym_to_number(&leach[n], board, mask_arg);
		  if (mask == SYM_INVALID)
		  {
		     printf("error: `%s' is an invalid mask\n", mask_arg);
		     return FAIL;
		  }
		  if (bitpos) mask = 1<<mask;
	       }
	    }
	    else
	    {
	       printf("error: example: wrm ltbM Y:0012AB 0xABCDEF [<mask>]\n");
	       return FAIL;
	    }

	    if ((mask&MASK24)==0x000000)
	    {
	       printf("error: Masks for wrmem must have at least 1 bit set in bits 0..23\n");
	       return FAIL;
	    }

	    data_old = -1;

#ifdef LIO_16BIT_READ
	    if (!leach[n].has_rm2[board])
	       rdmem_result = FAIL;
	    else
#endif
	       rdmem_result =
		  rdmem(&leach[n], board, addr,
			&data_old, /*verify=*/mask==MASK24?0:1, /*verbose=*/0);

	    if (mask != MASK24 && rdmem_result == FAIL)
	    {
	       printf("error: wrmem with mask failed because rdmem failed.\n");
	       /* === More trace info needed here */
	       return FAIL;
	    }

	    data_new = (data_old&~mask) | (data&mask);

	    if (rdmem_result == FAIL || data_old != data_new)
	    {
	       if (wrmem(&leach[n], board, addr, data_new, 0)==FAIL)
	       {
		  printf("error: wrmem command failed.\n");
		  /* === more trace info? */
		  return FAIL;
	       }
	       printf("logonly: %s\t%s\t%c:%06lX was 0x%06lX; now 0x%06lX (%ld)\n",
		      NAMESTR(leach[n].name_board[board]),
		      cargv[1], MEMBASE(addr), MEMMASK(addr),
		      data_old, data_new, data_new);
	    }
	 }
      }
   }
   return PASS;
}


unsigned long
sym_rdmem( const char* board_name, const char* sym_name)
{
   lb_set_t lbset = lb_parse(board_name);
   BoardID board;
   int n;

   for (n=0; n<n_controllers; n++)
   {
      for (board=0; board<MAX_BOARDS; board++)
      {
	 if (LBISSET(&lbset, n, board) && leach[n].name_board[board])
	 {
	    unsigned long sym_addr = sym_to_addr(&(leach[n]), board, sym_name);
	    unsigned long sym_data;
#ifdef LIO_16BIT_READ
	    leach[n].has_rm2[board] = 1; /* === */
#endif
	    if (rdmem( &(leach[n]), board,
		       sym_addr, &sym_data,
		       /*verify=*/0, /*verbose=*/0)==FAIL) /* === Some may be able to use verify */
	       return SYM_INVALID;
	    else
	       return sym_data;
	 }
      }
   }
   return SYM_INVALID;
}


PASSFAIL
com_rdmem( const char* args )
{
   static char** cargv = 0;
   static int cargc = 0;
   char** cargv1;
   unsigned long addr;
   BoardID board;
   lb_set_t lbset;
   int n;
   char* endptr;
   int verify = 1;

   if (cargv) { if (cargv[0]) free(cargv[0]); free(cargv); }
   cargv1 = cargv = cli_argv(&cargc, args);

   if (cargc>0 && (!strcmp(cargv1[0], "-volatile") || !strcmp(cargv1[0], "-v")))
   {
      verify = 0;
      cargc--;
      cargv1++;
   }

   if (cargc > 0) lbset = lb_parse(cargv1[0]);
   else
   {
      cli_help("rdmem");
      return FAIL;
   }
   
   for (n=0; n<n_controllers; n++)
   {
      for (board=0; board<MAX_BOARDS; board++)
      {
	 if (LBISSET(&lbset, n, board) && leach[n].name_board[board])
	 {
	    unsigned long data;
	    
	    /*
	     * Look up symbol names *per board* since each board
	     * could have a different address for the same symbol.
	     */
	    if (cargc == 3 && strlen(cargv1[1])==1 &&
		strchr("PXYpxy",cargv1[1][0]))
	    {
	       /*
		* Support old DetI-style memory addressing with white
		* space instead of a colon in the address.  This should
		* not be used.
		*/
	       unsigned long addr_offset;
	       char memtype;

	       printf("warning: `%s %s' is obsolete memory address syntax.\n",
		      cargv1[1], cargv1[2]);
	       memtype = cargv1[1][0];
	       addr_offset = strtoul(cargv1[2], &endptr, 16);
	       if (!endptr || *endptr!='\0')
	       {
		  printf("error: %s is not a valid address\n", cargv1[2]);
		  return FAIL;
	       }
	       addr = addr_offset|LEACH_MEMBASE(memtype);
	    }
	    else if (cargc == 2 && strlen(cargv1[1])>2)
	    {
	       /*
		* Check the address.
		*/
	       addr = sym_to_addr(&leach[n], board, cargv1[1]);
	       if (addr == SYM_INVALID)
	       {
		  printf("error: `%s' is not a valid address\n", cargv1[1]);
		  return FAIL;
	       }
	    }
	    else
	    {
	       cli_help("rdmem");
	       return FAIL;
	    }
	    
#ifdef LIO_16BIT_READ
	    leach[n].has_rm2[board] = 1; /* === */
#endif
	    if (rdmem(&leach[n], board, addr,
		      &data, verify, /*verbose=*/1)==FAIL)
	    {
	       printf("error: Read memory command failed for %s\n",
		      NAMESTR(leach[n].name_board[board]));
	       return FAIL;
	    }
	    else if (data&FLAG16)
	    {
	       printf("%s\t%s\t%c:%06lX = 0x**%04lX ( *%c%c %ld )\n",
		      NAMESTR(leach[n].name_board[board]),
		      cargv1[1], MEMBASE(addr), MEMMASK(addr),
		      data & ~FLAG16,
		      TOCHAR((data&0xFF00)>>8), TOCHAR(data&0x00FF),
		      data & ~FLAG16);
	    }
	    else
	    {
	       printf("%s\t%s\t%c:%06lX = 0x%06lX ( %c%c%c %ld )\n",
		      NAMESTR(leach[n].name_board[board]),
		      cargv1[1], MEMBASE(addr), MEMMASK(addr),
		      data,
		      TOCHAR(data>>16), TOCHAR((data&0xFF00)>>8), TOCHAR(data&0x00FF),
		      data);
	    }
	 }
      }
   }

   return PASS;
}

/* End of leach_mem.c */
