/*                                             -*- c-file-style: "Ellemtel" -*-

`leach_power.c' - Power related commands and diagnostics (pon, voltage).

This file is part of version 3 of the CFHT Detector Host Command Interface.
Read the `License' file for terms of use and distribution.
Copyright 1998-2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "cli/cli.h"
#include "leachcont.h"
#include "detcom.h" /* %%% Why? */

#define VSCALE_36 19.18
#define VSCALE_15 7.64
#define VSCALE_5 2.46
#define VOLTAGE_FORMAT "%6.2f"

static double
vscale(unsigned long dspval, double scale)
{
   return (-3.0 + (dspval*6.0/4096.0))*scale;
}

static unsigned long
vunscale(double val, double scale)
{
   return ( val / scale + 3.0 ) * 4096.0 / 6.0;
}

/*
 * The PASSFAIL return from this routine is a hack:
 * it indicates whether or not the voltage being printed
 * was within the PON tolerance or not.
 */

static PASSFAIL
vprint(const char* mtype_ok, const char* mtype_err, const char* name, double scale,
       unsigned long tar, unsigned long tol, unsigned long pon, unsigned long cur)
{
   const char* mtype = mtype_ok;
   PASSFAIL rtn = PASS;

   tol &= 0xFFFFFF;
   tar &= 0xFFFFFF;
   pon &= 0xFFFFFF;
   cur &= 0xFFFFFF;

   /*
    * Check to see if the DSP code detected a problem at PON
    */
   if (pon != 0xFFFFFF &&
       (pon < tar - tol || pon > tar + tol))
   {
      mtype = mtype_err;
      rtn = FAIL;
   }

   /*
    * Check to see if the current value is still within tolerance
    * (although, the DSP code should trap this and shut off power!)
    */
   if (cur != 0xFFFFFF &&
       ((cur ^ 0x800) < tar - tol || (cur ^ 0x800) > tar + tol))
   {
      mtype = mtype_err;
      rtn = FAIL;
   }

   tar = (tar ^ 0x800) & 0x0FFF;
   pon = (pon ^ 0x800) & 0x0FFF;
   tol = (tol ^ 0x800) & 0x0FFF;

   if (mtype)
   {
      printf("%s: %s controller target="VOLTAGE_FORMAT"+/-%4.2f "
	     "pon="VOLTAGE_FORMAT" current="VOLTAGE_FORMAT"\n",
	     mtype, name,
	     vscale(tar, scale),
	     vscale(tol, scale),
	     vscale(pon, scale),
	     vscale(cur, scale));
   }

   return rtn;
}

PASSFAIL
voltage_check(int n, const char* mtype_ok, const char* mtype_err)
   /*
    * This routine reads the memory locations in the utility boards
    * that are supposed to contain a value representing each of
    * the voltages checked by PON.  If the interrupt routine that
    * updates this memory is not running, the results may be incorrect.
    *
    * mtype_ok and mtype_err are message types to be used to log
    * successes and/or failures.  If either or both are NULL, no]
    * message will be logged.
    *
    * In any case, the two info: lines are updated with each call
    * to this routine, even if there is some kind of failure (in which
    * case the info line is set to "voltages not available").
    */
{
   static char* dsp_version_ = 0;       /* Last known DSP code version IDs */
   static int dsp_support = 1;		/* DSP has support for voltage check */
   unsigned long v36tar=0, v36tol=0, v36pon=0, v36cur=0;
   unsigned long v15ptar=0, v15ptol=0, v15ppon=0, v15pcur=0;
   unsigned long v15mtar=0, v15mtol=0, v15mpon=0, v15mcur=0;
   unsigned long v5cur=0;
   PASSFAIL rtn = PASS;
   const char* lub;

   /*
    * %%% This could print a lot of warnings if you have one DSP
    * with and another without the voltage symbols.
    */
   if (!dsp_version_ || strcmp(dsp_version_, dsp_version()))
   {
      if (dsp_version_) free(dsp_version_);
      dsp_version_ = strdup(dsp_version());
      if (sym_to_addr(&leach[n], LUB, "T_HV")==SYM_INVALID ||
	  sym_to_addr(&leach[n], LUB, "C_P5")==SYM_INVALID)
      {
	 printf("logonly: No DSP support for voltage check.\n");
	 dsp_support = 0;
      }
      else if (dsp_support == 0)
      {
	 printf("logonly: DSP has support for voltage check.\n");
	 dsp_support = 1;
      }
   }

   if (!dsp_support) return FAIL;

   lub = n?"lubS":"lubM";

   v36tar = sym_rdmem(lub, "T_HV");
   if (v36tar == SYM_INVALID)
   {
      if (mtype_err)
	 printf("%s: voltage rdmem failed for %s\n", mtype_err, leach[n].name_contr);
      printf("info%d: %s voltages not available\n", n+3, leach[n].name_contr);
      return FAIL;
   }
   if ((v36tol = sym_rdmem(lub, "K_HV")) == SYM_INVALID ||
       (v36pon = sym_rdmem(lub, "I_HV")) == SYM_INVALID ||
       (v36cur = sym_rdmem(lub, "C_HV")) == SYM_INVALID ||
       (v15ptar = sym_rdmem(lub, "T_P15")) == SYM_INVALID ||
       (v15ptol = sym_rdmem(lub, "K_P15")) == SYM_INVALID ||
       (v15ppon = sym_rdmem(lub, "I_P15")) == SYM_INVALID ||
       (v15pcur = sym_rdmem(lub, "C_P15")) == SYM_INVALID ||
       (v15mtar = sym_rdmem(lub, "T_M15")) == SYM_INVALID ||
       (v15mtol = sym_rdmem(lub, "K_M15")) == SYM_INVALID ||
       (v15mpon = sym_rdmem(lub, "I_M15")) == SYM_INVALID ||
       (v15mcur = sym_rdmem(lub, "C_M15")) == SYM_INVALID ||
       (v5cur = sym_rdmem(lub, "C_P5")) == SYM_INVALID)
   {
      if (mtype_err)
	 printf("%s: Unexpected error while reading voltages for %s\n",
		mtype_err, leach[n].name_contr);
      printf("info%d: %s error reading voltages\n", n+3, leach[n].name_contr);
      return FAIL;
   }

   if (mtype_ok || mtype_err)
   {
      if (vprint(mtype_ok, mtype_err, leach[n].name_contr, VSCALE_36,
		 v36tar, v36tol, v36pon, v36cur)==FAIL)
	 rtn = FAIL;
      if (vprint(mtype_ok, mtype_err, leach[n].name_contr, VSCALE_15,
		 v15ptar, v15ptol, v15ppon, v15pcur)==FAIL)
	 rtn = FAIL;
      if (vprint(mtype_ok, mtype_err, leach[n].name_contr, VSCALE_15,
		 v15mtar, v15mtol, v15mpon, v15mcur)==FAIL)
	 rtn = FAIL;
      if (vprint(mtype_ok, mtype_err, leach[n].name_contr, VSCALE_5,
		 vunscale(5.0, VSCALE_5) ^ 0x800, /* 5V target missing from DSP code */
		 vunscale(0.2, VSCALE_5) ^ 0x800, /* 5V tolerance missing from DSP */
		 0xFFFFFF, /* Hack; 5V at PON missing from DSP too */
		 v5cur)==FAIL)
	 rtn = FAIL;
   }

   printf("info%d: %s\tvoltages: "
	  VOLTAGE_FORMAT" "VOLTAGE_FORMAT" "VOLTAGE_FORMAT" "VOLTAGE_FORMAT"\n",
	  n+3, leach[n].name_contr,
	  vscale(v36cur & 0xFFFFFF, VSCALE_36),
	  vscale(v15pcur & 0xFFFFFF, VSCALE_15),
	  vscale(v15mcur & 0xFFFFFF, VSCALE_15),
	  vscale(v5cur & 0xFFFFFF, VSCALE_5));

   return rtn;
}


PASSFAIL
com_pon( const char* args )
{
   PASSFAIL rtn = PASS;
   lb_set_t lbset;
   int n;

   LBZERO(&lbset);
   if (args && *args) lbset = lb_parse(args); else
   {
      /*
       * Default is to PON all utility boards
       */
      for (n=0; n<n_controllers; n++)
	 LBSET(&lbset, n, LUB);
   }

   for (n=0; n<n_controllers; n++)
   {
      if (LBISSET(&lbset, n, LUB))
      {
	 char cmdbuf[16];

	 sprintf(cmdbuf, "lub%d PON", n);
	 if (com_cmd(cmdbuf)==FAIL)
	 {
	    rtn = FAIL;
	    voltage_check(n, "status", "warning");
	    printf("error: High voltages not enabled at utility board for %s.\n", leach[n].name_contr);
	 }
	 else
	 {
	    if (voltage_check(n, "logonly", "warning")!=PASS)
	    {
	       printf("warning: High voltages enabled at utility board for %s but the "
		      "voltage(s) listed above are marginal and may affect reliability.\n",
		      leach[n].name_contr);
	    }
	    else
	       printf("status: High voltages enabled at utility board for %s.\n",
		      leach[n].name_contr);
	 }
      }
   }
   if (rtn==PASS) camera_state(CAMERA_READY);
   return rtn;
}

PASSFAIL
com_pof( const char* args )
{
   PASSFAIL rtn = PASS;
   lb_set_t lbset;
   int n;

   LBZERO(&lbset);
   if (args && *args) lbset = lb_parse(args); else
   {
      /*
       * Default is to PON all utility boards
       */
      for (n=0; n<n_controllers; n++)
	 LBSET(&lbset, n, LUB);
   }

   for (n=0; n<n_controllers; n++)
   {
      if (LBISSET(&lbset, n, LUB))
      {
	 char cmdbuf[16];

	 sprintf(cmdbuf, "lub%d POF", n);
	 if (com_cmd(cmdbuf)==FAIL)
	 {
	    rtn = FAIL;
	    voltage_check(n, "warning", "warning");
	    printf("error: Failed to disable high voltages for %s.\n", leach[n].name_contr);
	 }
	 else
	 {
	    printf("status: High voltages disabled for %s.\n", leach[n].name_contr);
	    voltage_check(n, "logonly", "logonly");
	 }
      }
   }
   if (rtn==PASS) camera_state(CAMERA_POF);
   return rtn;
}

PASSFAIL
com_power( const char* args )
{
   static char** cargv = 0;	/* Args given on CLI to power ... */
   static int cargc = 0;	/* Argument count */

   lb_set_t lbset;
   int n;
   PASSFAIL rtn = PASS;
   BOOLEAN quiet = FALSE;

   if (args && !strncmp(args, "-q", 2))
   {
      quiet = TRUE;
      args += 2;
      while (*args <= ' ' && *args) args++;
   }

   if (cargv) { if (cargv[0]) free(cargv[0]); free(cargv); }
   cargv = cli_argv(&cargc, args);

   if (cargc == 2)
      lbset = lb_parse(cargv[1]);
   else
   {
      /*
       * Default is to check all utility boards' voltages
       */
      for (n=0; n<n_controllers; n++)
	 LBSET(&lbset, n, LUB);
   }

   if (cargc == 1 || cargc==2)
   {
      if (!xstrcasecmp(cargv[0], "on"))
      {
	 rtn = com_pon(cargc==2?cargv[1]:0);
	 return rtn;
      }
      if (!xstrcasecmp(cargv[0], "off"))
      {
	 rtn = com_pof(cargc==2?cargv[1]:0);
	 return rtn;
      }
      if (!xstrcasecmp(cargv[0], "test"))
      {
	 if (n_controllers > 1)
	 {
	    if (!LBISSET(&lbset, 0, LUB))
	    {
	       return com_cmd("lubS PTS");
	    }
	    else if (!LBISSET(&lbset, 1, LUB))
	    {
	       return com_cmd("lubM PTS");
	    }
	 }
	 rtn = com_cmd("-quiet lub PTS");
	 if (rtn==PASS)
	 {
	    if (camera_state(CAMERA_STATE_QUERY)==CAMERA_POF ||
		camera_state(CAMERA_STATE_QUERY)==CAMERA_UNKNOWN)
	       camera_state(CAMERA_READY);
	 }
	 else
	 {
	    if (!quiet)
	       printf("warning: Power may be off at utility board (try \"verbose on\" for details).\n");
	    camera_state(CAMERA_UNKNOWN);
	 }
	 return rtn;
      }
      else if (cargc==1)
      {
	 lbset = lb_parse(cargv[0]);
      }
   }

   for (n=0; n<n_controllers; n++)
   {
      if (LBISSET(&lbset, n, LUB))
	 if (voltage_check(n, quiet?0:"status", quiet?0:"warning")==FAIL) rtn = FAIL;
   }

   return rtn;
}

/* End of leach_power.c */
