/*                                             -*- c-file-style: "Ellemtel" -*-

`libedt.c' - See libedt.h.

This file is part of version 3 of the CFHT Detector Host Command Interface.
Read the `License' file for terms of use and distribution.
Copyright 1998-2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#ifdef __STDC__
#pragma ident "@(#)libedt.c	1.44 04/23/98 EDT"
#else
    static char SccsId_libedt_c[] = "@(#)libedt.c	1.44 04/23/98 EDT" ;
#endif

/*
 * EDT Library
 *
 * Copyright (c) 1996 by Engineering Design Team, Inc.
 *
 * DESCRIPTION
 *	Provides a 'C' language interface to the EDT SBus DMA cards
 *	to simplify the ring buffer method of reading data.
 *
 *	All routines access a specific device, whose handle is created
 *	and returned by the edt_open() routine.
 *
 *          1.  General purpose functions
 *
 *              edt_open()
 *              edt_close()
 *              edt_read()
 *              edt_write()
 *              edt_ring_buffer_overrun()
 * 		edt_configure_ring_buffers()
 * 		edt_configure_internal_ring_buffers()
 * 		edt_disable_ring_buffers()
 *              edt_set_defaults()
 *
 *          2.  Control functions
 *
 */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include "cli/cli.h"
#include "libedt.h"

static int  debug = 0;

/*
 * edt_open
 *
 * DESCRIPTION
 *	open the named unit and set up the device struct
 *
 * ARGUMENTS
 *	unit:	opens the named unit
 *
 * RETURNS
 *	Pointer to EdtDev struct, or NULL if error
 */
EdtDev         *
edt_open(device_name, unit)
int unit ;
char *device_name ;
{
    int             fd;		/* open file descriptor */
    EdtDev         *edt_p;		/* ptr video device struct */
    static char    *debug_env = NULL;
    char devname[80] ;

    if (debug_env == NULL && ((debug_env = (char *) getenv("EDTDEBUG")) != NULL)
			  && *debug_env != '0')
    {
	debug = atoi(debug_env);
	fprintf(stderr, "edtlib: environment DEBUG set to %d: enabling debug\n", debug);
    }

    if (debug)
	fprintf(stderr, "edt_open(device_name %s, unit %d)\n", device_name, unit) ;

    if ((edt_p = (EdtDev *) calloc(1, sizeof(EdtDev))) == NULL)
	return(NULL) ;

    if (*device_name == '/')
	sprintf(devname, "%s%d", device_name, unit) ;
    else
	sprintf(devname, "/dev/%s%d", device_name, unit) ;

    if ((edt_p->fd = open(devname, O_RDWR)) < 0)
	return NULL;

    if (debug)
	fprintf(stderr, "edt_open for unit %d succeeded\n", unit) ;

    edt_p->donecount = 0 ;

    return edt_p;
}

/*
 * edt_close
 *
 * DESCRIPTION
 *	close the specified device, free the device struct
 *	and memory
 *
 * ARGUMENTS
 *	edt_p:	EDT device struct returned from edt_open()
 *
 * RETURNS
 *	0 on success, -1 on failure
 */
int
edt_close(edt_p)
EdtDev *edt_p ;
{
    int             ret = 0;

    if (debug)
	fprintf(stderr, "edt_close(edt_p 0x%x)\n", edt_p) ;

    if (edt_p->ring_buffers_configured)
	edt_disable_ring_buffers(edt_p);

    ret = close(edt_p->fd);
    free(edt_p);

    return ret;
}

static int initiate_concurrent_reads() ;
static int initiate_concurrent_writes() ;


/*
 * edt_configure_ring_buffers
 *
 * DESCRIPTION
 *	Configure ring buffer mode, allocating all resources.
 *
 * ARGUMENTS
 *	edt_p:	EDT device struct returned from edt_open()
 *
 * RETURNS
 *	0 on success, -1 on failure
 */
int
edt_configure_ring_buffers(edt_p, bufsize, numbufs, write_flag, bufarray)
EdtDev *edt_p ;
int bufsize ;
int numbufs ;
int write_flag ;
void **bufarray ;
{
    int i, j, arg ;
    int result ;

    if (debug && bufarray)
    {
	char **buf = (char **) bufarray ;

	fprintf(stderr, "edt_configure_ring_buffers(edt_p 0x%x, bufsize %d, numbufs %d, write_flag %d, bufarray 0x%x)\n", edt_p, bufsize, numbufs, write_flag, bufarray);

	for (i = 0; i < numbufs; i++)
	{
	    /* Traverse each buffer to trigger Purify if there is a problem */
	    arg = 0 ;
	    for (j = 0; j < bufsize; j++)
		arg += buf[i][j] ;

	    fprintf(stderr,
		"\tbuf[%d] 0x%x firstval 0x%x lastval 0x%x arg 0x%x\n",
			i, buf[i], buf[i][0], buf[i][bufsize-1], arg) ; 
	}
    }

    if (edt_p->ring_buffers_configured)
	edt_disable_ring_buffers(edt_p) ;

    if (bufarray == NULL)
    {
	if ((edt_p->ring_buffers = (void **) malloc(numbufs * sizeof(void *)))
								    == NULL)
	    return(-1) ;

	for (i = 0; i < numbufs ; i++)
	{
	    if (debug) fprintf(stderr, "valloc buffer size %d \n", bufsize) ;

	    if ((edt_p->ring_buffers[i] = (void *) valloc(bufsize)) == NULL)
		return(-1) ;
	    if (debug) fprintf(stderr, "size %d ring_buffer[%d] 0x%x\n",
					bufsize, i, edt_p->ring_buffers[i]) ;
	}
	puts("") ;

	edt_p->ring_buffers_allocated = 1 ;
    }
    else
    {
	edt_p->ring_buffers = bufarray ;
	edt_p->ring_buffers_allocated = 0 ;
    }

    edt_p->ring_buffer_count = numbufs ;
    edt_p->ring_buffer_size  = bufsize ;
    edt_p->donecount  	     = 0 ;
    edt_p->nextwbuf  	     = 0 ;
    edt_p->nextbuf  	     = 0 ;

    if (write_flag)
	result = initiate_concurrent_writes(edt_p) ;
    else
	result = initiate_concurrent_reads(edt_p) ;

    if (result == numbufs)
    {
	edt_p->ring_buffers_configured = 1 ;
	return 0 ;
    }
    else
    {
	edt_p->ring_buffers_configured = 0 ;
	fprintf(stderr, "Wanted %d buffers, got %d\n", numbufs, result) ;
	return -1 ;
    }
}

#ifdef EDTS_IBUF_MODE

/*
 * edt_configure_internal_ring_buffers
 *
 * DESCRIPTION
 *	Configure internal ring buffer mode, allocating all resources.
 *
 * ARGUMENTS
 *	edt_p:		EDT device struct returned from edt_open()
 *	bufsize: 	Size of individual ring buffers
 *	numbufs: 	Number of individual ring buffers
 *
 * RETURNS
 *	0 on success, -1 on failure
 */
int
edt_configure_internal_ring_buffers(edt_p, bufsize, numbufs)
EdtDev *edt_p ;
int bufsize ;
int numbufs ;
{
    int i, arg ;
    int enable = 1 ;
    char *data ;
    int  lockstep = 0 ;

    if (edt_p->ring_buffers_configured)
	edt_disable_ring_buffers(edt_p) ;

    edt_p->ring_buffer_count = numbufs ;
    edt_p->ring_buffer_size  = bufsize ;

    edt_p->donecount = 0 ;
    ioctl(edt_p->fd, EDT_CLEAR_DONECOUNT) ;
    ioctl(edt_p->fd, EDTS_LOCKSTEP, &lockstep);

    if ((edt_p->ring_buffers = (void **)malloc(numbufs*sizeof(void *))) == NULL)
	return(-1) ;

    if (ioctl(edt_p->fd, EDTS_IBUF_SIZE, &bufsize) == -1)
	return -1 ;

    if (ioctl(edt_p->fd, EDTS_IBUF_NUM, &numbufs) == -1)
	return -1 ;

    if (ioctl(edt_p->fd, EDTS_IBUF_MODE, &enable) == -1)
	return -1 ;

    /* Map in the driver resident ring buffers */
    data = mmap((caddr_t)0, bufsize * numbufs, PROT_READ|PROT_WRITE, MAP_SHARED,
							 edt_p->fd, IBUF_BASE) ;
    for (i = 0; i < numbufs ; i++)
	edt_p->ring_buffers[i] = (char *) (data + (i * bufsize)) ;

    edt_p->ring_buffers_configured = 1 ;
    return 0 ;

}
#endif /* EDTS_IBUF_MODE */


#ifdef SYSV
void *th_share_ring_buffers(arg)
void *arg ;
{
    EdtDev *edt_p = (EdtDev *) arg ;

    /* Will block until sharing is done */
    ioctl(edt_p->fd, EDT_SHARE_RING_BUFFERS, &edt_p->write_flag) ;

    thr_exit(NULL) ;
}


int
edt_share_ring_buffers(edt_p, bufsize, numbufs, write_flag, bufarray)
EdtDev *edt_p ;
int bufsize ;
int numbufs ;
int write_flag ;
void **bufarray ;
{
    int i, arg ;
    int result ;
    int lockstep = 0 ;

    if (edt_p->ring_buffers_configured)
	edt_disable_ring_buffers(edt_p) ;

    edt_p->ring_buffers = bufarray ;
    edt_p->ring_buffers_allocated = 0 ;

    edt_p->ring_buffer_count = numbufs ;
    edt_p->ring_buffer_size  = bufsize ;
    edt_p->write_flag	     = write_flag ;

    edt_p->donecount = 0 ;
    ioctl(edt_p->fd, EDT_CLEAR_DONECOUNT) ;
    ioctl(edt_p->fd, EDTS_LOCKSTEP, &lockstep);

    /* THR_BOUND is required here so that thr_jone releases thread memory */
    thr_create(NULL,NULL,th_share_ring_buffers, (void *)edt_p, THR_BOUND, NULL);
    ioctl(edt_p->fd, EDT_WAITRDY) ;

    if (debug)
    {
	int i ;

	fprintf(stderr,
"\nshare_ring_buffers: bsize %d nbufs %d ring_buffers 0x%x\n",
       edt_p->ring_buffer_size, edt_p->ring_buffer_count, edt_p->ring_buffers) ;

	for (i = 0; i < numbufs; i++)
	    printf("buf[%d]:  0x%x\n", i, edt_p->ring_buffers[i]) ;
    }

    edt_p->ring_buffers_configured = 2 ;  /* 2 means shared ring buffers */
    return 0 ;
}
#endif /* SYSV */

/*
 * edt_disable_ring_buffers
 *
 * DESCRIPTION
 *	Disable ring buffer mode, freeing all resources.
 *
 * ARGUMENTS
 *	edt_p:	EDT device struct returned from edt_open()
 *
 * RETURNS
 *	0 on success, -1 on failure
 */
int
edt_disable_ring_buffers(edt_p)
EdtDev *edt_p ;
{
    int i ;
    int mode ;

    if (debug)
	fprintf(stderr, "edt_disable_ring_buffers(edt_p 0x%x)\n", edt_p) ;

    ioctl(edt_p->fd, EDT_FREE_BUF) ;

#ifdef EDTS_IBUF_MODE

    if (ioctl(edt_p->fd, EDTG_IBUF_MODE, &mode) != -1)
    {
	if (mode)
	{
	    int numbufs = edt_p->ring_buffer_count ;
	    int bufsize = edt_p->ring_buffer_size ;

	    munmap(edt_p->ring_buffers[0], numbufs * bufsize) ;
	    free (edt_p->ring_buffers) ;

	    mode = 0 ;
	    ioctl(edt_p->fd, EDTS_IBUF_MODE, &mode) ;

	    edt_p->ring_buffers_configured = 0 ;
	    return 0 ;
	}
    }

#endif /* EDTS_IBUF_MODE */

#ifdef SYSV
    if (edt_p->ring_buffers_configured == 2)
    {
	ioctl(edt_p->fd, EDT_FREE_BUF) ;

	if (edt_p->ring_buffer_tids)
	{
	    thread_t *tids = (thread_t *) edt_p->ring_buffer_tids ;
	    thr_join(tids[0], NULL, NULL) ;
	    free(edt_p->ring_buffer_tids) ;
	}
    }
    else
#endif
    {
	/* Wait for all outstanding I/O threads to finish */
	if (edt_p->ring_buffer_tids)
	{
	    for(i = 0 ; i < edt_p->ring_buffer_count ; i++)
	    {
#ifdef SYSV
		thread_t *tids = (thread_t *) edt_p->ring_buffer_tids ;
		thr_join(tids[i], NULL, NULL) ;
#else
		(void) aiowait(NULL) ;
#endif
	    }

	    free(edt_p->ring_buffer_tids) ;
	}

	if (edt_p->ring_buffers_allocated)
	{
	    for(i = 0 ; i < edt_p->ring_buffer_count ; i++)
		if (edt_p->ring_buffers[i])
		    free(edt_p->ring_buffers[i]) ;
		else
		    return -1 ;

	    if (edt_p->ring_buffers)
		free(edt_p->ring_buffers) ;
	    else
		return -1 ;
	}
    }
    edt_p->ring_buffers_configured = 0 ;
    edt_p->ring_buffers_allocated = 0 ;

    return 0 ;
}

/*
 * edt_read
 *
 * DESCRIPTION
 *      Perform a UNIX read() on the edt. The UNIX 2GB file offset bug
 *      is avoided during large amounts of I/O, i.e. reading past
 *      2^31 will not fail.
 *
 * ARGUMENTS
 *      edt_p:  EDT device handle returned from edt_open()
 *      buf:    address of buffer to read into
 *      size:   size of read in bytes
 *
 * RETURNS
 *      returns the return value from read;  errno is set by read() on error
 *
 * NOTE
 *      This implementation may change.  Use edt_read() over using
 *      UNIX read().
 */
int
edt_read(edt_p, buf, size)
EdtDev *edt_p ;
void *buf ;
int size ;
{
    static u_int b_count = 0 ;

    if (debug)
	fprintf(stderr, "edt_read(edt_p 0x%x, buf 0x%x, size %d)\n", edt_p, buf, size) ;
 
    b_count += size ;
    if (b_count > 0x40000000) /* 1GB */
    {
        lseek(edt_p->fd, 0, 0) ;
        b_count = 0 ;
    }
    return read(edt_p->fd, buf, size) ;
}

/*
 * edt_write
 *
 * DESCRIPTION
 *      Perform a UNIX write() on the edt. The UNIX 2GB file offset bug
 *      is avoided during large amounts of I/O, i.e. writing past
 *      2^31 will not fail.
 *
 * ARGUMENTS
 *      edt_p:  EDT device handle returned from edt_open()
 *      buf:    address of buffer to write into
 *      size:   size of write in bytes
 *
 * RETURNS
 *      returns the return value from write;  errno is set by write() on error
 *
 * NOTE
 *      This implementation may change.  Use edt_write() over using
 *      UNIX write().
 */
int
edt_write(edt_p, buf, size)
EdtDev *edt_p ;
void *buf ;
int size ;
{
    static u_int b_count = 0 ;

    if (debug)
	fprintf(stderr, "edt_write(edt_p 0x%x, buf 0x%x, size %d)\n", edt_p, buf, size) ;
 
    b_count += size ;
    if (b_count > 0x40000000) /* 1GB */
    {
        lseek(edt_p->fd, 0, 0) ;
        b_count = 0 ;
    }
    return write(edt_p->fd, buf, size) ;
}


/* Returns address of next available sequential _write_ buffer, for
 * filling in with data when sending with edt_start_write for example
 */
void *
edt_next_writebuf(edt_p)
EdtDev *edt_p ;
{
    caddr_t retval;

    if (debug)
	fprintf(stderr, "edt_next_writebuf(edt_p 0x%x) ", edt_p) ;

    retval = edt_p->ring_buffers[edt_p->nextwbuf];
    edt_p->nextwbuf = ++edt_p->nextwbuf % edt_p->ring_buffer_count ;

    if (debug)
	fprintf(stderr, "returns 0x%x\n", retval) ;

    return retval;
}

/* Returns next sequential buffer */
static caddr_t
edt_next_buffer_after_dma(edt_p)
EdtDev *edt_p ;
{
    int      	    donecount;
    caddr_t	    retval ;

    retval = edt_p->ring_buffers[edt_p->nextbuf] ;
    edt_p->nextbuf = (edt_p->nextbuf + 1) % edt_p->nbufs ;
    return retval ;
}

void **
edt_buffer_addresses(edt_p)
EdtDev         *edt_p;
{
    return(edt_p->ring_buffers) ;
}

void *
edt_wait_for_buffers(edt_p, count)
EdtDev         *edt_p;
int count ;
{
    int bufnum ;

    if (debug)
	fprintf(stderr, "edt_wait_for_buffers(edt_p 0x%x, count %d)\n", edt_p, count) ;

    count += edt_p->donecount ;
    bufnum = (count - 1) % edt_p->ring_buffer_count ;
    if(debug) fprintf(stderr, "Wait for buffer %d ring buffer %d\n", count, bufnum) ;
    ioctl(edt_p->fd, EDTS_WAKEUP_DONECOUNT, &count);
    edt_p->donecount = count  ;
    return (edt_p->ring_buffers) ? edt_p->ring_buffers[bufnum] : 0 ;
}


void *
edt_wait_for_next_buffer(edt_p)
EdtDev         *edt_p;
{
    u_int count ;
    u_int bufnum ;

    ioctl(edt_p->fd, EDTG_DONECOUNT, &count);
    count++ ;
    bufnum = (count - 1) % edt_p->ring_buffer_count ;

    ioctl(edt_p->fd, EDTS_WAKEUP_DONECOUNT, &count);
    edt_p->donecount = count  ;
    return (edt_p->ring_buffers) ? edt_p->ring_buffers[bufnum] : 0 ;
}


void *
edt_check_for_buffers(edt_p, count)
EdtDev         *edt_p;
int count ;
{
    u_int driver_count ;

    if (debug)
	fprintf(stderr, "edt_check_for_buffers(edt_p 0x%x, count %d)\n", edt_p, count) ;

    count += edt_p->donecount ;
    ioctl(edt_p->fd, EDTG_DONECOUNT, &driver_count);
    if (driver_count >= count)
    {
	edt_p->donecount = count  ;
	return(edt_p->ring_buffers[(count - 1) % edt_p->ring_buffer_count]) ;
    }
    else
	return(NULL) ;
}


/*
 * edt_ring_buffer_overrun
 *
 * DESCRIPTION
 *	Detect the condition where DMA has outpaced application buffer
 *	processing and has circled around the ring buffer and caught
 *	up with and overwritten the buffer the application is working on.
 *
 *	This test does not detect past overruns, so must be called
 *	each time a new buffer is ready for processin.
 *
 * ARGUMENTS
 *	edt_p:	EDT device struct returned from edt_open()
 *
 * RETURNS
 *	True if an overrun condition is present, false otherwise.
 */
int
edt_ring_buffer_overrun(edt_p)
EdtDev         *edt_p;
{
    int dmacount ;

    ioctl(edt_p->fd, EDTG_DONECOUNT, &dmacount);

    if (dmacount >= edt_p->donecount + edt_p->ring_buffer_count)
	return(1) ;
    else
	return(0) ;
}

int
edt_start_buffers(edt_p, count)
EdtDev *edt_p ;
int count ;
{
    u_int lockstep ;

    if (debug)
	fprintf(stderr, "edt_start_buffers(edt_p 0x%x, count %d)\n", edt_p, count) ;

    if (count == 0)
	return ioctl(edt_p->fd, EDT_LOCKOFF) ;
    else
    {
	ioctl(edt_p->fd, EDTG_LOCKSTEP, &lockstep);
	if (lockstep)
	    return ioctl(edt_p->fd, EDTS_APPBUFS_COMPLETED, &count);
	else
	    return ioctl(edt_p->fd, EDTS_LOCKSTEP, &count);
    }
}

int
edt_stop_buffers(edt_p)
EdtDev *edt_p ;
{
    u_int lockstep = 0 ;

    ioctl(edt_p->fd, EDTS_LOCKSTEP, &lockstep);
}

int
edt_done_count(edt_p)
EdtDev         *edt_p;
{
    int             donecount;
    ioctl(edt_p->fd, EDTG_DEVBUFS_COMPLETED, &donecount);
    return(donecount) ;
}

edt_timeout(edt_p, val)
EdtDev *edt_p;
int val;
{
    u_short arg = val ;	 /* hundreths of seconds */
    int ret ;

    ret = ioctl(edt_p->fd, EDTS_RTIMEOUT, &arg);

    return(ret) ;
}


/*
 * following are internal support routines 
 */

/* 
 * set throttle and dmasize - not ususally needed
 */
int 
edt_throttle(edt_p, throttle) 
EdtDev *edt_p ;
u_int throttle ;
{
    return ioctl(edt_p->fd, EDTS_DMATHROT, &throttle);
}

int
edt_dma_chunk_size(edt_p, size) 
EdtDev *edt_p ;
u_int size ;
{
    return ioctl(edt_p->fd, EDTS_MAXDMASIZE, &size);
}

#ifdef SYSV

static void *
read_thread(arg)
void *arg;
{
    EdtDev *edt_p = (EdtDev *) arg;
    int k, ret ;

    mutex_lock(&edt_p->bufnum_mutex) ;
    k = edt_p->bufnum++ ;
    if (edt_p->bufnum == edt_p->ring_buffer_count)
	sema_post(&edt_p->bufnum_sema) ;
    mutex_unlock(&edt_p->bufnum_mutex) ;

    /*
     * Start a read which will block forever in this thread. aioread()
     * does the same thing but won't allow as much concurrency.  Lseek
     * orders the buffers inside the driver.
     */
    if(debug) fprintf(stderr, "start read for %d addr %x\n", k, edt_p->ring_buffers[k]) ;
    if (pread(edt_p->fd, edt_p->ring_buffers[k], edt_p->ring_buffer_size, k*512)
									  == -1)
    {
	if (debug)
	{
	    char errstr[128] ;
	    sprintf(errstr,"pread buffer %d at 0x%x",k,edt_p->ring_buffers[k]) ;
	    perror(errstr) ;
	}
	edt_p->ring_buffer_setup_error = 1 ;
    }
    if(debug) fprintf(stderr, "end read for %d\n", k) ;

    return(0) ;
}

static void *
write_thread(arg)
void *arg;
{
    EdtDev *edt_p = (EdtDev *) arg;
    int k ;

    mutex_lock(&edt_p->bufnum_mutex) ;
    k = edt_p->bufnum++ ;
    if (edt_p->bufnum == edt_p->ring_buffer_count)
	sema_post(&edt_p->bufnum_sema) ;
    mutex_unlock(&edt_p->bufnum_mutex) ;

    /*
     * Start a write which will block forever in this thread. aiowrite()
     * does the same thing but won't allow as much concurrency.  Lseek
     * orders the buffers inside the driver.
     */
    if(debug) fprintf(stderr, "start write for %d addr %x\n", k,
					 edt_p->ring_buffers[k]) ;
    if (pwrite(edt_p->fd, edt_p->ring_buffers[k], edt_p->ring_buffer_size,k*512)
									  == -1)
    {
	if (debug)
	{
	    char errstr[128] ;
	    sprintf(errstr,"pwrite buffer %d at 0x%x",k,edt_p->ring_buffers[k]);
	    perror(errstr) ;
	}
	edt_p->ring_buffer_setup_error = 1 ;
    }
    if(debug) fprintf(stderr, "end write for %d\n", k) ;

    return(0) ;
}

#endif

static int
initiate_concurrent_reads(edt_p)
EdtDev         *edt_p;
{
    int             i;
    int             done = 0;
    int		    lockstep = 0 ;

#ifdef SYSV
    edt_p->ring_buffer_tids = 
	    (thread_t *) malloc(sizeof(thread_t) * edt_p->ring_buffer_count);
    mutex_init(&edt_p->bufnum_mutex, NULL, NULL) ;
    sema_init(&edt_p->bufnum_sema, 0, USYNC_THREAD, NULL);
#else
    aio_result_t   *aioresult ;
    edt_p->ring_buffer_tids = (void *) aioresult =
	(aio_result_t *)malloc(sizeof(aio_result_t) * edt_p->ring_buffer_count);
#endif
    edt_p->bufnum = 0 ;

    edt_p->donecount = 0 ;
    ioctl(edt_p->fd, EDT_CLEAR_DONECOUNT) ;
    ioctl(edt_p->fd, EDTS_LOCKSTEP, &lockstep);
    ioctl(edt_p->fd, EDTS_NBUFIO, &edt_p->ring_buffer_count);
    edt_p->ring_buffer_setup_error = 0 ;


    for (i = 0; i < edt_p->ring_buffer_count; i++)
    {
#ifdef SYSV
	thread_t *tids = (thread_t *) edt_p->ring_buffer_tids ;
	/* THR_BOUND is required here so that thr_jone releases thread memory */
	if (thr_create(NULL, NULL, read_thread, edt_p, THR_BOUND, &tids[i]))
	    perror("thr_create");
#else
	if (aioread(edt_p->fd, edt_p->ring_buffers[i], edt_p->ring_buffer_size,
					      i * 512, 0, &aioresult[i]) == -1)
	    perror("Continuous aioread");
	else
	    edt_p->bufnum++ ;
#endif
    }
    /* puts("") ; === Why did they do this?  Thanks EDT. */

#ifdef SYSV
    /* Wait until all the threads are started */
    sema_wait(&edt_p->bufnum_sema) ;
#endif

    {
      int rtn;
      struct sigaction oldsigalrm;
      
      printf("debug: [libedt] Waiting for EDT card to be ready\n");
      sigaction(SIGALRM, 0, &oldsigalrm);
      cli_signal(SIGALRM, CLI_SIG_TRAP);
      alarm(30);
      rtn = ioctl(edt_p->fd, EDT_WAITRDY);
      alarm(0);
      if (cli_signal_event(SIGALRM))
	printf("warning: [libedt] Problem with EDT driver/card; Timeout during EDT_WAITRDY!\n");
      sigaction(SIGALRM, &oldsigalrm, 0);
      if (rtn != 0) printf("warning: [libedt] Problem with EDT driver/card; EDT_WAITRDY failed!\n");
      else printf("debug: [libedt] EDT card is ready for DMA\n");
    }

#ifdef SYSV
    mutex_destroy(&edt_p->bufnum_mutex) ;
    sema_destroy(&edt_p->bufnum_sema) ;
#endif
    return (edt_p->bufnum);
}


static int
initiate_concurrent_writes(edt_p)
EdtDev         *edt_p;
{
    int             i;
    int             done = 0;
    int		    lockstep = 0 ;

#ifdef SYSV
    edt_p->ring_buffer_tids = 
	    (thread_t *) malloc(sizeof(thread_t) * edt_p->ring_buffer_count);
    mutex_init(&edt_p->bufnum_mutex, NULL, NULL) ;
    sema_init(&edt_p->bufnum_sema, 0, USYNC_THREAD, NULL);
#else
    aio_result_t   *aioresult ;
    edt_p->ring_buffer_tids = (void *) aioresult =
	(aio_result_t *)malloc(sizeof(aio_result_t) * edt_p->ring_buffer_count);
#endif
    edt_p->bufnum = 0 ;

    edt_p->donecount = 0 ;
    ioctl(edt_p->fd, EDT_CLEAR_DONECOUNT) ;
    ioctl(edt_p->fd, EDTS_LOCKSTEP, &lockstep);
    ioctl(edt_p->fd, EDTS_NBUFIO, &edt_p->ring_buffer_count);
    edt_p->ring_buffer_setup_error = 0 ;


    for (i = 0; i < edt_p->ring_buffer_count; i++)
    {
#ifdef SYSV
	thread_t *tids = (thread_t *) edt_p->ring_buffer_tids ;
	/* THR_BOUND is required here so that thr_jone releases thread memory */
	if (thr_create(NULL, NULL, write_thread, edt_p, THR_BOUND, &tids[i]))
	    perror("thr_create");
#else
	if (aiowrite(edt_p->fd, edt_p->ring_buffers[i], edt_p->ring_buffer_size,
					       i * 512, 0, &aioresult[i]) == -1)
	    perror("Continuous aiowrite");
	else
	    edt_p->bufnum++ ;
#endif
    }

#ifdef SYSV
    /* Wait until all the threads are started */
    sema_wait(&edt_p->bufnum_sema) ;
#endif

    ioctl(edt_p->fd, EDT_WAITRDY) ;

#ifdef SYSV
    mutex_destroy(&edt_p->bufnum_mutex) ;
    sema_destroy(&edt_p->bufnum_sema) ;
#endif
    return (edt_p->bufnum);
}

int
edt_byte_count(edt_p)
EdtDev *edt_p;
{
    int count ;
    ioctl(edt_p->fd, EDTG_BYTECOUNT, &count);
    if (debug)
	fprintf(stderr, "edtlib: count %d\n", count);

    return(count) ;
}

/* just for convienence */

#ifndef _rasterfile_h
/* from 4.1.x rasterfile.h */
struct rasterfile
{
    int             ras_magic;	/* magic number */
    int             ras_width;	/* width (pixels) of image */
    int             ras_height;	/* height (pixels) of image */
    int             ras_depth;	/* depth (1, 8, or 24 bits) of pixel */
    int             ras_length;	/* length (bytes) of image */
    int             ras_type;	/* type of file; see RT_* below */
    int             ras_maptype;/* type of colormap; see RMT_* below */
    int             ras_maplength;	/* length (bytes) of following map */
    /* color map follows for ras_maplength bytes, followed by image */
};

#define RAS_MAGIC       0x59a66a95
#define RT_STANDARD     1	/* Raw pixrect image in 68000 byte order */
#define RMT_EQUAL_RGB   1	/* red[ras_maplength/3],green[],blue[] */
#define RMT_NONE	0
#endif

/*
 * edt_write_rasfile
 *
 * DESCRIPTION
 *      output a 1 band, 8 bit image to a sun raster format file
 *
 * ARGUMENTS
 *      fname		the name of the output file
 *      addr		the address of the image data (8 bits per pixel)
 *      x_size		width in pixels of image
 *      y_size		height in pixels of image
 *
 * RETURNS
 *	 0 on success, -1 on failure
 *
 */
edt_write_rasfile(fname, addr, x_size, y_size)
char           *fname;
caddr_t         addr;
int             x_size;
int             y_size;
{
    struct rasterfile ras;
    FILE           *fp;
    int             i;
    int		    rowsize ;
    int		    rows ;

    if ((fp = fopen(fname, "w")) == NULL)
    {
	printf("edtlib: can't open destination data file %s\n", fname);
	return -1;
    }

    rowsize = x_size ;
    if(rowsize & 1)
    {
	rowsize++ ;
    }
    ras.ras_magic = RAS_MAGIC;
    ras.ras_width = x_size;
    ras.ras_height = y_size;
    ras.ras_depth = 8;
    ras.ras_length = x_size * y_size;
    ras.ras_type = RT_STANDARD;
    ras.ras_maptype = RMT_NONE ;
    ras.ras_maplength = 0 ;

    fwrite(&ras, sizeof(ras), 1, fp);


    if (debug) fprintf(stderr, "before write %d %d\n",x_size,y_size*3) ;
    for(i = 0; i < y_size ; i++)
    {
	fwrite(addr, rowsize, 1, fp);
	addr += x_size ;
    }
    fclose(fp);
    return(0) ;
}

/*
 * edt_write_rasfile24
 *
 * DESCRIPTION
 *      output a 3 band, 8 bit image to a sun raster format file
 *
 * ARGUMENTS
 *       fname		the name of the output file
 *       addr		the address of the image data (24 bits per pixel)
 *       x_size		width in pixels of image
 *       y_size		height in pixels of image
 *
 * RETURNS
 *	 0 on success, -1 on failure
 *
 */
edt_write_rasfile24(fname, addr, x_size, y_size)
char           *fname;
caddr_t         addr;
int             x_size;
int             y_size;
{
    struct rasterfile ras;
    FILE           *fp;
    int             i;
    int		    rowsize ;
    int		    rows ;
    u_char	    *im0 ;
    u_char	    *im1 ;
    int		    x, y ;
    int		    srcline ;
    u_char *srcptr1, *srcptr2 ;
    int 	    ret ;

    if ((fp = fopen(fname, "w")) == NULL)
    {
	printf("edtlib: can't open destination data file %s\n", fname);
	return (-1);
    }

    rowsize = x_size ;
    if(rowsize & 1)
    {
	rowsize++ ;
    }
    ras.ras_magic = RAS_MAGIC;
    ras.ras_width = x_size;
    ras.ras_height = y_size;
    ras.ras_depth = 24;
    ras.ras_length = x_size * y_size;
    ras.ras_type = RT_STANDARD;
    ras.ras_maptype = RMT_NONE ;
    ras.ras_maplength = 0 ;

    fwrite(&ras, sizeof(ras), 1, fp);
    if (debug) fprintf(stderr, "before write %d %d\n",x_size,y_size*3) ;
    ret = fwrite(addr, x_size, y_size*3, fp);
    if(debug) fprintf(stderr, "done ret %d\n",ret) ;
    fclose(fp);
    return(0) ;
}

#ifdef SYSV
/*
 * edt_cache_flush
 *
 * DESCRIPTION
 *      Perform a dma flush to synchronize hardware caches on a ring buffer.
 *
 * ARGUMENTS
 *	edt_p		edt handle returned from edt_open()
 *	buf		pointer to a ring buffer element
 *
 * RETURNS
 *	 0 on success, -1 on failure
 *
 */
edt_cache_flush(edt_p, buf)
EdtDev *edt_p;
void   *buf ;
{
    int i ;

    for (i = 0; i < edt_p->ring_buffer_count; i++)
	if (edt_p->ring_buffers[i] == (caddr_t) buf)
	    break ;

    return(ioctl(edt_p->fd, EDTS_SYNC_RING_BUFFER, &i)) ;
}
#endif

int
edt_abort_dma(edt_p)
EdtDev *edt_p ;
{
    if (ioctl(edt_p->fd, EDT_ABORT_DMA) == -1)
	return -1 ;
    return 0 ;
}


int
edt_set_buffer(edt_p, bufnum)
EdtDev *edt_p ;
int bufnum ;
{
    if (bufnum >= edt_p->ring_buffer_count)
    {
	fputs("edt_set_buffer:	invalid	parameter\n", stderr) ;
	return 1 ;
    }

    ioctl(edt_p->fd, EDT_SET_BUFFER, &bufnum) ;

    edt_p->donecount = bufnum ;
    return 0 ;
}


int
edt_abort_current_dma(edt_p)
EdtDev *edt_p ;
{
    if (ioctl(edt_p->fd, EDT_ABORT_CURRENT) == -1)
	return -1 ;
    return 0 ;
}

int
edt_set_eodma_int(edt_p, signum)
EdtDev *edt_p;
u_short signum;
{
    if (ioctl(edt_p->fd, EDTS_EODMA_SIG, &signum) == -1)
	return -1 ;
    return 0 ;
}


#ifndef SYSV
microsleep(val)
int		val;
{
    usleep(val);
    return(0) ;
}

#else
microsleep(val)
int		val;
{
    struct timeval  t;
    fd_set fd1,	fd2, fd3 ;

    FD_ZERO(&fd1);
    FD_ZERO(&fd2);
    FD_ZERO(&fd3);

    t.tv_sec = val / 1000000;
    val	-= t.tv_sec * 1000000;
    t.tv_usec =	val;
    select(0, &fd1, &fd2, &fd3,	&t);
    return(0) ;
}

#endif

