/*                                             -*- c-file-style: "Ellemtel" -*-

`lio.h' - Leach controller low level input/output driver interface.

This file is part of version 3 of the CFHT Detector Host Command Interface.
Read the `License' file for terms of use and distribution.
Copyright 1998-2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#ifndef _INCLUDED_lio
#define _INCLUDED_lio 1

/*
 * These return codes are used by most of the lio_*() calls which
 * return an integer.  detcom only needs to make a distinction between
 * ERR/ABORTED in the case of a readout with lio_read16_parallel().
 */
#define LIO_OK 0       /* Call completed successfully */
#define LIO_ERR -1     /* An error occurred */
#define LIO_ABORTED -2 /* User abort */

/*
 * Routines leach_*.c use these calls to communicate with the hardware
 * layer connected to the SDSU controller.  "lio" stands for
 * "leach input/output"
 */

typedef void* Devp;
/*
 * Each of the following routines use the generic pointer to keep track
 * of open Leach devices.  Inside lio_*.c this pointer is cast to point
 * to some implementation-specific structure that is useful to the module.
 * To the outside world, it simply functions as a file handle.
 */

typedef void (*read16_handler)(int n,unsigned short* buf,unsigned int sz);
/*
 * read16_handler is a function type that expects to be passed raw,
 * scrambled buffers of 16-bit data which it will process.  Processing
 * typically consists of descrambling. `n' is the controller number
 * and `sz' is the number of pixels passed in `buf'.  The parallel
 * read call will invoke read16_handler some number of times as data
 * becomes available.
 *
 * A read16_handler is provided by detcom.  lio_xxx.c only needs to know
 * how to pass it raw data correctly.
 */

Devp lio_open(const char* fn);
/*
 * DESCRIPTION
 *
 * Opens a controller device file.  If and "fn" is given, a particular
 * device file should be opened.  If fn==NULL then this call should
 * search a list of implementation-specific device names that are
 * appropriate for the driver and use the next available one.
 *
 * RETURN VALUE
 *
 * Returns a Devp pointer to be used as a file handle, passed as the
 * first argument to all the other lio_* calls.  If an error occurrs,
 * lio_open should print an error message directly to STDOUT or STDERR
 * and return NULL.
 *
 * ERRORS
 *
 * Implementation specific.  Generally this call will need to call
 * open() internally, and should do a perror("error: open failed");
 * if that call fails and then return NULL.
 */

int lio_close(Devp devp);
/*
 * DESCRIPTION
 *
 * This closes the device previously returned by lio_open().
 *
 * RETURN VALUE
 *
 * The return code is the same as the return code from close().
 * A LIO_OK (0) is returned on success, and a LIO_ERR (-1) if an error
 * occurs.
 *
 * ERRORS
 *
 * Implementation specific, but usually the same as those from close().
 */

int lio_check(Devp devp);
/*
 * DESCRIPTION
 *
 * Tests the hardware connection between the host computer and the detector.
 * If you don't want to write one of these, just make a function that
 * silently returns LIO_OK.  If this function detects a problem, it should
 * output a message directly to the console explaining the problem.
 *
 * RETURN VALUE
 *
 * Returns LIO_OK on success, or LIO_ERR on error and prints some diagnostics.
 *
 * ERRORS
 *
 * Implementation specific.  In the case of the fiber parallel box,
 * this tests if the cables are properly plugged in, and warns about
 * this specific problem if they appear to be unplugged or if the
 * box is not powered up.
 */

int lio_reset(Devp devp, int sdsu_gen);
/*
 * DESCRIPTION
 *
 * Sets proper operating paramters for the hardware driver and sends and
 * initializes the interface.  Write a function that returns 0 if this
 * is not applicable.  The sdsu_gen parameter must either be a 1 or a 2.
 * Any other value, or a value not supported by the hardware must return
 * LIO_ERR.
 *
 * RETURN VALUE
 *
 * Returns LIO_OK on success, LIO_ERR on error or unsupported sdsu_gen.
 *
 * ERRORS
 *
 * Implementation specific.  If a reset fails, lio_reset should itself
 * print a message describing the problem when returning LIO_ERR.
 */

int lio_reset_peer(Devp devp);
/*
 * DESCRIPTION
 *
 * Attempt to reset the controller.  This is done with a special
 * driver ioctl() (version 1.6 or higher of lotupci required.)
 *
 * RETURN VALUE
 *
 * Returns LIO_OK, or LIO_ERR on error.
 *
 * ERRORS
 *
 * Currently, none are detected.  The real test if the reset worked
 * will come by doing a TDL or other higher level-command.
 */

int lio_write24(Devp devp, unsigned long* data, int nwords);
/*
 * DESCRIPTION
 *
 * Send 24-bit data to the lio (packed as an array of unsigned long).
 * For lio_fpbox only:
 *   This data is actually sent as 2 16-bit words which the lio will
 *   re-assemble, but higher level code need not be concerned with this.
 *
 * RETURN VALUE
 *
 * Returns the number of words actually written (should always be
 * the same as nwords) or LIO_ERR in case of an error.
 *
 * ERRORS
 *
 * Usually the same as write(), but lio_write24() should print its
 * own message describing and failures when returning a LIO_ERR.
 */

/*
 * Some defines used by detcom to figure out the data size for
 * responses to leach commands.  This does not include pixels:
 * They are always read (fast) as 16-bit data.
 */
#ifndef LIO_16BIT_READ
/*
 * This is the normal case, where we read "unsigned long" data.
 */
#define LREAD unsigned long
#define LREAD_MASK(a) ((a)&0x00FFFFFF)
#define LIO_LREAD lio_read24
#else
/*
 * This is the crippled case for the UH fpbox where we read "unsigned short"
 */
#define LREAD unsigned short
#define LREAD_MASK(a) ((a)&0x0000FFFF)
#define LIO_LREAD lio_read16
#endif

#ifndef LIO_16BIT_READ
int lio_read24(Devp devp,		/* Device to be read */
	       unsigned long* buf,	/* 32-bit data buffer */
	       int maxwords,		/* Maximum words in response buffer */
	       int words,		/* Expected words in response */
	       int response_msec);	/* Timeout for response to _begin_ */
#else
int lio_read16(Devp devp,		/* Device to be read */
	       unsigned short* buf,	/* 16-bit data buffer */
	       int maxwords,		/* Maximum words in response buffer */
	       int words,		/* Expected words in response */
	       int response_msec);	/* Timeout for response to _begin_ */
#endif
/*
 * DESCRIPTION
 *
 * Read up to maxwords into buf.  The first word must be received within
 * response_msec milliseconds.  Once the first word has been received,
 * if possible, a 10 millisecond timeout should be used to read subsequent
 * words.  So, when nothing has been received for 10 milliseconds, it
 * can be assumed that there is no more data waiting.  If this second
 * timeout cannot be implemented, try just returning the maximum data
 * that is waiting.  (I think the way Leach's drivers are written, this
 * will always return the "full" message.)
 *
 * Regardless of timeouts, if maxwords have been read, the function
 * must return them immediately.
 *
 * RETURN VALUE
 *
 * Return the actual number of words that were read, or LIO_ERR on error.
 * It is not considered an error to receive less than maxwords, or even
 * less than "words".  Note that the units here are the same as those
 * for the maxwords/words parameters, and not bytes.
 *
 * If there is no data ready, LIO_ERR should be returned.
 * A return value 0 indicates an end-of-file condition.
 *
 * ERRORS
 *
 * Typically the same as those from read(), but lio_read() should print
 * its own message to the console for errors _other than_ timeouts
 * and short reads.
 *
 * NOTES
 *
 * Because of a hardware limitation in the lio, the upper 8 bits of each
 * 24-bit value are lost if lio_read24 is not available (only the case for
 * lio_fpbox.c). 
 *
 * If the hardware driver supports a "non-blocking" mode, where one
 * can efficiently read _up to_ a certain number of words, then
 * we can ignore the `words' parameter, but in the case of lio_sbus.c,
 * there was no non-blocking mode, but I implemented something that
 * works (using an alarm timeout and reading just one word at a time.)
 * This method is really a hack and slows down the operation of detcom,
 * but it is the only way to poll for an unknown amount of data without
 * modifying the SBus driver.  To speed things up for these kind of
 * drivers, the `words' parameter can be used.  If it is non-zero,
 * do a *blocking* read (with timeout) on `words' instead.  The only
 * problem is:
 *
 * (1) We cannot distinguish a short response from no response.
 * (2) We cannot detect trailing garbage or out of sync conditions
 *     as reliably.
 *
 * So, in general, try to implement a non-blocking mode, but if it is
 * too slow, use the `words' parameter as a hint to speed things up.
 */

int lio_read16_init(int n_controllers,	/* Number of controllers */
		    Devp devp[],	/* Array of devices to read */
		    int maxwords);	/* Words per controller expected */
int lio_read16_init_host(int n_controllers,
			 Devp devp[],
			 int maxwords);
/*
 * DESCRIPTION
 *
 * This can be used to "prime" parallel/DMA mode.  detcom uses this sequence:
 *
 * lio_read16_init(...);
 * lio_write24(...some command that causes an immediate response...);
 * lio_read16_parallel(...);
 *
 * lio_read16_init() initializes the driver to return FITS byte-order 16-bits.
 * lio_read16_init_host() is an alternate form which requests the reverse
 *  byte order (if needed, i.e. alphas and pcs) so that the host can do math
 *  operations on the data.  It is up to the higher levels to figure out if
 *  the data needs to be swapped back before writing a FITS file in this case.
 *
 * If it is not possible to be ready for an immediate response after
 * the lio_write24(), lio_read16_init() should still attempt to set
 * up as much as possible (allocating ring buffers, etc.)  so that
 * the lio_read16_parallel() call will be ready as fast as possible.
 * In the case where this is not instant, the DSP code must know to delay
 * the response to a readout command long enough to safely be ready to
 * read here.
 *
 * RETURN VALUE
 *
 * Returns LIO_OK on success, or LIO_ERR on error and prints a message.
 *
 * ERRORS
 *
 * This call should not return a LIO_ERR without describing the problem
 * (failure to allocate memory or initialize DMA mode, for example.)
 */

int lio_read16_parallel(int n_controllers,	/* Number of controllers */
			Devp devp[],		/* Array of devices to read */
			read16_handler handler,	/* args: cont,buf,words */
			int maxwords,		/* Words per controller */
			int response_msec,	/* TO for response to begin */
			int timeout_msec);	/* Timeout for entire read */
/*
 * DESCRIPTION
 *
 * Start parallel readout mode.  As pixels arrive, this function
 * should call handler(), passing it chunks of data from the controller(s).
 * detcom must be configured with the correct handler to descramble or
 * properly process the data.
 *
 * RETURN VALUE
 *
 * Upon success, the TOTAL number of words should be returned.  This
 * should normally be = n_controllers * maxwords.  If an error occurs,
 * this functions should return a value less than 0, as described below:
 *
 * ERRORS
 *
 * If an error occurs with the driver, such as ring buffer overflow
 * or some other unexpected problem, a LIO_ERR (-1) should be returned.
 * detcom will attempt to salvage any data that have already been
 * sent to handler().
 *
 * If an exposure should be aborted (for example, the user sent an
 * interrupt signal, SIGINT while reading) a LIO_ABORTED (-2) should be
 * returned instead (and detcom will assume the data should not
 * be saved to disk.)
 */

#endif
