/*                                             -*- c-file-style: "Ellemtel" -*-

`lio_fpbox.c' - Read/write/reset routines to handle UH "fiber/parallel" box.

This file is part of version 3 of the CFHT Detector Host Command Interface.
Read the `License' file for terms of use and distribution.
Copyright 1998-2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#ifdef SOLARIS_PRIOCNTL
#include <sys/types.h>
#include <sys/priocntl.h>
#include <sys/rtpriocntl.h>
#include <sys/tspriocntl.h>
#endif

#ifdef TIME_WRITES
#include <sys/time.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <errno.h>

#include "cli/cli.h"

#include "libedt.h"
#include "s16d.h"

#include "lio.h"
#include "misc.h"

/*
 * All S16x_COMMAND ioctl's should include this:
 */
#define COMMAND_DEFAULT S16D_DISINT
#define ASSUME_COMMAND_BITS
#define RING_BUFFER_COUNT 4
#define RING_BUFFER_SIZE  (1024*1024)
#define LEACH_READ_TIMEOUT 1 /* hundreths of seconds */

#define TOCHAR(x) (((char)(x)<' '||(char)(x)>'~')?'.':(char)(x))


static inline void
set_edt_param(int fd, int param, unsigned short val)
{
   if (ioctl(fd, param, &val)!=0)
      perror("ioctl");
}

static inline void
set_edt_lparam(int fd, int param, unsigned long val)
{
   if (ioctl(fd, param, &val)!=0)
      perror("ioctl");
}



static unsigned short
get_edt_param(int fd, int param)
{
   unsigned short val = 0xFFFF;

   if (ioctl(fd, param, &val)!=0)
   {
      perror("ioctl");
      return 0xFFFF;
   }
   else return val;
}

static unsigned long
get_edt_lparam(int fd, int param)
{
   unsigned long val = 0xFFFFFFFF;

   if (ioctl(fd, param, &val)!=0)
   {
      perror("ioctl");
      return 0xFFFFFFFF;
   }
   else return val;
}


#define MAX_S16D_SEARCH 6

static Devp devs_used[MAX_S16D_SEARCH];
static int devs_initialized = 0;

Devp
lio_open(const char* fn)
{
   int i;
   Devp rtn = 0;

   if (!devs_initialized)
   {
      memset(devs_used, 0, sizeof(devs_used));
      devs_initialized = 1;
   }

   if (fn)
   {
      char* tmp = strdup(fn);
      int n, len;

      len = strlen(tmp);
      n = tmp[len-1] - '0';
      tmp[len-1] = '\0';
      rtn = edt_open(tmp, n);
      free(tmp);
      return rtn;
   }
   else
   {
      for (i=0; i<MAX_S16D_SEARCH; i++)
      {
	 if (devs_used[i]==0 && (rtn = (Devp)edt_open("s16d", i))!=0)
	 {
	    CLI_DEBUG(printf("debug: Found EDT card at /dev/s16d%d\n", i));
	    devs_used[i] = rtn;
	    return rtn;
	 }
      }
   }
   return 0;
}

int
lio_close(Devp devp)
{
   int i;
   EdtDev* edt = (EdtDev*)devp;

   for (i=0; i<MAX_S16D_SEARCH; i++)
      if (devs_used[i]==devp) devs_used[i] = 0;

   return edt_close(edt);
}


int
lio_reset(Devp devp, int sdsu_gen)
{
   EdtDev* edt = (EdtDev*)devp;

   if (sdsu_gen != 1) return LIO_ERR;
   set_edt_param(edt->fd, S16S_CONFIG, S16D_SWAP);
   set_edt_param(edt->fd, S16S_RTIMEOUT, LEACH_READ_TIMEOUT);
   set_edt_param(edt->fd, S16S_WTIMEOUT, 0);
#ifdef S16S_SHORT_XFER
   set_edt_param(edt->fd, S16S_SHORT_XFER, 1);
#endif
   set_edt_param(edt->fd, S16S_READ_COMMAND,  COMMAND_DEFAULT|S16D_FNCT1|S16D_FNCT2);
   set_edt_param(edt->fd, S16S_WRITE_COMMAND, COMMAND_DEFAULT|S16D_DIR);
   set_edt_param(edt->fd, S16S_COMMAND,       COMMAND_DEFAULT|S16D_FNCT1|S16D_RESET);
   set_edt_param(edt->fd, S16S_SEND_INIT_PULSE, 1);
   return LIO_OK;
}

int
lio_check(Devp devp)
{
   EdtDev* edt = (EdtDev*)devp;
   if (get_edt_param(edt->fd, S16G_DATA)==0xFFFF)
      return LIO_ERR; /* === header file says we should print a message */
   else
      return LIO_OK;
}

int
lio_read16(Devp devp, unsigned short* buf,
	   int maxwords, int words, int response_msec)
{
   EdtDev* edt = (EdtDev*)devp;
   int timeouts_left = (response_msec+50) / 100 + 1;
   int rtn = 2;
   int total = 0;
#ifdef FPBOX_TRACE
   unsigned short* bufp;
   int i;
#endif

   if (words) ; /* === lio_fpbox doesn't need the words hint because
		 * non-blocking reads work correctly.  Keep gcc happy.
		 */

#ifdef S16S_SHORT_XFER
   do
   {
      rtn = read(edt->fd, buf, maxwords*2);
      if (rtn==0 && --timeouts_left) xusleep(90000);
   } while (rtn==0 && timeouts_left);

   if (rtn>0) total += rtn/2;
#else
#warning No SHORT_XFER mode
   for (i = maxwords, bufp = buf; rtn==2 && i--; bufp++)
   {
      rtn = read(edt->fd, bufp, 2);
      if (rtn==2) total++;
   }
#endif
   if (rtn==-1 && errno!=ETIMEDOUT)
   {
      perror("[lio_read]");
      return LIO_ERR;
   }
   if ((rtn&1)) /* This should never happen, but check anyway... */
      printf("warning: Got odd number of bytes from EDT driver.\n");

   if (total == 0) return LIO_ERR;

#ifdef FPBOX_TRACE
#define MAX_SHOW 100
   CLI_DEBUG(printf("debug: [lio_read] <- %s%d*16-bits",
		    total==maxwords?"all ":"", total));
   for (i = ((total>MAX_SHOW)?MAX_SHOW:total), bufp = buf; i--; bufp++)
   {
      printf(" 0x**%04X", *bufp);
   }
   if (total>MAX_SHOW) printf("...\n");
   else
   {
      printf(" ( ");
      for (i = total, bufp = buf; i--; bufp++)
      {
	 printf("*%c%c ",
		TOCHAR((*bufp&0xFF00)>>8),
		TOCHAR(*bufp&0x00FF));
      }
      printf(")\n");
   }
#endif
   return total;
}


/*
 * Some static data allocated and initialized by read16_init
 * and used by read16_parallel...  write24 also checks whether
 * Ring buffer mode is active and causes the fpbox to latch the final
 * word if this is the case.  It doesn't actually get sent down the
 * fiber until the next call to read16ring.
 */
typedef struct
{
   EdtDev* edt;
   BOOLEAN ring_mode;
   unsigned int words_read;
   unsigned int words_copied;
   unsigned int buf_pos;
   unsigned short* buf[RING_BUFFER_COUNT];
} ring_info_t;

static ring_info_t* ring_info = 0;
static int ring_n_controllers = 0;
static BOOLEAN ring_successful = FALSE;

int
lio_write24(Devp devp, unsigned long* data, int nwords)
   /*
    * Write `nwords' 24-bit words using programmed-i/o.  The values
    * are sent as 2 16-bit writes each and the S16 is placed in input mode
    * at the same time that the interface board will send off the
    * re-assembled 24-bit value of the very last word so that an immediate
    * response from the leach system will not be lost.
    *
    * In cases where communication is out of sync, early responses may be lost.
    *
    * Each ioctl seems to take about 10us.
    */
{
#ifdef SOLARIS_PRIOCNTL
   static long prio_rtn = 0;
   static pcinfo_t pcinfo_rt;
   static pcinfo_t pcinfo_ts;
   static pcparms_t pcparms_rt;
   static pcparms_t pcparms_ts;
#endif
#ifdef TIME_WRITES
   struct timeval tp_start, tp_finish;
#endif
   EdtDev* edt = (EdtDev*)devp;
   unsigned short tmp;
   unsigned long* datap;
   int hold_last = 0;
   int i;
#ifdef ASSUME_COMMAND_BITS
   /* See comments in code where these are used: */
   static unsigned short s16_init =  COMMAND_DEFAULT|S16D_DIR|S16D_BCLR|S16D_INIT;
   static unsigned short s16_send1 = COMMAND_DEFAULT|S16D_DIR;
   static unsigned short s16_send2 = COMMAND_DEFAULT|S16D_FNCT1;
   static unsigned short s16_recv  = COMMAND_DEFAULT|S16D_FNCT1|S16D_FNCT2;
#else
   unsigned short s16_command_default;
   unsigned short s16_init;
   unsigned short s16_send1;
   unsigned short s16_send2;
   unsigned short s16_recv;

   ioctl(edt->fd, S16G_COMMAND, &s16_command_default);
   s16_command_default &= (S16D_SAVE_MASK&~(S16D_FNCT1|S16D_FNCT2|S16D_DIR));
   CLI_DEBUG(printf("debug: s16_command_default=%04X\n", s16_command_default));
   
   s16_init =  s16_command_default|S16D_DIR|S16D_BCLR|S16D_INIT;
   s16_send1 = s16_command_default|S16D_DIR;
   s16_send2 = s16_command_default|S16D_FNCT1;
   s16_recv  = s16_command_default|S16D_FNCT1|S16D_FNCT2;
#endif

   for (i=0; i<ring_n_controllers; i++)
   {
      if (ring_info[i].edt == edt && ring_info[i].ring_mode == TRUE)
      {
	 hold_last = 1;
	 break;
      }
   }

#ifdef FPBOX_TRACE
   CLI_DEBUG(printf("debug: [lio_write] ->"));
   for (i = nwords, datap = data; i--; datap++)
   {
      printf(" 0x%06lX", *datap);
   }
   printf(" ( ");
   for (i = nwords, datap = data; i--; datap++)
   {
      printf("%c%c%c ",
	     TOCHAR(((*datap&0xFF0000)>>16)),
	     TOCHAR((*datap&0x00FF00)>>8),
	     TOCHAR(*datap&0x0000FF));
   }
   printf(")\n");
#endif

   if (WRITE_PREDELAY_USEC) xusleep(WRITE_PREDELAY_USEC);

#ifdef SOLARIS_PRIOCNTL
   if (prio_rtn == 0)
   {
      strcpy(pcinfo_rt.pc_clname, "RT");
      strcpy(pcinfo_ts.pc_clname, "TS");
      if ((prio_rtn = priocntl(P_PID, P_MYID, PC_GETCID, (void*)&pcinfo_rt))>0)
	 if ((prio_rtn = priocntl(P_PID, P_MYID, PC_GETCID, (void*)&pcinfo_ts))>0)
	 {
	    rtparms_t* rtparms = (rtparms_t*)pcparms_rt.pc_clparms;

	    pcparms_ts.pc_cid = pcinfo_ts.pc_cid;
	    prio_rtn = priocntl(P_PID, P_MYID, PC_GETPARMS, (void*)&pcparms_ts);
	    pcparms_rt.pc_cid = pcinfo_rt.pc_cid;
	    rtparms->rt_pri = 20;
	    rtparms->rt_tqsecs = 0;
	    rtparms->rt_tqnsecs = 100000; /* .1 msec time quantum */
	 }
   }

   /*
    * Switch to realtime mode, if possible
    */
   if (prio_rtn > 0)
   {
      seteuid(0);
      if ((prio_rtn = priocntl(P_PID, P_MYID, PC_SETPARMS, (void*)&pcparms_rt))==-1)
	 perror("warning: Failed to set realtime priority");
      else prio_rtn = 1;
      seteuid(getuid());
   }
#endif

#ifdef TIME_WRITES
   gettimeofday(&tp_start, 0);
#endif

   /*
    * Make sure the box and edt board are ready to accept data from the sparc
    */
   ioctl(edt->fd, S16S_COMMAND, &s16_init);
   ioctl(edt->fd, S16S_COMMAND, &s16_send1);

   for (i = nwords, datap = data; i--; datap++)
   {
      /*
       * Send the high word (upper 8-bits)
       */
      tmp = (*datap&0xFF0000)>>16; ioctl(edt->fd, S16S_DATA, &tmp);
      tmp = (*datap&0x00FFFF);
      if (!i)
      {
	 /*
	  * Before sending the lower 16-bits of the final 24-bit word,
	  * get ready to switch to input mode and don't send right away.
	  */
	 ioctl(edt->fd, S16S_COMMAND, &s16_send2);
	 ioctl(edt->fd, S16S_DATA, &tmp);
	 if (hold_last==0)
	    ioctl(edt->fd, S16S_COMMAND, &s16_recv); /* Tell the box to send it off */
	 else
	    CLI_DEBUG(printf("debug: [lio_write24] Holding last 16 bits in fpbox\n"));
      }
      else
      {
	 ioctl(edt->fd, S16S_DATA, &tmp);
      }
   }

#ifdef TIME_WRITES
   {
      long elapsed;
      gettimeofday(&tp_finish, 0);
      elapsed = (tp_finish.tv_sec - tp_start.tv_sec)*1000000
	 + (tp_finish.tv_usec - tp_start.tv_usec);
      if (elapsed > 400)
      {
	 printf("%s: fpbox write of %d words took %ld usec\n",
		elapsed>2000?"logonly":"logonly", nwords, elapsed); /* %%% Should be a warning */
      }
   }
#endif

#ifdef SOLARIS_PRIOCNTL
   /*
    * Switch back to timeshare mode
    */
   if (prio_rtn > 0)
   {
      seteuid(0);
      if ((prio_rtn = priocntl(P_PID, P_MYID, PC_SETPARMS, (void*)&pcparms_ts))==-1)
	 perror("warning: Failed to return to timesharing priority mode");
      seteuid(getuid());
   }
#endif
   return nwords; /* PIO mode, assume it worked :-/ */
}


int
lio_read16_init(int n_controllers, Devp devp[], int maxwords)
{
   int i, n;
   BOOLEAN same = ring_successful;

   if (maxwords) ; /* Keep gcc happy.  For fpbox, maxwords expected
		    * doesn't affect the size of the buffers we choose
		    * for DMA, so we don't use this parameter.
		    */

   if (ring_n_controllers != n_controllers) same = FALSE;
   for (n = 0; n < n_controllers && same!=FALSE; n++)
      if (ring_info[n].edt != (EdtDev*)(devp[n])) same = FALSE;
   
   if (same==FALSE)
   {
      printf("logonly: [lio_read16_init] Setting up ring buffers\n");
      /*
       * Free the buffers allocated last time around if called
       * again and edt devices were not the same or there were more/less.
       */
      if (ring_info != 0)
      {
	 for (n = 0; n < ring_n_controllers; n++)
	 {
	    if (ring_info[n].buf[0] != 0) free(ring_info[n].buf[0]);
	 }
	 free(ring_info);
      }
      
      /*
       * (Re)allocate buffers for specified number of controllers.
       */
      ring_n_controllers = n_controllers;

      ring_info = (ring_info_t*)malloc(n_controllers * sizeof(ring_info_t));
      memset(ring_info, 0, sizeof(ring_info));

      for (n = 0; n < n_controllers; n++)
      {
	 ring_info[n].ring_mode = FALSE;
	 ring_info[n].edt = (EdtDev*)devp[n];
	 /*
	  * Allocate temporary ring buffer space for the EDT driver.
	  * === Unfortunately this requires another copy of all the
	  * data to be performed by the CPU to get it into the user's
	  * buffer space.
	  */
	 ring_info[n].buf[0] = (unsigned short*)valloc(RING_BUFFER_SIZE * RING_BUFFER_COUNT);
	 if (ring_info[n].buf[0]==0)
	 {
	    printf("error: Unable to allocate %d MB ring buffer space for EDT\n",
		   RING_BUFFER_SIZE * RING_BUFFER_COUNT / (1024*1024));
	    return LIO_ERR;
	 }
	 CLI_DEBUG(printf("debug: Clearing ring buffers\n")); /* To keep purify happy */
	 memset(ring_info[n].buf[0], 0, RING_BUFFER_SIZE * RING_BUFFER_COUNT);
	 CLI_DEBUG(printf("debug: Done clearing\n"));
	 for (i = 0; i < RING_BUFFER_COUNT; i++)
	 {
	    ring_info[n].buf[i] =
	       (unsigned short*)((void*)(ring_info[n].buf[0]) + RING_BUFFER_SIZE*i);
	 }
      }
      printf("logonly: [lio_read16_init] Buffer space allocated\n");
   }
   
   
   for (n=0; n<n_controllers; n++) if (ring_info[n].ring_mode == FALSE)
   {
      /*
       * Read timeout doesn't work in ring mode, so turn it off.
       */
      set_edt_param(ring_info[n].edt->fd, S16S_RTIMEOUT, 0);

      /*
       * Zero the counters in the ring_info structs.
       */
      ring_info[n].words_read = 0;
      ring_info[n].words_copied = 0;
      ring_info[n].buf_pos = 0;

      /*
       * Pass buffer addresses to the driver.
       */
      if (edt_configure_ring_buffers(ring_info[n].edt,
				     RING_BUFFER_SIZE, RING_BUFFER_COUNT,
				     EDT_READ, (void*)ring_info[n].buf)!=0)
      {
	 perror("error: edt_configure_ring_buffers");
	 return LIO_ERR;
      }
      
      /*
       * Save a flag so this doesn't have to be repeated.
       */
      ring_info[n].ring_mode = TRUE;
   }

   ring_successful = TRUE;
   return LIO_OK;
}

static void
lio_read16_exit(void)
{
   int n;

   for (n = 0; n < ring_n_controllers; n++) if (ring_info[n].ring_mode == TRUE)
   {
      ring_info[n].ring_mode = FALSE;
      edt_disable_ring_buffers(ring_info[n].edt);
      set_edt_param(ring_info[n].edt->fd, S16S_RTIMEOUT, LEACH_READ_TIMEOUT);
   }
}

int
lio_read16_parallel(int n_controllers,
		    Devp devp[],
		    read16_handler handler,
		    int maxwords,
		    int response_msec,
		    int timeout_msec)
{
   int time_stop;
   int time_response;
   int last_words_read = 0;
   int curr_words_read = 0;
   int n;
   static unsigned short s16_recv  = S16D_FNCT1|S16D_FNCT2;

   time_response = time_stop = time(0);
   time_response += (response_msec + 900) / 1000;
   time_stop += (timeout_msec + 900) / 100;

   /*
    * Make sure the card is in ring buffer mode.  To be ready for an
    * immediate response, the higher level code should already have
    * called this, in which case this call does nothing here.
    */
   lio_read16_init(n_controllers, devp, maxwords);

   /*
    * Tell the driver to start using the ring buffers.
    */
   for (n=0; n<n_controllers; n++)
   {
      if (edt_start_buffers(ring_info[n].edt, 0)!=0)
      {
	 perror("edt_start_buffers");
	 lio_read16_exit();
	 /* Abort the read if things can't be set up. */
	 return LIO_ABORTED;
      }
      set_edt_lparam(ring_info[n].edt->fd, S16S_BYTECOUNT, 0); /* Zero the bytecounter */
   }

   /*
    * Toggle a control line on the fpbox which causes it to finally
    * send the last latched word down the fiber, now that the driver
    * is ready for an immediate response.
    */
   for (n=0; n<n_controllers; n++)
   {
      if (ioctl(ring_info[n].edt->fd, S16S_COMMAND, &s16_recv)==-1)
      {
	 printf("warning: ioctl to set EDT card %d in receive mode failed\n", n);
      }
   }

   CLI_DEBUG(printf("debug: [lio_read16_parallel] ready for input\n"));

   do
   {
      int delay;
      
      delay = 100000;
      last_words_read = curr_words_read;
      curr_words_read = 0;

      for (n=0; n<n_controllers; n++)
      {
	 void* ptr;

	 /*
	  * Update progress statistics for each controller
	  *
	  * NOTE: The "-2" is needed to compensate for an offset in the EDT
	  *       driver.  Discussed possibility of fixing this in the driver
	  *       with EDT but there are probably too many applications that
	  *       now rely on this "bug", so... here's another one!
	  */
	 ring_info[n].words_read = (edt_byte_count(ring_info[n].edt)-2)/2;
	 curr_words_read += ring_info[n].words_read;

	 /*
	  * Make sure processing is keeping up with the DMA rate
	  */
	 if (edt_ring_buffer_overrun(ring_info[n].edt))
	 {
	    printf("error: Ring buffer overrun; can't process data fast enough!\n");
	    lio_read16_exit();
	    return LIO_ERR; /* detcom will try to salvage the exposure */
	 }

	 /*
	  * Check if an entire buffer is ready to be copied yet.
	  */
	 if ((ptr = edt_check_for_buffers(ring_info[n].edt, 1)))
	 {
	    int words_to_copy = RING_BUFFER_SIZE/sizeof(short);

	    delay = 0;
	    if (ptr != ring_info[n].buf[ring_info[n].buf_pos])
	    {
	       printf("error: EDT driver returned the wrong buffer!\n");
	       lio_read16_exit();
	       return LIO_ERR; /* detcom will try to salvage the exposure */
	    }
	    ring_info[n].buf_pos = (ring_info[n].buf_pos + 1)%RING_BUFFER_COUNT;

	    if (ring_info[n].words_copied + words_to_copy > maxwords)
	    {
	       printf("logonly: Discarding %d extra words from controller %d\n",
		      ring_info[n].words_copied + words_to_copy - maxwords, n);
	       words_to_copy = maxwords - ring_info[n].words_copied;
	    }

	    if (words_to_copy)
	    {
	       handler(n, ptr, words_to_copy);
	       ring_info[n].words_copied += words_to_copy;
	    }
	 }
      }

      if (curr_words_read==last_words_read && curr_words_read>0)
      {
	 CLI_DEBUG(printf("debug: No more data\n"));
	 break;
      }

#if 0
      /*
       * === This can happen, because the counter is a volatile memory
       * location.  The only thing to check is that curr_words_read is
       * constantly changing.  Once it stops changing, then the final
       * value is reliable.
       */
      if (curr_words_read < last_words_read)
      {
	 printf("error: Problem with EDT driver: words read changed "
		"from %d to %d; data is suspect!!!\n",
		last_words_read, curr_words_read);
	 lio_read16_exit();
	 return LIO_ERR; /* detcom will try to salvage the exposure */
      }
#endif

      if (curr_words_read == 0)
	 printf("progress: Waiting for pixels to begin flowing\r");
      else
      {
	 if (maxwords >= 100 * 100)
	    printf("progress: Reading out camera (%d%%)\r",
		   curr_words_read/(maxwords*n_controllers/100));
	 else
	    printf("progress: Reading out camera (%d%%)\r",
		   curr_words_read*100/(maxwords*n_controllers));
      }
      fflush(stdout);

      /*
       * Process data if enough, else...
       */
      cli_signal_unblock(SIGINT);
      cli_signal_unblock(SIGQUIT);
      cli_signal_unblock(SIGTSTP);
      xusleep(delay);
      cli_signal_block(SIGINT);
      cli_signal_block(SIGQUIT);
      cli_signal_block(SIGTSTP);
      if (cli_signal_event(SIGINT))
      {
#ifdef ALLOW_READOUT_ABORTS
	 printf("error: Readout aborted by user!\n");
	 lio_read16_exit();
	 return LIO_ABORTED;
#else
	 printf("warning: Cannot abort now.  Please wait for readout to finish.\n");
#endif
      }
      if (cli_signal_event(SIGTSTP))
      {
	 printf("warning: Readout already in progress; `stop' command ignored.\n");
      }
   } while (time(0)<((curr_words_read==0)?time_response:time_stop));

   if (time(0) >= time_stop)
   {
      CLI_DEBUG(printf("debug: reached maximum time of %d msec\n", timeout_msec));
   }
   else if (time(0) >= time_response && curr_words_read==0)
   {
      CLI_DEBUG(printf("debug: no response within %d msec\n", response_msec));
   }


	 
   /*
    * === Hide the EDT's inability to read less than a 16 word chunk
    * when in DMA mode.  This hack will stuff zeroes in the last few
    * pixels to make the handler think it got exactly the expected
    * number of pixels, if needed.  This is usually only needed if
    * an odd raster size is selected.
    */
   curr_words_read = 0;
   for (n=0; n<n_controllers; n++)
   {
      ring_info[n].words_read = (edt_byte_count(ring_info[n].edt)-2)/2;
      {
	 int words_left_over = maxwords % 16;
	 
	 if (words_left_over != 0 &&
	     ring_info[n].words_read == maxwords - words_left_over)
	 {
	    printf("logonly: Odd read size; EDT hack is padding %d words\n",
		   words_left_over);
	    memset(ring_info[n].buf[ring_info[n].buf_pos] + ring_info[n].words_read,
		   0, words_left_over * 2);
	    ring_info[n].words_read += words_left_over;
	    last_words_read += words_left_over;
	 }
      }
      
      curr_words_read += ring_info[n].words_read;
   }
   
   if (curr_words_read!=last_words_read)
   {
      printf("warning: Byte count changed from %d to %d after end of data!\n",
	     last_words_read, curr_words_read);
   }

   /*
    * Flush last partial buffer, if any.
    */
   for (n = 0; n<n_controllers; n++)
   {
      if (ring_info[n].words_read == ring_info[n].words_copied)
      {
	 CLI_DEBUG(printf("debug: nothing to flush\n"));
      }
      else if (ring_info[n].words_read < ring_info[n].words_copied)
      {
	 printf("error: internal inconsistency: words_copied > words_read!\n");
      }
      else
      {
	 int words_to_copy = ring_info[n].words_read - ring_info[n].words_copied;

	 if (words_to_copy > RING_BUFFER_SIZE/sizeof(short))
	 {
	    printf("error: internal inconsistency: more than one ring buffer left over\n");
	 }
	 else
	 {
	    if (ring_info[n].words_read > maxwords)
	    {
	       words_to_copy = maxwords - ring_info[n].words_copied;
	       printf("logonly: Discarding %d extra pixels from controller %d\n",
		      ring_info[n].words_read - maxwords, n);
	    }

	    CLI_DEBUG(printf("debug: flushing %d final words\n", words_to_copy));
	    handler(n, ring_info[n].buf[ring_info[n].buf_pos], words_to_copy);
	    ring_info[n].words_copied += words_to_copy;
	 }
      }
   }

   lio_read16_exit();
   return curr_words_read;
}

void
print_edt_status_bits(Devp devp)
{
   EdtDev* edt = (EdtDev*)devp;

   printf("info1: STATUS=0x%04X, CONFIG=0x%04X, R/WTIMEOUT=%d/%d, "
	         "DATA=0x%04X COMMAND=0x%04X,\n"
	  "info2: DEBUG/INTR=%d/%d READ(DEF)=0x%04X(0x%04X) WRITE(DEF)=0x%04X(0x%04X)\n"
	  "info3: DMATHROT=%lX MAXDMASIZE=%lX\n",
	  get_edt_param(edt->fd, S16G_STATUS),
	  get_edt_param(edt->fd, S16G_CONFIG),
	  get_edt_param(edt->fd, S16G_RTIMEOUT),
	  get_edt_param(edt->fd, S16G_WTIMEOUT),
	  get_edt_param(edt->fd, S16G_DATA),
	  get_edt_param(edt->fd, S16G_COMMAND),
	  get_edt_param(edt->fd, S16G_DEBUG_LEVEL),
	  get_edt_param(edt->fd, S16G_DEBUG_INTR),
	  get_edt_param(edt->fd, S16G_READ_COMMAND),
	  get_edt_param(edt->fd, S16G_DEF_READ_COMMAND),
	  get_edt_param(edt->fd, S16G_WRITE_COMMAND),
	  get_edt_param(edt->fd, S16G_DEF_WRITE_COMMAND),
	  get_edt_lparam(edt->fd, S16G_DMATHROT),
	  get_edt_lparam(edt->fd, S16G_MAXDMASIZE)
	  );
}
