/*                                             -*- c-file-style: "Ellemtel" -*-

`lio_lotus.c' - Read/write/reset routines for SDSU's PCI fiber interface.

This file is part of version 3.1 of the CFHT Detector Host Command Inteface.
Read the `License' file for terms of use and distribution.
Copyright 1998-2000, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#include "lio.h"
#include "misc.h"
#include "cli/cli.h" /* For CLI_DEBUG() define */
#include <stddef.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include "lotuspci.h" /* %%% */
#include "detcom.h" /* %%% For DETCOM_BSWAP only */


#define MAX_CONTROLLER_SEARCH 2

typedef struct
{
      int in_use;
      int fd;
} Dev;

static Dev devs[MAX_CONTROLLER_SEARCH];
static int devs_initialized = 0;
static int ltpci_version_code = -1;

/*
 * This routine exists to support the controller simulator.
 */
static int
open_named_socket(const char* filename)
{
   struct sockaddr_un name; /* socket address, server's listening socket */
   int sock;
   size_t size;

   sock = socket (PF_UNIX, SOCK_STREAM, 0);
   if (sock < 0) return sock;

   name.sun_family = AF_FILE;
   strcpy (name.sun_path, filename);

   /* The size of the address is
      the offset of the start of the filename,
      plus its length,                                                         
      plus one for the terminating null byte.  */
   size = (offsetof (struct sockaddr_un, sun_path)
           + strlen (name.sun_path) + 1);

   if (connect(sock, (struct sockaddr *)&name, size) != 0)
      return -1;

   return sock;
}

Devp
lio_open(const char* fn)
{
   int i;

   if (!devs_initialized)
   {
      memset(devs, 0, sizeof(devs));
      devs_initialized = 1;
   }

   if (fn)
   {
      /* %%% Malloc a new dev instead? */
      for (i=0; i < MAX_CONTROLLER_SEARCH; i++)
	 if (devs[i].in_use == 0) break;
      if (i >= MAX_CONTROLLER_SEARCH)
      {
	 printf("error: Recompile lio_lotus.c with higher MAX_CONTROLLER_SEARCH\n");
	 return 0;
      }
      /*
       * First try opening fn as a named socket.
       */
      devs[i].fd = open_named_socket(fn);
      /* %%% Need to set NONBLOCK for socket too! */
      if (devs[i].fd < 0)
	 devs[i].fd = open(fn, O_RDWR|O_NONBLOCK, 0);
      if (devs[i].fd < 0)
      {
	 perror(fn);
	 return 0;
      }
      if (fcntl(devs[i].fd, F_SETFL,
		fcntl(devs[i].fd, F_GETFL, 0)|O_NONBLOCK) < 0)
	 perror("fcntl O_NONBLOCK failed");

      printf("logonly: Found SDSU card at %s, fd=%d\n", fn, devs[i].fd);
      devs[i].in_use = 1;
      return (Devp)(&(devs[i]));
   }

   for (i=0; i<MAX_CONTROLLER_SEARCH; i++)
   {
      char tmp[80];
      
      sprintf(tmp, "/dev/lotuspci%d", i);
      if (devs[i].in_use==0 && (devs[i].fd = open(tmp, O_RDWR|O_NONBLOCK, 0))!=-1)
      {
	 printf("logonly: Found SDSU card at %s, fd=%d\n", tmp, devs[i].fd);
	 devs[i].in_use = 1;
	 return (Devp)(&(devs[i]));
      }
   }
   return 0;
}

int
lio_close(Devp devp)
/*
 * This closes the device previously returned by lio_open.  The
 * return code is the same as the return code from close().
 */
{
   Dev* dp = (Dev*)devp;

   if (dp==0 || dp->in_use!=1) return LIO_ERR;
   dp->in_use = 0;
   return close(dp->fd);
}


int
lio_check(Devp devp) { return LIO_OK; }

#ifdef DETCOM_BSWAP
/* Dec, PC */
#define SWAP_MODE_FITS LTPCI_16SWAP /* Byte order suitable for FITS */
#define SWAP_MODE_HOST LTPCI_NOSWAP /* Byte order for host math ops */
#else
/* Sun, HP */
#define SWAP_MODE_FITS LTPCI_16SWAP /* Byte order suitable for FITS */
#define SWAP_MODE_HOST LTPCI_16SWAP /* Byte order for host math ops */
#endif

int
lio_reset(Devp devp, int sdsu_gen)
{
   Dev* dp = (Dev*)devp;
   int gen;

   switch (sdsu_gen)
   {
      /*
       * This driver supports both SDSU1 and SDSU2 fiber protocols.
       */
      case 1: gen = LTPCI_GEN1_MODE; break;
      case 2: gen = LTPCI_GEN2_MODE; break;
      default: return LIO_ERR;
   }

   if (ioctl(dp->fd, LTPCI_GET_VERSION, &ltpci_version_code) != 0)
   {
      ltpci_version_code = -1;
      printf("logonly: Version 1.x lotuspci.o driver\n");
      printf("warning: You are using an old version of the lotuspci driver.\n");
      printf("warning: Version 1.6, 3.04, or better should be used.\n");
   }
   else
   {
      printf("logonly: Version %g lotuspci.o driver\n",
	     (double)ltpci_version_code/1000);
      if (ltpci_version_code < 1006)
      {
	 printf("warning: You are using an old version (%g) of the lotuspci driver.\n",
		(double)ltpci_version_code/1000);
	 printf("warning: Version 1.6, 3.04, or better should be used.\n");
      }
      if ((ltpci_version_code >= 2000 && ltpci_version_code < 3002) ||
	  (ltpci_version_code > 3002 && (ltpci_version_code&1)))
      {
	 if (ltpci_version_code >= 2016)
	 {
	    printf("warning: Using experimental lotuspci driver!\n");
	 }
	 else
	 {
	    printf("error: This version of DetCom requires lotuspci-1 or 3 "
		   "(version 2 for IR is loaded)\n");
	    return LIO_ERR;
	 }
      }
   }

   if (ioctl(dp->fd, LTPCI_DETCOM, 0) != 0 ||
       ioctl(dp->fd, gen, 0) != 0 ||
       ioctl(dp->fd, SWAP_MODE_FITS, 0) != 0)
   {
      perror("lio_reset ioctl failed");
      return LIO_ERR;
   }

   printf("logonly: Fiber reset to SDSU GEN%d mode\n", sdsu_gen);

   return LIO_OK;
}

int
lio_reset_peer(Devp devp)
{
   Dev* dp = (Dev*)devp;

   printf("logonly: Sending special controller-reset to timing board\n");

   if (ioctl(dp->fd, LTPCI_RESET_CONTROLLER, 0) != 0)
      return LIO_ERR;

   return LIO_OK;
}

int
lio_write24(Devp devp, unsigned long* data, int nwords)
/*
 * Send 24-bit data to the lio.
 */
{
   Dev* dp = (Dev*)devp;
   int rtn;

   if (dp==0 || dp->in_use!=1) return LIO_ERR;

   rtn = write(dp->fd, data, nwords*4);
   if (rtn!=nwords*4)
   {
      perror("error: write failed"); /* %%% */
      return LIO_ERR;
   }
   return nwords;
}

int
lio_read24(Devp devp, unsigned long* buf,
	   int maxwords, int words, int response_msec)
{
   Dev* dp = (Dev*)devp;
   struct timeval tv;
   fd_set readfds;
   int i, rtn;

   tv.tv_sec = response_msec / 1000;
   tv.tv_usec = (response_msec % 1000) * 1000;

   FD_ZERO(&readfds);
   FD_SET(dp->fd, &readfds);
   switch (TEMP_FAILURE_RETRY(select(dp->fd + 1, &readfds, 0, 0, &tv)))
   {
      case -1: perror("select failed"); return LIO_ERR;
      case 0: /* timeout */ return LIO_ERR;
   }
   rtn = read(dp->fd, buf, sizeof(long) * maxwords);
   switch (rtn)
   {
      case -1: perror("read failed"); return LIO_ERR;
      case 0:  fprintf(stderr, "logonly: End of file on character device.\n"); return LIO_ERR;
   }
   if (rtn % sizeof(long) != 0)
   {
      printf("error: got strange number of bytes (%d) from leach driver\n",
	     rtn);
      return LIO_ERR;
   }
   rtn /= sizeof(long);
   for (i=0; i<words; i++)
      buf[i] &= 0xFFFFFF;
   return rtn;
}

int
lio_read16_init(int n_controllers, Devp devp[], int maxwords)
{
   int i;

   /*
    * 1.x versions of the driver are not capable of sending out the
    * RDC command (or any command) once DMA mode has been selected,
    * so no initialization is possible yet (and we must rely on DEL_RDT
    * being long enough in the DSP code.)
    */
   if (ltpci_version_code < 2000) return LIO_OK;

   for (i=0 ; i<n_controllers; i++)
   {
      if (ioctl(((Dev*)devp[i])->fd, LTPCI_DMA_READ, &maxwords) < 0)
      {
	 perror("ioctl(LTPCI_DMA_READ) failed");
	 return LIO_ERR;
      }
   }

   return LIO_OK;
}

int
lio_read16_init_host(int n_controllers, Devp devp[], int maxwords)
{
   int i;

   for (i=0 ; i<n_controllers; i++)
   {
      if (ioctl(((Dev*)devp[i])->fd, SWAP_MODE_HOST, 0) < 0)
      {
	 perror("lio_read16_init_host: ioctl to select host-byte order failed");
	 return LIO_ERR;
      }
   }
   return lio_read16_init(n_controllers, devp, maxwords);
}

void
lio_read16_exit(void)
{
   if (ioctl(devs[0].fd, SWAP_MODE_FITS, 0) != 0) /* %%% other devs */
   {
      perror("lio_read16_exit: ioctl to select FITS byte order failed");
   }
   /* %%% DMA abort */
}

int
lio_read16_parallel(int n_controllers, Devp devp[], read16_handler handler,
		    int maxwords, int response_msec, int timeout_msec)
{
   fd_set readfds_full, readfds;
   int i, max_fd = 1;
   unsigned int words_left[MAX_CONTROLLER_SEARCH];
   unsigned int words_left_total = 0;

   if (ltpci_version_code < 2000)
   {
      for (i=0 ; i<n_controllers; i++)
      {
	 if (ioctl(((Dev*)devp[i])->fd, LTPCI_DMA_READ, &maxwords) < 0)
	 {
	    perror("ioctl(LTPCI_DMA_READ) failed");
	    return LIO_ERR;
	 }
      }
   }

   /*
    * Build the list of file descriptors (PCI devices) to select
    * during the waiting loop.
    */
   FD_ZERO(&readfds_full);
   for (i=0; i<n_controllers; i++)
   {
      int fd;

      words_left[i] = maxwords;
      words_left_total += maxwords;
      fd = ((Dev*)devp[i])->fd; /* %%% */
      if (fd + 1 > max_fd) max_fd = fd + 1;
      FD_SET(fd, &readfds_full);
   }

   while (words_left_total > 0)
   {
      static int timeout = 5;
      struct timeval tv;

      tv.tv_sec = 1;
      tv.tv_usec = 0;

      readfds = readfds_full;
      switch (TEMP_FAILURE_RETRY(select(max_fd, &readfds, 0, 0, &tv)))
      {
	 case -1:
	 {
	    perror("lio_read16_parallel select failed");
	    for (i=0 ; i<n_controllers; i++) /* %%% */
	       ioctl(((Dev*)devp[i])->fd, LTPCI_RESET, 0);
	    return LIO_ERR; /* %%% Abort */
	    break;
	 }
	 case 0:
	 {
	    if (timeout-- == 0)
	    {
	       fprintf(stderr, "\rerror: No data for 5 seconds\n");
	       for (i=0 ; i<n_controllers; i++) /* %%% */
		  ioctl(((Dev*)devp[i])->fd, LTPCI_RESET, 0);
	       return LIO_ERR; /* Timeout */
	    }
	    break;
	 }
	 default: timeout = 5;
      }

      cli_signal_unblock(SIGINT); /* %%% Enable BEFORE select? */
      cli_signal_unblock(SIGQUIT);
      cli_signal_unblock(SIGTSTP);
      cli_signal_block(SIGINT);
      cli_signal_block(SIGQUIT);
      cli_signal_block(SIGTSTP);
      if (cli_signal_event(SIGINT))
      {
	 raise(SIGQUIT);
	 cli_signal_unblock(SIGQUIT);
	 cli_signal_block(SIGQUIT);
#ifdef ALLOW_READOUT_ABORTS
	 fprintf(stderr, "error: Readout aborted by user!\n");
	 lio_read16_exit();
	 return LIO_ERR;
#else
	 fprintf(stderr, "\rwarning: Cannot abort now.  Please wait for readout to finish.\n");
#endif
      }
      if (cli_signal_event(SIGTSTP))
      {
	 fprintf(stderr, "\rwarning: Readout already in progress.\n");
	 raise(SIGQUIT);
	 cli_signal_unblock(SIGQUIT);
	 cli_signal_block(SIGQUIT);
      }

      for (i=0; i<n_controllers; i++)
      {
	 int fd;

	 fd = ((Dev*)devp[i])->fd; /* %%% */
	 if (FD_ISSET(fd, &readfds))
	 {
	    unsigned char buf[32 * 4096];
	    int count;

	    if ((words_left[i]*2) > sizeof(buf))
	    {
	       count = read(fd, buf, sizeof(buf));
	    }
	    else
	    {
	       count = read(fd, buf, words_left[i]*2);
	    }

	    if (count==-1 || count & 1) /* %%% Count & 1 */
	    {
	       perror("read failed"); /* %%% abort */
	       for (i=0 ; i<n_controllers; i++) /* %%% */
		  ioctl(((Dev*)devp[i])->fd, LTPCI_RESET, 0);
	       return LIO_ERR;
	    }

	    if (count > sizeof(buf))
	    {
	       fprintf(stderr, "\rerror: %d > %d\n", count, sizeof(buf)); /* %%% */
	    }

	    if ((words_left[i] -= count/2) == 0)
	       FD_CLR(fd, &readfds_full);

	    words_left_total -= count/2;
	    handler(i, (unsigned short*)buf, count/2); /* %%% Hack cast */
	 }
	 fprintf(stderr, "\rprogress: Reading %s (%d%%)\r", filebase(),
		 (maxwords*n_controllers - words_left_total)/(maxwords*n_controllers/100));
      }
   }
   if (!words_left_total)
      fprintf(stderr, "\rprogress: Reading complete (100%%)\n");

   for (i=0; i<n_controllers; i++)
   {
      unsigned long don = 0xFFFFFFFF;
      int rdcount;

      rdcount = lio_read24(devp[i], &don, 1, 1, 1000);

      if (rdcount != 1)
      {
	 fprintf(stderr, "error: Got %d words instead of single DON from PCI board %d after the readout.\n",
		 rdcount, i);
      }
      if (don != 0x444F4E)
      {
#if 0
	 ioctl(((Dev*)devp[i])->fd, LTPCI_RESET, 0); /* %%% Clean up the casts */
#endif
	 fprintf(stderr,
		 "error: No DON (0x%06lx) after readout from PCI board %d; pixels may be missing!\n", don, i);
	 /* %%% Return -1? */
      }
   }
   lio_read16_exit(); /* %%% */
   return maxwords * n_controllers - words_left_total; /* %%% */
}

/* End of lio_lotus.c */
