/*                                             -*- c-file-style: "Ellemtel" -*-

`lio_sbus.c' - Read/write/reset routines for SDSU's SBus fiber interface.

This file is part of version 3 of the CFHT Detector Host Command Interface.
Read the `License' file for terms of use and distribution.
Copyright 1998-2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#ifdef SOLARIS_PRIOCNTL
#include <sys/types.h>
#include <sys/priocntl.h>
#include <sys/rtpriocntl.h>
#include <sys/tspriocntl.h>
#endif
/* %%% Only for mlock */
#include <sys/mman.h>

#include "lio.h"
#include "misc.h"
#include "cli/cli.h" /* For CLI_DEBUG() define */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/asynch.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#define MAX_CONTROLLER_SEARCH 2
#define RING_BUFFER_COUNT 2
/* %%% MAJOR HACK */
#define RING_BUFFER_COUNT_MALLOC 1
/* #define RING_BUFFER_SIZE  (1024*1024) */
/* #define RING_BUFFER_SIZE  (2080*4128*4*2) */
/* #define RING_BUFFER_SIZE  (PIXELS_PER_CONTROLLER*2) */

typedef struct
{
      int in_use;
      int fd;
} Dev;

/*
 * Some static data allocated and initialized by read16_init
 * and used by read16_parallel.
 */
typedef struct
{
   Dev* dev;
   BOOLEAN ring_mode;
   unsigned int words_requested;
   unsigned int words_read;
   unsigned int buf_pos;
   unsigned int buf_next;
   unsigned short* buf[RING_BUFFER_COUNT];
   aio_result_t result[RING_BUFFER_COUNT];
} ring_info_t;

static ring_info_t* ring_info = 0;
static int ring_n_controllers = 0;
static int ring_maxwords = 0;
static int ring_buffer_size = 0;
static BOOLEAN ring_successful = FALSE;

static Dev devs[MAX_CONTROLLER_SEARCH];
static int devs_initialized = 0;

Devp
lio_open(const char* fn)
{
   int i;

   if (!devs_initialized)
   {
      memset(devs, 0, sizeof(devs));
      devs_initialized = 1;
   }
   
   if (fn)
   {
      printf("error: Filename not supported yet.\n");
      return 0;
   }
   else
   {
      for (i=0; i<MAX_CONTROLLER_SEARCH; i++)
      {
	 char tmp[80];

	 sprintf(tmp, "/dev/astro%d", i);
	 if (devs[i].in_use==0 && (devs[i].fd = open(tmp, O_RDWR|O_NONBLOCK, 0))!=-1)
	 {
	    printf("logonly: Found Leach card at %s, fd=%d\n", tmp, devs[i].fd);
	    devs[i].in_use = 1;
	    return (Devp)(&(devs[i]));
	 }
      }
   }
   return 0;
}

int
lio_close(Devp devp)
{
   Dev* dp = (Dev*)devp;

   if (dp==0 || dp->in_use!=1) return LIO_ERR;
   dp->in_use = 0;
   return close(dp->fd);
}

/*
 * This closes the device previously returned by lio_open.  The
 * return code is the same as the return code from close().
 */

int
lio_check(Devp devp) { return LIO_OK; }

union reply_buffer
{
   char ch[256];
   u_int uin[64];
};

union   reply_buffer dat;
int     *reply_arg = (int *)&dat.uin[0];
char    *reply_byte = (char *)&dat.ch[0];

#define DEFAULT_SBUS_REQUEST	0x80000003
#define SBUS_REQUEST_TIMEOUT	0x80040007
#define SBUS_REQUEST_WORD_WIDTH	0x80040004
#define SBUS_RESET		0x80000009

int
lio_reset(Devp devp, int sdsu_gen)
{
   Dev* dp = (Dev*)devp;
   int timeout_msec = 1000 * 20;
   int zero = 0;
   int rtn;

   if (dp==0 || dp->in_use!=1) return LIO_ERR;
   if (sdsu_gen != 2) return LIO_ERR;
   CLI_DEBUG(printf("debug: Resetting fd %d\n", dp->fd));
   rtn = ioctl(dp->fd, SBUS_RESET, reply_arg);
   if (rtn==-1)
   {
      perror("error: ioctl(SBUS_RESET)"); /* %%% */
      return LIO_ERR;
   }
   CLI_DEBUG(printf("debug: Setting timeout to %dms on fd %d\n", timeout_msec, dp->fd));
   if (ioctl(dp->fd, SBUS_REQUEST_TIMEOUT, &timeout_msec)==-1)
   {
      perror("error: ioctl(SBUS_REQUEST_TIMEOUT)"); /* %%% */
      return LIO_ERR;
   }
   if (ioctl(dp->fd, SBUS_REQUEST_WORD_WIDTH, &zero)==-1)
   {
      perror("error: ioctl(SBUS_REQUEST_WORD_WIDTH)"); /* %%% */
      return LIO_ERR;
   }
   return LIO_OK;
}

int
lio_write24(Devp devp, unsigned long* data, int nwords)
/*
 * Send 24-bit data to the lio.
 */
{
   Dev* dp = (Dev*)devp;
   int rtn;

   if (dp==0 || dp->in_use!=1) return LIO_ERR;
   rtn = ioctl(dp->fd, (nwords*4)<<16 | DEFAULT_SBUS_REQUEST, data);
   if (rtn==-1)
   {
      perror("error: ioctl(SBUS WRITE)"); /* %%% */
      return LIO_ERR;
   }
   return nwords;
}

int
lio_read24(Devp devp, unsigned long* buf,
	   int maxwords, int words, int response_msec)
{
   Dev* dp = (Dev*)devp;
   int rtn, rtn_errno, words_read = 0, keep_reading = 1;
   int i;
   struct sigaction oldsigalrm;

   if (dp==0 || dp->in_use!=1) return LIO_ERR;

   /* === Hack because leach driver can't do non-blocking reads */

   if (words > 0)
   {
      BOOLEAN alrm_was_blocked = cli_signal_isblocked(SIGALRM);

      sigaction(SIGALRM, 0, &oldsigalrm);
      cli_signal(SIGALRM, CLI_SIG_TRAP);
      if (alrm_was_blocked) cli_signal_unblock(SIGALRM);
      ualarm(response_msec * 1000 * words + 100, 100);
      rtn = read(dp->fd, buf, words * sizeof(long));
      rtn_errno = errno;
      ualarm(0, 0);
      sigaction(SIGALRM, &oldsigalrm, 0);
      if (alrm_was_blocked) cli_signal_block(SIGALRM);

      if (rtn == -1)
      {
	 if (rtn_errno !=  ETIME)
	 {
	    errno = rtn_errno;
	    perror("error: read");
	 }
	 return LIO_ERR;
      }

      if (rtn == 0)
      {
         keep_reading = 0;
	 printf("logonly: End-of-file!?\n");
	 return LIO_ERR;
      }

      if (rtn % sizeof(long) != 0)
      {
	 printf("error: got strange number of bytes (%d) from leach driver\n",
		rtn);
	 return LIO_ERR;
      }


      for (i=0; i<words; i++)
	 buf[i] &= 0xFFFFFF;
      return words;
   }

   /* Else, emulate a non-blocking read using lots of timeouts.  Yuk. */

   while (words_read < maxwords && keep_reading)
   {
      BOOLEAN alrm_was_blocked = cli_signal_isblocked(SIGALRM);
      unsigned long tmp;

      sigaction(SIGALRM, 0, &oldsigalrm);
      cli_signal(SIGALRM, CLI_SIG_TRAP);
      if (alrm_was_blocked) cli_signal_unblock(SIGALRM);
      ualarm(response_msec * 1000 + 100, 100);
      rtn = read(dp->fd, &tmp, sizeof(long));
      rtn_errno = errno;
      ualarm(0, 0);
      sigaction(SIGALRM, &oldsigalrm, 0);
      if (alrm_was_blocked) cli_signal_block(SIGALRM);

      switch (rtn)
      {
	 case -1: keep_reading = 0; 
	 if (rtn_errno != EINTR && rtn_errno != ETIME)
	 {
	    errno = rtn_errno;
	    perror("error: read");
	    return LIO_ERR;
	 }
	 break;
	 case 0: keep_reading = 0; printf("logonly: End-of-file!?\n"); break;
	 case 4: buf[words_read++] = tmp & 0xFFFFFF; break;
	 default: keep_reading = 0;
	 printf("error: Got odd read-size from leach driver\n");
      }
   }

   if (words_read == 0) return LIO_ERR; 
   return words_read;
}

#if 0
static short* tmp_buf = 0;
static size_t tmp_buf_size = 0;
/* %%% What's this for? */
#endif

#ifdef SOLARIS_PRIOCNTL
static long prio_rtn = 0;
static pcinfo_t pcinfo_rt;
static pcinfo_t pcinfo_ts;
static pcparms_t pcparms_rt;
static pcparms_t pcparms_ts;
#endif

int
lio_read16_init(int n_controllers, Devp devp[], int maxwords)
{
   int i, n;
   BOOLEAN same = ring_successful;

#ifdef SOLARIS_PRIOCNTL
   if (prio_rtn == 0)
   {
      strcpy(pcinfo_rt.pc_clname, "RT");
      strcpy(pcinfo_ts.pc_clname, "TS");
      if ((prio_rtn = priocntl(P_PID, P_MYID, PC_GETCID, (void*)&pcinfo_rt))>0)
	 if ((prio_rtn = priocntl(P_PID, P_MYID, PC_GETCID, (void*)&pcinfo_ts))>0)
	 {
	    rtparms_t* rtparms = (rtparms_t*)pcparms_rt.pc_clparms;
	    
	    pcparms_ts.pc_cid = pcinfo_ts.pc_cid;
	    prio_rtn = priocntl(P_PID, P_MYID, PC_GETPARMS, (void*)&pcparms_ts);
	    pcparms_rt.pc_cid = pcinfo_rt.pc_cid;
	    rtparms->rt_pri = 20;
	    rtparms->rt_tqsecs = 0;
	    rtparms->rt_tqnsecs = 100000; /* .1 msec time quantum */
	 }
   }
   /*
    * Switch to realtime mode, if possible
    */
   if (prio_rtn > 0)
   {
      seteuid(0);
      if ((prio_rtn = priocntl(P_PID, P_MYID, PC_SETPARMS, (void*)&pcparms_rt))==-1)
	 perror("warning: Failed to set realtime priority");
      else prio_rtn = 1;
      seteuid(getuid());
   }
#endif
   
   if (ring_n_controllers != n_controllers) same = FALSE;
   if (ring_maxwords != maxwords) same = FALSE;
   for (n = 0; n < n_controllers && same!=FALSE; n++)
      if (ring_info[n].dev != (Dev*)(devp[n])) same = FALSE;
   
   if (same==FALSE)
   {
      printf("logonly: [lio_read16_init] Setting up ring buffers\n");
      /*
       * Free the buffers allocated last time around if called
       * again and edt devices were not the same or there were more/less.
       */
      if (ring_info != 0)
      {
	 for (n = 0; n < ring_n_controllers; n++)
	 {
	    if (ring_info[n].buf[0] != 0) free(ring_info[n].buf[0]);
	 }
	 free(ring_info);
      }
      
      /*
       * (Re)allocate buffers for specified number of controllers.
       */
      ring_n_controllers = n_controllers;
      ring_maxwords = maxwords;
      ring_buffer_size = maxwords * sizeof(short);

      ring_info = (ring_info_t*)malloc(n_controllers * sizeof(ring_info_t));
      memset(ring_info, 0, sizeof(ring_info));

#if 0
      {
	 struct rlimit rl;
	 getrlimit(RLIMIT_DATA, &rl);
	 printf("RLIMIT_DATA = %ld/%ld\n", rl.rlim_cur, rl.rlim_max);
	 getrlimit(RLIMIT_VMEM, &rl);
	 printf("RLIMIT_VMEM = %ld/%ld\n", rl.rlim_cur, rl.rlim_max);
	 getrlimit(RLIMIT_AS, &rl);
	 printf("RLIMIT_AS = %ld/%ld\n", rl.rlim_cur, rl.rlim_max);
      }
#endif

      for (n = 0; n < n_controllers; n++)
      {
	 ring_info[n].ring_mode = FALSE;
	 ring_info[n].dev = (Dev*)devp[n];
	 /*
	  * Allocate temporary ring buffer space for the EDT driver.
	  * === Unfortunately this requires another copy of all the
	  * data to be performed by the CPU to get it into the user's
	  * buffer space.
	  */
	 ring_info[n].buf[0] = (unsigned short*)valloc(ring_buffer_size * RING_BUFFER_COUNT_MALLOC);
	 if (ring_info[n].buf[0]==0)
	 {
	    perror("error: valloc valloc");
	    printf("error: Unable to allocate %d MB ring buffer space\n",
		   ring_buffer_size * RING_BUFFER_COUNT_MALLOC / (1024*1024));
	    return LIO_ERR;
	 }

	 seteuid(0);
	 if (mlock((void*)(ring_info[n].buf[0]), ring_buffer_size * RING_BUFFER_COUNT_MALLOC) < 0)
	 {
	    perror("error: mlock failed");
	 }
	 seteuid(getuid()); /* %%% */

	 CLI_DEBUG(printf("debug: Clearing ring buffers\n")); /* To keep purify happy */
	 memset(ring_info[n].buf[0], 0, ring_buffer_size * RING_BUFFER_COUNT_MALLOC );
	 CLI_DEBUG(printf("debug: Done clearing\n"));
	 for (i = 0; i < RING_BUFFER_COUNT; i++)
	 {
	    ring_info[n].result[i].aio_return = 0;
	    ring_info[n].buf[i] =
	       (unsigned short*)((void*)(ring_info[n].buf[0]) + ring_buffer_size*i);
	 }
      }
      printf("logonly: [lio_read16_init] Buffer space allocated\n");
   }
   else
   {
      for (n=0; n<n_controllers; n++)
      {
	 CLI_DEBUG(printf("debug: Clearing ring buffers\n")); /* To keep purify happy */
	 memset(ring_info[n].buf[0], 0, ring_buffer_size * RING_BUFFER_COUNT_MALLOC );
	 CLI_DEBUG(printf("debug: Done clearing\n"));
      }
   }

   for (n=0; n<n_controllers; n++) if (ring_info[n].ring_mode == FALSE)
   {
      /*
       * Zero the counters in the ring_info structs.
       */
      ring_info[n].words_requested = 0;
      ring_info[n].words_read = 0;
      ring_info[n].buf_pos = 0;
      ring_info[n].buf_next = 0;

      /*
       * Save a flag so this doesn't have to be repeated.
       */
      ring_info[n].ring_mode = TRUE;
      {
	 Dev* dp = (Dev*)(devp[n]);
	 int timeout_msec = 1000 * 300;
	 int one = 1;
	 
	 if (dp==0 || dp->in_use!=1) return LIO_ERR;
	 CLI_DEBUG(printf("debug: Setting timeout to %dms on fd %d\n", timeout_msec, dp->fd));
	 if (ioctl(dp->fd, SBUS_REQUEST_TIMEOUT, &timeout_msec)==-1)
	 {
	    perror("error: ioctl(SBUS_REQUEST_TIMEOUT)"); /* %%% */
	    return LIO_ERR;
	 }
	 if (ioctl(dp->fd, SBUS_REQUEST_WORD_WIDTH, &one)==-1)
	 {
	    perror("error: ioctl(SBUS_REQUEST_WORD_WIDTH)"); /* %%% */
	    return LIO_ERR;
	 }
      }
   }

   ring_successful = TRUE;
   return LIO_OK;
}

void
lio_read16_exit(void)
{
   int i, n;

   for (n = 0; n < ring_n_controllers; n++) if (ring_info[n].ring_mode == TRUE)
   {
      {
	 Dev* dp = ring_info[n].dev;
	 int zero = 0;
	 
	 if (ioctl(dp->fd, SBUS_REQUEST_WORD_WIDTH, &zero)==-1)
	 {
	    perror("error: ioctl(SBUS_REQUEST_WORD_WIDTH)"); /* %%% */
	 }
      }

      ring_info[n].ring_mode = FALSE;

      for (i = 0; i < RING_BUFFER_COUNT; i++)
      {
	 if (ring_info[n].result[i].aio_return == AIO_INPROGRESS)
	 {
	    if (aiocancel(&(ring_info[n].result[i]))!=0)
	    {
	       perror("warning: aiocancel");
	    }
	    ring_info[n].result[i].aio_return = 0;
	 }
      }
   }
#ifdef SOLARIS_PRIOCNTL
   /*
    * Switch back to timeshare mode
    */
   if (prio_rtn > 0)
   {
      seteuid(0);
      if ((prio_rtn = priocntl(P_PID, P_MYID, PC_SETPARMS, (void*)&pcparms_ts))==-1)
	 perror("warning: Failed to return to timesharing priority mode");
      seteuid(getuid());
   }
#endif
}

int
lio_read16_parallel(int n_controllers, Devp devp[], read16_handler handler,
		    int maxwords, int response_msec, int timeout_msec)
{
   int time_stop;
   int time_response;
   int curr_words_read = 0;
   int n;

   /* %%% Hack: Leach driver returns entire block at once, so use
   * timeout_msec for response_msec also.
   */
   response_msec = timeout_msec;
   time_response = time_stop = time(0);
   time_response += response_msec / 1000;
   time_stop += timeout_msec / 1000;

#if 0
   {
      if (tmp_buf_size != maxwords*sizeof(short))
      {
	 if (tmp_buf) free(tmp_buf);
	 tmp_buf_size = maxwords*sizeof(short);
	 CLI_DEBUG(printf("debug: Allocating tmp_buf %d bytes...\n", tmp_buf_size));
	 tmp_buf = (short*)malloc(tmp_buf_size);
	 memset(tmp_buf, 0, tmp_buf_size);
	 CLI_DEBUG(printf("debug: done.\n"));
      }
      for (n=0; n<n_controllers; n++)
      {
	 int rtn = read(((Dev*)devp[n])->fd, tmp_buf, tmp_buf_size);
	 
	 if (rtn == -1) perror("error: read");
	 else if (rtn != tmp_buf_size)
	 {
	    printf("error: got %d instead of %d bytes\n", rtn, tmp_buf_size);
	 }
	 else handler(n, tmp_buf, maxwords);
      }
   }
   lio_read16_exit();
   return maxwords*n_controllers;
#endif

   /*
    * Make sure the card is in ring buffer mode.  To be ready for an
    * immediate response, the higher level code should already have
    * called this, in which case this call does nothing here.
    */
   lio_read16_init(n_controllers, devp, maxwords);

   if (maxwords==0)
   {
      printf("error: [lio_read16_parallel] request for 0 words.\n");
      lio_read16_exit();
      return LIO_ERR;
   }

   CLI_DEBUG(printf("debug: [lio_read16_parallel] ready for input\n"));

   do
   {
      int delay;
      int more;

      delay = 100000;

      /*
       * Start asynchronous reads into as many of the buffers as
       * are ready (all 4 on first pass, then 1 at a time.)
       */
      do
      {
	 more = 0;
	 for (n=0; n<n_controllers; n++)
	 {
	    unsigned int bn;

	    bn = (ring_info[n].buf_next + 1)%RING_BUFFER_COUNT;
	    if (bn != ring_info[n].buf_pos)
	    {
	       int sz, bp;
	       
	       bp = ring_info[n].buf_next;
	       sz = maxwords - ring_info[n].words_requested;
	       if (sz > ring_buffer_size/sizeof(short))
		  sz = ring_buffer_size/sizeof(short);
	       if (sz >  0)
	       {
		  CLI_DEBUG(printf("debug: Request %d pixels from controller %d fd %d in buffer %d\n",
				   sz, n, ring_info[n].dev->fd, bp));
		  more = 1;
		  ring_info[n].result[bp].aio_return = AIO_INPROGRESS;
		  if (aioread(ring_info[n].dev->fd,
			      (char*)ring_info[n].buf[bp], sz * sizeof(short),
			      0, 0, &(ring_info[n].result[bp]))!=0)
		  {
		     perror("error: aioread");
		     lio_read16_exit();
		     return LIO_ERR;
		  }
		  xusleep(1000000); /* %%% */
		  CLI_DEBUG(printf("debug: Done with request\n"));
		  ring_info[n].words_requested += sz;
		  ring_info[n].buf_next = bn;
	       }
	    }
	 }
      } while (more);

      /*
       * Wait for the requests to complete.
       */
      {
	 struct timeval tv = { 0, 0 };
	 aio_result_t* result;
	 
	 result = aiowait(&tv); /* tv is 0, so this is a poll */
	 if (result == (aio_result_t*)(-1))
	 {
	    perror("error: aiowait");
	    lio_read16_exit();
	    return LIO_ERR;
	 }

	 if (result != 0)
	 {
	    int found = 0;

	    if (result->aio_return == -1)
	    {
	       errno = result->aio_errno;
	       switch (errno)
	       {
		  case ENOTEMPTY:
		  fprintf(stderr,
			  "error: Leach SBus driver says: \"Bad interrupt\" "
			  "(CPU too slow?)\n");
		  break;
		  case ENOTBLK:
		  fprintf(stderr,
			  "error: Leach SBus driver says: \"Device not busy\"\n");
		  break;
		  case EDEADLK:
		  fprintf(stderr,
			  "error: Leach SBus driver says: \"Image buffer not busy\"\n");
		  break;
		  case ENOLCK:
		  fprintf(stderr,
			  "error: Leach SBus driver says: \"Data transmission error\"\n");
		  break;
		  case ETIME:
		  fprintf(stderr,
			  "error: Leach SBus driver says: \"Timeout\"\n");
		  break;
		  default:
		  perror("error: read");
	       }
	       lio_read16_exit();
	       return LIO_ERR;
	    }

	    /*
	     * Find out which controller's asynchoronous read came back
	     */
	    for (n=0; n<n_controllers; n++)
	    {
	       int bp;

	       if (ring_info[n].buf_pos != ring_info[n].buf_next)
	       {
		  bp = ring_info[n].buf_pos;
		  if (result == &(ring_info[n].result[bp]))
		  {
		     int sz = maxwords - ring_info[n].words_read;

		     if (sz > ring_buffer_size/sizeof(short))
			sz = ring_buffer_size/sizeof(short);

		     if (result->aio_return != sz * sizeof(short))
		     {
			printf("error: read returned %d bytes, expected %d\n",
			       result->aio_return, sz * sizeof(short));
			lio_read16_exit();
			return LIO_ERR;
		     }
		     CLI_DEBUG(printf("debug: Receive %d pixels from controller %d in buffer %d (%x)\n",
				      sz, n, bp, (unsigned int)result));
		     handler(n, ring_info[n].buf[bp], sz);
	 	     ring_info[n].words_read += sz;
		     curr_words_read += sz;
		     ring_info[n].buf_pos = (bp + 1) % RING_BUFFER_COUNT;
		     found = 1;
		  }
	       }
	    }
	    if (!found)
	    {
	       printf("error: aiowait returned wrong buffer!\n");
	       printf("logonly: Buffer=%x\n", (unsigned int)result);
	       
	       for (n=0; n<n_controllers; n++)
	       {
		  int bp;

		  for (bp = 0; bp<RING_BUFFER_COUNT; bp++)
		     printf("logonly: controller %d buffer %d%s = %x\n",
			    n, bp, (bp==ring_info[n].buf_pos)?"(NEXT)":"",
			    (unsigned int)&(ring_info[n].result[bp]));
	       }
	       lio_read16_exit();
	       return LIO_ERR;
	    }
	 }
      }

      if (curr_words_read == maxwords * n_controllers)
      {
	 printf("progress: Readout complete\n");
	 break;
      }

      /* %%% aiocancel? */
      /* %%% timeout? */

      if (curr_words_read == 0)
	 printf("progress: Waiting for pixels\r");
      else
      {
	 if (maxwords >= 100 * 100)
	    printf("progress: Reading out camera (%d%%)\r",
		   curr_words_read/(maxwords*n_controllers/100));
	 else
	    printf("progress: Reading out camera (%d%%)\r",
		   curr_words_read*100/(maxwords*n_controllers));
      }
      fflush(stdout);

      cli_signal_unblock(SIGINT);
      cli_signal_unblock(SIGQUIT);
      cli_signal_unblock(SIGTSTP);
      xusleep(delay);
      cli_signal_block(SIGINT);
      cli_signal_block(SIGQUIT);
      cli_signal_block(SIGTSTP);
      if (cli_signal_event(SIGINT))
      {
#ifdef ALLOW_READOUT_ABORTS
	 printf("error: Readout aborted by user!\n");
	 lio_read16_exit();
	 return LIO_ERR;
#else
	 printf("warning: Cannot abort now.  Please wait for readout to finish.\n");
#endif
      }
      if (cli_signal_event(SIGTSTP))
      {
	 printf("warning: Readout already in progress; `stop' command ignored.\n");
      }
   } while (time(0)<((curr_words_read==0)?time_response:time_stop));

   if (time(0) >= time_stop)
   {
      CLI_DEBUG(printf("debug: reached maximum time of %g seconds\n", timeout_msec/1000.));
   }
   else if (time(0) >= time_response && curr_words_read==0)
   {
      CLI_DEBUG(printf("debug: no response within %d msec\n", response_msec));
   }

   lio_read16_exit();
   return curr_words_read;
}

