#ifndef _INCLUDED_lotuspci
#define _INCLUDED_lotuspci 1

#define DEVICE_NAME "lotuspci"
#define DEVICE_NAME_LEN 8
#define DEVICE_MAJOR 123 /* Set to 0 to be dynamically assigned */

typedef struct
{
      unsigned long phys_addr;
      unsigned long size;
} ssp_dma_mem_t;

/*
 * DMA ring-buffer mode %%% Not implemented in this driver
 * --------------------
 *
 * There is still only one contiguous bigphysarea buffer in this mode.
 * It is divided into equal sized  regions with separate DMA transfers.
 * Eventually, user code could be signalled when each DMA transfer
 * is complete.  This is not implemented, but the driver does catch
 * the interrupt and move the DMA window.
 *
 * Example 1: Let's say the total bigphysarea is 1000K and as it
 * gets filled, we want to re-initialize the DMA transfer at
 * 250K, 500K, and 750K.  (Right now, user code would have to
 * poll the driver to see how many of the 4 chunks have been
 * completed.)
 *
 *   dma_buf_size = 1000K;
 *   ring.n_bufs   = 4;
 *   ring.buf_size = ring.stride     = 250K;
 *   ring.bytes_total = ring.n_bufs * ring.buf_size;
 *   ring.start = 0;
 *   ring.shift = 0;
 *
 * Example 2: Just as example 1, except when we complete the 4th
 * iteration, we want to wrap around and re-use the space used by
 * the 1st iteration.  (User code could be polling the current
 * buffer number and processing data... a mechanism to detect
 * overruns needs to be implemented.)
 * 
 *   dma_buf_size = 1000K;
 *   ring.n_bufs = 4;
 *   ring.buf_size = ring.stride     = 250K;
 *   ring.bytes_total = 0; // Continuous mode; only difference from example 1
 *   ring.start = 0;
 *   ring.shift = 0;
 *
 * NOTE -- For the above examples:
 *
 *   stride==buf_size
 *   shift==0
 *   start==0, and
 *   bytes_total=n_bufs*buf_size (or 0 for continuous mode.)
 *
 * These parameters exist to handle a specific kind of block
 * descrambling for multi-readout CCD acquisition systems, which
 * bring us to Example 3: 44 CCD amplifiers will produce 8M of
 * data each.  The data will be interleaved as 44 2K blocks from
 * each amplifier in a repeating cycle.  The data must arrive in
 * memory sorted into contiguous 8M buffers corresponding to
 * each amplifier.  This kind of descrambling is accomplished
 * with the following settings:
 *
 *   dma_buf_size = 44*8M;
 *   ring.n_bufs = 44;
 *   ring.buf_size = 2K;
 *   ring.stride = 8M;
 *   ring.bytes_total = 44*4000*2K;
 *   ring.start = 0;
 *   ring.shift = 2K;
 *
 * If you don't require this kind of descrambling, just follow the
 * recipe for setting stride, shift, start, and iterations given
 * in the note above.
 */

typedef struct
{
      int bytes_total;		/* usually=n_bufs*buf_size (0=continuous) */
      off_t start;		/* Starting offset in dma_buf (usually 0) */
      off_t stride;		/* buffer1 - buffer0 (0=same as buf_size) */
      off_t shift;		/* Offset to add after wrapping (usually 0) */
      size_t buf_size;		/* Size of each DMA buffer (0=use maximum) */
      unsigned int n_bufs;	/* Number of DMA ring buffers (0=calculate) */
} ssp_ring_t;

typedef struct
{
      enum { DSP_MEM_P=0x100000,
	     DSP_MEM_X=0x200000,
	     DSP_MEM_Y=0x400000,
      	     DSP_MEM_R=0xF00000} type;
      unsigned int addr;
      unsigned int data;
} dsp_mem_t;

#define LTMAGIC 0xdf

#define LTPCI_RESET	  _IO(LTMAGIC, 0) /* Hard reset AMCC and S-Link */
#define LTPCI_DWORD_32	  _IO(LTMAGIC, 1) /* Select 32-bit data (default) */
#define LTPCI_DWORD_16	  _IO(LTMAGIC, 2) /* Select 16-bit data */
#define LTPCI_DWORD_8	  _IO(LTMAGIC, 3) /* Select 8-bit data */
#define LTPCI_UTDO_ACTIVE _IO(LTMAGIC, 4) /* Select SLINK test data output */
#define LTPCI_UTDO_NORMAL _IO(LTMAGIC, 5) /* Normal operation (default) */
#define LTPCI_SET_URL	  _IOW(LTMAGIC, 6, int) /* Set URL bits (4) */
#define LTPCI_GET_URL	  _IOR(LTMAGIC, 7, int) /* Get current URL bits */
#define LTPCI_GET_STATUS  _IOR(LTMAGIC, 8, ssp_status_t) /* Get status bits */
#define LTPCI_SET_DMAMEM  _IOW(LTMAGIC, 9, ssp_dma_mem_t) /* Set DMA buffer */
#define LTPCI_DMA_READ	  _IOW(LTMAGIC, 10, unsigned int) /* DMA read */
#define LTPCI_DMA_RING	  _IOW(LTMAGIC, 11, ssp_ring_t) /* Ring buffer mode */
#define LTPCI_DMA_WAIT	  _IOW(LTMAGIC, 12, unsigned int) /* DMA wait */
#define LTPCI_GET_LATENCY _IOR(LTMAGIC, 13, unsigned char) /* Get latency */
#define LTPCI_SET_LATENCY _IOW(LTMAGIC, 14, unsigned char) /* Set latency */
#define LTPCI_RDMEM	  _IOWR(LTMAGIC, 15, dsp_mem_t) /* PCI DSP read */
#define LTPCI_WRMEM	  _IOWR(LTMAGIC, 16, dsp_mem_t) /* PCI DSP write */
#define LTPCI_DOWNLOAD_START _IO(LTMAGIC, 17)
#define LTPCI_DOWNLOAD_STOP  _IO(LTMAGIC, 18)
#define LTPCI_DOWNLOAD_WORD  _IOW(LTMAGIC, 19, unsigned int)
#define LTPCI_NOSWAP        _IO(LTMAGIC, 20) /* No byte-swapping */
#define LTPCI_16SWAP        _IO(LTMAGIC, 21) /* Select 16-bit (2-byte) swapping */
#define LTPCI_GEN1_MODE     _IO(LTMAGIC, 22) /* Select SDSU GEN1 fiber mode */
#define LTPCI_GEN2_MODE     _IO(LTMAGIC, 23) /* Select SDSU GEN2 fiber mode */
#define LTPCI_DETCOM	    _IO(LTMAGIC, 24) /* Put driver in DetCom mode */
#define LTPCI_RESET_CONTROLLER    _IO(LTMAGIC, 25) /* Reset timing board */
#define LTPCI_GET_VERSION   _IOR(LTMAGIC, 99, int) /* Read back driver's version */
#endif
