/*                                             -*- c-file-style: "Ellemtel" -*-

`fhtool.c' - Validates and converts basic and multi-extension FITS

This file is part of version 1 of the CFHT FITS Validation and Conversion tool.
Read the `License' file for terms of use and distribution.
Copyright 2001, Canada-France-Hawaii Telescope, daprog@cfht.hawaii.edu.

___This header automatically generated from `Index'. Do not edit it here!___ */

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <dirent.h>

#include <missing_protos.h> /* Only needed on SUNOS 4.x so far */

#include "fh/fh.h" /* Requires FITS header library */

#define MAX_EXTENSIONS 999

/*
 * Usage:

fhtool subdir new-mef.fits
fhtool subdir > new-mef.fits
fhtool mef[.fits] new-subdir
fhtool -Vaq subdir-or-file

*/

/*
 * ----------------------------------------------------------
 * Internal Implementation (these are at the end of the file)
 * ----------------------------------------------------------
 */

static int xstrcasecmp(const char *s, const char *d);
/*
 * Case-insensitive string-compare function.
 */

static int pstrcasecmp(const void* p1, const void* p2);
/*
 * For passing to qsort()
 */

static char* get_basename(const char* pathname, const char* ext, int strip_path);
/*
 * Allocates a new string and puts the basename of pathname inside.
 * Any path elements containing '/' are removed, along with the
 * optional string `ext' if it occurs on the end of the name.
 */

static HeaderUnit phu_from_phu_file(const char* filename);
/*
 * Read a .phu file into a HeaderUnit.  The return of this function
 * (if non-zero) should be passed to fh_destroy() when no longer needed.
 */

static int open_file(const char* filename_requested, const char* default_ext,
		     char** filename_found, int default_fd, int open_mode);
/*
 * Returns -1 if file is not found.
 * Returns -2 if file found but is a directory.
 * Opens file and returns file descriptor in other cases.
 */

static void close_file(int fd, const char* filename, int repair);
/*
 * Will exit!!! if the file is not at EOF when this is called. %%%
 */

static int quiet_flg = 0;
static int sync_flg = 0;

typedef char* charptr;

/*
 * The following is for statistics only.
 */
static struct timeval time_start;
size_t bytes_written = 0;
static void
print_statistics(void)
{
   int milliseconds;
   struct timeval time_stop;

   if (sync_flg)
   {
      if (!quiet_flg)
	 fprintf(stderr, "fhtool: Synching disk\n");
      sync();
   }
   gettimeofday(&time_stop,NULL);
   milliseconds = (time_stop.tv_sec - time_start.tv_sec) * 1000 +
      (time_stop.tv_usec - time_start.tv_usec) / 1000;
   if (milliseconds && bytes_written)
   {
      fprintf(stderr,
	      "fhtool: Approximate data rate %4.2f MBytes/sec. Use `-q' for quiet operation.\n",
	      (double)((double)(bytes_written / 1024.0) / (double)milliseconds));
   }
}

int
main(int argc, const char* argv[])
{
   char* filename_in;	/* File name of the input file */
   char* filename_out;	/* File name of the output file (or subdirectory) */
   int fd_in = -1;	/* File descriptor of the input file */
   int fd_out = -1;	/* File descriptor of the output file */
   int i;
   enum { SPLIT_TO_MEF, MEF_TO_SPLIT } mode = MEF_TO_SPLIT;
   /*
    * Option parsing stuff:
    */
   int c, errflg = 0, phu_reconstruct_flg = 0, repair_flg = 0, validate_flg = 0;

   if (!isatty(STDERR_FILENO)) quiet_flg = 1;
   else quiet_flg = 0;

   /*
    * Parse the command line options.
    */
   while ((c = getopt(argc, (char**)argv, "qsPRhVv")) != -1)
   {
      switch (c)
      {
	 case 'q': quiet_flg = 1; break;
         case 'v': quiet_flg = 0; break;
	 case 's': sync_flg = 1; break;
#define REPAIR_FILES
#ifdef REPAIR_FILES
	 case 'P': phu_reconstruct_flg = 1; break;
	 case 'R': repair_flg = 1; break;
#endif
	 case 'h': errflg++; break;
	 case 'V': validate_flg++; break;
	 case ':':
         case '?': errflg++; break;
      }
   }

   /*
    * Skip over the optional arguments.
    */
   if (optind)
   {
      argc -= (optind - 1);
      argv += (optind - 1);
   }

   if (errflg || argc < 1 || (argc < 2 && !validate_flg) || argc > 3)
   {
      fprintf(stderr,
	      "fhtool: CFHT MEF conversion utility version "VERSION_STRING"\n"
	      "usage: fhtool [options] inputfile-or-dir [outputfile-or-dir]\n"
	      "command line options:\n"
	      "     -q       Run quietly (the default if not running on a terminal)\n"
              "     -v       Run verbosely (the default if running on a terminal)\n"
	      "     -s       Sync disks before and after (for measuring performance)\n"
#ifdef REPAIR_FILES
	      "     -P       Reconstruct Primary Header Units if *.phu file is missing\n"
	      "     -R       Repair padding errors in source FITS files\n"
#endif
              "     -V       Validate integrity of the input file(s) only\n"
	      "\n"
	      "example of MEF->SPLIT conversion:\n"
	      "# fhtool file.fits [dir]\n"
	      "  `file.fits' (MEF) will be split into `dir/file.phu', `dir/file00.fits' ...\n"
	      "  `file.fits' is not modified (it is opened read-only).\n"
	      "  If the `dir' argument is omitted, the basename of the source file is used.\n"
	      "  If an old `dir' exists, it must be removed before running fhtool.\n"
	      "\n"
	      "example of SPLIT->MEF conversion:\n"
	      "# fhtool dir file.fits\n"
	      "  `dir/dir.phu' (if present) and `dir/*.fits' are assembled into `file.fits'\n"
	      "  `dir/*' are not modified (files are opened read-only).\n"
	      "\n"
	      "Generating SPLIT from a mef file on another machine:\n"
	      "# runon -n somehost cat file.fits | fhtool - dir\n"
	      "  This can be used to read an MEF file from another machine and save it in\n"
	      "  split format on the fly.\n"
	      "\n"
	      "Generating MEF from split files on another machine:\n"
	      "# runon -n otherhost fhtool dir > file.fits\n");
      exit(EXIT_FAILURE);
   }

   fd_in = open_file((argc<2)?"-":argv[1], ".fits", &filename_in, STDIN_FILENO, O_RDONLY);

   switch (fd_in)
   {
      case -1:
	 perror("error: fhtool: open failed");
	 fprintf(stderr, "error: fhtool: Failed to open `%s' for input.\n",
		 (argc<2)?"-":argv[1]);
	 exit(EXIT_FAILURE);
	 break;
      case -2:  mode = SPLIT_TO_MEF; break;
      default: mode = MEF_TO_SPLIT;
   }

   if (!quiet_flg)
      fprintf(stderr, "fhtool: Version "VERSION_STRING" ("BUILDDATE") Use `-q' for quiet operation.\n");

   if (sync_flg)
   {
      if (!quiet_flg)
	 fprintf(stderr, "fhtool: Synching disk\n");
      sync();
   }

   gettimeofday(&time_start, 0);
   /*
    * -------------------
    * Handle SPLIT -> MEF
    * -------------------
    */
   if (mode == SPLIT_TO_MEF)
   {
      HeaderUnit phu;
      DIR* dir;
      struct dirent* dirent;
      charptr dir_listing[MAX_EXTENSIONS];
      int extensions = 0;
      char* filename_base;
      char* filename_tmp;

      /*
       * Rather than assuming the files are numbered 00 01 02 ...,
       * read the directory instead, and then sort it with qsort.
       */
      if (!(dir = opendir(filename_in)))
      {
	 perror(filename_in);
	 exit(EXIT_FAILURE);
      }
      while ((dirent = readdir(dir)))
      {
	 int namelen;
	 
	 if (dirent->d_name[0] == '.') continue; /* Ignore ., .. and other dot-files */
	 /*
	  * Only count *.fits or *.fit files
	  */
	 namelen = strlen(dirent->d_name);
	 if ((namelen > 5 && !xstrcasecmp(dirent->d_name + namelen - 5, ".fits")) ||
	     (namelen > 4 && !xstrcasecmp(dirent->d_name + namelen - 4, ".fit")))
	 {
	    if (extensions >= MAX_EXTENSIONS)
	    {
	       fprintf(stderr,
		       "error: fhtool: Cannot make MEF file with more than %d extensions\n",
		       MAX_EXTENSIONS);
	       exit(EXIT_FAILURE);
	    }
	    dir_listing[extensions++] = strdup(dirent->d_name);
	 }
	 /* else === Non-fits file in directory */
      }
      if (closedir(dir) < 0)
      {
	 perror("error: fhtool: closedir failed");
	 exit(EXIT_FAILURE);
      }

      if (!extensions)
      {
	 fprintf(stderr, "error: fhtool: No *.fits files found in `%s/'\n",
		 filename_in);
	 exit(EXIT_FAILURE);
      }

      /*
       * Sort the directory listing.
       */
      qsort(dir_listing, extensions, sizeof(charptr), pstrcasecmp);

      filename_base = get_basename(filename_in, ".fits", /*strip_path=*/1);
      filename_tmp = (char*)malloc(strlen(filename_in) + strlen(filename_base) + 32);

      /*
       * Look for a .phu file first.  Only lower-case *.phu is matched.
       */
      sprintf(filename_tmp, "%s/%s.phu", filename_in, filename_base);
      phu = phu_from_phu_file(filename_tmp);

      if (phu)
      {
	 int phu_extensions = 0;

	 if (fh_get_int(phu, "NEXTEND", &phu_extensions) != FH_SUCCESS ||
	     phu_extensions != extensions)
	 {
	    fprintf(stderr, "error: fhtool: `%s' says %d extensions but found %d FITS files.\n",
		    filename_tmp, phu_extensions, extensions);
	    exit(EXIT_FAILURE);
	 }
      }
      else
      {
	 if (!phu_reconstruct_flg)
	 {
#ifdef REPAIR_FILES
	    fprintf(stderr,
		    "error: fhtool: Run again with -P to reconstruct a primary header.\n");
#endif
	    exit(EXIT_FAILURE);
	 }
      }

      if (validate_flg)
      {
	 filename_out = strdup("/dev/null");
	 fd_out = open("/dev/null", O_RDWR);
      }
      else
      {
	 filename_out = 0;
	 fd_out = open_file((argc<3)?"-":argv[2], ".fits", &filename_out,
			    STDOUT_FILENO, O_RDWR|O_CREAT|O_EXCL);
      }

      if (fd_out < 0)
      {
	 switch (errno)
	 {
	    case EEXIST:
	       fprintf(stderr, "error: fhtool: A file `%s' already exists.  Remove it first.\n",
		       filename_out);
	       break;
	    case EPERM:
	       fprintf(stderr, "error: fhtool: No permission to create `%s' here.\n",
		       filename_out);
	       break;
	    default:
	       if (!filename_out)
		  fprintf(stderr, "error: No output file given.\n");
	       else
		  perror(filename_out);
	 }
	 exit(EXIT_FAILURE);
      }

      if (!quiet_flg)
      {
	 fprintf(stderr, "fhtool: Reading files `%s/%s*' ",
		 filename_in, filename_base);
	 if (validate_flg)
	    fprintf(stderr, "(validating only)\n");
	 else
	    fprintf(stderr, "and creating MEF file `%s'\n", filename_out);
      }

      for (i=0; i<extensions; i++)
      {
	 HeaderUnit ehu;
	 char detsize[FH_MAX_STRLEN];	/* Save DETSIZE from PHU */
	 double idx_detector;	/* And insert after DETECTOR if missing from an EHU */
	 double idx_naxis2 = 0;

	 sprintf(filename_tmp, "%s/%s", filename_in, dir_listing[i]);
	 if ((fd_in = open(filename_tmp, O_RDONLY, 0))==-1)
	 {
	    perror(filename_tmp);
	    exit(EXIT_FAILURE);
	 }
	 if (!quiet_flg)
	    fprintf(stderr, "fhtool: Reading file `%s'\n", filename_tmp);
	 ehu = fh_create();
	 if (fh_read(ehu, fd_in, FH_AUTO) != FH_SUCCESS)
	 {
	    fprintf(stderr, "error: fhtool: Failed to read extension header unit from %s\n",
		    filename_tmp);
	    exit(EXIT_FAILURE);
	 }
	 if (i==0)
	 {
	    /*
	     * Generate a phu from the first extension, if needed.
	     */
	    if (!phu)
	    {
	       double idx;
	       char detector[80];
	       
	       phu = fh_create();
	       fh_merge(phu, ehu);
	       /*
		* The following are not applicable to the primary header:
		*/
	       fh_remove(phu, "EXTNAME");
	       fh_remove(phu, "NAXIS1");
	       fh_remove(phu, "NAXIS2");
	       fh_remove(phu, "NAXIS3");
	       fh_remove(phu, "PCOUNT");
	       fh_remove(phu, "GCOUNT");
	       fh_remove(phu, "INHERIT");
	       fh_remove(phu, "IMAGEID");
	       fh_remove(phu, "CHIPID");
	       fh_remove(phu, "GAIN");
	       fh_remove(phu, "BZERO");
	       fh_remove(phu, "BSCALE");
	       fh_remove(phu, "DATASEC");
	       fh_remove(phu, "DETSEC");
	       fh_remove(phu, "BIASSEC");
	       fh_remove(phu, "CTYPE1");
	       fh_remove(phu, "CTYPE2");
	       fh_remove(phu, "CRVAL1");
	       fh_remove(phu, "CRVAL2");
	       fh_remove(phu, "CRPIX1");
	       fh_remove(phu, "CRPIX2");
	       fh_remove(phu, "CD1_1");
	       fh_remove(phu, "CD1_2");
	       fh_remove(phu, "CD2_1");
	       fh_remove(phu, "CD2_2");
	       fh_remove(phu, "LIM1_1");
	       fh_remove(phu, "LIM2_2");
	       fh_remove(phu, "WAT0_001");
	       fh_remove(phu, "WAT1_001");
	       fh_remove(phu, "WAT2_001");
	       
	       /*
		* Remove these also, before re-setting them, otherwise
		* the sorting number has no effect.
		*/
	       fh_remove(phu, "SIMPLE");
	       fh_remove(phu, "BITPIX");
	       fh_remove(phu, "NAXIS");
	       
	       /*
		* %%% Set DETSIZE NCCDS NAMPS for UH8K also.
		*/
	       
	       fh_set_bool(phu, 0, "SIMPLE", 1,           "Standard FITS");
	       fh_set_int( phu, 1, "BITPIX", 16,          "Bits per pixel (not applicable)");
	       fh_set_int( phu, 2, "NAXIS",   0,          "No image data with primary header");
	       fh_set_bool(phu, 3, "EXTEND",  1,          "File contains extensions");
	       fh_set_int( phu,3.1,"NEXTEND", extensions, "Number of extensions");
	       fh_set_com( phu, 99,"COMMENT", " Primary Header Unit generated by fhtool v"VERSION_STRING);
	       fh_set_com( phu, 99.00001,"COMMENT", " Some comments in this header may only apply to CCD0");
	       if (fh_search(phu, "DETECTOR", &idx) == FH_SUCCESS &&
		   fh_get_str(phu, "DETECTOR", detector, sizeof(detector)) == FH_SUCCESS &&
		   !strcmp(detector, "CFH12K"))
	       {
		  if (extensions != 12)
		  {
		     fprintf(stderr, "error: fhtool: `%s' looks like 12K data but has %d FITS files.\n",
			     filename_tmp, extensions);
		     exit(EXIT_FAILURE);
		  }
		  if (!quiet_flg)
		     fprintf(stderr, "fhtool: This looks like a CFH12K image\n");
		  fh_set_str(phu, idx + 0.000001, "DETSIZE", "[1:12450,1:8230]", "Total data pixels in full mosaic");
		  fh_set_int(phu, idx + 0.000002, "NCCDS",   12, "Number of devices");
		  fh_set_int(phu, idx + 0.000003, "NAMPS",   12, "Number of amplifiers");
	       }
	    }
	    if (fh_get_str(phu, "DETSIZE", detsize, sizeof(detsize)) != FH_SUCCESS)
	       detsize[0] = '\0';

	    if (fh_validate(phu) != FH_SUCCESS)
	    {
	       fprintf(stderr, "error: fhtool: Failed to validate primary header unit for `%s'\n",
		       filename_in);
	       exit(EXIT_FAILURE);
	    }
	    if (fh_write(phu, fd_out) != FH_SUCCESS)
	    {
	       fprintf(stderr, "error: fhtool: Failed to write primary header unit to `%s'\n",
		       filename_out);
	       exit(EXIT_FAILURE);
	    }
	    bytes_written += fh_header_blocks(phu) * FH_BLOCK_SIZE;
	 }

	 fh_remove(ehu, "EXTEND"); /* Silently remove the extend keyword from the extension */

	 if (fh_remove(ehu, "SIMPLE") != FH_SUCCESS)
	 {
	    fprintf(stderr, "error: fhtool: `%s' does not start with \"SIMPLE  = T...\"\n",
		    filename_tmp);
	    exit(EXIT_FAILURE);
	 }
	 fh_set_str(ehu, 0, "XTENSION", "IMAGE", "Image extension");

	 /*
	  * This patch is needed to make it possible to display split files as a full
	  * mosaic in SAO DS9.
	  */
	 if (detsize[0] != '\0' &&
	     fh_search(ehu, "DETSIZE", 0) != FH_SUCCESS &&
	     fh_search(ehu, "DETECTOR", &idx_detector) == FH_SUCCESS)
	 {
	    static int warned = 0;

	    if (!warned)
	    {
	       fprintf(stderr, "warning: fhtool: Copying DETSIZE from Primary Header into Extension(s)\n");
	       warned = 1;
	    }
	    fh_set_str(ehu, idx_detector + 0.000001,
		       "DETSIZE", detsize, "Total data pixels in full mosaic");
	    /* %%% Ok for B8 extensions? */
	 }

	 if (fh_search(ehu, "PCOUNT", 0) != FH_SUCCESS &&
	     fh_search(ehu, "NAXIS2", &idx_naxis2) == FH_SUCCESS)
	 {
	    static int warned = 0;
	    
	    if (!warned)
	    {
	       fprintf(stderr, "warning: fhtool: Inserting PCOUNT and GCOUNT into Extension(s)\n");
	       warned = 1;
	    }
	    fh_set_int(ehu, idx_naxis2 + 0.0000001, "PCOUNT", 0, "No 'random' parameters");
	    fh_set_int(ehu, idx_naxis2 + 0.0000002, "GCOUNT", 1, "Only one group");
	 }

	 if (fh_validate(ehu) != FH_SUCCESS)
	 {
	    fprintf(stderr, "error: fhtool: Failed to validate extension header unit from `%s'\n",
		    filename_tmp);
	    exit(EXIT_FAILURE);
	 }
	 if (fh_write(ehu, fd_out) != FH_SUCCESS)
	 {
	    fprintf(stderr, "error: fhtool: Failed to write extension header %d to `%s'\n",
		    i, filename_out);
	    exit(EXIT_FAILURE);
	 }
	 bytes_written += fh_header_blocks(ehu) * FH_BLOCK_SIZE;
	 if (!fh_image_blocks(ehu))
	 {
	    fprintf(stderr, "error: fhtool: No image data found for extension %d\n", i);
	    exit(EXIT_FAILURE);
	 }
	 switch (fh_copy_and_verify_padded_image(ehu, fd_out, fd_in))
	 {
	    case FH_SUCCESS: break;
	    case FH_BAD_VALUE:
	    {
	       if (validate_flg)
	       {
		  fprintf(stderr, "error: fhtool: File may be damaged\n");
		  exit(EXIT_FAILURE);
	       }
	       /* else, copy anyway */
	       break;
	    }
	    case FH_BAD_PADDING:
#ifdef REPAIR_FILES
	    if (repair_flg) break;
	    fprintf(stderr,
		    "error: fhtool: Run again with `-R' option to repair damaged file\n");
#endif
	    /* Fall through */
	    default:
	    {
	       fprintf(stderr, "error: fhtool: Failed to %s data blocks for extension %d\n",
		       validate_flg?"validate":"copy", i);
	       exit(EXIT_FAILURE);
	    }
	 }
	 bytes_written += fh_image_blocks(ehu) * FH_BLOCK_SIZE;
	 /*
	  * Make sure the file is at EOF
	  */
	 close_file(fd_in, filename_tmp, repair_flg);
	 fh_destroy(ehu);
      }
      fh_destroy(phu);

      if (!quiet_flg)
      {
	 if (validate_flg)
	    fprintf(stderr, "fhtool: Validated files in `%s/'\n", filename_in);
	 else
	    fprintf(stderr, "fhtool: Wrote MEF file `%s'\n", filename_out);
	 print_statistics();
      }
      exit(EXIT_SUCCESS);
   }

   /*
    * -------------------
    * Handle MEF -> SPLIT
    * -------------------
    */
   else if (mode == MEF_TO_SPLIT)
   {
      HeaderUnit phu;
      char* filename_base;
      char* filename_tmp;
      int extver_used[100];
      
      memset(extver_used, 0, sizeof(extver_used));

      if (argc >= 3)
	 filename_out = strdup(argv[2]);
      else
	 filename_out = get_basename(filename_in, ".fits", /*strip_path=*/0);

      if (!validate_flg && !strcmp(filename_out, filename_in))
      {
	 fprintf(stderr, "error: fhtool: Please specify a unique output filename\n");
	 exit(EXIT_FAILURE);
      }

      filename_base = get_basename(filename_out, "", /*strip_path=*/1);
      filename_tmp = (char*)malloc(strlen(filename_out) + strlen(filename_base) + 32);

      if (!quiet_flg)
      {
	 if (validate_flg)
	    fprintf(stderr, "fhtool: Validating file `%s'\n",
		    filename_in);
	 else
	    fprintf(stderr, "fhtool: Reading file `%s'\n", filename_in);
      }

      /*
       * Read the Primary Header Unit (PHU) from the file.
       */
      phu = fh_create();
      if (fh_read(phu, fd_in, FH_AUTO) != FH_SUCCESS)
      {
	 fprintf(stderr, "error: fhtool: Failed to read primary header unit from %s\n",
		 filename_in);
	 exit(EXIT_FAILURE);
      }

      if (fh_validate(phu) != FH_SUCCESS)
      {
	 fprintf(stderr, "error: fhtool: Failed to validate primary header unit for `%s'\n",
		 filename_in);
	 exit(EXIT_FAILURE);
      }

      /*
       * Make sure there is at least one extension inside the file.
       */
      if (fh_extensions(phu) < 1)
      {
	 if (!quiet_flg || !validate_flg)
	    fprintf(stderr, "fhtool: `%s' is a basic FITS file (No extensions found)\n",
		    filename_in);

	 if (!validate_flg) exit(EXIT_FAILURE); /* Nothing to do!? */

	 /*
	  * Only keep running past this point if validate is requested.
	  */

	 if (!quiet_flg)
	    fprintf(stderr, "fhtool: Validating FITS data for `%s'\n", filename_in);
	 fd_out = open("/dev/null", O_RDWR);
	 if (fh_copy_padded_image(phu, fd_out, fd_in) != FH_SUCCESS)
	 {
	    fprintf(stderr, "error: fhtool: Failed to validate data for `%s'\n", filename_in);
	    exit(EXIT_FAILURE);
	 }
	 bytes_written += fh_image_blocks(phu) * FH_BLOCK_SIZE;
	 if (close(fd_out) < 0)
	 {
	    perror(filename_tmp);
	    exit(EXIT_FAILURE);
	 }
      }

      /*
       * Try to create the subdirectory for the output.
       */
      if (!validate_flg && mkdir(filename_out, 0775)!=0)
      {
	 switch (errno)
	 {
	    case EEXIST:
	       fprintf(stderr, "error: fhtool: A subdirectory `%s/' already exists.  Remove it first.\n",
		       filename_out);
	       break;
	    case EPERM:
	       fprintf(stderr, "error: fhtool: No permission to create `%s/' here.\n",
		       filename_out);
	       break;
	    default:
	       perror(filename_out);
	 }
	 exit(EXIT_FAILURE);
      }

      if (!validate_flg)
      {
	 /*
	  * Write the PHU to a ".phu" file.
	  */
	 sprintf(filename_tmp, "%s/%s.phu", filename_out, filename_base);
	 if ((fd_out = open(filename_tmp, O_RDWR|O_CREAT, 0664 )) < 0)
	 {
	    perror(filename_tmp);
	    exit(EXIT_FAILURE);
	 }
	 if (fh_write(phu, fd_out) != FH_SUCCESS)
	 {
	    fprintf(stderr, "error: fhtool: Failed to write primary header unit to `%s'\n",
		    filename_tmp);
	    exit(EXIT_FAILURE);
	 }
	 bytes_written += fh_header_blocks(phu) * FH_BLOCK_SIZE;
	 if (close(fd_out) < 0)
	 {
	    perror(filename_tmp);
	    exit(EXIT_FAILURE);
	 }
      }

      /*
       * Iterate through each of the extensions, writing standard FITS files.
       */
      for (i=0; i<fh_extensions(phu); i++)
      {
	 static int extver_previous = -1;
	 static int extver_warned = 0;
	 int extver;
	 HeaderUnit ehu;
	 char xtension[80];

	 ehu = fh_create();
	 if (fh_read(ehu, fd_in, FH_AUTO) != FH_SUCCESS)
	 {
	    fprintf(stderr, "error: fhtool: Failed to read extension %d headers from `%s'\n",
		    i, filename_in);
	    exit(EXIT_FAILURE);
	 }

	 /*
	  * Before removing XTENSION, make sure the header unit is valid as a MEF.
	  * (Otherwise missing PCOUNT or GCOUNT is not caught as an error, because
	  * these are only required when you have an XTENSION.)
	  */
	 if (validate_flg)
	 {
	    if (!quiet_flg)
	       fprintf(stderr, "fhtool: Validating extension %d (%s)\n", i, filename_tmp);
	    if (fh_validate(ehu) != FH_SUCCESS)
	    {
	       fprintf(stderr, "error: fhtool: Failed to validate extension header unit for `%s'\n",
		       filename_in);
	       exit(EXIT_FAILURE);
	    }
	 }

	 if (fh_get_str(ehu, "XTENSION", xtension, sizeof(xtension)) != FH_SUCCESS)
	 {
	    fprintf(stderr, "error: fhtool: `%s[%d]' does not begin with \"XTENSION= 'IMAGE...\"\n",
		    filename_in, i);
	    exit(EXIT_FAILURE);
	 }

	 if (fh_get_int(ehu, "EXTVER", &extver) != FH_SUCCESS)
	 {
	    fprintf(stderr, "error: fhtool: `%s[%d]' does not have an EXTVER keyword\n",
		    filename_in, i);
	    exit(EXIT_FAILURE);
	 }

	 if (!extver_warned && extver < extver_previous)
	 {
	    fprintf(stderr, "warning: fhtool: `%s' has EXTVER out of order\n",
		    filename_in);
	    fprintf(stderr, "warning: fhtool: Splitting and re-mefing will fix the problem.\n");
	    extver_warned++;
	 }
	 extver_previous = extver;

	 if (extver_used[extver]++ != 0)
	 {
	    fprintf(stderr, "error: fhtool: `%s' uses EXTVER %d more than once\n",
		    filename_in, extver);
	    exit(EXIT_FAILURE);
	 }

	 if (strcmp(xtension, "IMAGE"))
	 {
	    if (!validate_flg)
	    {
	       fprintf(stderr, "error: fhtool: cannot split MEF with '%s' extensions\n",
		       xtension);
	       exit(EXIT_FAILURE);
	    }
	 }
	 else
	 {
	    /*
	     * Remove XTENSION='IMAGE   ' and replace it with SIMPLE = T
	     */
	    fh_remove(ehu, "XTENSION");
	    fh_set_bool(ehu, 0, "SIMPLE", 1, "Standard FITS");
	 }

	 sprintf(filename_tmp, "%s/%s%02d.fits", filename_out, filename_base, extver);

	 if (validate_flg)
	 {
	    fd_out = open("/dev/null", O_RDWR);
	 }
	 else
	 {
	    if (!quiet_flg)
	       fprintf(stderr, "fhtool: Writing extension %d to `%s'\n", i, filename_tmp);
	    fd_out = open(filename_tmp, O_RDWR|O_CREAT, 0664);
	 }
	 if (fd_out < 0)
	 {
	    perror(filename_tmp);
	    exit(EXIT_FAILURE);
	 }
	 /*
	  * Validate again, after replacing XTENSION with SIMPLE = T
	  */
	 if (fh_validate(ehu) != FH_SUCCESS)
	 {
	    fprintf(stderr, "error: fhtool: Failed to validate extension header unit for `%s'\n",
		    filename_in);
	    exit(EXIT_FAILURE);
	 }
	 if (fh_write(ehu, fd_out) != FH_SUCCESS)
	 {
	    fprintf(stderr, "error: fhtool: Failed to write extension header unit %d to %s\n",
		    i, filename_tmp);
	    exit(EXIT_FAILURE);
	 }
	 bytes_written += fh_header_blocks(ehu) * FH_BLOCK_SIZE;

	 if (fh_image_blocks(ehu) == 0)
	 {
	    fprintf(stderr, "error: fhtool: No data in extension %d\n", i);
	    exit(EXIT_FAILURE);
	 }
	 if (fh_copy_padded_image(ehu, fd_out, fd_in) != FH_SUCCESS)
	 {
	    fprintf(stderr, "error: fhtool: Failed to %s data for extension %d\n",
		    validate_flg?"validate":"copy", i);
	    exit(EXIT_FAILURE);
	 }
	 bytes_written += fh_image_blocks(ehu) * FH_BLOCK_SIZE;
	 if (close(fd_out) < 0)
	 {
	   perror(filename_tmp);
	   exit(EXIT_FAILURE);
	 }
	 fh_destroy(ehu);
      }
      close_file(fd_in, filename_in, repair_flg);
      fh_destroy(phu);
      if (!quiet_flg)
      {
	 if (validate_flg)
	    fprintf(stderr, "fhtool: Validated MEF file `%s'\n", filename_in);
	 else
	    fprintf(stderr, "fhtool: Wrote %d files in subdirectory `%s/'\n", i+1, filename_out);
	 print_statistics();
      }

      /*
       * This should happen before all the exits, but... */
      free(filename_base);
      free(filename_tmp);
      free(filename_out);

      exit(EXIT_SUCCESS);
   }
   fprintf(stderr, "error: fhtool: Unknown mode\n");
   exit(EXIT_FAILURE);
}

static int
xstrcasecmp(const char *s, const char *d)
/* Copyright (C) 1995,1996 Robert de Bath <rdebath@cix.compulink.co.uk>
 * This file is part of the Linux-8086 C library and is distributed
 * under the GNU Library General Public License.
 *
 * 1998-8-10 ANSIfied and renamed to xstrcasecmp by isani@cfht.hawaii.edu.
 */
{
   for(;;)
   {
      if( *s != *d )
      {
         if( tolower(*s) != tolower(*d) )
            return *s - *d;
      }
      else if( *s == '\0' ) break;
      s++; d++;
   }
   return 0;
}

static int
pstrcasecmp(const void* p1, const void* p2)    /* compare strings through pointers */
{
   return xstrcasecmp(*(char **)p1, *(char **)p2);
}

static char*
get_basename(const char* pathname, const char* ext, int strip_path)
{
   char* p = 0;
   char* rtn;

   if (strip_path)
      p = strrchr(pathname, '/');

   if (p) rtn = strdup(++p);
   else rtn = strdup(pathname);

   if (ext && strlen(rtn) >= strlen(ext) &&
       !xstrcasecmp(rtn + strlen(rtn) - strlen(ext), ext))
      rtn[strlen(rtn) - strlen(ext)] = '\0';
   return rtn;
}


static HeaderUnit
phu_from_phu_file(const char* filename)
{
   HeaderUnit phu = 0;
   int fd_in;

   if ((fd_in = open(filename, O_RDONLY, 0)) < 0)
   {
      switch (errno)
      {
	 case ENOENT:
	 {
	    if (!quiet_flg)
	       fprintf(stderr,
		       "warning: fhtool: No .phu file found.\n");
	    break;
	 }
	 case EACCES:
	 {
	    fprintf(stderr,
		    "error: fhtool: Cannot access `%s'.  Check permissions.\n",
		    filename);
	    exit(EXIT_FAILURE);
	 }
	 default:
	 {
	    perror("error: fhtool: Cannot read PHU");
	    fprintf(stderr,
		    "error: fhtool: Cannot access `%s'.\n",
		    filename);
	    exit(EXIT_FAILURE);
	 }
      }
   }
   else
   {
      phu = fh_create();
      if (fh_read(phu, fd_in, 0) != FH_SUCCESS)
      {
	 if (!quiet_flg)
	    fprintf(stderr, "warning: fhtool: Failed to read primary header unit from %s\n",
		    filename);
	 fh_destroy(phu); phu = 0;
      }
      close_file(fd_in, filename, 0);
   }
   return phu;
}

static int
open_file(const char* filename_requested, const char* default_ext,
	  char** filename_found, int default_fd, int open_mode)
{
   struct stat st_buf;
   char* filename_tmp;
   int fd;

   if (!filename_requested || !strcmp(filename_requested, "-"))
   {
      /*
       * If no filename given, or filename is "-", try to open default_fd
       * (which is typically STDIN or STDOUT).
       */
      if (isatty(default_fd))
      {
	 fprintf(stderr,
		 "error: Cannot read or write a file on a terminal device.\n"
		 "error: Specify a filename or redirect a file to/from standard in/out.\n");
	 return -1;
      }
      if (filename_found)
	 *filename_found = strdup("stdio");
      return default_fd;
   }

   filename_tmp = strdup(filename_requested);
   if (!(open_mode & O_CREAT) && stat(filename_tmp, &st_buf) != 0)
   {
      int saved_errno = errno;

      /*
       * Try "filename.ext" if "filename" doesn't exist...
       */
      if (!default_ext)
      {
	 perror(filename_requested);
	 return -1;
      }
      filename_tmp = malloc(strlen(filename_requested) + strlen(default_ext) + 1);
      sprintf(filename_tmp, "%s%s", filename_requested, default_ext);
      if (stat(filename_tmp, &st_buf) != 0)
      {
	 free(filename_tmp);
	 errno = saved_errno; /* Hack to get original errno again */
	 return -1;
      }
   }

   if (filename_found)
      *filename_found = filename_tmp;

   /*
    * Input file exists.  Figure out if it is a subdirectory.
    * This implies converting MEF->SPLIT.
    */
   if (!(open_mode & O_CREAT) && S_ISDIR(st_buf.st_mode))
   {
      if (!filename_found) free(filename_tmp);
      return -2;
   }

   fd = open(filename_tmp, open_mode, 0664);
   if (!filename_found) free(filename_tmp);
   return fd;
}

static void
close_file(int fd, const char* filename, int repair)
{
   static char buf[FH_BLOCK_SIZE];
   int rtn;

   rtn = read(fd, buf, sizeof(buf));
   switch (rtn)
   {
      case -1:
      {
	 perror("read failed");
	 exit(EXIT_FAILURE);
	 break;
      }
      case 0: break;
      default:
      {
	 int all_zero = 1;
	 int at_least = 1;
	 int n;
	 
	 /*
	  * See if the extra junk is all zeroes
	  */
	 for (n = 0; n < rtn; n++)
	    if (buf[n] != '\0') all_zero = 0;
	 
	 /*
	  * See if there is more than just one buffer extra.
	  */
	 if (read(fd, buf, 1) == 0)
	    at_least = 0;
	 
	 fprintf(stderr,
		 "warning: fhtool: `%s' has %s %d bytes of unexpected %s.\n",
		 filename, at_least?"at least":"exactly", rtn,
		 (at_least==0 && all_zero==1)?"zero padding":"garbage");
	 if (at_least==0 && all_zero==1)
	 {
	    if (!repair)
	    {
#ifdef REPAIR_FILES
	       fprintf(stderr,
		       "error: fhtool: Run again with `-R' option to repair damaged file\n");
#endif
	       exit(EXIT_FAILURE);
	    }
	 }
	 else
	 {
	    fprintf(stderr,
		    "error: fhtool: FITS data is damaged or incomplete.\n");
	    exit(EXIT_FAILURE);
	 }
      }
   }
   if (close(fd) < 0)
   {
      perror(filename);
      exit(EXIT_FAILURE);
   }
}

/* End of fhtool.c */
