/* Copyright (C) 1991, 94, 95, 97    Canada-France-Hawaii Telescope Corp. */
/* This program is distributed WITHOUT any warranty, and is under the     */
/* terms of the GNU General Public License, see the file COPYING          */
/*!****************************************************************************
 * FILE
 * cfht_exec.c
 * $Header: cfht_exec.c,v 1.11 97/12/30 16:41:10 thomas Exp $
 * $Locker:  $
 *
 * ROUTINES
 *
 *  cfht_exec()
 *    Execute a program using a given argument string, possibly with variable
 *    substitution from a par file.  The program can just be executed without
 *    forking (CFHT_EXEC), executed in the foreground with an exit code return
 *    (CFHT_EXECFG), executed in the background with no future communication
 *    (CFHT_EXECBG), executed in the background with an exit code updated on
 *    process termination (value on return is -1) (CFHT_EXECBGC), or in the
 *    background with a callback on completion (CFHT_EXECBGH).
 *
 *   NOTES
 *
 *    This depends heavily on the implementation of cfht_signal .  It assumes
 *    that cfht_signal uses sigaction for its signal handling.  If that
 *    assumption is incorrect, it is probable this will break.  If someone has
 *    a real problem with this (it obviously bothers me), replace the
 *    cfht_signal and cfht_signal_block calls in the
 *    code with the same code extracted from cfht_signal.c .
 *
 *    The caller can block SIGCHLD around a call to cfht_exec , but this block
 *    should be removed as soon as possible.  The caller can NOT have a
 *    handler for SIGCHLD (none of its business, plus the waitpid(2) calls do
 *    not nest).
 *
 *   INVOKED
 *    pid = cfht_exec(file_id, system_str, exec_type, addr);
 *
 *   PARAMETERS IN
 *    FILE_ID file_id  - par file for [%$]name substitutions in system_str
 *                       -1 for no substitutions
 *    char *system_str - command string for actual exec call
 *    int exec_type    - type of exec
 *                       CFHT_EXEC    - just do exec, don't return
 *                       CFHT_EXECFG  - foreground, return program's exit code
 *                                                         or FAIL
 *                       CFHT_EXECBG  - background, return child pid
 *                                                         or FAIL
 *                       CFHT_EXECBGC - background, return child pid
 *                                                         or FAIL
 *                       CFHT_EXECBGH - background, return child pid
 *                                                         or FAIL
 *
 *   PARAMETERS IN/OUT
 *    void *addr - ignored (for CFHT_EXEC, CFHT_EXECFG, CFHT_EXECBG)
 *                 input (PFVexec) function (for CFHT_EXECBGH)
 *                 output (int *) status code (for CFHT_EXECBGC)
 *
 *   FUNCTION RETURNS
 *    int - process id (background)
 *          process exit code (foreground)
 *          or FAIL
 *
 *   GLOBALS
 *    cfht_exec_data   - table of active children
 *    cfht_exec_whoami - my pid (to clear table after fork)
 *    cfht_exec_argv   - argv array built for exec
 *    cfht_exec_args   - argv strings (arg1\0arg2\0... pointed to by argv)
 *
 * HISTORY
 * who          when            what
 * ---------    ------------    ----------------------------------------------
 * jab          13 May 1988     Original coding
 * jrw          20 Dec 1990     sun port
 * jrw          14 Feb 1991     check for null pointers
 * jk           27 Feb 1991     fixed unions for exec functions
 *
 * $Log:	cfht_exec.c,v $
 * Revision 1.11  97/12/30  16:41:10  16:41:10  thomas (Jim Thomas)
 * Added some debug statements for tracing child deaths
 * 
 * Revision 1.10  97/09/29  12:40:53  12:40:53  thomas (Jim Thomas)
 * Saved old SIGCHLD info so mask could be restored, added check for handler
 * Added more error messages
 * Added length check in argument substitution
 * 
 * Revision 1.9  97/03/06  22:26:53  22:26:53  isani (Sidik Isani)
 * Rehashed handling of signals using the new cfht_signal(3) routine.
 *   This routine calls sigaction(2) underneath, but should provide the
 *   same behavior as the old, less portable, sigvector(2) call.
 * Added blocking of sigchld until we are really ready to deal with it
 *   replacing JAB's correct but more complicated race condition avoidance.
 * 
 * Revision 1.8  1995/08/16  13:39:15  thomas
 * Clear signal mask in child after fork (unblock any signals parent blocked)
 *
 * Revision 1.7  95/08/14  14:50:27  14:50:27  thomas (Jim Thomas)
 * Added DEBUG output from parent at exec (to match at exit)
 * 
 * Revision 1.6  95/01/11  18:41:00  18:41:00  thomas (Jim Thomas)
 * Changed exit value from unix version to exits.h form
 * Changed some DEBUG messages, added explanation of wait3 return value
 * 
 * Revision 1.5  95/01/06  15:32:26  15:32:26  thomas (Jim Thomas)
 * Added a bunch of documentation, reformatted
 * 
 * Revision 1.4  94/12/30  10:34:43  10:34:43  john (John Kerr)
 * fixed the return of exitcode; pid switched from int to pid_t; some
 * quick documenting.
 * 
 * Revision 1.3  94/12/14  11:16:36  11:16:36  thomas (Jim Thomas)
 * Added #ifdef __LINT__ around RCS header
 * Added include of unistd.h, took out sys/types.h
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h
 * Changed function prototypes to use P macro
 * Changed CFHT_ERROR to CFHT_LOGONLY
 * Changed type casting on p->addr to match definition
 * Added check for -1 as par file ID before argv substitution
 * 
 * Revision 1.2  92/03/30  10:08:16  10:08:16  steve (Steven Smith)
 * made the CFHT_EXECBGC work as described (treated the addr parameter
 * as a int * rather than as a mess.
 * 
 * Revision 1.1  91/09/24  16:32:16  16:32:16  steve (Steven S Smith)
 * Initial revision
 * 
 ***************************************************************************!*/

/*
 * NOTE: real-time priority operations only apply to HP-UX.  The arguments
 * must be present for other systems, but they have no effect.
 */

#include <stdio.h>		  /* for stderr */
#include <string.h>		  /* for string handling defines */
#include <unistd.h>		  /* for process control */
#include <signal.h>		  /* for child death handling */
#include <errno.h>		  /* for error codes and errno */
#include <sys/wait.h>		  /* for waitpid */
#ifdef hpux
#  include <sys/rtprio.h>
#endif /* hpux */
#if defined(sun) && !defined(__svr4__)
#include <vfork.h>
#endif

#include "cfht.h"
#include "cfp.h"		  /* parfile stuff */
#include "exits.h"		  /* for exit codes */

RCSID("$Header: cfht_exec.c,v 1.11 97/12/30 16:41:10 thomas Exp $");

/*
 * The active children are remembered in a table containing their pid, the
 * execution type (type from cfht_exec call), and the address for exit code or
 * callback routine (addr from cfht_exec call).
 */
volatile struct cfht_exec_data cfht_exec_data[CFHT_EXEC_NUM];

/* our pid - used to clear above table after fork */
static pid_t cfht_exec_whoami = 0;

/*
 * An argv array is built before execution.  The pointers are stored in
 * cfht_exec_argv .  The strings are stored in cfht_exec_args , nul separated
 * ("arg1\0arg2\0...") and pointed to by the cfht_exec_argv entries.
 */
char *cfht_exec_argv[CFHT_ARGV_NUM];
char cfht_exec_args[CFHT_ARGS_SIZE];

#ifdef hpux
int cfht_exec_rtprio = RTPRIO_RTOFF; /* rtprio */
#else
int cfht_exec_rtprio = 0;	  /* not applicable */
#endif

/* forward declarations */
static void cfht_exec_catch P((void)); /* CHLD handler */
static char **cfht_argv P((FILE_ID id, char *rstring));	/* argv builder */

/*****************************************************************************
 *
 * cfht_exec
 *
 *****************************************************************************/

int				  /* return: exit code, pid or FAIL */
cfht_exec(id, string, type, addr)
    FILE_ID id;                   /* par file for %name subs, -1 -> ignore */
    char *string;                 /* system() type string to exec */
    int type;                     /* type of exec (foregrnd, backgrnd, etc.) */
    volatile void *addr;          /* exit code *int or callback *func */
{
    static int idx;		  /* index of available cfht_exec_data[] */
    volatile int exitcode;	  /* exit code from foreground process */
    char **argv;		  /* pointer to cfht_exec_argv */
    pid_t newwhoami;		  /* current process's pid */
    pid_t pid;			  /* pid of forked child */
    struct sigaction oldaction;	  /* parent's current SIGCHLD handling */

    (void) sigemptyset(&oldaction.sa_mask); /* in case, see sigemptyset(2) */
    (void) sigaction(SIGCHLD, (struct sigaction *) NULL, &oldaction);

    /*
     * Make sure the pid table is cleared initially and after forks (the user
     * code may have forked and we don't want to run out of pid entries).
     *
     * Set up SIGCHLD interrupt handler.
     */

    newwhoami = getpid();
    if (cfht_exec_whoami != newwhoami) {
	for (idx = 0; idx < CFHT_EXEC_NUM; idx++)
            cfht_exec_data[idx].pid = 0;
	cfht_exec_whoami = newwhoami;

	if (oldaction.sa_handler != (void *) NULL) {
	    cfht_logv(CFHT_LIBCFHT, CFHT_ERROR,
		      "(cfht_exec:%d) Error, child signal handler will not work for exec of '%.100s'",
		      __LINE__, string);
	    return(FAIL);
	}
	cfht_signal(SIGCHLD, cfht_exec_catch);
    }

    /*
     * get the argv list built from the provided string
     */

    if ((argv = cfht_argv(id, string)) == (char **) NULL) {
	cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(cfht_exec:%d) error parsing argument string '%.100s'",
		  __LINE__, string);
	return(FAIL);
    }

    /*
     * Look for an open pid entry -- ie, look through data table
     * until we find a 0 in the pid variable.  If they're all non-0,
     * wow, we have a bunch of processes running!
     */

    for (idx = 0; idx < CFHT_EXEC_NUM; idx++) {
	if (!cfht_exec_data[idx].pid) {
	    break;
	}
    }
    if (idx >= CFHT_EXEC_NUM) {
	cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(cfht_exec:%d) no pid space", __LINE__);
	errno = ENOSPC;
	return(FAIL);
    }

    /*
     * Fill out pid entry
     */

    cfht_exec_data[idx].pid = -1;
    cfht_exec_data[idx].type = type;
    cfht_exec_data[idx].addr = (void *) (-1);

    switch (type) {
      case CFHT_EXECFG:
	exitcode = CFHT_EXEC_NOT_EXITED;
	cfht_exec_data[idx].addr = (void *) &exitcode;
	break;
      case CFHT_EXECBGC:
	*((int *)addr) = CFHT_EXEC_NOT_EXITED;
	cfht_exec_data[idx].addr = addr; /* here addr is address for exit code */
	break;
      case CFHT_EXECBGH:
	cfht_exec_data[idx].addr = addr; /* here addr is function address */
	break;
      /* default case is already set up */
    }

    /*
     * fork a child
     */

    if (type == CFHT_EXEC)
      pid = cfht_exec_data[idx].pid = 0;	 /* fake fork child side */
    else {
        /*
	 * start blocking SIGCHLD now for two reasons:
	 *
	 * (1) for EXECFG types, we want to make sure the handler doesn't run
	 *     before sigsuspend() starts waiting to keep things simple
	 * (2) it is possible for child thread to exit and trigger
	 *     the handler in the parent before pid has been stored into
	 *     cfht_exec_data[idx].pid
	 */
        cfht_signal_block(SIGCHLD);

	pid = vfork();
	if (pid == (pid_t) -1) {
  	    cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
	  	      "(cfht_exec:%d) vfork() failed", __LINE__);
	    (void) sigprocmask(SIG_SETMASK, &oldaction.sa_mask,
			       (sigset_t *) NULL);
  	    return(FAIL);
	}
    }

    /*
     * child side, execute the desired program
     */

    if (pid == 0) {
        sigset_t emptymask;	  /* empty mask for child; nothing blocked */

	sigemptyset(&emptymask);
        (void) sigprocmask(SIG_SETMASK, &emptymask, (sigset_t *) NULL);

#ifdef hpux
	rtprio(0, cfht_exec_rtprio);
#endif
	execvp(argv[0], argv);
	fprintf(stderr, "(cfht_exec:%d) exec of %s failed, %s\n",
		__LINE__, argv[0], strerror(errno));
	fflush(stderr);
	if (type == CFHT_EXEC) {
	    /*
	     * We are not in a forked thread if this is a CFHT_EXEC
	     * style operation, so it is ok to return something from here.
	     */
	    return(FAIL);
	} else {
	    /*
	     * This one should be _exit(), not exit()! Especially for vfork()
	     * (see the man page for vfork for further details).
	     */
	    _exit(EXIT_FAIL);
	}
    }

    /*
     * parent side, either wait for foreground child to finish or return
     * background pid
     */

    cfht_logv(CFHT_LIBCFHT, CFHT_DEBUG,
	      "(cfht_exec:%d) started '%s' on pid=%ld with idx=%d",
	      __LINE__, string, pid, idx);
#ifdef DEBUG
    cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
	      "(cfht_exec:%d) started '%s' on pid=%ld with idx=%d",
	      __LINE__, string, pid, idx);
#endif

    cfht_exec_data[idx].pid = pid;

    /*
     * foreground type waits for child termination
     */

    if (type == CFHT_EXECFG) {
	/*
	 * wait for p->pid to become zero (during interrupt)
	 *
	 * (I don't think the exitcode test is needed - JT)
	 */
	while ((exitcode == CFHT_EXEC_NOT_EXITED)
	       && (cfht_exec_data[idx].pid != 0))
	  (void) sigsuspend(&oldaction.sa_mask);

	/*
	 * put the mask back without SIGCHLD blocked
	 */
	(void) sigprocmask(SIG_SETMASK, &oldaction.sa_mask, (sigset_t *) NULL);

	cfht_logv(CFHT_LIBCFHT, CFHT_DEBUG,
		  "(cfht_exec:%d) returning %#08x", __LINE__, exitcode);
	return(exitcode);
    } else {

        /*
         * it is now safe to unblock the child signal for EXECBG cases
	 */
	(void) sigprocmask(SIG_SETMASK, &oldaction.sa_mask, (sigset_t *) NULL);

	/*
	 * background types return pid of child
	 */
	return(pid);
    }
}

/*****************************************************************************
 *
 *
 *  cfht_exec_catch()
 *    This is called when a child process the parent has started terminates.
 *
 *  INVOKED
 *    via address stored in main program's sigaction block
 *
 *  PARAMETERS IN
 *   NONE
 *
 *  PARAMETERS OUT
 *   NONE
 *
 *  FUNCTION RETURNS
 *   NONE
 *
 *  GLOBALS
 *   cfht_exec_data   - table of active children
 *
 *****************************************************************************/

static void
cfht_exec_catch()
{
    int old_errno = errno;	  /* back this up so waitpid() can use errno */
    int idx;			  /* index of available cfht_exec_data[] */
    int exitcode;		  /* exit code returned by child */
    pid_t pid;			  /* pid of terminating child */

    /*
     * take care of all dead child processes
     */

    do {
	/* get the pid and exit code of the next one */
#ifndef WAIT_MYPGRP
#define WAIT_MYPGRP 0
#endif
        do {
	    errno = 0;
	    pid = waitpid(WAIT_MYPGRP, &exitcode, WNOHANG);
	} while ((pid <= 0) && (errno == EINTR));

	/* pid < 0 --> no more active children */
	/* pid = 0 --> more active children, but none exited yet */
	/* pid > 0 --> this is one that just exited */
	if (pid > 0) {
	    	cfht_logv(CFHT_LIBCFHT, CFHT_DEBUG,
			  "(cfht_exec_catcher:%d) pid= %ld exit= %#08x",
			  __LINE__, pid, exitcode);
	} else {
	    	cfht_logv(CFHT_LIBCFHT, CFHT_DEBUG,
			  "(cfht_exec_catcher:%d) pid= %ld",
			  __LINE__, pid);
	}

	if (pid > 0) {

	    /*
	     * convert the returned exit into exit code or exit error
	     */
	    if (WIFEXITED(exitcode) == 0) {
		/* child did not exit(), assume error (note, not EXIT_FAIL) */
		exitcode = CFHT_EXEC_FAILED;
		cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
			  "(cfht_exec_catch:%d) child died w/o exit? (%#08x)",
			  __LINE__, exitcode);
	    } else {
		/* exit()'ed, convert code to exits.h value (see wait3(2)) */
		exitcode = WEXITSTATUS(exitcode);
	    }

	    /*
	     * Look through the process table for a matching pid.
	     */
#ifdef DEBUG
	    cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		      "(cfht_exec_catch:%d) looking for pid=%ld",
		      __LINE__, pid);
#endif
	    for (idx = 0; idx < CFHT_EXEC_NUM; idx++) {
#ifdef DEBUG
		cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
			  "(cfht_exec_catch:%d) idx=%d pid=%ld",
			  __LINE__, idx, cfht_exec_data[idx].pid);
#endif
		if (cfht_exec_data[idx].pid == pid)
		  break;
	    }

	    /*
	     * This should never be able to happen... something's smelly
	     * if you get this error.
	     */
	    if (idx >= CFHT_EXEC_NUM) {
	        cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
			  "(cfht_exec_catcher:%d) She says I am the one, but the kid is not my son. pid=%ld'",
			  __LINE__, pid);
  	        continue;
	    }

	    /*
	     * Found the table entry, record the child's death
	     */

	    switch (cfht_exec_data[idx].type) {

		/* for foreground and background processes wanting exit code */
	      case CFHT_EXECFG:
	      case CFHT_EXECBGC:
		if ((cfht_exec_data[idx].addr != (void *) NULL)
		    && (cfht_exec_data[idx].addr != (void *) -1))
		  *((int *)cfht_exec_data[idx].addr) = exitcode;
		break;

		/* for background processes wanting callback */
	      case CFHT_EXECBGH:
		if ((cfht_exec_data[idx].addr != (void *) NULL)
		    && (cfht_exec_data[idx].addr != (void *) -1))
		  ((PFVexec)(cfht_exec_data[idx].addr))(pid, exitcode);
		break;
	    }

	    /*
	     * release table entry
	     */

	    cfht_exec_data[idx].addr = (void *) NULL;
	    cfht_exec_data[idx].pid = 0;
	}
    } while (pid > 0);
    errno = old_errno;
}

/*****************************************************************************
 *
 * cfht_argv()
 *   Take a string and change it into an argv list.  If a par file id is
 *   provided, also use cfp_cursub to do %name and $name substitution using
 *   the referenced file.
 *
 *  INVOKED
 *   argv_list = cfht_argv(parfile_id, argstring);
 *
 *  PARAMETERS IN
 *   FILE_ID parfile_id - file used to look up %name and $name substitutions
 *                        -1 if no substitution is wanted
 *   char *rstring - string to convert
 *
 *  PARAMETERS OUT
 *   NONE
 *
 *  FUNCTION RETURNS
 *   char ** - pointer to the argv array
 *
 *  GLOBALS
 *   cfht_exec_argv   - argv array built for exec
 *   cfht_exec_args   - argv strings (arg1\0arg2\0... pointed to by argv)
 *
 *****************************************************************************/

static char **			  /* return: argv, or NULL */
cfht_argv(id, rstring)
    FILE_ID id;			  /* file to find %name, $name in */
    char *rstring;		  /* string to convert */
{
    int curnum;			  /* index into argv array */
    char *item;			  /* pointer to current token in string */
    char *curarg;		  /* pointer to current position in args */
    char *maxarg;		  /* pointer to far end of args */
    char *p;			  /* temporary pointer into args */
    char stringdata[CFHT_ARGS_SIZE]; /* copy of rstring for cfht_tok to munge */
    char *string;		  /* pointer into stringdata */

    if (rstring == (char *) NULL) {
	cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(cfht_argv:%d) missing string", __LINE__);
	errno = EINVAL;
	return((char **) NULL);
    }


    /* get a copy so cfp_cursub and cfht_tok won't change original */

    if (strlen(rstring) >= (sizeof stringdata) ) {
	cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(cfht_argv:%d) argument string '%.100s' too long",
		  __LINE__, rstring);
	return((char **) NULL);
    }
    (void) strcpy(stringdata, rstring);

    /* see if the caller wants variable substitution before execution */

    if (id != -1) {
	/* do variable substitution */
	if (cfp_cursub(id, stringdata, sizeof stringdata) == (char *) NULL) {
	    cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		      "(cfht_argv:%d) error substituting variables in '%.100s'",
		      __LINE__, stringdata);
	    return((char **) NULL);
	}
    }

    /* now parse string into separate argv pieces */
    curnum = 0;
    curarg = cfht_exec_args;
    maxarg = &cfht_exec_args[CFHT_ARGS_SIZE];
    string = stringdata;
    item = cfht_tok(string, " \t", &string);

    while (item) {
	if (curnum >= CFHT_ARGV_NUM) {
	    cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		      "(cfht_argv:%d) run string has too many args", __LINE__);
	    errno = E2BIG;
	    return((char **) NULL);
	}

	p = curarg;
	curarg += strlen(item) + 1;
	if (curarg > maxarg) {
	    cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		      "(cfht_argv:%d) run string too big", __LINE__);
	    errno = E2BIG;
	    return((char **) NULL);
	}
	strcpy(p, item);
	cfht_exec_argv[curnum] = p;
	curnum++;
	item = cfht_tok(string, " \t", &string);
    }

    cfht_exec_argv[curnum] = (char *) NULL;
    return(cfht_exec_argv);
}
