/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* cfht_frs.c
* $Header: cfht_frs.c,v 1.3 96/11/19 22:46:51 isani Exp $
* $Locker:  $
*
* ROUTINES
*
* cfht_frs()
*    Convert from a number of seconds to HH:MM:SS.SSS... format.  This
*    will work equally well for time conversions as well as angular
*    conversions.  (Note: "frs" seems to have been pulled out of thin air.)
*
*    INVOKED
*    result_string = cfht_frs(dest_string, value, scale, precision);
*
*    PARAMETERS IN
*    double val - value to convert, in seconds
*    int scale - 12 for am/pm, 24 for military, 180 for +/-, 360 for [0,360],
*                0 to leave alone (no scaling or annotations)
*    int prec - number of decimal places of precision for seconds
*
*    PARAMETERS OUT
*    char *dest - location for result or NULL
*
*    FUNCTION RETURNS
*    char * - pointer to result; if dest is NULL then points to internal
*             static area which is reused each time a NULL is passed in
*
*    GLOBALS
*    ???, a static internal string area may be used
*
* ifrs()
*    Internal formatting routine used by cfht_frs().  This actually does
*    the calculation of hours, minutes and seconds and formats the string.
*
*    INVOKED
*    ifrs(dest_string, value, precision);
*
*    PARAMETERS IN
*    double val - number of seconds, scaled as desired
*    int prec - number of decimal places of precision to use for seconds
*
*    PARAMETERS OUT
*    char *dest - formatted string in HH:MM:SS.SSS... form
*
*    FUNCTION RETURNS
*    NONE
*
*    GLOBALS
*    NONE
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jab          11 Mar 1988     Original coding
* jrw          15 Feb 1991     check for null pointers
*
* $Log:	cfht_frs.c,v $
 * Revision 1.3  96/11/19  22:46:51  isani
 * cleaned up rcsid's
 * changed a "%lf" format string to "%f"
 * 
 * Revision 1.2  1994/12/14  11:24:29  thomas
 * Added #ifdef __LINT__ around RCS header
 * Changed cfht/cfht.h to cfht.h
 * Changed function prototypes to use P macro
 *
 * Revision 1.1  91/09/24  16:47:42  16:47:42  steve (Steven Smith)
 * Original checkin
 * 
*
****************************************************************************!*/

#include <math.h>         /* for fmod */
#include <stdio.h>	  /* for sprintf() */
#include <string.h>       /* for string handling defines */
#include <errno.h>        /* for error codes and errno define */
#include "cfht.h"		  /* libcfht defines */

RCSID("$Header: cfht_frs.c,v 1.3 96/11/19 22:46:51 isani Exp $");

static char cfht_frs_s[40];

static void ifrs P((char *dest, double val, int prec));

/*****************************************************************************
 *
 * cfht_frs
 *
 *****************************************************************************/

    char *             /* return: destination */
cfht_frs(dest, val, scale, prec)
    char *dest;        /* place to put the answer, if NULL use our space */
    double val;        /* value to convert in seconds */
    int scale;         /* 12 -> produces am/pm, 24, 180 -> produces +/-, 
                          360 -> scales to 360, 0 -> leave it alone */
    int prec;          /* precision for fractional seconds */
{
    if (!dest)
        dest = cfht_frs_s;

    switch (scale) {
    case 12:
    case 24:
        {
        char stamp[3];

        stamp[0] = 0;
        stamp[1] = 0;
        stamp[2] = 0;
        val = fmod(val, (24.0 * 3600.0));
        if (val < 0.0)
            val += (24.0 * 3600.0);
        if (scale == 12) {
            if (val >= (12.0 * 3600.0)) {
                strcpy(stamp, "pm");
                val -= (12.0 * 3600.0);
            } else {
                strcpy(stamp, "am");
            }
            if (val < 3600.0)
                val += (12.0 * 3600);
        }
        ifrs(dest, val, prec);
        strcat(dest, stamp);
        break;
        }
    case 180:
    case 360:
        val = fmod(val, (360.0 * 3600.0));
        if (val < 0)
            val += (360 * 3600);
        if (scale == 180 && val > (180.0 * 3600.0))
            val -= (360 * 3600);
        ifrs(dest, val, prec);
        break;
    default:
        ifrs(dest, val, prec);
        break;
    }
    return(dest);
}

/*****************************************************************************
 *
 * ifrs -- internal to this routine, not to be called from elsewhere
 *
 *****************************************************************************/

    static void
ifrs(dest, val, prec)
    char *dest;        /* place to put the answer */
    double val;        /* value to convert in seconds */
    int prec;          /* precision for fractional seconds */
{
    char *p;
    int mm, hh;
    int size;

    p = dest;
    if (val < 0) {
        *p++ = '-';
        val = -val;
    }
    hh = val / 3600;
    val -= hh * 3600;
    mm = val / 60;
    val -= mm * 60;
    if (prec > 0) {
        size = 3 + prec;
    } else {
        size = 2;
        prec = 0;
    }
    sprintf(p, "%d:%02d:%0*.*f", hh, mm, size, prec, val);
}
