/* Copyright (C) 1991-1998    Canada-France-Hawaii Telescope Corp.       */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING         */
/*!****************************************************************************
 * FILE cfht_genh.c
 * $Header: cfht_genh.c,v 1.5 98/11/25 16:14:59 thomas Exp $
 * $Locker:  $
 *
 * NOTE
 *    This routine has been superceded by what's in libgenhandler .
 *    Do not use this for any new code.
 *
 * ROUTINES
 *
 * cfht_genh()
 *    This routine serves as the routine for a generic handler program.
 *    Programmers must supply the following additional routines, which
 *    are called by this routine:
 *      beginfits(id) - any fits cards required at the beginning of the
 *                      data collection sequence should be set up here.
 *      cleanup(id)   - any end of program clean up can be done here.
 *      endfits(id)   - any fits cards required at the end of the
 *                      data collection sequence should be set up here.
 *      init(id)      - any initialization may be done here.
 *      otheropts(id) - non-standard options and optionless modes
 *                      should be handled by this routine.
 *      reset(id)     - any resetting should be done here.
 *      status(id)    - report current status..
 *      myCleanup()   - signal handler routine (routine does exit).
 *    In almost all cases the functions return PASSFAIL and "id" is the
 *    parse file FILE_ID for further checking and manipulation of the
 *    .par file variables.
 *
 *    INVOKED
 *    cfht_genh(argv, argc, envp);
 *
 *    PARAMETERS IN
 *    int argc - number of parameters
 *    char **argv - list of parameter strings
 *    char **envp - list of environment strings
 *
 *    PARAMETERS OUT
 *    NONE
 *
 *    FUNCTION RETURNS
 *    NONE
 *
 *    GLOBALS
 *    ???
 *    extern char *progname - our invocation name, stripped of leading path
 *    extern char *gen_msg[] - list of strings to print as usage message
 *
 * HISTORY
 * who          when            what
 * ---------    ------------    ----------------------------------------------
 * jab          17 Mar 1988     Original coding
 * sss          15 Jul 1988     changed to cfht_genh (libcfht), init error fix
 * sss          21 Aug 1989     added genh.h - for standard options
 * jk           24 Jan 1991     provided signal handling capabilities
 *
 * $Log:	cfht_genh.c,v $
 * Revision 1.5  98/11/25  16:14:59  16:14:59  thomas (Jim Thomas)
 * Changed to call cleanup() on failures after init() done
 * Removed bogus if (!myCleanup) , changed FATAL to ERROR
 * 
 * Revision 1.4  96/11/19  22:48:03  22:48:03  isani (Sidik Isani)
 * cleaned up rcsid
 * changed over to use cfht_signal*()
 * 
 * Revision 1.3  1994/12/14  11:27:55  thomas
 * Added #ifdef __LINT__ around RCS header
 * Changed cfht includes from <> to ""
 * Recovered older, pre-client/server version
 *
 * Revision 1.2  92/10/15  10:24:48  10:24:48  jwright (Jim Wright)
 * addition of global variable "progname" so handlers can know who they are
 * 
 * Revision 1.1  91/09/24  16:55:29  16:55:29  steve (Steven Smith)
 * Original checkin
 * 
 **************************************************************************!*/

#include <errno.h>		  /* for errno defines */
#include <unistd.h>		  /* to keep solaris quiet */
#include <signal.h>		  /* for errno defines */

#include "cfht.h"		  /* for cfht generic defines */
#include "cfp.h"		  /* for parse file defines */
#include "genh.h"		  /* for standard arg defines */
#include "exits.h"		  /* for standard exit defines */
#include "states.h"		  /* for standard signal defines */

RCSID("$Header: cfht_genh.c,v 1.5 98/11/25 16:14:59 thomas Exp $");

extern char *progname;
extern char *genh_msg[];

static void myCleanup P((void));
int G_id;         /* global for containing the id of par file */

/***************************************************************************
 *
 * main
 *
 ***************************************************************************/

void cfht_genh(argc, argv, envp)
    int argc;			  /* number of parameters */
    char **argv;		  /* list of parameter strings */
    char **envp UNUSED;		  /* list of environment strings */
{
    FILE_ID id;
    BOOLEAN bool;
    int ExitCode;		  /* EXIT_PASS unless error then EXIT_FAIL */

    ExitCode = EXIT_PASS;

    /*****************
     * parse options *
     *****************/

    if ((id = cfp_parse(CFP_PAR, argc, argv)) == FAIL) {
        cfht_log(CFHT_MAIN, CFHT_ERROR, "parsing failed, exiting");
        exit(EXIT_FAIL);
    }

    progname = argv[0];
    G_id = id;

    /*************
     * handle -H *
     *************/

    if (cfp_getcur(id, GENH_HELP_ARG, &bool, CFP_BOOLEAN, sizeof(bool)) == FAIL)
        getcurdeath(GENH_HELP_ARG);
    if (bool) {
        usage(genh_msg);
        cfht_log(CFHT_MAIN, CFHT_DONE, "completing successfully");
        exit(EXIT_PASS);
    }

    /*************
     * handle -V *
     *************/

    if (cfp_getcur(id, GENH_VERBOSE_ARG, &bool, CFP_BOOLEAN, sizeof(bool)) 
                   == FAIL)
        getcurdeath(GENH_VERBOSE_ARG);
    if (bool)
        cfht_debug = TRUE;

    /************************************************
     * allow specific code to do any initialization *
     ************************************************/

    /* set up sigterm and sigint handlers and block the signals until
     * ready to handle them (blocking done within these routines).
     */

    cfht_signal_block(SIGINT);
    cfht_signal_block(SIGTERM);
    cfht_signal(SIGINT, myCleanup);
    cfht_signal(SIGTERM, myCleanup);

    if (init(id) == FAIL) {
        cfht_log(CFHT_MAIN, CFHT_ERROR, "init() error, exiting");
        exit(EXIT_FAIL);
    }
    
    /* Unblock signals now and check to see if one is present */

    cfht_signal_unblock(SIGINT);
    cfht_signal_unblock(SIGTERM);

    /*************
     * handle -R *
     *************/

    if (cfp_getcur(id, GENH_RESET_ARG, &bool, CFP_BOOLEAN, sizeof(bool)) 
                   == FAIL)
        getcurdeath(GENH_RESET_ARG);
    if (bool) {
        if (reset(id) == FAIL) {
            cfht_log(CFHT_MAIN, CFHT_ERROR, "fatal error, exiting");
            ExitCode = EXIT_FAIL;
        }
        goto docleanup;
    }

    /*************
     * handle -B *
     *************/

    if (cfp_getcur(id, GENH_BEGINFITS_ARG, &bool, CFP_BOOLEAN, sizeof(bool)) 
                   == FAIL)
        getcurdeath(GENH_BEGINFITS_ARG);
    if (bool) {
        if (beginfits(id) == FAIL) {
            cfht_log(CFHT_MAIN, CFHT_ERROR, "fatal error, exiting");
            ExitCode = EXIT_FAIL;
        }
        goto docleanup;
    }

    /*************
     * handle -E *
     *************/

    if (cfp_getcur(id, GENH_ENDFITS_ARG, &bool, CFP_BOOLEAN, sizeof(bool)) 
                   == FAIL)
        getcurdeath(GENH_ENDFITS_ARG);
    if (bool) {
        if (endfits(id) == FAIL) {
            cfht_log(CFHT_MAIN, CFHT_ERROR, "fatal error, exiting");
            ExitCode =EXIT_FAIL;
        }
        goto docleanup;
    }

    /*************
     * handle -S *
     *************/

    if (cfp_getcur(id, GENH_STATUS_ARG, &bool, CFP_BOOLEAN, sizeof(bool)) 
                   == FAIL)
        getcurdeath(GENH_STATUS_ARG);
    if (bool) {
        if (status(id) == FAIL) {
            cfht_log(CFHT_MAIN, CFHT_ERROR, "fatal error, exiting");
            ExitCode = EXIT_FAIL;
        }
        goto docleanup;
    }

    /***********************************************
     * allow specific code to handle other options *
     ***********************************************/

    if (otheropts(id) == FAIL) {
        cfht_log(CFHT_MAIN, CFHT_ERROR, "fatal error, exiting");
        ExitCode = EXIT_FAIL;
    }

    /******************************************
     * allow specific code to do any clean up *
     ******************************************/

docleanup:
    if (cleanup(id) == FAIL) {
        cfht_log(CFHT_MAIN, CFHT_ERROR, "cleanup() error, exiting");
        exit(EXIT_FAIL);
    }

    /***********************
     * do our own clean up *
     ***********************/

    cfht_logv(CFHT_MAIN, CFHT_DONE, "completing %ssuccessfully",
	      (ExitCode == EXIT_PASS) ? "" : "un");
    exit(ExitCode);
}

/***************************************************************************
 *
 * myCleanup -- internal to this routine, not to be called from elsewhere
 *
 * This routine is the signal handler for genh.  It will in turn call
 * the usual cleanup.
 *
 *  jk  910124  Original
 ***************************************************************************/

static void myCleanup()
{
    if (cleanup(G_id) == FAIL) {
        cfht_log(CFHT_MAIN, CFHT_ERROR, "cleanup() error, exiting");
        exit(EXIT_FAIL);
    }

    /***********************
     * do our own clean up *
     ***********************/

    cfht_log(CFHT_MAIN, CFHT_DONE, "completing successfully");
    exit(EXIT_PASS);
}
