/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* cfht_julian.c
* $Header: cfht_julian.c,v 1.3 96/11/19 23:06:20 isani Exp $
* $Locker:  $
*
* ROUTINES
*
* cfht_julian()
*    This routine will return a double which is the Julian date of the date
*    passed in, something on the order of 2400000.  Remember to input
*    UT values and not HST values.  Note that years procede ...-2, -1, 0,
*    1, 2,...  Calendar reform (i.e. switch from julian calendar to
*    gregorian calendar) is accounted for.
*
*    INVOKED
*    jd = cfht_julian(year, month, day, hour, minute, sec);
*
*    PARAMETERS IN
*    int year - calendar year (for example, 1990)
*    int month - month of year, [1,12]
*    int day - day of month, [1,31]
*    int hour - hour of day, [0,23]
*    int minute - minute of hour, [0,59]
*    double sec - second of minute, [0.0,59.999]
*
*    PARAMETERS OUT
*    none
*
*    FUNCTION RETURNS
*    double - julian date, including fractional part for seconds into day
*
*    GLOBALS
*    none
*
* cfht_fromjulian()
*    This routine converts from a julian date back into y/m/d/h/m/s form.
*
*    INVOKED
*    cfht_fromjulian(jd, year, month, day, hour, minute, sec);
*
*    PARAMETERS IN
*    double jd - julian date to be converted
*
*    PARAMETERS OUT
*    int *year - calendar year (for example, 1990)
*    int *month - month of year, [1,12]
*    int *day - day of month, [1,31]
*    int *hour - hour of day, [0,23]
*    int *minute - minute of hour, [0,59]
*    double *sec - second of minute, [0.0,59.999]
*
*    FUNCTION RETURNS
*    none
*
*    GLOBALS
*    none
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jrw          13 Jul 90       original coding from textbook provided by bg
*
* $Log:	cfht_julian.c,v $
 * Revision 1.3  96/11/19  23:06:20  isani
 * cleaned up rcsid's
 * 
 * Revision 1.2  1994/12/14  11:32:00  thomas
 * Added #ifdef __LINT__ around RCS header
 *
 * Revision 1.1  91/09/24  16:57:25  16:57:25  steve (Steven Smith)
 * Original checkin
 * 
*
****************************************************************************!*/

#include <time.h>

#include "cfht.h"

RCSID("$Header: cfht_julian.c,v 1.3 96/11/19 23:06:20 isani Exp $");

/* EXAMPLES
 * -584 May 28 12:00:00    1507900.0
 * 1957 Oct 04 19:26:24    2436116.31    (launch of Sputnik 1)
 *  333 Jan 27 12:00:00    1842713.0
 * 1977 Apr 26 09:36:00    2443259.9
 */

#define EPSILON 0.0
/*
 * machine has only finite precision, so we use this to avoid translating
 * 119.999 seconds into (1 minute, 60 seconds) instead of (2 minutes 0 seconds)
 *
 * turned out to be a bad idea, change back to 0
 */


double cfht_julian(year,month,day,hour,minute,sec)
int year, month, day, hour, minute;
double sec;
{
    int a,b;
    int y,m;
    double jd;

    if (month <= 2) {
        y = year - 1;
        m = month + 12;
    } else {
        y = year;
        m = month;
    }

    /*
     * figure the julian date based on the julian calendar
     */
    jd = 1720994.5;        /* -1 Oct 30 0:00:00 */
    jd += (double)( (int)((double)y * 365.25 - (y<0 ? 0.75 : 0.0)) +
                (int)((double)(m+1) * 30.6001                ) + day);
    jd += (((double)hour*3600.0) + ((double)minute*60.0) + sec) / 86400.0;

    /*
     * If we are in the gregorian calendar and not the julian
     * calendar, we need to adjust.
     */
    if ((year > 1582) || 
        ((year == 1582) && (month > 10)) ||
        ((year == 1582) && (month == 10) && (day >= 15))) {
        a = y / 100;
        b = 2 - a + (a / 4);
        jd += (double)b;
    }

    return (jd);
}

void cfht_fromjulian(jd,year,month,day,hour,minute,sec)
int *year, *month, *day, *hour, *minute;
double jd, *sec;
{
    long z,a,alpha,b,c,d,e;
    double f,q;

    if (jd < 0) {
        *year = *month = *day = *hour = *minute = 0;
        *sec = 0.0;
    } else {
        jd += 0.5;
        z = (long)jd;
        f = jd - (double)z;
        if (z < 2299161) {
            a = z;
        } else {
            alpha = (long)(((double)z - 1867216.25) / 36524.25);
            a = z + 1 + alpha - (alpha/4);
        }
        b = a + 1524;
        c = (long)(((double)b - 122.1) / 365.25);
        d = (long)((double)c * 365.25);
        e = (long)(((double)b - (double)d) / 30.6001);
        q = (double)(b - d - (long)((double)e * 30.6001)) + f;
        *month = (e<14 ? (int)(e-1) : (int)(e-13));
        *year = (*month<3 ? (int)(c-4715) : (int)(c-4716));
        *day = (int)q;
        q = (q - (double)*day) * 86400.0;
        *hour = (int)((q / 3600.0) + EPSILON);
        q -= (double)*hour * 3600.0;
        *minute = (int)((q / 60.0) + EPSILON);
        *sec = q - ((double)*minute * 60.0);
    }
}
