/*****************************************************************************/
/* Copyright (C) 1991-2003 Canada-France-Hawaii Telescope Corp.              */
/* This program is distributed WITHOUT any warranty, and is under the        */
/* terms of the GNU General Public License, see the file COPYING             */
/*!****************************************************************************
 * $Id: cfht_log.c,v 1.25 2003/01/22 23:57:43 thomas Exp $
 * $Locker:  $
 *
 * ROUTINES
 *
 * cfht_logv()
 *    This is a varargs version of the logging routine.  As the code says,
 *    read up on vprintf(3) and varargs(5).  It simply handles the varargs
 *    munging and then calls cfht_log().
 *
 *    INVOKED
 *    cfht_logv(who, type, fmt, arg, arg, ... );
 *
 *    PARAMETERS IN
 *    int who - who is calling this
 *    int type - what kind of message is this
 *    char *fmt - a format string for the args that follow
 *    [any] arg - some argument; the number and type must agree with fmt
 *
 *    PARAMETERS OUT
 *    NONE
 *
 *    FUNCTION RETURNS
 *    NONE
 *
 *    GLOBALS
 *    none
 *
 * cfht_logpv()
 *    Same as cfht_logv but it logs the parent's PID instead
 *
 * cfht_log()
 *    This routine performs the logging of informational messages for
 *    the everybody else.  It sends information to a pipe, for display
 *    in an X window, and it also writes to a file.  The behavior of
 *    this routine is greatly influenced by the setting of various
 *    environment variables.
 *
 *    INVOKED
 *    cfht_log(who, type, mess);
 *
 *    PARAMETERS IN
 *    int who - who is calling this
 *    int type - what kind of message is this
 *    char *mess - the message to be displayed
 *
 *    PARAMETERS OUT
 *    NONE
 *
 *    FUNCTION RETURNS
 *    NONE
 *
 *    GLOBALS
 *    none
 *
 * HISTORY
 * who          when            what
 * ---------    ------------    ----------------------------------------------
 * jab          12 Mar 1988     Original coding
 *                              (This is just a primitive do for now job)
 * sss          12 Apr 1988     implemented time stamping.
 * sss          08 May 1988     output to named pipe
 * jab          08 Aug 1988     rewrite for log file, and new xout
 * sss          17 Mar 1988     patched so CFHTLIBxxx=TRUE environment works
 * jab          01 Jun 1989     added cfht_logv() function
 * sss          17 Nov 1990     bumped cfht_logv internal buffer, still can be
 *                              exceeded.
 * jrw          07 Dec 1990     made internal buffer size a #define
 * jrw          15 Feb 1991     check for null pointers
 * jrw          28 Feb 1991     add bells for error, fatal and errno
 * sss          08 Jun 1991     added no real need to be inited
 *
 * $Log: cfht_log.c,v $
 * Revision 1.25  2003/01/22 23:57:43  thomas
 * Changed logu_prefix and logu_postfix to arrays to avoid HP-UX UMR errors
 *
 * Revision 1.24  2003/01/22 23:43:08  isani
 * Added TYPESTRING defines.
 *
 * Revision 1.23  2002/07/23 09:27:02  isani
 * Use blocking I/O to write pipes.
 *
 * Revision 1.22  2002/01/23 16:15:52  isani
 * Rewrote the way log strings are built.  They are now written to
 * the log pipe (user log) with a single write() call, protected
 * by a retry in case a signal interrupts the write.
 * Several potential buffer overrun problems were fixed as well.
 *
 * Revision 1.21  2001/10/14  20:18:06  isani
 * Pass START and DONE messages to director so they reach the system log.
 * 
 * Revision 1.20  2000/04/13  13:01:14  isani
 * Separated relevant defines into clidefs.h instead of cli.h
 * to avoid adding P() defines to everything in cli.h.
 * 
 * Revision 1.19  2000/04/10  11:31:45  isani
 * Added new logging behavior if DIRECTOR is in environment.
 * 
 * Revision 1.18  2000/02/10  13:14:57  isani
 * Turned off color escape codes if not on an HP.
 * Added "error:" "warning:" prefix strings instead (color added by director.)
 * Added a few type casts to avoid print format warnings.
 * 
 * Revision 1.17  1998/02/25 16:36:36  thomas
 * Added printing of PID to log file
 * Added cfht_logpv to print parent PID for cfht_log program
 * Removed network version of code
 *
 * Revision 1.16  97/12/09  00:12:53  00:12:53  healey (SueAnn Healey)
 * added CFHT_LIBSSX
 * 
 * Revision 1.15  1997/03/24 23:57:25  thomas
 * Added LIBOMS
 *
 * Revision 1.14  97/02/07  21:57:58  21:57:58  isani (Sidik Isani)
 * Added color support for xterms
 * 
 * Revision 1.13  96/11/19  23:08:04  isani
 * cleaned up rcsid's
 * added a forward declaration
 * 
 * Revision 1.12  1996/08/16  04:21:37  isani
 * added fallback to "unknown" for empty usernames
 *
 * Revision 1.11  1996/05/20 16:00:28  thomas
 * Added LIBCX
 *
 * Revision 1.10  95/06/13  12:03:53  12:03:53  thomas (Jim Thomas)
 * Added debug logging entries for CS, DUCK, HELP, IDS, PIXD, TCS, and UI
 * libraries
 * 
 * Revision 1.9  95/01/30  14:59:01  14:59:01  thomas (Jim Thomas)
 * Change msgio.h to music.h in network version
 * 
 * Revision 1.8  95/01/05  15:54:51  15:54:51  thomas (Jim Thomas)
 * If compiling on sun don't include color escape sequence output
 * (only deleted output so if someone gets xterm color there won't have been
 *  lots of places changed)
 * 
 * Revision 1.7  94/12/14  11:43:07  11:43:07  thomas (Jim Thomas)
 * Moved LogThisDebugQ earlier, used it for LOGONLY as well as DEBUG
 * Moved default output (neither LOGS nor LOGU defined) to LOGU+DEBUG+LOGONLY
 * Fixed initialization of fd_file and fd_pipe for reentrancy
 * 
 * Revision 1.6  94/09/20  18:53:34  18:53:34  thomas (Jim Thomas)
 * added CFHT_LOGONLY, CFHT_LIBOCS
 * changed meaning of CFHT_DEBUG calls
 * added check for CFHT_DEBUG in cfht_logv
 * used symbols instead of numbers in several places
 * renamed CFHT_FATAL to be a synonym for CFHT_ERROR (on its way out)
 * NOTE - above only done to old code, NOT network version
 * 
 * 
 * Revision 1.5  94/08/14  10:28:32  10:28:32  jwright (Jim Wright)
 * add warning, add color (ONLY IN OLD VERSION, NOT YET IN NET LOGGING!)
 * 
 * Revision 1.4  94/07/11  15:50:12  15:50:12  veran (Jean-Pierre Veran)
 * Added net logging capabilities
 *
 * Revision 1.3  94/01/20  02:03:40  02:03:40  jwright (Jim Wright)
 * rename 'time' to avoid redefinition hides earlier one
 *
 * Revision 1.2  93/06/29  15:47:03  15:47:03  steve (Steven Smith)
 * added timestamping capability
 *
 * Revision 1.1  91/09/24  16:59:08  16:59:08  steve (Steven Smith)
 * Original checkin
 *
 ***************************************************************************!*/

#include <stdio.h>		  /* for bomb error messages */
#include <unistd.h>		  /* for write() */
#ifdef __STDC__
#include <stdarg.h>		  /* for handling logv */
#else
#include <varargs.h>		  /* for handling logv */
#endif
#include <fcntl.h>		  /* for O_WRONLY */
#include <string.h>		  /* for string handling defines */
#include <errno.h>		  /* for named pipe failure stuffs */
#include <sys/time.h>
#include <stdlib.h>		  /* for getenv() */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* CFP_PATHSIZE */
#include "cfht_log.h"		  /* message symbols */
#include "cli/clidefs.h"	  /* for director-style log formatting */

RCSID("$Id: cfht_log.c,v 1.25 2003/01/22 23:57:43 thomas Exp $");

#define LogFdUnassigned 0	  /* File descriptor needs to be assigned */
#define LogFdDontUse -1		  /* File descriptor failed or not used */

extern int sys_nerr;		  /* not in errno.h unless C++ :-( */

/* the following two output file descriptors are initialized to 0 (stdin) */
/* the zero value is used as a flag indicating a real open is needed      */
static int fdfile = LogFdDontUse; /* system log file */
static int fdpipe = LogFdDontUse; /* user feedback window */

static char LOGUvalue[CFP_PATHSIZE + 1]; /* copy of CFHTLOGU value */
static char LOGSvalue[CFP_PATHSIZE + 1]; /* copy of CFHTLOGS value */

static const char* whoarewe = "";  /* user name */
static char log_id[CFHT_LOG_ID_SIZE + 1]; /* program identification */

BOOLEAN cfht_debug;		  /* debug environment variable set */
BOOLEAN cfht_warn;		  /* warnings off environment variable set */
BOOLEAN cfht_error;		  /* errors off environment variable set */

BOOLEAN cfht_nodisp = FALSE;	  /* user messages off environment variable
				   * set */
static BOOLEAN ready = FALSE;	  /* false until LOG_ID call seen then true */

static BOOLEAN director = FALSE;  /* TRUE when running in "agent" mode. */
static BOOLEAN debugging = FALSE; /* FALSE unless both LOGS and LOGU undef */
				  /*  then TRUE to indicate debugging mode */

BOOLEAN cfht_lib = FALSE;	  /* do not debug in general */
static BOOLEAN cfht_libcamac = FALSE; /* do not debug CAMAC */
static BOOLEAN cfht_libccd = FALSE; /* do not debug ccd */
static BOOLEAN cfht_libcfht = FALSE; /* do not debug cfht */
static BOOLEAN cfht_libcfp = FALSE; /* do not debug cfp */
static BOOLEAN cfht_libcs = FALSE; /* do not debug client/server */
static BOOLEAN cfht_libcx = FALSE; /* do not debug coordinate library */
static BOOLEAN cfht_libduck = FALSE; /* do not debug DUCK interface */
static BOOLEAN cfht_libff = FALSE; /* do not debug FITS */
static BOOLEAN cfht_libhelp = FALSE; /* do not debug help */
static BOOLEAN cfht_libhh = FALSE; /* do not debug high level HPIB */
static BOOLEAN cfht_libids = FALSE; /* do not debug IDS communication */
static BOOLEAN cfht_libmusic = FALSE; /* do not debug music */
static BOOLEAN cfht_libocs = FALSE; /* do not debug OCS */
static BOOLEAN cfht_liboms = FALSE; /* do not debug OMS */
static BOOLEAN cfht_libpixd = FALSE; /* do not debug pixel library */
static BOOLEAN cfht_libret = FALSE; /* do not debug reticon */
static BOOLEAN cfht_libssx = FALSE; /* do not debug ssx motor controller */
static BOOLEAN cfht_libtcs = FALSE; /* do not debug telescope control */
static BOOLEAN cfht_libui = FALSE; /* do not debug user interface */

#ifdef hpux
static BOOLEAN use_color = TRUE;  /* use terminal-specific color sequences */
#else
/*
 * On other platforms, don't use color by default (can still be changed
 * by environment variable.)  It is not necessary to have cfht_log try
 * to generate terminal specific escape sequences when it can just tack
 * things like "error:" on the front of the message and let director
 * add the colors to the feedback window.
 */
static BOOLEAN use_color = FALSE; /* use terminal-specific color sequences */
#endif

static struct timeval starttime;  /* time LOG_ID called */
static struct timeval lasttime;	  /* last time cfht_log called */

/*
 * This library requires a SYSV style sprintf, which returns the number
 * of bytes added to the string, just like printf and fprintf.
 */
#ifdef SUNOS
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
static int sysv_sprintf(char* s, const char* fmt, ...)
{
   va_list args;
#ifdef __STDC__
   va_start(args, fmt);
#else
   va_start(args);
#endif
   vsprintf(s, fmt, args);
   return strlen(s);
}
#else
#define sysv_sprintf sprintf
#endif

/***********************************************************************
 *
 *   SetUpLogging
 *
 * DESCRIPTION
 *    given a CFHT_LOG_ID call, initialize all logging state
 *
 * INVOKED
 *    SetUpLogging(mess);
 *
 * PARAMETERS IN
 *    char *mess - should contain the program arguments
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    none
 *
 * GLOBALS
 *    cfht_* in this file
 *
 ***********************************************************************/

static void
SetUpLogging Q((char *mess), (mess), char *mess;)
{
    char *p, *q;		  /* scratch */
    static struct timezone tzp;	  /* local time zone */

    /* clear out the output file descriptors (fails in daemons if not) */
    if (fdfile != LogFdDontUse) close(fdfile);
    if (fdpipe != LogFdDontUse) close(fdpipe);
    /*
     * The value of LogFdUnassigned means nothing is selected yet.  By
     * the time this function returns, there will either be a different,
     * valid file descriptor value here or LogFdDontUse if the file/pipe is
     * not to be used.
     */
    fdfile = LogFdUnassigned;
    fdpipe = LogFdUnassigned;

    /* who are we */
    if ((p = getenv("HOME"))) {
	if ((q = strrchr(p, '/')))
	  p = q + 1;
	whoarewe = p;
    }
    if (whoarewe == NULL)
        whoarewe = "unknown";

    /* what program is it */
    if (strlen(mess) >= CFHT_LOG_ID_SIZE - 2)
      (void) strcpy(&mess[CFHT_LOG_ID_SIZE / 2], "--- ID too long ----");
    (void) strcpy(log_id, mess);

    /* get environment variable flags, if any */
    cfht_debug = cfht_tob(getenv(CFHT_EDEBUG));
    cfht_lib = cfht_tob(getenv(CFHT_ELIB));
    cfht_libcamac = cfht_tob(getenv(CFHT_ELIBCAMAC));
    cfht_libccd = cfht_tob(getenv(CFHT_ELIBCCD));
    cfht_libcfht = cfht_tob(getenv(CFHT_ELIBCFHT));
    cfht_libcfp = cfht_tob(getenv(CFHT_ELIBCFP));
    cfht_libcs = cfht_tob(getenv(CFHT_ELIBCS));
    cfht_libcx = cfht_tob(getenv(CFHT_ELIBCX));
    cfht_libduck = cfht_tob(getenv(CFHT_ELIBDUCK));
    cfht_libff = cfht_tob(getenv(CFHT_ELIBFF));
    cfht_libhelp = cfht_tob(getenv(CFHT_ELIBHELP));
    cfht_libhh = cfht_tob(getenv(CFHT_ELIBHH));
    cfht_libids = cfht_tob(getenv(CFHT_ELIBIDS));
    cfht_libmusic = cfht_tob(getenv(CFHT_ELIBMUSIC));
    cfht_libocs = cfht_tob(getenv(CFHT_ELIBOCS));
    cfht_liboms = cfht_tob(getenv(CFHT_ELIBOMS));
    cfht_libpixd = cfht_tob(getenv(CFHT_ELIBPIXD));
    cfht_libret = cfht_tob(getenv(CFHT_ELIBRET));
    cfht_libssx = cfht_tob(getenv(CFHT_ELIBSSX));
    cfht_libtcs = cfht_tob(getenv(CFHT_ELIBTCS));
    cfht_libui = cfht_tob(getenv(CFHT_ELIBUI));

    /* 
     * if ERROR variable is there and it says OFF then turn off errors,
     * otherwise make sure they appear in feedback
     */
    p = getenv (CFHT_EERROR);
    if (p && !cfht_tob (p))
      cfht_error = FALSE;
    else
      cfht_error = TRUE;

    /* 
     * if WARN variable is there and it says OFF then turn off warnings,
     * otherwise make sure they appear in feedback
     */
    p = getenv (CFHT_EWARN);
    if (p && !cfht_tob (p))
      cfht_warn = FALSE;
    else
      cfht_warn = TRUE;

    /*
     * if program has already set nodisplay, honor that
     * otherwise set to NODISP
     */
    if (!cfht_nodisp)		  /* allow early override */
      cfht_nodisp = cfht_tob (getenv (CFHT_ENODISP));

    if (getenv(DIRECTOR_ENV) != NULL &&
	*getenv(DIRECTOR_ENV) != '\0')
      director = TRUE;
    else
      director = FALSE;

    if (director) {
	/*
	 * When running as a "agent" of director, send all messages to stderr.
	 * CFHT_DEBUG, CFHT_ERROR, CFHT_WARN, etc. still have an effect, but
	 * if they are not set, the default is all messages ON so that director
	 * can decide which to filter.  This is a performance hit for programs
	 * that generate a lot of debugging messages.
	 */
	LOGUvalue[0] = '\0';
	LOGSvalue[0] = '\0';
	use_color = FALSE;
	debugging = TRUE; /* CFHTDEBUG still has to get set too for debug: */
	fdfile = LogFdDontUse;
	fdpipe = dup(STDERR_FILENO);
    } else {
	/* if COLORLOG says OFF then honor that */
	p = getenv (CFHT_ECOLORLOG);
	if (p) use_color = cfht_tob(p);
	
	/* figure out what logging is desired */
	debugging = FALSE;
	if ((p = getenv(CFHT_LOGU)) == (char *) NULL) {
	    LOGUvalue[0] = '\0';
	} else {
	    (void) strncpy(LOGUvalue, p, CFP_PATHSIZE);
	    LOGUvalue[CFP_PATHSIZE] = '\0';
	}
	if ((p = getenv(CFHT_LOGS)) == (char *) NULL) {
	    LOGSvalue[0] = '\0';
	} else {
	    (void) strncpy(LOGSvalue, p, CFP_PATHSIZE);
	    LOGSvalue[CFP_PATHSIZE] = '\0';
	}
	if ((LOGUvalue[0] == '\0') && (LOGSvalue[0] == '\0')) {
	    /* if neither is defined, we're debugging, put LOGU on stderr */
	    fdfile = LogFdDontUse;
	    fdpipe = dup(STDERR_FILENO);
	    debugging = TRUE;
	} else {
	    if (LOGUvalue[0] == '\0') {
		/* if no LOGU, suppose they don't want to be bothered? */
		fdpipe = LogFdDontUse;
	    }
	    if (LOGSvalue[0] == '\0') {
		/* no LOGS?  ignore it too?? */
		fdfile = LogFdDontUse;
	    }
	}
	
	/* open files */
	if (fdpipe == LogFdUnassigned) {
	    fdpipe = open(LOGUvalue,
			  O_CREAT | O_WRONLY | O_APPEND | O_NDELAY, 0664);
	    if (fdpipe >= 0) {
	        fcntl(fdpipe, F_SETFD, 1); /* set close on exec */
		fcntl(fdpipe, F_SETFL,
		      fcntl(fdpipe, F_GETFL, 0)&~(O_NONBLOCK|O_NDELAY));
	    }
	    else {
		perror("(cfht_log) user log open failed");
		fprintf(stderr, "(...)      file was '%s'\n", LOGUvalue);
	    }
	}
	
	if (fdfile == LogFdUnassigned) {
	    fdfile = open(LOGSvalue,
			  O_CREAT | O_WRONLY | O_APPEND | O_NDELAY, 0664);
	    if (fdfile >= 0) {
	        fcntl(fdfile, F_SETFD, 1); /* set close on exec */
		fcntl(fdfile, F_SETFL,
		      fcntl(fdfile, F_GETFL, 0)&~(O_NONBLOCK|O_NDELAY));
	    } else {
		perror("(cfht_log) system log open failed");
		fprintf(stderr, "(...)     file was '%s'\n", LOGSvalue);
	    }
	}
	
	/* one last check, if both open's failed, use stderr */
	if ((fdpipe < 0) && (fdfile < 0)) {
	    fdfile = dup(STDERR_FILENO);
	}
    }

    /* get starting time and save as last time for delta */
    gettimeofday (&starttime, &tzp);
    memcpy (&lasttime, &starttime, sizeof (struct timeval));
    
    ready = TRUE;
    return;
}

/***********************************************************************
 *
 *   LogThisDebugQ
 *
 * DESCRIPTION
 *    given that a message is CFHT_DEBUG, decide whether to log it
 *
 * INVOKED
 *    if(LogThisDebugQ(who) == FALSE) return;
 *
 * PARAMETERS IN
 *    cfht_log_who who - who is logger?
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    BOOLEAN - TRUE if logging is on for this caller
 *              FALSE if message should not be logged
 *
 * GLOBALS
 *    cfht_debug, cfht_lib*
 *
 ***********************************************************************/

static BOOLEAN
LogThisDebugQ Q((cfht_log_who who), (who), cfht_log_who who;)
{
    /* setting LIB debugs all libraries */
    if ((who != CFHT_MAIN) && (cfht_lib == TRUE))
      return(TRUE);

    switch (who) {
      case CFHT_MAIN:
	return(cfht_debug);
	/*NOTREACHED*/
	break;
      case CFHT_LIB:
	return(cfht_lib);
	/*NOTREACHED*/
	break;
      case CFHT_LIBCAMAC:
	return(cfht_libcamac);
	/*NOTREACHED*/
	break;
      case CFHT_LIBCCD:
	return(cfht_libccd);
	/*NOTREACHED*/
	break;
      case CFHT_LIBCFHT:
	return(cfht_libcfht);
	/*NOTREACHED*/
	break;
      case CFHT_LIBCFP:
	return(cfht_libcfp);
	/*NOTREACHED*/
	break;
      case CFHT_LIBCS:
	return(cfht_libcs);
	/*NOTREACHED*/
	break;
      case CFHT_LIBCX:
	return(cfht_libcx);
	/*NOTREACHED*/
	break;
      case CFHT_LIBDUCK:
	return(cfht_libduck);
	/*NOTREACHED*/
	break;
      case CFHT_LIBFF:
	return(cfht_libff);
	/*NOTREACHED*/
	break;
      case CFHT_LIBHELP:
	return(cfht_libhelp);
	/*NOTREACHED*/
	break;
      case CFHT_LIBHH:
	return(cfht_libhh);
	/*NOTREACHED*/
	break;
      case CFHT_LIBIDS:
	return(cfht_libids);
	/*NOTREACHED*/
	break;
      case CFHT_LIBMUSIC:
	return(cfht_libmusic);
	/*NOTREACHED*/
	break;
      case CFHT_LIBOCS:
	return(cfht_libocs);
	/*NOTREACHED*/
	break;
      case CFHT_LIBOMS:
	return(cfht_liboms);
	/*NOTREACHED*/
	break;
      case CFHT_LIBPIXD:
	return(cfht_libpixd);
	/*NOTREACHED*/
	break;
      case CFHT_LIBRET:
	return(cfht_libret);
	/*NOTREACHED*/
	break;
      case CFHT_LIBSSX:
	return(cfht_libssx);
	/*NOTREACHED*/
	break;
      case CFHT_LIBTCS:
	return(cfht_libtcs);
	/*NOTREACHED*/
	break;
      case CFHT_LIBUI:
	return(cfht_libui);
	/*NOTREACHED*/
	break;
    }
    return(TRUE);
}

/***********************************************************************
 *
 *   LogIt
 *
 * DESCRIPTION
 *    do most of the actual logging operation
 *
 * INVOKED
 *    LogIt(who, id, pid, type, mess);
 *
 * PARAMETERS IN
 *    cfht_log_who who - what part of system is requesting message?
 *    char *id - name of program doing the logging.
 *    int pid - process ID which generated the log message.
 *    cfht_log_type type - what kind of message is this?
 *    char *mess - message string to output
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    none
 *
 * GLOBALS
 *    cfht_error, cfht_warn, cfht_nodisp
 *    starttime, lasttime
 *
 * NOTE NOTE NOTE NOTE NOTE
 *    When I go back and look at distributed logging, remember to handle
 *    CFHT_LIBMUSIC calls specially so there's no infinite recursion :-)
 *
 ***********************************************************************/

static char nullmess[]= "(no message provided)";

/*
 * Forward declarations
 */
static void LogIt P((int who, char* id, int pid, int type, char *mess));

/* color codes for HPTERM and ansi-xterm error messages */

#define MAX_COLOR_SIZE 40	/* length of color setting string */

/* these go in the first dim of cfht_log_color[][] */
#define COL_EXTRA 0
#define COL_ERROR 1
#define COL_WARN 2

/* these go in the second dim of cfht_log_color[][] */
#define COL_START 0
#define COL_STOP 1

static char *cfht_log_color[3][2] = {
#ifdef hpux
    { "\033&v1S","" },	/* use color set 1 */
    { "\033&v2S","" },	/* use color set 2 */
    { "\033&v3S","" }	/* use color set 3 */
#else
/*
 * Else assume we are on a color xterm window.  ANSI color escape
 * sequences accept the following codes according to the color ls
 * man mages on linux:
 *
 *        0 to restore default color
 *        1 for brighter colors
 *        4 for underlined text
 *        5 for flashing text
 *       30 for black foreground
 *       31 for red foreground
 *       32 for green foreground
 *       33 for yellow (or brown) foreground
 *       34 for blue foreground
 *       35 for purple foreground
 *       36 for cyan foreground
 *       37 for white (or gray) foreground
 *
 * I also found that 7 is reverse video...
 */
    { "\033[01;32m","\033[0m" },	/* use color set 1 */
    { "\033[01;31m","\033[0m" },	/* use color set 2 */
    { "\033[01;33m","\033[0m" }		/* use color set 3 */
#endif
};

static void
LogIt(who, id, pid, type, mess)
    int who;
    char* id;
    int pid;
    int type;			/* message type */
    char *mess;
{
    static char logstr[CFHT_LOG_BUFFER_SIZE]; /* buffer for whole message */
    int len;			  /* Length of current string in logstr */
    struct timeval thistime;	  /* current time */
    static struct timezone tzp;	  /* local time zone */
    char logtail[CFHT_LOG_TAIL_SIZE]; /* [timing info] . newline */
    char *log_time;		  /* date or time of this call */
    char *types;		  /* printable message type */
    char *tails;		  /* pointer to [timing info] . newline */
    /*
     * The following two could be pointers, and the strcpy's could be
     * avoided.  But then on HP-UX (at least) purify shows UMR errors on the
     * sprintf call that formats the whole string.  :-(
     */
    char logu_prefix[MAX_COLOR_SIZE]; /* sequence or token to start color */
    char logu_postfix[MAX_COLOR_SIZE]; /* sequence or token to stop color */
    int elapsed_sec, elapsed_centsec, delta_sec, delta_centsec;

    /***************************************
     * take care of library initialization *
     ***************************************/

    if (mess == (char *) NULL)
	mess = nullmess;

    if (type == CFHT_LOG_ID) {
	SetUpLogging(mess);
	return;
    }

    if (!ready) {
	/* giving some good defaults - sss 910608 */
        SetUpLogging("try calling 'cfht_log(CFHT_MAIN, CFHT_LOG_ID, argv[0])'");
    }

    /********************************
     * set up type specific strings *
     ********************************/

    switch (type) {
     case CFHT_START:
	 log_time = cfht_date (NULL);
	 types = CFHT_LOGS_TYPESTRING_START;
	 tails = "\n";
	 break;

     case CFHT_STATUS:
	 log_time = cfht_time (NULL);
	 types = CFHT_LOGS_TYPESTRING_STATUS;
	 tails = ".\n";
	 break;

     case CFHT_LOGONLY:
	 log_time = cfht_time(NULL);
	 types = CFHT_LOGS_TYPESTRING_LOGONLY;
	 tails = ".\n";
	 break;

     case CFHT_WARN:
	 if (!cfht_warn || !cfht_error)
	     return;
	 log_time = cfht_time (NULL);
	 types = CFHT_LOGS_TYPESTRING_WARN;
	 tails = ".\n";
	 break;

     case CFHT_ERROR:
     case CFHT_ERRNO:
#if CFHT_FATAL != CFHT_ERROR
     case CFHT_FATAL:
#endif
	 if (!cfht_error)
	     return;
	 log_time = cfht_time (NULL);
	 types = CFHT_LOGS_TYPESTRING_ERROR;
	 if (type == CFHT_ERRNO) {
	     if (errno < 0 || errno > sys_nerr)
		 tails = " (Bad errno).\n";
	     else {
		 sprintf(logtail, " (%s).\n", strerror(errno));
		 tails = logtail;
	     }
	 } else
	     tails = ".\n";
	 break;

     case CFHT_DONE:
	 log_time = cfht_time (NULL);
	 types = CFHT_LOGS_TYPESTRING_DONE;
	 tails = ".\n";
	 break;

     case CFHT_DEBUG:
	 log_time = cfht_time (NULL);
	 types = CFHT_LOGS_TYPESTRING_DEBUG;
	 tails = ".\n";
	 break;

     case CFHT_TIMING:
	 log_time = cfht_time (NULL);
	 types = CFHT_LOGS_TYPESTRING_TIMING;
	 gettimeofday (&thistime, &tzp);
	 elapsed_sec = thistime.tv_sec - starttime.tv_sec;
	 elapsed_centsec = (thistime.tv_usec - starttime.tv_usec) / 10000;
	 /* handle second borrow */
	 if (elapsed_centsec < 0) {
	     elapsed_centsec += 100;
	     elapsed_sec--;
	 }
	 delta_sec = thistime.tv_sec - lasttime.tv_sec;
	 delta_centsec = (thistime.tv_usec - lasttime.tv_usec) / 10000;
	 /* handle second borrow */
	 if (delta_centsec < 0) {
	     delta_centsec += 100;
	     delta_sec--;
	 }
	 sprintf (logtail, " (%d.%02d, %d.%02d).\n",
		  elapsed_sec, elapsed_centsec,
		  delta_sec, delta_centsec);

	 memcpy (&lasttime, &thistime, sizeof (struct timeval));

	 tails = logtail;
	 break;

     default:
	 return;
    }

    /*********************
     * update system log *
     *********************/

    if (fdfile != LogFdDontUse) {
        len = sysv_sprintf(logstr, "%.*s %.*s%.*s%.*s(%d): %.*s%.*s",
			   CFHT_LOG_NAME_SIZE, whoarewe,
			   CFHT_LOG_TIME_SIZE, log_time,
			   CFHT_LOG_TYPE_SIZE, types,
			   CFHT_LOG_ID_SIZE, id,
			   pid,
			   CFHT_LOG_MESSAGE_SIZE, mess,
			   CFHT_LOG_TAIL_SIZE, tails);

	if (TEMP_FAILURE_RETRY(write(fdfile, logstr, len)) != len)
	    perror("(cfht_log) can not write to system log");
    }

    /*******************
     * update user log *
     *******************/

    /*
     * If no valid user log pipe is open, skip the rest.
     */
    if (fdpipe == LogFdDontUse)
        return;

    /*
     * START and DONE messages never go to the user log.
     * Note: Always pass them to director, because in this mode,
     * director is responsible for passing these messages to
     * the system log, and they would otherwise be lost completely.
     */
    if ((director == FALSE) &&
	((type == CFHT_START) || (type == CFHT_DONE)))
        return;

    /*
     * LOGONLY messages do not go to the user log when there is
     * a system log (debugging is FALSE) or when CFHTDEBUG is FALSE.
     * For director, however, LOGONLY messages must always be printed
     * here, even if CFHTDEBUG is FALSE.
     */
    if ((type == CFHT_LOGONLY)
	&& ((debugging == FALSE) || (LogThisDebugQ(who) == FALSE))
	&& (director == FALSE))
        return;

    /*
     * STATUS is suppressed from the user log when CFHTNODISP is TRUE.
     */
    if ((type == CFHT_STATUS) && (cfht_nodisp == TRUE))
        return;

    /*
     * DEBUG messages will only be generated if LogThisDebugQ() is true,
     * so only check the value of "debugging" to decide what to do with
     * DEBUG messages that make it this far.
     */
    if ((type == CFHT_DEBUG) && (debugging == FALSE))
        return;

    logu_prefix[0] = '\0';
    logu_postfix[0] = '\0';
    if (use_color) {
        if ((type == CFHT_ERROR) || (type == CFHT_ERRNO)
	    || (type == CFHT_FATAL)) {
	    (void) strcpy(logu_prefix, cfht_log_color[COL_ERROR][COL_START]);
	    (void) strcpy(logu_postfix, cfht_log_color[COL_ERROR][COL_STOP]);
	} else if (type == CFHT_WARN) {
	    (void) strcpy(logu_prefix, cfht_log_color[COL_WARN][COL_START]);
	    (void) strcpy(logu_postfix, cfht_log_color[COL_WARN][COL_STOP]);
	}
    }
    else {
        switch (type) {
	  case CFHT_ERRNO:
#if (CFHT_FATAL != CFHT_ERROR)
	  case CFHT_FATAL:
#endif
	  case CFHT_ERROR:
	    (void) strcpy(logu_prefix, "\r" ERROR_MSG);
	    break;
	  case CFHT_WARN:
	    (void) strcpy(logu_prefix, "\r" WARNING_MSG);
	    break;
	  /* The rest don't always get sent */
	  case CFHT_STATUS:
	    (void) strcpy(logu_prefix, "\r" STATUS_MSG);
	    break;
	  case CFHT_START:
	    (void) strcpy(logu_prefix, "\r" START_MSG);
	    break;
	  case CFHT_DONE:
	    (void) strcpy(logu_prefix, "\r" DONE_MSG);
	    break;
	  case CFHT_TIMING: /* Log these same as logonly's with director */
	  case CFHT_LOGONLY:
	    (void) strcpy(logu_prefix, "\r" LOGONLY_MSG);
	    break;
	  case CFHT_DEBUG:
	    (void) strcpy(logu_prefix, "\r" DEBUG_MSG);
	    break;
	}
    }

    len = sysv_sprintf(logstr, "%s%.*s: %.*s%.*s%s",
		       logu_prefix,
		       CFHT_LOG_ID_SIZE, id,
		       CFHT_LOG_MESSAGE_SIZE, mess,
		       CFHT_LOG_TAIL_SIZE, tails,
		       logu_postfix);

    if (TEMP_FAILURE_RETRY(write(fdpipe, logstr, len)) != len)
        perror ("(cfht_log) can not write to user log pipe");
}

/*****************************************************************************
 *
 * cfht_logv
 *
 * variable arg list version on cfht_log()
 * read up on vprintf(3) and varargs(5)
 *
 *****************************************************************************/

/*VARARGS3*/
void
cfht_logv Q((int who, int type, char *fmt, ...),
	    (who, type, fmt, va_alist),
	    int who; int type; char *fmt; va_dcl)
{
    va_list args;
    static char mess[CFHT_LOG_MESSAGE_SIZE];

/*
 * for any debug message, return immediately unless the environment variable
 * associated with the caller is specifically set to TRUE
 */
    if ((ready == TRUE)
	&& (type == CFHT_DEBUG) && (LogThisDebugQ(who) == FALSE))
      return;

#ifdef __STDC__
    va_start (args, fmt);
#else
    va_start (args);
#endif

    vsprintf (mess, fmt, args);
/*
 * Note - on an HP PA-RISC machine the above does not return if the resulting
 * string is too long.  The extra bytes have clobbered the return address from
 * vsprintf so there is no use even checking.  Thus, as the man page says, the
 * best we can do for that error is a core dump :-)
 */
    va_end (args);
    LogIt(who, log_id, (int)getpid(), type, mess);
}

/* version that logs parent PID instead for the cfht_log program to use */
/*VARARGS3*/
void
cfht_logpv Q((int who, int type, char *fmt, ...),
	     (who, type, fmt, va_alist),
	     int who; int type; char *fmt; va_dcl)
{
    va_list args;
    static char mess[CFHT_LOG_MESSAGE_SIZE];

/*
 * for any debug message, return immediately unless the environment variable
 * associated with the caller is specifically set to TRUE
 */
    if ((ready == TRUE)
	&& (type == CFHT_DEBUG) && (LogThisDebugQ(who) == FALSE))
      return;

#ifdef __STDC__
    va_start (args, fmt);
#else
    va_start (args);
#endif

    vsprintf (mess, fmt, args);
/*
 * Note - on an HP PA-RISC machine the above does not return if the resulting
 * string is too long.  The extra bytes have clobbered the return address from
 * vsprintf so there is no use even checking.  Thus, as the man page says, the
 * best we can do for that error is a core dump :-)
 */
    va_end (args);
    LogIt(who, log_id, (int)getppid(), type, mess);
}

/*****************************************************************************
 *
 * cfht_log
 *
 *****************************************************************************/

void
cfht_log Q((int who, int type, char *mess),
	    (who, type, mess),
	    int who; int type; char *mess;)
{
/*
 * for any debug message, return immediately unless the environment variable
 * associated with the caller is specifically set to TRUE
 */
    if ((ready == TRUE)
	&& (type == CFHT_DEBUG) && (LogThisDebugQ(who) == FALSE))
      return;

    LogIt(who, log_id, (int)getpid(), type, mess);
}

/*****************************************************************************
 *
 * cfht_log_proxy
 *
 * This does the same as cfht_log() but allows director to override
 * the log_id name and PID with each call, so that everything passing
 * through director doesn't get logged as if it came from director
 * itself.  Using this, director can make it seem as if an agent running
 * on a remote machine was running locally and logging directly to the
 * local roll.
 *
 *****************************************************************************/

void
cfht_log_proxy Q((int who, int type, char *mess, char* id, int pid),
		 (who, type, mess, id, pid),
		 int who; int type; char *mess; char* id; int pid;)
{
    /*
     * For this version, there is also no check for CFHT_DEBUG.
     * If director received a CFHT_DEBUG message from an agent,
     * it means debugging was turned on at that level and it should
     * get passed through regardless of director's own CFHT_DEBUG
     * environment setting.
     */

    /*
     * It's ok for id to be NULL here.  It just means log
     * with the usual log_id name.
     */
    LogIt(who,
	  (id != NULL) ? id : log_id,
	  (pid != 0) ? pid : (int)getpid(),
	  type, mess);
}
