/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* cfht_number.c
* $Header: cfht_number.c,v 1.6 96/11/19 23:09:06 isani Exp $
* $Locker:  $
*
* ROUTINES
*
* cfht_int()
*    This routine will attempt to parse a string, extract the first number
*    present in the string, and return the number as an integer.  The number
*    is actually parsed by cfht_double so it can be any format accepted by
*    that routine.  Optionally, this routine will also perform range checking
*    on the number.
*
*    INVOKED
*    success = cfht_int(in_str, out_addr, rangeinfo);
*
*    PARAMETERS IN
*    char *in_str - string to parse
*    cfht_range_t rangeinfo - pointer to structure containing range checking
*                             information, or NULL for no range checking
*
*    PARAMETERS OUT
*    int *out_addr;
*
*    FUNCTION RETURNS
*    PASSFAIL - indicates parse or range error, cfht_errno set on FAIL
*
*    GLOBALS
*    none
*
* cfht_double()
*    This routine will attempt to parse a string, extract the first number
*    present in the string, and return the number as a double.  The number can
*    be any standard floating point representation, hex (0xaaa), or a time
*    (hh:mm:ss[.fff]).  Optionally, this routine will also perform range
*    checking on the number.
*
*    INVOKED
*    success = cfht_double(in_str, out_addr, rangeinfo);
*
*    PARAMETERS IN
*    char *in_str - string to parse
*    cfht_range_t rangeinfo - pointer to structure containing range checking
*                             information, or NULL for no range checking
*
*    PARAMETERS OUT
*    double *out_addr;
*
*    FUNCTION RETURNS
*    PASSFAIL - indicates parse or range error, cfht_errno set on FAIL
*
*    GLOBALS
*    none
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jrw          27 sep 90       original coding
*
* $Log:	cfht_number.c,v $
 * Revision 1.6  96/11/19  23:09:06  isani
 * cleaned up rcsid's
 * moved sun hacks to "cfht.h"
 * renamed variable "time" to "timeval" because there's a function named "time"
 * 
 * Revision 1.5  1995/02/27  20:05:18  thomas
 * Made cfht_int do its own scan (not use cfht_double), moved hex to cfht_int
 * Changed both int and double to require a valid numeric field in the input
 *    (leading and trailing whitespace OK)
 * Changed time to allow MM:SS again
 *
 * Revision 1.4  95/01/05  16:53:42  16:53:42  thomas (Jim Thomas)
 * sun's need explicit definition of strtod
 * 
 * Revision 1.3  94/12/14  11:55:50  11:55:50  thomas (Jim Thomas)
 * Added #ifdef __LINT__ around RCS headers
 * Added includes of stdlib.h, limits.h, errno.h, took out math.h, values.h
 * Changed cfht/cfht.h to cfht.h
 * Fixed int range check so it accepts INT_MIN
 * Rewrote cfht_double to use strtol and strtod, removed parse_* routines
 * 
 * Revision 1.2  93/03/18  16:11:01  16:11:01  jwright (Jim Wright)
 * fix bug if fractional part is missing (e.g. "123." would fail)
 * 
 * Revision 1.1  91/09/24  17:00:14  17:00:14  steve (Steven Smith)
 * Original checkin
 * 
****************************************************************************!*/

#include <ctype.h>		  /* isascii(), etc. */
#include <stdlib.h>		  /* strtod & strtol */
#include <string.h>		  /* strspn() */
#include <limits.h>		  /* INT_MAX */
#include <errno.h>		  /* errno and ERANGE */

#include "cfht.h"		  /* libcfht difines */

RCSID("$Header: cfht_number.c,v 1.6 96/11/19 23:09:06 isani Exp $");

#define HexitBase 16		  /* base for hexadecimal */
#define DecimalBase 10		  /* base for decimal */
#define SexidecimalBase 60	  /* base for time */

/* #define VERBOSE_DEBUG */

#ifdef VERBOSE_DEBUG
#define _report_ printf("char '%c' rest '%s' result %f or %d\n", *str, str, \
			result, result);
#else
#define _report_
#endif

#define BOMB(reason) { cfht_errno=(cfht_errno ? cfht_errno : reason) ; \
                       _report_ ; return (FAIL); }

PASSFAIL cfht_int (in_str, out_addr, rangeinfo)
    char *in_str;
    int *out_addr;
    cfht_range_t rangeinfo;
{
    int result;			  /* resulting number */
    char *str;			  /* pointer into input string */
    char *end;			  /* pointer to end of strtol scan */

    str = in_str + strspn(in_str, " \t\n"); /* skip leading white space */
    if (*str == '\0') {
	BOMB (PARSE_FAILED);	  /* blank is not a number? */
    }
    errno = 0;			  /* strtol returns ERANGE on overflow */


    /* first try hex number - no decimal or exponent allowed */
    if ( (*str+1)
	&& (*str == '0') && ((*(str+1) == 'x') || (*(str+1) == 'X')) ) {
        result = strtol(str, &end, HexitBase);
    } else {
	/* not hex, must be decimal, don't allow octal */
	result = strtol(str, &end, DecimalBase);
    }
    
    if (errno != 0) {
	BOMB (OUT_OF_RANGE);	  /* too big */
    }

    str = end + strspn(end, " \t\n"); /* skip trailing spaces */
    if (*str != '\0') {
	BOMB (PARSE_FAILED);	  /* junk after number */
    }
    
    if (rangeinfo != (cfht_range_t) NULL) {
        if (rangeinfo->cfht_int_range.min_inclusive) {
            if (result < rangeinfo->cfht_int_range.min)
                BOMB (OUT_OF_RANGE);
        } else {
            if (result <= rangeinfo->cfht_int_range.min)
                BOMB (OUT_OF_RANGE);
        }
        if (rangeinfo->cfht_int_range.max_inclusive) {
            if (result > rangeinfo->cfht_int_range.max)
                BOMB (OUT_OF_RANGE);
        } else {
            if (result >= rangeinfo->cfht_int_range.max)
                BOMB (OUT_OF_RANGE);
        }
    }
    *out_addr = result;
#ifdef VERBOSE_DEBUG
    printf ("result is %d, *out_addr is %d\n", result, *out_addr);
#endif
    return (PASS);
}


PASSFAIL cfht_double (in_str, out_addr, rangeinfo)
    char *in_str;
    double *out_addr;
    cfht_range_t rangeinfo;
{
    char *str;			  /* scan pointer through input string */
    char *start;		  /* string start for strtod */
    char *end;			  /* string end from strtod */
    int sign = 1;		  /* sign this is sign of result */
    int timeval;		  /* integer part of time */
    int itemp;			  /* integer temp for time */
    char save;			  /* save character after seconds */
    double result = 0.0;	  /* working version of final answer */
    double temp = 0.0;		  /* float temp */

    if (in_str == (char *) NULL)  /* protect from null pointer */
        BOMB (PARSE_FAILED);

    str = in_str;		  /* preserve pointer to beginning of string */
    cfht_errno = 0;		  /* so return value is correct */

    /* skip any whitespace */
    str += strspn(str, " \t\n");
    if (*str == '\0')
        BOMB (PARSE_FAILED);    /* no number encountered! */

    /* check for (one only) sign */
    if (*str == '-') {
        sign = -1;
        str++;
    } else if (*str == '+') {
        str++;
    }
    if ((*str == '\0') || (*str == '+') || (*str == '-'))
        BOMB (PARSE_FAILED);

    /* the strto[ld] routines set errno to ERANGE if they get an overflow */
    errno = 0;

    /* see if this is a time 0:0:0[.0] */
    start = str;
    end = str + strspn(str, "0123456789");
    if (*end == ':') {
	timeval = strtol(str, &end, DecimalBase);
	str = end + 1;		  /* has to be the : */
	itemp = strtol(str, &end, DecimalBase);
	/* if another : shift over minutes else rescan as seconds */
	if (*end == ':') {
	    if (itemp >= SexidecimalBase)
	      BOMB (OUT_OF_RANGE);
	    timeval = (timeval * SexidecimalBase) + itemp;
	    str = end + 1;
	}
	result = timeval * SexidecimalBase;
	/* scan seconds as decimal, no exponent allowed */
	start = str;
	end = str + strspn(str, " 0123456789.");
	save = *end;
	*end = '\0';		  /* make sure strtod won't see any exponent */
	temp = strtod(start, &str);
	*end = save;
	if (temp >= SexidecimalBase)
	  BOMB (OUT_OF_RANGE);
	result += temp;

    } else {

	/* it is just a floating point number */
	result = strtod(str, &end);
	str = end;
    }

    /* did we get to the end of the input? */
    str += strspn(str, " \t\n");
    if (*str != '\0')
      BOMB (PARSE_FAILED);

    /* now see if any of the strto[dl] calls got an overflow */
    if (errno == ERANGE)
      BOMB (TOO_BIG);

    if (sign == -1)
        result = -result;

    if (rangeinfo != (cfht_range_t) NULL) {
        if (rangeinfo->cfht_double_range.min_inclusive) {
            if (result < rangeinfo->cfht_double_range.min)
                BOMB (OUT_OF_RANGE);
        } else {
            if (result <= rangeinfo->cfht_double_range.min)
                BOMB (OUT_OF_RANGE);
        }
        if (rangeinfo->cfht_double_range.max_inclusive) {
            if (result > rangeinfo->cfht_double_range.max)
                BOMB (OUT_OF_RANGE);
        } else {
            if (result >= rangeinfo->cfht_double_range.max)
                BOMB (OUT_OF_RANGE);
        }
    }
    *out_addr = result;
#ifdef VERBOSE_DEBUG
    printf ("result is %f, *out_addr is %f\n", result, *out_addr);
#endif

    return (PASS);
}
