/* Copyright (C) 1991-2000    Canada-France-Hawaii Telescope Corp.     */
/* This program is distributed WITHOUT any warranty, and is under the  */
/* terms of the GNU General Public License, see the file COPYING       */
/*!****************************************************************************
 * FILE
 * cfht_od.c
 * $Id: cfht_od.c,v 1.7 2002/01/23 16:17:20 isani Exp $
 * $Locker:  $
 *
 * ROUTINES
 *
 * cfht_od()
 *    This routine generates a number ("odometer" reading) that is
 *    guaranteed to be unique.  Consectutive calls will generate
 *    consecutive numbers.
 *
 *    INVOKED
 *    seq_num = cfht_od();
 *
 *    PARAMETERS IN
 *    none
 *
 *    PARAMETERS OUT
 *    none
 *
 *    FUNCTION RETURNS
 *    long - the next odometer number, or -1 indicating failure
 *
 *    GLOBALS
 *    none
 *
 * HISTORY
 * who          when            what
 * ---------    ------------    ----------------------------------------------
 * jk           21 Mar 1988     Original coding
 * jk           25 May 1988     hard coded path to odom.
 * jab          28 Jul 1988     went to cfp_file to get odometer path
 * jab          29 Jul 1988     drop year stamped numbers
 * jk           22 Jun 1989     changed lock file technology.
 * jk           02 Aug 1989     added auditing capabilities.
 *
 * $Log:	cfht_od.c,v $
 * Revision 1.7  2002/01/23  16:17:20  isani
 * Replaced cuserid with getpwuid(geteuid())
 * 
 * Revision 1.6  2000/11/02  03:16:46  thomas
 * Added error checks on read() and write() :-(
 * 
 * Revision 1.5  1998/12/11 18:22:22  thomas
 * Changed from cfp_file to looking for odometer file directly
 *
 * Revision 1.4  96/11/19  23:10:09  23:10:09  isani (Sidik Isani)
 * cleaned up rcsid's
 * changed a few of "%d"'s to "%ld" format strings to match argument
 * 
 * Revision 1.3  1995/02/28  20:16:06  thomas
 * Fix memory access error, stuff null after bytes read from odometer file
 *
 * Revision 1.2  94/12/14  12:00:12  12:00:12  thomas (Jim Thomas)
 * Added #ifdef __LINT__ around RCS headers
 * Added includes of stdio.h, string.h, errno.h
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h
 * Changed log messages to CFHT_LOGONLY with line numbers
 * 
 * Revision 1.1  91/09/24  17:01:04  17:01:04  steve (Steven Smith)
 * Original checkin
 * 
 **************************************************************************!*/

#include <stdio.h>		  /* for sscanf */
#include <fcntl.h>		  /* for open */
#include <unistd.h>		  /* for process control */
#include <stdlib.h>		  /* for getenv */
#include <string.h>		  /* for string routines */
#include <errno.h>		  /* for errno */
#include <pwd.h>		  /* for getpwuid */
#include <sys/types.h>		  /* same */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* parser defines */

RCSID("$Id: cfht_od.c,v 1.7 2002/01/23 16:17:20 isani Exp $");

#define MAX 99999999    /* maximum odometer reading */
#define BUF_LEN 80

long cfht_od()
{
    long exp_num;
    int fd;
    int bytes_read;		  /* number of bytes read from odometer file */
    char buf[BUF_LEN];
    char name_buf[CFP_PATHSIZE];  /* place for building path to file */
    char *name;
    int logfd;

    /**************************
     * find the odometer file *
     **************************/

    if ((name = getenv(CFHT_EOBSERVERPATH)) == (char *) NULL) {
	cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(cfht_od:%d) no observer definition in environment",
		  __LINE__);
	(void) strcpy(name_buf, CFHT_OBSERVERPATH);
    } else {
	(void) strcpy(name_buf, name);
    }
    (void) strcat(name_buf, CFHT_ODOMETERPATH);
    (void) strcat(name_buf, CFHT_ODOMETER);

    /**************************
     * open the odometer file *
     **************************/

    if ((fd = open(name_buf, O_RDWR)) < 0) {
        cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(cfht_od:%d) could not open odometer file '%s' (%s)",
		  __LINE__, name_buf, strerror(errno));
        return FAIL;
    }

    /*  Lock the file so that we are guaranteed a unique number */

    if (cfht_setlockwait(fd, TRUE) == FAIL) {
        /* possible sleep situation here */
        cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(cfht_od:%d) setlockwait on odometer file failed (%s)",
		  __LINE__, strerror(errno));
        return FAIL;
    }

    /*  Get the number inside and update it with the proper range checks */

    if ((bytes_read = read(fd, buf, BUF_LEN)) < 0) {
	cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(cfht_od:%d) read of odometer file failed (%s)",
		  __LINE__, strerror(errno));
	return FAIL;
    }
    buf[bytes_read] = '\0';
    if (sscanf (buf, "%ld", &exp_num) != 1) {
	cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(cfht_od:%d) did not find number in -%.100s-",
		  __LINE__, buf);
	return FAIL;
    }

    exp_num++;
    if ((exp_num > MAX) || (exp_num < 0)) {
        cfht_logv(CFHT_LIBCFHT, CFHT_STATUS,
                 "(cfht_od:%d) resetting odometer file to zero", __LINE__);
        exp_num = 0;
    }

    sprintf(buf, "%ld\n", exp_num);
    lseek(fd, 0L, 0);
    if (write(fd, buf, (unsigned)strlen (buf)) < 0) {
	cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(cfht_od:%d) failed writing number back out (%s)",
		  __LINE__, strerror(errno));
	return FAIL;
    }
   
    /*  Cleanup and get out */

    sprintf(buf,"%s.log", name_buf);
    if ((logfd = open(buf, O_RDWR|O_APPEND|O_CREAT, 0666)) < 0)
        cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(cfht_od:%d) could not open odometer log file (%s)",
		  __LINE__, strerror(errno));
    else {
        struct passwd* pw = getpwuid(geteuid());

        sprintf(buf,"(%s) %s took odometer number %ld\n", cfht_date(NULL),
		pw ? pw->pw_name : "unknown", exp_num);
        if (write(logfd, buf, strlen (buf)) < 0) {
	    cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		      "(cfht_od:%d) failed appending to log file (%s)",
		      __LINE__, strerror(errno));
	}
        close(logfd);
    }

    close(fd);  /* this will release lock */
    return exp_num;
}
