/* Copyright (C) 1996-2002    Canada-France-Hawaii Telescope Corp.      */
/* This program is distributed WITHOUT any warranty, and is under the   */
/* terms of the GNU General Public License, see the file COPYING        */
/*
 * cfht_signal.c - block, unblock, install handler, 
 *                 and check for signal event facilities
 *
 * $Id: cfht_signal.c,v 1.6 2002/04/03 21:23:03 thomas Exp $
 * $Log: cfht_signal.c,v $
 * Revision 1.6  2002/04/03 21:23:03  thomas
 * Added printing of error message on sigaction failure
 *
 * Revision 1.5  2000/02/10 13:09:54  isani
 * Added cast to sa_handler to avoid warnings from gcc about type difference.
 *
 * Revision 1.4  1997/05/28  02:50:29  isani
 * renamed builtin_signal_handler to cfht_signal_handler and made public
 *
 * Revision 1.3  97/04/07  20:11:14  thomas
 * Changed cfht_signal event routines to return BOOLEAN
 * 
 * Revision 1.2  97/02/07  22:27:18  22:27:18  isani (Sidik Isani)
 * Fixed a typo in a comment
 * 
 * Revision 1.1  96/11/22  03:58:19  isani
 * Initial revision
 * 
 *
 * DESCRIPTION
 *
 * Call cfht_signal() just as you would call signal(), with two arguments
 * (the signal, and a function pointer to your handler.)  See the man page
 * for details and examples.
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#include "cfht.h"		  /* libcfht defines */

RCSID("$Id: cfht_signal.c,v 1.6 2002/04/03 21:23:03 thomas Exp $");

/*
 * This table contains an entry for each signal.  When a signal handler
 * is installed, cfht_signal clears the flag for that signal.  If the
 * builtin signal handler is used, then it gets set to 1 and this can
 * be tested with cfht_signal_peek(), or tested and cleared with
 * cfht_signal_event().
 */
static volatile BOOLEAN builtin_signal_state[NSIG];

/*
 * Use this function to start blocking a particular signal.  As long
 * as the signal is blocked, caught signals will be queued up.  (But
 * you can only determine whether a signal happened, and not how many
 * times it might have happened.)  If you actually want to DISCARD
 * signals, use cfht_signal(signo,SIG_IGN).
 */
void cfht_signal_block(signo)
    int signo;
{
    sigset_t mask;
    
    (void) sigemptyset(&mask);
    (void) sigaddset(&mask, signo);
    (void) sigprocmask(SIG_BLOCK, &mask, 0);
}

/*
 * Use this function to allow the handler for signo to run again.  If
 * and events are queued up, the handler will run immediately.
 */
void cfht_signal_unblock(signo)
    int signo;
{
    sigset_t mask;
    
    (void) sigemptyset(&mask);
    (void) sigaddset(&mask, signo);
    (void) sigprocmask(SIG_UNBLOCK, &mask, 0);
}


/*
 *
 * This function sets up the signal catching for any signal.
 * `signo' is the signal you want to trap.  If you want to block
 * the signal initially, you might want to block it first, then
 * call cfht_signal() to install the handler, and then unblock
 * when you're ready to handle the signal.
 *
 * If desired, a default builtin handler can be used
 * which just records whether an event (or multiple events ... there
 * is no reliable way to distinguish) has occurred.  If you go this
 * route, you should use the function cfht_signal_event() to check
 * if the signal happened.
 */
void cfht_signal(signo, handler)
    int signo;
    PFV handler;
{
    struct sigaction vec_trap;
  
    if (signo >= NSIG) {
        cfht_logv(CFHT_LIBCFHT, CFHT_WARN,
		  "(cfht_signal:%d) signo %d out of range (0-%d)",
		  __LINE__, signo, NSIG);
	return;
    }
    builtin_signal_state[signo] = FALSE;
    /*
     * cfht_signal()'s handlers don't make use of the (int) argument
     * but sa_handler still needs the function to be case to the
     * proper type to avoid a warning here.
     */
    vec_trap.sa_handler = (void(*)P((int)))handler;
    (void) sigemptyset(&vec_trap.sa_mask);
#ifdef SA_INTERRUPT
    /*
     * If this particular macro is defined, then it means we are compiling
     * on a BSD system where the default (even with sigaction()) might be to
     * restart system calls after handling a signal.  Setting this flag in
     * the sa_flags field tells BSD systems to return EINTR right away
     * instead, whenever any system call is interrupted.
     */
    vec_trap.sa_flags = SA_INTERRUPT;
#else
    vec_trap.sa_flags = 0;
#endif /* SA_INTERRUPT */
    
    /*
     * install the handler
     */
    if (sigaction(signo, &vec_trap, NULL) != 0) {
        cfht_logv(CFHT_LIBCFHT, CFHT_WARN,
		  "(cfht_signal:%d) sigaction failed to install handler (%s)",
		  __LINE__, strerror(errno));
    }
}

/*
 *
 * cfht_signal_handler -- saves a flag so cfht_signal_event() can tell if
 *                        the signal signo has been trapped yet.
 *
 */

void
cfht_signal_handler(signo)
    int signo;
{
    builtin_signal_state[signo] = TRUE;
}

/*
 *
 *  NAME:  cfht_signal_event
 *
 *  This function tests and clears the global flag "builtin_signal_state"
 *  
 *  The idea here is that if a SIGXYZ was set up and subsequently trapped in
 *  the handler, the global flag would have been set. When a process that is
 *  pausing (see pause(2)) on this event is awaken, it should first check 
 *  that the wakeup call was really for SIGCXYZ.
 *
 *  INVOKED: while not (cfht_signal_event (SIGXYZ))
 *               sleep
 *           end while
 *           proceed with state checking
 *
 *  FUNCTION RETURNS:
 *      TRUE - global flag set, indicating at least 1 SIGXYZ event
 *      FALSE - SIGXYZ not caught since last time you called cfht_signal_event
 *
 *  GLOBALS: builtin_signal_state[]
 */

BOOLEAN
cfht_signal_event(signo)
    int signo;
{
    if (signo >= NSIG) {
	cfht_logv(CFHT_LIBCFHT, CFHT_WARN,
		  "(cfht_signal_event:%d) signo %d out of range (0-%d)",
		  __LINE__, signo, NSIG);
	return TRUE;
    }
    if (builtin_signal_state[signo] == TRUE) {
        builtin_signal_state[signo] = FALSE;
        return TRUE;
    } else {
        return FALSE;
    }
}

/*
 * NAME: cfht_signal_peek - see if there's been an event, but don't
 *                          reset the indicator flag
 *  INVOKED: if(cfht_signal_peek(signo)) printf("we have had a signo\n");
 *
 *  FUNCTION RETURNS:
 *      TRUE if signo has happened since last of cfht_signal or cfht_event
 *      FALSE if signal has not happened since install or reset
 *      and garbage if no handler installed yet
 *      and TRUE if signo out of range
 */

BOOLEAN
cfht_signal_peek(signo)
    int signo;
{
    if (signo >= NSIG) {
	cfht_logv(CFHT_LIBCFHT, CFHT_WARN,
		  "(cfht_signal_peek:%d) signo %d out of range (0-%d)",
		  __LINE__, signo, NSIG);
	return TRUE;
    }
    return (builtin_signal_state[signo]);
}

/* end of cfht_signal.c */
