/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* cfht_string.c
* $Header: cfht_string.c,v 1.3 96/11/19 23:11:31 isani Exp $
* $Locker:  $
*
* ROUTINES
*
* cfht_string()
*    This routine performs a number of manipulations on strings.  The
*    intention is to provide sanity checking when dealing with a human
*    at the keyboard.
*
*    INVOKED
*    out_str = cfht_string(in_str, operation, rangeinfo);
*
*    PARAMETERS IN
*    char *in_str - string to examine; this copy is never modified
*    int operation - specify string operation to perform
*    cfht_range_t rangeinfo - pointer to structure containing range checking
*                             information, or NULL for no range checking
*
*    PARAMETERS OUT
*    none
*
*    FUNCTION RETURNS
*    char * - modified string or NULL on failure with cfht_errno set
*
*    GLOBALS
*    none
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jrw          01 Nov 90       original coding
* jrw          15 Feb 1991     check for null pointers
*
* $Log:	cfht_string.c,v $
 * Revision 1.3  96/11/19  23:11:31  isani
 * cleaned up rcsid
 * changed a couple of NULL's to '\0'
 * 
 * Revision 1.2  1994/12/14  12:05:44  thomas
 * Added #ifdef __LINT__ around RCS headers
 * Added includes of string.h, stdlib.h(HP), malloc.h(Sun)
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h
 *
 * Revision 1.1  91/09/24  17:02:42  17:02:42  steve (Steven Smith)
 * Original checkin
 * 
*
****************************************************************************!*/

#include <ctype.h>		  /* isascii(), etc. */
#include <errno.h>
#include <string.h>		  /* for string routines */
#include <stdlib.h>		  /* for calloc */

#include "cfht.h"		  /* libcfht defines */

RCSID("$Header: cfht_string.c,v 1.3 96/11/19 23:11:31 isani Exp $");

/* #define VERBOSE_DEBUG */

#ifdef VERBOSE_DEBUG
#define _report_ printf("char '%c' rest '%s' result %s\n", *str, str, result);
#else
#define _report_
#endif

#define BOMB(reason) { cfht_errno=(cfht_errno ? cfht_errno : reason) ; \
                       if (result) free(result); _report_ ; return (NULL); }

#define NUMCHARS 256        /* how many possible character codes in char */


char *cfht_string (in_str, operation, rangeinfo)
char *in_str;
int operation;
cfht_range_t rangeinfo;
{
    char *str;              /* preserve pointer passed in */
    char *result = NULL;    /* pointer to start of result string */
    char *rp;               /* working pointer within result string */
    size_t origlen;         /* length of string passed in */
    char *p;                /* general purpose pointer to string */
    char **pp;              /* general purpose pointer to array of string */
    int n;                  /* general purpose int */
    int lookup[NUMCHARS];   /* true/false lookup table */

    if (in_str == NULL)
        BOMB(OUT_OF_RANGE);
    str = in_str;
    origlen = strlen(in_str);
    /* most options will not increase size of string, so grab some memory */
    result = (char *)calloc(origlen+1, sizeof(char));
    if (result == NULL) {
        if (errno == 0)
            errno = ENOMEM;
        BOMB(CHECK_ERRNO);
    }
    rp = result;

    switch (operation) {
    case CFHT_STR_MATCH:
        if (rangeinfo == NULL)
            BOMB(INVALID_RANGE);
        for (pp=rangeinfo->cfht_string_range.string_array ; *pp ; pp++) {
            if (strcmp(in_str, *pp) == 0) {
                /* got a match! */
                (void)strcpy(result, in_str);
                break;  /* out of loop */
            }
        }
        break;
    case CFHT_STR_INCLUDE:
        if (rangeinfo == NULL)
            BOMB(INVALID_RANGE);
        for (n=0 ; n<NUMCHARS ; n++) {
            lookup[n] = FALSE;
        }
        for (p=rangeinfo->cfht_char_range.list_of_chars ; *p ; p++) {
            lookup[(unsigned char)(*p)] = TRUE;
        }
        while (*str) {
            if (lookup[(unsigned char)(*str)]) {
                *rp++ = *str++;
            } else {
                str++;
            }
        }
        *rp = '\0';
        break;
    case CFHT_STR_EXCLUDE:
        if (rangeinfo == NULL)
            BOMB(INVALID_RANGE);
        for (n=0 ; n<NUMCHARS ; n++) {
            lookup[n] = TRUE;
        }
        for (p=rangeinfo->cfht_char_range.list_of_chars ; *p ; p++) {
            lookup[(unsigned char)(*p)] = FALSE;
        }
        while (*str) {
            if (lookup[(unsigned char)(*str)]) {
                *rp++ = *str++;
            } else {
                str++;
            }
        }
        *rp = '\0';
        break;
    case CFHT_STR_NOWHITE:
        while (*str) {
            if (! isspace(*str)) {
                *rp++ = *str++;
            } else {
                str++;
            }
        }
        *rp = '\0';
        break;
    case CFHT_STR_NOTRAIL:
        while (*str)
            *rp++ = *str++;
        *rp = '\0';
        while ((rp > result) && isspace(*(rp-1)))
            *(--rp) = '\0';
        break;
    case CFHT_STR_ESCAPE:
        BOMB(UNIMPLEMENTED);
	/*NOTREACHED*/
        break;
    case CFHT_STR_UNESCAPE:
        BOMB(UNIMPLEMENTED);
	/*NOTREACHED*/
        break;
    default:
        BOMB(BAD_OPTION);
	/*NOTREACHED*/
        break;
    }

    return (result);
}
