/* Copyright (C) 1991-1998    Canada-France-Hawaii Telescope Corp.       */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING         */

/*****************************************************************************
 *
 * file: cfp.h
 * $Header: cfp.h,v 2.12 2000/05/24 20:09:43 isani Exp $
 * $Locker:  $
 *
 * This file contains public defines for the configuration file routines.
 * (cfp stands for Configuration File Parsing.)
 *
 * HISTORY
 *
 * who        when           what
 * -------    -----------    ----------------------------------------------
 * jab        27 Jan 1988    Original coding
 * jab        29 Apr 1988    Added already included check
 * jab         8 Aug 1988    Added baud rate to ioconfig stuff
 * jab        17 Aug 1988    change auto environment to file name driven
 * jk         16 Apr 1991    added macros for cfp routines
 * jrw        09 May 1991    added function definitions (ANSI and crufty flavors)
 *
 * $Log:	cfp.h,v $
 * Revision 2.12  2000/05/24  20:09:43  isani
 * Changed <cfht/cfht.h> to "cfht.h"
 * 
 * Revision 2.11  98/12/16  11:47:05  thomas
 * Added CFP_READFIXED for cfp_file
 * 
 * Revision 2.10  98/12/11  18:22:48  18:22:48  thomas (Jim Thomas)
 * Added CFHT_ODOMETERPATH
 * 
 * Revision 2.9  98/07/07  15:06:53  15:06:53  thomas (Jim Thomas)
 * Added CFP_READONLY and CFP_READWRITE
 * 
 * Revision 2.8  97/09/10  17:46:49  17:46:49  thomas (Jim Thomas)
 * Added cfp_reread
 * 
 * Revision 2.7  97/05/17  01:32:25  01:32:25  thomas (Jim Thomas)
 * Changed CFP_* macros to be CFHT_LIBCFP , added CFP_PUTVAL
 * 
 * Revision 2.6  96/11/19  23:14:01  23:14:01  isani (Sidik Isani)
 * added prototypes for cfp_getinvokes, cfp_getvars, cfp_editvalue, and cfp_rmline
 * 
 * Revision 2.5  1995/02/27  20:06:08  thomas
 * Took out private P definition, added cfp_rangep and cfp_editstring
 *
 * Revision 2.4  95/01/04  09:51:00  09:51:00  thomas (Jim Thomas)
 * Added cfp_rangemess prototype
 * 
 * Revision 2.3  94/12/28  17:05:54  17:05:54  thomas (Jim Thomas)
 * Moved strerror definition for suns to cfht.h
 * 
 * Revision 2.2  94/12/21  12:39:34  12:39:34  thomas (Jim Thomas)
 * Added cfp_lookup for checking on a name's existence in a par file
 * 
 * Revision 2.1  94/12/14  12:24:23  12:24:23  thomas (Jim Thomas)
 * Added definitions for cfp version 2
 * Moved out internal cfp definitions
 * Changed function prototypes to use P macro
 * 
 * Revision 1.3  94/02/02  14:12:14  14:12:14  jwright (Jim Wright)
 * remove typedef of FILE_ID form cfp.h and put it in cfht.h.  reason is
 * that cfht_exec() has an argument of type FILE_ID and thus needs to have
 * this type visible so that it can correctly declare an ANSI prototype.
 * it is presumed that any file including cfp.h will first have included
 * cfht.h.
 * 
 * Revision 1.2  92/05/14  09:21:31  09:21:31  john (John Kerr)
 * added constants CFP_VARIABLES, CFP_INVOKES, CFP_FORMS
 * 
 * Revision 1.1  91/09/24  17:06:26  17:06:26  steve (Steven Smith)
 * Original checkin
 * 
 *****************************************************************************/

#ifndef CFPDOTH
#define CFPDOTH

#include "cfht.h" /* This works even when installed if "gcc -I." is used. */

/*************
 * constants *
 *************/

#define CFP_PAR         (-33157)  /* fakeout for standard .par file */
#define CFP_STRSIZE     80        /* temporary string size */
#define CFP_FILESIZE	32	  /* parse file name size */
#define CFP_PATHSIZE    256       /* parse file path size */
#define CFP_MAXINST     10        /* maximum number of active parse files */
#define CFP_READONLY    "r"	  /* cfp_file open read only */
#define CFP_READWRITE   "rw"	  /*               read/write */
#define CFP_READFIXED   "rf"	  /*               read fixed - don't copy */

/* ioconfig stuff */
#define CFP_HPIB        0
#define CFP_RS232       1
#define CFP_CAMAC       2
#define CFHT_IOCONFIG   "ioconfig"

/* par file keywords */

#define CFP_VARIABLES "variables"
#define CFP_INVOKES "invokes"
#define CFP_FORMS "forms"

/* odometer */
#define CFHT_ODOMETERPATH "/conf/" /* subdirectory containing below */
#define CFHT_ODOMETER   "odometer"


/********************
 * type definitions *
 ********************/

/*
 * Variable history values
 *
 * For each variable, a code is maintained indicating the source of the
 * current value.  Internally this is used to decide whether to update the
 * value when a file is written back out (at least historically).
 */

typedef enum {
    CFP_FAILED = -1,		  /* not really history, failure return value */
    CFP_UNCHECKED = 0,		  /* untouched (historical) */
    CFP_CURRENT,		  /* value came from current field of file */
    CFP_DEFAULT,		  /* value came from default field of file */
    CFP_ENVIRONMENT,		  /* value came from environment */
    CFP_FOUND_CURRENT,		  /* command arg present without value
				       value from file current field */
    CFP_FOUND_DEFAULT,		  /* command arg present without value
				       value from default */
    CFP_FOUND_SUPPLIED,		  /* command arg present without value
				       value from supplied */
    CFP_FOUND_ARGV,		  /* command arg present
				       and value also from command line */
    CFP_SET			  /* value set by program (putcur) */
} HistoryType;


/*
 * Varable definition line index values
 *
 * A variable definition line has the format:
 *      <name> = { <mode>, <current>, <default>, <supplied>, <min>, <max> }
 * .  The fields within the {} are accessible to the program using the index
 * values defined by this type and the getval/putval calls.  In addition, the
 * environment can be specified as the source or destination.
 */

typedef enum {
    CFP_EnvironmentIndex = -1,	  /* reference the environment */
    CFP_ModeIndex = 0,		  /* reference the mode field */
    CFP_CurrentIndex,		  /* reference the current value */
    CFP_DefaultIndex,		  /* reference the default value */
    CFP_SuppliedIndex,		  /* reference the supplied value */
    CFP_RangeIndex		  /* reference the range value */
} ItemIndexType;

/* parse line type definitions */
#define CFP_BOOLEAN          0
#define CFP_INTEGER          1
#define CFP_DOUBLE           2
#define CFP_STRING           3

/*************************************
 * macros -- Added by JK 16 Apr 1991 *
 *************************************/

#define CFP_GETCUR(id, key, addr, type, size) \
    { if (cfp_getcur(id, key, addr, type, size) == FAIL) \
        { cfht_log(CFHT_LIBCFP, CFHT_LOGONLY, "getcur failed"); return(FAIL); } }

#define CFP_GETVAL(id, key, addr, type, size, item) \
    { if (cfp_getval(id, key, addr, type, size, item) == FAIL) \
        { cfht_log(CFHT_LIBCFP, CFHT_LOGONLY, "getval failed"); return(FAIL); } }

#define CFP_GETSTR(id, key, addr, size) \
    { if (cfp_getstr(id, key, addr, size) == FAIL) \
        { cfht_log(CFHT_LIBCFP, CFHT_LOGONLY, "getstr failed"); return(FAIL); } }

#define CFP_PUTCUR(id, key, addr, type) \
    { if (cfp_putcur(id, key, addr, type) == FAIL) \
        { cfht_log(CFHT_LIBCFP, CFHT_LOGONLY, "putcur failed"); return(FAIL); } }

#define CFP_PUTVAL(id, key, addr, type, item) \
    { if (cfp_putval(id, key, addr, type, item) == FAIL) \
        { cfht_log(CFHT_LIBCFP, CFHT_LOGONLY, "putval failed"); return(FAIL); } }


/*************************
 * structure definitions *
 *************************/

/*
 * This definition describes the public part of the CFP database.  It is
 * provided for backwards compatibility with CFP version 1.
 */

struct cfp_inst {
    FILE_ID id;                   /* real id for those who have pointers */
    char file[CFP_FILESIZE];      /* file name with no ., or .par */
    char path[CFP_PATHSIZE];      /* path name of file */
};

struct cfp_inst_GXE {
    FILE_ID id;                   /* real id for those who have pointers */
    char file[CFP_FILESIZE];      /* file name with no ., or .par */
    char path[CFP_PATHSIZE];      /* path name of file */
    char *data;			  /* copy of file - GXE ONLY GXE ONLY */
};

struct cfp_io {
    int type;                     /* CFP_HPIB, CFP_RS232, or CFP_CAMAC */
    char file[80];                /* device node name */
    int hpib;                     /* HP-IB device address, or -1 */
    int branch;                   /* CAMAC branch number, or -1 */
    int crate;                    /* CAMAC crate number, or -1 */
    int module;                   /* CAMAC module number, or -1 */
    int cflag;                    /* baud rate | bits | parity, or -1 */
};

/*********************
 * global references *
 *********************/

char *cfp_cursub P((FILE_ID id, char *string, int size));
char *cfp_file P((char *addr, char *name, char *suff, char *mode));
PASSFAIL cfp_finish P((FILE_ID id));
HistoryType cfp_getcur P((FILE_ID id, char *name, void *addr, int type,
			  int size));
FILE_ID cfp_getf P((FILE_ID id));
char *cfp_getln P((char *dest, char *key, char *name, char *suff));
int cfp_getsiz P((FILE_ID id, char *name));
PASSFAIL cfp_getinvokes P((FILE_ID id, char **list, int *count));
PASSFAIL cfp_getvars P((FILE_ID id, char **list, int *count));
PASSFAIL cfp_getstr P((FILE_ID id, char *name, char *addr, int size));
PASSFAIL cfp_getval P((FILE_ID id, char *name, void *addr, int type, int size,
		       ItemIndexType valnum));
FILE_ID cfp_init P((char *name));
struct cfp_io *cfp_io P((char *name, struct cfp_io *data, int type));
PASSFAIL cfp_lookup P((FILE_ID id, char *name));
int cfp_match P((char *name, char **addrs));
int cfp_parse P((FILE_ID id, int argc, char **argv));
PASSFAIL cfp_putcur P((FILE_ID id, char *name, void *addr, int type));
PASSFAIL cfp_putf P((FILE_ID id));
PASSFAIL cfp_putln P((char *src, char *key, char *name, char *suff));
PASSFAIL cfp_putstr P((FILE_ID id, char *name, char *string));
PASSFAIL cfp_putval P((FILE_ID id, char *name, void *addr, int type,
		       ItemIndexType valnum));
BOOLEAN cfp_rangep P((FILE_ID id, char *name));
PASSFAIL cfp_reread P((FILE_ID id));
PASSFAIL cfp_editvalue P((FILE_ID id, char *name, char *addr, int type));
PASSFAIL cfp_editstring P((FILE_ID id, char *name, char *addr, int type));
PASSFAIL cfp_range P((FILE_ID id, char *name, char *addr, int type));
PASSFAIL cfp_rmline P((FILE_ID id, char *Key));
char *cfp_rangemess P((FILE_ID id, char *name));

/* holdovers from cfp version 1 */
struct cfp_inst *cfp_chid P((FILE_ID id));
struct cfp_inst_GXE *cfp_chidForGXE P((FILE_ID id));

#endif /* CFPDOTH */
