/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* cfp_cursub.c
* $Header: cfp_cursub.c,v 2.2 96/11/19 23:14:10 isani Exp $
* $Locker:  $
*
* ROUTINES
*
*   cfp_cursub()
*
* DESCRIPTION
*    This function takes a string and and replaces variable values.  Tokens
*    that start with % will be cfp_getcur()'ed for the real token.  Tokens
*    that start with $ will be cfp_getcur()'ed and if TRUE -token will be
*    left.  $% or %% will be replaced with %.  $$ or %$ will be replaced with
*    $.
* 
*    An error occurs if the translation results in more characters than will
*    fit in the space provided or if the input string is outrageously long.
*
*    The input string is replaced by the translation and a pointer to it is
*    returned.
*
* INVOKED
*    dest_str = cfp_cursub(id, string, size);
*
* PARAMETERS IN
*    FILE_ID id - the current par file
*    char *string - the string to be translated
*    int size - the maximum size of the string
*
* PARAMETERS OUT
*    char *string - the translated string
*
* FUNCTION RETURNS
*    char * - a pointer to the translated string, or NULL on error
*
* GLOBALS
*    none
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jab          15 Jun 1989     Tear out of cfht_exec and rename
*
* $Log:	cfp_cursub.c,v $
 * Revision 2.2  96/11/19  23:14:10  isani
 * cleaned up rcsid
 * 
 * Revision 2.1  1994/12/14  12:32:33  thomas
 * Added #ifdef __LINT__ around RCS headers
 * Removed include of sys/types.h
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h, added cfp_internal.h
 * Changed CFHT_ERROR to CFHT_LOGONLY, added debug messages for quiet failures
 *
 * Revision 1.1  91/09/24  17:07:11  17:07:11  steve (Steven Smith)
 * Original checkin
 * 
*
****************************************************************************!*/

#include <errno.h>		  /* for errno */
#include <string.h>		  /* for string handling defines */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* parfile stuff */
#include "cfp_internal.h"	  /*   and internals */

RCSID("$Header: cfp_cursub.c,v 2.2 96/11/19 23:14:10 isani Exp $");

char *				  /* dest string or NULL */
cfp_cursub(id, string, size)
    FILE_ID id;
    char *string;		  /* input AND output string */
    int size;			  /* max size of string */
{
    FileEntryType *FileEntry;	  /* pointer to file data for errors */
    char tstringdata[CFHT_ARGS_SIZE]; /* copy of original string */
    char *tstring;		  /* pointer into input string */
    BOOLEAN useit;		  /* for $name translations */
    char *item;			  /* pointer to a name to be looked up */
    char type;			  /* $ or % */
    char term;			  /* character at end of name (save from nul) */
    char curval[CFHT_VALUE_SIZE + 1]; /* translated value for %name */
    char *curarg;		  /* pointer into result string */
    char *maxarg;		  /* pointer to far end of result */
    char *p;			  /* scratch pointer into result string */

    if (!string) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_cursub:%d) missing string", __LINE__);
        errno = EINVAL;
        return((char *) NULL);
    }

    (void) strncpy(curval, string, CFHT_VALUE_SIZE);
    curval[CFHT_VALUE_SIZE] = '\0';
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_cursub:%d) starting, string = '%s'",
	      __LINE__, curval);

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_cursub:%d) cfp_CheckID failed on '%s'",
		  __LINE__, curval);
        return((char *) NULL);
    }

    if (strlen(string) >= (CFHT_ARGS_SIZE)) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_cursub:%d) input string too long (%d>=%d) for '%s'",
		  __LINE__, strlen(string), CFHT_ARGS_SIZE,
		  FileEntry->FilePath);
	return((char *) NULL);
    }

    tstring = tstringdata;
    strcpy(tstring, string);    /* so src and dest don't collide */

    curarg = string;
    maxarg = &string[size];

    while (*tstring) {
        size = strcspn(tstring, "$%");

        /* copy unmatched stuff to destination string */
        if (size > 0) {
            p = curarg;
            curarg += size;
            if (curarg >= maxarg) {
                cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_cursub:%d) string too big for '%s'",
			  __LINE__, FileEntry->FilePath);
                errno = E2BIG;
                return((char *) NULL);
            }
            strncpy(p, tstring, (size_t)size);
            tstring += size;
            if (!*tstring)
	      continue;
        }
        type = *tstring++;
    
        /* check for escaped % or $ */
        if (*tstring == '$' || *tstring == '%') {
            if (curarg >= maxarg) {
                cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_cursub:%d) result too big for '%s'",
			  __LINE__, FileEntry->FilePath);
                errno = E2BIG;
                return(NULL);
            }
            *curarg++ = *tstring++;
            continue;
        }

        /* take a good guess at chars that could terminate a name */
        size = strcspn(tstring, " \t'\",;:><?*/`~\n");
        item = tstring;
        term = tstring[size];
        tstring = &tstring[size];
        *tstring = '\0';

	useit = FALSE;

	/* do boolean substitution */
	if (type == '$') {
	    if (cfp_getcur(id, item, (char *) &useit,
			   CFP_BOOLEAN, sizeof(useit))
		== FAIL) {
		cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
			  "(cfp_cursub:%d) lookup boolean '%s' failed for '%s'",
			  __LINE__, item, FileEntry->FilePath);
		return(NULL);
	    }
	    item--;
	    *item = '-';
	}

	/* do string substitution */
	if (type == '%') {
	    if (cfp_getcur(id, item, curval, CFP_STRING, sizeof(curval))
		== FAIL) {
		cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
			  "(cfp_cursub:%d) lookup string '%s' failed for '%s'",
			  __LINE__, item, FileEntry->FilePath);
		return(NULL);
	    }
	    item = curval;
	    useit = TRUE;
	}

	/* copy token back to destination string */
	if (useit) {
	    p = curarg;
	    curarg += strlen(item);
	    if (curarg >= maxarg) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_cursub:%d) result too big for '%s'",
			  __LINE__, FileEntry->FilePath);
		errno = E2BIG;
		return(NULL);
	    }
	    strcpy(p, item);
	}

        *tstring = term;
    }
    *curarg = '\0';

    (void) strncpy(curval, string, CFHT_VALUE_SIZE);
    curval[CFHT_VALUE_SIZE] = '\0';
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_cursub:%d) ending, string = '%s'",
	      __LINE__, curval);

    return(string);
}
