/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* cfp_getcur.c
* $Header: cfp_getcur.c,v 2.4 96/11/19 22:57:39 isani Exp $
* $Locker:  $
*
* ROUTINES
*
*   cfp_getcur()
*
* DESCRIPTION
*    This routine will extract the current value of a variable and return
*    its value.  If the current value has not been set yet and is defined in
*    the environment (under the name
*      <par file name w/o ., or .par><variable name>
*    ), the current value is filled in from the environment first.  If the
*    current value has not been set yet and there is a supplied value, the
*    current value is set from the supplied value.  If the current value has
*    not been set yet and there is a default value, the current value is set
*    from the default value.
*
*    If a CFP_STRING is requested from a constant, this just returns the
*    value.
*
* INVOKED
*    history = cfp_getcur(id, name, addr, type , size);
*
* PARAMETERS IN
*    FILE_ID id - par file to operate on
*    char *name - name of variable to extract
*    int type - the type of the variable
*               CFP_BOOLEAN - convert string to TRUE/FALSE, size must be 4
*               CFP_INTEGER - convert string to int, size must be 4
*               CFP_DOUBLE  - convert string to double, size must be 8
*               CFP_STRING  - just return string, size must be >strlen(current)
*    int size - the maximum length of the returned string, including NULL
*
* PARAMETERS OUT
*    void *addr - the matching string
*
* FUNCTION RETURNS
*    int - history of the identifier or FAIL
*
* GLOBALS
*    none
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jab          27 Jan 1988     Original coding
* jab          09 Feb 1989     get string if not a variable and type string
*
* $Log:	cfp_getcur.c,v $
 * Revision 2.4  96/11/19  22:57:39  isani
 * cleaned up rcsid
 * type-cast an unsigned->signed and signed->unsigned that were ok
 * 
 * Revision 2.3  1995/02/27  20:07:17  thomas
 * Updated test for range checking to test RangeInfo for NULL
 *
 * Revision 2.2  94/12/21  12:42:34  12:42:34  thomas (Jim Thomas)
 * Added "not just a lookup" flag to cfp_FindName call
 * 
 * Revision 2.1  94/12/14  12:51:36  12:51:36  thomas (Jim Thomas)
 * Added #ifdef __LINT__ around RCS headers
 * Removed stdlib.h
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h, added cfp_internal.h
 * Changed CFHT_ERROR to CFHT_LOGONLY
 * Rewrote to use version 2 line structure directly
 * 
 * Revision 1.1  91/09/24  17:11:13  17:11:13  steve (Steven Smith)
 * Original checkin
 * 
*
****************************************************************************!*/

#include <string.h>		  /* for string handling defines */
#include <errno.h>		  /* for error codes and errno define */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */
#include "cfp_internal.h"	  /*   and internals */

RCSID("$Header: cfp_getcur.c,v 2.4 96/11/19 22:57:39 isani Exp $");

HistoryType			  /* return: history or FAIL */
cfp_getcur(id, name, addr, type, size)
    FILE_ID id;			  /* instance handle */
    char *name;			  /* identifier to look for */
    void *addr;			  /* return: place to put the string */
    int type;			  /* addrs actual type */
    int size;			  /* max size allowable including NULL */
{
    FileEntryType *FileEntry;	  /* pointer to file data */
    LineType *Line;		  /* pointer to a line in this file */
    char *Value;		  /* pointer to value to return */

    if ((name == (char *) NULL) || (addr == (void *) NULL)) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getcur:%d) null parameter passed in", __LINE__);
        return(FAIL);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_getcur:%d) start on '%s'",
	      __LINE__, name);

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getcur:%d) CheckID failed on %d for '%s'",
		  __LINE__, id, name);
        return(FAIL);
    }

    if ((Line = cfp_FindName(FileEntry, name, FALSE)) == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getcur:%d) name '%s' not found for '%s'",
		  __LINE__, name, FileEntry->FilePath);
	return(FAIL);
    }

    Value = Line->CurrentValue;

    /************************************************************
     * if not a variable and string is requested handle special *
     ************************************************************/

    if ((Line->TypeCode == ConstantLine) && (type == CFP_STRING)) {
	if (Line->OriginalValue == (char *) NULL) {
	    strcpy(addr, "");
	} else {
	    if ((unsigned)size > Line->OriginalValueLength) {
		strcpy((char *) addr, Line->OriginalValue);
	    } else {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_getcur:%d) size %d for const %s < length %d",
			  __LINE__, size, Line->NameString,
			  Line->OriginalValueLength);
		return(FAIL);
	    }
	}
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_getcur:%d) end (const, value='%s')",
		  __LINE__, (char *) addr);

	return(CFP_CURRENT);	  /* NOTE NOTE NOTE NOTE :-( */
    }

    else if (Line->TypeCode != VariableLine) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getcur:%d) trying to get non-variable '%s' for '%s'",
		  __LINE__, Line->NameString, FileEntry->FilePath);
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(...)           type code is %d, data type is %d",
		  Line->TypeCode, type);
	return(FAIL);
    }

    /**********************************
     * check for environment override *
     **********************************/

    if (Line->History == CFP_UNCHECKED) {
        if ((Value = cfp_GetEnvValue(FileEntry, FileEntry->FileName,
				     (int)strlen(FileEntry->FileName)))
	    != (char *) NULL) {
            if (cfp_CopyToCurrent(Line, Value) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_getcur:%d) update from env failed for '%s'",
			  __LINE__, FileEntry->FilePath);
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(...)           trying '%s=%s'",
			  Line->NameString, Value);
		return(FAIL);
	    }
            Line->History = CFP_ENVIRONMENT;
        }
    }

    /******************************
     * check for default override *
     ******************************/

    if (Line->History == CFP_UNCHECKED) {
        if (Line->UseDefault == TRUE) {
            if (cfp_GetItem(Line, CFP_DefaultIndex,
			    &Value, (unsigned int *) NULL)
		!= PASS) {
                cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_getcur:%d) get default for '%s' failed on '%s'",
			  __LINE__, Line->NameString, FileEntry->FilePath);
                return(FAIL);
            }
            if (cfp_CopyToCurrent(Line, Value) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_getcur:%d) update from default failed for '%s'",
			  __LINE__, FileEntry->FilePath);
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(...)           trying '%s=%s'",
			  Line->NameString, Value);
		return(FAIL);
	    }
            Line->History = CFP_DEFAULT;
        }
    }

    if (Line->History == CFP_UNCHECKED) {
	Value = Line->CurrentValue;
        Line->History = CFP_CURRENT;
    }

    /*
     * if there is no value just use a null string
     * this loses the distinction between no value and ""
     *   but there did not used to be any difference anyway......
     */
    if (Value == (char *) NULL)
      Value = "";
    
    if (cfp_ConvertFromString(type, Value, addr, size) != PASS) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getcur:%d) convert failed for '%s' (type=%d) in '%s'",
		  __LINE__, Line->NameString, type, FileEntry->FilePath);
        return(FAIL);
    }

    if (Line->RangeInfo != (RangeInfoType *) NULL) {
	switch (type) {
	  case CFP_INTEGER:
	  case CFP_DOUBLE:
	  case CFP_STRING:
	    if (cfp_CheckRange(Line, type, Value) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_getcur:%d) '%s' bad or out of range for '%s'",
			  __LINE__, Value, Line->NameString);
		return(FAIL);
	    }
	    break;

	  default:
	    /* don't know what to do with strings, booleans don't care */
	    break;
	}
    }

    switch (type) {
      case CFP_INTEGER:
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_getcur:%d) end (var, current=%d)",
		  __LINE__, *((int *) addr));
	break;
      case CFP_DOUBLE:
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_getcur:%d) end (var, current=%g)",
		  __LINE__, *((double *) addr));
	break;
      case CFP_BOOLEAN:
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_getcur:%d) end (var, current=%s)",
		  __LINE__, (*((int *) addr) == TRUE) ? "True" : "False");
	break;
      case CFP_STRING:
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_getcur:%d) end (var, current='%s')",
		  __LINE__, (char *) addr);
	break;
      default:
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_getcur:%d) end, but unknown type???",
		  __LINE__);
    }

    return(Line->History);
}
