/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */

/*!**********************************************************************
 * FILE
 * cfp_getinvokes.c
 * $Header: cfp_getinvokes.c,v 2.3 96/11/19 22:58:27 isani Exp $
 * $Locker:  $
 *
 * ROUTINES
 *
 * cfp_getinvokes
 *
 * DESCRIPTION
 *    return all the invokes lines from a file
 *
 *    This routine looks up the invokes line in a file and then finds all
 *    invoke lines described.  The invokes line looks like:
 *         invokes = +invokea +invokeb +invokec
 *    So the invoke lines would be, for example:
 *         +invokea = programa -a -b
 *         +invokeb = programb -1
 *         +invokec = programc -geometry 80x25-0-0
 *    The three above lines would be concatenated into one string with
 *    newlines as separators (currently in a static internal buffer).  A
 *    pointer to the combined string is returned.
 *
 * INVOKED
 *    status = cfp_getinvokes(id, &list, &count);
 *
 * PARAMETERS IN
 *    FILE_ID id - file to look in
 *
 * PARAMETERS OUT
 *    char **list - address for string of invoke lines, newline terminated
 *    int *count - number of invokes found
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if invokes lines found OK, FAIL if not
 *
 * GLOBALS
 *    none
 *
 * $Log:	cfp_getinvokes.c,v $
 * Revision 2.3  96/11/19  22:58:27  isani
 * cleaned up rcsid
 * typecast a couple of unsigned's -> signed that were ok
 * 
 * Revision 2.2  1994/12/21  12:42:48  thomas
 * Added "not just a lookup" flag to cfp_FindName call
 *
 * Revision 2.1  94/12/14  12:58:00  12:58:00  thomas (Jim Thomas)
 * Added #ifdef __LINT__ around RCS headers
 * Removed sys/types.h, changed to stdlib.h on HP and malloc.h on Sun
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h, added cfp_internal.h
 * Changed CFHT_ERROR to CFHT_LOGONLY
 * Rewrote to use version 2 routines and malloc result buffer
 * 
 * Revision 1.2  94/01/20  02:12:29  jwright
 * remove unused variables
 *
 * Revision 1.1  92/05/22  15:07:10  steve
 * Initial revision
 *
 **********************************************************************!*/

#include <string.h>		  /* for string handling defines */
#include <errno.h>		  /* for error codes and errno define */
#include <stdlib.h>		  /* for malloc() */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */
#include "cfp_internal.h"	  /*   and internals */

RCSID("$Header: cfp_getinvokes.c,v 2.3 96/11/19 22:58:27 isani Exp $");

PASSFAIL
cfp_getinvokes(id, list, count)
    FILE_ID id;			  /* instance handle */
    char **list;		  /* return: pointer to array of invokes */
    int *count;			  /* number of invokes found */
{
#define EXTRA 160  /* 2 extra lines at realloc time (might save on reallocs) */

    FileEntryType *FileEntry;	  /* pointer to file data */
    LineType *Line;		  /* pointer to one variable */
    char *p;			  /* temporary string pointer */
    int tcount;			  /* total line counts */
    static char *tmpBuf;	  /* temporary buffer */
    char Invokes[CFHT_VALUE_SIZE + 1]; /* invokes line */
    char *Value;		  /* the line from one invoke */
    size_t Length;		  /* length of an individual line */
    static int first = 1;	  /* first time through init flag */
    static long bytes = EXTRA;	  /* counter of bytes malloc'd so far */

    if ((list == (char **) NULL) || (count == (int *) NULL)) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getinvokes:%d) given null pointer", __LINE__);
	return(FAIL);
    }

    if (first) {		  /* start a buffer if first time */
	if ((tmpBuf = (char *) malloc((size_t) bytes)) == NULL) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_getinvokes:%d) malloc %d failed", __LINE__, bytes);
	    return (FAIL);
	}
	first = 0;
    }

    cfht_log(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_getinvokes) start");

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getinvokes:%d) cfp_CheckID failed on %d",
		  __LINE__, id);
        return(FAIL);
    }

    tmpBuf[0] = '\0';
    tcount = 0;

/***************************
 * retreive invokes line *
 ***************************/

    if (cfp_getstr(id, CFP_INVOKES, Invokes, sizeof Invokes) != PASS) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getinvokes:%d) no invokes in '%s'",
		  __LINE__, FileEntry->FilePath);
	return (FAIL);
    }

/*************************************
 * extract invoke definition lines *
 *************************************/

    p = strtok (Invokes, " \t,");
    while (p != NULL) {

	if ((Line = cfp_FindName(FileEntry, p, FALSE)) == (LineType *) NULL) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY, 
                    "(cfp_getinvokes:%d) warning '%s' not found in '%s'",
		      __LINE__, p, FileEntry->FilePath);
	}
	else {

	    if (cfp_RebuildLine(Line, &Value, &Length) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_getinvokes:%d) RebuildLine failed for '%s' in '%s'",
			  __LINE__, Line->NameString, FileEntry->FilePath);
	    } else {
		
		/*
		 * Concatenate the line to tmpBuf;
		 * Dynamically keep a buffer with at least a couple extra
		 * lines available.
		 */
		
		if (strlen(tmpBuf) + Length >= (unsigned)bytes) {
		    bytes = bytes + Length + EXTRA;  
		    if ((tmpBuf = (char *) realloc(tmpBuf, (unsigned)bytes + 1))
			== NULL) {
			cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
				  "(cfp_getinvokes:%d): realloc %d failed for '%s'",
				  __LINE__, bytes + 1, FileEntry->FilePath);
			return(FAIL);
		    }
		}
		strcat(tmpBuf, Value);
		tcount++;
	    }
	}
        p = strtok (NULL, " \t,");
    }

    *list = tmpBuf;
    *count = tcount;

    cfht_log(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_getinvokes) end");

    return (PASS);
}
