/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* cfp_getstr.c
* $Header: cfp_getstr.c,v 2.3 96/11/19 22:59:16 isani Exp $
* $Locker:  $
*
* ROUTINES
*
*   cfp_getstr()
*
* DESCRIPTION
*    This routine retrieves a string value matching the given identifier
*    from the named file.
*
* INVOKED
*    result = cfp_getstr(id, name, addr, size);
*
* PARAMETERS IN
*    FILE_ID id - the file to examine
*    char *name - the name of the identifier to match
*    int size - number of characters which can be put in addr, including NULL
*
* PARAMETERS OUT
*    char *addr - the string which matches the given identifier
*
* FUNCTION RETURNS
*    PASSFAIL - PASS if string extracted successfully, FAIL if not
*
* GLOBALS
*    none
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jab          27 Jan 1988     Original coding
*
* $Log:	cfp_getstr.c,v $
 * Revision 2.3  96/11/19  22:59:16  isani
 * cleaned up rcsid
 * typecast a signed -> unsigned that was ok
 * 
 * Revision 2.2  1994/12/21  12:43:11  thomas
 * Added "not just a lookup" flag to cfp_FindName call
 *
 * Revision 2.1  94/12/14  13:06:45  13:06:45  thomas (Jim Thomas)
 * Added #ifdef __LINT__ around RCS headers
 * Removed sys/types.h
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h, added cfp_internal.h
 * Changed CFHT_ERROR to CFHT_LOGONLY
 * Rewrote to use version 2 routines
 * 
 * Revision 1.1  91/09/24  17:14:46  17:14:46  steve (Steven Smith)
 * Original checkin
 * 
****************************************************************************!*/

#include <string.h>		  /* for string handling defines */
#include <errno.h>		  /* for error codes and errno define */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */
#include "cfp_internal.h"	  /*   and internals */

RCSID("$Header: cfp_getstr.c,v 2.3 96/11/19 22:59:16 isani Exp $");

PASSFAIL			  /* return: PASS, or FAIL */
cfp_getstr(id, name, addr, size)
    FILE_ID id;			  /* instance handle */
    char *name;			  /* identifier to look for */
    char *addr;			  /* return: place to put the string */
    int size;			  /* max size allowable with NULL */
{
    FileEntryType *FileEntry;	  /* pointer to data on this file */
    LineType *Line;		  /* pointer to line for this name */
    char *Value;		  /* pointer to value that is found */
    unsigned int Length;	  /* length of the above value */
    char ValueCopy[CFP_STRSIZE + 1]; /* for return value print */

    if ((name == (char *) NULL) || (addr == (char *) NULL)) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getstr:%d) given null pointer", __LINE__);
	return(FAIL);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
              "(cfp_getstr) start (name = '%s')", name);

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getstr:%d) cfp_CheckID failed on %d for '%s'",
		  __LINE__, id, name);
        return(FAIL);
    }

    if ((Line = cfp_FindName(FileEntry, name, FALSE)) == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getstr:%d) name '%s' not found for '%s'",
		  __LINE__, name, FileEntry->FilePath);
	return(FAIL);
    }

    if (Line->TypeCode == VariableLine) {
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_getstr) requesting variable string of '%s' for '%s'",
		  Line->NameString, FileEntry->FilePath);
    } else if (Line->TypeCode != ConstantLine) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getstr:%d) line type (%d) not constant or variable",
		  __LINE__, Line->TypeCode);
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(...)           for '%s' for '%s'",
		  Line->NameString, FileEntry->FilePath);
	return(FAIL);
    }

    if (Line->TypeCode == ConstantLine) {
	Value = Line->OriginalValue;
	Length = Line->OriginalValueLength;
    } else {
	Value = Line->CurrentValue;
	Length = Line->CurrentValueLength;
    }

    /* make sure we don't pass back junk :-) */
    if (Value == (char *) NULL) {
	Value = "";
	Length = 0;
    }

    if (Length >= (unsigned)size) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_getstr:%d) string '%s' of '%s' too big in '%s'",
		  __LINE__, Value, Line->NameString, FileEntry->FilePath);
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(...)     string length %d > provided size %d",
		  Length + 1, size);
        errno = E2BIG;
        return(FAIL);
    }

    /* pass back string */
    strcpy(addr, Value);

    if (Length > CFP_STRSIZE) {
	strncpy(ValueCopy, Value, CFP_STRSIZE);
	ValueCopy[CFP_STRSIZE] = '\0';
	Value = ValueCopy;
    }
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_getstr) end (returning %s '%s')",
	      Line->TypeCode == ConstantLine ? "constant" : "variable",
	      Value);

    return(PASS);
}
