/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* cfp_getval.c
* $Header: cfp_getval.c,v 2.4 96/11/19 22:59:45 isani Exp $
* $Locker:  $
*
* ROUTINES
*
*   cfp_getval()
*
* DESCRIPTION
*    This routine will grab an individual token out of a string in a file,
*    or will get a string from the environment.
*
* INVOKED
*    result = cfp_getval(id, name, addr, type, size, valnum);
*
* PARAMETERS IN
*    FILE_ID id - the file to operate on
*    char *name - the identifier to look for
*    int type - what type of value are we getting
*    int size - maximum number of characters available in addr, including NULL
*    ItemIndexType valnum - which field of par file to return:
*                 CFP_ModeIndex for the options field,
*                 CFP_CurrentIndex for current,
*                 CFP_DefaultIndex for default,
*                 CFP_SuppliedIndex for supplied,
*                 CFP_MinIndex for minimum
*                 CFP_MaxIndes for maximum
*                 CFP_EnvironmentIndex for environment
*
* PARAMETERS OUT
*    char *addr - place to return string
*
* FUNCTION RETURNS
*    PASSFAIL - PASS if OK, FAIL if something wrong
*
* GLOBALS
*    none
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jab          17 Feb 1988     Original coding
*
* $Log:	cfp_getval.c,v $
 * Revision 2.4  96/11/19  22:59:45  isani
 * cleaned up rcsid
 * typecast a size_t to an int
 * 
 * Revision 2.3  1995/02/27  20:07:51  thomas
 * Updated test for range checking to test RangeInfo for NULL
 *
 * Revision 2.2  94/12/21  12:43:21  12:43:21  thomas (Jim Thomas)
 * Added "not just a lookup" flag to cfp_FindName call
 * 
 * Revision 2.1  94/12/14  13:08:57  13:08:57  thomas (Jim Thomas)
 * Added #ifdef __LINT__ around RCS headers
 * Removed sys/stdlib.h
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h, added cfp_internal.h
 * Changed CFHT_ERROR to CFHT_LOGONLY
 * Rewrote to use version 2 routines
 * Added call for range checking
 * 
 * Revision 1.1  91/09/24  17:15:18  17:15:18  steve (Steven Smith)
 * Original checkin
 * 
*
****************************************************************************!*/

#include <string.h>		  /* for string handling defines */
#include <errno.h>		  /* for error codes and errno define */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */
#include "cfp_internal.h"	  /*   and internals */

RCSID("$Header: cfp_getval.c,v 2.4 96/11/19 22:59:45 isani Exp $");

PASSFAIL			  /* return: PASS or FAIL */
cfp_getval(id, name, addr, type, size, valnum)
    FILE_ID id;			  /* instance handle */
    char *name;			  /* identifier to look for */
    void *addr;			  /* return: place to put the string */
    int type;			  /* addrs actual type */
    int size;			  /* max size allowable with NULL */
    ItemIndexType valnum;	  /* which val, -1 for env */
{
    FileEntryType *FileEntry;	  /* pointer to data on this file */
    LineType *Line;		  /* pointer to line for this name */
    char *Value;		  /* pointer to value that is found */
    char ValueCopy[CFP_STRSIZE + 1]; /* copy for return printout */

    if ((name == (char *) NULL) || (addr == (void *) NULL)) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getval:%d) got null pointer", __LINE__);
	return(FAIL);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_getval) start on %s", name);

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getval:%d) cfp_CheckID failed on %d for '%s'",
		  __LINE__, id, name);
        return(FAIL);
    }

    if ((Line = cfp_FindName(FileEntry, name, FALSE)) == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getval:%d) name '%s' not found for '%s'",
		  __LINE__, name, FileEntry->FilePath);
	return(FAIL);
    }

    if (Line->TypeCode != VariableLine) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getval:%d) '%s' not a variable (type=%d) for '%s'",
		  __LINE__, Line->NameString, Line->TypeCode,
		  FileEntry->FilePath);
	return(FAIL);
    }

    switch (valnum) {
      case CFP_EnvironmentIndex:
        Value = cfp_GetEnvValue(FileEntry, Line->NameString,
				(int)strlen(Line->NameString));
	break;

      case CFP_ModeIndex:
      case CFP_CurrentIndex:
      case CFP_DefaultIndex:
      case CFP_SuppliedIndex:
      case CFP_RangeIndex:
	if ((cfp_GetItem(Line, valnum, &Value, (unsigned int *) NULL))
	    != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_getval:%d) GetItem %d for '%s' failed for '%s'",
		      __LINE__, valnum, Line->NameString, FileEntry->FilePath);
	};
	break;

      default:
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getval:%d) unknown field index %d for '%s' in '%s'",
		  __LINE__, valnum, Line->NameString, FileEntry->FilePath);
	return(FAIL);
	/*NOTREACHED*/
	break;
    }
    if (Value == (char *) NULL)
        Value = "";
    
    if (cfp_ConvertFromString(type, Value, addr, size) != PASS) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getval:%d) convert failed for '%s' (type=%d) in '%s'",
		  __LINE__, Line->NameString, type, FileEntry->FilePath);
        return(FAIL);
    }

    if ((valnum == CFP_CurrentIndex)
	&& (Line->RangeInfo != (RangeInfoType *) NULL)) {
	switch (type) {
	  case CFP_INTEGER:
	  case CFP_DOUBLE:
	  case CFP_STRING:
	    if (cfp_CheckRange(Line, type, Value) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_getval:%d) '%s' bad or out of range for '%s'",
			  __LINE__, Value, Line->NameString);
		return(FAIL);
	    }
	    break;

	  default:
	    /* don't know what to do with strings, booleans don't care */
	    break;
	}
    }

    if ( strlen(Value) > CFP_STRSIZE) {
	strncpy(ValueCopy, Value, CFP_STRSIZE);
	ValueCopy[CFP_STRSIZE] = '\0';
	Value = ValueCopy;
    }
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_getval) end (field = '%s')",
	      Value);

    return(PASS);
}
