/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!**************************************************************************
 *
 * $Header: cfp_getvars.c,v 2.3 96/11/19 23:00:14 isani Exp $
 * $Locker:  $
 *
 * cfp_getvars
 *    retrieves: string consisting of all the variables in a par
 *    file, or NULL
 *    NOTE - the result string is reused on each call
 *
 * INVOKED
 *    status = cfp_getvars(id, &string, &count);
 *
 * PARAMETERS IN
 *    FILE_ID id - id of open par file
 *
 * PARAMETERS OUT
 *    char **string - address for pointer to resulting string
 *    int *count - length of returned string
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if string is OK, FAIL if not
 *
 * GLOBALS
 *    none
 * 
 * $Log:	cfp_getvars.c,v $
 * Revision 2.3  96/11/19  23:00:14  isani
 * cleaned up rcsid
 * typecase a couple of signed's to unsigned that were ok
 * 
 * Revision 2.2  1994/12/21  12:43:31  thomas
 * Added "not just a lookup" flag to cfp_FindName call
 *
 * Revision 2.1  94/12/14  16:03:33  16:03:33  thomas (Jim Thomas)
 * Added #ifdef __LINT__ around RCS headers
 * Removed sys/types.h, sun floatingpoint.h, moved malloc.h to sun
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h, added cfp_internal.h
 * Changed CFHT_ERROR to CFHT_LOGONLY
 * Rewrote to use version 2 routines
 * Added malloc for result (like getinvokes)
 * 
 * Revision 1.7  94/01/20  02:14:04  jwright
 * remove unused variables
 *
 * Revision 1.6  93/02/09  11:26:44  evans
 * added fix to malloc of tmpBuf
 *
 * Revision 1.5  92/05/15  14:49:59  john
 * fixed malloc bug.
 *
 * Revision 1.4  92/05/14  15:34:04  john
 * new malloc version
 *
 * Revision 1.3  92/05/14  09:20:55  john
 * new version that uses cfp calls instead of parsing on its own.
 *
 * Revision 1.2  92/05/12  10:49:29  john
 * working version before change in "techniques".  This version
 * parsed for variables char by char.  Next version will use
 * cfp calls only.
 *
 * Revision 1.1  92/04/16  20:41:24  john
 * Initial revision
 *
 ************************************************************************!*/

#include <string.h>		  /* for string handling defines */
#include <errno.h>		  /* for error codes and errno define */
#include <stdlib.h>		  /* for malloc() */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */
#include "cfp_internal.h"	  /*   and internals */

RCSID("$Header: cfp_getvars.c,v 2.3 96/11/19 23:00:14 isani Exp $");

PASSFAIL
cfp_getvars(id, list, count)
    FILE_ID id;			  /* instance handle */
    char **list;		  /* return: pointer to array of vars */
    int *count;			  /* return: length of list */
{
#define EXTRA 160  /* 2 extra lines at realloc time (might save on reallocs) */

    FileEntryType *FileEntry;	  /* pointer to file data */
    LineType *Line;		  /* pointer to one variable */
    char *p;			  /* temporary string pointer */
    int tcount;			  /* total line counts */
    static char *tmpBuf;	  /* temporary buffer */
    char vars[CFHT_VALUE_SIZE];	  /* variables line */
    char *Value;		  /* the line from one variable */
    size_t Length;		  /* length of an individual line */
    static int first = 1;	  /* first time through init flag */
    static long bytes = EXTRA;	  /* counter of bytes malloc'd so far */

    if ((list == (char **) NULL) || (count == (int *) NULL)) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getvars:%d) given null pointer", __LINE__);
	return(FAIL);
    }

    if (first) {		  /* start a buffer if first time */
	if ((tmpBuf = (char *) malloc((size_t) bytes)) == NULL) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_getvars:%d) malloc %d failed", __LINE__, bytes);
	    return (FAIL);
	}
	first = 0;
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_getvars) start");

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getvars:%d) cfp_CheckID failed on %d", __LINE__, id);
        return(FAIL);
    }

    tmpBuf[0] = '\0';
    tcount = 0;

/***************************
 * retreive variables line *
 ***************************/

    if (cfp_getstr(id, CFP_VARIABLES, vars, sizeof vars) != PASS) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getvars:%d) no variables in '%s'",
		  __LINE__, FileEntry->FilePath);
	return (FAIL);
    }

/*************************************
 * extract variable definition lines *
 *************************************/

    p = strtok (vars, " \t,");
    while (p != NULL) {

	if ((Line = cfp_FindName(FileEntry, p, FALSE)) == (LineType *) NULL) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY, 
                    "(cfp_getvars:%d) warning %s not found", __LINE__, p);
	}
	else {

	    if (cfp_RebuildLine(Line, &Value, &Length) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_getvars:%d) RebuildLine failed for '%s' in '%s'",
			  __LINE__, Line->NameString, FileEntry->FilePath);
	    } else {
		
		/*
		 * Concatenate the line to tmpBuf;
		 * Dynamically keep a buffer with at least a couple extra
		 * lines available.
		 */
		
		if (strlen(tmpBuf) + Length >= (unsigned)bytes) {
		    bytes = bytes + Length + EXTRA;  
		    if ((tmpBuf = (char *) realloc(tmpBuf, (unsigned)bytes + 1))
			== NULL) {
			cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
				  "(cfp_getvars:%d): realloc %d failed for '%s'",
				  __LINE__, bytes + 1, FileEntry->FilePath);
			return(FAIL);
		    }
		}
		strcat(tmpBuf, Value);
		tcount++;
	    }
	}
        p = strtok (NULL, " \t,");
    }

    *list = tmpBuf;
    *count = tcount;

    cfht_log(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_getvars) end");

    return (PASS);
}

