/* Copyright (C) 1994    Canada-France-Hawaii Telescope Corp.            */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* cfp_init.c
* $Header: cfp_init.c,v 2.4 96/11/19 23:01:08 isani Exp $
* $Locker:  $
*
* ROUTINES
*
*   cfp_init()
*
* DESCRIPTION
*    Check for a valid file, then read the file into memory for further
*    manipulation by other routines.  The file is opened and closed in
*    this routine; it is not left hanging open.  The raw data is
*    parsed out into individual lines (via another routine).
*
* INVOKED
*    id = cfp_init(name);
*
* PARAMETERS IN
*    char *name - file name
*
* PARAMETERS OUT
*    NONE
*
* FUNCTION RETURNS
*    FILE_ID - the index into an array of buffers to find this file's
*              data; effectively this identifier can be thought of as
*              referring to the file itself
*            - on failure this returns FAIL (which had better be negative)
*
* GLOBALS
*    FileEntryType FileEntryTable - array of open file entries
*    int errno - global error indicator
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jab          27 Jan 1988     Original coding
* jab          13 JAN 1989     Fix nasty user file name mod bug
*
* $Log:	cfp_init.c,v $
 * Revision 2.4  96/11/19  23:01:08  isani
 * cleaned up rcsid
 * moved sun hacks to "cfht.h"
 * typecast read() result to size_t for comparison with another size_t
 * 
 * Revision 2.3  1995/01/19  09:32:34  thomas
 * Change start message back to DEBUG, generated too many lines
 *    (didn't actually change end, contrary to previous log entry)
 *
 * Revision 2.2  95/01/17  12:17:10  12:17:10  thomas (Jim Thomas)
 * Changed start and stop messages to include file name and be LOGONLY to help
 * with debugging file update race conditions
 * 
 * Revision 2.1  94/12/14  16:08:05  16:08:05  thomas (Jim Thomas)
 * Added #ifdef __LINT__ around RCS headers
 * Removed sys/types.h
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h, added cfp_internal.h
 * Changed CFHT_ERROR to CFHT_LOGONLY
 * Rewrote to use version 2 routines
 * Added more error checking of arguments
 * 
 * Revision 1.1  91/09/24  17:16:50  17:16:50  steve (Steven Smith)
 * Original checkin
 * 
****************************************************************************!*/

#include <string.h>		  /* for string defines */
#include <unistd.h>		  /* read() */
#include <fcntl.h>		  /* for O_RDWR */
#include <sys/stat.h>		  /* for file size and modification time */
#include <errno.h>		  /* for error codes and errno defines */
#include <stdlib.h>		  /* for malloc */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */
#include "cfp_internal.h"	  /* cfp private defines */

RCSID("$Header: cfp_init.c,v 2.4 96/11/19 23:01:08 isani Exp $");

FILE_ID				  /* return: instance handle, or FAIL */
cfp_init(name)
    char *name;			  /* file name */
{
    FileEntryType *FileEntry;	  /* description of current file */
    int fd;			  /* file descriptor for file */
    struct stat statbuf;	  /* for stat(2)ing file */
    FILE_ID inst;		  /* index into open file table */
    size_t FileSize;		  /* file size claimed by stat */
    char *FileData;		  /* pointer to a copy of the file */
    char *p;
    int size;

    /************************
     * verify name is valid *
     ************************/

    if (name) {
        if ((size = strlen(name)) > CFP_PATHSIZE) {
            cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                      "(cfp_init:%d) name '%s' too big", __LINE__, name);
            errno = E2BIG;
            return(FAIL);
        }
	else {
	    if (size == 0) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_init:%d) null name", __LINE__);
		errno = ENOTTY;
		return(FAIL);
	    }
	}
    } else {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_init:%d) no name passed in", __LINE__);
        errno = ENODEV;
        return(FAIL);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_init) start, name = '%s'", name);

    /**************************
     * find an empty instance *
     **************************/

    for (inst = 0; inst <= CFP_MAXINST; inst++)
        if (FileEntryTable[inst].InUseCode == NotInUse)
            break;

    if (inst > CFP_MAXINST) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_init:%d) no room in table for '%s'", __LINE__, name);
        errno = ENOSPC;
        return(FAIL);
    }

    FileEntry = &FileEntryTable[inst];
    FileEntry->FileID = inst;

    strcpy(FileEntry->FilePath, name);
    if ((p = strrchr(name, '/')))
        p++;
    else
        p = name;
    p += strspn(p, ".,");
    strcpy(FileEntry->FileName, p);
    size = strcspn(p, ".");
    FileEntry->FileName[size] = '\0';

    /********************
     * read in the file *
     ********************/

    /* get data space */
    if (stat(name, &statbuf) < 0) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_init:%d) stat failed on '%s' (%s)",
		  __LINE__, name, strerror(errno));
        return(FAIL);
    }
    FileSize = (size_t) statbuf.st_size;
    if (FileSize == 0) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY, "(cfp_init:%d) '%s' zero size",
		  __LINE__, name);
	errno = ENOTTY;
	return(FAIL);
    }
    FileEntry->LastModified = statbuf.st_mtime;
    if (!(FileData = (char *) malloc((size_t) (FileSize + 1)))) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_init:%d) malloc of %d bytes for '%s' failed",
                  __LINE__, FileSize + 1, name);
        return(FAIL);
    }

    /* get data */
    if ((fd = open(name, O_RDONLY)) < 0) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_init:%d) open failed for '%s' (%s)",
		  __LINE__, name, strerror(errno));
        free((void *) FileData);
        return(FAIL);
    }
    if ((size_t)read(fd, (void *) FileData, FileSize) != FileSize) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_init:%d) read of '%s' failed (%s)",
		  __LINE__, name, strerror(errno));
        close(fd);
        free((void *) FileData);
        return(FAIL);
    }
    close(fd);

    /* null terminate */
    FileData[FileSize] = '\0';

    /* verify proper termination */
    if (FileData[FileSize - 1] != '\n') {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_init:%d) '%s' corrupt, no newline at end",
		  __LINE__, name);
        free((void *) FileData);
        errno = ENOTTY;
        return(FAIL);
    }

    /* build up the line linked list */
    FileEntry->LineCount = 0;
    FileEntry->Lines = (LineType *) NULL;
    if (cfp_InitLines(FileEntry, FileData) != PASS) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_init:%d) InitLines failed", __LINE__);
	free((void *) FileData);
	return(FAIL);
    }

    /* now we are sure all is OK, say it is in use */
    FileEntry->InUseCode = InUse;
    free((void *) FileData);

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_init) end, name = '%s'", name);

    return(inst);
}
