/* Copyright (C) 1994-1998    Canada-France-Hawaii Telescope Corp.       */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING         */

/***************************************************************************
 *
 * file: cfp_internal.h
 * $Header: cfp_internal.h,v 2.6 2000/02/10 13:13:46 isani Exp $
 * $Locker:  $
 *
 * This file is an include file for the generic library.
 *
 * In particular it contains the internal defines for the configuration file
 * routines.
 *
 * (cfp stands for Configuration File Parsing.)
 *
 *
 * HISTORY
 *
 * $Log:	cfp_internal.h,v $
 * Revision 2.6  2000/02/10  13:13:46  isani
 * Changed size_t to unsigned int.
 * 
 * Revision 2.5  1998/07/02 19:19:23  thomas
 * Removed IsAlias and IsVariable bits
 *
 * Revision 2.4  96/01/25  17:25:40  17:25:40  thomas (Jim Thomas)
 * Changed structure definitions for new symbol hash table
 * 
 * Revision 2.3  95/02/27  20:10:04  20:10:04  thomas (Jim Thomas)
 * Added Line struct field definitions for parsed range specifiers
 * Took out m and i mode specifiers
 * Added prototypes for cfp_EditValue and cfp_BuildRangeMessage
 * 
 * Revision 2.2  94/12/21  09:45:25  09:45:25  thomas (Jim Thomas)
 * Add LookupOnly to cfp_FindName
 * 
 * Revision 2.1  94/12/14  16:39:47  16:39:47  thomas (Jim Thomas)
 * Original check in
 *
 ***************************************************************************/

#ifndef CFPINTERNALDOTH
#define CFPINTERNALDOTH

#include <time.h>		  /* for file modification time */

#include "cfht.h"		  /* just for protection and P macro */
#include "cfp.h"		  /* ditto */

#define AssignmentCharacter '='	  /* indicates variable or constant def */
#define AliasCharacter1 '-'	  /* indicates first of alias digraph */
#define AliasCharacter2 '>'	  /* indicates second of alias digraph */
#define AliasIndirectCharacter '%' /* indicates alias through variable value */
#define VariableDefStartCharacter '{' /* indicates start of variable def */
#define VariableDefEndCharacter '}' /* indicates end of variable def */
#define QuoteCharacter '\"'	  /* indicates start/end of quoted string */
#define EscapeCharacter '\\'	  /* indicates next character has problems */
#define CommentCharacter '#'	  /* indicates rest of line is a comment */
#define LeftOpenCharacter '('	  /* left end of range is open */
#define LeftClosedCharacter '['	  /* left end of range is closed */
#define RightOpenCharacter ')'	  /* right end of range is open */
#define RightClosedCharacter ']'  /* right end of range is closed */

/*
 * The BOOLEAN type is currently just an int .  Because we are going to use
 * one bit fields, we need to have an unsigned int (one cannot use enum's
 * because they are signed too).  We did not want to chance changing the
 * system BOOLEAN definition at this stage.  The potential values are still
 * TRUE and FALSE .
 */

typedef unsigned int BooleanType;


/*
 * When a range field is encountered in the par file for a variable, one of
 * the following structures is build containing the parsed information
 * describing the range.  This is what is used internally for value editing
 * and range checking.  If the file is output again, the string is copied back
 * into the variable definition; it is NOT reconstructed from this structure.
 *
 * There are three possibilites for each of the minimum and maximum values.
 * The pointer cases and flag values are indicated in the following table.
 *
 *  {Min|Max}String                  {Min|Max}Seen    {Min|Max}Indirect
 *   1)  an actual value (e.g., 100)      TRUE              FALSE
 *   2)  an indirect (e.g., foo)         FALSE               TRUE
 *   3)  missing                         FALSE              FALSE
 *
 * +----------------------------------------------------------------------+
 * | Variable type (one of CFP_INTEGER, etc.)                             |
 * +----------------------------------------------------------------------+
 * | Pointer to a character string describing the minimum value           |
 * |   this can be one of:                                                |
 * |      1) a copy of the minimum value from the range string            |
 * |      2) the name of an indirect reference (no %, i.e., foo not %foo) |
 * |      3) "-inf" if no explicit minimum was given                      |
 * +----------------------------------------------------------------------+
 * | TRUE if the minimum is an indirect reference (%foo), FALSE if not    |
 * +----------------------------------------------------------------------+
 * | TRUE if the minimum is an actual, specified value, FALSE if not      |
 * +----------------------------------------------------------------------+
 * | Pointer to a character string describing the maximum value           |
 * |   this can be one of:                                                |
 * |      1) a copy of the maximum value from the range string            |
 * |      2) the name of an indirect reference (no %, i.e., foo not %foo) |
 * |      3) "+inf" if no explicit maximum was given                      |
 * +----------------------------------------------------------------------+
 * | TRUE if the maximum is an indirect reference (%foo), FALSE if not    |
 * +----------------------------------------------------------------------+
 * | TRUE if the maximum is an actual, specified value, FALSE if not      |
 * +----------------------------------------------------------------------+
 * | "Compiled" range structure filled in with open/closed flags and      |
 * | integer or double values (except for indirect references)            |
 * +----------------------------------------------------------------------+
 */

typedef struct {
    int Type;
    char *MinString;
    BooleanType MinIndirect;
    BooleanType MinSeen;
    char *MaxString;
    BooleanType MaxIndirect;
    BooleanType MaxSeen;
    union cfht_ranges Range;
} RangeInfoType;

/*
 * short names for range structure fields
 */

#define intRangeMin Range.cfht_int_range.min
#define intRangeMax Range.cfht_int_range.max
#define intRangeLeft Range.cfht_int_range.min_inclusive
#define intRangeRight Range.cfht_int_range.max_inclusive
#define doubleRangeMin Range.cfht_double_range.min
#define doubleRangeMax Range.cfht_double_range.max
#define doubleRangeLeft Range.cfht_double_range.min_inclusive
#define doubleRangeRight Range.cfht_double_range.max_inclusive

/*
 * Each line of a file is read into one of the following entries
 *
 * The lines are kept in order of appearance in the file in a singly linked
 * list with a head pointer.
 *
 * There are five kinds of lines.  The line can be blank.  The whole line can
 * be a comment (the first non-blank character on the line is a # ).  The line
 * can be an alias line with the format
 *      <name> -> <alias>
 * .  The line can be a "constant" definition with the format
 *      <name> = <value>
 * , where <value> does not begin with a { .  Constant definitions cannot be
 * changed by command line arguments.  The line can be a variable definition
 * with the format
 *      <name> = { <mode>, <current>, <default>, <specified>, <min>, <max> }
 * .
 *
 * A blank line has both Name and Comment pointers NULL and has a TypeCode of
 * BlankLine.  A comment line has the Name pointer NULL and the Comment pointer
 * non-NULL and has a TypeCode of CommentLine.  An alias line has the Name
 * pointer non-NULL and has a TypeCode of AliasLine.  A constant line has
 * the Name pointer non-NULL and has a TypeCode of ConstantLine.  A variable
 * line has the Name pointer non-NULL and has a TypeCode of VariableLine.
 *
 * In the first column of the following diagram, the following letters are used
 * to indicate that the given line type uses the field (i.e., the field
 * can contain useful information for the line type).
 *
 * 	B - blank line
 * 	# - comment line
 * 	A - alias line
 * 	C - constant line
 * 	V - variable line
 *
 * +----------------------------------------------------------------------+
 * | B#ACV | Pointer to next line in linked list of lines, NULL if none   |
 * +----------------------------------------------------------------------+
 * | B#ACV | Pointer to next symbol in linked list of lines, NULL if none |
 * +----------------------------------------------------------------------+
 * | B#ACV | Pointer to next symbol in hash linked list, NULL if none     |
 * +----------------------------------------------------------------------+
 * | B#ACV | Pointer back to FileEntry for this line                      |
 * +----------------------------------------------------------------------+
 * | B#ACV | Line type code                                               |
 * +----------------------------------------------------------------------+
 * |     V | TRUE if command line value not allowed ("n"), FALSE if not   |
 * |     V | TRUE if command line value optional ("o"), FALSE if not      |
 * |     V | TRUE if use default ("d"), FALSE if not                      |
 * |     V | TRUE if use supplied ("s"), FALSE if not                     |
 * |     V | TRUE if update value on output ("r"), FALSE if not           |
 * |    C  | TRUE if input line had "", FALSE if not (don't lose them)    |
 * +----------------------------------------------------------------------+
 * |     V | Code indicating the source of the current value              |
 * +----------------------------------------------------------------------+
 * |   ACV | Count of leading white space on the line for maintaining     |
 * |       | formatting                                                   |
 * +----------------------------------------------------------------------+
 * |   ACV | Length of the name string, not including null                |
 * +----------------------------------------------------------------------+
 * |   ACV | Pointer to a string containing the name, NULL if no name     |
 * +----------------------------------------------------------------------+
 * |   ACV | Column the - or = was in in input, for maintaining format    |
 * +----------------------------------------------------------------------+
 * |   ACV | Length of the original value string, not including null      |
 * +----------------------------------------------------------------------+
 * |   ACV | Pointer to a string containing the original value from the   |
 * |       | file - the alias for an alias, the value for a constant, and |
 * |       | the original current value for a variable - NULL if none     |
 * +----------------------------------------------------------------------+
 * |     V | Pointer to a string version of the mode bits for a           |
 * |       | variable, NULL if none                                       |
 * +----------------------------------------------------------------------+
 * |    CV | Length of the current value string, not including null       |
 * +----------------------------------------------------------------------+
 * |    CV | Pointer to a string containing the current value for a       |
 * |       | variable - the whole right hand side for a constant, the     |
 * |       | current field for a variable - NULL if none                  |
 * +----------------------------------------------------------------------+
 * |     V | Pointer to a string containing the default value for a       |
 * |       | variable, NULL if none                                       |
 * +----------------------------------------------------------------------+
 * |     V | Pointer to a string containing the supplied value for a      |
 * |       | variable, NULL if none                                       |
 * +----------------------------------------------------------------------+
 * |     V | Pointer to a string containing the range value for a         |
 * |       | variable, NULL if none                                       |
 * +----------------------------------------------------------------------+
 * |     V | Pointer to a range structure containing the parsed range     |
 * |       | string for a variable, NULL if none                          |
 * +----------------------------------------------------------------------+
 * |  #ACV | Pointer to a string containing the comment for the line,     |
 * |       | NULL if no comment                                           |
 * +----------------------------------------------------------------------+
 * |  #ACV | Target column to put the comment back into if file is        |
 * |       | written back out (put it here to to the right of this)       |
 * |       | NOTE - this number is 0-nn, i.e., it starts at zero not one  |
 * +----------------------------------------------------------------------+
 */

#define NumberOfModes 5		  /* number of different mode letters */
#define ModeOptionalCharacter 'o' /* command line value is optional */
#define ModeNeverCharacter 'n'	  /* command line value cannot be given */
#define ModeRequiredCharacter ' ' /* command line value is required */
				  /*  actually specified by neither o nor n */
#define ModeDefaultCharacter 'd'  /* command line value from default */
#define ModeSuppliedCharacter 's' /* command line value from supplied */
#define ModeUpdateCharacter 'r'	  /* update this value in file if changes */

typedef enum {			  /* the codes for the line types */
    BlankLine = 0,		  /* empty or only tabs and spaces */
    CommentLine,		  /* only a comment */
    AliasLine,			  /* contains an alias (foo -> bar) */
    ConstantLine,		  /* "constant" definition (foo = bar) */
    VariableLine		  /* "variable" definition (foo = {,bar}) */
} LineTypeCodeType;

typedef struct LineEntryStruct {
    struct LineEntryStruct *Next;
    struct LineEntryStruct *NextSymbol;
    struct LineEntryStruct *NextHash;
    struct FileEntryStruct *FileEntry;
    LineTypeCodeType TypeCode;
    BooleanType ValueNotAllowed : 1;
    BooleanType ValueOptional : 1;
    BooleanType UseDefault : 1;
    BooleanType UseSupplied : 1;
    BooleanType UpdateValue : 1;
    BooleanType WasQuoted : 1;
    HistoryType History;
    unsigned int LeadingSpaceCount;
    unsigned int NameStringLength;
    char *NameString;
    unsigned int EqualsColumn;
    unsigned int OriginalValueLength;
    char *OriginalValue;
    char *ModeString;
    unsigned int CurrentValueLength;
    char *CurrentValue;
    char *DefaultValue;
    char *SuppliedValue;
    char *RangeValue;
    RangeInfoType *RangeInfo;
    char *CommentString;
    unsigned int CommentColumn;
} LineType;

/*
 * The symbols are hashed into a table for faster searching
 */

typedef LineType *HashEntryType;
typedef HashEntryType *HashTableType;

/*
 * Each "open" file is described by one of the following entries
 *   (Files are not actually kept open.  A file is opened, read in, and
 *    closed.  If the contents are changed and later written out, the actual
 *    file is reopened.)
 *
 * All the lines in the file are in one linked list, in file order.
 * In addition, all the lines that are symbol definitions (constant, variable,
 * and alias definitions) in an additional linked list, in file order.  The
 * symbol list is used to build a hash table after the file has been read in.
 *
 * +----------------------------------------------------------------------+
 * | In use code - NotInUse if not, InUse if so                           |
 * +----------------------------------------------------------------------+
 * | File identifier (actually the index into an array of these struct's) |
 * +----------------------------------------------------------------------+
 * | File name (without ., or .par)                                       |
 * +----------------------------------------------------------------------+
 * | Full path to file                                                    |
 * +----------------------------------------------------------------------+
 * | Time file was last modified (used to make sure we don't overwrite    |
 * |   changes that were made while we were working on the file contents) |
 * +----------------------------------------------------------------------+
 * | Count of lines in the file (historic?)                               |
 * +----------------------------------------------------------------------+
 * | Cound of lines with symbol definitions                               |
 * +----------------------------------------------------------------------+
 * | Pointer to a linked list of all lines in the file                    |
 * +----------------------------------------------------------------------+
 * | Pointer to first line in list with a name field (for searching)      |
 * +----------------------------------------------------------------------+
 * | Size of hash table (must be prime)                                   |
 * +----------------------------------------------------------------------+
 * | Pointer to hash table of symbol definitions                          |
 * +----------------------------------------------------------------------+
 */

#define NotInUse 0		  /* entry not in use, must be zero */
#define InUse 05252		  /* entry in use, must be non-zero */

typedef struct FileEntryStruct {
    unsigned int InUseCode;
    FILE_ID FileID;
    char FileName[CFP_FILESIZE];
    char FilePath[CFP_PATHSIZE];
    time_t LastModified;
    unsigned int LineCount;
    unsigned int SymbolCount;
    LineType *Lines;
    LineType *FirstWithName;
    unsigned int HashSize;
    HashTableType HashTable;
} FileEntryType;

/*********************
 * global references *
 *********************/

/* private version of file instance array */
extern FileEntryType FileEntryTable[CFP_MAXINST + 1];

/*
 * cfp internal function definitions (all in cfp_util.c)
 */

FileEntryType *cfp_CheckID P((FILE_ID id));
void cfp_FreeVariableSpace P((LineType *Line));
void cfp_FreeLineSpace P((FileEntryType *FileEntry));
PASSFAIL cfp_DoVariableDef P((char **Input, unsigned int *Column,
			      LineType *Line));
PASSFAIL cfp_InitLines P((FileEntryType *FileEntry, char *FileData));
char *cfp_GetEnvValue P((FileEntryType *FileEntry, char *name, int size));
PASSFAIL cfp_PutEnvValue P((FileEntryType *FileEntry, char *name, int size,
			    char *Value));
unsigned int cfp_HashIt P((FileEntryType *FileEntry, char *Name));
LineType *cfp_FindName P((FileEntryType *FileEntry, char *Name,
			  BooleanType LookupOnly));
PASSFAIL cfp_CopyToCurrent P((LineType *Line, char *Value));
PASSFAIL cfp_GetItem P((LineType *Line, ItemIndexType Index,
			char **String, unsigned int *Length));
PASSFAIL cfp_PutItem P((LineType *Line, ItemIndexType Index, char *Value));
PASSFAIL cfp_RebuildVariableValue P((LineType *Line, char **String,
				     unsigned int *Length));
PASSFAIL cfp_RebuildVariableLine P((LineType *Line, char **String,
				    size_t *Length));
PASSFAIL cfp_RebuildLine P((LineType *Line, char **String, size_t *Length));
PASSFAIL cfp_RebuildFile P((FileEntryType *FileEntry, char **FileData,
			    unsigned int *Length));
PASSFAIL cfp_ConvertFromString P((int Type, char *SourceString,
				  void *Destination, int DestinationSize));
PASSFAIL cfp_CheckRange P((LineType *Line, int Type, char *Value));
PASSFAIL cfp_EditValue P((LineType *Line, int Type, char *Value));
PASSFAIL cfp_BuildRangeMessage P((LineType *Line, char *Message));

#endif /* CFPINTERNALDOTH */
