/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* cfp_io.c
* $Header: cfp_io.c,v 2.2 96/11/19 23:02:39 isani Exp $
* $Locker:  $
*
* ROUTINES
*
*   cfp_io()
*
* DESCRIPTION
*    This routine will fill in a cfp_io data structure using information
*    obtained from a file.
*
* INVOKED
*    io_struct = cfp_io(name, data, type);
*
* PARAMETERS IN
*    char *name - file name to extract information from
*    int type - type of device, or -1 to defer to type found in file
*
* PARAMETERS OUT
*    struct cfp_io *data - pointer to structure for this routine to fill
*                          in, or NULL to use static structure in routine
*
* FUNCTION RETURNS
*    struct cfp_io * - pointer to filled in data structure
*
* GLOBALS
*    none
*    the local static data structure is over-written each time it is used
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jab          15 Jun 1988     Original coding
* jab          08 Aug 1988     Added baud rate to ioconfig stuff
* jrw          20 Dec 1990     sun port
* jrw          15 Feb 1991     check for null pointers
*
* $Log:	cfp_io.c,v $
 * Revision 2.2  96/11/19  23:02:39  isani
 * cleaned up rcsid
 * added parens around a couple of assignments used as truth statements
 * changed a q that should have been a p to a p (around line 260)
 * 
 * Revision 2.1  1994/12/14  16:42:29  thomas
 * Added #ifdef __LINT__ around RCS headers
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h
 * Changed CFHT_ERROR to CFHT_LOGONLY
 *
 * Revision 1.1  91/09/24  17:18:29  17:18:29  steve (Steven Smith)
 * Original checkin
 * 
****************************************************************************!*/

#include <string.h>		  /* for string defines */
#include <errno.h>		  /* for error codes and errno defines */
#include <termio.h>		  /* for baud rate defines */
#include <stdlib.h>		  /* for getenv() */

#include "cfht.h"		  /* generic defines */
#include "cfp.h"		  /* parsing defines */

RCSID("$Header: cfp_io.c,v 2.2 96/11/19 23:02:39 isani Exp $");

static struct cfp_io dummy;

struct cfp_io *			  /* return: pointer to data or NULL */
cfp_io(name, data, type)
    char *name;			  /* symbolic name of device */
    struct cfp_io *data;	  /* optional pointer to data space */
    int type;			  /* CFP_CAMAC, CFP_RS232, CFP_HPIB, or -1 */
{
    char linedata[CFHT_VALUE_SIZE];
    char *line;
    char *ename;
    char *p, *q;

    if (!name) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY, "(cfp_io:%d) no name", __LINE__);
        errno = EINVAL;
        return(NULL);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_io) start, name = '%s'", name);

    if (!data)
        data = &dummy;

    /**********************
     * get name of device *
     **********************/

    ename = (char *)getenv(name);
    if (ename)
        name = ename;

    /*******************************
     * get line from ioconfig file *
     *******************************/

    line = linedata;
    if (!cfp_getln(line, name, CFHT_IOCONFIG, NULL)) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_io:%d) cfp_getln failed for '%s' in '%s'",
		  __LINE__, name, CFHT_IOCONFIG);
        return(NULL);
    }
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_io:%d) got '%s' for '%s' from ioconfig file '%s'",
	      __LINE__, linedata, name, CFHT_IOCONFIG);

    /************
     * get type *
     ************/

    p = cfht_tok(line, " \t", &line);
    if (!p) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_io:%d) no device type found in '%s' for '%s' in '%s'",
		  __LINE__, linedata, name, CFHT_IOCONFIG);
        errno = ENOTTY;
        return(NULL);
    }
    switch(p[0]) {
    case 'R':
        data->type = CFP_RS232;
        break;
    case 'C':
        data->type = CFP_CAMAC;
        break;
    case 'H':
        data->type = CFP_HPIB;
        break;
    default:
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_io:%d) found illegal type '%c' for '%s' in '%s'",
		  __LINE__, p[0], name, CFHT_IOCONFIG);
        errno = EINVAL;
        return(NULL);
    }

    if (type != -1 && type != data->type) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_io:%d) wrong type (call=%d, file=%d) for '%s' in '%s'",
		  __LINE__, type, data->type, name, CFHT_IOCONFIG);
        errno = ENOTTY;
        return(NULL);
    }

    /*******************
     * get device node *
     *******************/

    p = cfht_tok(line, " \t", &line);
    if (!p) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_io:%d) no device node found for '%s' in '%s'",
		  __LINE__, name, CFHT_IOCONFIG);
        errno = ENOTTY;
        return(NULL);
    }
    strcpy(data->file, p);

    /****************************
     * parse out other elements *
     ****************************/

    data->hpib = -1;
    data->branch = -1;
    data->crate = -1;
    data->module = -1;
    data->cflag = -1;

    while ((p = cfht_tok(line, " \t", &line))) {
        switch(p[0]) {
        case 'H':
            data->hpib = atoi(&p[1]);
            break;
        case 'B':
            data->branch = atoi(&p[1]);
            break;
        case 'C':
            data->crate = atoi(&p[1]);
            break;
        case 'N':
            data->module = atoi(&p[1]);
            break;
        case 'R':
            /* note the bizarre baud rates that HP-UX supports */
            q = cfht_tok(p, "R,", &p);
            switch(atoi(q)) {
            case 50:    data->cflag = B50;    break;
            case 75:    data->cflag = B75;    break;
            case 110:   data->cflag = B110;   break;
            case 134:   data->cflag = B134;   break;
            case 150:   data->cflag = B150;   break;
            case 200:   data->cflag = B200;   break;
            case 300:   data->cflag = B300;   break;
            case 600:   data->cflag = B600;   break;
#ifdef hpux
            case 900:   data->cflag = B900;   break;
#endif
            case 1200:  data->cflag = B1200;  break;
            case 1800:  data->cflag = B1800;  break;
            case 2400:  data->cflag = B2400;  break;
#ifdef hpux
            case 3600:  data->cflag = B3600;  break;
#endif
            case 4800:  data->cflag = B4800;  break;
#ifdef hpux
            case 7200:  data->cflag = B7200;  break;
#endif
            case 9600:  data->cflag = B9600;  break;
            case 19200: data->cflag = B19200; break;
            case 38400: data->cflag = B38400; break;
            default:
                cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                          "(cfp_io:%d) bad baud '%s' for '%s' in '%s'",
			  __LINE__, q, name, CFHT_IOCONFIG);
                errno = EIO;
                return(NULL);
	        /*NOTREACHED*/
                break;
            }
            if ((q = cfht_tok(p, ",", &p))) {
                switch(*q) {
                case '5':  data->cflag |= CS5;  break;
                case '6':  data->cflag |= CS6;  break;
                case '7':  data->cflag |= CS7;  break;
                case '8':  data->cflag |= CS8;  break;
                default:
                    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                              "(cfp_io:%d) bad bits '%s' for '%s' in '%s'",
			      __LINE__, q, name, CFHT_IOCONFIG);
                    errno = EIO;
                    return(NULL);
		    /*NOTREACHED*/
                    break;
                }
            }
            if ((q = cfht_tok(p, ",", &p))) {
                switch(*q) {
                case 'E':  data->cflag |= PARENB;           break;
                case 'O':  data->cflag |= PARENB | PARODD;  break;
                default:
                    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                              "(cfp_io:%d) bad parity '%s' for '%s' in '%s'",
			      __LINE__, q, name);
                    errno = EIO;
                    return(NULL);
		    /*NOTREACHED*/
                    break;
                }
            }
            break;
        default:
            cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                      "(cfp_io:%d) bad element '%s' for '%s' in '%s'",
		      __LINE__, p, name, CFHT_IOCONFIG);
            errno = EIO;
            return(NULL);
	    /*NOTREACHED*/
            break;
        }
    }

    cfht_log(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_io) end");

    return(data);
}
