/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* cfp_match.c
* $Header: cfp_match.c,v 2.2 96/11/19 23:02:59 isani Exp $
* $Locker:  $
*
* ROUTINES
*
*   cfp_match()
*
* DESCRIPTION
*    This routine looks for a string in a table.  The table is a vector of
*    pointers to strings to match against.  It is terminated by a null
*    pointer or a null string.  Abbreviations are allowed as long as they
*    are unique.  A leading '-' in the string is ignored.  The first exact
*    match takes precedence over any abbreviations.  If a match is found, the
*    index into the vector for the matching string is returned.  If none is
*    found, FAIL is returned.
*
* INVOKED
*    index = cfp_match(NameToMatch, ComparisonTable);
*
* PARAMETERS IN
*    char *name - identifier to look for
*    char **addrs - list of strings to match against, terminated by a pointer
*                   to a null string or a null pointer
*
* PARAMETERS OUT
*    none
* 
* FUNCTION RETURNS
*    int - index of matching entry in list if unambibuous match found
*          FAIL if no match or multiple matches
* 
* GLOBALS
*    none
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jab           1 Mar 1988     Original coding
* jrw          15 Feb 1991     check for null pointers
*
* $Log:	cfp_match.c,v $
 * Revision 2.2  96/11/19  23:02:59  isani
 * cleaned up rcsid
 * 
 * Revision 2.1  1994/12/14  16:43:47  thomas
 * Added #ifdef __LINT__ around RCS headers
 * Removed sys/types.h
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h
 * Changed CFHT_ERROR to CFHT_LOGONLY
 *
 * Revision 1.1  91/09/24  17:24:25  17:24:25  steve (Steven Smith)
 * Original checkin
 * 
*
****************************************************************************!*/

#include <string.h>		  /* for string handling defines */
#include <errno.h>		  /* for error codes and errno define */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */

RCSID("$Header: cfp_match.c,v 2.2 96/11/19 23:02:59 isani Exp $");

int				  /* return: index, or FAIL */
cfp_match(name, addrs)
    char *name;			  /* identifier to look for */
    char **addrs;		  /* list of strings to match against */
{
    int count;            /* line number of string */
    int tcount;
    char *p;              /* temporary string pointer */
    int size;             /* size of rest of string */
    int namelen;

    if (!name || !addrs || !*addrs) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_match:%d) null name or address or list provided",
		  __LINE__);
        errno = EIO;
        return(FAIL);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
              "(cfp_match) start with name = '%s'", name);

    count = FAIL;
    if (*name == '-')
        name++;

    /* for each line */
    namelen = strlen(name);
    for (tcount=0, p=addrs[tcount] ; p && *p ; tcount++, p=addrs[tcount]) {
        if (strncmp(p, name, (size_t)namelen))     /* test for name */
            continue;                              /* try next line */

        p += namelen;                              /* skip over name */
        size = strlen(p);                          /* how close a match ? */

        /* check for collisions and exact matches */
        if ((count == FAIL)  || size == 0) {
            count = tcount;                        /* found something */
            if (size == 0)
                break;                             /* exact match */
        } else {
            cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                      "(cfp_match:%d) collision for '%s'", __LINE__, name);
            errno = EIO;                           /* collision */
            return(FAIL);
        }
    }

    /**********************
     * check for no match *
     **********************/

    if (count == FAIL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_match:%d) no match for '%s'", __LINE__, name);
        errno = ENOENT;
        return(FAIL);
    }

    /*************************************************************
     * pass back index of zero length string if zero length name *
     *************************************************************/

    if (namelen == 0)
        count = tcount;

    cfht_log(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_match) end");

    return(count);
}
