/* Copyright (C) 1994-2002    Canada-France-Hawaii Telescope Corp.     */
/* This program is distributed WITHOUT any warranty, and is under the  */
/* terms of the GNU General Public License, see the file COPYING       */
/*!****************************************************************************
 * FILE cfp_parse.c
 * $Id: cfp_parse.c,v 2.5 2002/02/27 02:18:44 thomas Exp $
 * $Locker:  $
 *
 * ROUTINES
 *
 *   cfp_parse()
 *
 * DESCRIPTION
 *    This routine will parse the command line arguments present when the
 *    program is invoked.  This routine has two modes of operation.  In the
 *    first mode, the constant CFP_PAR is passed in as the first argument.
 *    This causes the program to be registered with the cfht_log() routines,
 *    the ".par" file to be opened and then parsed and finally the id of
 *    the par file is returned.
 *    In the second mode of operation, an already opened par file is
 *    re-parsed, with the return being PASSFAIL.
 *
 * INVOKED
 *    status = cfp_parse(id, argc, argv);
 *    or
 *    id = cfp_parse(CFP_PAR, argc, argv);
 *
 * PARAMETERS IN
 *    FILE_ID id - par file id for re-parsing, or the constant CFP_PAR
 *                 for an initial parsing
 *    int argc - the count of arguments as passed in to main() by shell
 *    char **argv - the array of arguments as passed in to main() by shell
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    FILE_ID - if the constant CFP_PAR is passed in, then this returns the
 *              FILE_ID of the par file or FAIL on error
 *    or
 *    PASSFAIL - if a previous FILE_ID is passed in, then this returns an
 *               indication of the re-parsing completion status
 *
 * GLOBALS
 *    The CFP_PAR mode causes a FileEntryTable entry to be allocated to this
 *    file.
 *
 * HISTORY
 * who          when            what
 * ---------    ------------    ----------------------------------------------
 * jab          27 Jan 1988     Original coding
 * jab          15 Mar 1988     Find par file and start logging
 * jab          15 Nov 1988     now optional, also kill envp
 * jrw          15 Feb 1991     check for null pointers
 *
 * $Log: cfp_parse.c,v $
 * Revision 2.5  2002/02/27 02:18:44  thomas
 * Added value to missing floater message
 *
 * Revision 2.4  1996/11/19 23:03:34  isani
 * cleaned up rcsid
 * added parens around assignment used as truth statement
 * moved sun hacks to "cfht.h"
 *
 * Revision 2.3  1995/02/27  20:10:28  thomas
 * Updated test for range checking to test RangeInfo for NULL
 *
 * Revision 2.2  94/12/21  12:43:57  12:43:57  thomas (Jim Thomas)
 * Added "not just a lookup" flag to cfp_FindName calls
 * 
 * Revision 2.1  94/12/14  16:48:02  16:48:02  thomas (Jim Thomas)
 * Added #ifdef __LINT__ around RCS headers
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h, added cfp_internal.h
 * Changed CFHT_ERROR to CFHT_LOGONLY
 * Rewrote to use version 2 routines
 * Added range check on provided argument values
 * 
 * Revision 1.1  91/09/24  17:24:49  17:24:49  steve (Steven Smith)
 * Original checkin
 * 
****************************************************************************!*/

#include <string.h>		  /* for string handling defines */
#include <errno.h>		  /* for error codes and errno define */
#include <stdio.h>		  /* sprinf() */
#include <stdlib.h>		  /* for malloc */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */
#include "cfp_internal.h"	  /*  and internals */

RCSID("$Id: cfp_parse.c,v 2.5 2002/02/27 02:18:44 thomas Exp $");

#define InvokeLength 128	  /* length for printable invocation string */

/*ARGSUSED*/
int				  /* return: PASS, or FAIL, or FILE_ID */
cfp_parse(FILE_ID id, int argc, char **argv)
{
    FileEntryType *FileEntry;	  /* pointer to file definition */
    LineType *ThisLine;		  /* pointer to variable def for arg */
    int arg;			  /* number of current arg */
    int floatnum;		  /* number of current floater arg */

    if (!argv || !*argv)
        return FAIL;

    /*****************************
     * handle a first time parse *
     *****************************/

    if (id == CFP_PAR) {
        char string[CFP_PATHSIZE + 1];
        char invoke[InvokeLength + 1];
        char *truename;
        int i;

        /*******************************
         * get the name of the program *
         *******************************/

        if ((truename = strrchr(argv[0], '/')))
            strcpy(argv[0], truename + 1);
        cfht_log(CFHT_LIBCFP, CFHT_LOG_ID, argv[0]);

        cfht_log(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_parse) CFP_PAR starting");

        /*****************************
         * log the invokation string *
         *****************************/

        strncpy(invoke, argv[0], InvokeLength);
	invoke[InvokeLength] = '\0';
        for (i=1; argv[i] && i<argc; i++) {
	    if ((strlen(invoke) + strlen(argv[i]) + 1) < InvokeLength) {
		strcat(invoke, " ");
		strcat(invoke, argv[i]);
	    }
        }
        cfht_log(CFHT_MAIN, CFHT_START, invoke);

        if (!cfp_file(string, argv[0], ".par", "rw")) {
            cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                     "(cfp_parse:%d) cfp_file failed on .par file '%s.par'",
		      __LINE__, argv[0]);
            return FAIL;
        }
        if ((id = cfp_init(string)) < 0) {
            cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                     "(cfp_parse:%d) cfp_init failed on .par file '%s'",
		      __LINE__, string);
            return FAIL;
        }

	/* call ourselves with the open file to use the code below :-) */
        if (cfp_parse(id, argc, argv) != PASS)
            return FAIL;

        if (cfp_putf(id) != PASS) {
            cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_parse:%d) cfp_putf failed on '%s'",
		      __LINE__, string);
            return FAIL;
        }

        cfht_log(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_parse) CFP_PAR ending");

        return id;
    }

    /************************************
     * handle the parsing of an open id *
     ************************************/

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_parse:%d) cfp_CheckID failed on id=%d",
		  __LINE__, id);
        return FAIL;
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_parse) start on '%s'",
	      FileEntry->FilePath);

    /* process each argument */
    floatnum = 0;
    arg = 1;
    while ((arg < argc) && (argv[arg] != (char *) NULL)) {
        /* does argument specify an option ? */
        if (argv[arg][0] == '-') {
            if ((ThisLine = cfp_FindName(FileEntry, &(argv[arg][1]), FALSE))
		== (LineType *) NULL) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_parse:%d) lookup '%s' in '%s' failed",
			  __LINE__, &argv[arg][1], FileEntry->FilePath);
		errno = EIO;
		return FAIL;
	    }
#ifdef DEBUGDEBUG
	    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		      "(cfp_parse) saw argument '%s', matched '%s'",
		      argv[arg], ThisLine->NameString);
#endif
	    arg++;
        } else {		  /* argument is a floater */
	    char floater[CFP_STRSIZE + 1];

	    sprintf(floater, "&floater%d", floatnum);
	    if ((ThisLine = cfp_FindName(FileEntry, floater, FALSE))
		== (LineType *) NULL) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			 "(cfp_parse:%d) no floater named '%.100s' for value '%.100s'",
			  __LINE__, floater, argv[arg]);
		errno = EIO;
		return FAIL;
	    }
#ifdef DEBUGDEBUG
	    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		      "(cfp_parse) saw floater %d using name '%s'",
		      floatnum, ThisLine->NameString);
#endif
	    floatnum++;
	}

	if (ThisLine->TypeCode != VariableLine) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_parse:%d) argument '-%s' is not a variable",
		      __LINE__, ThisLine->NameString);
	    errno = EIO;
	    return FAIL;
	}

	ThisLine->History = CFP_FOUND_CURRENT;

	/* now if we had "-<name> <value>" or "<floater value>" do the value */
	if ((arg < argc)
	    && (argv[arg] != (char *) NULL)
	    && argv[arg][0] != '-') {
	    if (ThisLine->ValueNotAllowed == TRUE) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_parse:%d) value not allowed for '-%s'",
			  __LINE__, ThisLine->NameString);
		errno = EIO;
		return FAIL;
	    }
	    if ((ThisLine->RangeInfo != (RangeInfoType *) NULL)
		&& (cfp_CheckRange(ThisLine, CFP_STRING, argv[arg]) != PASS)) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_parse:%d) range check error on '-%s %s'",
			  __LINE__, ThisLine->NameString, argv[arg]);
		return FAIL;
	    }
	    if (ThisLine->CurrentValue != (char *) NULL)
	        free(ThisLine->CurrentValue);
	    if ((ThisLine->CurrentValue
		 = (char *) malloc(strlen(argv[arg]) + 1)) == (char *) NULL) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_parse:%d) current malloc %d bytes failed",
			  __LINE__, strlen(argv[arg]) + 1);
		errno = EIO;
		return FAIL;
	    }
	    strcpy(ThisLine->CurrentValue, argv[arg]);
	    ThisLine->CurrentValueLength = strlen(ThisLine->CurrentValue);
	    ThisLine->History = CFP_FOUND_ARGV;
	    arg++;
#ifdef DEBUGDEBUG
	    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		      "(...)       and set current to '%s'",
		      ThisLine->CurrentValue);
#endif
	}
    }

    /* now go through all lines in the file and see if any defaults or
     * supplied values need to be filled in or if a required value is missing
     */
    ThisLine = FileEntry->Lines;
    while (ThisLine != (LineType *) NULL) {
	if (ThisLine->History == CFP_FOUND_CURRENT) {
	    if ((ThisLine->ValueOptional == TRUE)
		||
		(ThisLine->ValueNotAllowed == TRUE)) {
		if (ThisLine->UseSupplied == TRUE) {
		    if (cfp_CopyToCurrent(ThisLine, ThisLine->SuppliedValue)
			!= PASS) {
			cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
				  "(cfp_parse:%d) substitute supplied failed",
				  __LINE__);
			return FAIL;
		    }
		    ThisLine->History = CFP_FOUND_SUPPLIED;
#ifdef DEBUGDEBUG
		    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
			      "(cfp_parse) substituted supplied '%s'",
			      ThisLine->SuppliedValue);
#endif
		} else if (ThisLine->UseDefault == TRUE) {
		    if (cfp_CopyToCurrent(ThisLine, ThisLine->DefaultValue)
			!= PASS) {
			cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
				  "(cfp_parse:%d) substitute default failed",
				  __LINE__);
			return FAIL;
		    }
		    ThisLine->History = CFP_FOUND_DEFAULT;
#ifdef DEBUGDEBUG
		    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
			      "(cfp_parse) substituted default '%s'",
			      ThisLine->DefaultValue);
#endif
		}
	    } else {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_parse:%d) for '%s' -%s w/o required value",
			  __LINE__, FileEntry->FilePath, ThisLine->NameString);
		errno = EIO;
		return FAIL;
	    }
	}
	ThisLine = ThisLine->Next;
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_parse) end on '%s'",
	      FileEntry->FilePath);

    return PASS;
}
