/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* cfp_putcur.c
* $Header: cfp_putcur.c,v 2.4 96/11/19 23:04:04 isani Exp $
* $Locker:  $
*
* ROUTINES
*
*   cfp_putcur()
*
* DESCRIPTION
*    This routine will update the current value of a par file variable,
*    recording it in in-memory copy of the par file.
*
*    If a CFP_STRING is stored into a constant, this updates the constant
*    instead.
*
* INVOKED
*    status = cfp_putcur(id, name, addr, type);
*
* PARAMETERS IN
*    FILE_ID id - par file id
*    char *name - the name of the identifier to update
*    void *addr - pointer to the current value
*    int type - the identifier's type
*
* PARAMETERS OUT
*    FILE_ID id - the line data for this file is updated with a new current
*
* FUNCTION RETURNS
*    PASSFAIL - PASS if OK, FAIL if something wrong
*
* GLOBALS
*    none
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jab          17 Feb 1988     Original coding
* jrw          16 Feb 1991     check for null pointers
* jrw          20 Feb 1991     ugly hack since SPARCs don't know about union
*
* $Log:	cfp_putcur.c,v $
 * Revision 2.4  96/11/19  23:04:04  isani
 * cleaned up rcsid
 * changed a "%lg" format string to "%g"
 * 
 * Revision 2.3  1995/02/27  20:10:48  thomas
 * Updated test for range checking to test RangeInfo for NULL
 *
 * Revision 2.2  94/12/21  12:44:07  12:44:07  thomas (Jim Thomas)
 * Added "not just a lookup" flag to cfp_FindName call
 * 
 * Revision 2.1  94/12/14  16:51:16  16:51:16  thomas (Jim Thomas)
 * Added #ifdef __LINT__ around RCS headers
 * Added include of stdlib.h
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h, added cfp_internal.h
 * Changed CFHT_ERROR to CFHT_LOGONLY
 * Rewrote to use version 2 routines
 * Added range check on provided value
 * 
 * Revision 1.1  91/09/24  17:26:08  17:26:08  steve (Steven Smith)
 * Original checkin
 * 
****************************************************************************!*/

#include <string.h>		  /* for string handling defines */
#include <errno.h>		  /* for error codes and errno define */
#include <stdio.h>		  /* sprintf() */
#include <stdlib.h>		  /* for malloc */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */
#include "cfp_internal.h"	  /*   and internals */

RCSID("$Header: cfp_putcur.c,v 2.4 96/11/19 23:04:04 isani Exp $");

PASSFAIL
cfp_putcur(id, name, addr, type)
    FILE_ID id;			  /* instance handle */
    char *name;			  /* identifier to look for */
    void *addr;			  /* source address */
    int type;			  /* addrs actual type */
{
    FileEntryType *FileEntry;	  /* pointer to file data */
    LineType *Line;		  /* pointer to a line in this file */
    char *Result;		  /* pointer to resulting string value */
    char Scratch[CFP_STRSIZE + 1]; /* scratch string for converting */
    size_t Size;		  /* size for constant update */

    if ((name == (char *) NULL) || (addr == (void *) NULL)) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putcur:%d) null parameter passed in", __LINE__);
        return(FAIL);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_putcur) start to %s", name);

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putcur:%d) CheckID failed on %d for '%s'",
		  __LINE__, id, name);
        return(FAIL);
    }

    if ((Line = cfp_FindName(FileEntry, name, FALSE)) == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putcur:%d) name '%s' not found for '%s'",
		  __LINE__, name, FileEntry->FilePath);
	return(FAIL);
    }

    /************************************************************
     * if not a variable and string is requested handle special *
     ************************************************************/

    if ((Line->TypeCode == ConstantLine) && (type == CFP_STRING)) {
	if (Line->OriginalValue != (char *) NULL) {
	    free((void *) Line->OriginalValue);
	}
	Size = strlen((char *) addr);
	if ((Line->OriginalValue
	     = (char *) malloc(Size + 1)) == (char *) NULL) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_putcur:%d) malloc of %d failed for '%s' in '%s'",
		      __LINE__, Size + 1, Line->NameString,
		      FileEntry->FilePath);
	    return(FAIL);
	}
	strcpy(Line->OriginalValue, (char *) addr);
	Line->OriginalValueLength = Size;

	if (Size > CFP_STRSIZE) {
	    strncpy(Scratch, Line->OriginalValue, CFP_STRSIZE);
	    Scratch[CFP_STRSIZE] = '\0';
	    Result = Scratch;
	} else {
	    Result = Line->OriginalValue;
	}
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_putcur) end (value = '%s')", Result);

	return(PASS);		  /* NOTE NOTE NOTE NOTE :-( */
    }

    else if (Line->TypeCode != VariableLine) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putcur:%d) trying to update non-variable '%s' for '%s'",
		  __LINE__, Line->NameString, FileEntry->FilePath);
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(...)           type code is %d", Line->TypeCode);
	return(FAIL);
    }

    switch(type) {
      case CFP_BOOLEAN:
        if (*((BOOLEAN *)addr) == TRUE)
	  Result = "True";
        else
	  Result = "False";
        break;
      case CFP_DOUBLE:
        (void) sprintf(Scratch, "%g", *((double *)addr));
	Result = Scratch;
        break;
      case CFP_INTEGER:
        (void) sprintf(Scratch, "%ld", *((long *)addr));
	Result = Scratch;
        break;
      case CFP_STRING:
        Result = (char *) addr;
        break;
      default:
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_putcur:%d) '%s' unknown type %d, file '%s'",
		  __LINE__, name, type, FileEntry->FilePath);
        errno = EIO;
        return(FAIL);
	/*NOTREACHED*/
        break;
    }

    /* check range? */
    if (Line->RangeInfo != (RangeInfoType *) NULL) {
	switch (type) {
	  case CFP_INTEGER:
	  case CFP_DOUBLE:
	  case CFP_STRING:
	    if (cfp_CheckRange(Line, type, Result) != PASS) {
		if (Result != Scratch) {
		    strncpy(Scratch, Result, CFP_STRSIZE);
		    Scratch[CFP_STRSIZE] = '\0';
		}
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_putcur:%d) %s bad or out of range for '%s' in '%s'",
			  __LINE__, Scratch, Line->NameString,
			  FileEntry->FilePath);
		return(FAIL);
	    }
	    break;

	  default:
	    /* do not care about booleans */
	    break;
	}
    }

    if (cfp_CopyToCurrent(Line, Result) != PASS) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putcur:%d) update for '%s' to '%s' failed for '%s'",
		  __LINE__, Line->NameString, addr, FileEntry->FilePath);
	return(FAIL);
    }

    Line->History = CFP_SET;

    if (Line->CurrentValueLength > CFP_STRSIZE) {
	strncpy(Scratch, Line->CurrentValue, CFP_STRSIZE);
	Scratch[CFP_STRSIZE] = '\0';
	Result = Scratch;
    } else {
	Result = Line->CurrentValue;
    }
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
              "(cfp_putcur) end (history = %d, current = '%s')",
	      Line->History, Result);

    return(PASS);
}
