/* Copyright (C) 1994    Canada-France-Hawaii Telescope Corp.            */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* cfp_putf.c
* $Header: cfp_putf.c,v 2.6 96/11/19 23:04:50 isani Exp $
* $Locker:  $
*
* ROUTINES
*
*   cfp_putf()
*
* DSCRIPTION
*    This routine will write to disk the in-memory copy of the par file
*    associated with the id.
*
* INVOKED
*    status = cfp_putf(id);
*
* PARAMETERS IN
*    FILE_ID id - par file id
*
* PARAMETERS OUT
*    none
*
* FUNCTION RETURNS
*    PASSFAIL - PASS if successful, FAIL if there was any problem
*
* GLOBALS
*    indirectly references FileEntryTable for file contents
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jab          27 Jan 1988     Original coding
* jrw          07 Dec 1990     Made cfht_log calls more verbose
* jrw          20 Dec 1990     sun port
*
* $Log:	cfp_putf.c,v $
 * Revision 2.6  96/11/19  23:04:50  isani
 * cleaned up rcsid
 * type-cast a signed -> unsigned and size_t to int that were ok
 * 
 * Revision 2.5  1995/02/07  18:57:20  thomas
 * Commented out failure on file modified :-(
 *
 * Revision 2.4  95/01/19  09:34:31  09:34:31  thomas (Jim Thomas)
 * Change start/end messages back to DEBUG, generated too many lines
 * 
 * Revision 2.3  95/01/17  12:18:30  12:18:30  thomas (Jim Thomas)
 * Changed start and stop messages to include file name and be LOGONLY to help
 * with debugging file update race conditions
 * 
 * Revision 2.2  94/12/19  12:12:45  12:12:45  thomas (Jim Thomas)
 * Took out file locking part of stale file check under ifdef UseLocking
 * 
 * Revision 2.1  94/12/14  16:54:01  16:54:01  thomas (Jim Thomas)
 * Added #ifdef __LINT__ around RCS headers
 * Removed sys/types.h, added sys/stat.h, unistd.h
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h, added cfp_internal.h
 * Changed CFHT_ERROR to CFHT_LOGONLY
 * Rewrote to use version 2 routines
 * Added file locking and last update time check for modified files check
 * 
 * Revision 1.1  91/09/24  17:26:31  17:26:31  steve (Steven Smith)
 * Original checkin
 * 
****************************************************************************!*/

#include <sys/stat.h>		  /* for file modification time */
#include <unistd.h>		  /* for seeking and locking */
#include <stdlib.h>		  /* free() */
#include <fcntl.h>		  /* for O_WRONLY */
#include <string.h>		  /* for string handling defines */
#include <errno.h>		  /* for error codes and errno define */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */
#include "cfp_internal.h"	  /*   and internals */

RCSID("$Header: cfp_putf.c,v 2.6 96/11/19 23:04:50 isani Exp $");

/*
 * There needs to be a check on whether the file has been updated by some
 * other process since it was read in to this one.  The inode number can not
 * be used because this is an update, not a rewrite.  That leaves the
 * modification time, but that only has 1 second resolution.  Originally I put
 * a lock on the file before checking the time and doing the update, but that
 * failed on .,config because the session locks it too.  Since the possibility
 * of a race between the time check and the rewrite is much smaller than the 1
 * second accuracy of the times, the lock is out under an ifdef UseLocking .
 * (I note also that I should have used cfht_setlock .)
 */
/* #define UseLocking		  / * lock the file during stale check */

PASSFAIL			  /* return: PASS, or FAIL */
cfp_putf(id)
    FILE_ID id;			  /* instance handle */
{
    FileEntryType *FileEntry;	  /* info about the file */
    int fd;			  /* file descriptor to new file */
    struct stat statbuf;	  /* current file status */
    char *Buffer;		  /* pointer to reconstituted file */
    unsigned int Size;		  /* buffer size */
    int writeCount;		  /* return from buffer write */

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putf:%d) cfp_CheckID failed", __LINE__);
        return(FAIL);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_putf) start, name = '%s'",
	      FileEntry->FilePath);

    if ((fd = open(FileEntry->FilePath, O_WRONLY)) < 0) {
        cfht_logv (CFHT_LIBCFP, CFHT_LOGONLY,
                   "(cfp_putf:%d) open of '%s' failed (%s)",
		   __LINE__, FileEntry->FilePath, strerror(errno));
        return(FAIL);
    }

#ifdef UseLocking
    /* make SURE we are at start of file to lock the whole thing :-) */
    if (lseek(fd, (off_t) 0, SEEK_SET) < 0) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putf:%d) seek failed on '%s' (%s)",
		  __LINE__, FileEntry->FilePath, strerror(errno));
	close(fd);
	return(FAIL);
    }
    if (lockf(fd, F_TLOCK, (off_t) 0) < 0) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putf:%d) lock failed on '%s', error '%s'",
		  __LINE__, FileEntry->FilePath, strerror(errno));
	close(fd);
	return(FAIL);
    }
#endif

    if (fstat(fd, &statbuf) < 0) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putf:%d) stat failed on '%s', error '%s'",
		  __LINE__, FileEntry->FilePath, strerror(errno));
	close(fd);
	return(FAIL);
    }
    /* now, FINALLY, see if someone else has crunched the file since we read
     * in our copy */
    if (FileEntry->LastModified != statbuf.st_mtime) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putf:%d) file '%s' changed since this copy read",
		  __LINE__, FileEntry->FilePath);
/*
generated too many parse failed messages, need cfp_parsereadonly
	close(fd);
	return(FAIL);
*/
    }

    if (cfp_RebuildFile(FileEntry, &Buffer, &Size) != PASS) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putf:%d) failed rebuilding '%s'",
		  __LINE__, FileEntry->FilePath);
	close(fd);
	return(FAIL);
    }

    if ((unsigned)(writeCount = write(fd, Buffer, Size)) != Size) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_putf:%d) write %d bytes to '%s'; only %d written (%s)",
                  __LINE__, Size, FileEntry->FilePath, writeCount,
		  strerror(errno));
	free(Buffer);
        close(fd);
        return(FAIL);
    }

    free(Buffer);

    if (ftruncate(fd, (int)Size) < 0) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putf:%d) truncate '%s' to %d bytes failed (%s)",
		  __LINE__, FileEntry->FilePath, Size, strerror(errno));
	close(fd);
	return(FAIL);
    }

    /* fstat(5) claims that close will not change the time. */
    /* But the 1 second accuracy is not much good either :-( */
    if (fstat(fd, &statbuf) < 0) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putf:%d) after write stat failed on '%s' (%s)",
		  __LINE__, FileEntry->FilePath, strerror(errno));
	close(fd);
	return(FAIL);
    }
    FileEntry->LastModified = statbuf.st_mtime;

    close(fd);

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_putf) end, name = '%s'",
	      FileEntry->FilePath);

    return(PASS);
}
