/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* cfp_putval.c
* $Header: cfp_putval.c,v 2.4 96/11/19 22:50:59 isani Exp $
* $Locker:  $
*
* ROUTINES
*
*   cfp_putval()
*
* DESCRIPTITON
*    This routine will put a value in the in-memory copy of the par file
*    for the named identifier.  The in-memory copy of the par file will be
*    initialized if necessary.  [Comments in the code indicate that passing
*    in a valnum of -1 will update the environment, but inspection of the
*    code shows this will not work.]
*
* INVOKED
*    status = cfp_putval(id, name, addr, type, valnum);
*
* PARAMETERS IN
*    FILE_ID id - par file id
*    char *name - identifier to look for
*    void *addr - pointer to object to be put into par file
*    int type - type of object to be put into par file
*    ItemIndexType valnum - which field of par file to modify:
*                 CFP_ModeIndex for the options field,
*                 CFP_CurrentIndex for current,
*                 CFP_DefaultIndex for default,
*                 CFP_SuppliedIndex for supplied,
*                 CFP_MinIndex for minimum
*                 CFP_MaxIndes for maximum
*                 CFP_EnvironmentIndex for environment
*
* PARAMETERS OUT
*    FILE_ID id - one field in line database for this file is updated
*
* FUNCTION RETURNS
*    PASSFAIL - PASS if field updated, FAIL if something went wrong
*
* GLOBALS
*    none
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jab           4 Aug 1988     Original coding
* jrw          20 Feb 1991     ugly hack since SPARCs don't support union
* sss            8 Mar 1991     CFP_INTEGER cast fix
*
* $Log:	cfp_putval.c,v $
 * Revision 2.4  96/11/19  22:50:59  isani
 * cleaned up rcsid
 * changed a "%lg" format string to "%g"
 * 
 * Revision 2.3  1995/02/27  20:11:14  thomas
 * Updated test for range checking to test RangeInfo for NULL
 *
 * Revision 2.2  94/12/21  12:44:27  12:44:27  thomas (Jim Thomas)
 * Added "not just a lookup" flag to cfp_FindName call
 * 
 * Revision 2.1  94/12/14  17:02:36  17:02:36  thomas (Jim Thomas)
 * Added #ifdef __LINT__ around RCS headers
 * Added include of stdio.h
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h, added cfp_internal.h
 * Changed CFHT_ERROR to CFHT_LOGONLY
 * Rewrote to use version 2 routines
 * Added range check of current value update
 * 
 * Revision 1.1  91/09/24  17:32:55  17:32:55  steve (Steven Smith)
 * Original checkin
 * 
****************************************************************************!*/

#include <string.h>		  /* for string handling defines */
#include <stdio.h>		  /* for sprintf */
#include <errno.h>		  /* for error codes and errno define */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */
#include "cfp_internal.h"	  /*   and internals */

RCSID("$Header: cfp_putval.c,v 2.4 96/11/19 22:50:59 isani Exp $");

PASSFAIL			  /* return: PASS or FAIL */
cfp_putval(id, name, addr, type, valnum)
    FILE_ID id;			  /* instance handle */
    char *name;			  /* identifier to look for */
    void *addr;			  /* source address */
    int type;			  /* addr's actual type */
    ItemIndexType valnum;	  /* which field, -1 for env */
{
    FileEntryType *FileEntry;	  /* pointer to file data */
    LineType *Line;		  /* pointer to a line in this file */
    char *Result;		  /* pointer to resulting string value */
    char Scratch[CFP_STRSIZE + 1]; /* scratch string for converting */

    if ((name == (char *) NULL) || (addr == (void *) NULL)) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putval:%d) received null parameter", __LINE__);
        return(FAIL);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_putval) start to %s", name);

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putval:%d) CheckID failed on %d for '%s'",
		  __LINE__, id, name);
        return(FAIL);
    }

    if ((Line = cfp_FindName(FileEntry, name, FALSE)) == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putval:%d) name '%s' not found for '%s'",
		  __LINE__, name, FileEntry->FilePath);
	return(FAIL);
    }

    if (Line->TypeCode != VariableLine) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putval:%d) trying to update non-variable '%s' for '%s'",
		  __LINE__, Line->NameString, FileEntry->FilePath);
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(...)           type code is %d", Line->TypeCode);
	return(FAIL);
    }

    switch(type) {
      case CFP_BOOLEAN:
        if (*((BOOLEAN *)addr) == TRUE)
	  Result = "True";
        else
	  Result = "False";
        break;
      case CFP_DOUBLE:
        (void) sprintf(Scratch, "%g", *((double *)addr));
	Result = Scratch;
        break;
      case CFP_INTEGER:
        (void) sprintf(Scratch, "%ld", *((long *)addr));
	Result = Scratch;
        break;
      case CFP_STRING:
        Result = (char *) addr;
        break;
      default:
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_putval:%d) '%s' unknown type %d, file '%s'",
		  __LINE__, name, type, FileEntry->FilePath);
        errno = EIO;
        return(FAIL);
	/*NOTREACHED*/
        break;
    }

    switch (valnum) {
      case CFP_EnvironmentIndex:
        if (cfp_PutEnvValue(FileEntry, Line->NameString,
			    (int)strlen(Line->NameString), Result) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_putval:%d) update env '%s%s=%s failed for '%s'",
		      __LINE__, FileEntry->FileName, Line->NameString, Result,
		      FileEntry->FilePath);
	    errno = EIO;
	    return(FAIL);
	}
	break;

      case CFP_CurrentIndex:
	/* current is special because putcur updates history but putval
	 * doesn't */
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_putval:%d) used putval to change current '%s' for '%s'",
		  __LINE__, Line->NameString, FileEntry->FilePath);

	if ((valnum == CFP_CurrentIndex)
	    && (Line->RangeInfo != (RangeInfoType *) NULL)) {
	    switch (type) {
	      case CFP_INTEGER:
	      case CFP_DOUBLE:
	      case CFP_STRING:
		if (cfp_CheckRange(Line, type, Result) != PASS) {
		    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			      "(cfp_putval:%d) '%s' bad/out of range for '%s'",
			      __LINE__, Result, Line->NameString);
		    return(FAIL);
		}
		break;

	      default:
		/* don't know what to do with strings, booleans don't care */
		break;
	    }
	}

	if (cfp_PutItem(Line, valnum, Result) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_putval:%d) update field %d = '%s' failed for '%s'",
		      __LINE__, valnum, Result, FileEntry->FilePath);
	    errno = EIO;
	    return(FAIL);
	}
	break;

      case CFP_ModeIndex:
      case CFP_DefaultIndex:
      case CFP_SuppliedIndex:
      case CFP_RangeIndex:
	if (cfp_PutItem(Line, valnum, Result) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_putval:%d) update field %d = '%s' failed for '%s'",
		      __LINE__, valnum, Result, FileEntry->FilePath);
	    errno = EIO;
	    return(FAIL);
	}
	break;

      default:
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_getval:%d) unknown field index %d for '%s'",
		  __LINE__, valnum, FileEntry->FilePath);
	return(FAIL);
	/*NOTREACHED*/
	break;
    }

    if (strlen(Result) > CFP_STRSIZE) {
	strncpy(Scratch, Result, CFP_STRSIZE);
	Scratch[CFP_STRSIZE] = '\0';
	Result = Scratch;
    }
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_putval) end, (field %d = '%s')",
	      valnum, Result);

    return(PASS);
}
