/* Copyright (C) 1994    Canada-France-Hawaii Telescope Corp.            */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*****************************************************************************
 * FILE
 * cfp_range.c
 * $Header: cfp_range.c,v 2.4 96/11/19 22:51:11 isani Exp $
 * $Locker:  $
 *
 * ROUTINES
 *
 *   cfp_rangep()
 *    Given a variable name, see if range checking is enabled
 *
 *   cfp_editvalue()
 *    Given a variable name and a value, see if the value is legitimate for
 *    the type of the variable
 *
 *   cfp_range()
 *    Given a variable name and a value, check the value against the range
 *    field for that variable, if any
 *
 *   cfp_rangemess()
 *    Given a variable name, format a readable message indicating the range
 *    for that variable (to be used for an error message window most likely)
 *
 * HISTORY
 * $Log:	cfp_range.c,v $
 * Revision 2.4  96/11/19  22:51:11  isani
 * cleaned up rcsid's
 * 
 * Revision 2.3  1995/02/27  20:13:36  thomas
 * Added cfp_rangep and cfp_editvalue
 * Moved range field parsing to cfp_util at file parse time
 *
 * Revision 2.2  94/12/21  12:44:37  12:44:37  thomas (Jim Thomas)
 * Added "not just a lookup" flag to cfp_FindName calls
 * 
 * Revision 2.1  94/12/14  17:03:38  17:03:38  thomas (Jim Thomas)
 * Original check in
 * 
 ****************************************************************************/

#include <stdio.h>		  /* for sprintf */
#include <string.h>		  /* for string handling defines */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */
#include "cfp_internal.h"	  /*   and internals */

RCSID("$Header: cfp_range.c,v 2.4 96/11/19 22:51:11 isani Exp $");

/*!**********************************************************************
 *
 *   cfp_rangep()
 *
 * DESCRIPTION
 *    Given a variable name, see if range checking is enabled
 *    This looks up a variable and returns TRUE if range checking is defined
 *    for the variable, FALSE if not, if the variable does not exist, or if
 *    any other error happens.
 *
 * INVOKED
 *    if (cfp_rangep(id, name) == TRUE) { ...
 *
 * PARAMETERS IN
 *    FILE_ID id - par file to operate on
 *    char *name - name of variable to check
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    BOOLEAN - TRUE if range checking is enabled on name, FALSE otherwise
 *
 * GLOBALS
 *    none
 *
 **********************************************************************!*/

BOOLEAN
cfp_rangep(id, name)
    FILE_ID id;			  /* instance handle */
    char *name;			  /* identifier to look for */
{
    FileEntryType *FileEntry;	  /* pointer to file data */
    LineType *Line;		  /* pointer to a line in this file */

    if (name == (char *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_rangep:%d) null name parameter passed in", __LINE__);
        return(FALSE);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_rangep) start on %s", name);

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_rangep:%d) CheckID failed for %d", __LINE__, id);
        return(FALSE);
    }

    if ((Line = cfp_FindName(FileEntry, name, FALSE)) == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_rangep:%d) name '%s' not found for '%s'",
		  __LINE__, name, FileEntry->FilePath);
	return(FALSE);
    }

    if (Line->TypeCode != VariableLine) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_rangep:%d) range-p of non-variable '%s' in '%s'",
		  __LINE__, Line->NameString, FileEntry->FilePath);
        return(FALSE);
    }

    cfht_log(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_rangep) end, OK");

    return(Line->RangeInfo != (RangeInfoType *) NULL);
}

/*!**********************************************************************
 *
 *   cfp_editvalue
 *
 * DESCRIPTION
 *    check to see if a string is legal for the variable type it goes to
 *    Given a variable definition and an input string, see if the input is
 *    valid for the type of variable it is to be stored into.  This does not
 *    mean that the input has to be complete; just that it matches the format
 *    required for a variable of that type.  For example, "1" is valid and
 *    complete, but "-" is valid but incomplete.  "e" and "e+" and "-e-" are
 *    valid, incomplete double values.  "1." is invalid for an integer but
 *    valid for a double.  "50.." is invalid.
 *
 *    The type is determined from the range field.  If there is no range, any
 *    value is valid.  If the given type does not match the range field type,
 *    the range overrides since that is how the final range check will work.
 *
 * INVOKED
 *    Status = cfp_editvalue(id, name, addr, type);
 *
 * PARAMETERS IN
 *    FILE_ID id - par file to operate on
 *    char *name - name of variable to check
 *    char *addr - the value to be checked
 *    int type - the type of the variable
 *               CFP_BOOLEAN - ignored (PASS)
 *               CFP_INTEGER - do integer check
 *               CFP_DOUBLE  - do double check
 *               CFP_STRING  - hope we can guess type from range field
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if check was OK or there is no or a bad range field
 *               FAIL if check definitely failed
 *
 * GLOBALS
 *    none
 *
 **********************************************************************!*/

PASSFAIL
cfp_editvalue(id, name, addr, type)
    FILE_ID id;			  /* instance handle */
    char *name;			  /* identifier to look for */
    char *addr;			  /* value to check */
    int type;			  /* addrs actual type */
{
    FileEntryType *FileEntry;	  /* pointer to file data */
    LineType *Line;		  /* pointer to a line in this file */

    if ((name == (char *) NULL) || (addr == (char *) NULL)) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_editvalue:%d) null parameter passed in", __LINE__);
        return(FAIL);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_editvalue) start on %s", name);

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_editvalue:%d) CheckID failed for %d", __LINE__, id);
        return(FAIL);
    }

    if ((Line = cfp_FindName(FileEntry, name, FALSE)) == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_editvalue:%d) name '%s' not found for '%s'",
		  __LINE__, name, FileEntry->FilePath);
	return(FAIL);
    }

    if (Line->TypeCode != VariableLine) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_editvalue:%d) edit check of non-variable '%s' in '%s'",
		  __LINE__, Line->NameString, FileEntry->FilePath);
        return(PASS);
    }

    if (Line->RangeInfo != (RangeInfoType *) NULL) {
	switch (type) {
	  case CFP_INTEGER:
	  case CFP_DOUBLE:
	  case CFP_STRING:
	    if (cfp_EditValue(Line, type, addr) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_editvalue:%d) '%s' invalid for '%s' in '%s'",
			  __LINE__, addr, Line->NameString,
			  FileEntry->FilePath);
		return(FAIL);
	    }
	    break;

	  default:
	    /* don't know what to do with strings, booleans don't care */
	    break;
	}
    }

    cfht_log(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_editvalue) end, OK");

    return(PASS);
}

/*!***************************************************************************
 *
 *   cfp_range()
 *
 * DESCRIPTION
 *    This routine will check a value against the range field for a variable.
 *    Only integer or double values are actually checked.
 *    If the variable does not have range checking enabled (no legal range
 *    field), it PASSes.
 *    If the variable type is int and the range appears to be int, the check is
 *    an integer check.  If the range is double, a message is logged, but the
 *    check is done using double.
 *    If the variable type is double and the range appears to be double, the
 *    check is an double check.  If the range is int, a message is logged, and
 *    the check is done using int.  <----- NOTE NOTE
 *    If the variable type is string, the actual type is inferred from the
 *    range specification, and the range check is done using that type.
 *    Time values are handled (0:0:0).
 *
 * INVOKED
 *    status = cfp_range(id, name, addr, type);
 *
 * PARAMETERS IN
 *    FILE_ID id - par file to operate on
 *    char *name - name of variable to check
 *    char *addr - the value to be checked
 *    int type - the type of the variable
 *               CFP_BOOLEAN - ignored (PASS)
 *               CFP_INTEGER - do integer check
 *               CFP_DOUBLE  - do double check
 *               CFP_STRING  - hope we can guess type from range field
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if check was OK or range field was messed up
 *               FAIL if check definitely failed
 *
 * GLOBALS
 *    none
 *
 ***************************************************************************!*/

PASSFAIL
cfp_range(id, name, addr, type)
    FILE_ID id;			  /* instance handle */
    char *name;			  /* identifier to look for */
    char *addr;			  /* value to check */
    int type;			  /* addrs actual type */
{
    FileEntryType *FileEntry;	  /* pointer to file data */
    LineType *Line;		  /* pointer to a line in this file */

    if ((name == (char *) NULL) || (addr == (char *) NULL)) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_range:%d) null parameter passed in", __LINE__);
        return(FAIL);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_range) start on %s", name);

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_range:%d) CheckID failed for %d", __LINE__, id);
        return(FAIL);
    }

    if ((Line = cfp_FindName(FileEntry, name, FALSE)) == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_range:%d) name '%s' not found for '%s'",
		  __LINE__, name, FileEntry->FilePath);
	return(FAIL);
    }

    if (Line->TypeCode != VariableLine) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_range:%d) range check of non-variable '%s' in '%s'",
		  __LINE__, Line->NameString, FileEntry->FilePath);
        return(PASS);
    }

    if (Line->RangeInfo != (RangeInfoType *) NULL) {
	switch (type) {
	  case CFP_INTEGER:
	  case CFP_DOUBLE:
	  case CFP_STRING:
	    if (cfp_CheckRange(Line, type, addr) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_range:%d) '%s' out of range for '%s'",
			  __LINE__, addr, Line->NameString);
		return(FAIL);
	    }
	    break;

	  default:
	    /* don't know what to do with strings, booleans don't care */
	    break;
	}
    }

    cfht_log(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_range) end, OK");

    return(PASS);
}

/*!**********************************************************************
 *
 *   cfp_rangemess()
 *
 * DESCRIPTION
 *    Given a variable name, format a readable message indicating the range
 *    for that variable (to be used for an error message window most likely).
 *    The error message is returned in a static area that will be overwritten
 *    on the next call.
 *
 * INVOKED
 *    Message = cfp_rangemess(id, name);
 *
 * PARAMETERS IN
 *    FILE_ID id - par file to operate on
 *    char *name - name of variable to generate message for
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    char * - a string containing a readable form of the range
 *             if any error occurs, an appropriate message is stuffed here
 *
 * GLOBALS
 *    none
 *
 **********************************************************************!*/

char *
cfp_rangemess Q((FILE_ID id, char *name), (id, name), FILE_ID id; char *name;)
{
    FileEntryType *FileEntry;	  /* pointer to file data */
    LineType *Line;		  /* pointer to a line in this file */
    static char Message[CFP_STRSIZE + 1]; /* area for returned message */

    if (name == (char *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_rangemess:%d) null parameter passed in", __LINE__);
        return("Null parameter in call");
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_rangemess) start on %s", name);

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_rangemess:%d) CheckID failed for %d", __LINE__, id);
        return("Bad par file identifier");
    }

    if ((Line = cfp_FindName(FileEntry, name, FALSE)) == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_rangemess:%d) name '%s' not found for '%s'",
		  __LINE__, name, FileEntry->FilePath);
	(void) sprintf(Message, "Name '%s' not found in '%s'", name,
		       FileEntry->FilePath);
	return(Message);
    }

    if (Line->TypeCode != VariableLine) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_rangemess:%d) message for non-variable '%s' in '%s'",
		  __LINE__, Line->NameString, FileEntry->FilePath);
	(void) sprintf(Message,
		       "'%s' is not a variable and has no range in '%s'",
		       Line->NameString, FileEntry->FilePath);
        return(Message);
    }

    if (Line->RangeValue == (char *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_rangemess:%d) no range for '%s' in '%s'",
		  __LINE__, Line->NameString, FileEntry->FilePath);
	(void) sprintf(Message,
		       "No range field for '%s' in '%s'",
		       Line->NameString, FileEntry->FilePath);
	return(Message);
    }

    if (Line->RangeInfo == (RangeInfoType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_rangemess:%d) bad range '%s' for '%s' in '%s'",
		  __LINE__, Line->RangeValue, Line->NameString,
		  FileEntry->FilePath);
	(void) sprintf(Message, "Bad range '%s' for '%s' in '%s'",
		       Line->RangeValue, Line->NameString,
		       FileEntry->FilePath);
	return(Message);
    }

    if (cfp_BuildRangeMessage(Line, Message) != PASS) {
	(void) sprintf(Message,
		       "Invalid parameters in range message call for '%s' in '%s'",
		       Line->NameString, FileEntry->FilePath);
	return(Message);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_rangemess) end, returning '%s'", Message);

    return(Message);
}
