/* Copyright (C) 1997-2000    Canada-France-Hawaii Telescope Corp.       */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING         */
/*!****************************************************************************
 * FILE cfp_reread.c
 * $Id: cfp_reread.c,v 1.2 2000/08/15 06:43:03 thomas Exp $
 * $Locker:  $
 *
 * ROUTINES
 *
 *   cfp_reread()
 *
 * DESCRIPTION
 *    Clear out the old copy of a file, then read the file back into memory,
 *    updating the memory copy to match the disk copy.  The file is opened and
 *    closed in this routine; it is not left hanging open.  The raw data is
 *    parsed out into individual lines (via another routine).
 *
 *    This is a slightly lower level version of cfp_getf .  Here the same file
 *    id is used.
 *
 * INVOKED
 *    Status = cfp_reread(id);
 *
 * PARAMETERS IN
 *    FILE_ID id - internal file identifier
 *
 * PARAMETERS OUT
 *    FILE_ID id - memory data base updated from file
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if we could read it back again
 *               FAIL if there was a problem
 *
 * GLOBALS
 *    FileEntryType FileEntryTable - array of open file entries
 *    int errno - global error indicator
 *
 * HISTORY
 * who          when            what
 * ---------    ------------    ----------------------------------------------
 * jnt          10-Sep-97       stolen from cfp_init and cfp_finish
 *
 * $Log: cfp_reread.c,v $
 * Revision 1.2  2000/08/15 06:43:03  thomas
 * Removed unused variables
 *
 * Revision 1.1  1997/12/10 12:19:18  thomas
 * Original check in
 *
 **************************************************************************!*/

#include <string.h>		  /* for string defines */
#include <unistd.h>		  /* read() */
#include <fcntl.h>		  /* for O_RDWR */
#include <sys/stat.h>		  /* for file size and modification time */
#include <errno.h>		  /* for error codes and errno defines */
#include <stdlib.h>		  /* for malloc */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */
#include "cfp_internal.h"	  /* cfp private defines */

RCSID("$Id: cfp_reread.c,v 1.2 2000/08/15 06:43:03 thomas Exp $");

PASSFAIL cfp_reread(FILE_ID id)
{
    FileEntryType *FileEntry;	  /* description of current file */
    int fd;			  /* file descriptor for file */
    struct stat statbuf;	  /* for stat(2)ing file */
    size_t FileSize;		  /* file size claimed by stat */
    char *FileData;		  /* pointer to a copy of the file */

    cfht_log(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_reread) start");

    /**********************
     * verify id is valid *
     **********************/

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_reread:%d) CheckID failed on %d", __LINE__, id);
        errno = ENOTTY;
        return FAIL;
    }

    /*********************
     * free up resources *
     *********************/

    cfp_FreeLineSpace(FileEntry);
    FileEntry->InUseCode = NotInUse;
    if (FileEntry->HashTable != (HashTableType) NULL) {
	free(FileEntry->HashTable);
	FileEntry->HashTable = (HashTableType) NULL;
    }

    /*************************
     * read the file back in *
     *************************/

    /* get data space */
    if (stat(FileEntry->FilePath, &statbuf) < 0) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_reread:%d) stat failed on '%s' (%s)",
		  __LINE__, FileEntry->FilePath, strerror(errno));
        return FAIL;
    }
    FileSize = (size_t) statbuf.st_size;
    if (FileSize == 0) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY, "(cfp_reread:%d) '%s' zero size",
		  __LINE__, FileEntry->FilePath);
	errno = ENOTTY;
	return FAIL;
    }
    FileEntry->LastModified = statbuf.st_mtime;
    if (!(FileData = (char *) malloc((size_t) (FileSize + 1)))) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_reread:%d) malloc of %d bytes for '%s' failed",
                  __LINE__, FileSize + 1, FileEntry->FilePath);
        return FAIL;
    }

    /* get data */
    if ((fd = open(FileEntry->FilePath, O_RDONLY)) < 0) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_reread:%d) open failed for '%s' (%s)",
		  __LINE__, FileEntry->FilePath, strerror(errno));
        free((void *) FileData);
        return FAIL;
    }
    if ((size_t)read(fd, (void *) FileData, FileSize) != FileSize) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_reread:%d) read of '%s' failed (%s)",
		  __LINE__, FileEntry->FilePath, strerror(errno));
        (void) close(fd);
        free((void *) FileData);
        return FAIL;
    }
    (void) close(fd);

    /* null terminate */
    FileData[FileSize] = '\0';

    /* verify proper termination */
    if (FileData[FileSize - 1] != '\n') {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_reread:%d) '%s' corrupt, no newline at end",
		  __LINE__, FileEntry->FilePath);
        free((void *) FileData);
        errno = ENOTTY;
        return FAIL;
    }

    /* rebuild the line linked list */
    FileEntry->LineCount = 0;
    FileEntry->Lines = (LineType *) NULL;
    if (cfp_InitLines(FileEntry, FileData) != PASS) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_reread:%d) InitLines failed on '%s'",
		  __LINE__, FileEntry->FilePath);
	free((void *) FileData);
	return FAIL;
    }

    /* now we are sure all is OK, say it is in use again */
    FileEntry->InUseCode = InUse;
    free((void *) FileData);

    cfht_log(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_reread) end");

    return PASS;
}
