/* Copyright (C) 1991, 94, 95    Canada-France-Hawaii Telescope Corp.    */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */

/*!**********************************************************************
 * FILE
 * cfp_rmline.c
 * $Header: cfp_rmline.c,v 2.4 96/11/19 22:51:21 isani Exp $
 * $Locker:  $
 *
 * ROUTINES
 *   cfp_rmline
 * 
 * DESCRIPTION
 *    delete a line from a par file
 *    This removes all evidence of a variable or constant definition line
 *    from a file.  (Should it do aliases too??).
 *
 * INVOKED
 *    status = cfp_rmline(id, key);
 *
 * PARAMETERS IN
 *    FILE_ID id - file instance to delete a line from
 *    char *Key - name of constant/variable to delete
 *
 * PARAMETERS OUT
 *    FILE_ID id - updated line data to reflect deleted line
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if line gone, FAIL if not
 *
 * GLOBALS
 *    none
 *
 * $Log:	cfp_rmline.c,v $
 * Revision 2.4  96/11/19  22:51:21  isani
 * cleaned up rcsid
 * 
 * Revision 2.3  1996/01/25  17:27:00  thomas
 * Clean up symbol linked list and hash table on removal
 *
 * Revision 2.2  94/12/21  12:44:52  12:44:52  thomas (Jim Thomas)
 * Added "not just a lookup" flag to cfp_FindName call
 * 
 * Revision 2.1  94/12/14  17:07:25  17:07:25  thomas (Jim Thomas)
 * Added #ifdef __LINT__ around RCS headers
 * Removed include of sys/types.h, stdlib.h
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h, added cfp_internal.h
 * Changed CFHT_ERROR to CFHT_LOGONLY
 * Rewrote to use version 2 routines and actually work :-)
 * 
 **********************************************************************!*/

#include <string.h>		  /* for string handling defines */
#include <errno.h>		  /* for error codes and errno define */
#include <stdlib.h>		  /* free() */
#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */
#include "cfp_internal.h"	  /*   and internals */

RCSID("$Header: cfp_rmline.c,v 2.4 96/11/19 22:51:21 isani Exp $");

PASSFAIL
cfp_rmline(id, Key)
    FILE_ID id;			  /* instance handle */
    char *Key;			  /* name for soon to disappear line */
{
    FileEntryType *FileEntry;	  /* pointer to data on this file */
    LineType *Line;		  /* pointer to line for this name */
    LineType *PreviousLine;	  /* scan pointer to look for predecessor */
    unsigned int HashIndex;	  /* symbol index in hash table */

    if (Key == (char *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_rmline:%d) given null pointer", __LINE__);
	return(FAIL);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
              "(cfp_rmline) start (Key = '%s')", Key);

    if ((FileEntry = cfp_CheckID(id)) == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_rmline:%d) cfp_CheckID failed", __LINE__);
        return(FAIL);
    }

    if ((Line = cfp_FindName(FileEntry, Key, FALSE)) == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_rmline:%d) Key '%s' not found for '%s'",
		  __LINE__, Key, FileEntry->FilePath);
	return(FAIL);
    }

    if ((Line->TypeCode != ConstantLine) && (Line->TypeCode != VariableLine)) {
	cfht_logv(CFHT_LIBCFP, CFHT_ERROR,
		  "(cfp_rmline:%d) Jim made a bad choice for how this worked",
		  __LINE__);
	cfht_logv(CFHT_LIBCFP, CFHT_ERROR,
		  "(...)  delete type %d line %s for %s",
		  Line->TypeCode, Line->NameString, FileEntry->FilePath);
	return(FAIL);
    }

    /* scan through lines find out which one to delete */
    PreviousLine = FileEntry->Lines;
    if (PreviousLine == Line)
      FileEntry->Lines = Line->Next;
    else {
	while (PreviousLine->Next != Line)
	  PreviousLine = PreviousLine->Next;
	PreviousLine->Next = Line->Next;
    }

    /* clean up symbol linked list */
    PreviousLine = FileEntry->FirstWithName;
    if (PreviousLine == Line)
      FileEntry->FirstWithName = Line->NextSymbol;
    else {
	while (PreviousLine->NextSymbol != Line)
	  PreviousLine = PreviousLine->NextSymbol;
	PreviousLine->NextSymbol = Line->NextSymbol;
    }

    /* clean up hash table */
    HashIndex = cfp_HashIt(FileEntry, Line->NameString);
    if (FileEntry->HashTable[HashIndex] == Line) {
	FileEntry->HashTable[HashIndex] = Line->NextHash;
    } else {
	PreviousLine = FileEntry->HashTable[HashIndex];
	while (PreviousLine->NextHash != Line)
	  PreviousLine = PreviousLine->NextHash;
	PreviousLine->NextHash = Line->NextHash;
    }

    /* if it was first real name, clean up search start pointer */
    if (Line == FileEntry->FirstWithName)
      FileEntry->FirstWithName = Line->NextSymbol;
    
    /* now we can delete Line, it is unlinked */
    free((void *) Line->NameString);
    cfp_FreeVariableSpace(Line);
    free((void *) Line);

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_rmline) end");

    return (PASS);
}
