/* Copyright (C) 1994-2001    Canada-France-Hawaii Telescope Corp.       */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING         */
/***************************************************************************
 * FILE
 * cfp_util.c
 * $Id: cfp_util.c,v 2.13 2003/03/09 20:41:59 isani Exp $
 * $Locker:  $
 *
 * Various support routines for the config file parser.  Most of these
 * would not be called by an application program.
 *
 * ROUTINES
 *
 * cfp_chid()
 *    Checks an id for validity and returns a pointer to a block containing
 *    the file name and path.  On error, returns NULL.
 *    This is a holdover from version 1.
 *
 * cfp_CheckID()
 *    Internal version of cfp_chid().  It checks the given id for validity
 *    and returns a pointer to the file entry for the id.  On error, it
 *    returns NULL.
 *
 * cfp_FreeVariableSpace
 *    free all space allocated for the variable part of a line
 *
 * cfp_FreeLineSpace
 *    free all space allocated for the lines in a file
 *
 * cfp_InitLines()
 *    Allocates the memory needed to hold in memory the lines from the
 *    par file, and initializes associated structures.  This routine causes
 *    file data to be parsed out into line data structures.
 *
 * cfp_GetEnvValue()
 *    Get an environment variable.  This routine takes a string, prepends
 *    to it the name of the current par file (without ".," or ".par"),
 *    and then returns a pointer to the value of that environment variable.
 *    The length of the string plus the length of the par file name must
 *    be less than CFP_STRSIZE - 1 characters.  NULL is returned on error
 *    or if no such variable found.
 *
 * cfp_PutEnvValue()
 *    Set an environment variable.  This routine takes a string, prepends
 *    to it the name of the current par file (without ".," or ".par"),
 *    and then updates the value of that environment variable.
 *    The length of the string plus the length of the par file name must
 *    be less than CFP_STRSIZE - 1 characters.  FAIL
 *    is returned on error or if no such variable found.
 *
 * cfp_FindName()
 *    look up a variable by name in the line data base for the specified file
 *    the file must already be set up (cfp_init)
 *    the name can be a partial match as long as it is unique
 *    handle alias lookups silently
 *    return a pointer to the line if found, NULL if not
 *
 * cfp_CopyToCurrent()
 *    replace the current value with a new value
 *
 * cfp_GetItem()
 *    given a item index (e.g., CFP_CurrentIndex), return that item from
 *    the specified line
 *
 * cfp_PutItem()
 *    given a item index (e.g., CFP_CurrentIndex), update that item in
 *    the specified line
 *
 * cfp_RebuildFile
 *    reconstitute the file contents from the line information
 *
 *    combine the variable definitions back into lines and the lines back
 *    into a buffer that can be written to a file
 *
 * cfp_ConvertFromString()
 *    This does a type conversion from a string into a requested type
 *    (BOOLEAN, int, double, or string) and copies the result into
 *    a provided destination space.
 *
 * cfp_CheckRange()
 *    Given that a range field is specified, check the given string
 *    to make sure it is within the bounds specified by the min-max field.
 *    This checks CFP_INTEGER and CFP_DOUBLE values directly and CFP_STRING
 *    values by inference (if it looks like an integer use that, if it
 *    looks like a float or time use that, otherwise PASS).
 * 
 * HISTORY
 * who          when            what
 * ---------    ------------    ----------------------------------------------
 * jab          10 Feb 1988     Original coding
 * jrw          16 Feb 1991     check for null pointers
 *
 * $Log: cfp_util.c,v $
 * Revision 2.13  2003/03/09 20:41:59  isani
 * Fix crash that happened if CommentString was NULL.
 *
 * Revision 2.12  2001/10/14  20:19:22  isani
 * Linux keeps DBL_MIN DBL_MAX in float.h, not math.h apparently.
 * 
 * Revision 2.11  2001/08/23  03:52:49  thomas
 * Added [] to LegalNameSymbols
 * NOTE: this means a space is required in ranges with, e.g., "[ 0, %foo[a] ]"
 * 
 * Revision 2.10  2000/02/10 13:13:19  isani
 * Changed size_t's to unisigned int's.
 *
 * Revision 2.9  1998/07/02 20:08:20  thomas
 * Turned on EatBogusComments and made code to do it work
 * Fixed two function definitions and changed return format
 *
 * Revision 2.8  96/11/19  22:52:14  22:52:14  isani (Sidik Isani)
 * cleaned up rcsid
 * moved some sun hacks to "cfht.h"
 * commented out unused variable LineType* ThisLine
 * 
 * Revision 2.7  96/05/16  09:50:18  09:50:18  thomas (Jim Thomas)
 * Fixed HashIt bug on 1 character name
 *
 * Revision 2.6  96/01/26  17:57:22  17:57:22  thomas (Jim Thomas)
 * Tweaked hash function a bit
 * 
 * Revision 2.5  96/01/25  17:30:31  17:30:31  thomas (Jim Thomas)
 * Added code to build symbol linked list and hash table
 * Changed name lookup to try hash table first, then linear search
 * Changed alias loop detection to use counter instead of flag bits
 * 
 * Revision 2.4  95/03/21  18:48:41  18:48:41  thomas (Jim Thomas)
 * Fix non-fitting string sizes in error message
 * 
 * Revision 2.3  95/02/27  20:15:34  20:15:34  thomas (Jim Thomas)
 * Took out m and i mode specifiers
 * Moved range field parsing to file parse time
 * Added editing of values based on partial match to int/double type format
 * 
 * Revision 2.2  94/12/21  12:46:28  12:46:28  thomas (Jim Thomas)
 * Added third argument to cfp_FindName for look up without LOGONLY on fail
 * FALSE==just to lookup, TRUE==real search
 * 
 * Revision 2.1  94/12/14  17:26:18  17:26:18  thomas (Jim Thomas)
 * Complete rewrite for cfp version 2
 * 
 * Revision 1.4  93/03/10  13:02:03  13:02:03  evans
 * Revision 1.3  92/11/30  08:32:41  08:32:41  steve (Steven Smith)
 * Revision 1.2  92/05/22  15:05:51  15:05:51  steve (Steven Smith)
 * Revision 1.1  91/09/24  17:38:11  17:38:11  steve (Steven S Smith)
 *
 ****************************************************************************/

#include <stdio.h>		  /* for sprintf() */
#include <string.h>		  /* for string handling defines */
#include <ctype.h>		  /* for isdigit() */
#include <errno.h>		  /* for error codes and errno define */
#include <stdlib.h>		  /* for getenv, strtod, and malloc */
#include <limits.h>		  /* for INT_MIN and INT_MAX */
#ifdef LINUX
#include <float.h>		  /* for DBL_MIN and DBL_MAX */
#else
#include <math.h>		  /* for DBL_MIN and DBL_MAX */
#endif

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* configuration file parsing defines */
#include "cfp_internal.h"	  /*  and internal definitions */

RCSID("$Header: /cfht/src/pegasus/libcfht/RCS/cfp_util.c,v 2.13 2003/03/09 20:41:59 isani Exp $");

#define TabWidth 8		  /* NOTE - code assumes this is a power of 2 */
#define EatBogusComments TRUE	  /* define as TRUE after everything uses
				   * version 2 cfp and extra space is not
				   * needed for variables that get larger */

/*
 * Philosophical design thoughts that need recording somewhere :-)
 *
 * 1.  All strings are stored internally "dequoted".  I.e., when they are read
 * in from a file, any quotes are removed, all escaped characters are converted
 * to their "normal" format (e.g., \" -> " , \n -> a real newline, etc.), and
 * the string appears as the user sees it.  For example, in a file something
 * entered as 4.5" is stored as "4.5\"".  During program execution the string
 * is stored as 4.5" .  When the file is written out, each string is checked
 * for any characters requiring escaping.  If there are none, the string is
 * written to the file without quotes.  For variable definition strings,
 * commas, hashes, spaces, and right braces also require that the string be
 * quoted.  For constant definitions, if the value was quoted at input it will
 * be quoted again on output; otherwise hashes, backslashes, and newlines are
 * escaped in place.  The first time a file is read in and then written out
 * again, this process may change the form slightly.  Any changes done will be
 * backwards compatible with cfp version 1 as it exists (there are currently no
 * strange " in files).
 *
 * 2.  Null strings are considered different from missing values.  So "" in
 * the current field generates a zero length string that is present and should
 * not be replaced by the default or supplied value.  (This only matters in
 * command line argument parsing.)
 */

/*
 * symbols that are legal in a name
 */
static char LegalNameSymbols[] =
"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789%&._$+[]";

/*
 * constant strings (i.e., foo = afls slkd) are handled differently
 */

/*
 * symbols that are legal in a variable string (i.e., one of <mode>,
 * <current>, etc.  (Note: space is not, :; were equivalent to comma)
 */
static char LegalVariableSymbols[] =
"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789%&.-_~$+*/";

/*
 * strerror
 *
 * Sun doesn't have one :-(
 *
 */

#if defined(sun) && !defined(__svr4__)
extern int sys_nerr;		  /* the man pages lie */
extern char *sys_errlist[];	  /*   these aren't in errno.h */

char *
strerror Q((int ErrorNo), (ErrorNo), int ErrorNo;)
{
    if ((ErrorNo < 0) || (ErrorNo > sys_nerr))
      return "Error number out of range";
    else
      return sys_errlist[ErrorNo];
    /*NOTREACHED*/
}
#endif

/*!**********************************************************************
 *
 *   cfp_chid()
 *
 * DESCRIPTION
 *    Checks an id for validity and returns a pointer to a block containing
 *    the file name and path.  On error, returns NULL.
 *    This is a holdover from version 1.
 *
 * INVOKED
 *    instp = cfh_chid(id);
 *
 * PARAMETERS IN
 *    FILE_ID id - par file id
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    struct cfp_inst * - pointer to instance of the id, or NULL on error
 *
 * GLOBALS
 *    FileEntryType FileEntryTable[] - open file table
 *
 ********************************************************************!*/

struct cfp_inst *		  /* return: fast pointer */
cfp_chid Q((FILE_ID id), (id), FILE_ID id;)
{
    FileEntryType *FileEntry;	  /* pointer to private file entry */

    /* public file instance array */
    static struct cfp_inst cfp_inst_table[CFP_MAXINST + 1];
    struct cfp_inst *instp;	  /* pointer to public file instance */

    if (id < 0 || id > CFP_MAXINST) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_chid:%d) id (%d) out of range", __LINE__, id);
        errno = ENOTTY;
        return (struct cfp_inst *) NULL;
    }

    FileEntry = &FileEntryTable[id];
    instp = &cfp_inst_table[id];

    if (FileEntry->InUseCode != InUse) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_chid:%d) id (%d) appears uninitialized", __LINE__, id);
        errno = ENOTTY;
        return (struct cfp_inst *) NULL;
    }

    instp->id = FileEntry->FileID;
    strcpy(instp->file, FileEntry->FileName);
    strcpy(instp->path, FileEntry->FilePath);

    return instp;
}

/*!**********************************************************************
 *
 *   cfp_CheckID()
 *
 * DESCRIPTION
 *    Internal version of cfp_chid().  It checks the given id for validity
 *    and returns a pointer to the file entry for the id.  On error, it
 *    returns NULL.
 *
 * INVOKED
 *    FileEntry = cfp_CheckID(id);
 *
 * PARAMETERS IN
 *    FILE_ID id - par file id
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    FileEntryType * - pointer to file entry for this id, NULL if bad id
 *
 * GLOBALS
 *    FileEntryType FileEntryTable[] - open file table
 *
 *********************************************************************!*/

FileEntryType *
cfp_CheckID Q((FILE_ID id), (id), FILE_ID id;)
{
    FileEntryType *FileEntry;	  /* pointer to private file entry */

    if (id < 0 || id > CFP_MAXINST) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_CheckID:%d) id (%d) out of range", __LINE__, id);
        errno = ENOTTY;
        return (FileEntryType *) NULL;
    }

    FileEntry = &FileEntryTable[id];

    if (FileEntry->InUseCode != InUse) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_CheckID:%d) id (%d) appears uninitialized",
		  __LINE__, id);
        errno = ENOTTY;
        return (FileEntryType *) NULL;
    }

    return FileEntry;
}

/***********************************************************************
 *
 *   ShortString
 *
 * DESCRIPTION
 *    to protect calls to cfht_logv from long strings
 *    call this from the logv call instead of putting a string pointer
 *    creates possibly shortened string terminated by smallest of
 *       string length
 *       newline position
 *       CFP_STRSIZE
 *    returns pointer to static string clobbered on next call
 *
 * INVOKED
 *    cfht_logv(..., "... %s ...", ..., ShortString(String), ...);
 *
 * PARAMETERS IN
 *    char *String - pointer to string to be length checked
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    char * - pointer to not too long string
 *
 * GLOBALS
 *    none
 *
 ***********************************************************************/

static char *
ShortString Q((char *String), (String), char *String;)
{
    static char Line[CFP_STRSIZE + 1]; /* local buffer for piece of the line */
    char *End;			  /* pointer to newline, if any */
    if (String == (char *) NULL)
      Line[0] = '\0';
    else {
	strncpy(Line, String, CFP_STRSIZE);
	Line[CFP_STRSIZE] = '\0';
	if ((End = strchr(Line, '\n')) != (char *) NULL)
	  *End = '\0';
    }
    return Line;
}

/*****************************************************************************
 *
 * DumpLine
 *
 * DESCRIPTION
 *    dump the rest of the current input line to the logfile for debugging
 *    but only a line's worth if it's too long
 *
 * INVOKED
 *    DumpLine(Input);
 *
 * PARAMETERS IN
 *    char *Input - pointer into the line
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    none
 * 
 * GLOBALS
 *    log file updated
 *
 *****************************************************************************/

static void
DumpLine Q((char *Input), (Input), char *Input;)
{
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(...) next text is '%s'", ShortString(Input));
    return;
}

/*****************************************************************************
 *
 * cfp_SkipWhiteSpace
 *
 * DESCRIPTION
 *    skip spaces and/or tabs maintaing a column count
 *
 * INVOKED
 *    cfp_SkipWhiteSpace(&Input, &Column);
 *
 * PARAMETERS IN
 *    char **Input - address of pointer to current spot on line
 *    unsigned int *Column - address of current column
 *
 * PARAMETERS OUT
 *    char **Input - updated current spot on line after white space
 *    unsigned int *Column - updated current column after white space
 *
 * FUNCTION RETURNS
 *    none
 * 
 * GLOBALS
 *    none
 * 
 *****************************************************************************/

static void
cfp_SkipWhiteSpace Q((char **Input, unsigned int *Column),
		     (Input, Column),
		     char **Input; unsigned int *Column;)
{
	while ((**Input == ' ') || (**Input == '\t')) {
	    if (**Input == ' ') {
		(*Column)++;
	    }
	    else {		  /* get to next TAB stop */
		*Column |= (TabWidth) - 1;
		(*Column)++;
	    }
	    (*Input)++;
	}
}

/***********************************************************************
 *
 *   Spanint()
 *
 * DESCRIPTION
 *    skip over a legal integer value in a string
 *    Recognize a legal integer and march a pointer past it.  This matches
 *    decimal and hex optionally signed values. It does not skip any leading or
 *    trailing spaces.  It does not do a range check.  It does not require
 *    that any digits actually be present.
 *
 * INVOKED
 *    Spanint(&Scan);
 *
 * PARAMETERS IN
 *    char **Scan - address of pointer to the string to be scanned
 *
 * PARAMETERS OUT
 *    char **Scan - updated pointer past an integer value
 *
 * FUNCTION RETURNS
 *    none
 *
 * GLOBALS
 *    none
 *
 ***********************************************************************/

/* legal numeric characters for approximate legality scan */
char Digits[] = "0123456789";
char Hexits[] = "0123456789abcdefABCDEF";

static void
Spanint Q((char **Scan), (Scan), char **Scan;)
{
    if ((**Scan == '+') || (**Scan == '-'))
      (*Scan)++;
    if ((**Scan == '0')
	&& ((*(*Scan + 1) == 'x') || (*(*Scan + 1) == 'X'))) {
	*Scan += 2;
	*Scan += strspn(*Scan, Hexits);
    }
    else
      *Scan += strspn(*Scan, Digits);

    return;
}

/***********************************************************************
 *
 *   Spandouble()
 *
 * DESCRIPTION
 *    skip over a legal double value in a string
 *    Recognize a legal double and march a pointer past it.  This matches
 *    normal floating point values plus times.  It does not skip any leading or
 *    trailing spaces.  It does not do a range check.  It does not require
 *    that any digits actually be present.
 *
 * INVOKED
 *    Spandouble(&Scan);
 *
 * PARAMETERS IN
 *    char **Scan - address of pointer to the string to be scanned
 *
 * PARAMETERS OUT
 *    char **Scan - updated pointer past a double value
 *
 * FUNCTION RETURNS
 *    none
 *
 * GLOBALS
 *    none
 *
 ***********************************************************************/

static void
Spandouble Q((char **Scan), (Scan), char **Scan;)
{
    if ((**Scan == '+') || (**Scan == '-'))
      (*Scan)++;
    *Scan += strspn(*Scan, Digits);
    if (**Scan == ':') {
	(*Scan)++;
	*Scan += strspn(*Scan, Digits);
	if (**Scan == ':') {
	    (*Scan)++;
	    *Scan += strspn(*Scan, Digits);
	}
	if (**Scan == '.') {
	    (*Scan)++;
	    *Scan += strspn(*Scan, Digits);
	}
    } else {
	if (**Scan == '.')
	  (*Scan)++;
	*Scan += strspn(*Scan, Digits);
	if ((**Scan == 'e') || **Scan == 'E') {
	    (*Scan)++;
	    if ((**Scan == '+') || (**Scan == '-'))
	      (*Scan)++;
	    if (isdigit(**Scan)) {
		*Scan += strspn(*Scan, Digits);
	    }
	}
    }

    return;
}

/***********************************************************************
 *
 *   FreeRangeInfo
 *
 * DESCRIPTION
 *    free up any memory allocated to the line range information
 *
 * INVOKED
 *    FreeRangeInfo(&Line);
 *
 * PARAMETERS IN
 *    LineType *Line - the line needed cleaning
 *
 * PARAMETERS OUT
 *    LineType *Line - the cleaned line
 *
 * FUNCTION RETURNS
 *    none
 *
 * GLOBALS
 *    none
 *
 ***********************************************************************/

static void
FreeRangeInfo Q((LineType * Line), (Line), LineType *Line;)
{
    if (Line->RangeInfo != (RangeInfoType *) NULL) {
	if (Line->RangeInfo->MinString != (char *) NULL) {
	    free((void *) Line->RangeInfo->MinString);
	    Line->RangeInfo->MinString = (char *) NULL;
	}
	if (Line->RangeInfo->MaxString != (char *) NULL) {
	    free((void *) Line->RangeInfo->MaxString);
	    Line->RangeInfo->MaxString = (char *) NULL;
	}
	free((void *) Line->RangeInfo);
	Line->RangeInfo = (RangeInfoType *) NULL;
    }
    return;
}

/*!**********************************************************************
 *
 *   cfp_FreeVariableSpace
 *
 * DESCRIPTION
 *    free all space allocated for the variable part of a line
 *    basically all string parts except the name
 *
 * INVOKED
 *    cfp_FreeVariableSpace(Line);
 *
 * PARAMETERS IN
 *    LineType *Line - pointer to line to be cleared out
 *
 * PARAMETERS OUT
 *    LineType *Line - any previously malloc'ed fields free'ed
 *
 * FUNCTION RETURNS
 *    none
 *
 * GLOBALS
 *    none
 *
 **********************************************************************!*/

void
cfp_FreeVariableSpace Q((LineType *Line), (Line), LineType *Line;)
{
    if (Line->OriginalValue != (char *) NULL) {
	free((void *) Line->OriginalValue);
	Line->OriginalValue = (char *) NULL;
	Line->OriginalValueLength = 0;
    }
    if (Line->ModeString != (char *) NULL) {
	free((void *) Line->ModeString);
	Line->ModeString = (char *) NULL;
    }
    if (Line->CurrentValue != (char *) NULL) {
	free((void *) Line->CurrentValue);
	Line->CurrentValue = (char *) NULL;
	Line->CurrentValueLength = 0;
    }
    if (Line->DefaultValue != (char *) NULL) {
	free((void *) Line->DefaultValue);
	Line->DefaultValue = (char *) NULL;
    }
    if (Line->SuppliedValue != (char *) NULL) {
	free((void *) Line->SuppliedValue);
	Line->SuppliedValue = (char *) NULL;
    }
    if (Line->RangeValue != (char *) NULL) {
	free((void *) Line->RangeValue);
	Line->RangeValue = (char *) NULL;
    }
    FreeRangeInfo(Line);

    return;
}

/*!**********************************************************************
 *
 *   cfp_FreeLineSpace
 *
 * DESCRIPTION
 *    free all space allocated for the lines in a file
 *
 * INVOKED
 *    cfp_FreeLineSpace(FileEntry);
 *
 * PARAMETERS IN
 *    FileEntryType *FileEntry - pointer to file description block
 *
 * PARAMETERS OUT
 *    FileEntryType *FileEntry - line list now (LineType *) NULL
 *
 * FUNCTION RETURNS
 *    none
 *
 * GLOBALS
 *    none
 *
 **********************************************************************!*/

void
cfp_FreeLineSpace Q((FileEntryType *FileEntry),
		    (FileEntry),
		    FileEntryType *FileEntry;)
{
    LineType *Current;		  /* pointer to current line in scan */
    LineType *Next;		  /* pointer to next line */

    Current = FileEntry->Lines;
    while (Current != (LineType *) NULL) {
	if (Current->NameString != (char *) NULL)
	  free((void *) Current->NameString);
	cfp_FreeVariableSpace(Current);
	if (Current->CommentString != (char *) NULL)
	  free((void *) Current->CommentString);
	Next = Current->Next;
	free((void *) Current);
	Current = Next;
    }

    FileEntry->Lines = (LineType *) NULL;
    return;
}

/*****************************************************************************
 *
 * cfp_ScanName
 *
 * DESCRIPTION
 *    extract a name
 *
 *    starting at the first character of what is supposed to be a name field,
 *    scan across legal name symbols
 *    copy the name string into a malloc'ed area and point to it for the caller
 *
 * INVOKED
 *    status = cfp_ScanName(&Input, &NameString);
 *
 * PARAMETERS IN
 *    char **Input - address of pointer to line, assumed to be at first of name
 *
 * PARAMETERS OUT
 *    char **Input - updated pointer to line, now at first character after name
 *    char **Name - address of pointer to name string, filled in with pointer
 *                  to malloc'ed area containing a copy of the name
 *
 * FUNCTION RETURNS
 *    PASS if OK, FAIL if error (0 length name or malloc failed)
 *
 * GLOBALS
 *    none
 *
 *****************************************************************************/

static PASSFAIL
cfp_ScanName Q((char **Input, char **Name),
	       (Input, Name),
	       char **Input; char **Name;)
{
    size_t NameLength;		  /* length of name string seen */

    NameLength = strspn(*Input, LegalNameSymbols);
    if (NameLength == 0) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		 "(cfp_ScanName:%d) zero length name field", __LINE__);
	errno = ENOTTY;
	return FAIL;
    }
    if (NameLength > CFP_STRSIZE) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_ScanName:%d) too long (%d) a name '%s'",
		  __LINE__, NameLength, ShortString(*Input));
	errno = ENOTTY;
	return FAIL;
    }
    if ((*Name = (char *) malloc(NameLength + 1)) == (char *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_ScanName:%d) malloc of %d failed",
		  __LINE__, NameLength);
	errno = ENOTTY;
	return FAIL;
    }
    strncpy(*Name, *Input, NameLength);
    (*Name)[NameLength] = '\0';
    (*Input) += NameLength;

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_ScanName:%d) returned '%s'", __LINE__, ShortString(*Name));
#endif

    return PASS;
}

/*****************************************************************************
 *
 *   cfp_ScanQuotedString
 *
 * DESCRIPTION
 *    scan past a quoted string, updating the column, copying the string itself
 *    handling all escaped characters
 *
 *    called with the line pointer at the starting quote
 *    scan looking for an escape, the closing quote, or end of line
 *    for an escape, if the next character is legit, convert it
 *    otherwise copy the character
 *
 * INVOKED
 *    status = cfp_ScanQuotedString(&From, &Column, &To);
 *
 * PARAMETERS IN
 *    char **From - address of pointer to current line,
 *                  assumed to be pointing to the starting quote
 *    unsigned int *Column - pointer to current column number
 *    char **To - address of pointer to area to receive dequoted,
 *                deescaped string
 *
 * PARAMETERS OUT
 *    char **From - updated pointer to current line, following closing quote
 *    unsigned int *Column - updated column number
 *    char **To - updated pointer to copy
 *
 * FUNCTION RETURNS
 *    PASS if OK, FAIL if some problem
 *
 * GLOBALS
 *    none
 *
 *****************************************************************************/

static PASSFAIL
cfp_ScanQuotedString Q((char **From, unsigned int *Column, char **To),
		       (From, Column, To),
		       char **From; unsigned int *Column; char **To;)
{
    char Quote;			  /* the quote character to match */

#ifdef DEBUGDEBUG
    char *StartSource = *From;	  /* the beginning of the string in input */
    char Line[CFP_STRSIZE + 1];	  /* a copy of the string to dump */
#endif
    char *StartResult = *To;	  /* the beginning of the string in output */

    Quote = *(*From)++;
    (*Column)++;

    while (**From != Quote) {
	switch (**From) {

	  case '\n':
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		     "(cfp_ScanQuotedString:%d) no matching quote on line",
		      __LINE__);
	    errno = ENOTTY;
	    return FAIL;
	    /*NOTREACHED*/
	    break;

	  case '\t':
	    *Column |= (TabWidth) - 1; /* get to next TAB stop */
	    (*Column)++;
	    *(*To)++ = *(*From)++;
	    break;

	  case EscapeCharacter:
	    (*From)++;
	    (*Column)++;
	    switch (**From) {

	      case QuoteCharacter:
		*(*To)++ = QuoteCharacter;
		break;

#if QuoteCharacter != '\"'
	      case '\"':
		*(*To)++ = '\"';
		break;
#endif

#if QuoteCharacter != '\''
	      case '\'':
		*(*To)++ = '\'';
		break;
#endif

	      case EscapeCharacter:
		*(*To)++ = EscapeCharacter;
		break;

#if EscapeCharacter != '\\'
	      case '\\':
		*(*To)++ = '\\';
		break;
#endif

	      case 't':
		*(*To)++ = '\t';  /* make it be an actual tab from now on */
		break;

	      case 'n':
		*(*To)++ = '\n';
		break;

	      case 'r':
		*(*To)++ = '\r';
		break;

	      case CommentCharacter:
		*(*To)++ = CommentCharacter;
		break;

	      case '\n':
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_ScanQuotedString:%d) attempt to escape newline",
			  __LINE__);
		**To = '\0';
		cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
			  "(...) on line '%s'",
			  ShortString(StartResult));
		errno = ENOTTY;
		return FAIL;
		/*NOTREACHED*/
		break;

	      default:
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_ScanQuotedString:%d) unknown escape '\\%c' ignored",
			  __LINE__, **From);
		DumpLine(*From);
		break;
	    }
	    (*From)++;
	    (*Column)++;
	    break;

	  default:
	    *(*To)++ = *(*From)++;
	    (*Column)++;
	    break;
	}
    }

    (*From)++;			  /* skip trailing quote */
    (*Column)++;
    **To = '\0';

#ifdef DEBUGDEBUG
    strncpy(Line, StartSource, CFP_STRSIZE);
    Line[CFP_STRSIZE] = '\0';
    if ((*From - StartSource) < CFP_STRSIZE)
        Line[*From - StartSource] = '\0';
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_ScanQuotedString:%d) changed '%s'",
	      __LINE__, ShortString(Line));
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(....)                 to '%s'",
	      ShortString(StartResult));
#endif

    return PASS;
}

/*****************************************************************************
 *
 *   cfp_ClearModeBits
 *
 * DESCRIPTION
 *    set all mode bits to FALSE
 *
 * INVOKED
 *    cfp_ClearModeBits(Line);
 *
 * PARAMETERS IN
 *    none
 *
 * PARAMETERS OUT
 *    LineType *Line - line with mode bits cleared
 *
 * FUNCTION RETURNS
 *    none
 *
 * GLOBALS
 *    none
 *****************************************************************************/

static void
cfp_ClearModeBits Q((LineType *Line), (Line), LineType *Line;)
{
    Line->ValueNotAllowed = FALSE;
    Line->ValueOptional = FALSE;
    Line->UseDefault = FALSE;
    Line->UseSupplied = FALSE;
    Line->UpdateValue = FALSE;
    return;
}

/*****************************************************************************
 *
 * cfp_ModeStringToBits
 *
 * DESCRIPTION
 *    scan a mode string and set to TRUE any mode bits named in the string
 *    if ClearFirst is TRUE, zap the current settings first (do nothing
 *    if FALSE)
 *
 * INVOKED
 *    status = cfp_GetModeBits(ModeString, ClearFirst, Line);
 *
 * PARAMETERS IN
 *    char *ModeString - the mode in string form
 *    BooleanType ClearFirst - TRUE -> set all bits to FALSE first
 *                             FALSE -> leave current settings
 *
 * PARAMETERS OUT
 *    LineType *Line - line with mode bits and string updated
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if no problems, FAIL if malloc or bad mode letters
 *
 * GLOBALS
 *    none
 * 
 *****************************************************************************/

static PASSFAIL
cfp_ModeStringToBits Q((char *ModeString, BooleanType ClearFirst,
			LineType *Line),
		       (ModeString, ClearFirst, Line),
		       char *ModeString; BooleanType ClearFirst;
		       LineType *Line;)
{
    unsigned int Index;		  /* index into mode string */
    size_t StringLength;	  /* length of mode string */

    if (ClearFirst == TRUE)
      cfp_ClearModeBits(Line);

    StringLength = strlen(ModeString);
    for (Index = 0; Index < StringLength; Index++) {
	switch (ModeString[Index]) {

	  case ModeNeverCharacter:
	    if (Line->ValueNotAllowed == TRUE)
	        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_ModeStringToBits:%d) double %c mode for '%s'",
			  __LINE__, ModeNeverCharacter, Line->NameString);
	    if (Line->ValueOptional == TRUE)
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_ModeStringToBits:%d) both %c and %c for '%s'",
			  __LINE__, ModeNeverCharacter, ModeOptionalCharacter);
	    Line->ValueNotAllowed = TRUE;
	    break;

	  case ModeOptionalCharacter:
	    if (Line->ValueOptional == TRUE)
	        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_ModeStringToBits:%d) double %c mode for '%s'",
			  __LINE__, ModeOptionalCharacter, Line->NameString);
	    if (Line->ValueNotAllowed == TRUE)
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_ModeStringToBits:%d) both %c and %c for '%s'",
			  __LINE__, ModeNeverCharacter, ModeOptionalCharacter);
	    Line->ValueOptional = TRUE;
	    break;

	  case ModeDefaultCharacter:
	    if (Line->UseDefault == TRUE)
	        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_ModeStringToBits:%d) double %c mode for '%s'",
			  __LINE__, ModeDefaultCharacter, Line->NameString);
	    Line->UseDefault = TRUE;
	    break;

	  case ModeSuppliedCharacter:
	    if (Line->UseSupplied == TRUE)
	        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_ModeStringToBits:%d) double %c mode for '%s'",
			  __LINE__, ModeSuppliedCharacter, Line->NameString);
	    Line->UseSupplied = TRUE;
	    break;

	  case ModeUpdateCharacter:
	    if (Line->UpdateValue == TRUE)
	        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_ModeStringToBits:%d) double %c mode for '%s'",
			  __LINE__, ModeUpdateCharacter, Line->NameString);
	    Line->UpdateValue = TRUE;
	    break;

	  default:
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_ModeStringToBits:%d) unrecognized '%c' option for '%s'",
		      __LINE__, ModeString[Index], Line->NameString);
	    errno = ENOTTY;
	    return FAIL;
	    /*NOTREACHED*/
	    break;
	}
    }

    /* replace the mode string so the order is always consistent */
    if (Line->ModeString != (char *) NULL)
      free((void *) Line->ModeString);
    /* +5 in following for slack in case someone doesn't update symbol */
    if ((Line->ModeString = (char *) malloc(NumberOfModes + 1 + 5))
	== (char *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_ModeStringToBits:%d) rebuild malloc %d failed",
		  __LINE__, NumberOfModes + 1 + 5);
	errno = EIO;
	return FAIL;
    }
    Index = 0;
    /* most use alphabetical order now, so keep that */
    if (Line->UseDefault == TRUE)
      Line->ModeString[Index++] = ModeDefaultCharacter;
    if (Line->ValueNotAllowed == TRUE)
      Line->ModeString[Index++] = ModeNeverCharacter;
    if (Line->ValueOptional == TRUE)
      Line->ModeString[Index++] = ModeOptionalCharacter;
    if (Line->UpdateValue == TRUE)
      Line->ModeString[Index++] = ModeUpdateCharacter;
    if (Line->UseSupplied == TRUE)
      Line->ModeString[Index++] = ModeSuppliedCharacter;
    Line->ModeString[Index] = '\0';

    return PASS;
}

/***********************************************************************
 *
 *   ParseRange()
 *
 * DESCRIPTION
 *    parse out the pieces of a range specifier
 *    Given a string containing a range specifer (i.e., not a type - not int
 *    or double), figure out whether the ends are open or closed, separate out
 *    the min and max strings, if any, and record whether there were any.
 *    If not already determined, decide whether the range is for an int or a
 *    double.  Build a range check structure for the specifier.
 *
 *    Returns FAIL if there is any problem with the specifier; otherwise,
 *    PASS.
 *
 *    See cfp_CheckRange for the specifier format.
 *
 * INVOKED
 *    Status = ParseRange(&Line, RangeString, TypeKnown);
 *
 * PARAMETERS IN
 *    LineType *Line - pointer to the line, RangeValue already set up,
 *                     RangeInfo already allocated
 *    char *RangeString - current pointer into range string
 *                        (past int/double)
 *    BooleanType TypeKnown - TRUE if RangeInfo->Type already valid
 *                            FALSE if not
 *
 * PARAMETERS OUT
 *    LineType *Line - RangeInfo filled in (MinString, MinIndirect,
 *                     MaxString, MaxIndirect, Range, and Type)
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if ranges OK, FAIL if not
 *
 * GLOBALS
 *    none
 *
 ***********************************************************************/

static PASSFAIL
ParseRange Q((LineType *Line, char *RangeString, BooleanType TypeKnown),
	     (Line, RangeString, TypeKnown),
	     LineType *Line; char *RangeString; BooleanType TypeKnown;)
{
    RangeInfoType *Info;	  /* local copy of Line->RangeInfo */
    char *Start;		  /* left end of a range string */
    char *Scan;			  /* for scan through range string */
    BooleanType Isint;		  /* TRUE until character not [-+0-9] seen */
				  /* then FALSE */
    /* LeftClosed and RightClosed must have same sense as min/max_inclusive */
    BooleanType LeftClosed;	  /* TRUE if [ , FALSE if ( */
    BooleanType RightClosed;	  /* TRUE if ] , FALSE if ) */

    Isint = TRUE;		  /* assume so until shown otherwise */
    LeftClosed = TRUE;
    RightClosed = TRUE;
    
    if (Line == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(ParseRange:%d) null line argument",
		  __LINE__);
	return FAIL;
    }
    if (Line->RangeValue == (char *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(ParseRange:%d) no range string to parse for '%s'",
		  __LINE__, Line->NameString);
	return FAIL;
    }
    if (Line->RangeInfo == (RangeInfoType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(ParseRange:%d) range structure not set up for '%s'",
		  __LINE__, Line->NameString);
    }

    Info = Line->RangeInfo;
    Scan = RangeString;
    Scan += strspn(Scan, " \t");  /* not supposed to be any, but ... */

    if (*Scan == LeftClosedCharacter)
      LeftClosed = TRUE;
    else {
	if (*Scan == LeftOpenCharacter)
	  LeftClosed = FALSE;
	else {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "((cfp)ParseRange:%d) '%s' bad left end '%c'",
		      __LINE__, ShortString(Scan), *Scan);
	    return FAIL;
	}
    }
    Scan++;
    Scan += strspn(Scan, " \t");

    /* check minimum, see if an indirect, an int, a double, or nothing */
    Start = Scan;
    if (*Scan == AliasIndirectCharacter) {
	Scan++;
	if (cfp_ScanName(&Scan, &Info->MinString) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(ParseRange:%d) bad min indirect for '%s' in '%s'",
		      __LINE__, Line->NameString, ShortString(Start));
	    return FAIL;
	}
	Info->MinIndirect = TRUE;
	/* might check KnownType, but let's default to int */
    } else {
	Spanint(&Scan);
	if ((*Scan == '.') || (*Scan == ':')
	    || (*Scan == 'e') || (*Scan == 'E')) {
	    Spandouble(&Scan);	  /* min was double, scan rest of it */
	    Isint = FALSE;
	}
    
	if (Start == Scan) {
	    /* Info->MinSeen = FALSE; */ /* already done */
	} else {
	    if ((Info->MinString = (char *) malloc((size_t) (Scan - Start + 1)))
		== (char *) NULL) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(ParseRange:%d) malloc %d for min failed for '%s'",
			  __LINE__, (Scan - Start + 1), Line->NameString);
		return FAIL;
	    }
	    (void) strncpy(Info->MinString, Start, (size_t) (Scan - Start));
	    Info->MinString[Scan - Start] = '\0';
	    Info->MinSeen = TRUE;
	}
    }

    Scan += strspn(Scan, " \t");
    if (*Scan == ',') {
	Scan++;
	Scan += strspn(Scan, " \t");

	/* check maximum value, see what that is */
	Start = Scan;
	if (*Scan == AliasIndirectCharacter) {
	    Scan++;
	    if (cfp_ScanName(&Scan, &Info->MaxString) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(ParseRange:%d) bad max indirect for '%s' in '%s'",
			  __LINE__, Line->NameString, ShortString(Start));
		return FAIL;
	    }
	    Info->MaxIndirect = TRUE;
	    /* might check KnownType, but let's default to int */
	} else {
	    Spanint(&Scan);
	    if ((*Scan == '.') || (*Scan == ':')
		|| (*Scan == 'e') || (*Scan == 'E')) {
		Spandouble(&Scan); /* max was double, scan rest of it */
		Isint = FALSE;
	    }
	
	    if (Start == Scan) {
		/* Info->MaxSeen = FALSE; */ /* already done */
	    } else {
		if ((Info->MaxString
		     = (char *) malloc((size_t) (Scan - Start + 1)))
		    == (char *) NULL) {
		    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			      "(ParseRange:%d) malloc %d for max failed for '%s'",
			      __LINE__, (Scan - Start + 1), Line->NameString);
		    return FAIL;
		}
		(void) strncpy(Info->MaxString, Start, (size_t) (Scan - Start));
		Info->MaxString[Scan - Start] = '\0';
		Info->MaxSeen = TRUE;
	    }
	}

	Scan += strspn(Scan, " \t");
    } else {
	/* Info->MaxSeen = FALSE; */ /* already done */
    }

    /* OK, check right end  -- note any early nul above gets caught here */
    /* also any junk following any number in the range */
    if (*Scan == RightClosedCharacter)
      RightClosed = TRUE;
    else {
	if (*Scan == RightOpenCharacter)
	  RightClosed = FALSE;
	else {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "((cfp)ParseRange:%d) '%s' junk or bad right end '%c'",
		      __LINE__, ShortString(Scan), *Scan);
	    return FAIL;
	}
    }

    /* make sure no junk after */
    Scan++;
    Scan += strspn(Scan, " \t");
    if (*Scan != '\0') {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "((cfp)ParseRange:%d) junk '%s' after range",
		  __LINE__, ShortString(Scan));
	return FAIL;
    }

    if (TypeKnown == FALSE) {	  /* if we don't know yet, fill in guess */
	if (Isint == TRUE) {
	    Info->Type = CFP_INTEGER;
	} else {
	    Info->Type = CFP_DOUBLE;
	}
    }

    /* for values we've seen, the range check struct can now be filled in */
    if (Info->Type == CFP_INTEGER) {
	Info->intRangeLeft = LeftClosed;
	if (Info->MinSeen == TRUE) {
	    if (cfht_int(Info->MinString, &Info->intRangeMin,
			 (cfht_range_t) NULL) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(ParseRange:%d) int min convert of '%s' for '%s'",
			  __LINE__, ShortString(Info->MinString),
			  Line->NameString);
		return FAIL;
	    }
	}
	Info->intRangeRight = RightClosed;
	if (Info->MaxSeen == TRUE) {
	    if (cfht_int(Info->MaxString, &Info->intRangeMax,
			 (cfht_range_t) NULL) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(ParseRange:%d) int max convert of '%s' for '%s'",
			  __LINE__, ShortString(Info->MaxString),
			  Line->NameString);
		return FAIL;
	    }
	}
    } else {			  /* it's double */
	Info->doubleRangeLeft = LeftClosed;
	if (Info->MinSeen == TRUE) {
	    if (cfht_double(Info->MinString, &Info->doubleRangeMin,
			 (cfht_range_t) NULL) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(ParseRange:%d) double min convert of '%s' for '%s'",
			  __LINE__, ShortString(Info->MinString),
			  Line->NameString);
		return FAIL;
	    }
	}
	Info->doubleRangeRight = RightClosed;
	if (Info->MaxSeen == TRUE) {
	    if (cfht_double(Info->MaxString, &Info->doubleRangeMax,
			 (cfht_range_t) NULL) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(ParseRange:%d) double max convert of '%s' for '%s'",
			  __LINE__, ShortString(Info->MaxString),
			  Line->NameString);
		return FAIL;
	    }
	}
    }

    return PASS;
}

/*!**********************************************************************
 *
 *   cfp_BuildRangeValue
 *
 * DESCRIPTION
 *    Convert a range string into an internal range info structure.
 *    This parses a range string specification into its type and value and
 *    builds the internal information needed to do the actual range check.
 *    This attempts to get as much of this done as possible - the only case
 *    that is not finished is the indirect reference case (e.g., [1,%xmax])
 *    where the line for xmax may not even be read in yet when this is called
 *    and the value used really wants to be current anyway.
 *
 *    See cfp_CheckRange for the format.
 *
 * INVOKED
 *    status = cfp_BuildRangeValue(&Line);
 *
 * PARAMETERS IN
 *    LineType *Line - the line with Line->RangeValue containing the string
 *
 * PARAMETERS OUT
 *    LineType *Line - the line with Line->RangeInfo built and filled in
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if OK, FAIL if a problem (range struct cleared out)
 *
 * GLOBALS
 *    none
 *
 **********************************************************************!*/

/* type identifier name table for cfp_match */

static char *TypeNames[] = {
    "int",
    "double",
    "time",
    (char *) NULL
};

/* matching table of type codes, must be in same order as TypeNames */

static int TypeCodes[] = {
    CFP_INTEGER,
    CFP_DOUBLE,
    CFP_DOUBLE,
    0
};

#define MinusInf "-inf"		  /* string for no lower limit */
#define PlusInf "+inf"		  /* string for no upper limit */

static PASSFAIL
cfp_BuildRangeValue Q((LineType *Line), (Line), LineType *Line;)
{
    RangeInfoType *Info;	  /* a copy of the structure pointer */
    size_t Length;		  /* length of identifier if one there */
    char TypeName[CFP_STRSIZE + 1]; /* identifier */
    int TypeIndex;		  /* index of identifier in TypeNames */
    char *Scan;			  /* scan pointer into range string */
    BooleanType TypeKnown;	  /* TRUE if type identifier given */
				  /* FALSE if none there, have to guess */

    if (Line == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_BuildRangeValue:%d) null line pointer",
		  __LINE__);
	return FAIL;
    }
    if (Line->RangeValue == (char *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_BuildRangeValue:%d) no range string for '%s'",
		  __LINE__, Line->NameString);
	return FAIL;
    }

    Scan = Line->RangeValue;
    Scan += strspn(Scan, " \t");
    if (*Scan == '\0') {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_BuildRangeValue:%d) null range string for '%s'",
		  __LINE__, Line->NameString);
	return FAIL;
    }

    /* allocate a range information structure to put our answer into */
    if ((Info = (RangeInfoType *) malloc(sizeof(RangeInfoType)))
	== (RangeInfoType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_BuildRangeValue:%d) malloc %d for info failed",
		  __LINE__, sizeof(RangeInfoType));
	return FAIL;
    }
    Line->RangeInfo = Info;
    Info->Type = CFP_INTEGER;
    Info->MinString = (char *) NULL;
    Info->MinSeen = FALSE;
    Info->MinIndirect = FALSE;
    Info->MaxString = (char *) NULL;
    Info->MaxSeen = FALSE;
    Info->MaxIndirect = FALSE;
    TypeKnown = FALSE;

    /* see if there is a type identifier given */
    if (isalpha(*Scan)) {
	Length = 0;
	while ((Length < CFP_STRSIZE) && (isalnum(*Scan))) {
	    TypeName[Length++] = *Scan++;
	}
	/* don't bother to check Length > CFP_STRSIZE, won't be in table */
	TypeName[Length] = '\0';
	if ((TypeIndex = cfp_match(TypeName, TypeNames)) < 0) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_BuildRangeValue:%d) unknown type '%s' for '%s'",
		      __LINE__, TypeName, Line->NameString);
	    FreeRangeInfo(Line);
	    return FAIL;
	}
	Info->Type = TypeCodes[TypeIndex];
	TypeKnown = TRUE;
	Scan += strspn(Scan, " \t");
    }

    /* now see if there is a range specifier (let parse catch bad strings) */
    if (*Scan != '\0') {
	if (ParseRange(Line, Scan, TypeKnown) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_BuildRangeValue:%d) bad range '%s' for '%s'",
		      __LINE__, ShortString(Line->RangeValue),
		      Line->NameString);
	    FreeRangeInfo(Line);
	    return FAIL;
	}
    }

    /* if either only a type given or a min or max was missing, fill in inf */
        if (Info->Type == CFP_INTEGER) {
	if ((Info->MinSeen == FALSE) && (Info->MinIndirect == FALSE)) {
	    Info->intRangeLeft = TRUE;
	    Info->intRangeMin = INT_MIN;
	    if ((Info->MinString = (char *) malloc(strlen(MinusInf) + 1))
		== (char *) NULL) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_BuildRangeValue:%d) malloc %d for -inf for '%s'",
			  __LINE__, (strlen(MinusInf) + 1), Line->NameString);
		FreeRangeInfo(Line);
		return FAIL;
	    }
	    strcpy(Info->MinString, MinusInf);
	}
	if ((Info->MaxSeen == FALSE) && (Info->MaxIndirect == FALSE)) {
	    Info->intRangeRight = TRUE;
	    Info->intRangeMax = INT_MAX;
	    if ((Info->MaxString = (char *) malloc(strlen(PlusInf) + 1))
		== (char *) NULL) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_BuildRangeValue:%d) malloc %d for +inf for '%s'",
			  __LINE__, (strlen(PlusInf) + 1), Line->NameString);
		FreeRangeInfo(Line);
		return FAIL;
	    }
	    strcpy(Info->MaxString, PlusInf);
	}
    } else {			  /* it's double */
	if ((Info->MinSeen == FALSE) && (Info->MinIndirect == FALSE)) {
	    Info->doubleRangeLeft = TRUE;
	    Info->doubleRangeMin = -DBL_MAX;
	    if ((Info->MinString = (char *) malloc(strlen(MinusInf) + 1))
		== (char *) NULL) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_BuildRangeValue:%d) malloc %d for -inf for '%s'",
			  __LINE__, (strlen(MinusInf) + 1), Line->NameString);
		FreeRangeInfo(Line);
		return FAIL;
	    }
	    strcpy(Info->MinString, MinusInf);
	}
	if ((Info->MaxSeen == FALSE) && (Info->MaxIndirect == FALSE)) {
	    Info->doubleRangeRight = TRUE;
	    Info->doubleRangeMax = DBL_MAX;
	    if ((Info->MaxString = (char *) malloc(strlen(PlusInf) + 1))
		== (char *) NULL) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_BuildRangeValue:%d) malloc %d for +inf for '%s'",
			  __LINE__, (strlen(PlusInf) + 1), Line->NameString);
		FreeRangeInfo(Line);
		return FAIL;
	    }
	    strcpy(Info->MaxString, PlusInf);
	}
    }

    return PASS;
}

/*****************************************************************************
 *
 *   cfp_DoConstantDef
 *
 * DESCRIPTION
 *    parse the right hand side of a line like <name> = <value>
 *
 *    called with the line pointer at the first character of the value
 *    scan off the value (handling quotes) stopping at a comment or end of line
 *    chop off any trailing spaces
 *    point the line entry to a copy of the value
 *
 * INVOKED
 *    status = cfp_DoConstantDef(&Input, &Column, Line);
 *
 * PARAMETERS IN
 *    char **Input - address of pointer to current line,
 *               assumed to be pointing to the first character of <value>
 *    unsigned int *Column - pointer to current column number
 *
 * PARAMETERS OUT
 *    char **Input - updated pointer to current line, should be looking at a
 *               comment or end of line
 *    unsigned int *Column - updated column number
 *    LineType *Line - OriginalValue field filled in with pointer to
 *                     a copy of the <value>
 *
 * FUNCTION RETURNS
 *    PASS if OK, FAIL if some problem
 *
 * GLOBALS
 *    none
 *
 *****************************************************************************/

static PASSFAIL
cfp_DoConstantDef Q((char **Input, unsigned int *Column, LineType *Line),
		    (Input, Column, Line),
		    char **Input; unsigned int *Column; LineType *Line;)
{
    size_t Index;		  /* index into the line */
    char *Current;		  /* pointer to internal copy of string */
    char *Trim;			  /* pointer to input for trailing blank trim */

    /* first reserve enough space to hold the string for sure (we may not need
     * quite this much if there's a comment or escaped characters, but this is
     * faster and guaranteed to be enough) */
    Index = strcspn(*Input, "\n");
    if ((Current = (char *) malloc(Index + 1)) == (char *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_DoConstantDef:%d) malloc of %d failed",
		  __LINE__, Index + 1);
	errno = ENOTTY;
	return FAIL;
    }
    Line->OriginalValue = Current;

    /* first see if whole thing (or at least first part ?) is quoted */
    if (**Input == QuoteCharacter) {
	if (cfp_ScanQuotedString(Input, Column, &Current) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_DoConstantDef:%d) bad quoted string after '%s'",
		      __LINE__, ShortString(Line->OriginalValue));
	    return FAIL;
	}
	Line->WasQuoted = TRUE;
    }

    /* now do rest of line (should this be else ?) */
    while ((**Input != '\n') && (**Input != CommentCharacter)) {
	switch (**Input) {

	  case EscapeCharacter:
	    (*Input)++;
	    (*Column)++;
	    switch (**Input) {

	      case QuoteCharacter:
		*Current++ = QuoteCharacter;
		break;

#if QuoteCharacter != '\"'
	      case '\"':
		*Current++ = '\"';
		break;
#endif

#if QuoteCharacter != '\''
	      case '\'':
		*Current++ = '\'';
		break;
#endif

	      case EscapeCharacter:
		*Current++ = EscapeCharacter;
		break;

#if EscapeCharacter != '\\'
	      case '\\':
		*Current++ = '\\';
		break;
#endif

	      case 't':
		*Current++ = '\t';  /* make it be an actual tab from now on */
		break;

	      case 'n':
		*Current++ = '\n';
		break;

	      case 'r':
		*Current++ = '\r';
		break;

	      case CommentCharacter:
		*Current++ = CommentCharacter;
		break;

	      case '\n':
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_DoConstantDef:%d) attempt to escape newline",
			  __LINE__);
		*Current = '\0';
		cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
			  "(...) on line '%s'",
			  ShortString(Line->OriginalValue));
		errno = ENOTTY;
		return FAIL;
		/*NOTREACHED*/
		break;

	      default:
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_DoConstantDef:%d) unknown escape '\\%c' ignored",
			  __LINE__, **Input);
		cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(...)  following %s",
			  ShortString(Line->OriginalValue));
		DumpLine(*Input);
		break;
	    }
	    (*Input)++;
	    (*Column)++;
	    break;

	  case '\t':
	    *Current++ = *(*Input)++;
	    *Column |= (TabWidth) - 1;
	    (*Column)++;
	    break;

	  default:
	    *Current++ = *(*Input)++;
	    (*Column)++;
	    break;
	}
    }
    *Current = '\0';

    if (strlen(Line->OriginalValue) > 0) {
	Trim = *Input - 1;
	while ((*Trim == ' ') || (*Trim == '\t')) {
	    *--Current = '\0';	  /* eat trailing spaces */
	    Trim--;
	}
    }

    Line->OriginalValueLength = strlen(Line->OriginalValue);

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_DoConstantDef:%d) got '%s = %s'",
	      __LINE__, Line->NameString, ShortString(Line->OriginalValue));
#endif

    return PASS;
}

/*****************************************************************************
 *
 *   cfp_ScanVariableString
 *
 * DESCRIPTION
 *    routine to scan off a single string from a variable definition
 *
 *    stops on space, tab, comma, (colon, semicolon,) }, #, newline
 *    handles quoted strings
 *    will eat white space before and after string
 *    allocates the space to put the string copy
 *
 *    A quoted string can be zero length.  Space for a nul will be allocated
 *    and To will point to that.  An unquoted string cannot be zero length.
 *    If there is no value present, To will contain (char *) NULL on return,
 *    and 0 will be returned.
 *
 * INVOKED
 *    length = cfp_ScanVariableString(&From, &Column, &To);
 *
 * PARAMETERS IN
 *    char **From - address of pointer to current position on line
 *    unsigned int *Column - address of current column
 *
 * PARAMETERS OUT
 *    char **From - updated pointer into line after string and any whitespace
 *    unsigned int *Column - updated current column
 *    char **To - address of a copy of the string w/o surrounding whitespace
 *                malloc'ed by this routine
 *
 * FUNCTION RETURNS
 *    int - length of the actual resulting string (i.e., without quotes or
 *          escapes or surrounding whitespace), possibly zero
 *          -1 if a malloc failed
 *
 * GLOBALS
 *    none
 *
 *****************************************************************************/

static int
cfp_ScanVariableString Q((char **From, unsigned int *Column, char **To),
			 (From, Column, To),
			 char **From; unsigned int *Column; char **To;)
{
    size_t StringLength;	  /* oversize or actual string length */
    char *QuoteCopy;		  /* pointer to oversize copy for quotes */
    char *QuoteString;		  /* current pointer into QuoteCopy */
    char *Copy;			  /* pointer to string copy */

    cfp_SkipWhiteSpace(From, Column);
    if (**From == QuoteCharacter) {
	/* start with more than enough space for string itself */
	StringLength = strcspn(*From, "\n");
	if ((QuoteCopy = (char *) malloc(StringLength + 1)) == (char *) NULL) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_ScanVariableString:%d) quote malloc %d bytes failed",
		      __LINE__, StringLength + 1);
	    errno = ENOTTY;
	    return -1;
	}
	QuoteString = QuoteCopy;
	if (cfp_ScanQuotedString(From, Column, &QuoteString) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		     "(cfp_ScanVariableString:%d) bad quoted string",
		      __LINE__);
	    return -1;
	}
	StringLength = strlen(QuoteCopy); /* now get actual size needed */
	if ((Copy = (char *) malloc(StringLength + 1)) == (char *) NULL) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_ScanVariableString:%d) shrink malloc %d bytes failed",
		      __LINE__, StringLength + 1);
	    errno = ENOTTY;
	    return -1;
	}
	strcpy(Copy, QuoteCopy);
	free(QuoteCopy);
    }
    else {
	StringLength = strspn(*From, LegalVariableSymbols);
	if (StringLength > 0) {
	    if ((Copy = (char *) malloc(StringLength + 1)) == (char *) NULL) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_ScanVariableString:%d) copy malloc %d failed",
			  __LINE__, StringLength + 1);
		errno = ENOTTY;
		return -1;
	    }
	    strncpy(Copy, *From, StringLength);
	    Copy[StringLength] = '\0';
	    *From += StringLength;
	    *Column += StringLength;
	} else {
	    Copy = (char *) NULL;
	}
    }
    cfp_SkipWhiteSpace(From, Column);

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_ScanVariableString:%d) got '%s'",
	      __LINE__, ShortString(Copy));
#endif

    *To = Copy;
    return (int) StringLength;
}

/*!***************************************************************************
 *
 *   cfp_DoVariableDef
 *
 * DESCRIPTION
 *    parse the right hand side of a line like <name> = { <mode>, <current> }
 *
 *    called with the line pointer at the {
 *    scan off the variable definition into the line entry block fields
 *
 * INVOKED
 *    status = cfp_DoVariableDef(&Input, &Column, Line);
 *
 * PARAMETERS IN
 *    char **Input - address of pointer to current line,
 *                   assumed to be pointing to the {
 *    unsigned int *Column - pointer to current column number
 *
 * PARAMETERS OUT
 *    char **Input - updated pointer to current line past }, should be looking
 *                   at a comment or end of line
 *    unsigned int *Column - updated column number
 *    LineType *Line - all the variable definition fields filled in
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if OK, FAIL if some problem
 *
 * GLOBALS
 *    none
 *
 ***************************************************************************!*/

PASSFAIL
cfp_DoVariableDef Q((char **Input, unsigned int *Column, LineType *Line),
		    (Input, Column, Line),
		    char **Input; unsigned int *Column; LineType *Line;)
{
    char *ModeLetters;		  /* any mode given */
    int StringLength;		  /* length of a scanned string (or -1) */

    if ((Input == (char **) NULL) || (Column == (unsigned int *) NULL)
	|| (Line == (LineType *) NULL)) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_DoVariableDef:%d) given null pointer", __LINE__);
	return FAIL;
    }

    if (**Input != VariableDefStartCharacter) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_DoVariableDef:%d) not looking at %c in '%s=%s'",
		  __LINE__, VariableDefStartCharacter, Line->NameString,
		  ShortString(*Input));
	return FAIL;
    }

    (*Input)++;			  /* skip { */
    (*Column)++;

    /* collect mode field, if any */
    cfp_SkipWhiteSpace(Input, Column);
    if ((**Input != ',') && (**Input != ';') && (**Input != ':')) {
	if ((StringLength =
	     cfp_ScanVariableString(Input, Column, &ModeLetters)) < 0) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		     "(cfp_DoVariableDef:%d) mode pickup failed",
		      __LINE__);
	    return FAIL;
	}
	if (StringLength > 0) {
	    if (cfp_ModeStringToBits(ModeLetters, TRUE, Line) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_DoVariableDef:%d) mode translate failed",
			  __LINE__);
		return FAIL;
	    }
	}
	if (ModeLetters != (char *) NULL)
	  free(ModeLetters);
    }

    /* collect current field, if any */
    if ((**Input == ',') || (**Input == ';') || (**Input == ':')) {
	(*Input)++;
	(*Column)++;
	if ((StringLength =
	     cfp_ScanVariableString(Input, Column, &Line->OriginalValue)) < 0) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		     "(cfp_DoVariableDef:%d) current pickup failed",
		      __LINE__);
	    return FAIL;
	}

	/* do not throw away present but zero length value, i.e., "" */

	Line->OriginalValueLength = StringLength;
	if (Line->OriginalValue != (char *) NULL) {
	    if ((Line->CurrentValue = (char *) malloc((unsigned)StringLength + 1))
		== (char *) NULL) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_DoVariableDef:%d) current malloc %d failed",
			  __LINE__, StringLength + 1);
		return FAIL;
	    }
	    strcpy(Line->CurrentValue, Line->OriginalValue);
	    Line->CurrentValueLength = StringLength;
	    Line->History = CFP_CURRENT;
	}
    }

    /* collect default field, if any */
    if ((**Input == ',') || (**Input == ';') || (**Input == ':')) {
	(*Input)++;
	(*Column)++;
	if ((StringLength =
	     cfp_ScanVariableString(Input, Column, &Line->DefaultValue)) < 0) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		     "(cfp_DoVariableDef:%d) default pickup failed",
		      __LINE__);
	    return FAIL;
	}

	/* do not throw away 0 length but present value, i.e., "" */
    }

    /* collect supplied field, if any */
    if ((**Input == ',') || (**Input == ';') || (**Input == ':')) {
	(*Input)++;
	(*Column)++;
	if ((StringLength =
	     cfp_ScanVariableString(Input, Column, &Line->SuppliedValue)) < 0) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		     "(cfp_DoVariableDef:%d) supplied pickup failed",
		      __LINE__);
	    return FAIL;
	}

	/* do not throw away 0 length but present value, i.e., "" */
    }

    /* collect range field, if any */
    if ((**Input == ',') || (**Input == ';') || (**Input == ':')) {
	(*Input)++;
	(*Column)++;
	if ((StringLength =
	     cfp_ScanVariableString(Input, Column, &Line->RangeValue)) < 0) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		     "(cfp_DoVariableDef:%d) range pickup failed",
		      __LINE__);
	    return FAIL;
	}

	/* throw away "" - but what if someone implements STRING ranges?? */
	if ((StringLength == 0) && (Line->RangeValue != (char *) NULL)) {
	    free(Line->RangeValue);
	    Line->RangeValue = (char *) NULL;
	}

	/* now convert range string into our structure */
	if (Line->RangeValue != (char *) NULL) {
	    if (cfp_BuildRangeValue(Line) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_DoVariableDef:%d) error building range for '%s', ignored",
			  __LINE__, Line->NameString);
	    }
	}
    }

    /* must be looking at the right brace at this point */
    if (**Input != VariableDefEndCharacter) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_DoVariableDef:%d) expecting '%c' at end for '%s'",
		  __LINE__, VariableDefEndCharacter, Line->NameString);
	DumpLine(*Input);
	return FAIL;
    }
    (*Input)++;
    (*Column)++;

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_DoVariableDef:%d) got definition for '%s'",
	      __LINE__, Line->NameString);
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(...)        mode = %c%c%c%c%c",
	      Line->ValueNotAllowed == TRUE ? ModeNeverCharacter : '-',
	      Line->ValueOptional == TRUE ? ModeOptionalCharacter : '-',
	      Line->UseDefault == TRUE ? ModeDefaultCharacter : '-',
	      Line->UseSupplied == TRUE ? ModeSuppliedCharacter : '-',
	      Line->UpdateValue == TRUE ? ModeUpdateCharacter : '-');
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(...)    original = %s",
	      Line->OriginalValue != (char *) NULL ?
	        ShortString(Line->OriginalValue) : "<NULL>");
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(...)     current = %s",
	      Line->CurrentValue != (char *) NULL ?
	        ShortString(Line->CurrentValue) : "<NULL>");
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(...)     default = %s",
	      Line->DefaultValue != (char *) NULL ?
	        ShortString(Line->DefaultValue) : "<NULL>");
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(...)    supplied = %s",
	      Line->SuppliedValue != (char *) NULL ?
	        ShortString(Line->SuppliedValue) : "<NULL>");
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(...)         range = %s",
	      Line->RangeValue != (char *) NULL ?
	        ShortString(Line->RangeValue) : "<NULL>");
#endif

    return PASS;
}

/*****************************************************************************
 *
 *   cfp_DoDefinitionLine
 *
 * DESCRIPTION
 *    the main part of cfp_InitLines - handle constant, variable, and alias
 *    lines
 *
 *    scan through one definition line, extracting all its information and
 *    filling in the line entry block with same
 *
 * INVOKED
 *    status = cfp_DoDefinitionLine(FileEntry, &Input, &Column, Line);
 *
 * PARAMETERS IN
 *    FileEntryType *FileEntry - pointer to file data (for duplicate name
 *                               search only)
 *    char **Input - address of pointer to current line,
 *               assumed to be pointing to the start of the name
 *    unsigned int *Column - pointer to current column number
 *
 * PARAMETERS OUT
 *    char **Input - updated pointer into line, pointing past definition,
 *                hopefully at spaces, #, or newline
 *    unsigned int *Column - updated column past definition
 *    LineType *Line - updated line entry block, filled in with definition
 *
 * FUNCTION RETURNS
 *    PASS if OK, FAIL if something went wrong
 *
 * GLOBALS
 *    none
 *
 *****************************************************************************/

static PASSFAIL
cfp_DoDefinitionLine Q((FileEntryType *FileEntry, char **Input,
			unsigned int *Column, LineType *Line),
		       (FileEntry, Input, Column, Line),
		       FileEntryType *FileEntry; char **Input;
		       unsigned int *Column; LineType *Line;)
{
    if (cfp_ScanName(Input, &Line->NameString) != PASS) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_DoDefinitionLine:%d) name scan", __LINE__);
	DumpLine(*Input);
	return FAIL;
    }
    Line->NameStringLength = strlen(Line->NameString);
    *Column += Line->NameStringLength;

    cfp_SkipWhiteSpace(Input, Column);
    Line->EqualsColumn = *Column;
    switch (**Input) {

      case ':':
	/* problem child with : instead of =, fix it */
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_DoDefinitionLine:%d) variable '%s' followed by : in '%s'",
		  __LINE__, Line->NameString, FileEntry->FilePath);
	DumpLine(*Input);
	cfht_log(CFHT_LIBCFP, CFHT_LOGONLY, "(...)  converting to =");

	/* NOTE NOTE NOTE - fall into following code - NOTE NOTE NOTE */

      case AssignmentCharacter:
	/* do variable or constant */
	(*Input)++;
	(*Column)++;
	cfp_SkipWhiteSpace(Input, Column);
	if (**Input == VariableDefStartCharacter) {
	    Line->TypeCode = VariableLine;
	    if (cfp_DoVariableDef(Input, Column, Line) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_DoDefinitionLine:%d) bad variable for '%s'",
			  __LINE__, Line->NameString);
		DumpLine(*Input);
		return FAIL;
	    }
	}
	else {
	    Line->TypeCode = ConstantLine;
	    if (cfp_DoConstantDef(Input, Column, Line) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_DoDefinitionLine:%d) bad constant for '%s'",
			  __LINE__, Line->NameString);
		DumpLine(*Input);
		return FAIL;
	    }
	}
	break;

      case AliasCharacter1:
	/* do alias */
	Line->TypeCode = AliasLine;
	(*Input)++;
	(*Column)++;
	if (**Input == AliasCharacter2) {
	    (*Input)++;
	    (*Column)++;
	    cfp_SkipWhiteSpace(Input, Column);
	    if (cfp_ScanName(Input, &Line->OriginalValue) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_DoDefinitionLine:%d) bad alias for '%s'",
			  __LINE__, Line->NameString);
		DumpLine(*Input);
		return FAIL;
	    }
	    Line->OriginalValueLength = strlen(Line->OriginalValue);
	    *Column += Line->OriginalValueLength;
#ifdef DEBUGDEBUG
	    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		      "(cfp_DoDefinitionLine:%d) got alias '%s -> %s'",
		      __LINE__, Line->NameString,
		      ShortString(Line->OriginalValue));
#endif
	}
	else {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_DoDefinitionLine:%d) no > after - for '%s'",
		      __LINE__, Line->NameString);
	    DumpLine(*Input);
	    return FAIL;
	}
	break;

      default:
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_DoDefinitionLine:%d) junk after name '%s' in '%s'",
		  __LINE__, Line->NameString, FileEntry->FilePath);
	DumpLine(*Input);
	cfht_log(CFHT_LIBCFP, CFHT_LOGONLY,
		 "(...)  hoping it is supposed to be a constant with =");
	Line->TypeCode = ConstantLine;
	if (cfp_DoConstantDef(Input, Column, Line) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_DoDefinitionLine:%d) bad constant for '%s'",
		      __LINE__, Line->NameString);
	    DumpLine(*Input);
	    return FAIL;
	}
	break;
    }

    return PASS;
}

/*****************************************************************************
 *
 *   cfp_CopyComment
 *
 * DESCRIPTION
 *    copy the rest of the line into the comment field of the line entry
 *
 * INVOKED
 *    status = cfp_CopyComment(&p, Line);
 *
 * PARAMETERS IN
 *    char **Input - address of pointer to # in current line
 *
 * PARAMETERS OUT
 *    char **Input - updated pointer now looking at newline
 *    LineType *Line - updated comment field pointer
 *
 * FUNCTION RETURNS
 *    PASS if OK, FAIL if malloc lost out
 *
 * GLOBALS
 *    none
 *
 *****************************************************************************/

static PASSFAIL
cfp_CopyComment Q((char **Input, LineType *Line),
		  (Input, Line),
		  char **Input;	LineType *Line;)
{
    size_t CommentLength;	  /* space needed for comment */

#if EatBogusComments
    CommentLength = strspn((*Input) + 1, " \t");
    if ((Line->TypeCode != CommentLine)
	&& (*((*Input) + CommentLength + 1) == '\n')) {
	*Input += CommentLength + 1;
	return PASS;
    }
#endif

    CommentLength = strcspn(*Input, "\n");
    if ((Line->CommentString = (char *) malloc(CommentLength + 1))
	== NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_CopyComment:%d) malloc of %d failed",
		  __LINE__, CommentLength + 1);
	return FAIL;
    }
    strncpy(Line->CommentString, *Input, CommentLength);
    Line->CommentString[CommentLength] = '\0';
    *Input += CommentLength;
    return PASS;
}

/*!**********************************************************************
 *
 *   cfp_HashIt
 *
 * DESCRIPTION
 *   hash function for par file symbols
 *
 * INVOKED
 *   Index = cfp_HashIt(FileEntry, Name);
 *
 * PARAMETERS IN
 *   FileEntryType *FileEntry - par file descriptor entry
 *   char *Name - symbol to be hashed
 *
 * PARAMETERS OUT
 *   none
 *
 * FUNCTION RETURNS
 *   unsigned int - index into this file's hash table for this symbol
 *
 * GLOBALS
 *   none
 *
 **********************************************************************!*/

unsigned int
cfp_HashIt Q((FileEntryType *FileEntry, char *Name), (FileEntry, Name),
	     FileEntryType *FileEntry; char *Name;)
{
    unsigned int HashIndex;
    unsigned int Front;
    unsigned int Back;

    /* generate sum of characters working from both ends */
    Front = 0;
    Back = strlen(Name);
    if (Back == 0) {
	return 0;		  /* ? */
    }
    HashIndex = 0;
    Back--;			  /* convert to index */
    if ((Back & 1) == 0) {	  /* if odd length, get oddball first */
	HashIndex = Name[Back] % FileEntry->HashSize;
	if (Back > 0) {
	    Back--;
	}
    }
    while (Front < Back) {
#define HashWindow 64		  /* that's what Sedgwick says :-) */
	HashIndex = (HashIndex * HashWindow) + Name[Front++];
	HashIndex = (HashIndex * HashWindow) + Name[Back--];
	HashIndex = HashIndex % FileEntry->HashSize;
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_HashIt:%d) '%s' hashed to %d",
	      __LINE__, Name, HashIndex);
#endif

    return HashIndex;
}

/***********************************************************************
 *
 *   cfp_BuildHash
 *
 * DESCRIPTION
 *   Set up a hash table for symbol lookup in this file.
 *
 * INVOKED
 *   status = cfp_BuildHash(FileEntry);
 *
 * PARAMETERS IN
 *   FileEntryType *FileEntry - file entry with lines read in
 *
 * PARAMETERS OUT
 *   FileEntryType *FileEntry - file entry with hash table pointer added
 *                              lines have hash linked list entry filled in
 *
 * FUNCTION RETURNS
 *   PASSFAIL - PASS if OK, FAIL if malloc didn't or symbol count outrageous
 *
 * GLOBALS
 *   none
 *
 ***********************************************************************/

/* pick the lowest prime from this arbitrary set */
static unsigned int Primes[] = {
    29, 103, 211, 401, 601, 809, 1021, 2011, 10099, 100003, 0
};

static PASSFAIL
cfp_BuildHash Q((FileEntryType *FileEntry), (FileEntry),
		FileEntryType *FileEntry;)
{
    unsigned int PrimeIndex;	  /* index into Primes table */
    unsigned int Prime;		  /* chosen prime number for hash table size */
    unsigned int HashIndex;	  /* index into hash table for current symbol */
    unsigned int Collisions;	  /* count of hash collisions */
    LineType *Line;		  /* current line in scan through lines */
    LineType *PreviousLine;	  /* previous line in scan through hash list */

    /* quit gracefully if no symbols */
    if (FileEntry->SymbolCount == 0) {
	FileEntry->HashSize = 0;
	FileEntry->HashTable = (HashTableType) NULL;
	return PASS;
    }
    
    /* find a prime for the size of the hash array */
    PrimeIndex = 0;
    while ((Primes[PrimeIndex] != 0)
	   && ((FileEntry->SymbolCount * 2) > Primes[PrimeIndex])) {
	PrimeIndex++;
    }
    if (Primes[PrimeIndex] == 0) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_BuildHash:%d) too many symbols (%d) in '%s'",
		  __LINE__, FileEntry->SymbolCount, FileEntry->FilePath);
	return FAIL;
    }
    Prime = Primes[PrimeIndex];
    FileEntry->HashSize = Prime;

    /* allocate the hash table */
    if ((FileEntry->HashTable
	 = (HashTableType) malloc(Prime * sizeof(HashEntryType)))
	== (HashTableType) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_BuildHash:%d) hash table malloc %d failed",
		  __LINE__, Prime * sizeof(HashEntryType));
	return FAIL;
    }
    for (HashIndex = 0; HashIndex < Prime; HashIndex++) {
	FileEntry->HashTable[HashIndex] = (HashEntryType) NULL;
    }

    /* build the hash table entries */
    Line = FileEntry->FirstWithName;
    Collisions = 0;
    while (Line != (LineType *) NULL) {
	HashIndex = cfp_HashIt(FileEntry, Line->NameString);
	if (FileEntry->HashTable[HashIndex] == (HashEntryType) NULL) {
	    FileEntry->HashTable[HashIndex] = Line;
	} else {
	    Collisions++;
	    PreviousLine = FileEntry->HashTable[HashIndex];
	    while ((strcmp(Line->NameString, PreviousLine->NameString) != 0)
		   && (PreviousLine->NextHash != (HashEntryType) NULL)) {
		PreviousLine = PreviousLine->NextHash;
	    }
	    if (PreviousLine->NextHash == (HashEntryType) NULL) {
		PreviousLine->NextHash = Line;
	    } else {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_BuildHash:%d) duplicate definition of '%s' in '%s' ignored",
			  __LINE__, Line->NameString, FileEntry->FileName);
	    }
	}
	Line = Line->NextSymbol;
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_BuildHash:%d) %d symbol%s, size %d, %d collision%s in '%s'",
	      __LINE__,
	      FileEntry->SymbolCount, (FileEntry->SymbolCount == 1) ? "" : "s",
	      FileEntry->HashSize,
	      Collisions, (Collisions == 1) ? "" : "s",
	      FileEntry->FileName);
#endif
    return PASS;
}

/*!**********************************************************************
 *
 *   cfp_InitLines()
 *
 * DESCRIPTION
 *    Allocates the memory needed to hold in memory the lines from the
 *    par file, and initializes associated structures.  This routine causes
 *    file data to be parsed out into line data structures.
 *
 * INVOKED
 *    status = cfp_InitLines(&FileEntry, FileData);
 *
 * PARAMETERS IN
 *    FileEntryType *FileEntry - pointer to file structure for this open file
 *    char *FileData - pointer to file contents
 *
 * PARAMETERS OUT
 *    FileEntryType *FileEntry - updated file entry, pointing to line structure
 *
 * FUNCTION RETURNS
 *    PASSFAIL - indicates completion status
 *
 * GLOBALS
 *    none
 *
 *********************************************************************!*/

PASSFAIL
cfp_InitLines Q((FileEntryType *FileEntry, char *FileData),
		(FileEntry, FileData),
		FileEntryType *FileEntry; char *FileData;)
{
    unsigned int Column;	  /* current line column for spacing */
    char *Input;		  /* scan pointer to input buffer */
    LineType *LastLine;		  /* last line entry built (tail pointer) */
    LineType *LastSymbol;	  /* last line entry with symbol */
    LineType *ThisLine;		  /* line entry being built */

    if (FileEntry == (FileEntryType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                 "(cfp_InitLines:%d) null instance pointer", __LINE__);
        return FAIL;
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_InitLines:%d) starting '%s'",
	      __LINE__, FileEntry->FilePath);
#endif

    Input = FileData;
    LastLine = (LineType *) NULL;
    LastSymbol = (LineType *) NULL;
    FileEntry->FirstWithName = (LineType *) NULL;
    FileEntry->SymbolCount = 0;
    FileEntry->HashTable = (HashTableType) NULL;

    while (*Input != '\0') {
	FileEntry->LineCount++;
	Column = 0;
	if ((ThisLine = (LineType *) malloc(sizeof(LineType)))
	    == (LineType *) NULL) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_InitLines:%d) malloc of %d failed for '%s'",
		      __LINE__, sizeof(LineType), FileEntry->FilePath);
	    cfp_FreeLineSpace(FileEntry);
	    errno = ENOTTY;
	    return FAIL;
	}
	/* purify complains if not zeroed 'cause of bit manipulation */
	(void) memset((void *) ThisLine, '\0', sizeof(LineType));

	if (LastLine == (LineType *) NULL) {
	    LastLine = ThisLine;
	    FileEntry->Lines = ThisLine;
	}
	else {
	    LastLine->Next = ThisLine;
	    LastLine = ThisLine;
	}
	ThisLine->Next = (LineType *) NULL;
	ThisLine->NextSymbol = (LineType *) NULL;
	ThisLine->NextHash = (LineType *) NULL;
	ThisLine->FileEntry = FileEntry;
	ThisLine->TypeCode = BlankLine;
	cfp_ClearModeBits(ThisLine);
	ThisLine->WasQuoted = FALSE;
	ThisLine->History = CFP_UNCHECKED;
	ThisLine->LeadingSpaceCount = 0;
	ThisLine->NameStringLength = 0;
	ThisLine->NameString = (char *) NULL;
	ThisLine->EqualsColumn = 0;
	ThisLine->OriginalValueLength = 0;
	ThisLine->OriginalValue = (char *) NULL;
	ThisLine->CurrentValueLength = 0;
	ThisLine->CurrentValue = (char *) NULL;
	ThisLine->ModeString = (char *) NULL;
	ThisLine->DefaultValue = (char *) NULL;
	ThisLine->SuppliedValue = (char *) NULL;
	ThisLine->RangeValue = (char *) NULL;
	ThisLine->RangeInfo = (RangeInfoType *) NULL;
	ThisLine->CommentString = (char *) NULL;
	ThisLine->CommentColumn = 0;

	cfp_SkipWhiteSpace(&Input, &Column);
	ThisLine->LeadingSpaceCount = Column;

	switch (*Input) {

	  case '\n':
	    ThisLine->LeadingSpaceCount = 0;
	    ThisLine->TypeCode = BlankLine;
	    break;

	  case CommentCharacter:
	    /* catch comment itself at end */
	    ThisLine->TypeCode = CommentLine;
	    break;

	  default:
	    if (cfp_DoDefinitionLine(FileEntry, &Input, &Column, ThisLine)
		!= PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_InitLines:%d) def failed for line %d in '%s'",
			  __LINE__, FileEntry->LineCount, FileEntry->FilePath);
		cfp_FreeLineSpace(FileEntry);
		return FAIL;
	    }

	    /* build symbol list */
	    FileEntry->SymbolCount++;
	    if (FileEntry->FirstWithName == (LineType *) NULL) {
		FileEntry->FirstWithName = ThisLine;
		LastSymbol = ThisLine;
	    } else {
		LastSymbol->NextSymbol = ThisLine;
		LastSymbol = ThisLine;
	    }
	    break;
	}

	cfp_SkipWhiteSpace(&Input, &Column);
	if (*Input == CommentCharacter) {
	    ThisLine->CommentColumn = Column;
	    if (cfp_CopyComment(&Input, ThisLine) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_InitLines:%d) comment failed for line %d in '%s'",
			  __LINE__, FileEntry->LineCount, FileEntry->FilePath);
		cfp_FreeLineSpace(FileEntry);
		return FAIL;
	    }
	}

	if (*Input != '\n') {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_InitLines:%d) junk at end of line %d in '%s'",
		      __LINE__, FileEntry->LineCount, FileEntry->FilePath);
	    DumpLine(Input);
	    cfp_FreeLineSpace(FileEntry);
	    return FAIL;
	}

	Input++;

#ifdef DEBUGDEBUG
	{
	    char *TypeCode;
	    switch (ThisLine->TypeCode) {

	      case BlankLine:
		TypeCode = "blank";
		break;

	      case CommentLine:
		TypeCode = "comment";
		break;

	      case AliasLine:
		TypeCode = "alias";
		break;

	      case ConstantLine:
		TypeCode = "constant";
		break;

	      case VariableLine:
		TypeCode = "variable";
		break;

	      default:
		TypeCode = "unknown ????";
		break;
	    }
	    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		      "(cfp_InitLines:%d) got %s line", __LINE__, TypeCode);
	}
#endif

    }

    /* set up hash table now that we have line count */
    if (cfp_BuildHash(FileEntry) != PASS) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_InitLines:%d) failed building hash table for '%s'",
		  __LINE__, FileEntry->FilePath);
	cfp_FreeLineSpace(FileEntry);
	return FAIL;
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_InitLines:%d) finished '%s'",
	      __LINE__, FileEntry->FilePath);
#endif

    return PASS;
}

/*!**********************************************************************
 *
 *   cfp_GetEnvValue()
 *
 * DESCRIPTION
 *    Get an environment variable.  This routine takes a string, prepends
 *    to it the name of the current par file (without ".," or ".par"),
 *    and then returns a pointer to the value of that environment variable.
 *    The length of the string plus the length of the par file name must
 *    be less than CFP_STRSIZE - 1 characters.  NULL is returned on error
 *    or if no such variable found.
 *
 * INVOKED
 *    env_string = cfp_GetEnvValue(FileEntry, name, size);
 *
 * PARAMETERS IN
 *    FileEntryType *FileEntry - containing file name to prepend
 *    char *name - appended to file name to create environment variable to look
 *                 for 
 *    int size - length of "name"
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    char * - pointer to value of environment variable or NULL
 *
 * GLOBALS
 *    none
 *
 *********************************************************************!*/

char *				  /* return: string , or NULL */
cfp_GetEnvValue Q((FileEntryType *FileEntry, char *name, int size),
		  (FileEntry, name, size),
		  FileEntryType *FileEntry; char *name; int size;)
{
    char ename[CFP_STRSIZE];

    if ((FileEntry == (FileEntryType *) NULL)
	|| (FileEntry->FileName == (char *) NULL)
	|| (name == (char *) NULL)) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_GetEnvValue:%d) null pointer",
		  __LINE__);
        return (char *) NULL;
    }

    if ((size + strlen(FileEntry->FileName)) < (CFP_STRSIZE - 1)) {
        strcpy(ename, FileEntry->FileName);
        strcat(ename, name);
        return getenv(ename);
    }
    return (char *) NULL;
}

/*!**********************************************************************
 *
 *   cfp_PutEnvValue()
 *
 * DESCRIPTION
 *    Set an environment variable.  This routine takes a string, prepends
 *    to it the name of the current par file (without ".," or ".par"),
 *    and then updates the value of that environment variable.
 *    The length of the string plus the length of the par file name must
 *    be less than CFP_STRSIZE - 1 characters.  FAIL
 *    is returned on error or if no such variable found.
 *
 * NOTE - better not to use this unless you need it.  It will cause a
 *        memory leak if the same variable is repeatedly updated.
 *
 * INVOKED
 *    status = cfp_PutEnvValue(FileEntry, name, size, Value);
 *
 * PARAMETERS IN
 *    FileEntryType *FileEntry - containing file name to prepend
 *    char *name - appended to file name to create environment variable to look
 *                 for 
 *    int size - length of "name"
 *    char *Value - new value for variable
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if environment updated OK, FAIL if not
 *
 * GLOBALS
 *    indirectly changes environment (through environ but not envp)
 *
 *********************************************************************!*/

PASSFAIL
cfp_PutEnvValue Q((FileEntryType *FileEntry, char *name, int size, char *Value),
		  (FileEntry, name, size, Value),
		  FileEntryType *FileEntry; char *name; int size; char *Value;)
{
    char ename[CFP_STRSIZE];	  /* for building environment name */
    char *NewVariable;		  /* points to string built for putenv */

    if ((FileEntry == (FileEntryType *) NULL)
	|| (FileEntry->FileName == (char *) NULL)
	|| (name == (char *) NULL)
	|| (Value == (char *) NULL)) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_PutEnvValue:%d) null pointer",
		  __LINE__);
        return FAIL;
    }

    if ((size + strlen(FileEntry->FileName)) < (CFP_STRSIZE - 1)) {
        strcpy(ename, FileEntry->FileName);
        strcat(ename, name);
	if ((NewVariable
	     = (char *) malloc(strlen(ename) + 1 + strlen(Value) + 1))
	    == (char *) NULL) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_PutEnvValue:%d) env malloc %d bytes failed for '%s'",
		      __LINE__, strlen(ename) + 1 + strlen(Value) + 1,
		      FileEntry->FilePath);
	    return FAIL;
	}
	strcpy(NewVariable, ename);
	strcat(NewVariable, "=");
	strcat(NewVariable, Value);
	if (putenv(NewVariable) != 0) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_PutEnvValue:%d) putenv '%s' failed for '%s'",
		      __LINE__, ShortString(NewVariable), FileEntry->FilePath);
	    return FAIL;
	}
        return PASS;
    }
    return FAIL;
}

/*!**********************************************************************
 *
 *   cfp_FindName()
 *
 * DESCRIPTION
 *    look up a variable by name in the line data base for the specified file
 *    the file must already be set up (cfp_init)
 *    the name can be a partial match as long as it is unique
 *    handle alias lookups silently
 *    return a pointer to the line if found, NULL if not
 *
 * INVOKED
 *    Line = cfp_FindName(FileEntry, name, FALSE);
 *
 * PARAMETERS IN
 *    FileEntryType *FileEntry - pointer to file entry to search
 *    char *name - identifier to look for
 *    BooleanType LookupOnly - FALSE for normal usage
 *                             TRUE if lookup only, don't complain if not found
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    LineType * - pointer to line defining this variable, NULL if not found
 *
 * GLOBALS
 *    none
 *
 ********************************************************************!*/

/* alias loops are stopped by reaching a count of times through this code */
#define AliasLoopStop 20

/*
 * The symbol search is a (possibly) two step process.  A hash table is used
 * for speed, but hash tables don't allow abbreviations.  Since most lookups
 * are for the full symbol, we try a hash first.  If that does not match, then
 * we do a linear search of the list of symbols.
 *
 * On an alias, this recurses to look up the alias.  To detect loops, a
 * counter is maintained and checked.
 */

LineType *
cfp_FindName Q((FileEntryType *FileEntry, char *Name, BooleanType LookupOnly),
	       (FileEntry, Name, LookupOnly),
	       FileEntryType *FileEntry; char *Name; BooleanType LookupOnly;)
{
    char *LookFor;		  /* name we are looking for this pass */
    size_t LookForLength;	  /* length of LookFor for abbreviations */
    unsigned int HashIndex;	  /* hash index for this symbol */
    LineType *ThisLine;		  /* pointer to current line in scan */
    LineType *FirstMatch;	  /* pointer to first match */
    static unsigned int AliasLoopCount = 0; /* count to stop infinite loops */
    char AliasIndirectName[CFP_STRSIZE]; /* for foo -> %bar's current value */

    if ((FileEntry == (FileEntryType *) NULL) || (Name == (char *) NULL)) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY, "(cfp_FindName:%d) null pointer",
		  __LINE__);
        return (LineType *) NULL;
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_FindName:%d) starting search for '%s'",
	      __LINE__, Name);
#endif

    if (FileEntry->FirstWithName == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_FindName:%d) no symbols in '%s'??",
		  __LINE__, FileEntry->FilePath);
	return (LineType *) NULL;
    }

    LookFor = Name;
    LookForLength = strlen(LookFor);
    if (LookForLength == 0) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_FindName:%d) zero length name being looked up?",
		  __LINE__);
	return (LineType *) NULL;
    }

    /* first try the hash table for an exact match */
    HashIndex = cfp_HashIt(FileEntry, LookFor);
    ThisLine = FileEntry->HashTable[HashIndex];
    while ((ThisLine != (LineType *) NULL)
	   && ((LookForLength != ThisLine->NameStringLength)
	       || (strcmp(LookFor, ThisLine->NameString) != 0))) {
#ifdef DEBUGDEBUG
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_FindName:%d) hash loop, '%s' != '%s'",
		  __LINE__, LookFor, ThisLine->NameString);
#endif
	ThisLine = ThisLine->NextHash;
    }
#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_FindName:%d) hashing found %s",
	      __LINE__,
	      (ThisLine == (LineType *) NULL) ? "<null>" :
	      ThisLine->NameString);
#endif

    /* if that didn't work, then search the whole list for abbreviations */
    if ((FirstMatch = ThisLine) == (LineType *) NULL) {
	ThisLine = FileEntry->FirstWithName;
	while (ThisLine != (LineType *) NULL) {
#ifdef DEBUGDEBUG
	    if (ThisLine->NameString != (char *) NULL) {
		cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
			  "(cfp_FindName:%d) comparing to %s",
			  __LINE__, ThisLine->NameString);
	    }
#endif
	    if ((ThisLine->NameStringLength >= LookForLength)
		&&
		(strncmp(ThisLine->NameString, LookFor, LookForLength) == 0)) {
#ifdef DEBUGDEBUG
		cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
			  "(cfp_FindName:%d) was at least partial match",
			  __LINE__);
#endif
		if (FirstMatch == (LineType *) NULL) {
		    FirstMatch = ThisLine;
		} else {
		    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			      "(cfp_FindName:%d) finding '%s', '%s' ambiguous",
			      __LINE__, Name, LookFor);
		    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			      "(...)          matches at least '%s' and '%s'",
			      FirstMatch->NameString, ThisLine->NameString);
		    return (LineType *) NULL;
		}
	    } /* end of it matched something */

	    ThisLine = ThisLine->Next;
	} /* end of search loop */
    } /* end of if hash did not work */
	
    if (FirstMatch == (LineType *) NULL) {
	if (LookupOnly == FALSE) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_FindName:%d) for '%s', no match for '%s' in '%s'",
		      __LINE__, Name, LookFor, FileEntry->FilePath);
	    errno = ENOENT;
	}
	return (LineType *) NULL;
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_FindName:%d) for '%s', found '%s' for '%s'",
	      __LINE__, Name, FirstMatch->NameString, LookFor);
#endif
    if (FirstMatch->TypeCode == AliasLine) {
	if (++AliasLoopCount > AliasLoopStop) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_FindName:%d) alias loop for '%s'",
		      __LINE__, Name);
	    AliasLoopCount--;
	    errno = EIO;
	    return (LineType *) NULL;
	}
#ifdef DEBUGDEBUG
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(...)          and that is alias for '%s'",
		  FirstMatch->OriginalValue);
#endif

	LookFor = FirstMatch->OriginalValue;

	/* foo -> %bar # requires getting bar's current value as name */
	if (LookFor[0] == AliasIndirectCharacter) {
	    LookFor++;		  /* skip the % */
	    if (cfp_getcur(FileEntry->FileID, LookFor, AliasIndirectName,
			   CFP_STRING, sizeof AliasIndirectName)
		== FAIL) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_FindName:%d) indirect lookup of '%s' for '%s' in '%s' failed",
			  __LINE__, LookFor, FirstMatch->NameString,
			  FileEntry->FileName);
		AliasLoopCount--;
		return (LineType *) NULL;
	    }
#ifdef DEBUGDEBUG
	    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		      "(...)          which has indirect value '%s'",
		      AliasIndirectName);
#endif
	    LookFor = AliasIndirectName;
	}
	FirstMatch = cfp_FindName(FileEntry, LookFor, LookupOnly);
	AliasLoopCount--;
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_FindName:%d) returning '%s' at alias level %d",
	      __LINE__,
	      ((FirstMatch->NameString == (char *) NULL)
	       ? "<null>" : FirstMatch->NameString),
	      AliasLoopCount);
#endif

    return FirstMatch;
}

/*!**********************************************************************
 *
 *   cfp_CopyToCurrent()
 *
 * DESCRIPTION
 *    replace the current value with a new value
 *    discard the old value if there is one
 *    update the current length field
 *
 * INVOKED
 *    status = cfp_CopyToCurrent(&Line, String);
 *
 * PARAMETERS IN
 *    LineType *Line - pointer to line to change
 *    char *String - new value to store in the current field
 *
 * PARAMETERS OUT
 *    LineType *Line - updated line, new current value
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS unless malloc failed, then FAIL
 *
 * GLOBALS
 *    none
 *
 *********************************************************************!*/

PASSFAIL
cfp_CopyToCurrent Q((LineType *Line, char *String),
		    (Line, String),
		    LineType *Line; char *String;)
{
    if (Line == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_CopyToCurrent:%d) null pointer", __LINE__);
	errno = EIO;
	return FAIL;
    }

    if (Line->CurrentValue != (char *) NULL)
        free(Line->CurrentValue);
    if (String == (char *) NULL) {
	Line->CurrentValueLength = 0;
	Line->CurrentValue = String;
    } else {
	Line->CurrentValueLength = strlen(String);
	if ((Line->CurrentValue = (char *) malloc(Line->CurrentValueLength + 1))
	    == (char *) NULL) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_CopyToCurrent:%d) malloc %d bytes failed",
		      __LINE__, Line->CurrentValueLength + 1);
	    errno = EIO;
	    return FAIL;
	}
	strcpy(Line->CurrentValue, String);
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_CopyToCurrent:%d) '%s' now '%s'",
	      __LINE__, Line->NameString, ShortString(Line->CurrentValue));
#endif

    return PASS;
}

/*!**********************************************************************
 *
 *   cfp_GetItem()
 *
 * DESCRIPTION
 *    given a item index (e.g., CFP_CurrentIndex), return that item from
 *    the specified line
 *
 * INVOKED
 *    Status = cfp_GetItem(&Line, Index, &String, &Length);
 *
 * PARAMETERS IN
 *    LineType *Line - pointer to the current line
 *    ItemIndexType Index - the item index to extract
 *
 * PARAMETERS OUT
 *    char **String - address pointing to string
 *    unsigned int *Length - address for returned string length
 *                           can be NULL for "I don't care"
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if OK, FAIL if bad call
 *
 * GLOBALS
 *    none
 *
 *********************************************************************!*/

PASSFAIL
cfp_GetItem Q((LineType *Line, ItemIndexType Index,
	       char **String, unsigned int *Length),
	      (Line, Index, String, Length),
	      LineType *Line; ItemIndexType Index;
	      char **String; unsigned int *Length;)
{
    unsigned int DummyLength;	  /* in case they don't provide one */

    if (Line == (LineType *) NULL) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY, "(cfp_GetItem:%d) null pointer",
		  __LINE__);
        return FAIL;
    }

    if (Length == (unsigned int *) NULL)
        Length = &DummyLength;
    *String = (char *) NULL;

    switch (Index) {

      case CFP_ModeIndex:
	*String = Line->ModeString;
	break;

      case CFP_CurrentIndex:
	*String = Line->CurrentValue;
	break;

      case CFP_DefaultIndex:
	*String = Line->DefaultValue;
	break;

      case CFP_SuppliedIndex:
	*String = Line->SuppliedValue;
	break;

      case CFP_RangeIndex:
	*String = Line->RangeValue;
	break;

      default:
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY, "(cfp_GetItem:%d) unknown index %d",
		  __LINE__, Index);
	errno = EIO;
	return FAIL;
	break;
    }

    if (*String == (char *) NULL)
      *Length = 0;
    else
      *Length = strlen(*String);

    return PASS;
}

/*!**********************************************************************
 *
 *   cfp_PutItem()
 *
 * DESCRIPTION
 *    given a item index (e.g., CFP_CurrentIndex), update that item in
 *    the specified line
 *
 * INVOKED
 *    status = cfp_PutItem(&Line, Index, Value);
 *
 * PARAMETERS IN
 *    LineType *Line - pointer to the current line
 *    ItemIndexType Index - the item index to extract
 *    char *Value - new string value
 *
 * PARAMETERS OUT
 *    LineType *Line - updated with new item value
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if OK, FAIL if could not malloc space for string
 *
 * GLOBALS
 *    none
 *
 *********************************************************************!*/

PASSFAIL
cfp_PutItem Q((LineType *Line, ItemIndexType Index, char *Value),
	      (Line, Index, Value),
	      LineType *Line; ItemIndexType Index; char *Value;)
{
    char **Where;		  /* where to put the new value */
#ifdef DEBUGDEBUG
    char *PrintWhere;		  /* for trace printout */
#endif

    if ((Line == (LineType *) NULL) || (Value == (char *) NULL)) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY, "(cfp_PutItem:%d) null pointer",
		  __LINE__);
        return FAIL;
    }

    Where = (char **) NULL;
    switch (Index) {

      case CFP_ModeIndex:
	if (cfp_ModeStringToBits(Value, TRUE /* ?? */, Line) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_PutItem:%d) mode update of '%s' to '%s' failed",
		      __LINE__, Line->NameString, Value);
	    return FAIL;
	}
	/* do not update Where, ModeStringToBits already did it */
#ifdef DEBUGDEBUG
	PrintWhere = Line->ModeString;
#endif
	break;

      case CFP_CurrentIndex:
	if (cfp_CopyToCurrent(Line, Value) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_PutItem:%d) current update of '%s' failed",
		      __LINE__, Line->NameString);
	    return FAIL;
	}
	/* do not update Where, CopyToCurrent already did everything */
#ifdef DEBUGDEBUG
	PrintWhere = Line->CurrentValue;
#endif
	break;

      case CFP_DefaultIndex:
	Where = &Line->DefaultValue;
	break;

      case CFP_SuppliedIndex:
	Where = &Line->SuppliedValue;
	break;

      case CFP_RangeIndex:
	Where = &Line->RangeValue;
	break;

      default:
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_PutItem:%d) unknown index %d", __LINE__, Index);
	errno = EIO;
	return FAIL;
	/*NOTREACHED*/
	break;
    }

    if (Where != (char **) NULL) {
	size_t Length;
	if (*Where != (char *) NULL)
	  free(*Where);
	Length = strlen(Value);
	if ((*Where = (char *) malloc(Length + 1)) == (char *) NULL) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_PutItem:%d) malloc %d failed",
		      __LINE__, Length + 1);
	    errno = EIO;
	    return FAIL;
	}
	strcpy(*Where, Value);
#ifdef DEBUGDEBUG
	PrintWhere = *Where;
#endif
    }

    if (Index == CFP_RangeIndex) { /* after string is updated */
	FreeRangeInfo(Line);
	if (cfp_BuildRangeValue(Line) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_PutItem:%d) bad format range for '%s'",
		      __LINE__, Line->NameString);
	    return FAIL;
	}
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_PutItem:%d) updated item %d to '%s' for '%s'",
	      __LINE__, Index, ShortString(PrintWhere), Line->NameString);
#endif

    return PASS;
}

/*****************************************************************************
 *
 *   cfp_RebuildAliasLine
 *
 * DESCRIPTION
 *    put alias line back together
 *
 * NOTE - caller must free String when finished with it
 *
 * INVOKED
 *    status = cfp_RebuildAliasLine(Line, &String, &Length);
 *
 * PARAMETERS IN
 *    LineType *Line - pointer to line being rebuilt
 *
 * PARAMETERS OUT
 *    char **String - address for pointer to rebuild line,
 *                    must be free'ed when used
 *    unsigned int *Length - address for length of resulting line
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if OK, FAIL if malloc didn't
 *
 * GLOBALS
 *    none
 *
 * NOTE - malloc for String needs to leave space for newline and nul -> +2
 * 
 *****************************************************************************/

static PASSFAIL
cfp_RebuildAliasLine Q((LineType *Line, char **String, size_t *Length),
		       (Line, String, Length),
		       LineType *Line; char **String; size_t *Length;)
{
    unsigned int Index;		  /* index into line */

    /* if commented, length depends on where comment was + its length
     * if not, length depends on where -> was plus alias length,
     *         with three extra for -, >, and space */
    if (Line->CommentColumn == 0)
      *Length = Line->EqualsColumn + 3 + Line->OriginalValueLength;
    else
      *Length = Line->CommentColumn + 
	(Line->CommentString?strlen(Line->CommentString):0);
    if (((*String) = (char *) malloc(*Length + 2)) == (char *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_RebuildAliasLine:%d) alias malloc %d failed",
		  __LINE__, *Length + 2);
	errno = EIO;
	return FAIL;
    }

    Index = 0;
    if (Line->LeadingSpaceCount > 0) {
	memset((void *) *String, ' ', (size_t) Line->LeadingSpaceCount);
	Index += Line->LeadingSpaceCount;
    }
    (*String)[Index] = '\0';
    strcat((*String), Line->NameString);
    Index += strlen(Line->NameString);
#ifdef DEBUGDEBUG
    if (Index != strlen(*String)) {
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_RebuildAliasLine:%d) alias bad at name i=%d, l=%d",
		  __LINE__, Index, strlen(*String));
	Index = strlen(*String);
    }
#endif
    if (Line->EqualsColumn > Index) {
	memset((void *) &((*String)[Index]), ' ',
	       (size_t) (Line->EqualsColumn  - Index));
	Index = Line->EqualsColumn;
	(*String)[Index] = '\0';
    }
    strcat((*String), "-> ");
    Index += 3;			  /* 3 == strlen("-> ") */
    strcat((*String), Line->OriginalValue);
    Index += Line->OriginalValueLength;
#ifdef DEBUGDEBUG
    if (Index != strlen(*String)) {
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_RebuildAliasLine:%d) alias bad at alias i=%d, l=%d",
		  __LINE__, Index, strlen(*String));
	Index = strlen(*String);
    }
#endif
    if (Line->CommentString != (char *) NULL) {
	if (Line->CommentColumn > Index) {
	    memset((void *) &((*String)[Index]), ' ',
		   (size_t) (Line->CommentColumn - Index));
	    Index = Line->CommentColumn;
	    (*String)[Index] = '\0';
	}
	strcat((*String), Line->CommentString);
#ifdef DEBUGDEBUG
	Index += strlen(Line->CommentString);
#endif
    }

#ifdef DEBUGDEBUG
    if ((*Length != strlen(*String)) || (Index != *Length)) {
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_RebuildAliasLine:%d) alias bad at end i=%d,ll=%d,l=%d",
		  __LINE__, Index, *Length, strlen(*String));
	*Length = strlen(*String);
    }
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_RebuildAliasLine:%d) got '%s'",
	      __LINE__, ShortString(*String));
#endif

    return PASS;
}

/***********************************************************************
 *
 *   cfp_RebuildQuotedString
 *
 * DESCRIPTION
 *    put quotes around a string that needs quoting (either because it
 *    was quoted in input or because it has nasty characters in it)
 *    insert escapes in front of any special characters
 *
 * INVOKED
 *    cfp_RebuildQuotedString(UnQuotedString, Length, String, &Index);
 *
 * PARAMETERS IN
 *    char *UnQuotedString - internal form of constant string
 *    int Length - length of internal form
 *    unsigned int *Index - current index into String, start adding result
 *                          at this point in String
 *
 * PARAMETERS OUT
 *    char **String - possibly quoted result tacked on the end
 *    unsigned int *Index - updated index into String, pointing after
 *                          what was added on
 *
 * FUNCTION RETURNS
 *    none
 *
 * GLOBALS
 *    none
 *
 ***********************************************************************/

static void
cfp_RebuildQuotedString Q((char *UnQuotedString, unsigned int Length,
			   char **String, unsigned int *Index),
			  (UnQuotedString, Length, String, Index),
			  char *UnQuotedString; unsigned int Length;
			  char **String; unsigned int *Index;)
{
    unsigned int FromIndex;	  /* index into unquoted string */
#ifdef DEBUGDEBUG
    unsigned int StartIndex = *Index; /* Index at front of our result */
#endif

    if ((Index == (unsigned int *) NULL) || (String == (char **) NULL)) {
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_RebuildQuotedString:%d) given null pointer",
		  __LINE__);
    }

    /*
     * if UnQuotedString is null, there is no string, do nothing to output
     * if UnQuotedString is non-null, but Length is 0, there was a null
     *                    string, but variables were already done and
     *                    constants don't care so do nothing to output
     * if UnQuotedString is non-null and Length is > 0 , there was a string,
     *                   generate a quoted string with escapes in output
     */

    if ((UnQuotedString != (char *) NULL) && (Length > 0)) {
	(*String)[(*Index)++] = QuoteCharacter;
	for (FromIndex = 0; FromIndex < Length; FromIndex++) {
	    switch (UnQuotedString[FromIndex]) {

	      case QuoteCharacter:
	      case EscapeCharacter:
		(*String)[(*Index)++] = EscapeCharacter;
		(*String)[(*Index)++] = UnQuotedString[FromIndex];
		break;

	      case '\n':
		(*String)[(*Index)++] = EscapeCharacter;
		(*String)[(*Index)++] = 'n';
		break;

	      case '\r':
		(*String)[(*Index)++] = EscapeCharacter;
		(*String)[(*Index)++] = 'r';
		break;

	      default:
		(*String)[(*Index)++] = UnQuotedString[FromIndex];
		break;
	    }

	}
	(*String)[(*Index)++] = QuoteCharacter;
	(*String)[*Index] = '\0';
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_RebuildQuotedString:%d) got '%s'",
	      __LINE__, ShortString(&(*String)[StartIndex]));
#endif

    return;
}

/***********************************************************************
 *
 *   cfp_RebuildString
 *
 * DESCRIPTION
 *    check out a proffered string to see if it needs quoting
 *    get it quoted if so, just copy it if not
 *    copy result to output string
 *
 * INVOKED
 *    cfp_RebuildString(LegalSymbols, UnQuotedString, Length,
 *                      String, &Index);
 *
 * PARAMETERS IN
 *    char *LegalSymbols - the set of characters that do not need quoting
 *    char *UnQuotedString - internal form of constant string
 *    int Length - length of internal form
 *    unsigned int *Index - current index into String, start adding result
 *                          at this point in String
 *
 * PARAMETERS OUT
 *    char **String - possibly quoted result tacked on the end
 *    unsigned int *Index - updated index into String, pointing after
 *                          what was added on
 *
 * FUNCTION RETURNS
 *    none
 *
 * GLOBALS
 *    none
 *
 ***********************************************************************/

static void
cfp_RebuildString Q((char *LegalSymbols, char *UnQuotedString,
		     unsigned int Length, char **String,
		     unsigned int *Index),
		    (LegalSymbols, UnQuotedString, Length, String, Index),
		    char *LegalSymbols; char *UnQuotedString;
		    unsigned int Length; char **String;
		    unsigned int *Index;)
{
#ifdef DEBUGDEBUG
    unsigned int StartIndex = *Index; /* Index at front of our result */
#endif

    if ((LegalSymbols == (char *) NULL) || (String == (char **) NULL)
	|| (Index == (unsigned int *) NULL)) {
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_RebuildString:%d) given null pointer",
		  __LINE__);
    }

    /*
     * if UnQuotedString is null, there is no string, do nothing to output
     * if UnQuotedString is non-null, but Length is 0, there was a null
     *                   string, generate ""
     * if UnQuotedString is non-null and Length is > 0 , there was a string,
     *                   if it has any nasty characters in it, generate a
     *                      quoted string with escapes in output
     *                    or if it was constant quoted in input do so again
     *                   else just copy it to output
     */

    if ((UnQuotedString != (char *) NULL) && (Length > 0)) {
	if (strspn(UnQuotedString, LegalSymbols) == Length) {
	    strcat(*String, UnQuotedString);
	    *Index += Length;
	} else {
	    cfp_RebuildQuotedString(UnQuotedString, Length, String, Index);
	}
    } else {
	if (UnQuotedString != (char *) NULL) {
	    (*String)[(*Index)++] = QuoteCharacter;
	    (*String)[(*Index)++] = QuoteCharacter;
	    (*String)[*Index] = '\0';
	}
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_RebuildString:%d) got '%s'",
	      __LINE__, ShortString(&(*String)[StartIndex]));
#endif

    return;
}

/*****************************************************************************
 *
 *   cfp_RebuildConstantLine
 *
 * DESCRIPTION
 *    put constant line back together
 *
 * NOTE - caller must free String when finished with it
 *
 * INVOKED
 *    status = cfp_RebuildConstantLine(Line, &String, &Length);
 *
 * PARAMETERS IN
 *    LineType *Line - pointer to line being rebuilt
 *
 * PARAMETERS OUT
 *    char **String - address for pointer to rebuild line,
 *                    must be free'ed when used
 *    unsigned int *Length - address for length of resulting line
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if OK, FAIL if malloc didn't
 *
 * GLOBALS
 *    none
 *
 * NOTE - malloc for String needs to leave space for newline and nul -> +2
 * 
 *****************************************************************************/

static PASSFAIL
cfp_RebuildConstantLine Q((LineType *Line, char **String, size_t *Length),
			  (Line, String, Length),
			  LineType *Line; char **String; size_t *Length;)
{
    char *From;			  /* pointer into source string */
    unsigned int Index;		  /* index into line */

    /*
     * length depends on where = was, plus = space, plus up to two
     *    times the value length (if every character needs escaping), plus
     *    two for possible quotes
     * if commented, length also depends on where comment was + its length
     * reserve more than enough space to be safe
     */
    
    *Length = Line->EqualsColumn + 2 + (Line->OriginalValueLength * 2) + 2;
    if (Line->CommentColumn > 0)
      *Length += Line->CommentColumn + 
	(Line->CommentString?strlen(Line->CommentString):0);
    if (((*String) = (char *) malloc(*Length + 2)) == (char *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_RebuildConstantLine:%d) constant malloc %d failed",
		  __LINE__, *Length + 2);
	errno = EIO;
	return FAIL;
    }

    Index = 0;
    if (Line->LeadingSpaceCount > 0) {
	memset((void *) *String, ' ', (size_t) Line->LeadingSpaceCount);
	Index += Line->LeadingSpaceCount;
    }
    (*String)[Index] = '\0';
    strcat((*String), Line->NameString);
    Index += strlen(Line->NameString);
#ifdef DEBUGDEBUG
    if (Index != strlen(*String)) {
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_RebuildConstantLine:%d) constant bad at name i=%d, l=%d",
		  __LINE__, Index, strlen(*String));
	Index = strlen(*String);
    }
#endif
    if (Line->EqualsColumn > Index) {
	memset((void *) &((*String)[Index]), ' ',
	       (size_t) (Line->EqualsColumn  - Index));
	Index = Line->EqualsColumn;
	(*String)[Index] = '\0';
    }
    strcat((*String), "= ");
    Index += 2;			  /* 2 == strlen("= ") */

    if (Line->WasQuoted == TRUE) {
	cfp_RebuildQuotedString(Line->OriginalValue,
				Line->OriginalValueLength, String, &Index);
    } else {
	/* copy string, escaping anything bad */
	From = Line->OriginalValue;
	if (From != (char *) NULL) {
	    while (*From != '\0') {
		switch (*From) {

		  case EscapeCharacter:
		    (*String)[Index++] = EscapeCharacter;
		    (*String)[Index++] = EscapeCharacter;
		    break;

		  case CommentCharacter:
		    (*String)[Index++] = EscapeCharacter;
		    (*String)[Index++] = CommentCharacter;
		    break;

		  case '\n':
		    (*String)[Index++] = EscapeCharacter;
		    (*String)[Index++] = 'n';
		    break;

		  case '\r':
		    (*String)[Index++] = EscapeCharacter;
		    (*String)[Index++] = 'r';
		    break;

		  default:
		    (*String)[Index++] = *From;
		    break;
		}
		From++;
	    }
	    (*String)[Index] = '\0';
	}
    }
#ifdef DEBUGDEBUG
    if (Index != strlen(*String)) {
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_RebuildConstantLine:%d) constant bad at constant i=%d, l=%d",
		  __LINE__, Index, strlen(*String));
	Index = strlen(*String);
    }
#endif
    if (Line->CommentString != (char *) NULL) {
	if (Line->CommentColumn > Index) {
	    memset((void *) &((*String)[Index]), ' ',
		   (size_t) (Line->CommentColumn - Index));
	    Index = Line->CommentColumn;
	    (*String)[Index] = '\0';
	}
	strcat((*String), Line->CommentString);
	Index += strlen(Line->CommentString);
    }

#ifdef DEBUGDEBUG
    if ((Index != strlen(*String)) || (Index > *Length)) {
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_RebuildConstantLine:%d) constant bad at end i=%d,ll=%d,l=%d",
		  __LINE__, Index, *Length, strlen(*String));
    }
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_RebuildConstantLine:%d) got '%s'",
	      __LINE__, ShortString(*String));
#endif

    *Length = Index;
    return PASS;
}

/*!**********************************************************************
 *
 *   cfp_RebuildVariableValue
 *
 * DESCRIPTION
 *    rebuild the value (right hand side) of a variable definition
 *    This recreates the {mode, current, default, supplied, min, max}
 *    value for a variable from the internal form.
 *
 * NOTE - the caller is responsible for free'ing the returned string value
 *
 * INVOKED
 *    status = cfp_RebuildVariableValue(Line, &String, &Length);
 *
 * PARAMETERS IN
 *    LineType *Line - the line containing the variable definition
 *
 * PARAMETERS OUT
 *    char **String - address for pointer to result string value
 *                    caller must free string when finished
 *    int *Length - address for length of String
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if construction OK, FAIL if not
 *
 * GLOBALS
 *    none
 *
 **********************************************************************!*/

PASSFAIL
cfp_RebuildVariableValue Q((LineType *Line, char **String,
			    unsigned int *Length),
			   (Line, String, Length),
			   LineType *Line; char **String; unsigned int *Length;)
{
    unsigned int SpaceNeeded;	  /* for figuring out how big string might be */
    unsigned int Index;		  /* index into result string */

    if ((Line == (LineType *) NULL) || (String == (char **) NULL)
	|| (Length == (unsigned int *) NULL)) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_RebuildVariableValue:%d) null pointer given",
		  __LINE__);
	return FAIL;
    }

    /*
     * length depends on { space
     *    plus up to two times the value length (if every character needs
     *    escaping) plus two for possible quotes plus comma space
     *    for each of mode+current+default+supplied+min+max,
     *    plus space }
     * reserve more than enough space to be safe 
     */

    SpaceNeeded = 2;
    if (Line->ModeString != (char *) NULL)
      SpaceNeeded += strlen(Line->ModeString) + 2;
    else
      SpaceNeeded += 2;
    if (Line->OriginalValueLength > Line->CurrentValueLength) {
	SpaceNeeded += (Line->OriginalValueLength * 2) + 2 + 2;
    } else {
	SpaceNeeded += (Line->CurrentValueLength * 2) + 2 + 2;
    }
    if (Line->DefaultValue != (char *) NULL)
      SpaceNeeded += strlen(Line->DefaultValue) + 2 + 2;
    else
      SpaceNeeded += 2;
    if (Line->SuppliedValue != (char *) NULL)
      SpaceNeeded += strlen(Line->SuppliedValue) + 2 + 2;
    else
      SpaceNeeded += 2;
    if (Line->RangeValue != (char *) NULL)
      SpaceNeeded += strlen(Line->RangeValue) + 2 + 2;
    SpaceNeeded += 2;
    if (((*String) = (char *) malloc(SpaceNeeded + 1)) == (char *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_RebuildVariableValue:%d) variable malloc %d failed",
		  __LINE__, SpaceNeeded + 1);
	errno = EIO;
	return FAIL;
    }
    
    strcpy(*String, "{ ");
    Index = 2;			  /* 2 == strlen("{ ") */
    if (Line->ModeString != (char *) NULL) {
	strcat(*String, Line->ModeString);
	Index += strlen(Line->ModeString);
    }
    if (Line->UpdateValue == TRUE) {
	if ((Line->CurrentValue != (char *) NULL)
	    || (Line->DefaultValue != (char *) NULL)
	    || (Line->SuppliedValue != (char *) NULL)
	    || (Line->RangeValue != (char *) NULL)) {
	    strcat(*String, ", ");
	    Index += 2;		  /* 2 == strlen(", ") */

	}
	cfp_RebuildString(LegalVariableSymbols,
			  Line->CurrentValue, Line->CurrentValueLength,
			  String, &Index);
    } else {
	if ((Line->OriginalValue != (char *) NULL)
	    || (Line->DefaultValue != (char *) NULL)
	    || (Line->SuppliedValue != (char *) NULL)
	    || (Line->RangeValue != (char *) NULL)) {
	    strcat(*String, ", ");
	    Index += 2;		  /* 2 == strlen(", ") */

	}
	cfp_RebuildString(LegalVariableSymbols,
			  Line->OriginalValue, Line->OriginalValueLength,
			  String, &Index);

    }
    if ((Line->DefaultValue != (char *) NULL)
	|| (Line->SuppliedValue != (char *) NULL)
	|| (Line->RangeValue != (char *) NULL)) {
	strcat(*String, ", ");
	Index += 2;		  /* 2 == strlen(", ") */
	if (Line->DefaultValue != (char *) NULL)
	  cfp_RebuildString(LegalVariableSymbols,
			    Line->DefaultValue, strlen(Line->DefaultValue),
			    String, &Index);
    }
    if ((Line->SuppliedValue != (char *) NULL)
	|| (Line->RangeValue != (char *) NULL)) {
	strcat(*String, ", ");
	Index += 2;		  /* 2 == strlen(", ") */
	if (Line->SuppliedValue != (char *) NULL)
	  cfp_RebuildString(LegalVariableSymbols,
			    Line->SuppliedValue, strlen(Line->SuppliedValue),
			    String, &Index);
    }
    if (Line->RangeValue != (char *) NULL) {
	strcat(*String, ", ");
	Index += 2;		  /* 2 == strlen(", ") */
	cfp_RebuildString(LegalVariableSymbols,
			  Line->RangeValue, strlen(Line->RangeValue),
			  String, &Index);
    }
    strcat(*String, " }");
    Index += 2;			  /* 2 == strlen(" }") */

#ifdef DEBUGDEBUG
    if ((Index != strlen(*String)) || (Index > SpaceNeeded)) {
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_RebuildVariableValue:%d) value bad i=%d, sl=%d, ml=%d",
		  __LINE__, Index, strlen(*String), SpaceNeeded);
	Index = strlen(*String);
    }

    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_RebuildVariableValue) got '%s' for '%s'",
	      ShortString(*String), Line->NameString);
#endif

    *Length = Index;
    return PASS;
}

/*!****************************************************************************
 *
 *   cfp_RebuildVariableLine
 *
 * DESCRIPTION
 *    put variable line back together
 *
 * NOTE - caller must free String when finished with it
 *
 * INVOKED
 *    status = cfp_RebuildVariableLine(Line, &String, &Length);
 *
 * PARAMETERS IN
 *    LineType *Line - pointer to line being rebuilt
 *
 * PARAMETERS OUT
 *    char **String - address for pointer to rebuild line,
 *                    must be free'ed when used
 *    unsigned int *Length - address for length of resulting line
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if OK, FAIL if malloc didn't
 *
 * GLOBALS
 *    none
 *
 * NOTE - malloc for String needs to leave space for newline and nul -> +2
 * 
 ***************************************************************************!*/

PASSFAIL
cfp_RebuildVariableLine Q((LineType *Line, char **String, size_t *Length),
			  (Line, String, Length),
			  LineType *Line; char **String; size_t *Length;)
{
    unsigned int Index;		  /* index into line */
    char *Value;		  /* value portion of variable */
    unsigned int ValueLength;	  /* length thereof */

    if ((Line == (LineType *) NULL) || (String == (char **) NULL)
	|| (Length == (size_t *) NULL)) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_RebuildVariableLine:%d) received null pointer",
		  __LINE__);
    }

    if (cfp_RebuildVariableValue(Line, &Value, &ValueLength) != PASS) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_RebuildVariableLine:%d) failed for '%s'",
		  __LINE__, Line->NameString);
	return FAIL;
    }

    /*
     * length depends on where = was, plus = space, plus length of value
     * if commented, length also depends on where comment was + its length
     * reserve more than enough space to be safe 
     */
    
    *Length = Line->EqualsColumn + 2 + ValueLength;
    if (Line->CommentColumn > 0)
      *Length += Line->CommentColumn +
	(Line->CommentString?strlen(Line->CommentString):0);
    if (((*String) = (char *) malloc(*Length + 2)) == (char *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_RebuildVariableLine:%d) variable malloc %d failed",
		  __LINE__, *Length + 2);
	errno = EIO;
	return FAIL;
    }

    Index = 0;
    if (Line->LeadingSpaceCount > 0) {
	memset((void *) *String, ' ', (size_t) Line->LeadingSpaceCount);
	Index += Line->LeadingSpaceCount;
    }
    (*String)[Index] = '\0';
    strcat((*String), Line->NameString);
    Index += Line->NameStringLength;
#ifdef DEBUGDEBUG
    if (Index != strlen(*String)) {
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_RebuildVariableLine:%d) variable bad at name i=%d, l=%d",
		  __LINE__, Index, strlen(*String));
	Index = strlen(*String);
    }
#endif
    if (Line->EqualsColumn > Index) {
	memset((void *) &((*String)[Index]), ' ',
	       (size_t) (Line->EqualsColumn  - Index));
	Index = Line->EqualsColumn;
	(*String)[Index] = '\0';
    }

    strcat(*String, "= ");
    Index += 2;			  /* 2 == strlen("= ") */
    strcat(*String, Value);
    Index += ValueLength;
    free((void *) Value);

    if (Line->CommentString != (char *) NULL) {
	if (Line->CommentColumn > Index) {
	    memset((void *) &((*String)[Index]), ' ',
		   (size_t) (Line->CommentColumn - Index));
	    Index = Line->CommentColumn;
	    (*String)[Index] = '\0';
	}
	strcat((*String), Line->CommentString);
	Index += strlen(Line->CommentString);
    }

#ifdef DEBUGDEBUG
    if ((Index != strlen(*String)) || (Index > *Length)) {
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_RebuildVariableLine:%d) variable bad at end i=%d,ll=%d,l=%d",
		  __LINE__, Index, *Length, strlen(*String));
	Index = strlen(*String);
    }
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_RebuildVariableLine:%d) got '%s'",
	      __LINE__, ShortString(*String));
#endif

    *Length = Index;
    return PASS;
}

/*
 * on rebuilding the file contents (cfp_RebuildFile below)
 * 
 * The buffer is first rebuilt into a list of fixed sized chunks (since we
 * have no clue at this point how big it will be).  These chunks are then
 * combined into the single buffer that is returned.  Presently most of
 * the .par files are less than 4K.  The config file is about 40K.  So
 * making the chunks hold 4K will get most of the files that get written
 * back out.
 *
 * The data in the chunk is NOT nul terminated.  memcpy() must be used
 * with Chunk->Size .
 */
#define ChunkSize 4096
typedef struct Chunk {
    struct Chunk *Next;	  /* singly linked list next pointer */
    unsigned int Size;	  /* size of this piece */
    char Data[ChunkSize];	  /* the piece */
} ChunkType;

/***********************************************************************
 *
 *   cfp_FreeChunkList
 *
 * DESCRIPTION
 *    throw away the temporary copies of file pieces
 *
 * INVOKED
 *    cfp_FreeChunkList(Head);
 *
 * PARAMETERS IN
 *    ChunkType *Head - pointer to the front of the list
 *
 * PARAMETERS OUT
 *    ChunkType *Head - (ChunkType *) NULL  (i.e., list deleted)
 *
 * FUNCTION RETURNS
 *    none
 *
 * GLOBALS
 *    none
 *
 ***********************************************************************/

static void
cfp_FreeChunkList Q((ChunkType *Head), (Head), ChunkType *Head;)
{
    ChunkType *Current;		  /* local pointer to current list entry */
    ChunkType *Next;		  /* pointer to next list entry */

    Current = Head;
    while (Current != (ChunkType *) NULL) {
	Next = Current->Next;
	free((void *) Current);
	Current = Next;
    }
    return;
}

/*!**********************************************************************
 *
 *   cfp_RebuildLine
 *
 * DESCRIPTION
 *    reconsitute one line into a string
 *    generate the file string format for one line from the internal info
 *
 * INVOKED
 *    status = cfp_RebuildLine(Line, &String, &Length);
 *
 * PARAMETERS IN
 *    LineType *Line - pointer to the line to rebuild
 *
 * PARAMETERS OUT
 *    char **String - address for pointer to resulting string
 *                    caller must free this when finished
 *    int *Length - address for length of resulting string
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if converted OK, FAIL if not
 *
 * GLOBALS
 *    none
 *
 **********************************************************************!*/

PASSFAIL
cfp_RebuildLine Q((LineType *Line, char **String, size_t *Length),
		  (Line, String, Length),
		  LineType *Line; char **String; size_t *Length;)
{
    unsigned int Index;		  /* index into string result */

    switch (Line->TypeCode) {

      case BlankLine:
	if (((*String) = (char *) malloc(2)) == (char *) NULL) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_RebuildLine:%d) blank malloc 2 failed",
		      __LINE__);
	    errno = EIO;
	    return FAIL;
	}
	*Length = 0;
	break;

      case CommentLine:
	*Length = strlen(Line->CommentString) + Line->CommentColumn;
	if (((*String) = (char *) malloc(*Length + 2)) == (char *) NULL) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_RebuildLine:%d) comment malloc %d failed",
		      __LINE__, *Length + 2);
	    errno = EIO;
	    return FAIL;
	}
	if (Line->CommentColumn > 0)
	  memset((*String), ' ', Line->CommentColumn);
	Index = Line->CommentColumn;
	(*String)[Index] = '\0';
	strcat((*String), Line->CommentString);
#ifdef DEBUGDEBUG
	if (*Length != strlen(*String)) {
	    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		      "(cfp_RebuildLine:%d) comment length %d != strlen %d",
		      __LINE__, *Length, strlen(*String));
	    *Length = strlen(*String);
	}
#endif
	break;

      case AliasLine:
	if (cfp_RebuildAliasLine(Line, String, Length) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_RebuildLine:%d) alias rebuild failed",
		      __LINE__);
	    return FAIL;
	}
	break;

      case ConstantLine:
	if (cfp_RebuildConstantLine(Line, String, Length) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_RebuildLine:%d) constant rebuild failed",
		      __LINE__);
	    return FAIL;
	}
	break;

      case VariableLine:
	if (cfp_RebuildVariableLine(Line, String, Length) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_RebuildLine:%d) variable rebuild failed",
		      __LINE__);
	    return FAIL;
	}
	break;

      default:
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_RebuildLine:%d) unknown line type %d",
		  __LINE__, Line->TypeCode);
	return FAIL;
	/*NOTREACHED*/
	break;
    }

    /* append a newline (NOTE - all routines needed to leave space!!) */
    (*String)[(*Length)++] = '\n';
    (*String)[*Length] = '\0';

    return PASS;
}

/*!**********************************************************************
 *
 *   cfp_RebuildFile
 *
 * DESCRIPTION
 *    reconstitute the file contents from the line information
 *
 *    combine the variable definitions back into lines and the lines back
 *    into a buffer that can be written to a file
 *
 * NOTE - it is the responsibility of the caller to free the buffer when
 *        they are finished with it - it is malloc'ed
 *
 * NOTE - each of the lines allocated (into String) needs to leave room for a
 *        newline and a nul !!!!!!!!!!!!!!!!!
 * 
 * INVOKED
 *    status = cfp_RebuildFile(FileEntry, &Buffer, &Length);
 *
 * PARAMETERS IN
 *    FileEntryType *FileEntry - pointer to file needing rebuilding
 *
 * PARAMETERS OUT
 *    char **Buffer - address to receive pointer to built buffer, must be
 *                    free'ed by caller
 *    unsigned int *Length - address to receive size of buffer
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if OK, FAIL if something bad happened
 *
 * GLOBALS
 *    none
 *
 *********************************************************************!*/

PASSFAIL
cfp_RebuildFile Q((FileEntryType *FileEntry, char **Buffer, unsigned int *Size),
		  (FileEntry, Buffer, Size),
		  FileEntryType *FileEntry; char **Buffer; unsigned int *Size;)
{
    ChunkType *Head;		  /* head of chunk linked list */
    ChunkType *Current;		  /* current chunk being filled */
    unsigned int Total;		  /* total size of buffer so far */
    char *String;		  /* address of a rebuilt line */
    size_t Length;		  /* length of the above */
    LineType *ThisLine;		  /* pointer to current line in line list */
    char *BufferPointer;	  /* pointer into returned buffer */

    if ((FileEntry == (FileEntryType *) NULL) || (Buffer == (char **) NULL)
	|| (Size == (unsigned int *) NULL)) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_RebuildFile:%d) null pointer for '%s'",
		  __LINE__, FileEntry->FilePath);
	return FAIL;
    }
    if ((ThisLine = FileEntry->Lines) == (LineType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY, "(cfp_RebuildFile:%d) '%s' empty?",
		  __LINE__, FileEntry->FilePath);
	*Buffer = (char *) NULL;
	*Size = 0;
	return PASS;
    }

    /* initialize list with first chunk */
    if ((Head = (ChunkType *) malloc(sizeof(ChunkType)))
	== (ChunkType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_RebuildFile:%d) head malloc %d failed in '%s'",
		  __LINE__, sizeof(ChunkType), FileEntry->FilePath);
	return FAIL;
    }
    Head->Next = (ChunkType *) NULL;
    Head->Size = 0;
    Current = Head;
    Total = 0;

    /* do each line, extending chunk list when needed */
    while (ThisLine != (LineType *) NULL) {
	if (cfp_RebuildLine(ThisLine, &String, &Length) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_RebuildFile:%d) could not rebuild line for '%s'",
		      __LINE__, FileEntry->FilePath);
	    cfp_FreeChunkList(Head);
	    errno = EIO;
	    return FAIL;
	}

	/* have a line, see if it fits in current chunk, make new if not */
	if ((Current->Size + Length) > ChunkSize) {
	    if ((Current->Next = (ChunkType *) malloc(sizeof(ChunkType)))
		== (ChunkType *) NULL) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_RebuildFile:%d) next malloc %d failed for '%s'",
			  __LINE__, sizeof(ChunkType), FileEntry->FilePath);
		cfp_FreeChunkList(Head);
		errno = EIO;
		return FAIL;
	    }
	    Current = Current->Next;
	    Current->Next = (ChunkType *) NULL;
	    Current->Size = 0;
	}
	if ((Length) > ChunkSize) {	/* sanity check! */
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_RebuildFile:%d) line too long, %d>%d, in '%s'",
		      __LINE__, Length, ChunkSize, FileEntry->FilePath);
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(...)   start of line is '%s'",
		      ShortString(String));
	    cfp_FreeChunkList(Head);
	    errno = EIO;
	    return FAIL;
	}
	(void) memcpy((void *) &(Current->Data[Current->Size]),
		      (void *) String, Length);
	Current->Size += Length;
	Total += Length;

	free(String);
	ThisLine = ThisLine->Next;
    }

    /* now combine the chunks into one block to return (+1 for debug nul) */

    if ((*Buffer = (char *) malloc((size_t) Total + 1)) == (char *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_RebuildFile:%d) block malloc %d failed",
		  __LINE__, Total + 1);
	cfp_FreeChunkList(Head);
	errno = EIO;
	return FAIL;
    }
    BufferPointer = *Buffer;
    Current = Head;
    while (Current != (ChunkType *) NULL) {
	(void) memcpy((void *) BufferPointer, Current->Data, Current->Size);
	BufferPointer += Current->Size;
	Current = Current->Next;
    }
    cfp_FreeChunkList(Head);

#ifdef DEBUGDEBUG		  /* check length and dump result */
    *BufferPointer = '\0';
    if (strlen(*Buffer) != Total) {
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_RebuildFile:%d) buffer str length %d != tot %d in '%s'",
		  __LINE__, strlen(*Buffer), Total, FileEntry->FilePath);
	free(*Buffer);
	errno = ENOTTY;
	return FAIL;
    }
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_RebuildFile:%d) resulting file '%s' looks like:",
	      __LINE__, FileEntry->FilePath);
    BufferPointer = *Buffer;
    while (*BufferPointer != '\0') {
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(...)   %s", ShortString(BufferPointer));
	BufferPointer = strchr(BufferPointer, '\n');
	if (BufferPointer == (char *) NULL) {
	    cfht_log(CFHT_LIBCFP, CFHT_DEBUG,
		      "(...)  BUT DUMP LOST LAST NEWLINE????");
	    break;
	} else {
	    BufferPointer++;
	}
    }
#endif

    *Size = Total;
    return PASS;
}

/*!**********************************************************************
 *
 *   cfp_ConvertFromString()
 *
 * DESCRIPTION
 *    This does a type conversion from a string into a requested type
 *    (BOOLEAN, int, double, or string) and copies the result into
 *    a provided destination space.
 *
 * INVOKED
 *    status = cfp_ConvertFromString(Type, SourceString,
 *                                   Destination, DestinationSize);
 *
 * PARAMETERS IN
 *    int Type - type of data required
 *    char *SourceString - source string address
 *    int DestinationSize - size of destination space
 *
 * PARAMETERS OUT
 *    void *Destination - address for resulting value
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if conversion OK and there is room for result
 *               FAIL if not
 *
 * GLOBALS
 *    none
 *
 *********************************************************************!*/

PASSFAIL
cfp_ConvertFromString Q((int Type, char *SourceString,
			 void *Destination, int DestinationSize),
			(Type, SourceString, Destination, DestinationSize),
			int Type; char *SourceString;
			void *Destination; int DestinationSize;)
{
    size_t SourceLength;	  /* for checking null strings and validity */
    char *Scan;			  /* scan pointer for validity check */

    if ((SourceString == (char *) NULL) || (Destination == (void *) NULL)) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_ConvertFromString:%d) null pointer",
		  __LINE__);
        errno = EIO;
        return FAIL;
    }

    SourceLength = strlen(SourceString);

    switch (Type) {

      case CFP_BOOLEAN:
        if ((size_t) DestinationSize != sizeof(int)) {
            cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                      "(cfp_ConvertFromString:%d) wrong size for boolean",
                      __LINE__);
            errno = EIO;
            return FAIL;
        }
        if (SourceLength <= 0)
	  *((BOOLEAN *)Destination) = FALSE;
        else
	  *((BOOLEAN *)Destination) = cfht_tob(SourceString);
#ifdef DEBUGDEBUG
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_ConvertFromString:%d) returning boolean %s",
		  __LINE__,
		  (*((BOOLEAN *)Destination) == TRUE) ? "True" : "False");
#endif
        break;

      case CFP_DOUBLE:
	if ((size_t) DestinationSize != sizeof(double)) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_ConvertFromString:%d) wrong size for double",
		      __LINE__);
	    errno = EIO;
	    return FAIL;
	}
	if (SourceLength <= 0)
	  *((double *) Destination) = 0.0;
	else {
	    Scan = SourceString + strspn(SourceString, " \t");
	    Spandouble(&Scan);
	    Scan += strspn(Scan, " \t");

	    if (*Scan != '\0') {
		char Scratch[CFP_STRSIZE + 1];
		strcpy(Scratch, ShortString(Scan));
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_ConvertFromString:%d) junk '%s' after double\
 '%s'",
			  __LINE__, Scratch,
			  ShortString(SourceString));
		return FAIL;
	    }
	    if (cfht_double(SourceString, (double *) Destination,
			    (cfht_range_t) NULL) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_ConvertFromString:%d) cfht_double failed (%d)\
 on '%s'",
			  __LINE__, cfht_errno, ShortString(SourceString));
		return FAIL;
	    }
	}
#ifdef DEBUGDEBUG
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_ConvertFromString:%d) returning double %g",
		  __LINE__, *((double *) Destination));
#endif
	break;

      case CFP_INTEGER:
	if ((size_t) DestinationSize != sizeof(long)) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_ConvertFromString:%d) wrong size for integer",
		      __LINE__);
	    errno = EIO;
	    return FAIL;
	}
	if (SourceLength <= 0)
	  *((long *) Destination) = 0;
	else {
	    Scan = SourceString + strspn(SourceString, " \t");
	    Spanint(&Scan);
	    Scan += strspn(Scan, " \t");

	    if (*Scan != '\0') {
		char Scratch[CFP_STRSIZE + 1];
		strcpy(Scratch, ShortString(Scan));
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_ConvertFromString:%d) junk '%s' after int '%s'",
			  __LINE__, Scratch,
			  ShortString(SourceString));
		return FAIL;
	    }
	    if (cfht_int(SourceString, (int *) Destination,
			 (cfht_range_t) NULL) != PASS) {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(cfp_ConvertFromString:%d) cfht_int failed (%d)\
 on '%s'",
			  __LINE__, cfht_errno, ShortString(SourceString));
		return FAIL;
	    }
	}
#ifdef DEBUGDEBUG
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_ConvertFromString:%d) returning integer %d",
		  __LINE__, *((long *) Destination));
#endif
	break;

      case CFP_STRING:
        if ((size_t)DestinationSize <= SourceLength) {
            cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                      "(cfp_ConvertFromString:%d) the string '%s' is too big",
                      __LINE__, ShortString(SourceString));
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(...)    source length = %d, destination size = %d",
		      SourceLength + 1, DestinationSize);
            errno = EIO;
            return FAIL;
        }
        strcpy((char *) Destination, SourceString);
#ifdef DEBUGDEBUG
	cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
		  "(cfp_ConvertFromString:%d) returning string '%s'",
		  __LINE__, (char *) Destination);
#endif
        break;

      default:
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_ConvertFromString:%d) unknown type %d",
		  __LINE__, Type);
        errno = EIO;
        return FAIL;
	/*NOTREACHED*/
        break;
    }

    return PASS;
}

/***********************************************************************
 *
 *   ChaseIndirects
 *
 * DESCRIPTION
 *    Convert any indirect references into current values.
 *    Fill in ASCII versions of values for error messages.
 *
 * INVOKED
 *    ChaseIndirects(Line, MinString, MaxString);
 *
 * PARAMETERS IN
 *    LineType *Line - line of variable needing current values
 *
 * PARAMETERS OUT
 *    char *MinString - filled in with numeric value or -inf
 *    char *MaxString - filled in with numeric value of +inf
 *
 * FUNCTION RETURNS
 *    none
 *
 * GLOBALS
 *    none
 *
 ***********************************************************************/

static void
ChaseIndirects Q((LineType *Line, char *MinString, char *MaxString),
		 (Line, MinString, MaxString),
		 LineType *Line; char *MinString; char *MaxString;)
{
    RangeInfoType *Info;	  /* local copy of Line->RangeInfo */

    Info = Line->RangeInfo;

    if (Info->MinIndirect == TRUE) {
	if (Info->Type == CFP_INTEGER) {
	    if (cfp_getcur(Line->FileEntry->FileID, Info->MinString,
			   (void *) &Info->intRangeMin, CFP_INTEGER,
			   sizeof(int)) != FAIL) {
		sprintf(MinString, "%d", Info->intRangeMin);
	    } else {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(ChaseIndirects:%d) get int value of %%%s for '%s'",
			  __LINE__, Info->MinString, Line->NameString);
		Info->intRangeMin = INT_MIN;
		(void) strcpy(MinString, MinusInf);
	    }
	} else {
	    if (cfp_getcur(Line->FileEntry->FileID, Info->MinString,
			   (void *) &Info->doubleRangeMin, CFP_DOUBLE,
			   sizeof(double)) != FAIL) {
		sprintf(MinString, "%f", Info->doubleRangeMin);
	    } else {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(ChaseIndirects:%d) get double value of %%%s for '%s'",
			  __LINE__, Info->MinString, Line->NameString);
		Info->doubleRangeMin = -DBL_MAX;
		(void) strcpy(MinString, MinusInf);
	    }
	}
    } else {
	(void) strcpy(MinString, Info->MinString);
    }

    if (Info->MaxIndirect == TRUE) {
	if (Info->Type == CFP_INTEGER) {
	    if (cfp_getcur(Line->FileEntry->FileID, Info->MaxString,
			   (void *) &Info->intRangeMax, CFP_INTEGER,
			   sizeof(int)) != FAIL) {
		sprintf(MaxString, "%d", Info->intRangeMax);
	    } else {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(ChaseIndirects:%d) get int value of %%%s for '%s'",
			  __LINE__, Info->MaxString, Line->NameString);
		Info->intRangeMax = INT_MAX;
		(void) strcpy(MaxString, PlusInf);
	    }
	} else {
	    if (cfp_getcur(Line->FileEntry->FileID, Info->MaxString,
			   (void *) &Info->doubleRangeMax, CFP_DOUBLE,
			   sizeof(double)) != FAIL) {
		sprintf(MaxString, "%f", Info->doubleRangeMax);
	    } else {
		cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
			  "(ChaseIndirects:%d) get double value of %%%s for '%s'",
			  __LINE__, Info->MaxString, Line->NameString);
		Info->doubleRangeMax = DBL_MAX;
		(void) strcpy(MaxString, PlusInf);
	    }
	}
    } else {
	(void) strcpy(MaxString, Info->MaxString);
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(ChaseIndirects:%d) on '%s' min='%s' max='%s'",
	      __LINE__, Line->NameString, MinString, MaxString);
#endif
    return;
}

/*!**********************************************************************
 *
 *   cfp_EditValue
 *
 * DESCRIPTION
 *    Given that a range is specified, check to make sure the given string
 *    does not have any illegal characters or an invalid format.
 *
 *    This assumes that the string is only a partial value so far.  The string
 *    does not have to be acceptable to strtol/strtod, it just has to be such
 *    that it can be completed into such a string.  For example "+" is not
 *    valid as a number, but it can become such.  Similarly "1e" and "1e-" are
 *    not valid as numbers, but they can become so.
 *
 * INVOKED
 *    Status = cfp_EditValue(Line, Type, Value);
 *
 * PARAMETERS IN
 *    LineType *Line - the line defining the variable
 *    int Type - cfp variable type (CFP_INTEGER, CFP_DOUBLE, or CFP_STRING)
 *    char *Value - characters to check
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if string OK so far, FAIL if not
 *
 * GLOBALS
 *    none
 *
 **********************************************************************!*/

PASSFAIL
cfp_EditValue Q((LineType *Line, int Type, char *Value),
		(Line, Type, Value),
		LineType *Line; int Type; char *Value;)
{
#ifdef DEBUGDEBUG
    char RangeString[CFP_STRSIZE + 1];
#endif
    RangeInfoType *Info;	  /* local copy of Line->RangeInfo */
    char *Scan;			  /* pointer into value string */

    if ((Line == (LineType *) NULL) || (Value == (char *) NULL)) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_EditValue:%d) given null pointer",
		  __LINE__);
	return FAIL;
    }
    if (Line->RangeInfo == (RangeInfoType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_EditValue:%d) null range for '%s'",
		  __LINE__, Line->NameString);
	return PASS;
    }

    Info = Line->RangeInfo;
    if (((Type == CFP_INTEGER) && (Info->Type == CFP_DOUBLE))
	||
	((Type == CFP_DOUBLE) && (Info->Type == CFP_INTEGER))) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_EditValue:%d) types != (par=%d call=%d) for '%s' in '%s'",
		  __LINE__, Info->Type, Type, Line->NameString,
		  Line->FileEntry->FileName);
    }

#ifdef DEBUGDEBUG
    strcpy(RangeString, ShortString(Line->RangeValue));
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_EditValue:%d) starting name=%s, range='%s', value='%s'",
	      __LINE__, Line->NameString, RangeString, ShortString(Value));
#endif

    /* now check the string for valid characters and format */
    if (Info->Type == CFP_INTEGER) {
	Scan = Value + strspn(Value, " \t");
	Spanint(&Scan);
	Scan += strspn(Scan, " \t");
	if (*Scan != '\0') {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_EditValue:%d) '%s' invalid int format for '%s'",
		      __LINE__, ShortString(Value), Line->NameString);
	    return FAIL;
	}
    } else {			  /* double case */
	Scan = Value + strspn(Value, " \t");
	Spandouble(&Scan);
	Scan += strspn(Scan, " \t");
	if (*Scan != '\0') {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_EditValue:%d) '%s' invalid double format for '%s'",
		      __LINE__, ShortString(Value), Line->NameString);
	    return FAIL;
	}
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_EditValue:%d) end", __LINE__);
#endif

    return PASS;
}

/*!**********************************************************************
 *
 *   cfp_CheckRange()
 *
 * DESCRIPTION
 *
 *    Given that a range is specified, check the given string
 *    to make sure it is within the bounds specified by the min-max field.
 *    This checks CFP_INTEGER and CFP_DOUBLE values directly and CFP_STRING
 *    values by inference (if it looks like an integer use that, if it
 *    looks like a float or time use that, otherwise punt).
 *
 *    cfp_getcur usually has a good idea of type (though not for 0:0:0) and
 *    will pass this routine the right type.  cfp_putcur may have the right
 *    type but is also called from a text edit window with a string.  Given
 *    that someone has included a range checking in the par file, presumably
 *    the min/max values should be an indication of what type the value really
 *    is.  Therefore this routine attempts to grok the right type for a
 *    CFP_STRING value by looking at the range specifier.  If it looks like
 *
 *         int
 *    or
 *         [[(][-+]?[0-9]*,[-+]?[0-9]*[])]
 *    where spaces outside the numbers are allowed, "int" can also precede the
 *    range values, and human readable examples include
 *         int [ 1 ]                 integer from 1 through INT_MAX
 *         [ 0, 100 ]                integer from 1 through 100
 *         [ , 0 )                   integer from INT_MIN to 0
 *    , the value is presumed to be an integer and all values are converted to
 *    int's for range checking.  If it looks like
 *
 *         double
 *         time
 *    or
 *         [[(][-+]?[0-9]*{.[0-9]*}{[eE]?[-+]?[0-9]+},
 *             [-+]?[0-9]*{.[0-9]*}{[eE]?[-+]?[0-9]+}[])]
 *    or
 *         [[(][-+]?[0-9]*:[0-9]*:[0-9]*{.[0-9]*},
 *             [-+]?[0-9]*:[0-9]*:[0-9]*{.[0-9]*}[])]
 *    where spaces outside the numbers are allowed, "double" or "time" can
 *    also precede the range values,  and human readable examples are
 *         [ 0.0, 90.0 ]
 *         [ -90.0, 90.0 ]
 *         time [ 0, 24:0:0 )
 *    meaning values from 0.0 through 90.0 inclusive, from -90.0 through +90.0
 *    inclusive, or from 0 hours inclusive to 24 hours non-inclusive are
 *    allowable, the value is presumed to be a double or time and all values
 *    are converted to double for range checking.
 *
 *    Either the minimum and or the maximum value can also be given as
 *         %variablename
 *    which substitutes the current value of variablename at the time of the
 *    range check.  If both minimum and maximum are such indirect references,
 *    the variable type is assumed to be integer unless a type is explicitly
 *    given.  The existence of variablename is not checked until the first
 *    range check happens.
 *
 *    Invalid range specifications are logged but ignored as far as the user
 *    is concerned.
 *
 * INVOKED
 *    status = cfp_CheckRange(Line, Type, Value);
 *
 * PARAMETERS IN
 *    LineType *Line - pointer to the line information for the value being
 *                     checked
 *    int Type - CFP_INTEGER, CFP_DOUBLE, or CFP_STRING
 *    char *Value - pointer to the string form of the value to be checked
 *
 * PARAMETERS OUT
 *    none
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if OK or only bogus range specifier (?), FAIL if problem
 *
 * GLOBALS
 *    none
 *
 *********************************************************************!*/

PASSFAIL
cfp_CheckRange Q((LineType *Line, int Type, char *Value),
		 (Line, Type, Value),
		 LineType *Line; int Type; char *Value;)
{
#ifdef DEBUGDEBUG
    char RangeString[CFP_STRSIZE + 1];
#endif
    RangeInfoType *Info;	  /* local copy of Line->RangeInfo */
    char MinString[CFP_STRSIZE + 1]; /* ASCII min and max versions for */
    char MaxString[CFP_STRSIZE + 1]; /*   error messages */

    int Dummyint;		  /* dummy for cfht_int call */
    double Dummydouble;		  /* dummy for cfht_double call */

    if ((Line == (LineType *) NULL) || (Value == (char *) NULL)) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_CheckRange:%d) given null pointer",
		  __LINE__);
	return FAIL;
    }
    if (Line->RangeInfo == (RangeInfoType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_CheckRange:%d) null range for '%s' in '%s'",
		  __LINE__, Line->NameString, Line->FileEntry->FileName);
	return PASS;
    }

    Info = Line->RangeInfo;
    if (((Type == CFP_INTEGER) && (Info->Type == CFP_DOUBLE))
	||
	((Type == CFP_DOUBLE) && (Info->Type == CFP_INTEGER))) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_CheckRange:%d) types != (par=%d call=%d) for '%s' in '%s'",
		  __LINE__, Info->Type, Type, Line->NameString,
		  Line->FileEntry->FileName);
    }

#ifdef DEBUGDEBUG
    strcpy(RangeString, ShortString(Line->RangeValue));
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_CheckRange:%d) starting name=%s, range='%s', value='%s'",
	      __LINE__, Line->NameString, RangeString, ShortString(Value));
#endif

    /* see if any indirect references need to be chased */
    ChaseIndirects(Line, MinString, MaxString);

    /* now check the range */
    if (Info->Type == CFP_INTEGER) {
	if (cfht_int(Value, &Dummyint, (cfht_range_t) &Info->Range) != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_CheckRange:%d) '%s' out of range for '%s'",
		      __LINE__, ShortString(Value), Line->NameString);
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(...)   %s <%s x <%s %s",
		      MinString,
		      (Info->intRangeLeft == TRUE)
		        ? "=" : "",
		      (Info->intRangeRight == TRUE)
		        ? "=" : "",
		      MaxString);
	    return FAIL;
	}
    } else {			  /* double case */
	if (cfht_double(Value, &Dummydouble, (cfht_range_t) &Info->Range)
	    != PASS) {
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(cfp_CheckRange:%d) '%s' out of range for '%s'",
		      __LINE__, ShortString(Value), Line->NameString);
	    cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		      "(...)   %s <%s x <%s %s",
		      MinString,
		      (Info->doubleRangeLeft == TRUE)
		        ? "=" : "",
		      (Info->doubleRangeRight == TRUE)
		        ? "=" : "",
		      MaxString);
	    return FAIL;
	}
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_CheckRange:%d) end", __LINE__);
#endif

    return PASS;
}

/*!**********************************************************************
 *
 *   cfp_BuildRangeMessage
 *
 * DESCRIPTION
 *    Construct a "human readable" message from the range information
 *    This takes a line's range info structure and transforms it into a more
 *    or less readable form for output to a user.
 *
 * INVOKED
 *    Status = cfp_BuildRangeMessage(Line, Message);
 *
 * PARAMETERS IN
 *    LineType *Line - pointer to the line needing explaining
 *
 * PARAMETERS OUT
 *    char *Message - filled in message area (must be at least CFP_STRSIZE+1)
 *
 * FUNCTION RETURNS
 *    PASSFAIL - PASS if message OK (message can describe an error)
 *               FAIL if bad parameters
 *
 * GLOBALS
 *    none
 *
 **********************************************************************!*/

PASSFAIL
cfp_BuildRangeMessage Q((LineType *Line, char *Message),
			(Line, Message),
			LineType *Line; char *Message;)
{
#ifdef DEBUGDEBUG
    char RangeString[CFP_STRSIZE + 1];
#endif
    RangeInfoType *Info;	  /* local copy of Line->RangeInfo */
    char MinString[CFP_STRSIZE + 1]; /* ASCII min and max versions for */
    char MaxString[CFP_STRSIZE + 1]; /*   message */
    BooleanType LeftClosed;	  /* TRUE if [  FALSE if ( */
    BooleanType RightClosed;	  /* TRUE if ]  FALSE if ) */

    if ((Line == (LineType *) NULL) || (Message == (char *) NULL)) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_BuildRangeMessage:%d) given null pointer",
		  __LINE__);
	return FAIL;
    }
    if (Line->RangeInfo == (RangeInfoType *) NULL) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_BuildRangeMessage:%d) null range for '%s' in '%s'",
		  __LINE__, Line->NameString, Line->FileEntry->FileName);
	sprintf(Message, "Check but no range field for '%s' in '%s'",
		Line->NameString, Line->FileEntry->FileName);
	return PASS;
    }

    Info = Line->RangeInfo;

#ifdef DEBUGDEBUG
    strcpy(RangeString, ShortString(Line->RangeValue));
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG,
	      "(cfp_BuildRangeMessage:%d) starting name=%s, range='%s'",
	      __LINE__, Line->NameString, RangeString);
#endif

    /* see if any indirect references need to be chased */
    ChaseIndirects(Line, MinString, MaxString);

    /* get ends out where they're accessible */
    if (Info->Type == CFP_INTEGER) {
	LeftClosed = Info->intRangeLeft;
	RightClosed = Info->intRangeRight;
    } else {
	LeftClosed = Info->doubleRangeLeft;
	RightClosed = Info->doubleRangeRight;
    }

    /* now build the message */
    if ((Info->MinSeen == FALSE) && (Info->MinIndirect == FALSE)
	&& (Info->MaxSeen == FALSE) && (Info->MaxIndirect == FALSE)) {
	if (Info->Type == CFP_INTEGER) {
	    (void) strcpy(Message, "Value can be any integer");
	} else {
	    (void) strcpy(Message, "Value can be any real value or time");
	}
    } else {
	(void) strcpy(Message, "Valid values from ");
	(void) strcat(Message, MinString);
	if (LeftClosed == TRUE)
	  (void) strcat(Message, " inclusive to ");
	else
	  (void) strcat(Message, " exclusive to ");
	(void) strcat(Message, MaxString);
	if (RightClosed == TRUE)
	  (void) strcat(Message, " inclusive");
	else
	  (void) strcat(Message, " exclusive");
    }

#ifdef DEBUGDEBUG
    cfht_logv(CFHT_LIBCFP, CFHT_DEBUG, "(cfp_BuildRangeMessage:%d) end", __LINE__);
#endif

    return PASS;
}


/*****************************************************************************
 *
 * cfp_chidForGXE
 *
 * check id for validity and return pointer to block with file name and path
 * this routine is mainly for backwards compatibility with cfp version 1
 * just for GXE since it wants a copy of the file
 * 
 * GXE is responsible for free'ing the data block after it finished making its
 * own copy of the file
 *
 *****************************************************************************/

struct cfp_inst_GXE *		  /* return: fast pointer */
cfp_chidForGXE Q((FILE_ID id), (id), FILE_ID id;)
{
    FileEntryType *FileEntry;	  /* pointer to private file entry */

    /* public file instance array - GXE only version */
    static struct cfp_inst_GXE cfp_inst_table[CFP_MAXINST + 1];
    struct cfp_inst_GXE *instp;	  /* pointer to public file instance */
    unsigned int Size;		  /* dummy size for RebuildFile */

    if (id < 0 || id > CFP_MAXINST) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_chidForGXE:%d) id (%d) out of range", __LINE__, id);
        errno = ENOTTY;
        return (struct cfp_inst_GXE *) NULL;
    }

    FileEntry = &FileEntryTable[id];
    instp = &cfp_inst_table[id];

    if (FileEntry->InUseCode != InUse) {
        cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
                  "(cfp_chidForGXE:%d) id (%d) appears uninitialized", __LINE__, id);
        errno = ENOTTY;
        return (struct cfp_inst_GXE *) NULL;
    }

    instp->id = FileEntry->FileID;
    strcpy(instp->file, FileEntry->FileName);
    strcpy(instp->path, FileEntry->FilePath);
    if (cfp_RebuildFile(FileEntry, &instp->data, &Size) != PASS) {
	cfht_logv(CFHT_LIBCFP, CFHT_LOGONLY,
		  "(cfp_chidForGXE:%d) rebuild file failed for '%s'",
		  __LINE__, FileEntry->FilePath);
	return (struct cfp_inst_GXE *) NULL;
    }

    return instp;
}
