/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
* FILE
* $Header: exp_controller.c,v 1.3 96/11/19 22:52:26 isani Exp $
* $Locker:  $
*
* ROUTINES
*
* exp_controller()
*    This routine will return a value from the enumeration exp_controller_t
*    indicating what sort of detector controller is in use.  This is an
*    expensive call due to the number of files needed to be parsed.
*
*    INVOKED
*    controller = exp_controller();
*
*    PARAMETERS IN
*    none
*
*    PARAMETERS OUT
*    none
*
*    FUNCTION RETURNS
*    exp_controller_t - type of controller in use, or FAIL
*
*    GLOBALS
*    none
*
* HISTORY
* $Log:	exp_controller.c,v $
 * Revision 1.3  96/11/19  22:52:26  isani
 * cleaned up rcsid
 * 
 * Revision 1.2  1994/12/14  17:32:12  thomas
 * Added RCS header
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h, cfht/states.h to states.h
 * Changed CFHT_ERROR to CFHT_LOGONLY
 *
 * Revision 1.1  94/01/20  04:51:59  04:51:59  jwright (Jim Wright)
 * Initial revision
 * 
****************************************************************************!*/

#include <string.h>		  /* for string routines */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* parser defines */
#include "states.h"		  /* exposure states */

RCSID("$Header: exp_controller.c,v 1.3 96/11/19 22:52:26 isani Exp $");

/*
 * Note that some of the definitions below also appear in <cfht/ccd.h>.
 * We can't include that here, because of problems related to build order.
 * An argument might be made that this function should be in libccd, but
 * it is not specific to ccds, and that would force anyone who wants to
 * know about the controller to include all of libccd.
 */

#define CCD_TYPE          "ccd_type"     /* card name in detector cfg file */
#define CCD_TYPE_GENIII   "ccdIII"       /* new generation III */
#define CCD_TYPE_CC200    "cc200"        /* old photometrics cc200 */
#define CCD_TYPE_FTS      "fts"          /* fts custom controller */
#define CCD_TYPE_RETICON  "reticon"      /* reticon custom controller */


enum exp_controller_t exp_controller()
{
    char *detector;
    char *cfg_path;
    FILE_ID cfp_id;
    char str[256];
    enum exp_controller_t controller;

    /* get detector name */

    if ((detector = exp_detector()) == NULL) {
        cfht_log (CFHT_LIBCFHT, CFHT_LOGONLY, "no detector name");
        return (FAIL);
    }
    cfht_logv (CFHT_LIBCFHT, CFHT_DEBUG, "detector is %s", detector);

    /* get cfg file name */

    if ((cfg_path = cfp_file(NULL, detector, ".cfg", "rw")) == NULL) {
        cfht_logv (CFHT_LIBCFHT, CFHT_LOGONLY, "no config file for %s", detector);
        return (FAIL);
    }

    /* initialize cfg file for cfp I/O */

    if ((cfp_id = cfp_init(cfg_path)) == FAIL) {
        cfht_logv (CFHT_LIBCFHT, CFHT_LOGONLY,
		   "can't init config file for %s", detector);
        return (FAIL);
    }
    if (cfp_getstr (cfp_id, CCD_TYPE, str, 80) == FAIL) {
        /* if ccd_type not in .cfg file, then must be old */
        return (FAIL);
    } else if (strncmp (CCD_TYPE_GENIII, str, strlen(CCD_TYPE_GENIII)) == 0) {
	controller = CONTROLLER_GENIII;
    } else if (strncmp (CCD_TYPE_CC200, str, strlen(CCD_TYPE_CC200)) == 0) {
	controller = CONTROLLER_CC200;
    } else if (strncmp (CCD_TYPE_FTS, str, strlen(CCD_TYPE_FTS)) == 0) {
	controller = CONTROLLER_FTS;
    } else if (strncmp (CCD_TYPE_RETICON, str, strlen(CCD_TYPE_RETICON)) == 0) {
	controller = CONTROLLER_RETICON;
    } else {
        return (FAIL);
    }

    (void) cfp_finish(cfp_id);
    return (controller);
}
