/* Copyright (C) 1991, 94, 97    Canada-France-Hawaii Telescope Corp.    */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */

/*!****************************************************************************
 *
 * $Header: exp_sigusr1.c,v 1.1 97/03/06 22:27:51 isani Exp $
 * $Locker:  $
 *
 * This is an extraction of exp_sigusr1() from its original
 * home in exp_states.c.  exp_sig stuff will not compile under
 * Solaris, but we wanted the rest of the stuff in exp_states,
 * so the two had to be split.
 *
 * The comments in the other exp_sig* files apply to this file
 * as well.  The main difference is that exp_sigusr1 sets
 * SV_INTERRUPT to make the old Suns behave like the HP's in
 * that interrupted slow system calls will not automatically
 * restart.  Instead, EINTR is returned.
 *
 * The previous RCS log, when this file was still in exp_states,
 * included the following messages:
 *
 *   Revision 1.6  93/06/21  15:28:51  15:28:51  steve (Steven Smith)
 *   added sun stuff for interuptable read/writes
 *
 *   Revision 1.5  93/03/05  16:24:34  16:24:34  steve (Steven Smith)
 *   added exp_sig1event_fastcheck() and the header for it (I want a raise)
 * 
 * ... there were others that didn't apply to exp_sigusr1.
 * If in doubt, check the exp_states.c file for more log entries.
 *
 * $Log:	exp_sigusr1.c,v $
 * Revision 1.1  97/03/06  22:27:51  isani
 * Initial revision
 * 
 * 
 ***************************************************************************!*/

#include <stdio.h>
#include <signal.h>

#include "cfht.h"		  /* libcfht defines */
#include "states.h"		  /* exposure states */

RCSID("$Header: exp_sigusr1.c,v 1.1 97/03/06 22:27:51 isani Exp $");

#define MASK(s) (1L<<((s) - 1))
#define UNMASK(s) (0L<<((s) - 1))

#ifdef sun
#define sigvector(sig,vec,ovec) sigvec((sig),(vec),(ovec))
#endif

/* global to these functions */

static int exp_global_event=0;
static void exp_trapper P((int sig));

/*!****************************************************************************
*
*  This function sets up the signal catching for SIGUSR1.
*
*  INVOKED: exp_sigusr1 (block)
*
*  PARAMETERS IN:
*      BOOLEAN block -> TRUE is block SIGUSR1
*                       FALSE is pass SIGUSR1
*
*  PARAMETERS OUT:
*      none
*
*  FUNCTION RETURNS:
*      void
*
*  GLOBALS: exp_global_event
*
*  -Date- -by- - D e s c r i p t i o n -
*
*  890123  jk  - Original version
*  890322 sss  - allow other signal libs to function
*
****************************************************************************!*/

void exp_sigusr1 (block)
BOOLEAN block;
{
    static int init=0;

    struct sigvec vec_ignore ;
    struct sigvec vec_trap ;

    /**************************
    *  Init section
    **************************/

    if (!init) {

        /*  Ignore SIGINT */

        vec_ignore.sv_handler = SIG_IGN;
        vec_ignore.sv_mask = 0;
        vec_ignore.sv_onstack = 0;
        sigvector (SIGINT, &vec_ignore, (struct sigvec *)NULL);

        /* In setting up the event, SIGUSR1, block the signal initially */

        vec_trap.sv_handler = exp_trapper;
        vec_trap.sv_mask = MASK (SIGUSR1);
        vec_trap.sv_onstack = 0;
#ifdef sun
	/* since SunOS 4.x by default has non-interuptable read/writes */
	/* see the sigvec(2), read(2) man pages */
	vec_trap.sv_flags = SV_INTERRUPT;
#endif      
        sigvector (SIGUSR1, &vec_trap, (struct sigvec *)NULL);

        init = 1;
    }

    /***********************************
    *  Handle hold/passing the signal
    ***********************************/

    if (block == TRUE) {
        (void)sigsetmask(sigsetmask(~0L)|sigmask(SIGUSR1));
    } else {
        (void)sigsetmask(sigsetmask(~0L)&(~sigmask(SIGUSR1)));
    }

    return;
}

/*!****************************************************************************
*
* exp_trapper -- internal to this routine, not to be called from elsewhere
*
***************************************************************************!*/

static void exp_trapper (sig)
int sig;
{
    if (sig == SIGUSR1)
        exp_global_event = 1;
}

/*!****************************************************************************
*
*  NAME:  int exp_sig1event_fastcheck()
*
*  This function returns the global flag "exp_global_event"
*  It is used as a fast check to see if dakine has happened, without
*  all the overhead associated with exp_sig1event()
*
*  INVOKED: if(exp_sig1event_fastcheck()) printf("we have had a sig1\n");
*
*  PARAMETERS IN:
*      none
*
*  PARAMETERS OUT:
*      none
*
*  FUNCTION RETURNS:
*	value of exp_global_even - zero if no sig1 lately
*
*  GLOBALS: exp_global_event
*
*  -Date- -by- - D e s c r i p t i o n -
*
*  930305 sss  - created
*
***************************************************************************!*/

int exp_sig1event_fastcheck()
{
  return(exp_global_event);
}


/*!****************************************************************************
*
*  NAME:  BOOLEAN exp_sig1event
*
*  This function tests and clears the global flag "exp_global_event"
*
*  The idea here is that if a SIGUSR1 was set up and subsequently trapped in
*  exp_sigusr1(), the global flag would have been set. When a process that is
*  pausing (see pause(2)) on this event is awaken, it should first check
*  that the wakeup call was really for SIGUSR1.
*
*  That is this function's job.  It will test that the flag is set, and if
*  so, clear it and return true.  Else return false.
*
*  INVOKED: while not (exp_sig1event ())
*               sleep
*           end while
*           proceed with state checking
*
*  PARAMETERS IN:
*      none
*
*  PARAMETERS OUT:
*      none
*
*  FUNCTION RETURNS:
*      TRUE if global flag set
*      FALSE  if spurrious event
*
*  GLOBALS: exp_global_event
*
*  -Date- -by- - D e s c r i p t i o n -
*
*  880930  jk  - Original version
*  890322 sss  - added blocking of sigs during critical section
*
***************************************************************************!*/

BOOLEAN exp_sig1event ()
{
    BOOLEAN rtn;
    long oldmask;

    oldmask=sigsetmask(~0L);    /* block all sigs */

    if (exp_global_event) {
        exp_global_event = 0;
        rtn=TRUE;
    } else {
        rtn=FALSE;
    }

    (void)sigsetmask(oldmask);        /* reset all sigs */

    return(rtn);
}

/* end of exp_sigusr1.c */
