/* Copyright (C) 1991, 94    Canada-France-Hawaii Telescope Corp.        */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */

/*!****************************************************************************
* FILE
*
* $Header: exp_states.c,v 1.11 2000/02/10 13:16:46 isani Exp $
* $Locker:  $
*
* ROUTINES
*
*    INVOKED
*
*    PARAMETERS IN
*
*    PARAMETERS OUT
*
*    FUNCTION RETURNS
*
*    GLOBALS
*
* HISTORY
* who          when            what
* ---------    ------------    ----------------------------------------------
* jrw          16 Feb 1991     check for null pointers
* jk           23 may 1991     removed ifdef sun
* jk           13 Jun 1991     added lseek error checking
* jk           14 Jun 1991     made lock_fd global
*
* $Log:	exp_states.c,v $
 * Revision 1.11  2000/02/10  13:16:46  isani
 * Rewrote if-statement to avoid warning about assignment used as truth statement.
 * 
 * Revision 1.10  1997/03/06  22:55:48  isani
 * Moved exp_sigusr1 to exp_states
 * Cleaned up RCSIDs
 *
 * Revision 1.9  95/02/28  20:16:32  thomas
 * Fix memory access error, stuff null after bytes read from lock file
 * 
 * Revision 1.8  94/12/14  17:48:24  17:48:24  thomas (Jim Thomas)
 * Added RCS headers
 * Removed sys/types.h
 * Changed cfht/cfht.h to cfht.h, cfht/cfp.h to cfp.h, cfht/states.h to states.h
 * Changed function prototypes to use P macro
 * Made functions match prototype definitions
 * Changed CFHT_MAIN to CFHT_LIBCFHT and standardized message format
 * Changed CFHT_ERROR to CFHT_LOGONLY
 * Added messages on some FAIL returns
 * Changed lockf error check to < 0
 * 
 * Revision 1.7  94/01/20  04:53:20  04:53:20  jwright (Jim Wright)
 * extract various routines and put them in their own file
 * fix some variable names to avoid name-space conflicts
 * only include one instance of rcs_id variable
 * 
 * Revision 1.6  93/06/21  15:28:51  15:28:51  steve (Steven Smith)
 * added sun stuff for interuptable read/writes
 * 
 * Revision 1.5  93/03/05  16:24:34  16:24:34  steve (Steven Smith)
 * added exp_sig1event_fastcheck() and the header for it (I want a raise)
 * 
 * Revision 1.4  93/02/16  16:21:32  16:21:32  john (John Kerr)
 * made exp_instrument return CFHT_INSTRUMENT_TYPE
 * 
 * Revision 1.3  93/02/16  15:47:56  15:47:56  john (John Kerr)
 * added exp_instrument to get the instrument line from the config file.
 * 
 * Revision 1.2  91/09/27  12:19:24  12:19:24  john (John Kerr)
 * Added additional check for non-existent signal handlers. Report error and
 * just return.
 * 
 * Revision 1.1  91/09/24  17:51:40  17:51:40  steve (Steven Smith)
 * Original checkin
 * 
****************************************************************************!*/

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#ifndef F_TLOCK
#include <unistd.h>
#endif
#include <signal.h>
#include <errno.h>
#include <stdlib.h>   /* for getenv() */

#include "cfht.h"		  /* libcfht defines */
#include "cfp.h"		  /* parsing defines */
#include "states.h"		  /* exposure states */

RCSID("$Header: exp_states.c,v 1.11 2000/02/10 13:16:46 isani Exp $");

#define MASK(s) (1L<<((s) - 1)) === sigvector too
#define UNMASK(s) (0L<<((s) - 1))

#define BUF_LEN 256

/* global to these functions */

static PASSFAIL exp_maybeopen P((char *detector));
static void exp_maybeclose P((int my_lock_fd));

static int exp_lfopen = FALSE;
static int exp_canclose = TRUE;

static int lock_fd = -1;  /* jk made global 14 June 1991 */

/******************************************************************************
*******************************************************************************
******************************************************************************/

/*!****************************************************************************
*
*   Check for a valid exposure in progress with a specified detector.
*
*   If something is truly happening, state will be updated in the lockfile
*   and the pid of the process using the detector will be returned by the
*   function.
*
*  INVOKED: if ((state=exp_curstate (detector, pid))==NULL)
*
*  PARAMETERS IN:
*      detector - char*  ascii id of detector being used (see exp_detector(3))
*
*  PARAMETERS OUT:
*      pid - int id of process using detector
*
*  FUNCTION RETURNS:
*      NULL - no exposure happening
*           - char* state from the lock file IF an exposure is happening.
*
*  GLOBALS: lock_fd
*
*  -Date- -by- - D e s c r i p t i o n -
*
*  890123  jk  - Original version (from ccd package).
*
****************************************************************************!*/

#define STATELEN 20

char *exp_curstate (detector, pid)
char *detector;
int *pid;
{
    static char exp_state[STATELEN];
    int my_pid, exp_pid;
    char pid_buf[80];
    char *state;
    int dummypid;
    int bytes_read;		  /* number of bytes read from lock file */

    if (!detector)
        return (NULL);

    if (!pid)
        pid = &dummypid;

    exp_pid = 0;

    /* IF file doesn't exist, then there is no exposure in progress */

    if ((lock_fd = exp_maybeopen(detector)) == FAIL) {
        *pid = exp_pid;
        return (NULL);
    }

  /*  If the pid in the lockfile happens to be ours, then don't check stale */
  /* *note* need more error checking to insure not scragged strings in file */

    my_pid = getpid();
    bytes_read = read (lock_fd, pid_buf, 80);
    pid_buf[bytes_read] = '\0';
    sscanf (pid_buf, "%d %s", &exp_pid, exp_state);
    state = exp_state;
    *pid = exp_pid;

    /*  If the pid isn't mine, see if the lock file is writeable; if so,
     *  the other guy has died
     */
    if (my_pid != exp_pid) {
        if (stale(lock_fd)) {
            *pid = 0;
            state = NULL;
        }
    }

    exp_maybeclose (lock_fd);
    return (state);
}

/*!****************************************************************************
*
*  NAME:  int exp_newstate (detector, newstate)
*
*  This function is part of the machinery to change states on in exposure.
*  It operates on a detector-dependent lockfile.
*
*  Essentially, the exp is first locked by opening a lockfile and putting a pid
*  in it (all done elsewhere).  If the file exists and the pid in it isn't
*  stale, then an exposure is truly in progress. We then write the new
*  state into it.
*
*  Note: if the file is created, it will be locked here.
*
*  INVOKED: if (!exp_newstate (detector, newstate))
*              whatever
*
*  PARAMETERS IN:
*      detector - char* name of detector installed.
*      newstate - char* character string with new state as defined in states.h
*
*  PARAMETERS OUT:
*      none
*
*  FUNCTION RETURNS:
*       0 if ok
*      -1 and errno = anything else if it is a system problem
*
*  GLOBALS: lock_fd
*
*  -Date- -by- - D e s c r i p t i o n -
*
*  890123  jk  - Original version
*
****************************************************************************!*/

int exp_newstate (detector, newstate)
char *detector;
char *newstate;
{
    char path[BUF_LEN];
    int my_pid, temp_pid;
    /* jk removed 14 June 1991
    int lock_fd;
    */
    char *state;
    char temp[BUF_LEN];
    BOOLEAN newfile = FALSE;

    /* Essentially lock the exp by opening the lockfile and putting pid in
     * it.  If the file exists and the pid in it isn't stale, then an
     * exposure is truly in progress.
     */

    if (!detector || !newstate) {
	cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(exp_newstate:%d) null argument",
		  __LINE__);
        errno = ENOTTY;
        return (-1);
    }

    my_pid = getpid();

    state = exp_curstate (detector, &temp_pid);

    /********************************************************************
    *  If temp_pid <> 0, then a valid exposure state exists already.
    *  Be careful to preserve the file's pid.
    ********************************************************************/

    if (temp_pid != 0) {
        /* don't nuke pid */
        cfht_logv (CFHT_LIBCFHT, CFHT_DEBUG,
                  "(exp_newstate) exposure in progress; process %d in state '%s'",
                  temp_pid, state);
        if ((lock_fd = exp_maybeopen(detector)) == FAIL) {
            cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
                      "(exp_newstate:%d) lock file failure", __LINE__);
            return (-1);
        } else {
            newfile = FALSE;
        }
    } else {
        /* create the file and lock it */
        /*** if (strcmp (detector, "fake")) *** jk mod 28 mar 89 ***/
        if (!exp_isfake(detector))
            sprintf (path, "%s%s", LF_PATH, detector);
        else
            sprintf (path, "%s/lock%s", getenv("HOME"), detector);

        cfht_logv (CFHT_LIBCFHT, CFHT_DEBUG,
                  "(exp_newstate) no exposure in progress; creating '%s' state '%s'",
                  path, state);
        unlink (path);
        if ((lock_fd = open(path, O_RDWR|O_CREAT|O_TRUNC, 0666)) == -1) {
            cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
                      "(exp_newstate:%d) can't open lock file", __LINE__);
            return (-1);
        } else {
            newfile = TRUE;
            exp_lfopen = TRUE;
            (void)exp_lockf (detector);
        }
    }

    if (newfile) {
        sprintf (temp, "%d\n%s\n", my_pid, newstate);
    } else {
        char *ptr;

        read (lock_fd, temp, BUF_LEN);
        if ((ptr = strchr (temp, '\n')) != 0) {
            /* bump past first line in file and overwrite old state */
            ptr++;
            sprintf (ptr, "%s    \n", newstate);
        } else {
            /* ERROR, only one line found, terminate it and append our line */
            ptr = temp + strlen(temp);
            sprintf (ptr, "\n%s    \n", newstate);
        }

        if (lseek (lock_fd, (off_t)0, SEEK_SET) < 0) {
            cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
                      "(exp_newstate:%d) lseek failed for %s (%s)",
		      __LINE__, detector, strerror(errno));
            return (-1);
	}
    }

    write (lock_fd, temp, strlen (temp));

    cfht_logv (CFHT_LIBCFHT, CFHT_DEBUG,
              "(exp_newstate) OK, state now '%s'", newstate);
    exp_maybeclose (lock_fd);

    return (0);
}

/*!****************************************************************************
*
*  NAME:  int exp_lockf (detector)
*
*  The purpose of this function is to attempt to do a file lock on the
*  logical exposure lockfile.
*
*  This function is part of the machinery to change states on in exposure.
*  It operates on a detector-dependent lockfile.
*
*  INVOKED: if (!exp_lockf (detector))
*              whatever
*
*  PARAMETERS IN:
*      detector - char* name of exp detector installed.
*
*  PARAMETERS OUT:
*      none
*
*  FUNCTION RETURNS:
*       PASS if ok
*       FAIL and errno = anything else if it is a system problem
*
*  GLOBALS: lock_fd
*
*  -Date- -by- - D e s c r i p t i o n -
*
*  881005  jk  - Original version
*
****************************************************************************!*/

int exp_lockf (detector)
char *detector;
{
    /* Essentially lock the exp by opening the lockfile and putting pid in */

    if (!detector)
        return (FAIL);

    if ((lock_fd = exp_maybeopen(detector)) == FAIL)
        return (FAIL);

    if (lockf(lock_fd, F_TLOCK, 0L) == -1) {
        exp_canclose = TRUE;
        exp_maybeclose(lock_fd);
        return (FAIL);
    }

    exp_canclose = FALSE;
    return (PASS);
}

/*!****************************************************************************
*
*  This function sends the signal SIGUSR1.
*
*  INVOKED: IF (exp_kill (pid) != PASS)
*
*  PARAMETERS IN:
*      pid - int process id to get signal
*
*  PARAMETERS OUT:
*      none
*
*  FUNCTION RETURNS:
*      void
*
*  GLOBALS: exp_global_event
*
*  -Date- -by- - D e s c r i p t i o n -
*
*  890123  jk  - Original version
*
****************************************************************************!*/

PASSFAIL exp_kill (pid)
int pid;
{
    if (kill((pid_t)pid, SIGUSR1) == 0)
        return (PASS);

    return (FAIL);
}

/*!****************************************************************************
*
* exp_maybeopen -- internal to this routine, not to be called from elsewhere
*
***************************************************************************!*/

static PASSFAIL exp_maybeopen (detector)
char *detector;
{
    char path[BUF_LEN];

    if (!detector) {
	cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(exp_maybeopen:%d) null argument", __LINE__);
	        return (FAIL);
    }

    if (!exp_lfopen) {
        if (!exp_isfake(detector))
            sprintf (path, "%s%s", LF_PATH, detector);
        else
            sprintf (path, "%s/lock%s",getenv ("HOME"),detector);

        if ((lock_fd = open (path, O_RDWR)) == -1) {
            exp_lfopen = FALSE;
            return (FAIL);
        }
        exp_lfopen = TRUE;
        exp_canclose = TRUE;
    }
    else {
        if (lseek (lock_fd, (off_t)0, SEEK_SET) < 0) {
            cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
                      "(exp_maybeopen:%d) lseek failed for '%s' (%s)",
		      __LINE__, detector, strerror(errno));
            return (FAIL);
	}
    }

    return (lock_fd);
}

/*!****************************************************************************
*
* exp_maybeclose -- internal to this routine, not to be called from elsewhere
*
***************************************************************************!*/

static void exp_maybeclose (my_lock_fd)
int my_lock_fd;
{
    if (exp_canclose) {
        close (my_lock_fd);
        exp_lfopen = FALSE;
    }
    return;
}

/*!****************************************************************************
*
* stale
*
***************************************************************************!*/

int
stale (fd)
int fd;
{
    if (lseek (fd, (off_t)0, SEEK_SET) < 0) {
	cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(exp_states(stale):%d) lseek failed (%s)",
		  __LINE__, strerror(errno));
	return (0);
    }
    if (lockf(fd, F_TEST, 0L) < 0) {
        /* file not lockable, why? */
        if ((errno == EACCES) || (errno == EAGAIN)) {
            cfht_logv(CFHT_LIBCFHT, CFHT_DEBUG,
		      "(exp_states(stale):%d) detector already in use (%s)",
		      __LINE__, strerror(errno));
            return (0); /* because other process lives */
        } else {
            cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		      "(exp_states(stale):%d) fcntl failed (%s)",
		      __LINE__, strerror(errno));
            return (0);
        }
    }
    return (1);
}

/*!****************************************************************************
*
*  File: exp_stop
*
*  Send stop signal
*
*  RETURNS:
*      PASSFAIL
*
*  -Date- -by- - D e s c r i p t i o n -
*
*  880413  jk  - Original version
*
****************************************************************************!*/

PASSFAIL
exp_stop ()
{
    int pid;
    char *state;
    char *detector;

    detector = exp_detector();

    /***********************************************************
     *  Check for an exposure in progress--signified by a pid
     *  returned from exp_curstate.
     **********************************************************/

    if ((state=exp_curstate (detector, &pid)) == NULL) {
        cfht_logv(CFHT_LIBCFHT, CFHT_DEBUG,
		  "(exp_stop:%d) no exposure in progress on '%s'",
		  __LINE__, detector);
        return (FAIL);
    }

    /********************************************
    *  change the state and send SIGUSR1
    ********************************************/

    cfht_logv (CFHT_LIBCFHT, CFHT_DEBUG,
               "(exp_stop) kill being sent to %d; state '%s'", pid, state);

    if (!exp_newstate (detector, EXP_STOP) && !exp_kill(pid)) {
        cfht_log(CFHT_LIBCFHT, CFHT_DEBUG, "(exp_stop) signal sent");
        return (PASS);
    }

    cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY, "(exp_stop:%d) failed on '%s'",
	      __LINE__, detector);
    return (FAIL);
}

/*!****************************************************************************
*
*  File: exp_abort
*
*  Send abort signal to an exposure.
*
*  RETURNS:
*      PASS / FAIL
*
*  -Date- -by- - D e s c r i p t i o n -
*
*  890203  jk  - Original version
*
****************************************************************************!*/

PASSFAIL
exp_abort ()
{
    int pid;
    char *state;
    char *detector;

    detector = exp_detector();

    /***********************************************************
     *  Check for an exposure in progress--signified by a pid
     *  returned from exp_curstate.
     **********************************************************/

    if ((state=exp_curstate (detector, &pid)) == NULL) {
        cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(exp_abort:%d) no exposure in progress on '%s'",
		  __LINE__, detector);
        return (FAIL);
    }

    /********************************************
    *  change the state and send SIGUSR1
    ********************************************/

    cfht_logv (CFHT_LIBCFHT, CFHT_DEBUG,
              "(exp_abort) kill being sent to %d; state '%s'", pid, state);

    if (!exp_newstate (detector, EXP_ABORT) && !exp_kill(pid)) {
        cfht_log(CFHT_LIBCFHT, CFHT_DEBUG, "(exp_abort) signal sent");
        return (PASS);
    }

    cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
	      "(exp_abort:%d) failed for '%s'", __LINE__, detector);
    return (FAIL);
}

/*!****************************************************************************
*
*  File: exp_hold
*
*  Send hold signal to a exposure
*
*  RETURNS:
*      PASS / FAIL
*
*  -Date- -by- - D e s c r i p t i o n -
*
*  890203  jk  - Original version
*
****************************************************************************!*/

PASSFAIL
exp_hold ()
{
    int pid;
    char *state;
    char *detector;

    detector = exp_detector();

    /***********************************************************
     *  Check for an exposure in progress--signified by a pid
     *  returned from exp_curstate.
     **********************************************************/

    if ((state=exp_curstate (detector, &pid)) == NULL) {
        cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
		  "(exp_hold:%d) no exposure in progress on '%s'",
		  __LINE__, detector);
        return (FAIL);
    }

    if (!strcmp (state, EXP_HOLD)) {
        cfht_logv(CFHT_LIBCFHT, CFHT_DEBUG,
		  "(exp_hold:%d) already holding on '%s'",
		  __LINE__, detector);
        return (FAIL);
    }

    /********************************************
    *  change the state and send SIGUSR1
    ********************************************/

    cfht_logv (CFHT_LIBCFHT, CFHT_DEBUG,
               "(exp_hold) kill being sent to %d; state '%s'", pid, state);

    if (!exp_newstate (detector, EXP_HOLD) && !exp_kill(pid)) {
        cfht_log(CFHT_LIBCFHT, CFHT_DEBUG, "(exp_hold) signal sent");
        return (PASS);
    }

    cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
	      "(exp_hold:%d) failed on '%s'", __LINE__, detector);
    return (FAIL);
}

/*!****************************************************************************
*
*  File: exp_restart
*
*  Send restart signal to a exposure
*
*  RETURNS:
*      PASS / FAIL
*
*  -Date- -by- - D e s c r i p t i o n -
*
*  880413  jk  - Original version
*
****************************************************************************!*/

PASSFAIL
exp_restart ()
{
    int pid;
    char *state;
    char *detector;

    detector = exp_detector();

    /***********************************************************
     *  Check for an exposure in progress--signified by a pid
     *  returned from exp_curstate.
     **********************************************************/

    if ((state=exp_curstate (detector, &pid)) == NULL) {
        cfht_logv(CFHT_LIBCFHT, CFHT_DEBUG,
		  "(exp_restart:%d) no exposure in progress on '%s'",
		  __LINE__, detector);
        return (FAIL);
    }

    if (strcmp (state, EXP_HOLD)) {
        cfht_logv(CFHT_LIBCFHT, CFHT_DEBUG,
		  "(exp_restart:%d) no exposure holding on '%s'",
		  __LINE__, detector);
        return (FAIL);
    }

    /********************************************
    *  change the state and send SIGUSR1
    ********************************************/

    cfht_logv (CFHT_LIBCFHT, CFHT_DEBUG,
               "(exp_restart) kill being sent to %d; state '%s'", pid, state);

    if (!exp_newstate (detector, EXP_RESTART) && !exp_kill(pid)) {
        cfht_log(CFHT_LIBCFHT, CFHT_DEBUG, "(exp_restart) signal sent");
        return (PASS);
    }

    cfht_logv(CFHT_LIBCFHT, CFHT_LOGONLY,
	      "(exp_restart:%d) failed on '%s'", __LINE__, detector);
    return (FAIL);
}
