/*
 * Test driver program for testing cfht_exec()
 *
 * Arguments:
 *
 *   If this is the call from test.sh, it should have no arguments.
 *   If this is us calling ourselves to check exec, there should be one
 *    argument (what doesn't matter).
 */

#include <stdio.h>		  /* for printf */
#include <unistd.h>		  /* for sleep, pid_t */

#include "cfht.h"		  /* for CFHT_EXEC*, cfht_exec() */
#include "cfp.h"		  /* for par file to do substitution */
#include "exits.h"		  /* for exit codes */

int ChildCounter;		  /* communication of which child is dying */
int BGExitCode;			  /* exit code for simple BG test */
unsigned int ErrorCount;	  /* how many failures we've seen */

/* Handler routines for background executions */

/*
 * This one is for CFHT_EXECBGH where we just want to see that the exit code
 * comes back right.
 */

void SimpleBGHandler(pid_t Child, int ExitCode)
{
    printf("simple BGH got exit code %d\n", ExitCode);
    BGExitCode = ExitCode;
    return;
}

/*
 * This one is for checking background/foreground interaction.
 * It expects a sequence of background jobs to complete with exit codes
 * starting at the current value in ChildCounter and increasing by one each
 * time.
 */

void SequenceBGHandler(pid_t Child, int ExitCode)
{
    if (ChildCounter == ExitCode) {
	printf("sequence BGH got exit code %d\n", ExitCode);
	ChildCounter++;
    } else {
	printf("FAIL - sequence BGH got exit code %d when expecting %d\n",
	       ExitCode, ChildCounter);
	ErrorCount++;
    }
    return;
}

int main (int argc, char *argv[])
{
    pid_t ChildPID;		  /* return value from cfht_exec() */
    int ExitCode;		  /* exit code for BGC */
    FILE_ID cfpID;		  /* par file ID for substitute */

    ErrorCount = 0;

    /* see if this is the test.sh initial call or our replacement of us */
    if (argc == 1) {
	/* see that plain exec works */
	if ((ChildPID = cfht_exec(-1, "./execor 1", CFHT_EXEC, (void *) NULL))
	    < 0) {
	    printf("FAIL - plain EXEC failed\n");
	    ErrorCount++;
	}
    } else {
	argc--;			  /* fake out parse */
	if ((cfpID = cfp_parse(CFP_PAR, argc, argv)) == FAIL) {
	    printf("FAIL - parse file failed\n");
	    exit(EXIT_FAIL);
	}
    }

    /* see that broken exec fails */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) exec of NoSuchFoo will fail",
	      __LINE__);
    if ((ChildPID = cfht_exec(-1, "./NoSuchFoo", CFHT_EXEC, (void *) NULL))
	!= FAIL) {
	printf("FAIL - plain EXEC of non-existent worked?\n");
	ErrorCount++;
    }

    /* see that plain foreground works with various exit codes */
    if ((ChildPID = cfht_exec(-1, "./execee 0 0", CFHT_EXECFG, (void *) NULL))
	!= 0) {
	printf("FAIL - EXECFG EXIT_PASS returned %d\n", ChildPID);
	ErrorCount++;
    }
    if ((ChildPID = cfht_exec(-1, "./execee 0 -1", CFHT_EXECFG, (void *) NULL))
	!= 255) {
	printf("FAIL - EXECFG -1 returned %d, not 255\n", ChildPID);
	ErrorCount++;
    }
    if ((ChildPID = cfht_exec(-1, "./execee 0 1", CFHT_EXECFG, (void *) NULL))
	!= 1) {
	printf("FAIL - EXECFG EXIT_FAIL returned %d\n", ChildPID);
	ErrorCount++;
    }
    if ((ChildPID = cfht_exec(-1, "./execee 0 2", CFHT_EXECFG, (void *) NULL))
	!= 2) {
	printf("FAIL - EXECFG 2 returned %d\n", ChildPID);
	ErrorCount++;
    }
    if ((ChildPID = cfht_exec(-1, "./execee 0 513", CFHT_EXECFG, (void *) NULL))
	!= 1) {
	printf("FAIL - EXECFG 513 returned %d, not 1\n", ChildPID);
	ErrorCount++;
    }

    /* see that missing foreground fails */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) exec of NoSuchFoo will fail",
	      __LINE__);
    if ((ChildPID = cfht_exec(-1, "./NoSuchFoo", CFHT_EXECFG, (void *) NULL))
	!= FAIL) {
	printf("FAIL - EXECFG non-existent worked?\n");
	ErrorCount++;
    }

    /* check out substitute mode on foreground */
    if ((ChildPID = cfht_exec(cfpID, "%GoodName 0 0", CFHT_EXECFG,
			      (void *) NULL))
	!= 0) {
	printf("FAIL - EXECFG %%GoodName failed (%d)\n", ChildPID);
	ErrorCount++;
    }
    if ((ChildPID = cfht_exec(cfpID, "./execee 0 %EXIT_PASS", CFHT_EXECFG,
			      (void *) NULL))
	!= 0) {
	printf("FAIL - EXECFG %%EXIT_PASS failed (%d)\n", ChildPID);
	ErrorCount++;
    }
    if ((ChildPID = cfht_exec(cfpID, "./execee 0 %EXIT_FAIL", CFHT_EXECFG,
			      (void *) NULL))
	!= 1) {
	printf("FAIL - EXECFG %%EXIT_FAIL failed (%d)\n", ChildPID);
	ErrorCount++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) exec of NoSuchFoo will fail",
	      __LINE__);
    if ((ChildPID = cfht_exec(cfpID, "%BadName 0 0", CFHT_EXECFG,
			      (void *) NULL))
	!= FAIL) {
	printf("FAIL - EXECFG %%BadName did not fail??\n");
	ErrorCount++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) lookup of EXIT_NoSuch will fail",
	      __LINE__);
    if ((ChildPID = cfht_exec(cfpID, "%GoodName 0 %EXIT_NoSuch", CFHT_EXECFG,
			      (void *) NULL))
	!= FAIL) {
	printf("FAIL - EXECFG %%EXIT_NoSuch did not fail\n");
	ErrorCount++;
    }

    /* now check background with just exit code return */
    ExitCode = 55555;		  /* return 0..255 */
    if ((ChildPID = cfht_exec(-1, "./execee 0 25", CFHT_EXECBGC,
			      (void *) &ExitCode))
	== FAIL) {
	printf("FAIL - EXECBGC exec failed\n");
	ErrorCount++;
    }
    (void) sleep(1);
    if (ExitCode != 25) {
	printf("FAIL - EXECBGC returned %d instead of 25\n", ExitCode);
	ErrorCount++;
    }

    /* now try to do a background with callback */
    BGExitCode = 55555;
    if ((ChildPID = cfht_exec(-1, "./execee 0 25", CFHT_EXECBGH,
			      (void *) SimpleBGHandler))
	== FAIL) {
	printf("FAIL - simple EXECBGH exec failed\n");
	ErrorCount++;
    }
    (void) sleep(1);
    if (BGExitCode != 25) {
	printf("FAIL - simple EXECBGH returned %d instead of 25\n",
	       BGExitCode);
	ErrorCount++;
    }

    /*
     * OK, now try to really hit FG and BG to see that the wait loop works
     *
     * The way I'm trying to do this is by starting up a bunch of BGH processes
     * and then waiting on an FG process.  The exit code of the BGH process is
     * used to tell what order they are supposed to finish in.  The exit codes
     * go up as the sleep time goes up, and the processes are started longest
     * sleep first.  The FG process must have the longest sleep time.
     */
    ChildCounter = 1;
    if ((ChildPID = cfht_exec(-1, "./execee 5 5", CFHT_EXECBGH,
			      (void *) SequenceBGHandler))
	== FAIL) {
	printf("FAIL - sequence EXECBGH exec 5 failed\n");
	ErrorCount++;
    }
    if ((ChildPID = cfht_exec(-1, "./execee 4 4", CFHT_EXECBGH,
			      (void *) SequenceBGHandler))
	== FAIL) {
	printf("FAIL - sequence EXECBGH exec 4 failed\n");
	ErrorCount++;
    }
    if ((ChildPID = cfht_exec(-1, "./execee 3 3", CFHT_EXECBGH,
			      (void *) SequenceBGHandler))
	== FAIL) {
	printf("FAIL - sequence EXECBGH exec 3 failed\n");
	ErrorCount++;
    }
    if ((ChildPID = cfht_exec(-1, "./execee 2 2", CFHT_EXECBGH,
			      (void *) SequenceBGHandler))
	== FAIL) {
	printf("FAIL - sequence EXECBGH exec 2 failed\n");
	ErrorCount++;
    }
    if ((ChildPID = cfht_exec(-1, "./execee 1 1", CFHT_EXECBGH,
			      (void *) SequenceBGHandler))
	== FAIL) {
	printf("FAIL - sequence EXECBGH exec 1 failed\n");
	ErrorCount++;
    }
    if ((ChildPID = cfht_exec(-1, "./execee 6 6", CFHT_EXECFG, (void *) NULL))
	!= 6) {
	printf("FAIL - sequence EXECFG exec failed, returned %d\n", ChildPID);
	ErrorCount++;
    }
    if (ChildCounter != 6) {
	printf("FAIL - sequence EXECBGH child count ended at %d not 6\n",
	       ChildCounter);
	ErrorCount++;
    }

    exit(EXIT_PASS);
}
