#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include "cfht.h"
#include "exits.h"

static void errhandler P((pid_t pid, int exitcode));
static void handler P((pid_t pid, int exitcode));
static void report P((char* what));
static void bail P((char* what));


static volatile int errhandler_ran=0;
static volatile int handler_ran=0;
static volatile int handler2_ran=0;

static void
errhandler(pid,exitcode)
  pid_t pid;
  int exitcode;
{
  if (errhandler_ran<0) return;
  if (exitcode!=0 || pid<=0) { errhandler_ran=-1; return; }
  errhandler_ran++;
}

static void
handler(pid,exitcode)
  pid_t pid;
  int exitcode;
{
  if (handler_ran<0) return;
  if (exitcode!=0 || pid<=0) { handler_ran=-1; return; }
  handler_ran++;
}

static void
handler2(pid_t pid, int exitcode)
{
  if (handler2_ran<0) return;
  if (exitcode!=0 || pid<=0) { handler2_ran=-1; return; }
  handler2_ran++;
}

static int iterations=0;

static void
report(what)
  char* what;
{
  fprintf(stdout,"...%s",what);
  fflush(stdout);
}

static void
bail(what)
  char* what;
{
  fprintf(stderr,"\n\n\007*** Error: %s failed (pid=%d gen=%d errno=%d)\n\n",
	  what,(int)getpid(),4-iterations,errno);
  exit(EXIT_FAIL);
}

int
main(int argc, const char* argv[])
{
  volatile int sleepexit=-1;
  int maxchild=0, giveup=0;
  int i;

  cfht_log(CFHT_MAIN, CFHT_LOG_ID, (char*)argv[0]);

  if (argc==3)
  {
    if (sleep(1)!=0)
      bail("unexpected signal");
    else
      exit(EXIT_PASS);
  }

  if (argc<2) iterations = 3; 
  else
  {
    iterations = atoi(argv[1]);
    srand(time(0));
    cfht_sleep(rand()%1000,0);
  }

  for (i=0; i<iterations; i++)
  {
    if (i==0)
    {
      char cmdstr[256];
      handler_ran=0;
      maxchild=0;
      giveup=0;
      sprintf(cmdstr,"%s %d sleep",argv[0],iterations-1);
      while (++giveup<=300 && 
	     cfht_exec(-1,cmdstr,CFHT_EXECBGH,(void*)handler)>0)
	maxchild++;
      if (errno!=ENOSPC) 
	bail("EXEC_NUM");
      while ((giveup+=10)<=300 && handler_ran<maxchild) sleep(1);
      sleep(1);
      if (handler_ran!=maxchild || maxchild<CFHT_EXEC_NUM)
      {
	fprintf(stderr,"...EXEC_NUM(%d)...failed (%d,%d)\n",
		CFHT_EXEC_NUM,maxchild,handler_ran);
	bail("EXEC_NUM");
      }
      report("EXEC_NUM");
    }

    if (cfht_exec(-1,"/bin/false",CFHT_EXECFG,NULL)<=0 ||
        cfht_exec(-1,"/bin/true",CFHT_EXECFG,NULL)!=EXIT_PASS)
      bail("EXECFG");
    else
      report("EXECFG");

    if (cfht_exec(-1,"/bin/false",CFHT_EXECBG,NULL)<0)
      bail("EXECBG");
    else
      report("EXECBG");

#ifndef NO_RECURSE
    {
      char cmdline[256];
      sprintf(cmdline,"%s %d",argv[0],iterations-1);
      if (cfht_exec(-1,cmdline,CFHT_EXECBGC,(void*)&sleepexit)>0)
      {
        while (sleepexit==-1) cfht_sleep(50,0);
        if (sleepexit!=0)
	{
	  fprintf(stderr,"...EXECBGC,sleepexit=%d...\n",sleepexit);
          bail("EXECBGC");
	}
        else
          report("EXECBGC");
      }
      else
        bail("EXECBGC");

      /*
      if (cfht_exec(-1,cmdline,CFHT_EXECBGH,(void*)errhandler)<=0)
        bail("EXECBGH");
	*/
      if (cfht_exec(-1,cmdline,CFHT_EXECBGH,(void*)errhandler)<=0)
        bail("EXECBGH");
      else
        report("EXECBGH");
    }
#endif

    {
      char cmdstr[256];

      handler2_ran=0;
      sprintf(cmdstr,"%s %d sleep",argv[0],iterations-1);
      if (cfht_exec(-1,cmdstr,CFHT_EXECBGH,(void*)handler2)>0)
      {
          while (!handler2_ran) cfht_sleep(50,0);
          report("EXECBGH");
      }
      else
        bail("EXECBGH");
    }
  }

  {
    int pipefd[2];
    char cmdstr[256];

    pipe(pipefd);

    if (!fork())
    {
      close(pipefd[0]);
      while (sleep(10)!=0);
      write(pipefd[1],"X",1);
      _exit(0);
    }

    close(pipefd[1]);
    
    sprintf(cmdstr,"%s %d sleep",argv[0],iterations-1);
    if (cfht_exec(-1,cmdstr,CFHT_EXECBG,NULL)>0)

    {
      char rd;

      errno=0;
      if (read(pipefd[0],&rd,1)<=0)
      {
	if (errno!=EINTR)
          bail("syscall interrupt");
        else
          report("EINTR");
      }
      else
	bail("syscall interrupt");
    }
  }

  if (errhandler_ran==-1)
    bail("recursive test");

  cfht_exec(-1,"/bin/echo ...EXEC",CFHT_EXEC,NULL);
  bail("EXEC");
  exit(EXIT_FAIL);
}
