#!/bin/sh
#
# Copyright (C) 1994 Canada-France-Hawaii Telescope Corporation
#
# Test script to drive testing of cfht_log.c
#
# History
#
# JT	20-Sep-94	created
# JT	28-Nov-94	changed to match debugging output moving to LOGU
#			and added LIBMUSIC
#

#
# function to clear out all the CFHT environment variables
#
# call:
#  ClearEnviroment
#
ClearEnvironment() {
  unset CFHTLOGS
  unset CFHTLOGU
  CFHTCOLORLOG=false;export CFHTCOLORLOG #display options
  CFHTERROR=true;export CFHTERROR
  CFHTWARN=true;export CFHTWARN
  CFHTNODISP=false; export CFHTNODISP
  CFHTDEBUG=false;export CFHTDEBUG
  for Lib in $Libraries
    do
      eval CFHT$Lib=false;export CFHT$Lib
    done;
}

#
# function to build and run a test program with a given set of log files
# and environment
#
# call:
#  TestIt <program-name.c> <system-log> <user-log> <environment-list>
# e.g.:
#  TestIt lib-all.c $NullLog $UserLog WARN=false LIBFF=true
#
TestIt () {
  Program=$1			# save arguments
  SLog=$2
  ULog=$3
  shift 3
  EnvList=$*

  ClearEnvironment		# set up desired environment
  if [ $SLog != $DummyLog ]
    then
      CFHTLOGS=$SLog;export CFHTLOGS
  fi
  if [ $ULog != $DummyLog ]
    then
      CFHTLOGU=$ULog;export CFHTLOGU
  fi
  for Env in $EnvList
    do
      SaveIFS=$IFS
      IFS="="
      set $Env
      IFS=$SaveIFS
      eval CFHT$Env;export CFHT$1
    done

  echo "Testing $Program $SLog $ULog $EnvList"
  Executable=`basename $Program .c`
  make $Executable
  Status=$?
  if [ $Status -ne 0 ]
    then
      echo "******* Error making $Program, skipping"
    else
      echo "------- Running $Program, stdout and stderr follow"
      ./$Executable
      Status=$?
      if [ $Status -ne 0 ]
	then
	  echo "xxxxxxx $Executable failed, code=$Status"
      fi
      echo "+++++++   user log output $ULog"
      if [ $ULog = $NullLog ]
	then
	  echo "-- went to $NullLog"
	else
	  if [ -f $ULog ]
	    then
	      if [ -s $ULog ]
		then
		  cat $ULog
		else
		  echo "-- empty file"
	      fi
	      /bin/rm $ULog
	    else
	      echo "-- no file created"
	  fi
      fi
      echo "+++++++   system log output $SLog"
      if [ $SLog = $NullLog ]
	then
	  echo "-- went to $NullLog"
	else
	  if [ -f $SLog ]
	    then
	      if [ -s $SLog ]
		then
		  cat $SLog
		else
		  echo "-- empty file"
	      fi
	      /bin/rm $SLog
	    else
	      echo "-- no file created"
	  fi
      fi
  fi
  echo "------- Done $Program";
}

#
# set up known environment
#
LogFilePrefix=/usr/tmp/Test-
SystemLog=${LogFilePrefix}$$-system	# log file names
UserLog=${LogFilePrefix}$$-user
NullLog=/dev/null
DummyLog="=dummy="			# for testing missing LOGS/LOGU
Libraries="LIB LIBCFHT LIBCFP LIBCAMAC LIBCCD LIBFF LIBHH LIBMUSIC LIBOCS LIBRET"

ClearEnvironment

#
# first test basic errors
#
echo "==================================================================="
echo "should be no program output, short system and user logs"
TestIt BasicCheck.c $SystemLog $UserLog
echo "==================================================================="
echo "should be user log to stderr, no system or user logs"
TestIt BasicCheck.c $DummyLog $DummyLog
echo "==================================================================="
echo "should be failed system log, missing system log, short user log"
TestIt BasicCheck.c /tmp/A/B/C/D/E/F/G $UserLog
echo "==================================================================="
echo "should be failed user log, short system log, missing user log"
TestIt BasicCheck.c $SystemLog /tmp/A/B/C/D/E/F/G
echo "==================================================================="
echo "should be missing LOG_ID call"
TestIt NoLOG_ID.c $SystemLog $UserLog
echo "==================================================================="
echo "should be too long identifier in LOG_ID call"
TestIt LongLOG_ID.c $SystemLog $UserLog
#
# now test all the message types
#
echo "==================================================================="
echo "should be one of each message type (except DEBUG)"
TestIt AllTypes.c $SystemLog $UserLog
echo "==================================================================="
echo "should turn off error and warn messages"
TestIt AllTypes.c $SystemLog $UserLog ERROR=false
echo "==================================================================="
echo "should turn off warn messages"
TestIt AllTypes.c $SystemLog $UserLog WARN=false
echo "==================================================================="
echo "should turn off status messages to LOGU"
TestIt AllTypes.c $SystemLog $UserLog NODISP=true
#
# now test all callers (who)
#
echo "==================================================================="
echo "should have one status from main and each library"
TestIt AllWhos.c $SystemLog $UserLog
echo "==================================================================="
echo "should have all above status messages plus main debug"
TestIt AllWhos.c $SystemLog $UserLog DEBUG=true
echo "==================================================================="
echo "should have main debug"
TestIt AllWhos.c $SystemLog $UserLog DEBUG=true NODISP=true
echo "==================================================================="
echo "should have all library debugs"
TestIt AllWhos.c $SystemLog $UserLog LIB=true NODISP=true
echo "==================================================================="
echo "should have cfht library debug"
TestIt AllWhos.c $SystemLog $UserLog LIBCFHT=true NODISP=true
echo "==================================================================="
echo "should have cfp library debug"
TestIt AllWhos.c $SystemLog $UserLog LIBCFP=true NODISP=true
echo "==================================================================="
echo "should have CAMAC library debug"
TestIt AllWhos.c $SystemLog $UserLog LIBCAMAC=true NODISP=true
echo "==================================================================="
echo "should have ccd library debug"
TestIt AllWhos.c $SystemLog $UserLog LIBCCD=true NODISP=true
echo "==================================================================="
echo "should have FITS library debug"
TestIt AllWhos.c $SystemLog $UserLog LIBFF=true NODISP=true
echo "==================================================================="
echo "should have high HP-IB library debug"
TestIt AllWhos.c $SystemLog $UserLog LIBHH=true NODISP=true
echo "==================================================================="
echo "should have Music library debug"
TestIt AllWhos.c $SystemLog $UserLog LIBMUSIC=true NODISP=true
echo "==================================================================="
echo "should have OCS library debug"
TestIt AllWhos.c $SystemLog $UserLog LIBOCS=true NODISP=true
echo "==================================================================="
echo "should have reticon library debug"
TestIt AllWhos.c $SystemLog $UserLog LIBRET=true NODISP=true
echo "==================================================================="
echo "should have all library debugs to stderr only"
TestIt AllWhos.c $DummyLog $DummyLog LIB=true NODISP=true
echo "==================================================================="
echo "try several logv calls"
TestIt Hit-logv.c $SystemLog $UserLog
echo "==================================================================="
echo done
#
# now clean up
#
make clean
rm -f /usr/tmp/${LogFilePrefix}*
