/*
 * Check_file.c - test cfp_file
 *
 * This program is used to make sure that cfp_file() works correctly.  Its
 * expected environment is that the calling shell script has copied a .par file
 * somewhere and expects this routine to find it there.  This is called with
 * three arguments, the .par file name, the expected path to the file (without
 * the file name), and the mode ("rf", "r", or "rw").  This routine calls
 * cfp_file three times, once with .par file name as the cfp_file name
 * parameter, once as the suffix parameter, and once split into name and
 * suffix.  All three answers are compared to the expected path.  If any of
 * them do not match, a message is written to stdout (any cfht_log messages
 * are expected to go to stderr), and this exits non-zero, with the status
 * value the count of errors.  If all match this exits zero.  If there are
 * any errors in the call to this program, it exits 99.
 *
 * CFHTLOGS and CFHTLOGU are expected to be undefined (though this doesn't
 * care).
 *
 * Note that if "rw" causes the file to be copied on the first cfp_file call,
 * the next two calls will find that new copy and return the same value, if
 * everything is correct.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "cfht.h"
#include "cfp.h"

main (argc, argv)
    int argc;
    char *argv [];
{
    char Name[CFP_PATHSIZE];	  /* name portion of argv[1] */
    char Suffix[CFP_PATHSIZE];	  /* "extension" portion of argv[1] */
    size_t Index;		  /* index into argv[1] for splitting */
    char Path[CFP_PATHSIZE];	  /* a place to put the answer */
    char *Answer;		  /* cfp_file's return value */
    char Expected[CFP_PATHSIZE];  /* what cfp_file should return */
    char *SessionDef;		  /* value of $Session */
    char SessionHOME[CFP_PATHSIZE]; /*   /h/$Session */
    int ReturnValue;		  /* starts at zero, +1 for each error */

    if (argc != 4) {
	printf("usage: %s <.par file name> <expected path> <mode>\n", argv[0]);
	return 99;
    }

    cfht_log(CFHT_MAIN, CFHT_LOG_ID, "file");

    /* split .par file name apart */
    if (strlen(argv[1]) > CFP_PATHSIZE) {
	printf(".par file name '%s' too long\n", argv[1]);
	return 99;
    }
    Index = strcspn(argv[1], ".");
    if (Index == strlen(argv[1])) {
	printf("par file name '%s' has no .\n", argv[1]);
	return 99;
    }
    argv[1][Index] = '\0';
    strcpy(Name, argv[1]);
    argv[1][Index] = '.';
    strcpy(Suffix, &argv[1][Index]);

    /* combine name and path into one */
    if ((strlen(argv[1]) + strlen(argv[2]) + 2) > CFP_PATHSIZE) {
	printf("expected path '%s/%s' too long\n", argv[1], argv[2]);
	return 99;
    }
    (void) strcpy(Expected, argv[2]);
    if (strlen(Expected) > 0) {
	(void) strcat(Expected, "/");
    }
    if (strcmp(argv[2], getenv("HOME")) == 0) {
	(void) strcat(Expected, ".,");
    }
    if ((SessionDef = getenv("Session")) == (char *) NULL) {
	SessionDef = "pegasus";
    }
    (void) strcpy(SessionHOME, "/h/");
    (void) strcat(SessionHOME, SessionDef);
    if (strcmp(argv[2], SessionHOME) == 0) {
	(void) strcat(Expected, ".,");
    }
    (void) strcat(Expected, argv[1]);
    ReturnValue = 0;

    /* try with name as name arg to cfp_file */
    Answer = cfp_file(Path, argv[1], (char *) NULL, argv[3]);
    if (Answer == (char *) NULL) {
	ReturnValue++;
	printf("cfp_file failed on %s as name argument\n", argv[1]);
    } else {
	if (Answer != Path) {
	    ReturnValue++;
	    printf("cfp_file name return (%08x) different from arg (%08x)\n",
		   Answer, Path);
	    if (strcmp(Answer, Path) != 0) {
		ReturnValue++;
		printf("cfp_file name return (%s) different from arg (%s)\n",
		       Answer, Path);
	    }
	}
	if (strcmp(Answer, Expected) != 0) {
	    ReturnValue++;
	    printf("cfp_file name return (%s) different from expected (%s)\n",
		   Answer, Expected);
	}
    }

    /* try with name as suff arg to cfp_file */
    Answer = cfp_file(Path, (char *) NULL, argv[1], argv[3]);
    if (Answer == (char *) NULL) {
	ReturnValue++;
	printf("cfp_file failed on %s as suff argument\n", argv[1]);
    } else {
	if (Answer != Path) {
	    ReturnValue++;
	    printf("cfp_file suff return (%08x) different from arg (%08x)\n",
		   Answer, Path);
	    if (strcmp(Answer, Path) != 0) {
		ReturnValue++;
		printf("cfp_file suff return (%s) different from arg (%s)\n",
		       Answer, Path);
	    }
	}
	if (strcmp(Answer, Expected) != 0) {
	    ReturnValue++;
	    printf("cfp_file suff return (%s) different from expected (%s)\n",
		   Answer, Expected);
	}
    }

    /* try with name split as name and suff args to cfp_file */
    Answer = cfp_file(Path, Name, Suffix, argv[3]);
    if (Answer == (char *) NULL) {
	ReturnValue++;
	printf("cfp_file failed on %s as split argument\n", argv[1]);
    } else {
	if (Answer != Path) {
	    ReturnValue++;
	    printf("cfp_file split return (%08x) different from arg (%08x)\n",
		   Answer, Path);
	    if (strcmp(Answer, Path) != 0) {
		ReturnValue++;
		printf("cfp_file split return (%s) different from arg (%s)\n",
		       Answer, Path);
	    }
	}
	if (strcmp(Answer, Expected) != 0) {
	    ReturnValue++;
	    printf("cfp_file split return (%s) different from expected (%s)\n",
		   Answer, Expected);
	}
    }

    return ReturnValue;
    /*NOTREACHED*/
}
