/*
 * Check_getcur.c - test cfp_getcur()
 *
 * This program is used to make sure that cfp_getcur() works correctly.  Its
 * expected environment is that the calling shell script has created a .par
 * file as defined below and called this program with no arguments.  This runs
 * through a set of tests outputing any unexpected results to stdout.  Each
 * output starts with "FAIL" for the calling shell script to find.  If any
 * getcur errors occurred, this exits with a positive error count.  If any
 * parse errors occurred, this exits with -1.
 *
 * The expected .par file:
 *
 * const = some value
 * bool1 = { , True }
 * bool2 = { nd, , True }
 * bool3 = { ns, , , True }
 * bool4 = { nds, , True, On }
 * bool5 = { nds, , , On }
 * int1 = { , 1 }
 * int2 = { , 2, , , int }
 * int3 = { , 3, , , "[0,10]" }
 * int4 = { , -1, , , "[0,10]" }
 * int5 = { , 11, , , "[0,10]" }
 * int6 = { , abc, , , int }
 * double1 = { , 1 }
 * double2 = { , 2, , , double }
 * double3 = { , 3, , , "[0.0,10.0]" }
 * double4 = { , -0.001, , , "[0.0,10.0]" }
 * double5 = { , 10.0001, , , "[0.0,10.0]" }
 * double6 = { , abc, , , double }
 * time1 = { , "0:1:0" }
 * time2 = { , "1:0:0", , , double }
 * time3 = { , "0:0:0", , , "(0:0:0,1:0:0)" }
 * time4 = { , "5:5:5", , , "[0:0:0,1:0:0]" }
 * time5 = { , "0:5:0abc", , , double }
 * string = { , foo }
 *
 *
 * CFHTLOGS and CFHTLOGU are expected to be undefined (though this doesn't
 * care)
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "cfht.h"
#include "cfp.h"
#include "exits.h"

main(argc, argv)
    int argc;
    char *argv[];
{
    FILE_ID FileId;		  /* file index from parse */
    BOOLEAN TestBoolean;	  /* scratch areas  */
    int TestInteger;		  /*   for          */
    double TestDouble;		  /*     getcur     */
    double TestTime;		  /*       provided */
    char TestString[CFP_STRSIZE]; /*         values */
    HistoryType History;	  /* getcur return value */
    int ReturnValue;		  /* starts at zero, +1 for each error */

    ReturnValue = 0;

    if ((FileId = cfp_parse(CFP_PAR, argc, argv)) == FAIL) {
	printf("FAIL - cfp_parse died on %s\n", argv[0]);
	return(EXIT_FAIL);
    }

    /* const = some value */
    if ((History = cfp_getcur(FileId, "const", (void *) TestString,
			     CFP_STRING, sizeof TestString)) == FAIL) {
	printf("FAIL - cfp_getcur of constant failed\n");
	ReturnValue++;
    } else {
	if (strcmp(TestString, "some value") != 0) {
	    printf("FAIL - cfp_getcur of const returned '%s' not '%s'\n",
		   TestString, "some value");
	    ReturnValue++;
	}
    }

    /* bool1 = { , True } */
    if ((History = cfp_getcur(FileId, "bool1", (void *) &TestBoolean,
			      CFP_BOOLEAN, sizeof TestBoolean)) == FAIL) {
	printf("FAIL - cfp_getcur of bool1 failed\n");
	ReturnValue++;
    } else {
	if (TestBoolean != TRUE) {
	    printf("FAIL - cfp_getcur of bool1 returned %d not %d\n",
		   TestBoolean, TRUE);
	    ReturnValue++;
	}
	/* check History? */
    }

    /* bool2 = { nd, , True } */
    if ((History = cfp_getcur(FileId, "bool2", (void *) &TestBoolean,
			      CFP_BOOLEAN, sizeof TestBoolean)) == FAIL) {
	printf("FAIL - cfp_getcur of bool2 failed\n");
	ReturnValue++;
    } else {
	if (TestBoolean != TRUE) {
	    printf("FAIL - cfp_getcur of bool2 returned %d not %d\n",
		   TestBoolean, TRUE);
	    ReturnValue++;
	}
	/* check History? */
    }

    /* bool3 = { ns, , , True } */
    if ((History = cfp_getcur(FileId, "bool3", (void *) &TestBoolean,
			      CFP_BOOLEAN, sizeof TestBoolean)) == FAIL) {
	printf("FAIL - cfp_getcur of bool3 failed\n");
	ReturnValue++;
    } else {
	if (TestBoolean != FALSE) {
	    printf("FAIL - cfp_getcur of bool3 returned %d not %d\n",
		   TestBoolean, FALSE);
	    ReturnValue++;
	}
	/* check History? */
    }

    /* bool4 = { nds, , True, On } */
    if ((History = cfp_getcur(FileId, "bool4", (void *) TestString,
			      CFP_STRING, sizeof TestString)) == FAIL) {
	printf("FAIL - STRING cfp_getcur of bool4 failed\n");
	ReturnValue++;
    } else {
	if (strcmp(TestString, "True") != 0) {
	    printf("FAIL - STRING cfp_getcur of bool4 returned '%s' not '%s'\n",
		   TestString, "True");
	    ReturnValue++;
	}
	/* check History? */
    }
    if ((History = cfp_getcur(FileId, "bool4", (void *) &TestBoolean,
			      CFP_BOOLEAN, sizeof TestBoolean)) == FAIL) {
	printf("FAIL - BOOLEAN cfp_getcur of bool4 failed\n");
	ReturnValue++;
    } else {
	if (TestBoolean != TRUE) {
	    printf("FAIL - BOOLEAN cfp_getcur of bool4 returned %d not %d\n",
		   TestBoolean, TRUE);
	    ReturnValue++;
	}
	/* check History? */
    }

    /* bool5 = { nds, , , On } */
    if ((History = cfp_getcur(FileId, "bool5", (void *) TestString,
			      CFP_STRING, sizeof TestString)) == FAIL) {
	printf("FAIL - STRING cfp_getcur of bool5 failed\n");
	ReturnValue++;
    } else {
	if (strcmp(TestString, "") != 0) {
	    printf("FAIL - STRING cfp_getcur of bool5 returned '%s' not ''\n",
		   TestString);
	    ReturnValue++;
	}
	/* check History? */
    }

    /* int1 = { , 1 } */
    if ((History = cfp_getcur(FileId, "int1", (void *) &TestInteger,
			      CFP_INTEGER, sizeof TestInteger)) == FAIL) {
	printf("FAIL - cfp_getcur of int1 failed\n");
	ReturnValue++;
    } else {
	if (TestInteger != 1) {
	    printf("FAIL - cfp_getcur of int1 returned %d not %d\n",
		   TestInteger, 1);
	    ReturnValue++;
	}
	/* check History? */
    }

    /* int2 = { , 2, , , int } */
    if ((History = cfp_getcur(FileId, "int2", (void *) &TestInteger,
			      CFP_INTEGER, sizeof TestInteger)) == FAIL) {
	printf("FAIL - cfp_getcur of int2 failed\n");
	ReturnValue++;
    } else {
	if (TestInteger != 2) {
	    printf("FAIL - cfp_getcur of int2 returned %d not %d\n",
		   TestInteger, 2);
	    ReturnValue++;
	}
	/* check History? */
    }

    /* int3 = { , 3, , , "[0,10]" } */
    if ((History = cfp_getcur(FileId, "int3", (void *) &TestInteger,
			      CFP_INTEGER, sizeof TestInteger)) == FAIL) {
	printf("FAIL - cfp_getcur of int3 failed\n");
	ReturnValue++;
    } else {
	if (TestInteger != 3) {
	    printf("FAIL - cfp_getcur of int3 returned %d not %d\n",
		   TestInteger, 3);
	    ReturnValue++;
	}
	/* check History? */
    }

    /* int4 = { , -1, , , "[0,10]" } */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if ((History = cfp_getcur(FileId, "int4", (void *) &TestInteger,
			      CFP_INTEGER, sizeof TestInteger)) != FAIL) {
	printf("FAIL - cfp_getcur of int4 did not fail, returned %d\n",
	       TestInteger);
	ReturnValue++;
    }

    /* int5 = { , 11, , , "[0,10]" } */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if ((History = cfp_getcur(FileId, "int5", (void *) &TestInteger,
			      CFP_INTEGER, sizeof TestInteger)) != FAIL) {
	printf("FAIL - cfp_getcur of int5 did not fail, returned %d\n",
	       TestInteger);
	ReturnValue++;
    }

    /* int6 = { , abc, , , int } */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if ((History = cfp_getcur(FileId, "int6", (void *) &TestInteger,
			      CFP_INTEGER, sizeof TestInteger)) != FAIL) {
	printf("FAIL - cfp_getcur of int6 did not fail, returned %d\n",
	       TestInteger);
	ReturnValue++;
    }

    /* double1 = { , 1 } */
    if ((History = cfp_getcur(FileId, "double1", (void *) &TestDouble,
			      CFP_DOUBLE, sizeof TestDouble)) == FAIL) {
	printf("FAIL - cfp_getcur of double1 failed\n");
	ReturnValue++;
    } else {
	if (TestDouble != 1.0) {
	    printf("FAIL - cfp_getcur of double1 returned %g not %g\n",
		   TestDouble, 1.0);
	    ReturnValue++;
	}
	/* check History? */
    }

    /* double2 = { , 2, , , double } */
    if ((History = cfp_getcur(FileId, "double2", (void *) &TestDouble,
			      CFP_DOUBLE, sizeof TestDouble)) == FAIL) {
	printf("FAIL - cfp_getcur of double2 failed\n");
	ReturnValue++;
    } else {
	if (TestDouble != 2.0) {
	    printf("FAIL - cfp_getcur of double2 returned %g not %g\n",
		   TestDouble, 2.0);
	    ReturnValue++;
	}
	/* check History? */
    }

    /* double3 = { , 3, , , "[0.0,10.0]" } */
    if ((History = cfp_getcur(FileId, "double3", (void *) &TestDouble,
			      CFP_DOUBLE, sizeof TestDouble)) == FAIL) {
	printf("FAIL - cfp_getcur of double3 failed\n");
	ReturnValue++;
    } else {
	if (TestDouble != 3.0) {
	    printf("FAIL - cfp_getcur of double3 returned %g not %g\n",
		   TestDouble, 3.0);
	    ReturnValue++;
	}
	/* check History? */
    }

    /* double4 = { , -0.001, , , "[0.0,10.0]" } */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if ((History = cfp_getcur(FileId, "double4", (void *) &TestDouble,
			      CFP_DOUBLE, sizeof TestDouble)) != FAIL) {
	printf("FAIL - cfp_getcur of double4 did not fail, returned %g\n",
	       TestDouble);
	ReturnValue++;
    }

    /* double5 = { , 10.0001, , , "[0.0,10.0]" } */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if ((History = cfp_getcur(FileId, "double5", (void *) &TestDouble,
			      CFP_DOUBLE, sizeof TestDouble)) != FAIL) {
	printf("FAIL - cfp_getcur of double5 did not fail, returned %g\n",
	       TestDouble);
	ReturnValue++;
    }

    /* double6 = { , abc, , , double } */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if ((History = cfp_getcur(FileId, "double6", (void *) &TestDouble,
			      CFP_DOUBLE, sizeof TestDouble)) != FAIL) {
	printf("FAIL - cfp_getcur of double6 did not fail, returned %g\n",
	       TestDouble);
	ReturnValue++;
    }

    /* time1 = { , "0:1:0" } */
    if ((History = cfp_getcur(FileId, "time1", (void *) &TestTime,
			      CFP_DOUBLE, sizeof TestTime)) == FAIL) {
	printf("FAIL - cfp_getcur of time1 failed\n");
	ReturnValue++;
    } else {
	if (TestTime != 60.0) {
	    printf("FAIL - cfp_getcur of time1 returned %g not %g\n",
		   TestTime, 60.0);
	    ReturnValue++;
	}
	/* check History? */
    }

    /* time2 = { , "1:0:0", , , double } */
    if ((History = cfp_getcur(FileId, "time2", (void *) &TestTime,
			      CFP_DOUBLE, sizeof TestTime)) == FAIL) {
	printf("FAIL - cfp_getcur of time2 failed\n");
	ReturnValue++;
    } else {
	if (TestTime != 3600.0) {
	    printf("FAIL - cfp_getcur of time2 returned %g not %g\n",
		   TestTime, 3600.0);
	    ReturnValue++;
	}
	/* check History? */
    }

    /* time3 = { , "0:0:0", , , "(0:0:0,1:0:0)" } */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if ((History = cfp_getcur(FileId, "time3", (void *) &TestTime,
			      CFP_DOUBLE, sizeof TestTime)) != FAIL) {
	printf("FAIL - cfp_getcur of time3 did not fail, returned %g\n",
	       TestTime);
	ReturnValue++;
    }

    /* time4 = { , "5:5:5", , , "[0:0:0,1:0:0]" } */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if ((History = cfp_getcur(FileId, "time4", (void *) &TestTime,
			      CFP_DOUBLE, sizeof TestTime)) != FAIL) {
	printf("FAIL - cfp_getcur of time4 did not fail, returned %g\n",
	       TestTime);
	ReturnValue++;
    }

    /* time5 = { , "0:5:0abc", , , double } */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if ((History = cfp_getcur(FileId, "time5", (void *) &TestTime,
			      CFP_DOUBLE, sizeof TestTime)) != FAIL) {
	printf("FAIL - cfp_getcur of time5 did not fail, returned %g\n",
	       TestTime);
	ReturnValue++;
    }

    /* string = { , foo } */
    if ((History = cfp_getcur(FileId, "string", (void *) TestString,
			      CFP_STRING, sizeof TestString)) == FAIL) {
	printf("FAIL - cfp_getcur of string failed\n");
	ReturnValue++;
    } else {
	if (strcmp(TestString, "foo") != 0) {
	    printf("FAIL - cfp_getcur of string returned %s not %s\n",
		   TestString, "foo");
	    ReturnValue++;
	}
	/* check History? */
    }

    /* non-existent name */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with lookup failure",
	      __LINE__);
    if ((History = cfp_getcur(FileId, "nosuchvar", (void *) TestString,
			      CFP_STRING, sizeof TestString)) != FAIL) {
	printf("FAIL - cfp_getcur of nosuchvar did not fail, returned %s\n",
	       TestString);
	ReturnValue++;
    }

    if (cfp_finish(FileId) != PASS) {
	printf("FAIL - cfp_finish failed\n");
	ReturnValue++;
    }

    return(ReturnValue);
    /*NOTREACHED*/
}
