/*
 * Check_getval.c - test cfp_getval()
 *
 * This program is used to make sure that cfp_getval() works correctly.  Its
 * expected environment is that the calling shell script has created a .par
 * file and an environment variable as defined below and called this program
 * with no arguments.  This runs through a set of tests outputting any
 * unexpected results to stdout.  Each output starts with "FAIL" for the
 * calling shell script to find.  If any getval errors occurred, this exits
 * with a positive error count.  If any parse errors occurred, this exits with
 * -1.
 *
 * The expected .par file:
 *
 * const = this should fail
 * var = { dns, value, default, supplied, range }
 * bool = { dns, original, false, true }
 * int1 = { , 1 }
 * int2 = { , 2, , , "[0,10]" }
 * int3 = { , 30, , , "[0,10]" }
 * double = { , 5.0 }
 * TEST = {  }
 * INT = {  }
 * NOT_there = {  }
 *
 * The expected environment variables:
 *
 * Check_getvalTEST="abc 123"
 * Check_getvalINT=123
 *
 *
 * CFHTLOGS and CFHTLOGU are expected to be undefined (though this doesn't
 * care)
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "cfht.h"
#include "cfp.h"
#include "exits.h"

main(argc, argv)
    int argc;
    char *argv[];
{
    FILE_ID FileId;		  /* file index from parse */
    BOOLEAN TestBool;		  /* spots for    */
    int TestInt;		  /*   getval     */
    double TestDbl;		  /*     returned */
    char TestString[CFP_STRSIZE + 1]; /*   values */
    char *EnvAddress;		  /* pointer to environment string */
    int ReturnValue;		  /* starts at zero, +1 for each error */

    ReturnValue = 0;

    if ((FileId = cfp_parse(CFP_PAR, argc, argv)) == FAIL) {
	printf("FAIL - cfp_parse died on %s\n", argv[0]);
	return(EXIT_FAIL);
    }

    /* var = { dns, value, default, supplied, range } */
    /* make sure all the fields return the correct values */
    if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		   sizeof TestString, CFP_ModeIndex) != PASS) {
	printf("FAIL - getval failed getting mode string\n");
	ReturnValue++;
    } else {
	if (strcmp(TestString, "dns") != 0) {
	    printf("FAIL - getval mode value '%s' != 'dns'\n", TestString);
	    ReturnValue++;
	}
    }
    if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		   sizeof TestString, CFP_CurrentIndex) != PASS) {
	printf("FAIL - getval failed getting current string\n");
	ReturnValue++;
    } else {
	if (strcmp(TestString, "value") != 0) {
	    printf("FAIL - getval current value '%s' != 'value'\n", TestString);
	    ReturnValue++;
	}
    }
    if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		   sizeof TestString, CFP_DefaultIndex) != PASS) {
	printf("FAIL - getval failed getting default string\n");
	ReturnValue++;
    } else {
	if (strcmp(TestString, "default") != 0) {
	    printf("FAIL - getval default value '%s' != 'default\n",
		   TestString);
	    ReturnValue++;
	}
    }
    if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		   sizeof TestString, CFP_SuppliedIndex) != PASS) {
	printf("FAIL - getval failed getting supplied string\n");
	ReturnValue++;
    } else {
	if (strcmp(TestString, "supplied") != 0) {
	    printf("FAIL - getval supplied value '%s' != 'supplied'\n",
		   TestString);
	    ReturnValue++;
	}
    }
    if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		   sizeof TestString, CFP_RangeIndex) != PASS) {
	printf("FAIL - getval failed getting range string\n");
	ReturnValue++;
    } else {
	if (strcmp(TestString, "range") != 0) {
	    printf("FAIL - getval range value '%s' != 'range'\n", TestString);
	    ReturnValue++;
	}
    }

    /* bool = { dns, original, false, true } */
    /* make sure "ds" mode does not affect getval */
    if (cfp_getval(FileId, "bool", (void *) TestString, CFP_STRING,
		   sizeof TestString, CFP_CurrentIndex) != PASS) {
	printf("FAIL - getval failed getting boolean as string\n");
	ReturnValue++;
    } else {
	if (strcmp(TestString, "original") != 0) {
	    printf("FAIL - getval current '%s' != 'original'\n", TestString);
	    ReturnValue++;
	}
    }
    if (cfp_getval(FileId, "bool", (void *) &TestBool, CFP_BOOLEAN,
		   sizeof TestBool, CFP_CurrentIndex) != PASS) {
	printf("FAIL - getval failed getting boolean as boolean\n");
	ReturnValue++;
    } else {
	if (TestBool != FALSE) {
	    printf("FAIL - getval boolean %d != %d\n", TestBool, FALSE);
	    ReturnValue++;
	}
    }

    /* int1 = { , 1 } */
    if (cfp_getval(FileId, "int1", (void *) &TestInt, CFP_INTEGER,
		   sizeof TestInt, CFP_CurrentIndex) != PASS) {
	printf("FAIL - getval failed getting int1\n");
	ReturnValue++;
    } else {
	if (TestInt != 1) {
	    printf("FAIL - getval int %d != 1\n", TestInt);
	    ReturnValue++;
	}
    }

    /* int2 = { , 2, , , "[0,10]" } */
    if (cfp_getval(FileId, "int2", (void *) &TestInt, CFP_INTEGER,
		   sizeof TestInt, CFP_CurrentIndex) != PASS) {
	printf("FAIL - getval failed getting int2\n");
	ReturnValue++;
    } else {
	if (TestInt != 2) {
	    printf("FAIL - getval int %d != 2\n", TestInt);
	    ReturnValue++;
	}
    }

    /* int3 = { , 30, , , "[0,10]" } */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with bad range",
	      __LINE__);
    if (cfp_getval(FileId, "int3", (void *) &TestInt, CFP_INTEGER,
		   sizeof TestInt, CFP_CurrentIndex) != FAIL) {
	printf("FAIL - getval did not fail getting int3, returned %d\n",
	       TestInt);
	ReturnValue++;
    }

    /* double = { , 5.0 } */
    if (cfp_getval(FileId, "double", (void *) &TestDbl, CFP_DOUBLE,
		   sizeof TestDbl, CFP_CurrentIndex) != PASS) {
	printf("FAIL - getval failed getting double\n");
	ReturnValue++;
    } else {
	if (TestDbl != 5.0) {
	    printf("FAIL - getval double %g != 5.0\n", TestDbl);
	    ReturnValue++;
	}
    }

    /* try environment variables */
    if ((EnvAddress = getenv("Check_getvalTEST")) == (char *) NULL) {
	printf("FAIL - Check_getval environment broken, no TEST\n");
	ReturnValue++;
    } else {
	if (cfp_getval(FileId, "TEST", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_EnvironmentIndex) != PASS) {
	    printf("FAIL - get of environment variable TEST failed\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, EnvAddress) != 0) {
		printf("FAIL - getval of env (%s) not equal env var (%s)\n",
		       TestString, EnvAddress);
		ReturnValue++;
	    }
	}
    }
    if ((EnvAddress = getenv("Check_getvalINT")) == (char *) NULL) {
	printf("FAIL - Check_getval environment broken, no INT\n");
	ReturnValue++;
    } else {
	if (cfp_getval(FileId, "INT", (void *) &TestInt, CFP_INTEGER,
		       sizeof TestInt, CFP_EnvironmentIndex) != PASS) {
	    printf("FAIL - get of int environment variable failed\n");
	    ReturnValue++;
	} else {
	    if (atoi(EnvAddress) != TestInt) {
		printf("FAIL - getval of env (%d) not equal env var (%d)\n",
		       TestInt, atoi(EnvAddress));
		ReturnValue++;
	    }
	}
    }

    /* non-existent field index */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with bad index",
	      __LINE__);
    if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		   sizeof TestString, (ItemIndexType) 42) != FAIL) {
	printf("FAIL - cfp_getval of field 42 did not fail, returned '%s'\n",
	       TestString);
	ReturnValue++;
    }

    /* reference to a constant */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail - constant",
	      __LINE__);
    if (cfp_getval(FileId, "const", (void *) TestString, CFP_STRING,
		   sizeof TestString, CFP_ModeIndex) != FAIL) {
	printf("FAIL - cfp_getval of a constant did not fail, returned '%s'\n",
	       TestString);
	ReturnValue++;
    }

    /* non-existent name */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with lookup failure",
	      __LINE__);
    if (cfp_getval(FileId, "nosuchvar", (void *) TestString,
		   CFP_STRING, sizeof TestString, CFP_ModeIndex) != FAIL) {
	printf("FAIL - cfp_getval of nosuchvar did not fail, returned '%s'\n",
	       TestString);
	ReturnValue++;
    }

    /* try non-existent environment variable */
    if ((EnvAddress = getenv("Check_getvalNOT_there")) != (char *) NULL) {
	printf("FAIL - Check_getval environment broken, has NOT-there\n");
	ReturnValue++;
    } else {
	if ((cfp_getval(FileId, "NOT_there", (void *) TestString,
			CFP_STRING, sizeof TestString, CFP_EnvironmentIndex)
	     == PASS)
	    && (strlen(TestString) > 0)) {
	    printf("FAIL - get of missing environment variable worked (=%s)\n",
		   TestString);
	    ReturnValue++;
	}
    }

    if (cfp_finish(FileId) != PASS) {
	printf("FAIL - cfp_finish failed\n");
	ReturnValue++;
    }

    return(ReturnValue);
    /*NOTREACHED*/
}
