/*
 * Check_putcur.c - test cfp_putcur()
 *
 * This program is used to make sure that cfp_putcur() works correctly.  It
 * assumes that cfp_getcur() has already been thoroughly. Its expected
 * environment is that the calling shell script has created a .par file as
 * defined below and called this program with no arguments.  This runs through
 * a set of tests outputting any unexpected results to stdout.  Each output
 * starts with "FAIL" for the calling shell script to find.  If any putcur
 * errors occurred, this exits with a positive error count.  If any parse
 * errors occurred, this exits with -1.
 *
 * The expected .par file:
 *
 * const = some value
 * bool = { , True }
 * int1 = { , 1 }
 * int2 = { , 2, , , int }
 * int3 = { , 3, , , "[0,10]" }
 * double1 = { , 1 }
 * double2 = { , 2, , , double }
 * double3 = { , 3, , , "[0.0,10.0]" }
 * time1 = { , "1:0:0", , , double }
 * time2 = { , "0:0:0", , , "(0:0:0,1:0:0)" }
 * string = { , foo }
 *
 *
 * CFHTLOGS and CFHTLOGU are expected to be undefined (though this doesn't
 * care)
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include "cfht.h"
#include "cfp.h"
#include "exits.h"

main(argc, argv)
    int argc;
    char *argv[];
{
    FILE_ID FileId;		  /* file index from parse */
    BOOLEAN TestBool;		  /* scratch areas  */
    int TestInt;		  /*   for          */
    double TestDbl;		  /*     getcur     */
    double TestTime;		  /*       provided */
    char TestString[CFP_STRSIZE]; /*         values */
    HistoryType History;	  /* getcur return value */
    int ReturnValue;		  /* starts at zero, +1 for each error */

    ReturnValue = 0;

    if ((FileId = cfp_parse(CFP_PAR, argc, argv)) == FAIL) {
	printf("FAIL - cfp_parse died on %s\n", argv[0]);
	return(EXIT_FAIL);
    }

    /* const = some value */
    if (cfp_putcur(FileId, "const", (void *) "new value", CFP_STRING) != PASS) {
	printf("FAIL - putcur to constant failed\n");
	ReturnValue++;
    } else {
	if (History = cfp_getcur(FileId, "const", (void *) TestString,
				 CFP_STRING, sizeof TestString) == FAIL) {
	    printf("FAIL - getcur of constant failed\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "new value") != 0) {
		printf("FAIL - putcur to const put '%s' not 'new value'\n");
		ReturnValue++;
	    }
	}
    }

    /* bool = { , True } */
    TestBool = TRUE;
    if (cfp_putcur(FileId, "bool", (void *) &TestBool, CFP_BOOLEAN) != PASS) {
	printf("FAIL - cfp_putcur of bool to TRUE failed\n");
	ReturnValue++;
    } else {
	TestBool = FALSE;
	if ((History = cfp_getcur(FileId, "bool", (void *) &TestBool,
				  CFP_BOOLEAN, sizeof TestBool)) == FAIL) {
	    printf("FAIL - cfp_getcur of TRUE bool failed\n");
	    ReturnValue++;
	} else {
	    if (TestBool != TRUE) {
		printf("FAIL - cfp_getcur of bool returned %d not %d\n",
		       TestBool, TRUE);
		ReturnValue++;
	    }
	}
    }
    TestBool = FALSE;
    if (cfp_putcur(FileId, "bool", (void *) &TestBool, CFP_BOOLEAN) != PASS) {
	printf("FAIL - cfp_putcur of bool to FALSE failed\n");
	ReturnValue++;
    } else {
	TestBool = TRUE;
	if ((History = cfp_getcur(FileId, "bool", (void *) &TestBool,
				  CFP_BOOLEAN, sizeof TestBool)) == FAIL) {
	    printf("FAIL - cfp_getcur of FALSE bool failed\n");
	    ReturnValue++;
	} else {
	    if (TestBool != FALSE) {
		printf("FAIL - cfp_getcur of bool returned %d not %d\n",
		       TestBool, FALSE);
		ReturnValue++;
	    }
	}
    }

    /* int1 = { , 1 } */
    TestInt = 42;
    if (cfp_putcur(FileId, "int1", (void *) &TestInt, CFP_INTEGER) != PASS) {
	printf("FAIL - cfp_putcur of int1 to 42 failed\n");
	ReturnValue++;
    } else {
	TestInt = 0;
	if ((History = cfp_getcur(FileId, "int1", (void *) &TestInt,
				  CFP_INTEGER, sizeof TestInt)) == FAIL) {
	    printf("FAIL - cfp_getcur of int1 42 failed\n");
	    ReturnValue++;
	} else {
	    if (TestInt != 42) {
		printf("FAIL - cfp_getcur of int1 returned %d not %d\n",
		       TestInt, 42);
		ReturnValue++;
	    }
	}
    }
    TestInt = -42;
    if (cfp_putcur(FileId, "int1", (void *) &TestInt, CFP_INTEGER) != PASS) {
	printf("FAIL - cfp_putcur of int1 to -42 failed\n");
	ReturnValue++;
    } else {
	TestInt = 0;
	if ((History = cfp_getcur(FileId, "int1", (void *) &TestInt,
				  CFP_INTEGER, sizeof TestInt)) == FAIL) {
	    printf("FAIL - cfp_getcur of int1 -42 failed\n");
	    ReturnValue++;
	} else {
	    if (TestInt != -42) {
		printf("FAIL - cfp_getcur of int1 returned %d not %d\n",
		       TestInt, -42);
		ReturnValue++;
	    }
	}
    }

    /* int2 = { , 2, , , int } */
    TestInt = INT_MAX;
    if (cfp_putcur(FileId, "int2", (void *) &TestInt, CFP_INTEGER) != PASS) {
	printf("FAIL - cfp_putcur of int2 to INT_MAX failed\n");
	ReturnValue++;
    } else {
	TestInt = 0;
	if ((History = cfp_getcur(FileId, "int2", (void *) &TestInt,
				  CFP_INTEGER, sizeof TestInt)) == FAIL) {
	    printf("FAIL - cfp_getcur of int2 INT_MAX failed\n");
	    ReturnValue++;
	} else {
	    if (TestInt != INT_MAX) {
		printf("FAIL - cfp_getcur of int2 returned %d not INT_MAX\n",
		       TestInt);
		ReturnValue++;
	    }
	}
    }
    TestInt = INT_MIN;
    if (cfp_putcur(FileId, "int2", (void *) &TestInt, CFP_INTEGER) != PASS) {
	printf("FAIL - cfp_putcur of int2 to INT_MIN failed\n");
	ReturnValue++;
    } else {
	TestInt = 0;
	if ((History = cfp_getcur(FileId, "int2", (void *) &TestInt,
				  CFP_INTEGER, sizeof TestInt)) == FAIL) {
	    printf("FAIL - cfp_getcur of int2 INT_MIN failed\n");
	    ReturnValue++;
	} else {
	    if (TestInt != INT_MIN) {
		printf("FAIL - cfp_getcur of int2 returned %d not INT_MIN%d\n",
		       TestInt);
		ReturnValue++;
	    }
	}
    }
    if (cfp_putcur(FileId, "int2", (void *) "0x3210", CFP_STRING) != PASS) {
	printf("FAIL - cfp_putcur of int2 to 0x3210 failed\n");
	ReturnValue++;
    } else {
	TestInt = 0;
	if ((History = cfp_getcur(FileId, "int2", (void *) &TestInt,
				  CFP_INTEGER, sizeof TestInt)) == FAIL) {
	    printf("FAIL - cfp_getcur of int2 0x3210 failed\n");
	    ReturnValue++;
	} else {
	    if (TestInt != 0x3210) {
		printf("FAIL - cfp_getcur of int2 returned %d not 0x3210%d\n",
		       TestInt);
		ReturnValue++;
	    }
	}
    }
    if (cfp_putcur(FileId, "int2", (void *) "0x7654", CFP_STRING) != PASS) {
	printf("FAIL - cfp_putcur of int2 to 0x7654 failed\n");
	ReturnValue++;
    } else {
	TestInt = 0;
	if ((History = cfp_getcur(FileId, "int2", (void *) &TestInt,
				  CFP_INTEGER, sizeof TestInt)) == FAIL) {
	    printf("FAIL - cfp_getcur of int2 0x7654 failed\n");
	    ReturnValue++;
	} else {
	    if (TestInt != 0x7654) {
		printf("FAIL - cfp_getcur of int2 returned %d not 0x7654%d\n",
		       TestInt);
		ReturnValue++;
	    }
	}
    }
    if (cfp_putcur(FileId, "int2", (void *) "0xba98", CFP_STRING) != PASS) {
	printf("FAIL - cfp_putcur of int2 to 0xba98 failed\n");
	ReturnValue++;
    } else {
	TestInt = 0;
	if ((History = cfp_getcur(FileId, "int2", (void *) &TestInt,
				  CFP_INTEGER, sizeof TestInt)) == FAIL) {
	    printf("FAIL - cfp_getcur of int2 0xba98 failed\n");
	    ReturnValue++;
	} else {
	    if (TestInt != 0xba98) {
		printf("FAIL - cfp_getcur of int2 returned %d not 0xba98%d\n",
		       TestInt);
		ReturnValue++;
	    }
	}
    }
    if (cfp_putcur(FileId, "int2", (void *) "0xfedc", CFP_STRING) != PASS) {
	printf("FAIL - cfp_putcur of int2 to 0xfedc failed\n");
	ReturnValue++;
    } else {
	TestInt = 0;
	if ((History = cfp_getcur(FileId, "int2", (void *) &TestInt,
				  CFP_INTEGER, sizeof TestInt)) == FAIL) {
	    printf("FAIL - cfp_getcur of int2 0xfedc failed\n");
	    ReturnValue++;
	} else {
	    if (TestInt != 0xfedc) {
		printf("FAIL - cfp_getcur of int2 returned %d not 0xfedc%d\n",
		       TestInt);
		ReturnValue++;
	    }
	}
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if (cfp_putcur(FileId, "int2", (void *) "10e1", CFP_STRING) != FAIL) {
	printf("FAIL - cfp_putcur of int2 to 10e1 did not fail\n");
	ReturnValue++;
	if ((History = cfp_getcur(FileId, "int2", (void *) &TestString,
				  CFP_STRING, sizeof TestString)) == FAIL) {
	    ;			  /* getcur worked even if putcur didn't */
	} else {
	    printf("FAIL - putcur stored %s\n", TestString);
	    ReturnValue++;
	}
    }

    /* int3 = { , 3, , , "[0,10]" } */
    TestInt = 0;
    if (cfp_putcur(FileId, "int3", (void *) &TestInt, CFP_INTEGER) != PASS) {
	printf("FAIL - cfp_putcur of int3 to 0 failed\n");
	ReturnValue++;
    } else {
	TestInt = 0;
	if ((History = cfp_getcur(FileId, "int3", (void *) &TestInt,
				  CFP_INTEGER, sizeof TestInt)) == FAIL) {
	    printf("FAIL - cfp_getcur of int3 0 failed\n");
	    ReturnValue++;
	} else {
	    if (TestInt != 0) {
		printf("FAIL - cfp_getcur of int3 returned %d not %d\n",
		       TestInt, 0);
		ReturnValue++;
	    }
	}
    }
    TestInt = 10;
    if (cfp_putcur(FileId, "int3", (void *) &TestInt, CFP_INTEGER) != PASS) {
	printf("FAIL - cfp_putcur of int3 to 10 failed\n");
	ReturnValue++;
    } else {
	TestInt = 0;
	if ((History = cfp_getcur(FileId, "int3", (void *) &TestInt,
				  CFP_INTEGER, sizeof TestInt)) == FAIL) {
	    printf("FAIL - cfp_getcur of int3 10 failed\n");
	    ReturnValue++;
	} else {
	    if (TestInt != 10) {
		printf("FAIL - cfp_getcur of int3 returned %d not %d\n",
		       TestInt, 10);
		ReturnValue++;
	    }
	}
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    TestInt = 11;
    if (cfp_putcur(FileId, "int3", (void *) &TestInt, CFP_INTEGER) != FAIL) {
	printf("FAIL - cfp_putcur of int3 to 11 did not fail\n");
	ReturnValue++;
	if ((History = cfp_getcur(FileId, "int3", (void *) &TestInt,
				  CFP_INTEGER, sizeof TestInt)) == FAIL) {
	    ;			  /* getcur worked even if putcur didn't */
	} else {
	    if (TestInt == 11) {
		printf("FAIL - neither putcur nor getcur worked, int3=11\n");
		ReturnValue++;
	    } else {
		printf("FAIL - putcur stored %d instead of 11\n", TestInt);
		ReturnValue++;
	    }
	}
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    TestInt = -1;
    if (cfp_putcur(FileId, "int3", (void *) &TestInt, CFP_INTEGER) != FAIL) {
	printf("FAIL - cfp_putcur of int3 to -1 did not fail\n");
	ReturnValue++;
	if ((History = cfp_getcur(FileId, "int3", (void *) &TestInt,
				  CFP_INTEGER, sizeof TestInt)) == FAIL) {
	    ;			  /* getcur worked even if putcur didn't */
	} else {
	    if (TestInt == -1) {
		printf("FAIL - neither putcur nor getcur worked, int3=-1\n");
		ReturnValue++;
	    } else {
		printf("FAIL - putcur stored %d instead of -1\n", TestInt);
		ReturnValue++;
	    }
	}
    }

    /* double1 = { , 1 } */
    TestDbl = 6.023e+23;
    if (cfp_putcur(FileId, "double1", (void *) &TestDbl, CFP_DOUBLE) != PASS) {
	printf("FAIL - cfp_putcur of double1 to 6.023e23 failed\n");
	ReturnValue++;
    } else {
	TestDbl = 0.0;
	if ((History = cfp_getcur(FileId, "double1", (void *) &TestDbl,
				  CFP_DOUBLE, sizeof TestDbl)) == FAIL) {
	    printf("FAIL - cfp_getcur of double1 6.023e23 failed\n");
	    ReturnValue++;
	} else {
	    if (fabs(TestDbl - 6.023e+23) > 1e15) {
		printf("FAIL - cfp_getcur of double1 returned %g not 6.023e23\n",
		       TestDbl);
		ReturnValue++;
	    }
	}
    }
    TestDbl = -5.25e-20;
    if (cfp_putcur(FileId, "double1", (void *) &TestDbl, CFP_DOUBLE) != PASS) {
	printf("FAIL - cfp_putcur of double1 to -5.25e-20 failed\n");
	ReturnValue++;
    } else {
	TestDbl = 0.0;
	if ((History = cfp_getcur(FileId, "double1", (void *) &TestDbl,
				  CFP_DOUBLE, sizeof TestDbl)) == FAIL) {
	    printf("FAIL - cfp_getcur of double1 -5.25e-20 failed\n");
	    ReturnValue++;
	} else {
	    if (fabs(TestDbl + 5.25e-20) > 1e-30) {
		printf("FAIL - cfp_getcur of double1 returned %g not -5.25e-20\n",
		       TestDbl);
		ReturnValue++;
	    }
	}
    }

    /* double2 = { , 2, , , double } */
    TestDbl = DBL_MAX;
    if (cfp_putcur(FileId, "double2", (void *) &TestDbl, CFP_DOUBLE) != PASS) {
	printf("FAIL - cfp_putcur of double2 to DBL_MAX failed\n");
	ReturnValue++;
    } else {
	TestDbl = 0.0;
	if ((History = cfp_getcur(FileId, "double2", (void *) &TestDbl,
				  CFP_DOUBLE, sizeof TestDbl)) == FAIL) {
	    printf("FAIL - cfp_getcur of double2 DBL_MAX failed\n");
	    ReturnValue++;
	} else {
/*	    if (TestDbl != DBL_MAX)   <-- this comparison doesn't work */
	}
    }
    TestDbl = -DBL_MAX;
    if (cfp_putcur(FileId, "double2", (void *) &TestDbl, CFP_DOUBLE) != PASS) {
	printf("FAIL - cfp_putcur of double2 to -DBL_MAX failed\n");
	ReturnValue++;
    } else {
	TestDbl = 0.0;
	if ((History = cfp_getcur(FileId, "double2", (void *) &TestDbl,
				  CFP_DOUBLE, sizeof TestDbl)) == FAIL) {
	    printf("FAIL - cfp_getcur of double2 -DBL_MAX failed\n");
	    ReturnValue++;
	} else {
/*	    if (TestDbl != -DBL_MAX)  <-- this comparison doesn't work */
	}
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if (cfp_putcur(FileId, "double2", (void *) "5e10a", CFP_STRING) != FAIL) {
	printf("FAIL - cfp_putcur of double2 to 5e10a did not fail\n");
	ReturnValue++;
	if ((History = cfp_getcur(FileId, "double2", (void *) &TestString,
				  CFP_STRING, sizeof TestString)) == FAIL) {
	    ;			  /* getcur worked even if putcur didn't */
	} else {
	    printf("FAIL - putcur stored %s\n", TestString);
	    ReturnValue++;
	}
    }

    /* double3 = { , 3, , , "[0.0,10.0]" } */
    TestDbl = 0.0;
    if (cfp_putcur(FileId, "double3", (void *) &TestDbl, CFP_DOUBLE) != PASS) {
	printf("FAIL - cfp_putcur of double3 to 0.0 failed\n");
	ReturnValue++;
    } else {
	TestDbl = 0.0;
	if ((History = cfp_getcur(FileId, "double3", (void *) &TestDbl,
				  CFP_DOUBLE, sizeof TestDbl)) == FAIL) {
	    printf("FAIL - cfp_getcur of double3 0.0 failed\n");
	    ReturnValue++;
	} else {
	    if (TestDbl != 0.0) {
		printf("FAIL - cfp_getcur of double3 returned %g not 0.0\n",
		       TestDbl);
		ReturnValue++;
	    }
	}
    }
    TestDbl = 10.0;
    if (cfp_putcur(FileId, "double3", (void *) &TestDbl, CFP_DOUBLE) != PASS) {
	printf("FAIL - cfp_putcur of double3 to 10.0 failed\n");
	ReturnValue++;
    } else {
	TestDbl = 0.0;
	if ((History = cfp_getcur(FileId, "double3", (void *) &TestDbl,
				  CFP_DOUBLE, sizeof TestDbl)) == FAIL) {
	    printf("FAIL - cfp_getcur of double3 10.0 failed\n");
	    ReturnValue++;
	} else {
	    if (TestDbl != 10.0) {
		printf("FAIL - cfp_getcur of double3 returned %g not 10.0\n",
		       TestDbl);
		ReturnValue++;
	    }
	}
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    TestDbl = 10.0001;
    if (cfp_putcur(FileId, "double3", (void *) &TestDbl, CFP_DOUBLE) != FAIL) {
	printf("FAIL - cfp_putcur of double3 to 10.0001 did not fail\n");
	ReturnValue++;
	if ((History = cfp_getcur(FileId, "double3", (void *) &TestDbl,
				  CFP_DOUBLE, sizeof TestDbl)) == FAIL) {
	    ;			  /* getcur worked even if putcur didn't */
	} else {
	    if (TestDbl == 10.0001) {
		printf("FAIL - neither putcur nor getcur worked, double3=10.0001\n");
		ReturnValue++;
	    } else {
		printf("FAIL - putcur stored %g instead of 10.0001\n", TestDbl);
		ReturnValue++;
	    }
	}
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    TestDbl = -0.0001;
    if (cfp_putcur(FileId, "double3", (void *) &TestDbl, CFP_DOUBLE) != FAIL) {
	printf("FAIL - cfp_putcur of double3 to -0.0001 did not fail\n");
	ReturnValue++;
	if ((History = cfp_getcur(FileId, "double3", (void *) &TestDbl,
				  CFP_DOUBLE, sizeof TestDbl)) == FAIL) {
	    ;			  /* getcur worked even if putcur didn't */
	} else {
	    if (TestDbl == -0.0001) {
		printf("FAIL - neither putcur nor getcur worked, double3=-0.0001\n");
		ReturnValue++;
	    } else {
		printf("FAIL - putcur stored %g instead of -0.0001\n", TestDbl);
		ReturnValue++;
	    }
	}
    }

    /* time1 = { , "1:0:0", , , double } */
    if (cfp_putcur(FileId, "time1", (void *) "24:0:0.0", CFP_STRING) != PASS) {
	printf("FAIL - cfp_putcur of time1 to 24:0:0 failed\n");
	ReturnValue++;
    } else {
	TestTime = 0.0;
	if ((History = cfp_getcur(FileId, "time1", (void *) &TestTime,
				  CFP_DOUBLE, sizeof TestTime)) == FAIL) {
	    printf("FAIL - cfp_getcur of time1 24:0:0 failed\n");
	    ReturnValue++;
	} else {
	    if (TestTime != (24.0 * 60.0 * 60.0)) {
		printf("FAIL - cfp_getcur of time1 returned %g not 86400\n",
		       TestTime);
		ReturnValue++;
	    }
	}
    }
    if (cfp_putcur(FileId, "time1", (void *) "-1:0:0.1", CFP_STRING) != PASS) {
	printf("FAIL - cfp_putcur of time1 to -1:0:0.1 failed\n");
	ReturnValue++;
    } else {
	TestTime = 0.0;
	if ((History = cfp_getcur(FileId, "time1", (void *) &TestTime,
				  CFP_DOUBLE, sizeof TestTime)) == FAIL) {
	    printf("FAIL - cfp_getcur of time1 -1:0:0.1 failed\n");
	    ReturnValue++;
	} else {
	    if (TestTime != -(1.0 * 60.0 * 60.0 + 0.1)) {
		printf("FAIL - cfp_getcur of time1 returned %g not -3600.1\n",
		       TestTime);
		ReturnValue++;
	    }
	}
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if (cfp_putcur(FileId, "time1", (void *) "0:0:60", CFP_STRING) != FAIL) {
	printf("FAIL - cfp_putcur of time1 to 0:0:60 did not fail\n");
	ReturnValue++;
	if ((History = cfp_getcur(FileId, "time1", (void *) &TestTime,
				  CFP_DOUBLE, sizeof TestTime)) == FAIL) {
	    ;			  /* getcur worked even if putcur didn't */
	} else {
	    if (TestTime == 60.0) {
		printf("FAIL - neither putcur nor getcur worked, time1=0:0:60\n");
		ReturnValue++;
	    } else {
		printf("FAIL - putcur stored %g instead of 60\n", TestTime);
		ReturnValue++;
	    }
	}
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if (cfp_putcur(FileId, "time1", (void *) "0:60:0", CFP_STRING) != FAIL) {
	printf("FAIL - cfp_putcur of time1 to 0:60:0 did not fail\n");
	ReturnValue++;
	if ((History = cfp_getcur(FileId, "time1", (void *) &TestTime,
				  CFP_DOUBLE, sizeof TestTime)) == FAIL) {
	    ;			  /* getcur worked even if putcur didn't */
	} else {
	    if (TestTime == (60.0 * 60.0)) {
		printf("FAIL - neither putcur nor getcur worked, time1=0:60:0\n");
		ReturnValue++;
	    } else {
		printf("FAIL - putcur stored %g instead of 3600\n", TestTime);
		ReturnValue++;
	    }
	}
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if (cfp_putcur(FileId, "time1", (void *) "0:ab0:0", CFP_STRING) != FAIL) {
	printf("FAIL - cfp_putcur of time1 to 0:ab0:0 did not fail\n");
	ReturnValue++;
	if ((History = cfp_getcur(FileId, "time1", (void *) TestString,
				  CFP_STRING, sizeof TestString)) == FAIL) {
	    ;			  /* getcur worked even if putcur didn't */
	} else {
	    printf("FAIL - putcur stored %s\n", TestString);
	    ReturnValue++;
	}
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if (cfp_putcur(FileId, "time1", (void *) "0:0:0e10", CFP_STRING) != FAIL) {
	printf("FAIL - cfp_putcur of time1 to 0:0:0e10 did not fail\n");
	ReturnValue++;
	if ((History = cfp_getcur(FileId, "time1", (void *) &TestString,
				  CFP_STRING, sizeof TestString)) == FAIL) {
	    ;			  /* getcur worked even if putcur didn't */
	} else {
	    printf("FAIL - cfp_putcur stored %s\n", TestString);
	    ReturnValue++;
	}
    }

    /* time2 = { , "0:0:0", , , "(0:0:0,1:0:0)" } */
    if (cfp_putcur(FileId, "time2", (void *) "0:20:0.5", CFP_STRING) != PASS) {
	printf("FAIL - cfp_putcur of time2 to 0:20:0.5 failed\n");
	ReturnValue++;
    } else {
	TestTime = 0.0;
	if ((History = cfp_getcur(FileId, "time2", (void *) &TestTime,
				  CFP_DOUBLE, sizeof TestTime)) == FAIL) {
	    printf("FAIL - cfp_getcur of time2 0:20:0.5 failed\n");
	    ReturnValue++;
	} else {
	    if (TestTime != (20.0 * 60.0 + 0.5)) {
		printf("FAIL - cfp_getcur of time2 returned %g not 1200.5\n",
		       TestTime);
		ReturnValue++;
	    }
	}
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if (cfp_putcur(FileId, "time2", (void *) "0:0:0", CFP_STRING) != FAIL) {
	printf("FAIL - cfp_putcur of time2 to 0:0:0 did not fail\n");
	ReturnValue++;
	if ((History = cfp_getcur(FileId, "time2", (void *) &TestTime,
				  CFP_DOUBLE, sizeof TestTime)) == FAIL) {
	    ;			  /* getcur worked even if putcur didn't */
	} else {
	    if (TestTime == 0.0) {
		printf("FAIL - neither putcur nor getcur worked, time2=0:0:0\n");
		ReturnValue++;
	    } else {
		printf("FAIL - putcur stored %g instead of 0\n", TestTime);
		ReturnValue++;
	    }
	}
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with range error",
	      __LINE__);
    if (cfp_putcur(FileId, "time2", (void *) "1:0:0", CFP_STRING) != FAIL) {
	printf("FAIL - cfp_putcur of time2 to 1:0:0 did not fail\n");
	ReturnValue++;
	if ((History = cfp_getcur(FileId, "time2", (void *) &TestTime,
				  CFP_DOUBLE, sizeof TestTime)) == FAIL) {
	    ;			  /* getcur worked even if putcur didn't */
	} else {
	    if (TestTime == (60.0 * 60.0)) {
		printf("FAIL - neither putcur nor getcur worked, time2=1:0:0\n");
		ReturnValue++;
	    } else {
		printf("FAIL - putcur stored %g instead of 3600\n", TestTime);
		ReturnValue++;
	    }
	}
    }

    /* string = { , foo } */
    if (cfp_putcur(FileId, "string", (void *) "bar 2", CFP_STRING) != PASS) {
	printf("FAIL - cfp_putcur of string to 'bar 2' failed\n");
	ReturnValue++;
    } else {
	if ((History = cfp_getcur(FileId, "string", (void *) TestString,
				  CFP_STRING, sizeof TestString)) == FAIL) {
	    printf("FAIL - cfp_getcur of string failed\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "bar 2") != 0) {
		printf("FAIL - cfp_getcur of string returned '%s' not '%s'\n",
		       TestString, "bar 2");
		ReturnValue++;
	    }
	}
    }

    /* non-existent name */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with lookup failure",
	      __LINE__);
    if (cfp_putcur(FileId, "nosuchvar", (void *) "say what?",
		   CFP_STRING) != FAIL) {
	printf("FAIL - cfp_putcur of nosuchvar did not fail\n");
	ReturnValue++;
    }

    if (cfp_putf(FileId) != PASS) {
	printf("FAIL - cfp_putf failed\n");
	ReturnValue++;
    }

    if (cfp_finish(FileId) != PASS) {
	printf("FAIL - cfp_finish failed\n");
	ReturnValue++;
    }

    return(ReturnValue);
    /*NOTREACHED*/
}
