/*
 * Check_putval.c - test cfp_putval()
 *
 * This program is used to make sure that cfp_putval() works correctly.  Its
 * expected environment is that the calling shell script has created a .par
 * file and environment variables as defined below and called this program
 * with no arguments.  This runs through a set of tests outputting any
 * unexpected results to stdout.  Each output starts with "FAIL" for the
 * calling shell script to find.  If any putval errors occurred, this exits
 * with a positive error count.  If any parse errors occurred, this exits with
 * -1.
 *
 * The expected .par file:
 *
 * const = this should fail
 * var = { nds, 100, default, supplied, int }
 * bool = { nds, original, false, true }
 * int1 = { , 1 }
 * int2 = { , 2, , , "[0,10]" }
 * double = { , 5.0 }
 * TEST = {  }
 * INT = {  }
 * NOT_there = {  }
 *
 * The expected environment variables:
 *
 * Check_putvalTEST="abc 123"
 * Check_putvalINT=123
 *
 *
 * CFHTLOGS and CFHTLOGU are expected to be undefined (though this doesn't
 * care)
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "cfht.h"
#include "cfp.h"
#include "exits.h"

main(argc, argv)
    int argc;
    char *argv[];
{
    FILE_ID FileId;		  /* file index from parse */
    BOOLEAN TestBool;		  /* spots for    */
    int TestInt;		  /*   getval     */
    double TestDbl;		  /*     returned */
    char TestString[CFP_STRSIZE + 1]; /*   values */
    char *EnvAddress;		  /* pointer to environment string */
    int ReturnValue;		  /* starts at zero, +1 for each error */

    ReturnValue = 0;

    if ((FileId = cfp_parse(CFP_PAR, argc, argv)) == FAIL) {
	printf("FAIL - cfp_parse died on %s\n", argv[0]);
	return(EXIT_FAIL);
    }

    /* var = { dns, 100, default, supplied, int } */
    /* make sure all the fields can be set independently */
    if (cfp_putval(FileId, "var", (void *) "o", CFP_STRING, CFP_ModeIndex)
	!= PASS) {
	printf("FAIL - putval failed on mode string\n");
	ReturnValue++;
    } else {
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_ModeIndex) != PASS) {
	    printf("FAIL - getval failed getting mode string\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "o") != 0) {
		printf("FAIL - getval mode value '%s' != 'o'\n", TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_CurrentIndex) != PASS) {
	    printf("FAIL - getval failed getting current string for mode\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "100") != 0) {
		printf("FAIL - mode getval current value '%s' != '100'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_DefaultIndex) != PASS) {
	    printf("FAIL - getval failed getting default string for mode\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "default") != 0) {
		printf("FAIL - mode getval default value '%s' != 'default'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_SuppliedIndex) != PASS) {
	    printf("FAIL - getval failed getting supplied string for mode\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "supplied") != 0) {
		printf("FAIL - mode getval supplied value '%s' != 'supplied'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_RangeIndex) != PASS) {
	    printf("FAIL - getval failed getting range string for mode\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "int") != 0) {
		printf("FAIL - mode getval range value '%s' != 'int'\n",
		       TestString);
		ReturnValue++;
	    }
	}
    }

    if (cfp_putval(FileId, "var", (void *) "200", CFP_STRING,
		   CFP_CurrentIndex) != PASS) {
	printf("FAIL - putval failed on current string\n");
	ReturnValue++;
    } else {
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_CurrentIndex) != PASS) {
	    printf("FAIL - getval failed getting current string\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "200") != 0) {
		printf("FAIL - getval current value '%s' != '200'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_ModeIndex) != PASS) {
	    printf("FAIL - getval failed getting mode string for current\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "o") != 0) {
		printf("FAIL - current getval mode value '%s' != 'o'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_DefaultIndex) != PASS) {
	    printf("FAIL - getval failed getting default string for current\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "default") != 0) {
		printf("FAIL - current getval default value '%s' != 'default'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_SuppliedIndex) != PASS) {
	    printf("FAIL - getval failed getting supplied string for current\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "supplied") != 0) {
		printf("FAIL - current getval supplied value '%s' != 'supplied'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_RangeIndex) != PASS) {
	    printf("FAIL - getval failed getting range string for current\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "int") != 0) {
		printf("FAIL - current getval range value '%s' != 'int'\n",
		       TestString);
		ReturnValue++;
	    }
	}
    }

    if (cfp_putval(FileId, "var", (void *) "DEFAULT", CFP_STRING,
		   CFP_DefaultIndex) != PASS) {
	printf("FAIL - putval failed on default string\n");
	ReturnValue++;
    } else {
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_DefaultIndex) != PASS) {
	    printf("FAIL - getval failed getting default string\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "DEFAULT") != 0) {
		printf("FAIL - getval default value '%s' != 'DEFAULT'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_ModeIndex) != PASS) {
	    printf("FAIL - getval failed getting mode string for default\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "o") != 0) {
		printf("FAIL - default getval mode value '%s' != 'o'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_CurrentIndex) != PASS) {
	    printf("FAIL - getval failed getting current string for default\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "200") != 0) {
		printf("FAIL - default getval current value '%s' != '200'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_SuppliedIndex) != PASS) {
	    printf("FAIL - getval failed getting supplied string for default\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "supplied") != 0) {
		printf("FAIL - default getval supplied value '%s' != 'supplied'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_RangeIndex) != PASS) {
	    printf("FAIL - getval failed getting range string for default\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "int") != 0) {
		printf("FAIL - default getval range value '%s' != 'int'\n",
		       TestString);
		ReturnValue++;
	    }
	}
    }

    if (cfp_putval(FileId, "var", (void *) "SUPPLIED", CFP_STRING,
		   CFP_SuppliedIndex) != PASS) {
	printf("FAIL - putval failed on supplied string\n");
	ReturnValue++;
    } else {
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_SuppliedIndex) != PASS) {
	    printf("FAIL - getval failed getting supplied string\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "SUPPLIED") != 0) {
		printf("FAIL - getval supplied value '%s' != 'SUPPLIED'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_ModeIndex) != PASS) {
	    printf("FAIL - getval failed getting mode string for supplied\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "o") != 0) {
		printf("FAIL - supplied getval mode value '%s' != 'o'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_CurrentIndex) != PASS) {
	    printf("FAIL - getval failed getting current string for supplied\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "200") != 0) {
		printf("FAIL - supplied getval current value '%s' != '200'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_DefaultIndex) != PASS) {
	    printf("FAIL - getval failed getting default string for supplied\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "DEFAULT") != 0) {
		printf("FAIL - supplied getval default value '%s' != 'DEFAULT'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_RangeIndex) != PASS) {
	    printf("FAIL - getval failed getting range string for supplied\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "int") != 0) {
		printf("FAIL - supplied getval range value '%s' != 'int'\n",
		       TestString);
		ReturnValue++;
	    }
	}
    }
    if (cfp_putval(FileId, "var", (void *) "double", CFP_STRING,
		   CFP_RangeIndex) != PASS) {
	printf("FAIL - putval failed on range string\n");
	ReturnValue++;
    } else {
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_RangeIndex) != PASS) {
	    printf("FAIL - getval failed getting range string\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "double") != 0) {
		printf("FAIL - getval range value '%s' != 'double'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_ModeIndex) != PASS) {
	    printf("FAIL - getval failed getting mode string for range\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "o") != 0) {
		printf("FAIL - range getval mode value '%s' != 'o'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_CurrentIndex) != PASS) {
	    printf("FAIL - getval failed getting current string for range\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "200") != 0) {
		printf("FAIL - range getval current value '%s' != '200'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_DefaultIndex) != PASS) {
	    printf("FAIL - getval failed getting default string for range\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "DEFAULT") != 0) {
		printf("FAIL - range getval default value '%s' != 'DEFAULT'\n",
		       TestString);
		ReturnValue++;
	    }
	}
	if (cfp_getval(FileId, "var", (void *) TestString, CFP_STRING,
		       sizeof TestString, CFP_SuppliedIndex) != PASS) {
	    printf("FAIL - getval failed getting supplied string for range\n");
	    ReturnValue++;
	} else {
	    if (strcmp(TestString, "SUPPLIED") != 0) {
		printf("FAIL - range getval supplied value '%s' != 'SUPPLIED'\n",
		       TestString);
		ReturnValue++;
	    }
	}
    }

    /* bool = { nds, original, false, true } */
    TestBool = FALSE;
    if (cfp_putval(FileId, "bool", (void*) &TestBool, CFP_BOOLEAN,
		   CFP_CurrentIndex) != PASS) {
	printf("FAIL - putval FALSE failed on bool\n");
	ReturnValue++;
    } else {
	TestBool = TRUE;
	if (cfp_getval(FileId, "bool", (void *) &TestBool, CFP_BOOLEAN,
		       sizeof TestBool, CFP_CurrentIndex) != PASS) {
	    printf("FAIL - getval failed getting boolean as boolean\n");
	    ReturnValue++;
	} else {
	    if (TestBool != FALSE) {
		printf("FAIL - getval boolean %d != %d\n", TestBool, FALSE);
		ReturnValue++;
	    }
	}
    }

    /* int1 = { , 1 } */
    TestInt = 42;
    if (cfp_putval(FileId, "int1", (void *) &TestInt, CFP_INTEGER,
		   CFP_CurrentIndex) != PASS) {
	printf("FAIL - putval 42 failed on int1\n");
	ReturnValue++;
    } else {
	TestInt = 0;
	if (cfp_getval(FileId, "int1", (void *) &TestInt, CFP_INTEGER,
		       sizeof TestInt, CFP_CurrentIndex) != PASS) {
	    printf("FAIL - getval failed getting int1\n");
	    ReturnValue++;
	} else {
	    if (TestInt != 42) {
		printf("FAIL - getval int %d != 42\n", TestInt);
		ReturnValue++;
	    }
	}
    }

    /* int2 = { , 2, , , "[0,10]" } */
    TestInt = 5;
    if (cfp_putval(FileId, "int2", (void *) &TestInt, CFP_INTEGER,
		   CFP_CurrentIndex) != PASS) {
	printf("FAIL - putval 5 failed on int2\n");
	ReturnValue++;
    } else {
	TestInt = 0;
	if (cfp_getval(FileId, "int2", (void *) &TestInt, CFP_INTEGER,
		       sizeof TestInt, CFP_CurrentIndex) != PASS) {
	    printf("FAIL - getval failed getting int2\n");
	    ReturnValue++;
	} else {
	    if (TestInt != 5) {
		printf("FAIL - getval int %d != 5\n", TestInt);
		ReturnValue++;
	    }
	}
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with bad range",
	      __LINE__);
    TestInt = 30;
    if (cfp_putval(FileId, "int2", (void *) &TestInt, CFP_INTEGER,
		   CFP_CurrentIndex) != FAIL) {
	printf("FAIL - putval 30 to int2 did not fail\n");
	ReturnValue++;
	TestInt = 0;
	if (cfp_getval(FileId, "int2", (void *) &TestInt, CFP_INTEGER,
		       sizeof TestInt, CFP_CurrentIndex) != FAIL) {
	    printf("FAIL - getval bad too, returned %d\n", TestInt);
	    ReturnValue++;
	}
    }

    /* double = { , 5.0 } */
    TestDbl = 20.5;
    if (cfp_putval(FileId, "double", (void *) &TestDbl, CFP_DOUBLE,
		   CFP_CurrentIndex) != PASS) {
	printf("FAIL - putval 20.5 failed on double\n");
	ReturnValue++;
    } else {
	TestDbl = 0.0;
	if (cfp_getval(FileId, "double", (void *) &TestDbl, CFP_DOUBLE,
		       sizeof TestDbl, CFP_CurrentIndex) != PASS) {
	    printf("FAIL - getval failed getting double\n");
	    ReturnValue++;
	} else {
	    if (TestDbl != 20.5) {
		printf("FAIL - getval double %g != 20.5\n", TestDbl);
		ReturnValue++;
	    }
	}

    }

    /* try environment variables */
    if (cfp_putval(FileId, "TEST", (void *) "123 abc", CFP_STRING,
		   CFP_EnvironmentIndex) != PASS) {
	printf("FAIL - putval '123 abc' failed on TEST\n");
	ReturnValue++;
    } else {
	if ((EnvAddress = getenv("Check_putvalTEST")) == (char *) NULL) {
	    printf("FAIL - Check_putval environment broken, no TEST\n");
	    (void) system("set");
	    ReturnValue++;
	} else {
	    if (strcmp(EnvAddress, "123 abc") != 0) {
		printf("FAIL - putval to env stored '%s' not '%s'\n",
		       EnvAddress, "123 abc");
		(void) system("set");
		ReturnValue++;
	    }
	}
    }
    TestInt = 321;
    if (cfp_putval(FileId, "INT", (void *) &TestInt, CFP_INTEGER,
		   CFP_EnvironmentIndex) != PASS) {
	printf("FAIL - putval 321 failed on INT\n");
	(void) system("set");
	ReturnValue++;
    } else {
	if ((EnvAddress = getenv("Check_putvalINT")) == (char *) NULL) {
	    printf("FAIL - Check_putval environment broken, no INT\n");
	    ReturnValue++;
	} else {
	    if (strcmp(EnvAddress, "321") != 0) {
		printf("FAIL - putval to env stored '%s' not 321\n",
		       EnvAddress);
		(void) system("set");
		ReturnValue++;
	    }
	}
    }
    if ((EnvAddress = getenv("Check_putval_NOT_there")) != (char *) NULL) {
	printf("FAIL - Check_putval environment broken, has NOT-there\n");
	ReturnValue++;
    } else {
	if (cfp_putval(FileId, "NOT_there", (void *) "new var", CFP_STRING,
		       CFP_EnvironmentIndex) != PASS) {
	    printf("FAIL - putval 'new var' failed on NOT_there\n");
	    (void) system("set");
	    ReturnValue++;
	} else {
	    if ((EnvAddress = getenv("Check_putvalNOT_there"))
		== (char *) NULL) {
		printf("FAIL - putval to env broken, no NOT_there\n");
		(void) system("set");
		ReturnValue++;
	    } else {
		if (strcmp(EnvAddress, "new var") != 0) {
		    printf("FAIL - putval to new env var stored '%s' not '%s'\n",
			   EnvAddress, "new var");
		    (void) system("set");
		    ReturnValue++;
		}
	    }
	}
    }

    /* non-existent field index */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with bad index",
	      __LINE__);
    if (cfp_putval(FileId, "var", (void *) TestString, CFP_STRING,
		   (ItemIndexType) 42) != FAIL) {
	printf("FAIL - cfp_putval of field 42 did not fail, returned '%s'\n",
	       TestString);
	ReturnValue++;
    }

    /* reference to a constant */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail - constant",
	      __LINE__);
    if (cfp_putval(FileId, "const", (void *) TestString, CFP_STRING,
		   CFP_ModeIndex) != FAIL) {
	printf("FAIL - cfp_putval of a constant did not fail, returned '%s'\n",
	       TestString);
	ReturnValue++;
    }

    /* non-existent name */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with lookup failure",
	      __LINE__);
    if (cfp_putval(FileId, "nosuchvar", (void *) TestString,
		   CFP_STRING, CFP_ModeIndex) != FAIL) {
	printf("FAIL - cfp_putval of nosuchvar did not fail, returned '%s'\n",
	       TestString);
	ReturnValue++;
    }

    if (cfp_finish(FileId) != PASS) {
	printf("FAIL - cfp_finish failed\n");
	ReturnValue++;
    }

    return(ReturnValue);
    /*NOTREACHED*/
}
