/*
 * Check_range.c - test cfp_rangep(), cfp_editvalue(), cfp_range(),
 *   and cfp_rangemess()
 *
 * This program is used to make sure that cfp_range() works correctly.  Its
 * expected environment is that the calling shell script has created a .par
 * file as defined below and called this program with no arguments.  This runs
 * through a set of tests outputing any unexpected results to stdout.  Each
 * output starts with "FAIL" for the calling shell script to find.  If any
 * range errors occurred, this exits with a positive error count.  If any
 * parse errors occurred, this exits with -1.
 *
 * The expected .par file:
 *
 * int = { , , , , int }
 * intsp = { , , , , "  int  " }
 * double = { , , , , double }
 * linf = { , , , , " ( , 0 ] " }
 * rinf = { , , , , " [ 0, ) " }
 * cc = { , , , , " [ 0 , 10 ] " }
 * oc = { , , , , " ( 0 , 10 ] " }
 * oo = { , , , , " ( 0 , 10 ) " }
 * co = { , , , , " [ 0 , 10 ) " }
 * indintmin = { , , , , " [ %indint, 15 ] " }
 * indintmax = { , , , , " [ 1, %indint ] " }
 * indint = { , 5 }"
 * inddoublemin = { , , , , " [ %inddouble, 15. ] " }
 * inddoublemax = { , , , , " [ 1., %inddouble ] " }
 * inddouble = { , 5.0 }"
 * float = { , , , , " [ -0.1e-5 , +0.1E-5 ] " }
 * time = { , , , , " [ 0:0:0 , 10:0:0 ) " }
 * norange = { , }
 * const = foobar
 * editint = { , , , , int }
 * editdouble = { , , , , double }
 * edittime = { , , , , time }
 *
 *
 * CFHTLOGS and CFHTLOGU are expected to be undefined (though this doesn't
 * care)
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "cfht.h"
#include "cfp.h"
#include "exits.h"

main(argc, argv)
    int argc;
    char *argv[];
{
    FILE_ID FileId;		  /* file index from parse */
    char *Message;		  /* message pointer from rangemess */
    int ReturnValue;		  /* starts at zero, +1 for each error */
    int Status;			  /* for cfp_rangep return value */
    int intDummy;		  /* dummy for int putcur */
    double doubleDummy;		  /* dummy for double putcur */

    ReturnValue = 0;

    if ((FileId = cfp_parse(CFP_PAR, argc, argv)) == FAIL) {
	printf("FAIL - cfp_parse died on %s\n", argv[0]);
	return(EXIT_FAIL);
    }

    /* check cfp_rangep */
    if ((Status = cfp_rangep(FileId, "int")) != TRUE) {
	printf("FAIL - cfp_rangep with range returned %d not TRUE\n", Status);
	ReturnValue++;
    }
    if ((Status = cfp_rangep(FileId, "norange")) != FALSE) {
	printf("FAIL - cfp_rangep without range returned %d not FALSE\n",
	       Status);
	ReturnValue++;
    }

    /* int = { , , , , int } */
    if (cfp_range(FileId, "int", "40", CFP_STRING) == FAIL) {
	printf("FAIL - cfp_range of int versus 40 failed for int\n");
	ReturnValue++;
    }
    Message = cfp_rangemess(FileId, "int");
    if (strcmp(Message, "Value can be any integer") != 0) {
	printf("FAIL - Message was '%s' not '%s' for int\n",
	       Message, "Value can be any integer");
	ReturnValue++;
    }

    /* intsp = { , , , , "  int  " } */
    if (cfp_range(FileId, "intsp", "-100", CFP_STRING) == FAIL) {
	printf("FAIL - cfp_range of int versus -100 failed for intsp\n");
	ReturnValue++;
    }
    Message = cfp_rangemess(FileId, "intsp");
    if (strcmp(Message, "Value can be any integer") != 0) {
	printf("FAIL - Message was '%s' not '%s' for intsp\n",
	       Message, "Value can be any integer");
	ReturnValue++;
    }

    /* double = { , , , , double } */
    if (cfp_range(FileId, "double", "5.6", CFP_STRING) == FAIL) {
	printf("FAIL - cfp_range of double versus 5.6 failed for double\n");
	ReturnValue++;
    }
    Message = cfp_rangemess(FileId, "double");
    if (strcmp(Message, "Value can be any real value or time") != 0) {
	printf("FAIL - Message was '%s' not '%s' for double\n",
	       Message, "Value can be any real value or time");
	ReturnValue++;
    }

    /* linf = { , , , , " ( , 0 ] " } */
    if (cfp_range(FileId, "linf", "-10", CFP_STRING) == FAIL) {
	printf("FAIL - cfp_range of (,0] versus -10 failed for linf\n");
	ReturnValue++;
    }
    Message = cfp_rangemess(FileId, "linf");
    if (strcmp(Message,
	       "Valid values from -inf inclusive to 0 inclusive")
	!= 0) {
	printf("FAIL - Message was '%s' not '%s' for linf\n",
	       Message, "Valid values from -inf inclusive to 0 inclusive");
	ReturnValue++;
    }

    /* rinf = { , , , , " [ 0, ) " } */
    if (cfp_range(FileId, "rinf", "1000", CFP_STRING) == FAIL) {
	printf("FAIL - cfp_range of [0,) versus 1000 failed for rinf\n");
	ReturnValue++;
    }
    Message = cfp_rangemess(FileId, "rinf");
    if (strcmp(Message,
	       "Valid values from 0 inclusive to +inf inclusive")
	!= 0) {
	printf("FAIL - Message was '%s' not '%s' for rinf\n",
	       Message, "Valid values from 0 inclusive to +inf inclusive");
	ReturnValue++;
    }

    /* cc = { , , , , " [ 0 , 10 ] " } */
    if (cfp_range(FileId, "cc", "10", CFP_STRING) == FAIL) {
	printf("FAIL - cfp_range of [0,10] versus 10 failed for cc\n");
	ReturnValue++;
    }
    Message = cfp_rangemess(FileId, "cc");
    if (strcmp(Message,
	       "Valid values from 0 inclusive to 10 inclusive")
	!= 0) {
	printf("FAIL - Message was '%s' not '%s' for cc\n",
	       Message, "Valid values from 0 inclusive to 10 inclusive");
	ReturnValue++;
    }

    /* oc = { , , , , " ( 0 , 10 ] " } */
    if (cfp_range(FileId, "oc", "1", CFP_STRING) == FAIL) {
	printf("FAIL - cfp_range of (0,10] versus 1 failed for oc\n");
	ReturnValue++;
    }
    Message = cfp_rangemess(FileId, "oc");
    if (strcmp(Message,
	       "Valid values from 0 exclusive to 10 inclusive")
	!= 0) {
	printf("FAIL - Message was '%s' not '%s' for oc\n",
	       Message, "Valid values from 0 exclusive to 10 inclusive");
	ReturnValue++;
    }

    /* oo = { , , , , " ( 0 , 10 ) " } */
    if (cfp_range(FileId, "oo", "9", CFP_STRING) == FAIL) {
	printf("FAIL - cfp_range of [0,10) versus 9 failed for oo\n");
	ReturnValue++;
    }
    Message = cfp_rangemess(FileId, "oo");
    if (strcmp(Message,
	       "Valid values from 0 exclusive to 10 exclusive")
	!= 0) {
	printf("FAIL - Message was '%s' not '%s' for oo\n",
	       Message, "Valid values from 0 exclusive to 10 exclusive");
	ReturnValue++;
    }

    /* co = { , , , , " [ 0 , 10 ) " } */
    if (cfp_range(FileId, "co", "0", CFP_STRING) == FAIL) {
	printf("FAIL - cfp_range of [0,10) versus 0 failed for co\n");
	ReturnValue++;
    }
    Message = cfp_rangemess(FileId, "co");
    if (strcmp(Message,
	       "Valid values from 0 inclusive to 10 exclusive")
	!= 0) {
	printf("FAIL - Message was '%s' not '%s' for co\n",
	       Message, "Valid values from 0 inclusive to 10 exclusive");
	ReturnValue++;
    }

    /* indintmin = { , , , , " [ %indint, 15 ] " } */
    /* indintmax = { , , , , " [ 1, %indint ] " } */
    /* indint = { , 5 }" */
    intDummy = 10;
    if (cfp_putcur(FileId, "indintmin", (void *) &intDummy, CFP_INTEGER)
	!= PASS) {
	printf("FAIL - cfp_putcur of 10 into (indirect 5)..15 failed\n");
	ReturnValue++;
    }
    intDummy = 5;
    if (cfp_putcur(FileId, "indintmax", (void *) &intDummy, CFP_INTEGER)
	!= PASS) {
	printf("FAIL - cfp_putcur of 5 into 1..(indirect 5) failed\n");
	ReturnValue++;
    }
    intDummy = 1;
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following should be out of range",
	      __LINE__);
    if (cfp_putcur(FileId, "indintmin", (void *) &intDummy, CFP_INTEGER)
	!= FAIL) {
	printf("FAIL - cfp_putcur of 1 into (indirect 5)..15 did not fail\n");
	ReturnValue++;
    }
    intDummy = 10;
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following should be out of range",
	      __LINE__);
    if (cfp_putcur(FileId, "indintmax", (void *) &intDummy, CFP_INTEGER)
	!= FAIL) {
	printf("FAIL - cfp_putcur of 10 into 1..(indirect 5) did not fail\n");
	ReturnValue++;
    }

    /* inddoublemin = { , , , , " [ %inddouble, 15. ] " } */
    /* inddoublemax = { , , , , " [ 1., %inddouble ] " } */
    /* inddouble = { , 5.0 }" */
    doubleDummy = 10.0;
    if (cfp_putcur(FileId, "inddoublemin", (void *) &doubleDummy, CFP_DOUBLE)
	!= PASS) {
	printf("FAIL - cfp_putcur of 10.0 into (indirect 5)..15 failed\n");
	ReturnValue++;
    }
    doubleDummy = 5.0;
    if (cfp_putcur(FileId, "inddoublemax", (void *) &doubleDummy, CFP_DOUBLE)
	!= PASS) {
	printf("FAIL - cfp_putcur of 5.0 into 1..(indirect 5) failed\n");
	ReturnValue++;
    }
    doubleDummy = 1.0;
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following should be out of range",
	      __LINE__);
    if (cfp_putcur(FileId, "inddoublemin", (void *) &doubleDummy, CFP_DOUBLE)
	!= FAIL) {
	printf("FAIL - cfp_putcur of 1.0 into (indirect 5)..10 did not fail\n");
	ReturnValue++;
    }
    doubleDummy = 10.0;
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following should be out of range",
	      __LINE__);
    if (cfp_putcur(FileId, "inddoublemax", (void *) &doubleDummy, CFP_DOUBLE)
	!= FAIL) {
	printf("FAIL - cfp_putcur of 10.0 into 1..(indirect 5) did not fail\n");
	ReturnValue++;
    }

    /* float = { , , , , " [ -0.1e-5 , +0.1E-5 ] " } */
    if (cfp_range(FileId, "float", "0", CFP_STRING) == FAIL) {
	printf("FAIL - cfp_range of [-.1e-5,.1e-5] versus 0 failed for float\n");
	ReturnValue++;
    }
    Message = cfp_rangemess(FileId, "float");
    if (strcmp(Message,
	       "Valid values from -0.1e-5 inclusive to +0.1E-5 inclusive")
	!= 0) {
	printf("FAIL - Message was '%s' not '%s' for float\n",
	       Message,
	       "Valid values from -0.1e-5 inclusive to +0.1E-5 inclusive");
	ReturnValue++;
    }

    /* time = { , , , , " [ 0:0:0 , 10:0:0 ) " } */
    if (cfp_range(FileId, "time", "9:59:59.9", CFP_STRING) == FAIL) {
	printf("FAIL - cfp_range of [0:0:0,10:0:0) versus 9:59:59.9 failed for time\n");
	ReturnValue++;
    }
    Message = cfp_rangemess(FileId, "time");
    if (strcmp(Message,
	       "Valid values from 0:0:0 inclusive to 10:0:0 exclusive")
	!= 0) {
	printf("FAIL - Message was '%s' not '%s' for time\n",
	       Message, "Valid values from 0:0:0 inclusive to 10:0:0 exclusive");
	ReturnValue++;
    }

    /* norange = { , } */
    if (cfp_range(FileId, "norange", "10", CFP_STRING) == FAIL) {
	printf("FAIL - cfp_range with no range failed for norange\n");
	ReturnValue++;
    }
    Message = cfp_rangemess(FileId, "norange");
    if (strstr(Message, "No range field") == (char *) NULL) {
	printf("FAIL - Message '%s' did not include 'No range field' for norange\n",
	       Message);
	ReturnValue++;
    }

    /* try to set norange to " foo [ 0 , 10 ] " */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following has a bad range type",
	      __LINE__);
    if (cfp_putval(FileId, "norange", (void *) "foo[0,10]", CFP_STRING,
		   CFP_RangeIndex) != FAIL) {
	printf("FAIL - cfp_putval with bad left did not fail for norange\n");
	ReturnValue++;
    }

    /* try to set norange to " < 0 , 10 > " */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following has a bad range left end",
	      __LINE__);
    if (cfp_putval(FileId, "norange", (void *) "<0,10>", CFP_STRING,
		   CFP_RangeIndex) != FAIL) {
	printf("FAIL - cfp_putval with bad left did not fail for norange\n");
	ReturnValue++;
    }

    /* try to set norange to " ( 1, 10 > " */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following has a bad range right end",
	      __LINE__);
    if (cfp_putval(FileId, "norange", (void *) "(1,10>", CFP_STRING,
		   CFP_RangeIndex) != FAIL) {
	printf("FAIL - cfp_putval with bad right did not fail for norange\n");
	ReturnValue++;
    }

    /* try to set norange to " [ 0, 10 ] junk " */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following has a bad range - junk after",
	      __LINE__);
    if (cfp_putval(FileId, "norange", (void *) "[0,10] junk", CFP_STRING,
		   CFP_RangeIndex) != FAIL) {
	printf("FAIL - cfp_putval with junk after did not fail for norange\n");
	ReturnValue++;
    }

    /* const = foobar */
    if (cfp_range(FileId, "const", "10", CFP_STRING) == FAIL) {
	printf("FAIL - cfp_range on constant failed for const\n");
	ReturnValue++;
    }
    Message = cfp_rangemess(FileId, "const");
    if (strstr(Message, "'const' is not a variable") == (char *) NULL) {
	printf("FAIL - Message '%s' did not include 'not a var' for const\n",
	       Message);
	ReturnValue++;
    }

    /* non-existent name */
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following is supposed to fail with lookup failure",
	      __LINE__);
    if (cfp_range(FileId, "nosuchvar", "1", CFP_STRING) != FAIL) {
	printf("FAIL - cfp_range of nosuchvar did not fail\n");
	ReturnValue++;
    }
    Message = cfp_rangemess(FileId, "nosuchvar");
    if (strstr(Message, "Name 'nosuchvar' not found") == (char *) NULL) {
	printf("FAIL - Message '%s' did not include 'not found' for nosuchvar\n",
	       Message);
	ReturnValue++;
    }

    /* test value editing */

    if (cfp_editvalue(FileId, "nosuchvar", (void *) "whatever", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'nosuchvar' did not fail.\n");
	ReturnValue++;
    }

    if (cfp_editvalue(FileId, "editint", (void *) "+", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'editint' with '+' failed.\n");
	ReturnValue++;
    }
    if (cfp_editvalue(FileId, "editint", (void *) "-", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'editint' with '-' failed.\n");
	ReturnValue++;
    }
    if (cfp_editvalue(FileId, "editint", (void *) "123", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'editint' with '123' failed.\n");
	ReturnValue++;
    }
    if (cfp_editvalue(FileId, "editint", (void *) "+1234", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'editint' with '+1234' failed.\n");
	ReturnValue++;
    }
    if (cfp_editvalue(FileId, "editint", (void *) "-9876", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'editint' with '-9876' failed.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following editint '+-' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "editint", (void *) "+-", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'editint' with '+-' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following editint 'abc' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "editint", (void *) "abc", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'editint' with 'abc' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following editint '1 2' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "editint", (void *) "1 2", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'editint' with '1 2' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following editint '12abc' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "editint", (void *) "12abc", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'editint' with '12abc' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following editint '12:00' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "editint", (void *) "12:00", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'editint' with '12:00' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following editint '50.0' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "editint", (void *) "50.0", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'editint' with '50.0' did not fail.\n");
	ReturnValue++;
    }

    if (cfp_editvalue(FileId, "editdouble", (void *) "+", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'editdouble' with '+' failed.\n");
	ReturnValue++;
    }
    if (cfp_editvalue(FileId, "editdouble", (void *) "-", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'editdouble' with '-' failed.\n");
	ReturnValue++;
    }
    if (cfp_editvalue(FileId, "editdouble", (void *) "123", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'editdouble' with '123' failed.\n");
	ReturnValue++;
    }
    if (cfp_editvalue(FileId, "editdouble", (void *) "+12.34", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'editdouble' with '+12.34' failed.\n");
	ReturnValue++;
    }
    if (cfp_editvalue(FileId, "editdouble", (void *) "+12:", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'editdouble' with '+12:' failed.\n");
	ReturnValue++;
    }
    if (cfp_editvalue(FileId, "editdouble", (void *) "-9876e-", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'editdouble' with '-9876e-' failed.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following editdouble '+-' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "editdouble", (void *) "+-", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'editdouble' with '+-' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following editdouble 'abc' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "editdouble", (void *) "abc", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'editdouble' with 'abc' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following editdouble '1 2' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "editdouble", (void *) "1 2", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'editdouble' with '1 2' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following editdouble '12abc' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "editdouble", (void *) "12abc", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'editdouble' with '12abc' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following editdouble '12:00am' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "editdouble", (void *) "12:00am", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'editdouble' with '12:00am' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following editdouble '50..0' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "editdouble", (void *) "50..0", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'editdouble' with '50..0' did not fail.\n");
	ReturnValue++;
    }

    if (cfp_editvalue(FileId, "edittime", (void *) "+", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'edittime' with '+' failed.\n");
	ReturnValue++;
    }
    if (cfp_editvalue(FileId, "edittime", (void *) "-", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'edittime' with '-' failed.\n");
	ReturnValue++;
    }
    if (cfp_editvalue(FileId, "edittime", (void *) "123", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'edittime' with '123' failed.\n");
	ReturnValue++;
    }
    if (cfp_editvalue(FileId, "edittime", (void *) "+12.34", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'edittime' with '+12.34' failed.\n");
	ReturnValue++;
    }
    if (cfp_editvalue(FileId, "edittime", (void *) "12:13.", CFP_STRING)
	!= PASS) {
	printf("FAIL - cfp_editvalue of 'edittime' with '12:13.' failed.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following edittime '+-' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "edittime", (void *) "+-", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'edittime' with '+-' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following edittime 'abc' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "edittime", (void *) "abc", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'edittime' with 'abc' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following edittime '1 2' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "edittime", (void *) "1 2", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'edittime' with '1 2' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following edittime '12abc' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "edittime", (void *) "12abc", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'edittime' with '12abc' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following edittime '12:00am' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "edittime", (void *) "12:00am", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'edittime' with '12:00am' did not fail.\n");
	ReturnValue++;
    }
    cfht_logv(CFHT_MAIN, CFHT_LOGONLY,
	      "(main:%d) following edittime '50:::0' should fail",
	      __LINE__);
    if (cfp_editvalue(FileId, "edittime", (void *) "50:::0", CFP_STRING)
	!= FAIL) {
	printf("FAIL - cfp_editvalue of 'edittime' with '50:::0' did not fail.\n");
	ReturnValue++;
    }

    if (cfp_finish(FileId) != PASS) {
	printf("FAIL - cfp_finish failed\n");
	ReturnValue++;
    }

    return(ReturnValue);
    /*NOTREACHED*/
}
