/*
 * signaltest.c - conduct three tests of "signal" and "sigaction" behavior:
 *
 * 1. Does the calling process receive SIG_CHLD if the child is stopped?
 * 2. Does an interrupted system call restart instead of returning EINTR?
 * 3. Does the signal handler have to reinstall itself?
 */

#include <signal.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/types.h>
#include <string.h>

static volatile int sig_count=0;

#ifndef WAIT_ANY
#define WAIT_ANY -1
#endif

void
wait_for_children(void)
{
  static int w;
  while (waitpid(WAIT_ANY,&w,WNOHANG|WUNTRACED)>0);
}

void
signal_handler(int signo)
{
  sig_count++;
  wait_for_children();
}

#ifdef SIG_RESTART
void
signal_handler_scp(int signo, int code, struct sigcontext* scp)
{
  sig_count++;
  if (scp!=NULL)
    scp->sc_syscall_action = SIG_RESTART;
}
#endif

void
do_test_3(void)
{
  pid_t cpid=0;

  printf("Does the calling process receive SIG_CHLD if the child is stopped? ");
  fflush(stdout);
  sig_count=0;

  if (!(cpid=fork())) /* child */
  {
    sleep(1);
    kill(getpid(),SIGTSTP);
    sleep(2);
    _exit(0);
  }
  else /* parent */
  {
    int answer=0;
    sig_count=0;
    sleep(2);
    answer=sig_count;
    kill(cpid,SIGCONT);
    sleep(3);
    if (sig_count==0)
      printf("FAILED\n");
    else
      printf("%s\n",answer?"yes":"no");
  }
  wait_for_children();
}

void
do_test_2(void)
{
  pid_t cpid=0;
  int pipefd[2];

  printf("Does an interrupted system call restart instead of returning EINTR? ");
  fflush(stdout);
  sig_count=0;

  pipe(pipefd);

  if (!fork())
  {
    close(pipefd[0]);
    sleep(1);
    /* write(pipefd[1],"ab",2); */
    sleep(4);
    write(pipefd[1],"cd",2);
    _exit(0);
  }

  close(pipefd[1]);

  if (!(cpid=fork())) /* child */
  {
     sleep(2);
     _exit(0);
  }
  else /* parent */
  {
    int count=0, answer=1;
    char buf[4];

    if (cpid<=0) perror("fork");

    if ((count=read(pipefd[0],buf,4))<=0)
    {
      if (errno==EINTR)
        answer=0;
      else
      {
        perror("read");
	answer=2;
      }
    }
    if (sig_count==0)
      printf("FAILED\n");
    else
      if (answer==0) printf("no\n");
    else
      if (answer==1) printf("yes\n");
    else 
      printf("dunno\n");
  }
  wait_for_children();
}

void
do_test_1(void)
{
  printf("Does the signal handler have to reinstall itself? ");
  fflush(stdout);
  sig_count=0;

  if (!fork())
  {
    sleep(1);
    _exit(0);
  }
  sleep(2);
  if (sig_count==0)
  {
     printf("FAILED\n");
     wait_for_children();
     return;
  }

  if (!fork()) /* child */
  {
     sleep(1);
     _exit(0);
  }
  sig_count=0;
  sleep(2);
  printf("%s\n",sig_count==1?"no":"yes");
  wait_for_children();
}

int
main(void)
{
  static struct sigaction act;
#ifdef SV_RESETHAND
  static struct sigvec vec;
#endif

  printf ("\nTesting with sigaction()...\n\n");

  act.sa_handler=signal_handler;
  sigemptyset(&act.sa_mask);
  act.sa_flags=0;
  sigaction(SIGCHLD,&act,0);
  do_test_1();
  sigaction(SIGCHLD,&act,0);
  do_test_2();
  sigaction(SIGCHLD,&act,0);
  do_test_3();

#ifdef SA_RESETHAND
  printf("\nTesting with sigaction() w/SA_RESETHAND...\n\n");
  act.sa_handler=signal_handler;
  sigemptyset(&act.sa_mask);
  act.sa_flags=SA_RESETHAND;
  if (sigaction(SIGCHLD,&act,0)!=0) perror("sigaction");
  do_test_1();
  if (sigaction(SIGCHLD,&act,0)!=0) perror("sigaction");
  do_test_2();
  if (sigaction(SIGCHLD,&act,0)!=0) perror("sigaction");
  do_test_3();
#else
  printf("\nsigaction() with SA_RESETHAND not found.\n\n");
#endif

#ifndef SIG_RESTART
  printf("\nTesting with sigaction() w/SA_RESTART...\n\n");
  act.sa_handler=signal_handler;
  sigemptyset(&act.sa_mask);
  act.sa_flags=SA_RESTART;
  if (sigaction(SIGCHLD,&act,0)!=0) perror("sigaction");
  do_test_1();
  if (sigaction(SIGCHLD,&act,0)!=0) perror("sigaction");
  do_test_2();
  if (sigaction(SIGCHLD,&act,0)!=0) perror("sigaction");
  do_test_3();
  act.sa_handler=signal_handler;
#else
  printf("\nTesting with sigaction() w/SIG_RESTART...\n\n");
  act.sa_handler=signal_handler_scp;
  sigemptyset(&act.sa_mask);
  act.sa_flags=0;
  if (sigaction(SIGCHLD,&act,0)!=0) perror("sigaction");
  do_test_1();
  if (sigaction(SIGCHLD,&act,0)!=0) perror("sigaction");
  do_test_2();
  if (sigaction(SIGCHLD,&act,0)!=0) perror("sigaction");
  do_test_3();
  act.sa_handler=signal_handler;
#endif

#ifdef SA_INTERRUPT
  printf("\nTesting with sigaction() w/SA_INTERRUPT...\n\n");
  act.sa_handler=signal_handler;
  sigemptyset(&act.sa_mask);
  act.sa_flags=SA_INTERRUPT;
  if (sigaction(SIGCHLD,&act,0)!=0) perror("sigaction");
  do_test_1();
  if (sigaction(SIGCHLD,&act,0)!=0) perror("sigaction");
  do_test_2();
  if (sigaction(SIGCHLD,&act,0)!=0) perror("sigaction");
  do_test_3();
  act.sa_handler=signal_handler;
#else
  printf("\nsigaction() with SA_INTERRUPT not found.\n\n");
#endif

#ifdef SA_NOCLDSTOP
  printf("\nTesting with sigaction() w/SA_NOCLDSTOP...\n\n");
  act.sa_handler=signal_handler;
  sigemptyset(&act.sa_mask);
  act.sa_flags=SA_NOCLDSTOP;
  sigaction(SIGCHLD,&act,0);
  do_test_1();
  sigaction(SIGCHLD,&act,0);
  do_test_2();
  sigaction(SIGCHLD,&act,0);
  do_test_3();
#else
  printf("\nsigaction() with SA_NOCLDSTOP not found.\n\n");
#endif

  printf("\nTesting with signal()...\n\n");

  if (signal(SIGCHLD,signal_handler)==SIG_ERR)
    perror("signal");
  do_test_1();
  if (signal(SIGCHLD,signal_handler)==SIG_ERR)
    perror("signal");
  do_test_2();
  if (signal(SIGCHLD,signal_handler)==SIG_ERR)
    perror("signal");
  do_test_3();
  signal(SIGCHLD,SIG_DFL);

#ifdef SV_RESETHAND
#ifdef SV_BSDSIG
  printf ("\nTesting with sigvector()...\n\n");
  vec.sv_handler=signal_handler;
  vec.sv_mask=0;
  vec.sv_flags=0;
  sigvector(SIGCHLD,&vec,0);
  do_test_1();
  sigvector(SIGCHLD,&vec,0);
  do_test_2();
  sigvector(SIGCHLD,&vec,0);
  do_test_3();

  printf("\nTesting with sigvector() w/SV_RESETHAND...\n\n");
  vec.sv_flags=SV_RESETHAND;
  sigvector(SIGCHLD,&vec,0);
  do_test_1();
  sigvector(SIGCHLD,&vec,0);
  do_test_2();
  sigvector(SIGCHLD,&vec,0);
  do_test_3();

  printf("\nTesting with sigvector() w/SV_BSDSIG...\n\n");
  vec.sv_flags=SV_BSDSIG;
  sigvector(SIGCHLD,&vec,0);
  do_test_1();
  sigvector(SIGCHLD,&vec,0);
  do_test_2();
  sigvector(SIGCHLD,&vec,0);
  do_test_3();
#else
  printf ("\nTesting with sigvec()...\n\n");
  vec.sv_handler=signal_handler;
  vec.sv_mask=0;
  vec.sv_flags=0;
  sigvec(SIGCHLD,&vec,0);
  do_test_1();
  sigvec(SIGCHLD,&vec,0);
  do_test_2();
  sigvec(SIGCHLD,&vec,0);
  do_test_3();

  printf("\nTesting with sigvec() w/SV_RESETHAND...\n\n");
  vec.sv_flags=SV_RESETHAND;
  sigvec(SIGCHLD,&vec,0);
  do_test_1();
  sigvec(SIGCHLD,&vec,0);
  do_test_2();
  sigvec(SIGCHLD,&vec,0);
  do_test_3();

  printf("\nTesting with sigvec() w/SV_INTERRUPT...\n\n");
  vec.sv_flags=SV_INTERRUPT;
  sigvec(SIGCHLD,&vec,0);
  do_test_1();
  sigvec(SIGCHLD,&vec,0);
  do_test_2();
  sigvec(SIGCHLD,&vec,0);
  do_test_3();
#endif
#else
  printf("\nsigvec() not found.\n\n");
#endif

  return 0;
}







