/****************************************************************************
    Copyright (C) 1991-2002    Canada-France-Hawaii Telescope Corp.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact information:
    CFHT, PO Box 1597, Kamuela, HI 96743, USA
    daprog@cfht.hawaii.edu

****************************************************************************/
/*!****************************************************************************
 * FILE roll.c
 *
 * $Id: roll.c,v 2.7 2003/03/09 20:43:28 isani Locked $
 * $Locker: isani $
 *
 * ROUTINES
 *
 *  CFHT logging roll/unroll program main and argument processing
 *
 * HISTORY
 * who          when            what
 * ---------    ------------    ----------------------------------------------
 *
 * $Log:	roll.c,v $
 * Revision 2.7  2003/03/09  20:43:28  isani
 * Added options of (C)ooked and (R)eversed output on unroll.
 * 
 * Revision 2.6  2002/05/07  18:40:18  thomas
 * Make large -h values start at time 0 instead of wrapping
 * 
 * Revision 2.5  2002/05/05 22:44:41  isani
 * roll will create the pipe name given to '-p' as long as the directory it
 *   needs to be in exists, or is /tmp/pipes/ (e.g., /tmp/pipes/syslog.np.)
 *   If the pipe already exists, that is fine too.
 *
 * Revision 2.4  2002/05/05  21:33:22  isani
 * ANSI-fied and converted to gcc/Make.Common.
 * Removed non-functional setuid feature (it's in RCS).
 * Properly detach from terminal instead of just calling setuid().
 *   This happens only after roll is all ready, so startup scripts should
 *   no longer start roll in the background and sleep for 2 seconds (which
 *   was not always long enough.)  Instead, start it in the foreground and
 *   let it detach like a normal daemon process.
 * 
 * Revision 2.3  2000/12/12 02:53:24  thomas
 * Change to use getopt for Sidik :-)
 *
 * Revision 2.2  1996/11/19 02:59:26  isani
 * cleaned up rcsid's
 * moved some old-sun hacks to cfht.h
 *
 * Revision 2.1  1996/03/11  08:09:49  thomas
 * Disconnect from terminal if non-Verbose roll (for protection from rc)
 * A last bit of formatting
 *
 * Revision 2.0  95/11/22  09:31:44  09:31:44  thomas (Jim Thomas)
 * Changed -t to -h
 * Added -f -t
 * 
 * Revision 1.4  95/06/20  09:37:06  09:37:06  viloria (Ron Viloria)
 * Added time option [-h hours], where "hours" is the number of hours to list 
 * output lines from now backwards from the file. 
 * 
 * Revision 1.3  94/12/22  12:30:22  12:30:22  thomas (Jim Thomas)
 * Standardize exit codes to exits.h
 * 
 * Revision 1.2  94/02/14  16:10:43  16:10:43  jwright (Jim Wright)
 * fix it so it really uses the suid bit when running on either sun or hp
 * 
 * Revision 1.1  91/09/30  12:13:47  12:13:47  jwright (Jim Wright)
 * Original checkin
 * 
 ***************************************************************************!*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>		  /* for string functions */
#include <stdlib.h>		  /* for strtoul and EXIT_SUCCESS/FAILURE */
#include <ctype.h>		  /* for isalpha, isupper, tolower */
#if defined(sun) && !defined(__svr4__)
#define strtoul strtol		  /*   but sun doesn't have one */
#endif
#include <unistd.h>		  /* for setsid, getopt, and sysconf */
#include <time.h>		  /* for time functions */
#include <errno.h>		  /* for ERANGE */

#include "cfht/cfht.h"		  /* for RCSID and cfht_basename() */

RCSID("$Id: roll.c,v 2.7 2003/03/09 20:43:28 isani Locked $");

#include "roll.h"

void roll_usage P((char *s));	  /* usage exit */
PASSFAIL TimeStringToTime P((char *String, time_t *Time));

time_t TimeNow;			  /* current time */

typedef struct ReverseTag {
    struct ReverseTag* next;
    char line[1];
} ReverseBuf;

/*
 * child_pass and child_fail are run on SIGUSR1 and SIGUSR2
 * respectively, depending on whether the "roll" child
 * is able to get a lock.
 */
static void child_pass(void)
{
    fprintf(stderr, "roll: continuing as daemon process\n");
    exit(EXIT_SUCCESS);
}
static void child_fail(void)
{
    fprintf(stderr, "roll: daemon failed to get file lock\n");
    exit(EXIT_FAILURE);
}

int
main(int argc, char* argv[])
{
    /* command line argument values */
    char *OurName;		  /* name of executing program */
    char *RollName = (char *) NULL; /* name of roll file */
    char *Input = (char *) NULL;  /* name of [pipe] to read */
    char *SessionName = (char *) NULL; /* name of session to match */
    char *FromString = (char *) NULL; /* time to start outputting */
    char *ToString = (char *) NULL; /* time to stop outputting */
    BOOLEAN IsRoll = FALSE;	  /* is this roll or unroll? */
    BOOLEAN Verbose = FALSE;	  /* want lots of debug output? */
    BOOLEAN GoBackwards = TRUE;	  /* want LIFO or FIFO?  TRUE = LIFO */
    BOOLEAN ReverseOutput = FALSE; /* for unroll, reverse in memory */
    BOOLEAN CookedOutput = FALSE; /* Re-format like unix syslog */
    off_t FileSize = 0;		  /* desired roll file size if creating */
    unsigned int HoursBack = 0;	  /* start output this far back */
    time_t FromTime;		  /* from time in internal form */
    time_t ToTime;		  /* to time in internal form */

    RollType *roll_id;		  /* roll file description */

    ReverseBuf *ReversedLines = NULL;

    int c;			  /* character being copied to/from file */
				  /*   or getopt option character */
    
    extern char *optarg; 
    extern int optind, optopt; 

    /* use name (roll versus unroll) to default what we do (write or read) */
    OurName = cfht_basename((char *) NULL, argv[0], (char *) NULL);
    if (OurName != (char *) NULL) {
	if (strcmp(OurName, "roll") == 0) {
	    IsRoll = TRUE;
	}
    }

    time(&TimeNow);		  /* get current time */

    opterr = 0;
    while ((c = getopt(argc, argv, ":CRFLbvrup:Ss:a:f:t:h:")) > 0) {
	switch (c) {
	  case 'C':
	      CookedOutput = TRUE;
	      break;

	  case 'R':
	      CookedOutput = FALSE;
	      break;

	  case 'F':
	      GoBackwards = FALSE;
	      break;

	  case 'L':
	  case 'b':
	      GoBackwards = TRUE;
	      break;

	  case 'v':
	      Verbose = TRUE;
	      break;

	  case 'r':
	      IsRoll = TRUE;
	      break;

	  case 'u':
	      IsRoll = FALSE;
	      break;

	  case 'p':		  /* if -p filename then open r/w as input */
	      Input = optarg;
	      break;

	  case 'S':
	      /* Dosuid feature had no purpose.  Removed by Sidik. */
	      break;

	  case 's':
	      FileSize = strtol(optarg, (char **) NULL, 0);
	      break;

	  case 'a':
	      SessionName = optarg; /* get session from command line */
	      break;

	  case 'f':
	      FromString = optarg; /* get from time from command line */
	      break;

	  case 't':
	      ToString = optarg; /* get to time from command line */
	      break;

	  case 'h':
	      HoursBack = strtoul(optarg, (char **) NULL, 0); /* get hrs */
	      break;

	  case ':':
	      printf("%s: missing argument value for %c\n", argv[0], optopt);
	      exit(EXIT_FAILURE);
	      break;

	  case '?':
	      printf("%s: unknown option %c\n", argv[0], optopt);
	      exit(EXIT_FAILURE);
	      break;

	  default:
	      printf("%s: unknown option %c\n", argv[0], optopt);
	      exit(EXIT_FAILURE);
	}
    }

    if ((optind + 1) == argc) {
	RollName = argv[optind];
    } else {
	if (optind != argc) {
	    roll_usage(argv[0]);  /* else error */
	}
    }

    /* default roll file name if not given */
    if (RollName == (char *) NULL) {
	if ((RollName = (char *) malloc(strlen(CFHT_CONFPATH) + 1 +
					strlen(ROLL_NAME) + 1))
	    == (char *) NULL) {
	    fprintf(stderr, "%s: error getting space for default roll file name\n",
		    OurName);
	    exit(EXIT_FAILURE);
	}
	(void) strcpy(RollName, CFHT_CONFPATH);
	(void) strcat(RollName, "/");
	(void) strcat(RollName, ROLL_NAME);
    }

    /* make sure time options are consistent and start defaulting */
    if (HoursBack > 0) {
	if ((FromString != (char *) NULL) || (ToString != (char *) NULL))
	  roll_usage(argv[0]);	  /* then error */ 
    } else {
	if ((FromString == (char *) NULL) && (ToString == (char *) NULL))
	  HoursBack = DEFTIME;
    }

    /* if to and from times are to be used (meaningless if we're roll) */
    if (HoursBack == 0) {
	if (FromString == (char *) NULL) {
	    /* start at "zero" */
	    FromTime = 0;
	    FromString = "time 0";
	} else {
	    if (TimeStringToTime(FromString, &FromTime) != PASS) {
		fprintf(stderr, "%s: bad -f time '%s'\n", OurName,
			FromString);
		exit(EXIT_FAILURE);
	    }
	}
	if (ToString == (char *) NULL) {
	    /* end now */
	    ToTime = TimeNow;
	    ToString = "now";
	} else {
	    if (TimeStringToTime(ToString, &ToTime) != PASS) {
		fprintf(stderr, "%s: bad -t time '%s'\n", OurName,
			ToString);
		exit(EXIT_FAILURE);
	    }
	}

	/* if "from" time is later than "to" time */
	if (FromTime > ToTime) {
	    fprintf(stderr, "%s: -f time '%s' after -t time '%s' ?\n",
		    OurName, FromString, ToString);
	    exit(EXIT_FAILURE);
	}

    } else {
	/* Convert HoursBack into an absolute time */
	/* OneHour = OneAM - Midnight */
	/* FromTime = TimeNow - (OneHour * HoursBack) */
	struct tm tmScratch;
	struct tm *tmPointer;
	time_t MidNight;
	time_t OneAM;
	time_t OneHour;

	/* figure out when midnight is */
	tmPointer = localtime(&TimeNow);
	(void) memcpy((void *) &tmScratch, (void *) tmPointer, sizeof tmScratch);
	tmScratch.tm_sec = 0;
	tmScratch.tm_min = 0;
	tmScratch.tm_hour = 0;
	MidNight = mktime(&tmScratch);

	/* figure out when 1AM is */
	tmScratch.tm_hour = 1;
	OneAM = mktime(&tmScratch);

	/* get one hour in time_t form */
	/*  NOTE: probably = 3600, but that may not be guaranteed */
	OneHour = OneAM - MidNight;

	/* make sure we do not wrap 2038 :-) */
	if (HoursBack > (TimeNow / OneHour))
	{
	    fprintf(stderr, "-h value (%d) before unix time 0, using 0\n",
		    HoursBack);
	    FromTime = 0;
	}
	else
	{
	    /* get -h hours in absolute time_t form */
	    FromTime = TimeNow - (HoursBack * OneHour);
	}

	/* end now */
	ToTime = TimeNow;
    }
    
    if ((roll_id = roll_init(RollName, FileSize, GoBackwards, IsRoll, Verbose))
       == (RollType *) NULL) {
	fprintf(stderr, "%s: roll_init() of '%s' failed\n", OurName, RollName);
	exit(EXIT_FAILURE);
    }

    roll_id->From = FromTime;
    roll_id->To = ToTime;
    roll_id->Session = SessionName;

    if(IsRoll == TRUE) {
	if(Input != (char *) NULL) {
	    struct stat stbuf;

	    /*
	     * No harm making sure the CFHT-standard directory for CFHTLOGS
	     * exists in /tmp (but fail further down if there's a problem.)
	     */
  	    if ((mkdir("/tmp/pipes",
		      S_IRUSR|S_IWUSR|S_IXUSR|
		      S_IRGRP|S_IWGRP|S_IXGRP|
		      S_IROTH|S_IWOTH|S_IXOTH) < 0 && errno != EEXIST) ||
		chmod("/tmp/pipes",
		      S_IRUSR|S_IWUSR|S_IXUSR|
		      S_IRGRP|S_IWGRP|S_IXGRP|
		      S_IROTH|S_IWOTH|S_IXOTH) < 0)
	        perror("roll: problem setting up /tmp/pipes (continuing)");

	    if (stat(Input, &stbuf)==0) {
	        if (!S_ISFIFO(stbuf.st_mode)) {
		    fprintf(stderr, "%s: warning: %s is not a FIFO!\n",
			    OurName, Input);
		}
#if 0
		/* Uncomment this to enable a check that the pipe is not being
		 * read.  There is a race condition with this method, and the
		 * file lock below should take care of things.  Unfortunately,
		 * the file lock has to be done after the fork, so some hacking
		 * was needed to signal the parent, but it works, and remains
		 * compatible with older versions of roll which may get run by
		 * some sessions that will check for the lock.
		 */
		else {
		    /*
		     * Try to determine if another roll process already has this
		     * FIFO open for reading.  A non-blocking open() of a pipe
		     * that is NOT being read yet should FAIL.  I.e., it is a bad
		     * thing if this open call does not fail.
		     */
		    int fd = open(Input, O_WRONLY | O_NONBLOCK, 0);
		    if (fd != -1) {
			close(fd);
			fprintf(stderr,
				"%s: error: The named pipe `%s' is already being read by another process - EXITING!\n",
				OurName, Input);
			exit(EXIT_FAILURE);
		    }
		}
#endif
	    }
	    else {
	      if (Verbose == TRUE)
		fprintf(stderr, "%s: creating pipe %s\n", OurName, Input);
	      if (mknod(Input, S_IFIFO|S_IRUSR|S_IWUSR|S_IWGRP|S_IWOTH, 0) < 0 ||
		  chmod(Input,         S_IRUSR|S_IWUSR|S_IWGRP|S_IWOTH) < 0)
		perror("roll: failed to set up Input pipe");
	    }

	    if(Verbose == TRUE)
	      fprintf(stderr, "%s: re-opening %s\n", OurName, Input);
	    if(freopen(Input, "r+", stdin) == (FILE *) NULL)  {
		fprintf(stderr, "%s: re-opening stdin to %s failed\n",
			OurName, Input);
		exit(EXIT_FAILURE);
	    }
	    if (Verbose == TRUE)
	      fprintf(stderr,
		      "%s: reopened stdin to %s with mode 'r+'\n",
		      OurName, Input);
	}

	if (Verbose == FALSE) {
	    pid_t parent = getpid();
	    cfht_signal(SIGUSR1, child_pass);
	    cfht_signal(SIGUSR2, child_fail);
	    switch (fork()) {
	      case (pid_t)-1: {
		perror("fork failed");
		exit(EXIT_FAILURE);
	      }
	      case 0: {
	 	  int fd;
		  if (setsid() < 0)
		      perror("warning: setsid failed");
		
		  cfht_signal(SIGUSR1, CFHT_SIG_DFL);
		  cfht_signal(SIGUSR2, CFHT_SIG_DFL);
		  /*
		   * Special case for roll: don't close stdin.
		   */
		  for (fd = sysconf(_SC_OPEN_MAX); fd >= 1; fd--)
		      if (fd == roll_id->rfd) {
			  /* if rolling and a lock exists - exit */
			  if (lockf(roll_id->rfd, F_TLOCK, 0L) < 0) {
			      /* Note: this error is not expected, since we check for
			       * duplicate roll processes by seeing if the pipe is
			       * being read (see open O_WRONLY above.)  However,
			       * old versions of roll use the file lock, so this
			       * lock MUST be maintained, and it has to happen after
			       * the fork() because child processes do not inherit
			       * any file locks.  (The O_WRONLY hack suffers from a
			       * race condition also.)
			       */
			      if ((errno == EAGAIN) || (errno == EACCES)) { /* already locked */
				  fprintf(stderr, "(roll_init) %s is already open to another roll\n",
					  RollName);
			      }
			      else perror(RollName); /* Some other error.  Still bail. */
			      close(roll_id->rfd);
			      kill(parent, SIGUSR2); /* Could not get lock */
			      _exit(EXIT_FAILURE);
			  }
		      }
		      else close(fd);
		  if (open("/dev/null", O_RDWR) == 1)
		      dup2(1, 2);
		  kill(parent, SIGUSR1); /* Daemon is off and running. */
		  break;
	      }
	      default: {
		  /*
		   * Parent must wait for child to ready before exiting because
		   * (1) The child could fail to get the lock (Parent cannot
		   *     manage the lock, since it will be release on fork()).
		   * (2) The subsequent commands are probably going to do
		   *     logging, so we want to wait for things to be ready.
		   *
		   * A SIGUSR1 or SIGUSR2 will arrive as soon as the child
		   * has initialized itself.
		   */
		  /* pause(); */
		  sleep(60); /* In case something goes wrong with child,
			      * time out after a minute and fail.
			      */
		  fprintf(stderr, "%s: Timeout waiting for daemon to start\n",
			  OurName);
		  exit(EXIT_FAILURE);
	      }
	    }
	}

	while ((c = getchar()) != EOF)
	  roll_putc(c, roll_id);
	roll_out(roll_id);
    } else { /* !IsRoll (unroll) */
	ReverseBuf* p;
	char *TimeString;
	char *s;

	if (roll_id->Direction == ROLL_BACK) {
	    if (GoBackwards == FALSE) ReverseOutput = TRUE;
	} else {
	    if (GoBackwards == TRUE) ReverseOutput = TRUE;
	}

	TimeString = ctime(&roll_id->From);
	if (TimeString[strlen(TimeString) - 1] == '\n')
	  TimeString[strlen(TimeString) - 1] = '\0';
	printf("Roll list from %s", TimeString);
	TimeString = ctime(&roll_id->To);
	if (TimeString[strlen(TimeString) - 1] == '\n')
	  TimeString[strlen(TimeString) - 1] = '\0';
	printf(" to %s\n", TimeString);
	
	while ((s = roll_gets(roll_id, CookedOutput)) != NULL) {
	    if (ReverseOutput == TRUE) {
		p = malloc(sizeof(ReverseBuf) + strlen(s) + 1);
		strcpy(p->line, s);
		p->next = ReversedLines;
		ReversedLines = p;
	    } else {
		fputs(s, stdout);
	    }
	}
	while ((p = ReversedLines) != NULL) {
	    fputs(p->line, stdout);
	    ReversedLines = p->next;
	    free(p);
	}
    }
    
    exit(EXIT_SUCCESS);
}

/* usage: prints default error message and then exits */
/* s is program name */
void
roll_usage(char* s)
{
    fprintf(stderr, "Usage: %s\n", s);
    fprintf(stderr, "          [-r|-u] [-b] [-s size]\n");
    fprintf(stderr, "          [-h hoursback] or [-f timefrom] [-t timeto]\n");
    fprintf(stderr, "          [-a name] [file]\n");
    exit(EXIT_FAILURE);
}

/*
 * copy of cfp_match without the cfht_log calls :-(
 */

int				  /* return: index, or FAIL */
Match(char *name,			  /* identifier to look for */
      char **addrs)		  /* list of strings to match against */
{
    int count;			  /* line number of string */
    int tcount;
    char *p;			  /* temporary string pointer */
    int size;			  /* size of rest of string */
    int namelen;

    count = FAIL;
    if (*name == '-')
        name++;

    /* for each line */
    namelen = strlen(name);
    for (tcount=0, p=addrs[tcount] ; p && *p ; tcount++, p=addrs[tcount]) {
        if (strncmp(p, name, (size_t) namelen))    /* test for name */
            continue;                              /* try next line */

        p += namelen;                              /* skip over name */
        size = strlen(p);                          /* how close a match ? */

        /* check for collisions and exact matches */
        if ((count == FAIL) || size == 0) {
            count = tcount;                        /* found something */
            if (size == 0)
                break;                             /* exact match */
        } else {
	    fprintf(stderr, "(Match) ambiguous abbreviation '%s'\n", name);
	    fprintf(stderr, "        matches '%s' and '%s'\n",
		    addrs[count], addrs[tcount]);
            return(FAIL);
        }
    }

    if (count == FAIL) {
	fprintf(stderr, "(Match) no match for '%s'\n", name);
        return(FAIL);
    }

    return(count);
}

/*
 * GetAlpha - get lower cased copy of a string input
 *
 * Call:
 *    GetAlpha(Scan, &Tail, Token, sizeof Token);
 * Input:
 *    char *Scan - pointer to string input
 *    size_t TokenSize - sizeof Token
 * Output:
 *    char **Tail - updated to point to first non-alpha character in input
 *    char *Token - filled in with lower cased copy of input
 *                  nul terminated
 */

void GetAlpha(char *Scan,
	      char **Tail,
	      char *Token,
	      size_t TokenSize)
{
    size_t Index;

    Index = 0;
    *Tail = Scan;
    while (isalpha(**Tail)) {
      if (Index < (TokenSize - 1)) {
	  if (isupper(**Tail))
	    Token[Index++] = tolower(**Tail);
	  else
	    Token[Index++] = **Tail;
	(*Tail)++;
      }
    }
    Token[Index] = '\0';
    return;
}

/*********************************************************************
 *
 *   TimeStringToTime - convert ASCII time specification into internal format
 *
 * DESCRIPTION
 *
 *    If the getdate routine is available, the input format is flexible, see
 *    the date file and getdate(3C).  Basically anything thought of is
 *    accepted.  If (i.e., sun) only strpdate is available, time or time and
 *    date are accepted.
 *
 *    The above would have been nice, but getdate is futurecentric.  It forces
 *    defaults to be future times (even worse than the documentation says - on
 *    Thursday at 1230, "thu 12:30" gets translated as next week Thursday :-(
 * 
 *    So instead we'll write our own parse routine that accepts as much stuff
 *    as possible.  It does not check for repetition so the results might not
 *    be as expected.  It does try to fill in only specified values, starting
 *    with the current time as the default.
 *
 * INVOKED
 *
 *   Status = TimeStringToTime(TimeString, &TimeValue);
 *
 * PARAMETERS IN
 *   char *TimeString - external time
 *
 * PARAMETERS OUT
 *   time_t *TimeValue - internal time
 *
 *
 * FUNCTION RETURNS
 *   PASSFAIL - PASS if no problems, FAIL if errors, TimeValue undefined
 *
 * GLOBALS
 *   none
 *
 *********************************************************************/

PASSFAIL TimeStringToTime(char *String, time_t *Value)
{
#define SecondsPerMinute 60	  /* should be in earth.h ? */
#define MinutesPerHour 60
#define HoursPerDay 24
#define DaysPerWeek 7
#define MidNightHour 0
#define SunRiseHour 6
#define NoonHour 12
#define SunSetHour 18
#define Base 10			  /* all numbers must be decimal */
#define MilitaryTimeSize 4	  /* for times given as 0010, e.g. */
#define MilitaryTimeShift 100	  /* divisor to get hours from 1200 */
#define FullYearSize 4		  /* for years given as 1995, e.g. */
#define YearZero 1900		  /* first year in struct tm's world */

    static char *MatchList[] = {
	"sunday",
#define SundayIndex 0		  /* must be 0 to match tm_wday definition */
#define FirstWeekDayIndex SundayIndex
	"monday",
#define MondayIndex (SundayIndex + 1)
	"tuesday",
#define TuesdayIndex (MondayIndex + 1)
	"wednesday",
#define WednesdayIndex (TuesdayIndex + 1)
	"thursday",
#define ThursdayIndex (WednesdayIndex + 1)
	"friday",
#define FridayIndex (ThursdayIndex + 1)
	"saturday",
#define SaturdayIndex (FridayIndex + 1)
#define LastWeekDayIndex SaturdayIndex
	"su",			  /* match for su == Sunday (see sunrise) */
#define SuIndex (SaturdayIndex + 1)
	"sun",			  /*  likewise for sun == Sunday */
#define SunIndex (SuIndex + 1)
	"now",
#define NowIndex (SunIndex + 1)
	"midnight",
#define MidNightIndex (NowIndex + 1)
	"noon",
#define NoonIndex (MidNightIndex + 1)
	"sunrise",		  /* 0600 */
#define SunRiseIndex (NoonIndex + 1)
	"sunset",		  /* 1800 */
#define SunSetIndex (SunRiseIndex + 1)
	"am",
#define AMIndex (SunSetIndex + 1)
	"pm",
#define PMIndex (AMIndex + 1)
	"today",		  /* == this day but don't change time */
#define TodayIndex (PMIndex + 1)
	"yesterday",		  /* == yesterday but don't change time */
#define YesterdayIndex (TodayIndex + 1)
	"january",
#define JanuaryIndex (YesterdayIndex + 1) /* must be 1st to match tm_mon */
#define FirstMonthIndex JanuaryIndex
	"february",
#define FebruaryIndex (JanuaryIndex + 1)
	"march",
#define MarchIndex (FebruaryIndex + 1)
	"april",
#define AprilIndex (MarchIndex + 1)
	"may",
#define MayIndex (AprilIndex + 1)
	"june",
#define JuneIndex (MayIndex + 1)
	"july",
#define JulyIndex (JuneIndex + 1)
	"august",
#define AugustIndex (JulyIndex + 1)
	"september",
#define SeptemberIndex (AugustIndex + 1)
	"october",
#define OctoberIndex (SeptemberIndex + 1)
	"november",
#define NovemberIndex (OctoberIndex + 1)
	"december",
#define DecemberIndex (NovemberIndex + 1)
#define LastMonthIndex DecemberIndex
	/* NOTE - months must be last in list, see "MonthList" below */
	(char *) NULL
    };
#define TokenSize 10		  /* length of longest string above + 1 */
#define MonthList &MatchList[FirstMonthIndex]

    char *Input;		  /* start of String after blank skip */
    char *Copy;			  /* our copy of String to munge */
    char *Scan;			  /* pointer into Copy */
    char *Tail;			  /* other end of number when one seen */
    char Token[TokenSize + 1];	  /* copy of word from string */
				  /*   lower case + nul */
    unsigned int Number;	  /* copy of number from string */
    struct tm tmTimeNow;	  /* tm version of current time */
    struct tm tmScratch;	  /* scratch time in tm form */
    struct tm *tmPointer;	  /* pointer to tmScratch, usually */
    struct tm tmValue;		  /* tm version of scanned string so far */
    time_t Scratch;		  /* scratch time in time_t form */
    time_t Midnight;		  /* 0000 today */
    time_t OneAM;		  /* 0100 today */
    time_t OneHour;		  /* 1 hour delta in time_t form */
    time_t OneDay;		  /* 1 day delta in time_t form */
    BOOLEAN KeepGoing;		  /* scan loop control */
    int MatchIndex;		  /* index to matched string or -1 if failed */

    if ((String == (char *) NULL) || (Value == (time_t *) NULL)) {
	fprintf(stderr, "(TimeStringToTime) null argument\n");
	return(FAIL);
    }
    Input = String + strspn(String, " \t");
    if (*Input == '\0') {	  /* ignore empty input */
	*Value = TimeNow;
	return(PASS);
    }
    if ((Copy = (char *) malloc(strlen(Input) + 1)) == (char *) NULL) {
	fprintf(stderr, "(TimeStringToTime) can't malloc copy\n");
	return(FAIL);
    }
    (void) strcpy(Copy, Input);
    Scan = Copy;
    while (*Scan != '\0') {
	if (*Scan == '\t')
	  *Scan = ' ';		  /* convert tabs to spaces */
	Scan++;
    }
    while (*--Scan == ' ') {
	;
    }
    *++Scan = '\0';		  /* trim trailing blanks */

    /* get various time values needed for conversions */
    tmPointer = localtime(&TimeNow);
    (void) memcpy(&tmTimeNow, tmPointer, sizeof tmTimeNow);

    (void) memcpy((void *) &tmScratch, (void *) &tmTimeNow,
		  sizeof tmScratch);
    /* get time of midnight */
    tmScratch.tm_sec = 0;
    tmScratch.tm_min = 0;
    tmScratch.tm_hour = 0;
    Midnight = mktime(&tmScratch);
    /*  and 0100 hours */
    tmScratch.tm_hour = 1;
    OneAM = mktime(&tmScratch);
    /*  get 1 hour in time_t form (should be 3600) */
    OneHour = OneAM - Midnight;
    /*  get 1 day in time_t form */
    OneDay = OneHour * HoursPerDay;

    /* start with now, for each piece of input update the part(s) it covers */

    (void) memcpy((void *) &tmValue, (void *) &tmTimeNow, sizeof tmValue);
    KeepGoing = TRUE;
    Scan = Copy;
    while ((KeepGoing == TRUE) && (*Scan != '\0')) {

	/* check character type to determine what to do next */
	if (isdigit(*Scan)) {

	    /* get the number then figure out what kind by what follows */
	    Number = strtoul(Scan, &Tail, Base);
	    switch (*Tail) {

		/* if end of field, assume time, hours unless 4 digits */
	      case '\0':
	      case ' ':
		if ((Tail - Scan) == MilitaryTimeSize) {
		    tmValue.tm_hour = Number / MilitaryTimeShift;
		    tmValue.tm_min = Number - (tmValue.tm_hour *
					       MilitaryTimeShift);
		    tmValue.tm_sec = 0;
		    if (tmValue.tm_hour >= HoursPerDay) {
			fprintf(stderr, "hours in %d too large\n", Number);
			KeepGoing = FALSE;
		    }
		    if (tmValue.tm_min >= MinutesPerHour) {
			fprintf(stderr, "minutes in %d too large\n", Number);
			KeepGoing = FALSE;
		    }
		} else {
		    if (Number < HoursPerDay) {
			tmValue.tm_hour = Number;
			tmValue.tm_min = 0;
			tmValue.tm_sec = 0;
		    } else {
			fprintf(stderr, "hours in %d too large\n", Number);
			KeepGoing = FALSE;
		    }
		}
		Scan = Tail;
		break;

		/* if followed by : must be time and Number must be hours */
	      case ':':
		tmValue.tm_hour = Number;
		tmValue.tm_min = 0;
		tmValue.tm_sec = 0;
		Scan = ++Tail;
		if (isdigit(*Scan)) {
		    tmValue.tm_min = strtoul(Scan, &Tail, Base);
		    Scan = Tail;
		    if (*Scan == ':') {
			if (isdigit(*++Scan)) {
			    tmValue.tm_sec = strtoul(Scan, &Tail, Base);
			    Scan = Tail;
			    if (*Scan == '.') {
				if (isdigit(*++Scan)) {
				    Number = strtoul(Scan, &Tail, Base);
				    Scan = Tail;
				}
			    }
			}
		    }
		}
		if (tmValue.tm_hour >= HoursPerDay) {
		    fprintf(stderr, "hours %d too large in '%s'\n",
			    tmValue.tm_hour, Copy);
		    KeepGoing = FALSE;
		}
		if (tmValue.tm_min >= MinutesPerHour) {
		    fprintf(stderr, "minutes %d too large in '%s'\n",
			    tmValue.tm_min, Copy);
		    KeepGoing = FALSE;
		}
		if (tmValue.tm_sec >= SecondsPerMinute) {
		    fprintf(stderr, "seconds %d too large in '%s'\n",
			    tmValue.tm_sec, Copy);
		    KeepGoing = FALSE;
		}
		break;

		/* if - or / must be date, assume day of month */
	      case '-':
	      case '/':
		tmValue.tm_mday = Number;
		tmValue.tm_mon = tmTimeNow.tm_mon;
		tmValue.tm_year = tmTimeNow.tm_year;
		Scan = ++Tail;
		if (isdigit(*Scan)) {
		    /* if all numeric, it's mm-dd-yy */
		    tmValue.tm_mon = Number;
		    tmValue.tm_mday = strtoul(Scan, &Tail, Base);
		    if (tmValue.tm_mon > 0)
		      tmValue.tm_mon--;
		    Scan = Tail;
		} else if (isalpha(*Scan)) {
		    /* if alpha month, it's dd/mmm/yy */
		    GetAlpha(Scan, &Tail, Token, sizeof Token);
		    Scan = Tail;
		    MatchIndex = Match(Token, MonthList);
		    if (MatchIndex >= 0)
		      tmValue.tm_mon = MatchIndex;
		    else {
			fprintf(stderr, "Bad month '%s' in '%s'\n",
				Token, Copy);
			KeepGoing = FALSE;
		    }
		} else {
		    fprintf(stderr, "Bad month field at '%s' in '%s'\n",
			    Scan, Copy);
		    KeepGoing = FALSE;
		}
		if ((*Scan == '-') || (*Scan == '/')) {
		    if (isdigit(*++Scan)) {
			Number = strtoul(Scan, &Tail, Base);
			if ((Tail - Scan) == FullYearSize) {
			    if (Number >= YearZero)
			      tmValue.tm_year = Number - YearZero;
			    else {
				fprintf(stderr, "Year %d too small in '%s'\n",
					Number, Copy);
				KeepGoing = FALSE;
			    }
			    
			} else {
			    tmValue.tm_year = Number;
			}
			Scan = Tail;
		    } else {
			fprintf(stderr, "Expecting year at '%s' in '%s'\n",
				Scan, Copy);
			KeepGoing = FALSE;
		    }
		}
		break;

		/* otherwise assume hours followed by AM/PM caught later */
	      default:
		if (Number < HoursPerDay) {
		    tmValue.tm_hour = Number;
		    tmValue.tm_min = 0;
		    tmValue.tm_sec = 0;
		} else {
		    fprintf(stderr, "hours in %d too large\n", Number);
		    KeepGoing = FALSE;
		}
		Scan = Tail;
		break;
	    }

	} else if (isalpha(*Scan)) {

	    /* some kind of word, see if we know what it is */
	    GetAlpha(Scan, &Tail, Token, sizeof Token);
	    Scan = Tail;
	    MatchIndex = Match(Token, MatchList);
	    switch (MatchIndex) {

		/* special case "Su" and "Sun" so can have sunrise/sunset */
	      case SuIndex:
	      case SunIndex:
		MatchIndex = SundayIndex;
		/* NOTE - no break, fall into regular weekday code */

		/* weekdays - set day-month-year */
	      case SundayIndex:
	      case MondayIndex:
	      case TuesdayIndex:
	      case WednesdayIndex:
	      case ThursdayIndex:
	      case FridayIndex:
	      case SaturdayIndex:
		/* do subtraction in time_t mode so don't have to worry */
		/*   about wrapping months or years */
		if (MatchIndex > tmTimeNow.tm_wday)
		  /* a day later in week has to back up past Sunday */
		  Scratch = TimeNow
		    - ((DaysPerWeek - (MatchIndex - tmTimeNow.tm_wday))
		       * OneDay);
		else
		  Scratch = TimeNow
		    - ((tmTimeNow.tm_wday - MatchIndex) * OneDay);
		tmPointer = localtime(&Scratch);
		tmValue.tm_mday = tmPointer->tm_mday;
		tmValue.tm_mon = tmPointer->tm_mon;
		tmValue.tm_year = tmPointer->tm_year;
		break;

		/* special time names just set time */
	      case NowIndex:
		tmValue.tm_hour = tmTimeNow.tm_hour;
		tmValue.tm_min = tmTimeNow.tm_min;
		tmValue.tm_sec = tmTimeNow.tm_sec;
		break;
	      case MidNightIndex:
		tmValue.tm_hour = MidNightHour;
		tmValue.tm_min = 0;
		tmValue.tm_sec = 0;
		break;
	      case NoonIndex:
		tmValue.tm_hour = NoonHour;
		tmValue.tm_min = 0;
		tmValue.tm_sec = 0;
		break;
	      case SunRiseIndex:
		tmValue.tm_hour = SunRiseHour;
		tmValue.tm_min = 0;
		tmValue.tm_sec = 0;
		break;
	      case SunSetIndex:
		tmValue.tm_hour = SunSetHour;
		tmValue.tm_min = 0;
		tmValue.tm_sec = 0;
		break;

		/* special days change month-day-year */
	      case TodayIndex:
		tmValue.tm_mday = tmTimeNow.tm_mday;
		tmValue.tm_mon = tmTimeNow.tm_mon;
		tmValue.tm_year = tmTimeNow.tm_year;
		break;
	      case YesterdayIndex:
		Scratch = TimeNow - OneDay;
		tmPointer = localtime(&Scratch);
		tmValue.tm_mday = tmPointer->tm_mday;
		tmValue.tm_mon = tmPointer->tm_mon;
		tmValue.tm_year = tmPointer->tm_year;
		break;

		/* ignore AM unless it's 12 something AM */
	      case AMIndex:
		if (tmValue.tm_hour == NoonHour)
		  tmValue.tm_hour = 0;
		break;
		/* for PM, add 12 unless already in afternoon */
	      case PMIndex:
		if (tmValue.tm_hour < NoonHour)
		  tmValue.tm_hour += NoonHour;
		break;

		/* do something with month names */
	      case JanuaryIndex:
	      case FebruaryIndex:
	      case MarchIndex:
	      case AprilIndex:
	      case MayIndex:
	      case JuneIndex:
	      case JulyIndex:
	      case AugustIndex:
	      case SeptemberIndex:
	      case OctoberIndex:
	      case NovemberIndex:
	      case DecemberIndex:
		tmValue.tm_mon = MatchIndex - FirstMonthIndex;
		break;		  /* want to do "mar 1, 1995" ???? */
		
	      case -1:
		fprintf(stderr, "Unknown token '%s' in '%s'\n", Token, Copy);
		KeepGoing = FALSE;
		break;

	      default:
		fprintf(stderr, "Program error, missed index for '%s'\n",
			Token);
		KeepGoing = FALSE;
	    }
	} else {
	    fprintf(stderr, "Unexpected '%c' at '%s' in '%s'\n", *Scan, Scan,
		    Copy);
	    KeepGoing = FALSE;
	}

	/* skip spaces between fields */
	while (*Scan == ' ') {
	    Scan++;
	}
    }

#ifdef DEBUGDEBUG
    printf("for '%s' got %d:%d:%d %d-%d-%d\n", Copy, tmValue.tm_hour,
	   tmValue.tm_min, tmValue.tm_sec, tmValue.tm_mday, tmValue.tm_mon+1,
	   tmValue.tm_year);
#endif

    (void) free(Copy);
    if (KeepGoing == FALSE)
      return(FAIL);
    
    errno = 0;
    *Value = mktime(&tmValue);
    if (errno == ERANGE) {
	/* note - no newline in following, asctime makes one */
	fprintf(stderr, "(TimeStringToTime) from '%s' error converting %s",
		Copy, asctime(&tmValue));
	return(FAIL);
    }
    return(PASS);
}
