/* Copyright (C) 1991, 1995    Canada-France-Hawaii Telescope Corp.      */
/* This program is distributed WITHOUT any warranty, and is under the    */
/* terms of the GNU General Public License, see the file COPYING    */
/*!****************************************************************************
 * FILE roll.h
 * $Header: /cfht/src/pegasus/roll/RCS/roll.h,v 2.2 2003/03/09 20:44:04 isani Exp $
 * $Locker:  $
 *
 *    header file for roll/unroll routines
 *
 * HISTORY
 * who          when            what
 * ---------    ------------    ----------------------------------------------
 * sss          880831          original
 * sss          880902          changed to forward/backwards one mark
 *
 * $Log: roll.h,v $
 * Revision 2.2  2003/03/09 20:44:04  isani
 * Updated prototypes for cooked and reversed options.
 *
 * Revision 2.1  2002/05/05  21:30:53  isani
 * Removed DummyRollFile and used offsetof() instead.
 * 
 * Revision 2.0  1995/11/22 09:55:37  thomas
 * Added roll file structure, redid roll memory structure
 * Changed defines to match
 *
 * Revision 1.3  95/06/20  09:04:44  09:04:44  viloria (Ron Viloria)
 * Added time option [-h hours], where "hours" is the number of hours to list 
 * output lines from now backwards from the file. 
 * 
 * Revision 1.2  95/02/14  14:44:09  14:44:09  thomas (Jim Thomas)
 * Changed default roll file size to 1M
 * 
 * Revision 1.1  91/09/30  12:13:51  12:13:51  jwright (Jim Wright)
 * Original checkin
 * 
 ***************************************************************************!*/

#ifndef ROLLDOTH
#define ROLLDOTH
#include <stddef.h>		  /* for offsetof() macro */
#include <time.h>
#include "cfht/cfht.h"

#define ROLL_NAME          "syslog.@"

/* line lengths longer than ROLLBUFSIZ will be messed up when viewed */
#define ROLLBUFSIZ         1024	  /* size of roll buffer */
#define LINEBUFSIZ         (ROLLBUFSIZ + 1 + 1) /* size of line buffer */
				  /* +1 for possible added newline */
				  /* +1 for nul */

#define ROLL_DEF_SIZE      (sizeof(RollFileType) + (1*1024*1024))
#define ROLL_MIN_SIZE      (sizeof(RollFileType) + (ROLLBUFSIZ * 2))

#define ROLL_FORWARD       0xaa	  /* Direction byte for forward roll */
#define ROLL_BACK          0x55	  /* Direction byte for backward roll */
#define ROLL_EOD           0xff	  /* end of data byte */
#define ROLL_CREATE_MODE   0666

#define ROLL_FIRST_BYTE(a) (offsetof(RollFileType, Data))
#define ROLL_LAST_BYTE(a)  (a->Size - 1)
#define ROLL_IS_BACK(a)    ((a)->Direction == ROLL_BACK)

/*
 * The -h option specifies how far back to start outputting lines.  The normal
 * use is to look at recent messages in the log; the following symbol defines
 * the default for "recent", in hours.
 */
#define DEFTIME 6

/*
 * On disk the roll file consists of a current file pointer, a direction flag
 * and lines of log information.  The direction flag indicates whether lines
 * in the file are saved in FIFO (forward) or LIFO (backwards) order.  The
 * normal order is LIFO.
 *
 * The file itself is a circular buffer.  If the order is LIFO, new lines are
 * added in front of the file pointer (which is then moved backwards).  If the
 * order is FIFO, new lines are written starting at the file pointer (which is
 * then moved forwards).  In either case, lines are read starting at the
 * current file pointer going forwards.  When the edge of the circular buffer
 * is reached, activity continues from the other edge.  There is an end of data
 * byte written into the file at the current file pointer position and
 * overwritten on the next write.
 *
 * The Offset byte is updated every time a line is added to the file.  Since
 * an unroll can happen at any time, for a forward file this must happen
 * before the data write, and for a backward file it must be after the data
 * write.
 *
 * +----------------------------------------------------------------------+
 * | Offset into file to current EOF point (including this header)        |
 * |   This should always point to the EOD byte                           |
 * +----------------------------------------------------------------------+
 * | Direction indicator (forward or backward)                            |
 * +----------------------------------------------------------------------+
 * | First data byte (rarely the first logical byte)                      |
 * | .......                                                              |
 * +----------------------------------------------------------------------+
 */

typedef struct {
    off_t Offset;
    unsigned char Direction;
    unsigned char Data;
} RollFileType;

/* define the relative position in the file of the Offset field */
#define OffsetOffset 0L

/*
 * The roll structure is the in memory representation of the roll file.
 *
 * +----------------------------------------------------------------------+
 * | File identifier of open roll file                                    |
 * +----------------------------------------------------------------------+
 * | In memory copy of offset into file of current EOF point              |
 * +----------------------------------------------------------------------+
 * | Copy of direction indicator                                          |
 * +----------------------------------------------------------------------+
 * | Buffer for a piece of the file                                       |
 * |   with room for the EOD byte on output                               |
 * |   NOTE - not nul terminated                                          |
 * +----------------------------------------------------------------------+
 * | Buffer for a line extracted from the file buffer                     |
 * |   with room for a newline in case a line is too long                 |
 * |   NOTE - not nul terminated                                          |
 * +----------------------------------------------------------------------+
 * | Count of unprocessed characters in file buffer                       |
 * +----------------------------------------------------------------------+
 * | Index to next byte in file buffer                                    |
 * +----------------------------------------------------------------------+
 * | Count of unprocessed characters in line buffer                       |
 * +----------------------------------------------------------------------+
 * | Index to next byte in line buffer                                    |
 * +----------------------------------------------------------------------+
 * | File size including header bytes                                     |
 * |   used as wrap value when writing                                    |
 * |   used as flag when reading (zeroed on first wrap)                   |
 * +----------------------------------------------------------------------+
 * | Verbosity level                                                      |
 * |   FALSE - be quiet, errors only to stderr                            |
 * |   TRUE  - debug, tell what's happening                               |
 * +----------------------------------------------------------------------+
 * | Beginning time of interest (default DEFTIME hours ago)               |
 * |   only lines between beginning and ending times are output           |
 * +----------------------------------------------------------------------+
 * | Ending time of interest (default now)                                |
 * +----------------------------------------------------------------------+
 * | Flag indicating whether the first time of interest has been seen yet |
 * |   FALSE - not seen yet                                               |
 * |   TRUE  - seen already (set initially if ending time is now)         |
 * +----------------------------------------------------------------------+
 * | Session name of interest                                             |
 * |   if not set (NULL), output all lines during time of interest        |
 * |   if set to a session name, output only lines from that session      |
 * |      during time of interest (session name is first token on line)   |
 * +----------------------------------------------------------------------+
 */

typedef struct {
    int rfd;
    off_t Offset;
    unsigned char Direction;
    unsigned char RollBuffer[ROLLBUFSIZ + 1];
    unsigned char LineBuffer[LINEBUFSIZ + 1];
    unsigned char* LineStart; /* Start of unparsed portion of message */
                              /* or NULL if no timestamp found. */
    struct tm LineTime; /* Full time information for whats in LineBuffer */
    int LineTimeHund; /* Hundreths of seconds to go with LineTime */
    BOOLEAN LineTimeValid; /* TRUE once LineTime has a value */
    unsigned int RollCount;
    unsigned int RollIndex;
    unsigned int LineCount;
    off_t Size;
    BOOLEAN Verbose;
    time_t From;
    time_t To;
    char *Session;
} RollType;

/* function prototypes */

RollType *roll_init(char *filename, off_t size, BOOLEAN backwards,
		    BOOLEAN IsRoll, BOOLEAN Verbose);
char* roll_gets(RollType *id, BOOLEAN cooked);
PASSFAIL roll_out(RollType *id);
PASSFAIL roll_putc(int c, RollType *id);

#endif /* ROLLDOTH */
