/* Copyright (C) 1991-1999    Canada-France-Hawaii Telescope Corp.    */
/* This program is distributed WITHOUT any warranty, and is under the */
/* terms of the GNU General Public License, see the file COPYING      */
/*****************************************************************************
 * FILE
 *
 * $Id: support.c,v 2.9 2003/03/09 20:44:34 isani Locked $
 * $Locker: isani $
 *
 * ROUTINES
 * 
 *   support routines for CFHT roll/unroll message logging system
 *
 * HISTORY
 * who          when            what
 * ---------    ------------    ----------------------------------------------
 * ???          ?? ??? ????     original
 * jrw          06 Feb 1991     sun port, general cleanup
 * jrw          25 Feb 1991     wrote own prealloc for suns, use it
 *
 * $Log:	support.c,v $
 * Revision 2.9  2003/03/09  20:44:34  isani
 * New support for cooked output on unroll.
 * Changed the way start/end times are calculated (more precise now.)
 * 
 * Revision 2.8  2002/05/06  05:42:10  isani
 * Support TZ (timezone) longer than three letters (just used CFHT_DATE_SIZE).
 * Fix bug in daylight savings areas which keeps you from seeing messages
 *   during a one hour waiting period.
 * 
 * Revision 2.7  2002/05/05 22:46:16  isani
 * Added message during creation of a new syslog.@ file, since this can take
 *   a long time (on systems without prealloc?)
 *
 * Revision 2.6  2002/05/05 21:36:34  isani
 * ANSI-fied.
 * Converted signal handling to use cfht_signal(), and don't try to catch
 *   *every* single signal from 1..NSIG-1 anymore!
 *
 * Revision 2.5  1999/07/25 09:22:27  thomas
 * Do not open roll file read/write for unroll !
 *
 * Revision 2.4  1998/02/19 11:10:10  thomas
 * Protect sscanf for start time against garbage
 *
 * Revision 2.3  96/11/19  03:00:09  03:00:09  isani (Sidik Isani)
 * cleaned up rcsid's
 * fixed exit with no arg
 * 
 * Revision 2.2  1996/03/11  08:10:48  thomas
 * A last bit of reformatting
 *
 * Revision 2.1  95/12/26  15:59:48  15:59:48  thomas (Jim Thomas)
 * Correct fencepost error on month name -> number conversion :-(
 * 
 * Revision 2.0  95/11/22  09:56:57  09:56:57  thomas (Jim Thomas)
 * Rewrote roll file output and input routines to maintain offset
 * Added support for -f and -t (based on work by Ron Viloria)
 * 
 * Revision 1.2  95/06/20  09:37:28  09:37:28  viloria (Ron Viloria)
 * Added time option [-h hours], where "hours" is the number of hours to list 
 * output lines from now backwards from the file. 
 * 
 * Revision 1.1  91/09/30  12:13:53  12:13:53  jwright (Jim Wright)
 * Original checkin
 * 
 ****************************************************************************/

#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>		  /* for time functions */
#if defined(sun) && !defined(__svr4__)
#define mktime timelocal	  /* different on sun */
#endif
#include <string.h>		  /* for string functions */

#include "cfht/cfht.h"
#include "cfht/cfht_log.h"

RCSID("$Id: support.c,v 2.9 2003/03/09 20:44:34 isani Locked $");

#include "roll.h"


/* forward function definitions */

static PASSFAIL roll_create(char *filename, off_t size, BOOLEAN backwards);
static PASSFAIL roll_flush(RollType *id);
static int line_fill(RollType *id);
static char* cook_line(RollType *id, const char* s);
static int roll_fill(RollType *id);
static void cleanup(int sig);
static int month_num(char* s);

/* globals */

RollType *RollFilePointer;	  /* copy of pointer to open roll file */
int linecounter = 0;		  /* debug */


RollType *roll_init(char* RollName, off_t Size,
		    BOOLEAN GoBackwards, BOOLEAN IsRoll, BOOLEAN Verbose)
{
    RollFileType RollBuffer;	  /* first part of roll file */
    struct stat stbuf;
    RollType *id;
    int i, j;
    off_t offset;
    unsigned char tmpc;

    if (stat(RollName, &stbuf) < 0) { /* file not there */
	fprintf(stderr, "(roll_init) %s not stat'ed\n", RollName);
	if ((IsRoll == TRUE)
	    && (roll_create(RollName, Size, GoBackwards) != PASS)) {
	    fprintf(stderr, "(roll_init) %s not created\n", RollName);
	    return((RollType *) NULL);
	}
	if (stat(RollName, &stbuf) < 0) { /* file not there */
	    fprintf(stderr, "(roll_init) stat failed - %s not created\n",
		    RollName);
	    return((RollType *) NULL);
	}
    }

    if ((id = (RollType *) malloc(sizeof(RollType))) == (RollType *) NULL) {
	fprintf(stderr, "(roll_init) malloc failed\n");
	return((RollType *) NULL);
    }
    RollFilePointer = id;

    if ((id->rfd = open(RollName,
			(IsRoll == TRUE) ? O_RDWR : O_RDONLY))
	< 0) {
	perror("(roll_init) open failed");
	fprintf(stderr, "            on file %s\n", RollName);
	free(id);
	return((RollType *) NULL);
    }

    if (read(id->rfd, &RollBuffer, sizeof RollBuffer) != sizeof RollBuffer) {
	fprintf(stderr, "(roll_init) initial read failed on %s\n", RollName);
	close(id->rfd);
	free(id);
	return((RollType *) NULL);
    }

    id->Offset = RollBuffer.Offset;
    id->Direction = RollBuffer.Direction;
    id->Size = stbuf.st_size;
    if (IsRoll == TRUE) {
	id->RollCount = 1;	  /* counting EOD */
	if (GoBackwards == TRUE)
	  id->RollIndex = 1;	  /* EOD goes in front */
	else
	  id->RollIndex = 0;	  /* EOD goes in back */
    } else {
	id->RollCount = 0;	  /* Nothing in buffers */
	id->RollIndex = 0;
	id->LineCount = 0;
    }
    id->Verbose = Verbose;
    id->From = 0;
    id->To = 0;
    id->LineTimeValid = FALSE;
    id->Session = (char *) NULL;

    /* old roll files always have Offset == 0 so fix that first */

    if (id->Offset == 0) {
	if (Verbose == TRUE)
	  fprintf(stderr, "(roll_init) 0 offset, looking for EOD\n");

	offset = ROLL_FIRST_BYTE(id);
	if (lseek(id->rfd, offset, SEEK_SET) == -1) {
	    perror("(roll_init) seek to start scan failed");
	    close(id->rfd);
	    free(id);
	    return((RollType *) NULL);
	}

	while ((i = read(id->rfd, id->RollBuffer, sizeof id->RollBuffer)) > 0) {
	    for (j = 0; j < i; j++) {
		if (id->RollBuffer[j] == ROLL_EOD) {
		    if (Verbose == TRUE)
		      fprintf(stderr, "(roll_init) EOD at %d\n", (int)offset + j);
		    if (id->Offset != 0L) { /* multiple marks */
			fprintf(stderr,
				"(roll_init) multiple EOD's, not valid roll\n");
			close(id->rfd);
			free(id);
			return((RollType *) NULL);
		    }
		    id->Offset = offset + j;
		}
	    }
	    offset += i;
	}

	if (id->Offset == 0L) {
	    fprintf(stderr, "(roll_init) no mark, not a valid roll file\n");
	    close(id->rfd);
	    free(id);
	    return((RollType *) NULL);
	}
    }

    /* if we're roll, make sure we find where the EOD marker is */
    /*  may have to look a bit if crash between Offset update and data write */

    if (IsRoll == TRUE) {
        cfht_signal(SIGHUP, CFHT_SIG_IGN);
	cfht_signal(SIGTERM, CFHT_SIG_IGN);
	cfht_signal(SIGSEGV, (PFV)cleanup);
	cfht_signal(SIGBUS, (PFV)cleanup);
	cfht_signal(SIGINT, (PFV)cleanup);
	cfht_signal(SIGTSTP, (PFV)cleanup);
	cfht_signal(SIGQUIT, (PFV)cleanup);

	offset = id->Offset;
	if (lseek(id->rfd, offset, SEEK_SET) == -1) {
	    perror("(roll_init) seek to EOD failed");
	    close(id->rfd);
	    free(id);
	    return((RollType *) NULL);
	}

	if (read(id->rfd, &tmpc, 1) != 1) {
	    perror("(roll_init) roll read of EOD failed");
	    close(id->rfd);
	    free(id);
	    return((RollType *) NULL);
	}

	if (tmpc != ROLL_EOD) {
	    fprintf(stderr, "(roll_init) EOD not at offset - looking\n");

	    /* if roll is forward, offset got moved forward but line */
	    /*   didn't get written - so EOD is behind offset */
	    /* if roll is backward, line got written (before offset) */
	    /*   but offset didn't get updated - so EOD is behind offset */

	    while ((tmpc != ROLL_EOD) && (--offset != id->Offset)) {
		if (ROLL_IS_BACK(id)
		    && (offset == ROLL_FIRST_BYTE(id) - 1)) {
		    offset = id->Size - 1;
		} else if ((!ROLL_IS_BACK(id)) && (offset == id->Size)) {
		    offset = ROLL_FIRST_BYTE(id);
		}
		if (lseek(id->rfd, offset, SEEK_SET) == -1) {
		    perror("(roll_init) EOD search wrap to front failed");
		    close(id->rfd);
		    free(id);
		    return((RollType *) NULL);
		}
		if (read(id->rfd, &tmpc, 1) != 1) {
		    perror("(roll_init) search read for EOD failed");
		    close(id->rfd);
		    free(id);
		    return((RollType *) NULL);
		}
	    }
	    if (tmpc == ROLL_EOD) {
		fprintf(stderr, "(roll_init) EOD found at %d\n", (int)offset);
	    } else {
		fprintf(stderr, "(roll_init) no EOD found in file\n");
		close(id->rfd);
		free(id);
		return((RollType *) NULL);
	    }
	    id->Offset = offset;
	}
    } else {

	/* if unroll, position after the EOD */
	/* the file is just read sequentially from there */

	if (lseek(id->rfd, id->Offset, SEEK_SET) == -1) {
	    perror("(roll_init) post EOD check seek failed");
	    close(id->rfd);
	    free(id);
	    return((RollType *) NULL);
	}

	if (read(id->rfd, &tmpc, 1) != 1) {
	    perror("(roll_init) unroll read of EOD failed");
	    close(id->rfd);
	    free(id);
	    return((RollType *) NULL);
	}
    }

    if (Verbose == TRUE)
      fprintf(stderr, "(roll_init) end=%d, size=%d, type=%s\n",
	      (int)id->Offset, (int)id->Size,
	      ROLL_IS_BACK(id) ? "LIFO" : "FIFO");

    return(id);
}


static PASSFAIL
roll_create(char* RollName, off_t Size, BOOLEAN GoBackwards)
{
    int fd;
    RollFileType NewFile;

    fprintf(stderr, "(roll_create) %s being created; please wait.\n", RollName);

    NewFile.Offset = ROLL_FIRST_BYTE(dummy);
    NewFile.Direction = (GoBackwards == TRUE) ? ROLL_BACK : ROLL_FORWARD;
    NewFile.Data = ROLL_EOD;

    if ((fd = open(RollName, O_WRONLY | O_CREAT, ROLL_CREATE_MODE)) < 0) {
	perror("(roll_create) open failed");
	fprintf(stderr, "              on roll file '%s'\n", RollName);
	return(FAIL);
    }

    if (Size == 0L) {		  /* if no size given - default */
	fprintf(stderr, "(roll_create) default size is %d\n", (int)ROLL_DEF_SIZE);
	Size = ROLL_DEF_SIZE;
    }

    if (Size < ROLL_MIN_SIZE) {	  /* if given too small - smallest */
	fprintf(stderr,
		"(roll_create) size increased from %d to %d (minimum)\n",
		(int)Size, (int)ROLL_MIN_SIZE);
	Size = ROLL_MIN_SIZE;
    }

    if (prealloc(fd, Size) == -1) {
	perror("(roll_create) prealloc failed");
	close(fd);
	return(FAIL);
    }

    if (write(fd, (void *) &NewFile, sizeof NewFile) != sizeof NewFile) {
	perror("(roll_create) initial write failed");
	fprintf(stderr, "              to roll file '%s'\n", RollName);
	close(fd);
	return(FAIL);
    }

    close(fd);

    fprintf(stderr, "(roll_create) %s created\n", RollName);

    return(PASS);
}


int roll_putc(c, id)
    int c;
    RollType *id;
{
    if ((c != ROLL_EOD) && (c != '\0')) {
	id->RollBuffer[id->RollIndex++] = c;
	id->RollCount++;
    }
    else
        fprintf(stderr,
                "(roll_putc) character 0x%X not allowed in roll file\n", c);

    if ((id->RollCount > ROLLBUFSIZ) && (c != '\n') && ROLL_IS_BACK(id)) {
        fprintf(stderr, "(roll_putc) line too long - scrambled\n");
    }

    if ((id->RollCount > ROLLBUFSIZ) || (c == '\n')) /* flush time */
        if (roll_flush(id) != PASS) {
            fprintf(stderr, "(roll_putc) flush failed\n");
            return(EOF);
        }

    return(c);
}

/*
 * roll_flush - dump a line from the roll buffer to the file
 *
 * Input:
 * id->RollCount contains the number of bytes to be written INCLUDING the EOD
 * id->Offset points to the last EOD written
 * id->RollBuffer contains the line
 *
 * Output:
 * id->RollCount is set to 1 (the EOD)
 * id->RollIndex is set to 0 or 1 (0 if forward - EOD at end,
 *                                 1 if backward - EOD at beginning)
 * id->Offset is updated to point at EOD just written
 *
 * file is updated with record and new offset value
 * new offset is written before line if forwards (so we lose the oldest line
 * if crash between that write and line write but file is intact)
 * new offset is written after line if backwards (so we lose the newest line)
 * Thus if offset doesn't point to EOD at roll_init time, we have to go
 * backwards if forward file and vice versa.
 */

PASSFAIL
roll_flush(RollType *id)
{
    off_t offset1, offset2;	  /* offsets into file for wrapped write */
    unsigned int outlen1, outlen2; /* lengths of each part of wrap */
				  /* outlen2 == 0 -> no wrap, only 1 write */
    unsigned int index1, index2;  /* index into line for each part */
    size_t i;

    if (id->RollCount <= 1)
      return(PASS);		  /* no need if no data to write */

    /* backwards or forwards? */
    if (ROLL_IS_BACK(id)) {	  /* backwards */
	id->RollBuffer[0] = ROLL_EOD;
	if ((ROLL_FIRST_BYTE(id) + id->RollCount) > id->Offset) {
	    outlen1 = id->Offset - ROLL_FIRST_BYTE(id) + 1;
	    offset1 = ROLL_FIRST_BYTE(id);
	    outlen2 = id->RollCount - outlen1;
	    offset2 = ROLL_LAST_BYTE(id) - outlen2 + 1;
	} else {
	    outlen1 = id->RollCount;
	    offset1 = id->Offset - outlen1 + 1;
	    outlen2 = 0;
	    offset2 = offset1;
	}
	index1 = outlen2;
	index2 = 0;
    } else {			  /* forwards */
	id->RollBuffer[id->RollIndex] = ROLL_EOD;
	if ((ROLL_LAST_BYTE(id) - id->RollCount) < id->Offset) {
	    outlen1 = ROLL_LAST_BYTE(id) - id->Offset + 1;
	    offset1 = id->Offset;
	    outlen2 = id->RollCount - outlen1;
	    offset2 = ROLL_FIRST_BYTE(id);
	} else {
	    outlen1 = id->RollCount;
	    offset1 = id->Offset;
	    outlen2 = 0;
	    offset2 = offset1 + outlen1;
	}
	index1 = 0;
	index2 = outlen1;
	id->Offset = offset2 + outlen2 - 1;

	/* write out new offset before writing data */
	if (lseek(id->rfd, OffsetOffset, SEEK_SET) == -1) {
	    perror("(roll_flush) offset write seek failed");
	    return(FAIL);
	}
	if ((i = write(id->rfd, (void *) &id->Offset, sizeof id->Offset))
	    != sizeof id->Offset) {
	    perror("(roll_flush) offset write failed");
	    fprintf(stderr, "             returned %d\n", (int)i);
	    return(FAIL);
	}
    }

    if (id->Verbose == TRUE)
      fprintf(stderr, "outlen1=%d, offset1=%d, index1=%d, outlen2=%d, offset2=%d, index2=%d\n",
	      outlen1, (int)offset1, index1, outlen2, (int)offset2, index2);

    if (lseek(id->rfd, offset1, SEEK_SET) == -1) {
	perror("(roll_flush) 1st seek failed");
	return(FAIL);
    }

    /* 1st write */
    if ((i = write(id->rfd, (void *) &id->RollBuffer[index1], outlen1))
	!= outlen1) {
	perror("(roll_flush) 1st write failed");
	fprintf(stderr, "             returned %d\n", (int)i);
	return(FAIL);
    }

    if (outlen2 > 0) {		  /* wrap around has occurred */
	if (id->Verbose == TRUE)
	  fprintf(stderr, "wrapping\n");
	if (lseek(id->rfd, offset2, SEEK_SET) == -1) {
	    perror("(roll_flush) 2nd seek failed");
	    return(FAIL);
	}

	/* 2nd write */
	if ((i = write(id->rfd, (void *) &id->RollBuffer[index2], outlen2))
	    != outlen2) {
	    perror("(roll_flush) 2nd write failed");
	    fprintf(stderr, "             returned %d\n", (int)i);
	    return(-1);
	}
    }

    if (ROLL_IS_BACK(id)) {
	id->Offset = offset2;

	/* write out new offset after writing data */
	if (lseek(id->rfd, OffsetOffset, SEEK_SET) == -1) {
	    perror("(roll_flush) offset write seek failed");
	    return(FAIL);
	}
	if ((i = write(id->rfd, (void *) &id->Offset, sizeof id->Offset))
	    != sizeof id->Offset) {
	    perror("(roll_flush) offset write failed");
	    fprintf(stderr, "             returned %d\n", (int)i);
	    return(FAIL);
	}
    }

    id->RollCount = 1;		  /* counting EOD */
    if (ROLL_IS_BACK(id))
      id->RollIndex = 1;	  /* EOD goes in front */
    else
      id->RollIndex = 0;	  /* EOD goes in back */

    return(PASS);
}

PASSFAIL roll_out(id)
    RollType *id;
{
    if (roll_flush(id) != PASS) {
	fprintf(stderr, "(roll_out) flush failed\n");
	close(id->rfd);
	free(id);
	return(FAIL);
    }

    if (id->Verbose == TRUE)
      fprintf(stderr, "(roll_out) end=%d, size=%d\n",
	      (int)id->Offset, (int)id->Size );

    close(id->rfd);
    free(id);

    return(PASS);
}

static char *month_name[] = {
  "Jan", "Feb", "Mar", "Apr", "May", "Jun",
  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

static BOOLEAN eat_token(unsigned char** Start, const char* token)
{
    char* p;

    for (p = *Start; *token && *token == *p; ) { token++; p++; }
    if (*token) return FALSE;
    *Start = p;  /* Advance start to point after the token */
    return TRUE; /* Found token */
}

/* this routine will return the "\n\0" term'd line in id->LineBuffer */

char* roll_gets(RollType* id, BOOLEAN cooked)
{
    do {
	if (line_fill(id) == EOF) return NULL;
    } while (id->LineCount == 0);

    if (cooked == TRUE) return cook_line(id, id->LineBuffer);

    return id->LineBuffer;
}

static char* cook_line(RollType* id, const char* s)
{
    static char buf[LINEBUFSIZ + 1];
    cfht_log_type logtype;
    const char* prefix = "";
    if (id->LineTime.tm_mon < 0 || id->LineTime.tm_mon > 11) {
	id->LineStart = NULL;
    }
    
    if (id->LineStart != NULL) {
	if (eat_token(&id->LineStart, CFHT_LOGS_TYPESTRING_START)) {
	    prefix="start";   logtype = CFHT_START;
	} else if (eat_token(&id->LineStart, CFHT_LOGS_TYPESTRING_STATUS)) {
	    prefix="status";  logtype = CFHT_STATUS;
	} else if (eat_token(&id->LineStart, CFHT_LOGS_TYPESTRING_LOGONLY)) {
	    prefix="logonly"; logtype = CFHT_LOGONLY; 
	} else if (eat_token(&id->LineStart, CFHT_LOGS_TYPESTRING_WARN)) {
	    prefix="warning"; logtype = CFHT_WARN;
	} else if (eat_token(&id->LineStart, CFHT_LOGS_TYPESTRING_ERROR)) {
	    prefix="error";   logtype = CFHT_ERROR;
	} else if (eat_token(&id->LineStart, CFHT_LOGS_TYPESTRING_DONE)) {
	    prefix="done"; logtype = CFHT_DONE;
	} else if (eat_token(&id->LineStart, CFHT_LOGS_TYPESTRING_DEBUG)) {
	    prefix="debug";   logtype = CFHT_DEBUG;
	} else if (eat_token(&id->LineStart, CFHT_LOGS_TYPESTRING_TIMING)) {
	    prefix="timing";  logtype = CFHT_TIMING;
	} else id->LineStart = NULL; /* Unusual line.  Don't try to parse */
    }
    
    if (id->LineStart != NULL) {
	char* space = strchr(id->LineBuffer, ' ');

	/*
	 * Warning: This modifies LineBuffer.  After this, it is no
	 * longer possible to return the original LineBuffer.!
	 */
	if (space)
	    *space = '\0'; /* LineBuffer is now just the "who" field */
	else
	    id->LineStart = NULL;
    }
    
    if (id->LineStart != NULL) {
	sprintf(buf, "%.3s %2d %2d:%02d:%02d.%02d %.99s %7s - %.900s",
		month_name[id->LineTime.tm_mon],
		id->LineTime.tm_mday,
		id->LineTime.tm_hour,
		id->LineTime.tm_min,
		id->LineTime.tm_sec,
		id->LineTimeHund,
		id->LineBuffer, /* Now "who" */
		prefix,
		id->LineStart);
	return buf;
    }
    /* Could not be parsed.  Return the original line. */
    return id->LineBuffer;
}

#define DOWSTRLEN 3		  /* string length for the day of the week */
#define ACCTNAMELEN 20		  /* name length for the account */
#define MONTHNAMELEN 3		  /* string length for the month name */
#define TZLEN CFHT_DATE_SIZE      /* string length for the time zone */
#define YEARZERO 1900		  /* starting point for tm_year */
#define MAXMONTHNUM 12		  /* count of months */

/*
 * this routine will fill id->LineBuffer from id->RollBuffer
 * it will also process line for time and session name options
 * id->LineCount will be set to 0 if line should not be printed
 *
 * Time options are handled by looking for (start) lines which are the only
 * lines with dates and times.
 */

static int
line_fill(RollType *id)
{
    unsigned char c;
    struct tm ts;		  /* time structure */
    time_t timein;		  /* time from input file */
 
    char dayofweek[DOWSTRLEN + 1];    /* weekday + NULL */
    char acctname[ACCTNAMELEN + 1];   /* name of account + NULL */
    char monthname[MONTHNAMELEN + 1]; /* name of month + NULL */
    char tz[TZLEN + 1];		      /* time zone + NULL */
    
    (void) memset((void *) &ts, 0, sizeof ts); /* zero out variable */

    if (!id->LineTimeValid) {
	id->LineTimeHund = 0;
	if (ROLL_IS_BACK(id)) timein = time(0);
	else timein = 0;
	id->LineTime = *localtime(&timein);
	id->LineTime.tm_wday = 0;
	id->LineTime.tm_yday = 0;
	id->LineTimeValid = TRUE;
    }

   /* printf("%d) ", ++linecounter); */ /* debug printf to identify lines */
    
    id->LineCount = 0;
    
    do { /* get data from buffer until '\n' */
       
        if (id->RollIndex >= id->RollCount && roll_fill(id) == EOF) {
            return EOF;
	}

	if ((c = id->RollBuffer[id->RollIndex++]) != '\0'){
	    if (id->LineCount < (LINEBUFSIZ - 1 - 1)) {
		id->LineBuffer[id->LineCount++] = c;
	    } /* save character if within buffer size */
	} /* and if it is not a null */
	
    } while (c != '\n');

    /* add a new line char if line was cut off */
    if (id->LineBuffer[id->LineCount - 1] != '\n') {
	id->LineBuffer[id->LineCount++] = '\n';
    }
    id->LineBuffer[id->LineCount] = '\0'; /* for sscanf and strchr */

    /* See if there is a comma (',') in the message */
    id->LineStart = strchr(id->LineBuffer, ',');
    
    /* split line into fields to see if (start) line with date and time */
    /* sscanf croaks on random garbage strings so protect a bit */
    if (id->LineStart
	&& (isalpha(id->LineBuffer[0]))
	&& (isalpha(id->LineBuffer[1]))
	&& (strstr((char *) id->LineBuffer, " (start) - ") != (char *) NULL)
	&& ((sscanf((char *) id->LineBuffer,
		    "%s %s %s %d %d:%d:%d %s %d, (start)",
		    acctname, dayofweek, monthname, &ts.tm_mday, &ts.tm_hour,
		    &ts.tm_min, &ts.tm_sec, tz, &ts.tm_year)) == 9)
	&& (ts.tm_mon = month_num(monthname)) >= 0) {
	ts.tm_year -= YEARZERO;	  /* note: = actual year - 1900 */
	ts.tm_wday = 0;		  /* set unused values to zero */
	ts.tm_yday = 0;
	if (strstr(tz, "HST"))
	    ts.tm_isdst = 0; /* do not even try to use DST in Hawaii! */
	else
	    ts.tm_isdst = -1; /* let mktime() attempt to figure it out */

	if (id->LineTime.tm_sec != ts.tm_sec ||
	    id->LineTime.tm_min != ts.tm_min ||
	    id->LineTime.tm_hour != ts.tm_hour) {
	    if (ROLL_IS_BACK(id))
	        id->LineTimeHund = 99;
	    else
	        id->LineTimeHund = 0;
	}
	id->LineTime = ts;
    } else if (isalpha(id->LineBuffer[0])
	       && isalpha(id->LineBuffer[1])
	       && isalpha(id->LineBuffer[2])
	       && id->LineBuffer[3] == ' '
	       && sscanf((char *) id->LineBuffer,
			 "%s %d %d:%d:%d.%d ",
			 monthname,
			 &ts.tm_mday,
			 &ts.tm_hour,
			 &ts.tm_min,
			 &ts.tm_sec,
			 &id->LineTimeHund) == 5
	       && (ts.tm_mon = month_num(monthname)) >= 0) {
	id->LineTime.tm_sec = ts.tm_sec;
	id->LineTime.tm_min = ts.tm_min;
	id->LineTime.tm_hour = ts.tm_hour;
	id->LineTime.tm_mday = ts.tm_mday;
	id->LineTime.tm_mon = ts.tm_mon;
	id->LineStart = NULL;
    } else if (sscanf((char *) id->LineBuffer,
		      "%s %d:%d:%d.%d, ", acctname,
		      &ts.tm_hour, &ts.tm_min, &ts.tm_sec, &id->LineTimeHund) == 5) {
	int day_change = 0;
	if (ROLL_IS_BACK(id)) {
	    if (ts.tm_hour > id->LineTime.tm_hour) day_change = -1;
	} else {
	    if (ts.tm_hour < id->LineTime.tm_hour) day_change = 1;
	}
	id->LineTime.tm_sec = ts.tm_sec;
	id->LineTime.tm_min = ts.tm_min;
	id->LineTime.tm_hour = ts.tm_hour;
	if (day_change != 0) {
	    timein = mktime(&id->LineTime) + 60 * 60 * 24 * day_change;
	    id->LineTime = *localtime(&timein);
	    id->LineTime.tm_wday = 0;
	    id->LineTime.tm_yday = 0;
	}
    } else {
      /* warning: line without timestamp in the log! */
      id->LineStart = NULL; /* There might have been a comma, but
			     * it isn't one delimiting the timestamp
			     */
    }

    timein = mktime(&id->LineTime); /* convert to arithmetic time */

    /* if LineTime is out of "to" and "from" range, ignore lines */
    if ((timein < id->From) || (timein > id->To)) {
	id->LineCount = 0;
	/* if we're past limit, quit completely for speed */
	if (ROLL_IS_BACK(id)) {
	    if (timein < id->From) {
		return(EOF);
	    }
	} else {
	    if (timein > id->To) {
		return(EOF);
	    }
	}
    }

    /* ignore lines from other sessions if so desired */
    if (id->Session != (char *) NULL) {
	if (sscanf((char *) id->LineBuffer, "%s", acctname) == 1) {
	    if (strcmp(id->Session, acctname) != 0) {
		id->LineCount = 0;
	    }
	}
    }
    
    return(~EOF); 
}

/* month_num: return number of month from its abbreviated name */
/* note: order is from 0-11 not 1-12, returns -1 if error */

static int
month_num(char* s)
{   
    int i;
     
    /* return index for the string */
    for (i = 0; i < MAXMONTHNUM; i++) {
	if (strcmp(s, month_name[i]) == 0) {
	    return i; 
	}
    }

    return(-1);			  /* error */
}

/*
 * this routine will fill id->RollBuffer from id->rfd
 * if needed, file will be rewound to first data byte
 * set id->RollIndex to 0, id->RollCount to # in rollbuf
 * returns EOF when EOD mark reached after wrapping around or on error,
 *        ~EOF otherwise
 * NULLs may be present
 */

static
int roll_fill(RollType *id)
{
    char* eod;

    if (id->Size == 0) {
	/* we have been around */
	return(EOF);
    }

    id->RollCount = read(id->rfd, (void *) id->RollBuffer, sizeof id->RollBuffer);
    if (id->RollCount == 0) {
	if (lseek(id->rfd, ROLL_FIRST_BYTE(id), SEEK_SET) == -1) {
	    perror("(roll_fill) seek to front of file failed");
	    return(EOF);
	}
	id->RollCount = read(id->rfd, (void *) id->RollBuffer,
			     sizeof id->RollBuffer);
	if (id->RollCount == 0) {
	    fprintf(stderr, "(roll_fill) read at front returned 0?\n");
	    return(EOF);
	} else if (id->RollCount < 0) {
	    perror("(roll_fill) wrapping read at front failed");
	    return(EOF);
	}
    } else if (id->RollCount < 0) {
	perror("(roll_fill) read failed");
	return(EOF);
    }
    
    eod = memchr(id->RollBuffer, ROLL_EOD, id->RollCount);
    if (eod != NULL) {
	id->RollCount = eod - (char*)id->RollBuffer;
	id->Size = 0;	  /* use as flag when wrapped */
	if (id->RollCount == 0)
	    return EOF;	  /* EOD was first thing */
    }

    id->RollIndex = 0;

    return(~EOF);
}


/* signal handler */
static void
cleanup(int sig)
{
    fprintf(stderr, "(roll) signal %d happend - cleanup\n", sig);
    fflush(stderr);
    roll_out(RollFilePointer);
    cfht_signal(sig, CFHT_SIG_DFL);
    kill(getpid(), sig);
    exit(EXIT_SUCCESS);	  /* just in case ? */
}
